%{
/*                               -*- Mode: C -*- 
 * field_l.l -- 
 * ITIID           : $ITI$ $Header $__Header$
 * Author          : Ulrich Pfeifer
 * Created On      : Tue Feb 15 16:03:22 1994
 * Last Modified By: Ulrich Pfeifer
 * Last Modified On: Thu Dec 22 14:56:14 1994
 * Update Count    : 225
 * Status          : Unknown, Use with caution!
 */


/* HISTORY 
 * 10-Mar-1994		Ulrich Pfeifer	
 *    added support for backslash escaped '/'.
 */

#include "field_token.h"
#include "cutil.h"
#ifdef FLEX_SCANNER
int yylineno = 0;
#define UPDATELINENO if (yytext[0]=='\n') yylineno++;
#undef yywrap
#define yywrap() (1)
#else
#define UPDATELINENO
#endif


#undef  ECHO
#define ECHO 
%}
WORD  [^ \t\n=()/]+
%%
"/"[^/]*"/" { ECHO; 
	      if (yytext[yyleng-2] == '\\') { 
                /* the ending '/' was escaped, lets get the rest */
                unput(yytext[yyleng-1]);/* yes yyless() is better */ 
                yymore(); 
	      } else {
                /* Flex bug?  When there is more than one escape, the
                   call to unput adds two chars to the from of yytext.
                   We now skip them :-(
                   */
                char *s = yytext;
                while (*s && *s != '/') s++;
                yylval.sval = strdup(s);

		{ char *s = yylval.sval, *t = yylval.sval;
		  while (*s) { 
		    if (*s == '\\') {
		      if (s[1] == '/') {
			s+=1; /* escaped '/' */
		      } else if (s[1] == '\\') { 
			s+=1; /* double \\ */
		      } else if (s[1] >= 'A' && s[1] <= 'Z') { 
			s[1] = s[1]-'A'+1;  /* map of ASCII character set */
			s+=1;
		      }	else if (s[1] == 'n') {
			s[1] = '\n'; /* newline */
			s+=1;
		      }	else if (s[1] == 'r') {
			s[1] = '\r'; /* cr */
			s+=1;
		      }	else if (s[1] == 'f') {
			s[1] = '\f'; /* form feed */
			s+=1;
		      } else {       /* let the escaped char unmodified */
			s+=1;
		      }
		    }
		    *(t++) = *(s++);
		  }
		  *t = *s; /* should be '\0' */
		  /* fprintf(stderr,"\nregexp:%s:\n", yylval.sval);  */
		  return(REGEXP); 
		}
	      }
	    }
[0-9]+            { ECHO; yylval.ival = atoi(yytext); return(INT); }
"<field>"         { ECHO; return(FIELD); }
"<end>"           { ECHO; return(END); }
"<numeric>"       { ECHO; return(NUMERIC); }
"<headline>"      { ECHO; return(HEADLINE); }
"<layout>"        { ECHO; return(LAYOUT); }
"stemming"        { ECHO; return(STEMMING); }
[ \t\n]           { ECHO; UPDATELINENO; }
"SOUNDEX"         { /* tung: it is very bad style to use the strings here! use a enum (up) */
                    ECHO; yylval.sval = strdup(yytext); return(ISOUNDEX); }
"PHONIX"          { ECHO; yylval.sval = strdup(yytext); return(IPHONIX); }
"TEXT"            { ECHO; yylval.sval = strdup(yytext); return(ITEXT); }
"LOCAL"           { ECHO; return(LOCAL); }
"IGNORE"          { ECHO; return(IGNORE); }
"GLOBAL"          { ECHO; return(GLOBAL); }
"BOTH"            { ECHO; return(BOTH); }
"<record-end>"    { ECHO; return(RECORD_END); }
"<date>"          { ECHO; /* yylval.sval = strdup(yytext); (up) */ return(DATE); }
"day"             { ECHO; yylval.sval = "day"; return(DAY); }
"month"           { ECHO; yylval.sval = "month"; return(MONTH); }
"year"            { ECHO; yylval.sval = "year"; return(YEAR); }
"string"          { ECHO; yylval.sval = "string";  return(STRING); }
\"[^"]*\"         { ECHO; 
                    yylval.sval = strdup(yytext+1); 
                    yylval.sval[strlen(yylval.sval)-1] = '\0'; 
                    return(DESCRIPTION); }
{WORD}            { ECHO; yylval.sval = strdup(yytext); return(WORD); }
