/* (C) Copyright 1993 by John G. Myers
 * All Rights Reserved.
 *
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without
 * fee, provided that the above copyright notice appear in all copies
 * and that both that copyright notice and this permission notice
 * appear in supporting documentation, and that the name of John G.
 * Myers not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior
 * permission.  John G. Myers makes no representations about the
 * suitability of this software for any purpose.  It is provided "as
 * is" without express or implied warranty.
 *
 * JOHN G. MYERS DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS, IN NO EVENT SHALL JOHN G. MYERS BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
 * OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#include <stdio.h>
#include <dir.h>
#include <errno.h>
#include "version.h"

extern int optind;
extern char *optarg;

extern int overwrite_files;

main(argc, argv)
int argc;
char **argv;
{
    int opt;
    FILE *file;
    char *p, *q, *path = 0;
    int done, gotone;
    struct ffblk ffblk;
    
    while ((opt = getopt(argc, argv, "fC:")) != EOF) {
	switch (opt) {
	case 'f':
	    overwrite_files = 1;
	    break;

	case 'C':
	    if (chdir(optarg)) {
		perror(optarg);
		exit(1);
	    }
	    break;

	default:
	    usage();
	}
    }

    if (optind == argc) {
	fprintf(stderr, "munpack: reading from standard input\n");
	handleMessage(stdin, "text/plain", 0);
	exit(0);
    }

    while (argv[optind]) {
	if (path) free(path);
	path = 0;

	p = argv[optind];
	if (p[1] == ':') p += 2;
	if (q = strrchr(p, '\\')) p = q+1;
	if (q = strrchr(p, '/')) p = q+1;

	if (p != argv[optind]) {
	    path = xmalloc(strlen(argv[optind])+20);
	    strcpy(path, argv[optind]);
	    p = path + (p - argv[optind]);
	}

	gotone = 0;
	done = findfirst(argv[optind], &ffblk, 0);
	while (!done) {
	    if (path) strcpy(p, ffblk.ff_name);
	    file = fopen(path ? path : ffblk.ff_name, "r");
	    if (!file) {
		perror(path ? path : ffblk.ff_name);
	    }
	    else {
		handleMessage(file, "text/plain", 0);
		fclose(file);
	    }
	    gotone = 1;
	    done = findnext(&ffblk);
	}
	if (!gotone) {
	    perror(argv[optind]);
	}
	optind++;
    }
    exit(0);
}

usage() {
    fprintf(stderr, "usage: munpack [-C directory] [files...]\n");
    exit(1);
}

warn(s)
char *s;
{
    fprintf(stderr, "munpack version %s\n", MPACK_VERSION);
    fprintf(stderr, "munpack: warning: %s\n", s);
}
