/*
 * tmConverters.c -
 *	This module implements all the Xt Converters to and
 *	from Strings needed by Tm
 *
 * Copyright 1993 Jan Newmarch, University of Canberra.
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appear in all copies.  The author
 * makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without
 * express or implied warranty.
 */

/* Copyright (C) 1992 by Gustaf Neumann, Stefan Nusser

 *      Wirtschaftsuniversitaet Wien, 
 *      Abteilung fuer Wirtschaftsinformatik
 *      Augasse 2-6, 
 *      A-1090 Vienna, Austria
 *      neumannM(wu)-wien.ac.at, nusserM(wu)-wien.ac.at

 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appears in all copies and that both that
 * copyright notice and this permission notice appear in all supporting
 * documentation.  This software is provided "as is" without expressed or
 * implied warranty.
 */

#include "tmFuncs.h"
#ifndef MOTIF11
#include <Xm/RepType.h>
#include <Xm/DragDrop.h>
#endif
#include <X11/IntrinsicP.h>
#include <X11/CoreP.h>		/* for CoreRec */
#include <string.h>
#include <ctype.h>

/* escape char for XmString formatting */
#define TM_FORMAT_CHAR '@'

#define done(address, type)                               \
    {                                                   \
        if(toVal->addr)                                 \
        {                                               \
            if(toVal->size < sizeof(type))              \
            {                                           \
                toVal->size = sizeof(type);             \
                return( FALSE) ;                        \
            }                                           \
            else                                        \
            {   *((type *) (toVal->addr)) = (type) (*address);   \
            }                                           \
        }                                               \
        else                                            \
        {                                               \
            toVal->addr = (XPointer) address;           \
        }                                               \
        toVal->size = sizeof(type);                     \
    }


/*
 * so that we can keep tabs on the size of an XmStringTable
 * see function Tm_MaybeSetXmStringTableSize
 */
int Tm_XmStringTableSize;

/*
 *--------------------------------------------------------------
 *
 * InstallReverseRepTypes --
 *
 *	Install the converters from internal representation
 *	types to their string values.
 *
 * Results:
 *
 * Side effects:
 *	Modifies the Motif RepType tables
 *
 *--------------------------------------------------------------
 */

static void
InstallReverseRepTypes()
{
#ifndef MOTIF11
    XmRepTypeId id;
    XmRepTypeList start, list;

    start = list = XmRepTypeGetRegistered();

    while (list->rep_type_name != NULL) {
        if ((id = XmRepTypeGetId(list->rep_type_name)) != XmREP_TYPE_INVALID) {
	    XmRepTypeAddReverse(id);
        } else {
	    fprintf(stderr, "Failed to install %s converter\n", 
			list->rep_type_name);
        }
	list++;
    }
    XtFree((char *)start);
#endif
}


/*
 *--------------------------------------------------------------
 *
 * Tm_CvtStringToWidget --
 *
 *	Converter from String to Widget type
 *
 * Results:
 *
 *	changes the "toVal" arg to contain the new value.
 *
 * Side effects:
 *
 *	none
 *--------------------------------------------------------------
 */

Boolean
Tm_CvtStringToWidget(display, args, num_args, fromVal, toVal, destructor_data)
    Display *display;
    XrmValuePtr args;
    Cardinal *num_args;
    XrmValuePtr fromVal;
    XrmValuePtr toVal;
    XtPointer *destructor_data;
{
    Tcl_Interp *interp = (Tcl_Interp *) args->addr;
    Tm_Widget *info;
    static Widget w;

    info = Tm_WidgetInfoFromPath(interp, (char *) fromVal->addr);
    if (info == NULL) {
	return False;
    }
    w = info->widget;

    done(&w, Widget);

    return True;
}

/*
 *--------------------------------------------------------------
 *
 * Tm_CvtWidgetToString --
 *
 *	Converter from Widget to String type
 *
 * Results:
 *
 *	changes the "toVal" arg to contain the new value.
 *
 * Side effects:
 *
 *	none
 *--------------------------------------------------------------
 */

Boolean
Tm_CvtWidgetToString(display, args, num_args, fromVal, toVal, destructor_data)
    Display *display;
    XrmValuePtr args;
    Cardinal *num_args;
    XrmValuePtr fromVal;
    XrmValuePtr toVal;
    XtPointer *destructor_data;
{
    Tm_Widget *info;
    Widget w = *(Widget *) fromVal->addr;
    static char *buf;
    extern char *Tm_Resource; 

    /* double hack going on here. because of Motif
       bug, some Widget resources are of type Window.
       When we *really* get a window to convert,
       just fail :-((
     */
    if (strcmp(Tm_Resource, "windowGroup") == 0)
	return False;

    /* it is valid to ask for a NULL widget e.g. topWidget in a Form
       when the thing is attached to the Form (default is NULL)
     */
    if (w == NULL) {
        buf = "";
        done(&buf, String);
	return True;
    }
    info = NULL;
    XtVaGetValues(w, XmNuserData, &info, NULL);
    if (info == NULL) {
	return False;
    }

    buf = info->pathName;
    done(&buf, String);

    return True;
}


/*
 *--------------------------------------------------------------
 *
 * Tm_CvtWidgetListToString --
 *
 *	Converter from WidgetList to String type
 *
 * Results:
 *
 *	changes the "toVal" arg to contain the new value.
 *
 * Side effects:
 *
 *	none
 *--------------------------------------------------------------
 */

Boolean
Tm_CvtWidgetListToString(display, args, num_args, fromVal, toVal, destructor_data)
    Display *display;
    XrmValuePtr args;
    Cardinal *num_args;
    XrmValuePtr fromVal;
    XrmValuePtr toVal;
    XtPointer *destructor_data;
{
    Tm_Widget *info;
    Widget w = *(Widget *) args[0].addr;
    WidgetList children = *(WidgetList *) fromVal->addr;
    Cardinal num_children, final_num_children;
    String *widgetNames;
    int n, len;
    char *end, *p;
    static char *buf = NULL;
    int buf_size;
    Widget child;

    buf_size = 32;
    buf = XtRealloc(buf, buf_size);

    if ( !XtIsComposite(w)) {
	buf[0] = '\0';
        done(&buf, char *);
        return True;
    }

    XtVaGetValues(w, 
	XmNnumChildren, &num_children,
	XmNchildren, &children, NULL);
    if (num_children == 0) {
	buf[0] = '\0';
        done(&buf, char *);
        return True;
    }

    widgetNames = (String *) XtMalloc(num_children * sizeof(String));
    len = 0;
    final_num_children = 0;
    for (n = 0; n < num_children; n++) {
	child = children[n];
	info = NULL;
        XtVaGetValues(child, XmNuserData, &info, NULL);
        if (info == NULL) {
	    continue;
	}

	/* if the widget is being destroyed, we don't want to know it */
	if (child->core.being_destroyed) {
	    continue;
	}
	
	widgetNames[final_num_children] = info->pathName;
	len += strlen(info->pathName) + 2; /* for ", " */
	final_num_children++;
    }

    /* grow if needed */
    if (buf_size < len + 1) {
	do {
	    buf_size *= 2;
	}   while (buf_size < len + 1);
	buf = XtRealloc(buf, buf_size);
    }

    if (len == 0) {
        *buf = '\0';
        done(&buf, char *);
        return True;
    }
    end = buf;
    for (n = 0; n < final_num_children; n++) {
        p = widgetNames[n];
        while ((*end++ = *p++))
	    ;
        *(end - 1) = ',';
        *end++ = ' ';
    }
    *(end - 2) = '\0';

    done(&buf, char *);
    XtFree((char *) widgetNames);

    return True;
}

/*
 *--------------------------------------------------------------
 *
 * Tm_CvtStringToTransferStatus --
 *
 *	Converter from String to TransferStatus type
 *	(see transferProc widget0
 *
 * Results:
 *
 *	changes the "toVal" arg to contain the new value.
 *
 * Side effects:
 *
 *	none
 *--------------------------------------------------------------
 */

Boolean
Tm_CvtStringToTransferStatus(display, args, num_args, fromVal, toVal, destructor_data)
    Display *display;
    XrmValuePtr args;
    Cardinal *num_args;
    XrmValuePtr fromVal;
    XrmValuePtr toVal;
    XtPointer *destructor_data;
{
    char *buf;
    int n;
    static unsigned char status;

    buf = (char *) (fromVal->addr);
    for (n = 0; n < strlen(buf); n++) {
	buf[n] = tolower(buf[n]);
    }
#   ifdef DEBUG
    fprintf(stderr, "converted string to %s\n", buf);
#   endif
    if (strcmp(buf, "transfer_success") == 0) {
	status = 1;
    } else {
	status = 0;
    }

    done(&status, unsigned char);

    return True;
}


/*
 *--------------------------------------------------------------
 *
 * Tm_CvtStringToDropTransfers--
 *
 *	Converter from String to DropTransfers type
 *	(see transferProc widget0
 *
 * Results:
 *
 *	changes the "toVal" arg to contain the new value.
 *
 * Side effects:
 *
 *	none
 *--------------------------------------------------------------
 */

Boolean
Tm_CvtStringToDropTransfers(display, args, num_args, fromVal, toVal, destructor_data)
    Display *display;
    XrmValuePtr args;
    Cardinal *num_args;
    XrmValuePtr fromVal;
    XrmValuePtr toVal;
    XtPointer *destructor_data;
{
#ifndef MOTIF11
    char *buf;
    char **argv, **argvv;
    int argc, argcc;
    int n;
    static XmDropTransferEntryRec transferEntries[32];
    static XmDropTransferEntryRec *p;
    Tcl_Interp *interp = (Tcl_Interp *) args->addr;

    buf = (char *) (fromVal->addr);
#   ifdef DEBUG
    fprintf(stderr, "converted string to %s\n", buf);
#   endif
    if (Tcl_SplitList(interp, buf, &argc, &argv) == TCL_ERROR) {
	fprintf(stderr, "failed to convert list\n");
    }
    for (n = 0; n < argc; n++) {
	if (Tcl_SplitList(interp, argv[n], &argcc, &argvv) == TCL_ERROR) {
	    fprintf(stderr, "failed to convert internal list\n");
	}
	if (argcc != 2) {
	    fprintf(stderr, "list should have two elements\n");
	}
	transferEntries[n].target = XInternAtom(display, argvv[0], False);
/*
	wPtr = Tm_WidgetInfoFromPath(interp, argvv[1]);
	transferEntries[n].client_data = wPtr->widget;
*/
	transferEntries[n].client_data = argvv[1];
    }

    p = transferEntries;
    done(&p, XtPointer);

#endif
    return True;
}


/*
 *--------------------------------------------------------------
 *
 * Tm_XmStringToString --
 *
 * Results:
 *
 *	returns the Ascii string rep of the XmString
 *
 * Side effects:
 *
 *	Ascii string in malloced space
 *--------------------------------------------------------------
 */
String
Tm_XmStringToString(xmstr)
    XmString xmstr;
{
    XmStringContext context;
    static char *buf = NULL;
    char *p, *text = NULL;
    XmStringCharSet charset = (XmStringCharSet)NULL;
    XmStringDirection direction;
    int size;
    int current_len, segment_len;
    XmStringComponentType tag;
    XmStringComponentType u_tag;
    unsigned short u_length;
    unsigned char *u_value = NULL;

    /* whatever buf has grown to, or started from, 
       make it a reasonable size
     */
    XtFree(buf);
    size = 128;
    buf = XtMalloc(size);
    p = buf;

    if (!XmStringInitContext(&context, xmstr)) {
	return "";
    }

#if 1
    while ((tag = XmStringGetNextComponent(context, &text, &charset,
		&direction, &u_tag, &u_length, &u_value))
	    != XmSTRING_COMPONENT_END) {
	switch (tag) {
	    case XmSTRING_COMPONENT_CHARSET:
	    case XmSTRING_COMPONENT_DIRECTION:
	    case XmSTRING_COMPONENT_UNKNOWN:
		/* should probably free stuff here - but what??? */
		break;
	    case XmSTRING_COMPONENT_LOCALE_TEXT:
	    case XmSTRING_COMPONENT_TEXT:
	        segment_len = strlen(text);
	        while (p + segment_len >= buf + size) {
	            /* string too large, double space */
	            current_len = p - buf;
	            size *= 2;
	            buf = XtRealloc(buf, size);
	            p = buf + current_len;
	        }
	        strcpy(p, text);
	        p += segment_len;
		XtFree(text);
		text = NULL;
		break;

	    case XmSTRING_COMPONENT_SEPARATOR:
               if (p + 1 >= buf + size) {
                    /* string too large, double space */
                    current_len = p - buf;
                    size *= 2;
                    buf = XtRealloc(buf, size);
                    p = buf + current_len;
                }
		*p++ = '\n';
		*p = '\0';
		break;
	    default:
		break;
	}
	if (text)
	    XtFree(text);
	if (charset)
	    XtFree(charset);
	if (u_value)
	    XtFree(u_value);
	text = NULL;
	charset = (XmStringCharSet)NULL;
	u_value = NULL;
    }
    if (text)
	XtFree(text);
    if (charset)
	XtFree(charset);
    if (u_value)
	XtFree(u_value);
    text = NULL;
    charset = (XmStringCharSet)NULL;
    u_value = NULL;

#else /* 1 */
    while (XmStringGetNextSegment(context, &text, &charset,
		&direction, &separator)) {
	XtFree(charset);
	segment_len = strlen(text);
	while (p + segment_len >= buf + size) {
	    /* string too large, double space */
	    current_len = p - buf;
	    size *= 2;
	    buf = XtRealloc(buf, size);
	    p = buf + current_len;
	}
	strcpy(p, text);
	p += segment_len;
	if (separator == True) {
	    *p++ = '\n';
	}
	*p = 0;
	XtFree(text);
    }
#endif /* 1 */
    XmStringFreeContext(context);

    return buf;
}

/*
 *--------------------------------------------------------------
 *
 * Tm_CvtXmStringToString --
 *
 *	Converter from XmString to String type
 *
 * Results:
 *
 *	changes the "toVal" arg to contain the new value.
 *
 * Side effects:
 *
 *	destroys the XmString
 *--------------------------------------------------------------
 */

Boolean
Tm_CvtXmStringToString(display, args, num_args, fromVal, toVal, destructor_data)
    Display *display;
    XrmValuePtr args;
    Cardinal *num_args;
    XrmValuePtr fromVal;
    XrmValuePtr toVal;
    XtPointer *destructor_data;
{
    static char *buf = NULL;
    XmString xmstr;
#if 0
    XmStringContext context;
    char *p, *text;
    XmStringCharSet charset;
    XmStringDirection direction;
    Boolean separator;
    int size;
    int current_len, segment_len;
#endif /* 0 */

    xmstr = *(XmString *) fromVal->addr;
    buf = Tm_XmStringToString(xmstr);

#if 0
    /* whatever buf has grown to, or started from, 
       make it a reasonable size
     */
    XtFree(buf);
    size = 128;
    buf = XtMalloc(size);
    p = buf;

    if (!XmStringInitContext(&context, xmstr)) {
	return False;
    }
    while (XmStringGetNextSegment(context, &text, &charset,
		&direction, &separator)) {
	XtFree(charset);
	segment_len = strlen(text);
	if (p + segment_len >= buf + size) {
	    /* string too large, double space */
	    current_len = p - buf;
	    size *= 2;
	    buf = XtRealloc(buf, size);
	    p = buf + current_len;
	}
	strcpy(p, text);
	p += segment_len;
	if (separator == True) {
	    *p++ = '\n';
	}
	*p = 0;
	XtFree(text);
    }
    XmStringFreeContext(context);
#endif /* 0 */

    /* destroy the XmString we started from */
    XmStringFree(xmstr);
 
    done(&buf, char *);

    return True;
}

/*
 *--------------------------------------------------------------
 *
 * Tm_CvtXmStringTableToString --
 *
 *	Converter from XmStringTable to String type
 *
 * Results:
 *
 *	changes the "toVal" arg to contain the new value.
 *
 * Side effects:
 *
 *	none
 *--------------------------------------------------------------
 */

Boolean
Tm_CvtXmStringTableToString(display, args, num_args, fromVal, toVal, destructor_data)
    Display *display;
    XrmValuePtr args;
    Cardinal *num_args;
    XrmValuePtr fromVal;
    XrmValuePtr toVal;
    XtPointer *destructor_data;
{
    String *items;
    static String buf = NULL;
    int n, len, item_count;
    XmString *xmstrs = *(XmString **) fromVal->addr;
    Arg arg;
    Widget w = *((Widget *) args[0].addr);
    char *size_resource;
    /* this should come in via args, but Xt doesn't give it */
    extern char *Tm_Resource;

    if (Tm_Resource== NULL)
	return False;
    if (strcmp (Tm_Resource, XmNitems) == 0) {
        size_resource = XmNitemCount;
    } else
    if (strcmp (Tm_Resource, XmNselectedItems) == 0) {
        size_resource = XmNselectedItemCount;
    } else
    return False;

    XtSetArg(arg, size_resource, &item_count);
    XtGetValues(w, &arg, 1);
    Tm_Resource = NULL; /* make sure under our control */

    len= 0;
    items = NULL;

    if (item_count != 0) {
        items = (String *) XtMalloc(item_count * sizeof(String));
        for (n = 0; n < item_count; n++) {
	    items[n] = Tm_XmStringToString(xmstrs[n]);
	    items[n] = XtNewString(items[n]);
            len += strlen(items[n]) + 2; /* for , */
        }
    }
#ifdef TM_MOTIF_XMSTRINGTABLE
    buf = XtRealloc(buf, len + 1); /* we overshoot by 1 */
    if (len == 0) {
        *buf = '\0';
        done(&buf, char *);
        return True;
    }
    end = buf;
    for (n = 0; n < item_count; n++) {
        p = items[n];
        while (*end++ = *p++)
    	;
        *(end - 1) = ',';
        *end++ = ' ';
        XtFree(items[n]);
    }
    *(end - 2) = '\0';
#else
    if (buf != NULL)
	free(buf);
    buf = Tcl_Merge(item_count, items);
    for (n = 0; n < item_count; n++) {
        XtFree(items[n]);
    }
#endif

    done(&buf, char *);

    if (items != NULL) {
	XtFree((char *) items);
    }

    return True;
}


/*
 *--------------------------------------------------------------
 *
 * Tm_CvtStringTableToString --
 *
 *	Converter from StringTable to String type
 *
 * Results:
 *
 *	changes the "toVal" arg to contain the new value.
 *
 * Side effects:
 *
 *	none
 *--------------------------------------------------------------
 */

Boolean
Tm_CvtStringTableToString(display, args, num_args, fromVal, toVal, destructor_data)
    Display *display;
    XrmValuePtr args;
    Cardinal *num_args;
    XrmValuePtr fromVal;
    XrmValuePtr toVal;
    XtPointer *destructor_data;
{
    static String buf = NULL;
    int item_count;
    String *strs = *(String **) fromVal->addr;
    Arg arg;
    Widget w = *((Widget *) args[0].addr);
    char *size_resource;
    /* this should come in via args, but Xt doesn't give it */
    extern char *Tm_Resource;

    if (Tm_Resource== NULL)
	return False;
    if (strcmp (Tm_Resource, XmNargv) == 0) {
        size_resource = XmNargc;
    } else
    return False;

    XtSetArg(arg, size_resource, &item_count);
    XtGetValues(w, &arg, 1);
    Tm_Resource = NULL; /* make sure under our control */

    if (buf != NULL)
	free(buf);
    buf = Tcl_Merge(item_count, strs);

    done(&buf, char *);

    return True;
}


/*
 *--------------------------------------------------------------
 *
 * Tm_XmStringConcatAndFree --
 *	concat two XmStrings and free them
 *
 * Result
 *	the concatenated XmString
 *
 *--------------------------------------------------------------
 */

static XmString
Tm_XmStringConcatAndFree(front, back)
    XmString front;
    XmString back;
{
    XmString result;

    if (front == NULL) {
	return back;
    }

    result = XmStringConcat(front, back);
    XmStringFree(front);
    XmStringFree(back);

    return result;
}

/*
 *--------------------------------------------------------------
 *
 * Tm_SimpleStringToXmString --
 *
 *	convert a simple piece of text to an XmString
 *	This converts internal \n to separators
 *
 *--------------------------------------------------------------
 */
static XmString
Tm_SimpleStringToXmString(text, charSet, dir, sep)
    char              *text;
    XmStringCharSet    charSet;
    XmStringDirection  dir;
    Boolean            sep;
{
    char *base, *p;
    XmString xmseg;
    XmString xmstr = XmStringCreate("", charSet);

    p = base = text;

    /* we cant use strtok to parse the string, because it will discard
     * leading \ns
     */
    while (*p) {
	if (*p == '\n') {
	    *p = '\0';
	    if (*base == '\0') {
		xmseg = XmStringSeparatorCreate();
	    } else {
	        xmseg = XmStringSegmentCreate(base, charSet, dir, True);
	    }
	    xmstr = Tm_XmStringConcatAndFree(xmstr, xmseg);
	    base = p + 1;
	}
	++p;
    }

    /* handle part of string after last \n */
    xmseg = XmStringSegmentCreate(base, charSet, dir, sep);
    xmstr = Tm_XmStringConcatAndFree(xmstr, xmseg);

    return xmstr;
}

/*
 *--------------------------------------------------------------
 *
 * Tm_XmStringAppendString --
 *
 *	append a piece of simple text to an XmString
 *	This obeys the semantics that embedded newlines
 *	become separators in the XmString
 *
 * Result
 *	the appended XmString
 *
 *--------------------------------------------------------------
 */

static XmString
Tm_XmStringAppendString(oldCompStr, text, charSet, dir, sep)
    XmString           oldCompStr;
    char              *text;
    XmStringCharSet    charSet;
    XmStringDirection  dir;
    Boolean            sep;
{
    XmString  addSeg, newString;
    
    if (*text == '\0' && ! sep)
	return(oldCompStr);

    
    if (oldCompStr == NULL) {
	return(Tm_SimpleStringToXmString(text, charSet, dir, sep));
    } else {
	addSeg = Tm_SimpleStringToXmString(text, charSet, dir, sep);
	newString = XmStringConcat(oldCompStr, addSeg);
	XmStringFree(addSeg);
	XmStringFree(oldCompStr);
	return(newString);
    }

#if 0
    if (oldCompStr == NULL) {
	return(XmStringSegmentCreate(text, charSet, dir, sep));
    } else {
	addSeg = XmStringSegmentCreate(text, charSet, dir, sep);
	newString = XmStringConcat(oldCompStr, addSeg);
	XmStringFree(addSeg);
	XmStringFree(oldCompStr);
	return(newString);
    }
#endif /* 0 */
}

/*
 *--------------------------------------------------------------
 *
 * Tm_StringToXmString --
 *
 *	Convert a string to an XmString. The string is a tcl list
 *	that may contain formatting control characters
 *
 * Result
 *	the converted XmString
 *
 *--------------------------------------------------------------
 */

#if XmVERSION < 2
XmString    
Tm_StringToXmString(interp, inString)
    char  *inString;
    Tcl_Interp *interp;
{
    char               *segBuf;
    XmStringCharSet    charSet = XmFONTLIST_DEFAULT_TAG; 
    XmStringDirection  dir = XmSTRING_DIRECTION_L_TO_R;
    XmString           result = XmStringCreate("", charSet);
    XmString	       xm_dir;
    char              *segPtr;
    char              *ptr;
    int		      argc;
    char	      **argv;
    int		      n;
    Boolean	      segment_started = True;

    segPtr = segBuf =  XtMalloc(strlen(inString) + 1);

    /* In Motif 2.0 the segment stuff will vanish, leaving
       the component stuff only
     */
    *segPtr = '\0';

    Tcl_SplitList(interp, inString, &argc, &argv);
    for (n = 0; n < argc; n++) {
	ptr = argv[n];
	switch (*ptr) {
#if 0
	case '\n':
	    /* NewLine: Finish current segment */

	    ++ptr;
	    *segPtr = '\0';
            result = Tm_XmStringAppendString(result, segBuf, charSet, dir, True);
	    segPtr = segBuf;
	    segment_started = True;
	    break;
#endif /* 0 */
	    
	case TM_FORMAT_CHAR:
            /* Enter command mode */
	    ptr++;

	    switch (*ptr) {
		case '\0':
		    fprintf(stderr, 
		        "XmString Conversion: Cannot handle trailing 'TM_FORMAT_CHAR'\n");
		    return NULL;
		    break;

		case TM_FORMAT_CHAR:
		    /* Two TM_FORMAT_CHAR => leave command mode, 
		       insert one TM_FORMAT_CHAR */
		    if (*segPtr != '\0')
		        *segPtr++ = ' ';
		    *segPtr++ = TM_FORMAT_CHAR;
		    break;

		case 'n':
	    	    *segPtr = '\0';
            	    result = Tm_XmStringAppendString(result, segBuf, 
					   charSet, dir, True);
	    	    segPtr = segBuf;
	    	    segment_started = True;
		    break;
		case 'F':
		case 'f':
		    /* save string collected so far */
		    *segPtr = '\0';
            	    result = Tm_XmStringAppendString(result, segBuf, 
					   charSet, dir, False);

		    /* now make new font */
		    segPtr = segBuf;
		    ptr++;
		    if (*ptr == '{') {
			/* M(f){font-name} */
			ptr++;
			while (*ptr && *ptr != '}') {
			    *segPtr++ = *ptr++;
			}
			if (*ptr == '}')
			    ptr++;
		    } else
		    if (*ptr == '(') {
			/* M(fXY) - 2 char font name XY */
			ptr++;
			if (ptr[0] && ptr[1]) {
			    *segPtr++ = *ptr++;
			    *segPtr++ = *ptr;
			} else {
			    *segPtr++ = 'X';
			    *segPtr++ = 'X';
			}
		    } else {
			/* M(fX) - 1 char font name X */
			*segPtr++ = *ptr ? *ptr : 'X';
		    }
		    *segPtr = '\0';
		    /* replace these 3 lines in 2.0 */
		    charSet = XtNewString(segBuf);
		    result = Tm_XmStringAppendString(result, "", 
                                           charSet, dir, False);

		    segPtr = segBuf;
	    	    segment_started = True;
		    *segPtr = '\0';
		    break;
		case 'L':
		case 'l':
	    	    *segPtr = '\0';
		    /* replace these 3 lines in 2.0 */
            	    result = Tm_XmStringAppendString(result, segBuf, 
					   charSet, dir, False);
		    dir = XmSTRING_DIRECTION_L_TO_R;
		    xm_dir = XmStringDirectionCreate(dir);
		    result = Tm_XmStringConcatAndFree(result, xm_dir);

		    segPtr = segBuf;
	    	    segment_started = True;
		    break;
		case 'R':
		case 'r':
	    	    *segPtr = '\0';
		    /* replace these 3 lines in 2.0 */
            	    result = Tm_XmStringAppendString(result, segBuf, 
					   charSet, dir, False);
		    dir = XmSTRING_DIRECTION_R_TO_L;
		    xm_dir = XmStringDirectionCreate(dir);
		    result = Tm_XmStringConcatAndFree(result, xm_dir);

		    segPtr = segBuf;
	    	    segment_started = True;
		    break;
		default:
		    break;
	    }
	    break;
	    	    
	default:
	    /* insert a space if not first */
	    if ( ! segment_started)
		*segPtr++ = ' ';
	    /* and copy string across */
	    while (*ptr)
	        *segPtr++ = *ptr++;
	    segment_started = False;
	    break;

	} /* switch */	    
    } /* for */

    *segPtr = '\0';
    result = Tm_XmStringAppendString(result, segBuf, charSet, dir, False);

    if (argc > 0)
	free((char *) argv);
    XtFree(segBuf);
    return result;
}

#else /* XmVERSION < 2 */

XmString    
Tm_StringToXmString(interp, inString)
    char  *inString;
    Tcl_Interp *interp;
{
    char              *segBuf;
    XmString          result = NULL;
    XmString	      xmstr;
    XmStringDirection  dir;
    char              *segPtr;
    char              *ptr;
    int 	      argc;
    char	      **argv;
    int		      n;
    Boolean	      segment_started = True;
    int		      len;

    len = strlen(inString); 
    segPtr = segBuf =  XtMalloc(len + 1);

    /* In Motif 2.0 the segment stuff will vanish, leaving
       the component stuff only
     */
    *segPtr = '\0';

    Tcl_SplitList(interp, inString, &argc, &argv);
    for (n = 0; n < argc; n++) {
	ptr = argv[n];
	switch (*ptr) {
	case '\n':
	    /* NewLine: Finish current segment */

	    *segPtr = '\0';
	    xmstr = XmStringComponentCreate(XmSTRING_COMPONENT_TEXT,
				strlen(segBuf), segBuf);
	    result = XmStringConcatAndFree(result, xmstr);

	    xmstr = XmStringComponentCreate(XmSTRING_COMPONENT_SEPARATOR,
				0, NULL);
	    result = XmStringConcatAndFree(result, xmstr);
	    segPtr = segBuf;
	    segment_started = True;
	    break;
	    
	case TM_FORMAT_CHAR:
            /* Enter command mode */
	    ptr++;

	    if (*ptr == '\0') {
		fprintf(stderr, 
		    "XmString Conversion: Cannot handle trailing 'TM_FORMAT_CHAR'\n");
		return NULL;
	    }

	    if (*ptr == TM_FORMAT_CHAR) {
		/* Two TM_FORMAT_CHAR => leave command mode, 
		   insert one TM_FORMAT_CHAR */
		if (*segPtr != '\0')
		    *segPtr++ = ' ';
		*segPtr++ = TM_FORMAT_CHAR;
	    } else

	    switch (*ptr) {
		case 'n':
		    *segPtr = '\0';
		    xmstr = XmStringComponentCreate(XmSTRING_COMPONENT_TEXT,
					strlen(segBuf), segBuf);
		    result = XmStringConcatAndFree(result, xmstr);

		    xmstr = XmStringComponentCreate(XmSTRING_COMPONENT_SEPARATOR,
				0, NULL);
		    result = XmStringConcatAndFree(result, xmstr);
		    segPtr = segBuf;
		    segment_started = True;
		    break;
		case 'F':
		case 'f':
		    /* save string collected so far */
		    *segPtr = '\0';
		    xmstr = XmStringComponentCreate(XmSTRING_COMPONENT_TEXT,
					strlen(segBuf), segBuf);
		    result = XmStringConcatAndFree(result, xmstr);

		    /* now make new font */
		    ptr++;
		    if (*ptr == '{') {
			/* M(f){font-name} */
			ptr[strlen(ptr) - 1] = '\0'; /* lose } */
			ptr++;
		    } else
		    if (*ptr == '(') {
			/* M(fXY) - 2 char font name XY */
			ptr++;
			ptr[2] = '\0'; /* should be anyway */
		    } else {
			/* M(fX) - 1 char font name X */
			ptr[1] = '\0'; /* should be anyway */
		    }
		    xmstr = XmStringComponentCreate(XmSTRING_COMPONENT_TAG,
					strlen(ptr), ptr);
		    result = XmStringConcatAndFree(result, xmstr);

		    segPtr = segBuf;
	    	    segment_started = True;
		    *segPtr = '\0';
		    break;
		case 'B':
		case 'b':
		    /* save string collected so far */
		    *segPtr = '\0';
		    xmstr = XmStringComponentCreate(XmSTRING_COMPONENT_TEXT,
					strlen(segBuf), segBuf);
		    result = XmStringConcatAndFree(result, xmstr);

		    /* now make new font */
		    ptr++;
		    if (*ptr == '{') {
			/* M(f){font-name} */
			ptr[strlen(ptr) - 1] = '\0'; /* lose } */
			ptr++;
		    } else
		    if (*ptr == '(') {
			/* M(fXY) - 2 char font name XY */
			ptr++;
			ptr[2] = '\0'; /* should be anyway */
		    } else {
			/* M(fX) - 1 char font name X */
			ptr[1] = '\0'; /* should be anyway */
		    }
		    xmstr = XmStringComponentCreate(
					XmSTRING_COMPONENT_RENDITION_BEGIN,
					strlen(ptr), ptr);
		    result = XmStringConcatAndFree(result, xmstr);

		    segPtr = segBuf;
	    	    segment_started = True;
		    *segPtr = '\0';
		    break;
		case 'E':
		case 'e':
		    /* save string collected so far */
		    *segPtr = '\0';
		    xmstr = XmStringComponentCreate(XmSTRING_COMPONENT_TEXT,
					strlen(segBuf), segBuf);
		    result = XmStringConcatAndFree(result, xmstr);

		    /* now make new font */
		    ptr++;
		    if (*ptr == '{') {
			/* M(f){font-name} */
			ptr[strlen(ptr) - 1] = '\0'; /* lose } */
			ptr++;
		    } else
		    if (*ptr == '(') {
			/* M(fXY) - 2 char font name XY */
			ptr++;
			ptr[2] = '\0'; /* should be anyway */
		    } else {
			/* M(fX) - 1 char font name X */
			ptr[1] = '\0'; /* should be anyway */
		    }
		    xmstr = XmStringComponentCreate(
					XmSTRING_COMPONENT_RENDITION_END,
					strlen(ptr), ptr);
		    result = XmStringConcatAndFree(result, xmstr);

		    segPtr = segBuf;
	    	    segment_started = True;
		    *segPtr = '\0';
		    break;
		case 'L':
		case 'l':
	    	    *segPtr = '\0';
		    xmstr = XmStringComponentCreate(XmSTRING_COMPONENT_TEXT,
					strlen(segBuf), segBuf);
		    result = XmStringConcatAndFree(result, xmstr);

		    dir = XmSTRING_DIRECTION_L_TO_R;
		    xmstr = XmStringDirectionCreate(dir);
		    result = XmStringConcatAndFree(result, xmstr);

		    segPtr = segBuf;
	    	    segment_started = True;
		    break;
		case 'R':
		case 'r':
	    	    *segPtr = '\0';
		    xmstr = XmStringComponentCreate(XmSTRING_COMPONENT_TEXT,
					strlen(segBuf), segBuf);
		    result = XmStringConcatAndFree(result, xmstr);

		    dir = XmSTRING_DIRECTION_R_TO_L;
		    xmstr = XmStringDirectionCreate(dir);
		    result = XmStringConcatAndFree(result, xmstr);

		    segPtr = segBuf;
	    	    segment_started = True;
		    break;
		default:
		    break;
	    }
	    break;
	    	    
	default:
	    /* insert a space if not first */
	    if ( ! segment_started)
		*segPtr++ = ' ';
	    /* and copy string across */
	    while (*ptr)
	        *segPtr++ = *ptr++;
	    segment_started = False;
	    break;

	} /* switch */	    
    } /* for */

    *segPtr = '\0';
    xmstr = XmStringComponentCreate(XmSTRING_COMPONENT_TEXT,
				strlen(segBuf), segBuf);
    result = XmStringConcatAndFree(result, xmstr);

    free((char *) argv);
    XtFree(segBuf);
    return result;
}

#endif /* XmVERSION < 2 */



/*
 *--------------------------------------------------------------
 *
 * Tm_CvtStringToXmString --
 *
 *	the String to XmString resource converter
 *
 *
 *--------------------------------------------------------------
 */

static Boolean
Tm_CvtStringToXmString(dpy, args, num_args, fromVal, toVal,
                                converter_data)
    Display   *dpy;
    XrmValue  *args;
    Cardinal  *num_args;
    XrmValue  *fromVal;
    XrmValue  *toVal;
    XtPointer *converter_data;
{
    static XmString result;
    Tcl_Interp *interp = (Tcl_Interp *) args->addr;


    if ((result = Tm_StringToXmString(interp, (char *)fromVal->addr)) == NULL) {
	XtDisplayStringConversionWarning(dpy, (char *)fromVal->addr, 
					 "XmString");
	return(False);
    }
    
    /* allow this to be garbage collected when possible */
    Tm_AddToFreeResourceList((char *) result,
		(Tm_FreeProc) XmStringFree);

    done(&result, XmString);
    return(True);
}


/*
 *--------------------------------------------------------------
 *
 * Tm_CvtStringToXmStringTable --
 *
 *
 *--------------------------------------------------------------
 */
Boolean
Tm_CvtStringToXmStringTable(display, args, num_args, 
		fromVal, toVal, destructor_data)
    Display *display;
    XrmValuePtr args;
    Cardinal *num_args;
    XrmValuePtr fromVal;
    XrmValuePtr toVal;
    XtPointer *destructor_data;
{
    Tcl_Interp *interp = (Tcl_Interp *) args->addr;
    char **argv;
    int argc;
    static XmStringTable tbl = NULL;
    int n;

    if (Tcl_SplitList(interp, (char *) fromVal->addr,
		&argc, &argv) == TCL_ERROR) {
	return False;
    }

    if (argc == 0) {
	tbl = NULL;
	done(&tbl, XmStringTable);
	free((char *) argv);
	return True;
    }

    tbl = (XmStringTable) XtMalloc((argc+1) * sizeof(XmString));
    Tm_AddToFreeResourceList((char *) tbl, (Tm_FreeProc) XtFree);

    for (n = 0; n < argc; n++) {
	tbl[n] = Tm_StringToXmString(interp, argv[n]);
        /* allow this to be garbage collected when possible */
        Tm_AddToFreeResourceList((char *) tbl[n],
                (Tm_FreeProc) XmStringFree);
    }
    tbl[argc] = NULL;
    done(&tbl, XmStringTable);
    free((char *) argv);

    return True;
}

/*
 *--------------------------------------------------------------
 *
 * Tm_CvtStringToString --
 *
 *	Converter from String to String type
 *
 * Results:
 *
 *	changes the "toVal" arg to contain the new value.
 *
 * Side effects:
 *
 *	none
 *--------------------------------------------------------------
 */

Boolean
Tm_CvtStringToString(display, args, num_args, fromVal, toVal, destructor_data)
    Display *display;
    XrmValuePtr args;
    Cardinal *num_args;
    XrmValuePtr fromVal;
    XrmValuePtr toVal;
    XtPointer *destructor_data;
{
    static char *buf;

    buf = *(char **) (fromVal->addr);
#   ifdef DEBUG
    fprintf(stderr, "converted string to %s\n", buf);
#   endif

    done(&buf, char *);

    return True;
}

/*
 *--------------------------------------------------------------
 *
 * Tm_CvtBooleanToString --
 *
 *	Converter from Boolean to String type
 *
 * Results:
 *
 *	changes the "toVal" arg to contain the new value.
 *
 * Side effects:
 *
 *	none
 *--------------------------------------------------------------
 */

Boolean
Tm_CvtBooleanToString(display, args, num_args, fromVal, toVal, destructor_data)
    Display *display;
    XrmValuePtr args;
    Cardinal *num_args;
    XrmValuePtr fromVal;
    XrmValuePtr toVal;
    XtPointer *destructor_data;
{
    static char *buf;
    Boolean boolVal;
 
    boolVal = *(Boolean *) (fromVal->addr);

    if (boolVal) {
	buf = "true";
    } else {
	buf = "false";
    } 
 
    done(&buf, char *);

    return True;
}


/*
 *--------------------------------------------------------------
 *
 * Tm_CvtBoolToString --
 *
 *	Converter from Bool to String type
 *
 * Results:
 *
 *	changes the "toVal" arg to contain the new value.
 *
 * Side effects:
 *
 *	none
 *--------------------------------------------------------------
 */

Boolean
Tm_CvtBoolToString(display, args, num_args, fromVal, toVal, destructor_data)
    Display *display;
    XrmValuePtr args;
    Cardinal *num_args;
    XrmValuePtr fromVal;
    XrmValuePtr toVal;
    XtPointer *destructor_data;
{
    static char *buf;
    Bool boolVal;
 
    boolVal = *(Bool *) (fromVal->addr);

    if (boolVal) {
	buf = "true";
    } else {
	buf = "false";
    } 
 
    done(&buf, char *);

    return True;
}

/*
 *--------------------------------------------------------------
 *
 * Tm_CvtIntToString --
 *
 *	Converter from Int to String type
 *
 * Results:
 *
 *	changes the "toVal" arg to contain the new value.
 *
 * Side effects:
 *
 *	none
 *--------------------------------------------------------------
 */

Boolean
Tm_CvtIntToString(display, args, num_args, fromVal, toVal, destructor_data)
    Display *display;
    XrmValuePtr args;
    Cardinal *num_args;
    XrmValuePtr fromVal;
    XrmValuePtr toVal;
    XtPointer *destructor_data;
{
    static char *buf = NULL;
    int intVal;
 
    intVal = *(int *) (fromVal->addr);

    if (buf == NULL) {
        buf = XtMalloc(16);
    }
    sprintf(buf, "%d", intVal);
 
    done(&buf, char *);

    return True;
}


/*
 *--------------------------------------------------------------
 *
 * Tm_CvtUnsignedIntToString --
 *
 *	Converter from UnsignedInt to String type
 *
 * Results:
 *
 *	changes the "toVal" arg to contain the new value.
 *
 * Side effects:
 *
 *	none
 *--------------------------------------------------------------
 */

Boolean
Tm_CvtUnsignedIntToString(display, args, num_args, fromVal, toVal, destructor_data)
    Display *display;
    XrmValuePtr args;
    Cardinal *num_args;
    XrmValuePtr fromVal;
    XrmValuePtr toVal;
    XtPointer *destructor_data;
{
    static char *buf = NULL;
    unsigned int intVal;
 
    intVal = *(int *) (fromVal->addr);

    if (buf == NULL)
        buf = XtMalloc(16);
    sprintf(buf, "%u", intVal);
 
    done(&buf, char *);

    return True;
}

/*
 *--------------------------------------------------------------
 *
 * Tm_CvtDimensionToString --
 *
 *	Converter from Dimension to String type
 *
 * Results:
 *
 *	changes the "toVal" arg to contain the new value.
 *
 * Side effects:
 *
 *	none
 *--------------------------------------------------------------
 */

Boolean
Tm_CvtDimensionToString(display, args, num_args, fromVal, toVal, destructor_data)
    Display *display;
    XrmValuePtr args;
    Cardinal *num_args;
    XrmValuePtr fromVal;
    XrmValuePtr toVal;
    XtPointer *destructor_data;
{
    static char *buf = NULL;
    Dimension dimVal;
 
    dimVal = *(Dimension *) (fromVal->addr);

    if (buf == NULL)
        buf = XtMalloc(16);
    sprintf(buf, "%hu", dimVal);
 
    done(&buf, char *);

    return True;
}

/*
 *--------------------------------------------------------------
 *
 * Tm_CvtShortToString --
 *
 *	Converter from Short to String type
 *
 * Results:
 *
 *	changes the "toVal" arg to contain the new value.
 *
 * Side effects:
 *
 *	none
 *--------------------------------------------------------------
 */

Boolean
Tm_CvtShortToString(display, args, num_args, fromVal, toVal, destructor_data)
    Display *display;
    XrmValuePtr args;
    Cardinal *num_args;
    XrmValuePtr fromVal;
    XrmValuePtr toVal;
    XtPointer *destructor_data;
{
    static char *buf = NULL;
    short shortVal;
 
    shortVal = *(short *) (fromVal->addr);

    if (buf == NULL)
        buf = XtMalloc(16);
    sprintf(buf, "%hd", shortVal);
 
    done(&buf, char *);

    return True;
}


/*
 *--------------------------------------------------------------
 *
 * Tm_CvtLongToString --
 *
 *	Converter from Long to String type
 *
 * Results:
 *
 *	changes the "toVal" arg to contain the new value.
 *
 * Side effects:
 *
 *	none
 *--------------------------------------------------------------
 */

Boolean
Tm_CvtLongToString(display, args, num_args, fromVal, toVal, destructor_data)
    Display *display;
    XrmValuePtr args;
    Cardinal *num_args;
    XrmValuePtr fromVal;
    XrmValuePtr toVal;
    XtPointer *destructor_data;
{
    static char *buf = NULL;
    long longVal;
 
    longVal = *(long *) (fromVal->addr);

    if (buf == NULL)
        buf = XtMalloc(16);
    sprintf(buf, "%ld", longVal);
 
    done(&buf, char *);

    return True;
}


/*
 *--------------------------------------------------------------
 *
 * Tm_CvtPixelToString --
 *
 *	Converter from Pixel to String type
 *	gives result in #RGB form
 *
 * Results:
 *
 *	changes the "toVal" arg to contain the new value.
 *
 * Side effects:
 *
 *	none
 *--------------------------------------------------------------
 */

Boolean
Tm_CvtPixelToString(display, args, num_args, fromVal, toVal, destructor_data)
    Display *display;
    XrmValuePtr args;
    Cardinal *num_args;
    XrmValuePtr fromVal;
    XrmValuePtr toVal;
    XtPointer *destructor_data;
{
    static char *buf = NULL;
    Screen *screen;
    Colormap colormap;
    XColor color;
 
    color.pixel = *(unsigned long *) fromVal->addr;
    screen = *((Screen **) args[0].addr);
    colormap = *((Colormap *) args[1].addr);

    XQueryColor(DisplayOfScreen(screen), colormap, &color);

    if (buf == NULL)
        buf = XtMalloc(32);
    sprintf(buf, "#%04x%04x%04x", color.red, color.green, color.blue);

    done(&buf, char *);

    return True;
}

/*
 *--------------------------------------------------------------
 *
 * Tm_CvtStringToPixmap --
 *
 *	Converter from String to Pixmap type
 *
 * Results:
 *
 *	changes the "toVal" arg to contain the new value.
 *
 * Side effects:
 *
 *	none
 *--------------------------------------------------------------
 */

Boolean
Tm_CvtStringToPixmap(display, args, num_args, fromVal, toVal, destructor_data)
    Display *display;
    XrmValuePtr args;
    Cardinal *num_args;
    XrmValuePtr fromVal;
    XrmValuePtr toVal;
    XtPointer *destructor_data;
{
    Widget w = *((Widget *) args[0].addr);

    /* use the Motif registered converter for this! (why reinvent?) */
    if (XtConvertAndStore(w, XmRString, fromVal, 
		XmRPrimForegroundPixmap, toVal)) {
        return True;
    }
    return False;
}

/*
 *--------------------------------------------------------------
 *
 * Tm_CvtStringToFunction --
 *
 *	Converter from String to XmRProc type
 *	(for dropProc in D&D)
 *
 * Results:
 *
 *	changes the "toVal" arg to contain the new value.
 *
 * Side effects:
 *
 *	none
 *--------------------------------------------------------------
 */

Boolean
Tm_CvtStringToFunction(display, args, num_args, fromVal, toVal, destructor_data)
    Display *display;
    XrmValuePtr args;
    Cardinal *num_args;
    XrmValuePtr fromVal;
    XrmValuePtr toVal;
    XtPointer *destructor_data;
{
#ifndef MOTIF11
    char *command;
    extern Tm_Widget *Tm_HackDragWidgetPtr;
    Tm_Widget *wPtr = Tm_HackDragWidgetPtr;
    Widget w = *((Widget *) args[1].addr);
    static XtConvertSelectionProc p = Tm_ConvertProcHandler;

    /* set our proc in widgets user data */
    command = (char *) (fromVal->addr);
/*
    XtVaGetValues(w, XmNuserData, &wPtr, NULL);
*/
    wPtr->convertProc = XtNewString(command);
    XtVaSetValues(w, XmNuserData, wPtr, NULL);
    
    done(&p, XtConvertSelectionProc);

    return True;
#endif
}


/*
 *--------------------------------------------------------------
 *
 * Tm_CvtStringToProc --
 *
 *	Converter from String to XmRFunc type
 *	(for convertProc resource in D&D)
 *
 * Results:
 *
 *	changes the "toVal" arg to contain the new value.
 *
 * Side effects:
 *
 *	none
 *--------------------------------------------------------------
 */

Boolean
Tm_CvtStringToProc(display, args, num_args, fromVal, toVal, destructor_data)
    Display *display;
    XrmValuePtr args;
    Cardinal *num_args;
    XrmValuePtr fromVal;
    XrmValuePtr toVal;
    XtPointer *destructor_data;
{
#ifndef MOTIF11
    char *command;
    Tm_Widget *wPtr;
    Widget w = *((Widget *) args[1].addr);
    static XtCallbackProc p = Tm_DropProcHandler;

   /* note these two hacks because XmRProc
      is not a unique enough type */
    extern Boolean Tm_SettingDropResources;
    extern Boolean Tm_SettingDropProc;
 
    if (Tm_SettingDropResources) {
        /* set our proc in widgets user data */
        command = (char *) (fromVal->addr);
        XtVaGetValues(w, XmNuserData, &wPtr, NULL);
	if (Tm_SettingDropProc) {
            wPtr->dropProc = XtNewString(command);
	} else {
	    return False;
	}
        XtVaSetValues(w, XmNuserData, wPtr, NULL);
    
        done(&p, XtCallbackProc);

        return True;
    }
#endif /* not MOTIF11 */
    return False;
}


/*
 *--------------------------------------------------------------
 *
 * Tm_CvtStringToCallbackProc --
 *
 *	Converter from String to XmRFunc type
 *
 * Results:
 *
 *	changes the "toVal" arg to contain the new value.
 *
 * Side effects:
 *
 *	none
 *--------------------------------------------------------------
 */

Boolean
Tm_CvtStringToCallbackProc(display, args, num_args, fromVal, toVal, destructor_data)
    Display *display;
    XrmValuePtr args;
    Cardinal *num_args;
    XrmValuePtr fromVal;
    XrmValuePtr toVal;
    XtPointer *destructor_data;
{
#ifndef MOTIF11
    char *command;
    extern Tm_Widget *Tm_HackWidgetPtr;
    Tm_Widget *wPtr = Tm_HackWidgetPtr;
    static XtSelectionCallbackProc p = Tm_DropTransferHandler;

    /* set our proc in widgets user data */
    command = (char *) (fromVal->addr);
    wPtr->transferProc = XtNewString(command);
    done(&p, XtCallbackProc);
    return True;
#endif
}

/*
 *--------------------------------------------------------------
 *
 * Tm_CvtStringToWidgetClass --
 *
 *      Converter from String to WidgetClass
 *
 * Results:
 *
 *      changes the "toVal" arg to contain the new value.
 *
 * Side effects:
 *
 *      none
 *--------------------------------------------------------------
 */

#if 1
Boolean
Tm_CvtStringToWidgetClass(display, args, num_args, fromVal,
		toVal, destructor_data)
    Display *display;
    XrmValuePtr args;
    Cardinal *num_args;
    XrmValuePtr fromVal;
    XrmValuePtr toVal;
    XtPointer *destructor_data;
{
    String str_class = (char *) fromVal->addr;
    static WidgetClass wc;
    int n = 0;
    extern Tm_CommandToClassType Tm_CommToClass[];
    extern Tm_CommandToClassType Tm_ExternCommandToClass[];

    while (Tm_CommToClass[n].command != NULL) {
	wc = *(Tm_CommToClass[n].clas);
	if (strcmp(wc->core_class.class_name, str_class) == 0) {
	    done(&wc, WidgetClass);
	    return True;
	}
	n++;
    }

    n = 0;
    while (Tm_ExternCommandToClass[n].command != NULL) {
	wc = *(Tm_ExternCommandToClass[n].clas);
	if (strcmp(wc->core_class.class_name, str_class) == 0) {
	    done(&wc, WidgetClass);
	    return True;
	}
	n++;
    }

    return False;
}
#endif

/*
 *--------------------------------------------------------------
 *
 * Tm_RegisterConverters --
 *
 *	Register all extra converters with Xt.
 *
 * Results:
 *
 * Side effects:
 *
 *	Modifies the Xt converter lists.
 *--------------------------------------------------------------
 */

void
Tm_RegisterConverters(interp, appContext)
    Tcl_Interp *interp;
    XtAppContext appContext;
{
    XtConvertArgRec convert_arg;
    XtConvertArgRec convert_args[128];
 
    /*
     * Register the type converters we need
     * This implementation installs everything, 
     * even if it is never used. A smarter method
     * would be to install on demand. Maybe someday...
     */

    InstallReverseRepTypes();

    /* 
     * String to Widget
     */
    convert_arg.address_mode = XtAddress;
    convert_arg.address_id = (XtPointer) interp;
    convert_arg.size = sizeof(XtPointer);

    XtSetTypeConverter(
		XtRString, XtRWidget,
		Tm_CvtStringToWidget,
		&convert_arg,
		1,
		XtCacheNone,
		NULL);

    /*
     * Motif 1.2.1 MainW.c has bugs in it - resources like menuBar are declared
     * as type Window not Widget. So we have to have another converter just
     * to handle this bug
     */
    convert_arg.address_mode = XtAddress;
    convert_arg.address_id = (XtPointer) interp;
    convert_arg.size = sizeof(XtPointer);

    XtSetTypeConverter(
		XtRString, XtRWindow,
		Tm_CvtStringToWidget,
		&convert_arg,
		1,
		XtCacheNone,
		NULL);

    /* 
     * String to MenuWidget
     */
    convert_arg.address_mode = XtAddress;
    convert_arg.address_id = (XtPointer) interp;
    convert_arg.size = sizeof(XtPointer);

    XtSetTypeConverter(
		XtRString, XmRMenuWidget,
		Tm_CvtStringToWidget,
		&convert_arg,
		1,
		XtCacheNone,
		NULL);

    /* 
     * Widget to String
     */

    XtSetTypeConverter(
		XtRWidget, XmRString,
		Tm_CvtWidgetToString,
		NULL,
		0,
		XtCacheNone,
		NULL);

    /* 
     * MenuWidget to String
     */

    XtSetTypeConverter(
		XmRMenuWidget, XmRString,
		Tm_CvtWidgetToString,
		NULL,
		0,
		XtCacheNone,
		NULL);

    /* 
     * Widget to String - this is registered as Window to String because
     * of the buggy resource decls mentioned before
     */

    XtSetTypeConverter(
		XtRWindow, XmRString,
		Tm_CvtWidgetToString,
		NULL,
		0,
		XtCacheNone,
		NULL);

    /* 
     * WidgetList to String
     */
    convert_args[0].address_mode = XtWidgetBaseOffset;
    convert_args[0].address_id = (XtPointer) XtOffsetOf(CoreRec, core.self);
    convert_args[0].size = sizeof(CoreWidget);

    XtSetTypeConverter(
		XtRWidgetList, XmRString,
		Tm_CvtWidgetListToString,
		convert_args,
		1,
		XtCacheNone,
		NULL);

#ifndef MOTIF11
    /* 
     * String to DropTransfers
     */
    convert_arg.address_mode = XtAddress;
    convert_arg.address_id = (XtPointer) interp;
    convert_arg.size = sizeof(XtPointer);

    XtSetTypeConverter(
		XtRString, XmRDropTransfers,
		Tm_CvtStringToDropTransfers,
		&convert_arg,
		1,
		XtCacheNone,
		NULL);

    /* 
     * String to TransferStatus (as in dropTransfer widget
     */

    XtSetTypeConverter(
		XtRString, XmRTransferStatus,
		Tm_CvtStringToTransferStatus,
		NULL,
		0,
		XtCacheNone,
		NULL);
#endif

    /* 
     * String to Pixmap
     */
    convert_args[0].address_mode = XtWidgetBaseOffset;
    convert_args[0].address_id = (XtPointer) XtOffsetOf(CoreRec, core.self);
    convert_args[0].size = sizeof(CoreWidget);

    XtSetTypeConverter(
		XmRString, XtRPixmap,
		Tm_CvtStringToPixmap,
		convert_args,
		1,
		XtCacheNone,
		NULL);
    /* 
     * XmString to String
     */
    XtSetTypeConverter(
		XmRXmString, XtRString,
		Tm_CvtXmStringToString,
		NULL,
		0,
		XtCacheNone,
		NULL);

    /* 
     * XmStringTable to String
     */
    convert_arg.address_mode = XtWidgetBaseOffset;
    convert_arg.address_id = (XtPointer) XtOffsetOf(CoreRec, core.self);
    convert_arg.size = sizeof(CoreWidget);

    XtSetTypeConverter(
		XmRXmStringTable, XtRString,
		Tm_CvtXmStringTableToString,
		&convert_arg,
		1,
		XtCacheNone,
		NULL);

    /* 
     * StringTable to String
     */
    convert_arg.address_mode = XtWidgetBaseOffset;
    convert_arg.address_id = (XtPointer) XtOffsetOf(CoreRec, core.self);
    convert_arg.size = sizeof(CoreWidget);

    XtSetTypeConverter(
		XmRStringArray, XtRString,
		Tm_CvtStringTableToString,
		&convert_arg,
		1,
		XtCacheNone,
		NULL);

    /* 
     * String to XmString
     */
    convert_arg.address_mode = XtAddress;
    convert_arg.address_id = (XtPointer) interp;
    convert_arg.size = sizeof(XtPointer);

    XtSetTypeConverter(
		XmRString, XmRXmString,
		Tm_CvtStringToXmString,
		&convert_arg,
		1,
		XtCacheNone,
		NULL);

    /*
     * String to XmStringTable
     */
    convert_arg.address_mode = XtAddress;
    convert_arg.address_id = (XtPointer) interp;
    convert_arg.size = sizeof(XtPointer);

    XtSetTypeConverter(
		XmRString, XmRXmStringTable,
		Tm_CvtStringToXmStringTable,
		&convert_arg,
		1,
		XtCacheNone,
		NULL);

    /* 
     * String to String
     * I don't know why there isn't already one of these,
     * but getting value out of Text seems to need it.
     *
     * it doesn't seem to work anyway - maybe something
     * special case away X to X convertions
     */
/*
    convert_arg.address_mode = XtAddress;
    convert_arg.address_id = (XtPointer) interp;
    convert_arg.size = sizeof(XtPointer);

    XtSetTypeConverter(
		XtRString, XtRString,
		Tm_CvtStringToString,
		&convert_arg,
		1,
		XtCacheNone,
		NULL);
*/

    /* 
     * Int to String
     */
    XtSetTypeConverter(
		XmRInt, XtRString,
		Tm_CvtIntToString,
		NULL,
		0,
		XtCacheNone,
		NULL);


    /* 
     * Cardinal to String
     */
    XtSetTypeConverter(
		XmRCardinal, XtRString,
		Tm_CvtUnsignedIntToString,
		NULL,
		0,
		XtCacheNone,
		NULL);

    /* 
     * HorizontalInt to String
     */
    XtSetTypeConverter(
		XmRHorizontalInt, XtRString,
		Tm_CvtIntToString,
		NULL,
		0,
		XtCacheNone,
		NULL);

    /* 
     * VerticalInt to String
     */
    XtSetTypeConverter(
		XmRVerticalInt, XtRString,
		Tm_CvtIntToString,
		NULL,
		0,
		XtCacheNone,
		NULL);


    /* 
     * Boolean to String
     */
    XtSetTypeConverter(
		XmRBoolean, XtRString,
		Tm_CvtBooleanToString,
		NULL,
		0,
		XtCacheNone,
		NULL);


    /* 
     * Bool to String
     */
    XtSetTypeConverter(
		XmRBool, XtRString,
		Tm_CvtBoolToString,
		NULL,
		0,
		XtCacheNone,
		NULL);

    /* 
     * Dimension to String
     */
    XtSetTypeConverter(
		XmRDimension, XtRString,
		Tm_CvtDimensionToString,
		NULL,
		0,
		XtCacheNone,
		NULL);

    /* 
     * BooleanDimension to String
     */
    XtSetTypeConverter(
		XmRBooleanDimension, XtRString,
		Tm_CvtDimensionToString,
		NULL,
		0,
		XtCacheNone,
		NULL);

    /* 
     * HorizontalDimension to String
     * Buggy? should use UnitType resource
     */
    XtSetTypeConverter(
		XmRHorizontalDimension, XtRString,
		Tm_CvtDimensionToString,
		NULL,
		0,
		XtCacheNone,
		NULL);

    /* 
     * VerticalDimension to String
     * Buggy? should use UnitType resource
     */

    XtSetTypeConverter(
		XmRVerticalDimension, XtRString,
		Tm_CvtDimensionToString,
		NULL,
		0,
		XtCacheNone,
		NULL);

    /* 
     * short to String
     */
    XtSetTypeConverter(
		XmRShort, XtRString,
		Tm_CvtShortToString,
		NULL,
		0,
		XtCacheNone,
		NULL);

    /* 
     * Position to String
     */
    XtSetTypeConverter(
		XmRPosition, XtRString,
		Tm_CvtShortToString,
		NULL,
		0,
		XtCacheNone,
		NULL);

    /* 
     * HorizontalPosition to String
     */
    XtSetTypeConverter(
		XmRHorizontalPosition, XtRString,
		Tm_CvtShortToString,
		NULL,
		0,
		XtCacheNone,
		NULL);

    /* 
     * VerticalPosition to String
     */
    XtSetTypeConverter(
		XmRVerticalPosition, XtRString,
		Tm_CvtShortToString,
		NULL,
		0,
		XtCacheNone,
		NULL);

    /* 
     * TextPosition to String
     */
    XtSetTypeConverter(
		XmRTextPosition, XtRString,
		Tm_CvtLongToString,
		NULL,
		0,
		XtCacheNone,
		NULL);


    /* 
     * ShellHorizPos to String
     */
    XtSetTypeConverter(
		XmRShellHorizPos, XtRString,
		Tm_CvtShortToString,
		NULL,
		0,
		XtCacheNone,
		NULL);


    /* 
     * ShellVertPos to String
     */
    XtSetTypeConverter(
		XmRShellVertPos, XtRString,
		Tm_CvtShortToString,
		NULL,
		0,
		XtCacheNone,
		NULL);


    /* 
     * ShellHorizDim to String
     */
    XtSetTypeConverter(
		XmRShellHorizDim, XtRString,
		Tm_CvtShortToString,
		NULL,
		0,
		XtCacheNone,
		NULL);


    /* 
     * ShellVertDim to String
     */
    XtSetTypeConverter(
		XmRShellVertDim, XtRString,
		Tm_CvtShortToString,
		NULL,
		0,
		XtCacheNone,
		NULL);

    /* 
     * String to XmProc
     */
    convert_args[0].address_mode = XtImmediate;
    convert_args[0].address_id = (XtPointer) interp;
    convert_args[0].size = sizeof(XtPointer);

    convert_args[1].address_mode = XtWidgetBaseOffset;
    convert_args[1].address_id = (XtPointer) XtOffsetOf(CoreRec, core.self);
    convert_args[1].size = sizeof(CoreWidget);

    convert_args[2].address_mode = XtWidgetBaseOffset;
    convert_args[2].address_id = (XtPointer) XtOffsetOf(CoreRec, core.widget_class);
    convert_args[2].size = sizeof(CoreWidget);

    XtSetTypeConverter(
		XmRString, XmRProc,
		Tm_CvtStringToProc,
		convert_args,
		3,
		XtCacheNone,
		NULL);

    /* 
     * String to XmCallbackProc
     */
    convert_args[0].address_mode = XtImmediate;
    convert_args[0].address_id = (XtPointer) interp;
    convert_args[0].size = sizeof(XtPointer);

    convert_args[1].address_mode = XtWidgetBaseOffset;
    convert_args[1].address_id = (XtPointer) XtOffsetOf(CoreRec, core.self);
    convert_args[1].size = sizeof(CoreWidget);

    convert_args[2].address_mode = XtWidgetBaseOffset;
    convert_args[2].address_id = (XtPointer) XtOffsetOf(CoreRec, core.widget_class);
    convert_args[2].size = sizeof(CoreWidget);

    XtSetTypeConverter(
		XmRString, XmRCallbackProc,
		Tm_CvtStringToCallbackProc,
		convert_args,
		3,
		XtCacheNone,
		NULL);

    /* 
     * String to XmFunc
     */
    convert_args[0].address_mode = XtImmediate;
    convert_args[0].address_id = (XtPointer) interp;
    convert_args[0].size = sizeof(XtPointer);

    convert_args[1].address_mode = XtWidgetBaseOffset;
    convert_args[1].address_id = (XtPointer) XtOffsetOf(CoreRec, core.self);
    convert_args[1].size = sizeof(CoreWidget);

    convert_args[2].address_mode = XtWidgetBaseOffset;
    convert_args[2].address_id = (XtPointer) XtOffsetOf(CoreRec, core.widget_class);
    convert_args[2].size = sizeof(CoreWidget);

    XtSetTypeConverter(
		XmRString, XmRFunction,
		Tm_CvtStringToFunction,
		convert_args,
		3,
		XtCacheNone,
		NULL);

    /* 
     * Pixel to String
     */
    convert_args[0].address_mode = XtWidgetBaseOffset;
    convert_args[0].address_id = (XtPointer) XtOffsetOf(WidgetRec, core.screen);
    convert_args[0].size = sizeof(Screen *);

    convert_args[1].address_mode = XtWidgetBaseOffset;
    convert_args[1].address_id = (XtPointer) XtOffsetOf(CoreRec, core.colormap);
    convert_args[1].size = sizeof(Colormap);

    XtSetTypeConverter(
		XmRPixel, XmRString,
		Tm_CvtPixelToString,
		convert_args,
		2,
		XtCacheNone,
		NULL);
#if 1

    /*
     * String to WidgetClass
     */
    XtSetTypeConverter(
		XmRString, XmRWidgetClass,
		Tm_CvtStringToWidgetClass,
		NULL, 0, XtCacheNone, NULL);
#endif
}
