
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: builtin.h,v 2.3 1995/06/17 20:33:21 davidsi Exp $
//
// $Log: builtin.h,v $
// Revision 2.3  1995/06/17  20:33:21  davidsi
// - Added CHAR_VALUE to support the char type.
//
// Revision 2.2  1995/06/09  20:26:50  davidsi
// - Added GNU copyright message.
//
// Revision 2.1  1995/04/16  19:56:30  davidsi
// - Added support for builtin functions that wish to return values rather
//   than passing via arguments.  The return_value member is NULL or a
//   pointer to the value that is to returned. It can be set with the
//   SetReturn() function member.
//
// Revision 2.0  1995/01/19  21:58:31  davidsi
// - Initial version.
//
//
//
#ifndef __BUILTIN_H
#define __BUILTIN_H
#include "context.h"
#include "value.h"
#include "node.h"

//
// Maximum number of arguments a built in function can handle
//
#define MAX_ARGS	32

//
// A builtin function inherits everything it needs from a stanard
// function declaration.  The only difference is that when
// executed, the control is passed to a standard function.
//
class node_builtin:public node_funcdecl {
	int (*funcaddr)(class node_builtin *);	// the C/C++ function to call

	class context *cur_context;
	class node_arg *cur_arg;
	//
	// Arguments
	//
        class value *val[MAX_ARGS];
        int num_args;
	class value *return_value;	// the return value for the function
private:
	void getarg();
	void freearg();
public:
	virtual class value *evaluate(class context *,class context *,class node_arg *);

	//
	// Using the [] operator gives access to the arguments
	//
	class value *operator[](int);
	int number_of_arguments();
	int types_of_arguments(int num,class value **req_type);
	void SetReturn(class value *);

	node_builtin(class context *,char *,int (*func)(class node_builtin &));
	virtual ~node_builtin();
};

//
// Reference types
//
extern class value_int *INT_VALUE;
extern class value_string *STRING_VALUE;
extern class value_real *REAL_VALUE;
extern class value_char *CHAR_VALUE;


#endif
