
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

// $Id: button.C,v 2.2 1995/06/24 15:01:04 davidsi Exp $
//
// $Log: button.C,v $
// Revision 2.2  1995/06/24  15:01:04  davidsi
// - Added GNU copyright message.
//
// Revision 2.1  1995/02/27  22:11:08  davidsi
// - Added new button creating functions that understand the Dialog class,
//   and removed those that required actual dialog references.
// - Added support for the Buttoninstance class. This allows information
//   about created buttons to be easily read and processed.
//
// Revision 2.0  1995/02/02  21:14:16  davidsi
// - First version.
//
//
//

#include <stdio.h>
#include <varargs.h>
#define _STDARG_H
#include <stdlib.h>
#include <string.h>

#define FULL_INCLUDE

#include "localmacros.h"

#include "button.h"


//
// ButtoninstanceDestroy
//
// Destroys a Buttoninstance. This does more than the destructor as it
// will also pull the instance off the list of active instances. Note 
// that the remove() couldn't be put in ~Buttoninstance() (giving
// a single delete/destroy function) because the destructor function 
// will be called by the list destructor and the remove() is then
// not applicable!
//
// INPUTS :	None
// RETURNS:	None
//
void
Buttoninstance::ButtoninstanceDestroy()
{
	remove();
	delete this;
}

//
// ButtoninstanceGet
//
// Returns the state of the button given by a Buttoninstance.
//
// INPUTS :	None
// RETURNS: 	A buttonstate_e value indicating the state of the button
//
buttonstate_e
Buttoninstance::ButtoninstanceGet()
{
	buttonstate_e flag;

	if ( dialog_button_value(d_button) )
                flag=BUTTON_ENABLED;
	else
                flag=BUTTON_DISABLED;

	return(flag);
}

//
// Constructor for the Buttoninstance class.  This builds the wrapper ready for
// use later on.  The code using the constructor is responsible for
// maintaining the list of wrapper instances.
//
Buttoninstance::Buttoninstance( class Button *ifaddr,unsigned long icbvalue)
{
        faddr=ifaddr;
        cbvalue=icbvalue;
}


//
// Destructor
//
Buttoninstance::~Buttoninstance()
{
}


//
// mainbutton
//
// This is the static function that button callbacks first come to. It
// unwraps the pointer to determine the actual instance of the class
// and an id that can be passed to the ButtonCallback() member.
// 
// INPUTS :	d 	The dialog causing the callback
//		iptr	The wrapped up values for this and the id.
// RETURNS:	None
//
void
Button::mainbutton(dialog d, char *iptr)
{
	class Buttoninstance *ptr;
	buttonstate_e flag;
	
	ptr=(class Buttoninstance *)iptr;
	if ( dialog_button_value(d) )
                flag=BUTTON_ENABLED;
	else
                flag=BUTTON_DISABLED;

        ptr->faddr->ButtonCallback(ptr->cbvalue,flag);
}



//
// ButtonCallback id buttonstate
//
// Standard button callback handler.  This function should nearly
// always be overriden.
//
// INPUTS :	id 	The id for the button that was pressed.
//		flag	The state of the button
// RETURNS:	None
//
void
Button::ButtonCallback(unsigned long id,buttonstate_e flag)
{
}

//
// ButtonPush dialog buttonlabel callbackid
//
// Creates a push button. When the button is pressed the ButtonCallback
// function member will be called with the callbackid argument.
//
// INPUTS :	dialog		The dialog in which the button is to be created
//		buttonlabel	The label to be displayed in the button
//		callbackid	The id for the callback
// RETURNS:	None
//
Buttoninstance *
Button::ButtonPush(
		Dialog &d,
		char *name,
		unsigned long icallback_val
	)
{
	class Buttoninstance *cb;
	static new_d_pbutton nb1 = {
		NULL,
		NULL,                     // CB_finish_action,
		"Name of button"
	};


	cb=new class Buttoninstance(this,icallback_val);

	//
	// Add to list, so it can be deleted later on.
	//
	callbacks->insert_head(cb);

	nb1.value=(char *)cb;
	nb1.name=name;
        nb1.callback=&Button::mainbutton;
        cb->d_button=dialog_pbutton(d.d,&nb1,NULL);

	return(cb);
}

//
// ButtonToggle dialog buttonlabel callbackid
//
// Creates a toggle button. When the button is pressed the ButtonCallback
// function member will be called with the callbackid argument and
// a flag describing the button's state.
//
// INPUTS :	dialog		The dialog in which the button is to be created
//		buttonlabel	The label to be displayed in the button
//		callbackid	The id for the callback
// RETURNS:	None
//
Buttoninstance *
Button::ButtonToggle(
	Dialog &d,
	char *name,
	unsigned long icallback_val,
	buttonstate_e initialvalue
	)
{
	class Buttoninstance *cb;
	static new_d_tbutton nb1 = {
		0,				// flags
                "Title",			// title
                1,                              // initial value is on
                NULL,                           // no call back
                0                               // no call back
	};


	cb=new class Buttoninstance(this,icallback_val);

	//
	// Add to list, so it can be deleted later on.
	//
	callbacks->insert_head(cb);

	nb1.value=(char *)cb;
	nb1.name=name;
        nb1.callback=&Button::mainbutton;
	nb1.flags=0;
	if (initialvalue==BUTTON_ENABLED)
        	nb1.ivalue=1;
	else
	        nb1.ivalue=0;
		
        cb->d_button=dialog_tbutton(d.d,&nb1,NULL);

	return(cb);
}

//
// ButtonTogglemutex dialog buttonlabel callbackid
//
// Creates a toggle button that will exclude all other buttons in
// the higher level container dialog. When the button is pressed the
// ButtonCallback function member will be called with the callbackid
// argument and a flag describing the button's state.
//
// INPUTS :	dialog		The dialog in which the button is to be created
//		buttonlabel	The label to be displayed in the button
//		callbackid	The id for the callback
// RETURNS:	None
//
Buttoninstance *
Button::ButtonTogglemutex(
	Dialog &d,
	char *name,
	unsigned long icallback_val,
	buttonstate_e initialvalue
	)
{
	class Buttoninstance *cb;
	static new_d_tbutton nb1 = {
		DIALOG_TBUTTON_MUTEX,		// flags
                "Title",			// title
                1,                              // initial value is on
                NULL,                           // no call back
                0                               // no call back
	};


	cb=new class Buttoninstance(this,icallback_val);

	//
	// Add to list, so it can be deleted later on.
	//
	callbacks->insert_head(cb);

	nb1.value=(char *)cb;
	nb1.name=name;
        nb1.callback=&Button::mainbutton;
	if (initialvalue==BUTTON_ENABLED)
        	nb1.ivalue=1;
	else
	        nb1.ivalue=0;
		
        cb->d_button=dialog_tbutton(d.d,&nb1,NULL);

	return(cb);
}



// 
// Constructor for the button class.
//
Button::Button()
{
	callbacks=new class list;
}

// 
// Destructor for the button class.
//
Button::~Button()
{
	//
	// Delete all remaining callbacks
	//
	delete callbacks;
}
