
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: button.h,v 2.2 1995/06/09 20:32:04 davidsi Exp $
//
// $Log: button.h,v $
// Revision 2.2  1995/06/09  20:32:04  davidsi
// - Added GNU copyright message.
//
// Revision 2.1  1995/02/27  22:15:38  davidsi
// - Added new button creating functions that understand the Dialog class,
//   and removed those that required actual dialog references.
// - Added support for the Buttoninstance class. This allows information
//   about created buttons to be easily read and processed.
//
// Revision 2.0  1995/02/02  21:14:16  davidsi
// - First version.
//
//
//
#ifndef __BUTTON_H
#define __BUTTON_H

#include "suplib.h"

#include "dialog.h"
#include "list.h"


//
// The states a button can have
//
enum buttonstate_e {
	BUTTON_ENABLED,
	BUTTON_DISABLED
	};


//
// A Buttoninstance describes a single button that has been created. It
// inherits from the list_node so that it can form part of a much larger
// list of created buttons - ready for deletion (etc)
//
class Buttoninstance : public list_node
{
	dialog d_button;
	class Button *faddr;
	unsigned long cbvalue;
public:
	void ButtoninstanceDestroy();
	buttonstate_e ButtoninstanceGet();

	Buttoninstance( class Button *faddr , unsigned long);
	virtual ~Buttoninstance();

friend class Button;
};

// 
// The Button class
//
class Button 
{
private:
	class list *callbacks;			// all attached callbacks
	static void mainbutton(dialog,char *);	// real callback function
	virtual void ButtonCallback(unsigned long id,buttonstate_e flag); // const=0;

public:
	Buttoninstance *ButtonPush(class Dialog &,char *name,unsigned long val);
	Buttoninstance *ButtonToggle(class Dialog &,char *name,
					unsigned long val,buttonstate_e);
	Buttoninstance *ButtonTogglemutex(class Dialog &,char *nm,
					unsigned long vl,buttonstate_e);
	Button(); 
	virtual ~Button(); 
};

#endif
