
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: callback.C,v 2.1 1995/05/19 18:51:00 davidsi Exp $
//
// $Log: callback.C,v $
// Revision 2.1  1995/05/19  18:51:00  davidsi
// - Added CallbackCopy() to all callbacks to be copied.
//
// Revision 2.0  1995/01/19  21:53:19  davidsi
// - callback::add_callback now returns the created node, so it can
//   be subsequently removed.
// - The add_callback now also allows callbacks to be added which will
//   be passed a specific value as the first argument.
//
// Revision 1.1  1994/06/27  19:50:18  davidsi
// Initial revision
//
//
//


#include "stdio.h"

#include "localmacros.h"

#include "callback.h"

//
/////////////////////////////////////////////////////////////////////////////
//
// Functions for the  mcallback class
//
/////////////////////////////////////////////////////////////////////////////
//


//
// Constructor for mcallback class. Simply create a place holder 
// for the relevant information.
//
mcallback::mcallback( char *faddr,unsigned long naction)
{
	addr=faddr;
	action=naction;
}

//
/////////////////////////////////////////////////////////////////////////////
//
// Functions for handling the callback class
//
/////////////////////////////////////////////////////////////////////////////
//



//
// callback::CallbackAdd function
//
// Adds a function that is to be called when the callbacks are
// made for this member of the callback class.
//
// INPUTS :	function	The function to add to the callback list
// RETURNS:	A pointer to the newly created node
//
Callbackid
callback::CallbackAdd(  void (*newfunc)(char *,va_list ) )
{
	class callback_node *ptr=new class callback_node;

	ptr->f=newfunc;
	ptr->cval=NULL;
	insert_head(ptr);
	ptr->id=nextid++;
	return(ptr->id);
}


//
// callback::CallbackAdd function callback_value
//
// Adds a function that is to be called when the callbacks are
// made for this member of the callback class.
//
// INPUTS :	function	The function to add to the callback list
//		val		The value to be passed to the callback func.
// RETURNS:	A pointer to the newly created node
//
Callbackid
callback::CallbackAdd(  void (*newfunc)( char *, va_list ), char *val )
{
	class callback_node *ptr=new class callback_node;

	ptr->f=newfunc;
	ptr->cval=val;
	insert_head(ptr);
	ptr->id=nextid++;
	return(ptr->id);
}

//
// callback::CallbackAdd function callback_value id
//
// Adds a function that is to be called when the callbacks are
// made for this member of the callback class.  Note that this
// function is private to the class and expects the id to already
// allocate and valid.
//
// INPUTS :	function	The function to add to the callback list
//		val		The value to be passed to the callback func.
//		id		The id for the callback.
// RETURNS:	A pointer to the newly created node
//
void
callback::CallbackAdd(  void (*newfunc)( char *, va_list ), 
		char *val,
		Callbackid rid 
		)
{
	class callback_node *ptr=new class callback_node;

	ptr->f=newfunc;
	ptr->cval=val;
	ptr->id=rid;
	insert_head(ptr);
}



//
// callback::CallbackRemove cbackid
//
// Removes the specified callback from the chain
//
// INPUTS :	cbackid		The callback to remove
// RETURNS:	None
//
void
callback::CallbackRemove(Callbackid remid)
{
	class callback_node *ptr;
	class callback_node *nptr;

	ptr=head();
	while (! eol(ptr) )  {
		nptr=ptr->next();
		if (ptr->id==remid)
			ptr->remove();
		ptr=nptr;
	}
}


//
// callback::CallbackMake dummy_var arguments
//
// Calls ALL the stored functions.  The dummy argument is simply a place
// holder for varargs though it is recommended to set it to 0.  The
// following arguments can be any valid C arguments.
//
// INPUTS :	dummy_var	Not used - only identifies position on stack
//		arguments	The arguments to pass to all the functions
//				being called back.
// RETURNS:	None
//
void
callback::CallbackMake( int num, ... )
{
	class callback_node *ptr;
	va_list ap;
	va_start( ap, num);

	ptr=head();
	while (! eol(ptr) )  {
		ptr->f(ptr->cval, ap );
		ptr=ptr->next();
	}
}


//
// callback::CallbackCopy cback
//
// Copies the callbacks from instance to another without deleting
// any that might already be in the destination instance.
//
// INPUTS :	cback		The source instance
// RETURNS:	None
//
void
callback::CallbackCopy(class callback &cback) 
{
	class callback_node *ptr;

	ptr=cback.head();
        while (! cback.eol(ptr) )  {
		CallbackAdd(ptr->f,ptr->cval,ptr->id);
		ptr=ptr->next();
        }
}

//
// Constructor
//
callback::callback()
{
	nextid=0;
}

//
// Destructor
//
callback::~callback()
{
}




//
// Function to save casts
//
class callback_node *
callback_node::next()
{
	return( LISTREMAP(callback_node,list_node,next,this) );
}
