
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: callback.h,v 2.2 1995/05/18 21:00:33 davidsi Exp $
//
// $Log: callback.h,v $
// Revision 2.2  1995/05/18  21:00:33  davidsi
// - Added head() function to callback class to prevent the need for
//   explicit casts.
// - Added CallbackCopy() to all callbacks to be copied.
// - Using pointers to callback nodes, creates problems when callbacks
//   are duplicated and the original callback list is deleted. Better
//   to use opaque handles, so added Callbackid type and this is now
//   returned by CallbackAdd (ne callback_node).
//
// Revision 2.1  1995/02/27  22:09:30  davidsi
// - mcallback instances can now be constructed without passing arguments.
//
// Revision 2.0  1995/01/19  21:52:51  davidsi
// - callback::add_callback now returns the created node, so it can
//   be subsequently removed.
// - The add_callback now also allows callbacks to be added which will
//   be passed a specific value as the first argument.
//
// Revision 1.1  1994/06/27  19:49:55  davidsi
// Initial revision
//
//
#ifndef __CALLBACK_H
#define __CALLBACK_H

#include "stdarg.h"
#include "list.h"
#include "localmacros.h"

//
// The mcallback class is a convenience class used for providing wrappers
// between C callbacks and C++ classes.  Construct members of this 
// class and then directly read addr and action in a C function to
// pass into the relevant C++ function.
//
// THIS CLASS IS *NOT* RELATED TO callback lists!
//
class mcallback {
public:
	char *addr;
	unsigned long action;

	mcallback() {};
	~mcallback() {};
	mcallback( char *addr , unsigned long);
};


//
// The callback identifier
//
typedef int Callbackid;


//
// The callback class
//
// The callback class allows lists of functions to be stored and
// for those functions to be called on demand.
//
class callback : public list {
	Callbackid nextid;

private:
        void CallbackAdd(  void (*f)( char *,va_list ),char *,Callbackid);

public:
        Callbackid CallbackAdd(  void (*f)( char *,va_list ) );
        Callbackid CallbackAdd(  void (*f)( char *,va_list ),char *val);
	void CallbackRemove(Callbackid);
	void CallbackMake(  int num, ... );
	void CallbackCopy(class callback &);
	class callback_node *head() {
                return( LISTREMAP(callback_node,list,head,this) );
        }

	callback();
	virtual ~callback();
};

class callback_node: public list_node {
	char *cval;		// value to be passed back
	void (*f)(char *, va_list );
	Callbackid id;
public:
	friend callback;
	//
	// Note that callback_nodes are removed by using the list
	// remove function member!
	//

	//
        // These are to save casts
        //
        class callback_node *next();

};
#endif
