
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: cmd.C,v 2.10 1995/06/17 20:50:45 davidsi Exp $
//
// $Log: cmd.C,v $
// Revision 2.10  1995/06/17  20:50:45  davidsi
// - Added basic operator functionality to the value_date class, so that
//   dates can be added, compared (etc).
// - Added basic operator functionality to the value_sec class.
// - Renamed the getsecurity function to dialogsecurity - which is
//   consistent with the other dialog commands that the macro
//   language provides.
// - Added dialogdaterange to allow input of date ranges.
// - Added date2days to allow dates to be converted to days.
// - Added isaprice() to allow checks to be made on the validity
//   of a price.
// - Added the NOPRICE constant which is guaranteed not to be a valid
//   price.
//
// Revision 2.9  1995/06/02  19:59:16  davidsi
// - BUG FIX.  Many of the functions were still writing strings to the
//   error stream and not using the enumerated type. Also removed remaining
//   printf() calls that were writing error information.
//
// Revision 2.8  1995/05/19  18:51:10  davidsi
// - BUG FIX. getprices() must check that the security pointer is valid,
//   otherwise it can indirect off NULL.
//
// Revision 2.7  1995/04/25  21:13:02  davidsi
// - Add name() function member to value and all value derived classes. This
//   identifies the type using an ascii string - great for debugging.
// - Remove release_storage() function member - it is not needed.
// - BUG FIX. In xsmbuiltin() the additional types should only be created
//   once, otherwise they will have multiple ids attached.
//
// Revision 2.6  1995/04/16  20:27:26  davidsi
// - Renamed the menu command to mainmenu, menu can now be used for creating
//   menus on user created windows.
//
// Revision 2.5  1995/04/03  20:42:05  davidsi
// - Moved all window handling functions to a new file "cmd_window.C".
//
// Revision 2.4  1995/03/08  21:25:12  davidsi
// - The extract() member of security was renamed to SecurityExtract(), so
//   so this module modified to reflect that.
// - No longer need extern declarations of main_sec_list and main_price_list
//   as the relevant include files provide these.
//
// Revision 2.3  1995/02/27  22:03:30  davidsi
// - Removed the dplot.h include since this file does not require it, instead
//   include drawdialog.h - which it does need!
// - Revised the usage of Drawdialog to reflect the ongoing changes to that
//   said class.
// - Uses dialog independent code.
//
// Revision 2.2  1995/02/05  17:58:25  davidsi
// - References to drawdialog are now Drawdialog (straight name change).
// - Support functions have now moved from the system specific suplib so
//   now include support.h.
//
// Revision 2.1  1995/01/31  22:16:08  davidsi
// - Many changes to reflect name changes (etc) to other classes.
//
// Revision 2.0  1995/01/19  22:01:52  davidsi
// *** empty log message ***
//
//
//


#include <stdio.h>
#include <varargs.h>
#define _STDARG_H
#include <stdlib.h>
#include <dial.h>

#include "localmacros.h"
#include "support.h"
#include "security.h"
#include "menu.h"
#include "node.h"
#include "builtin.h"
#include "error.h"

#include "date.h"
#include "price.h"
#include "setdate.h"
#include "drawdialog.h"

extern menu_bar mb;

extern errstream cerror;
extern diagstream cinfo;

//
// menu_env is used to bind function names and code modules so that
// the appropriate user function can be triggered at the right time.
//
struct menu_env {
	char *fncname;
	class node_header *node;
};




//
/////////////////////////////////////////////////////////////////////////////
//
// The value_date class provides users with a basic date type
//
/////////////////////////////////////////////////////////////////////////////
//
int value_date_id=0;
class value_date *DATE_VALUE;

class value_date : public value  {
	class date *dateptr;
	class date tmpdate;
public:
	virtual int size();
	virtual void to_mem(char *mem);
	virtual void from_mem(char *mem);
	virtual void assign_storage(char *mem);
	virtual char *fetch_storage();
	virtual char *name();
	virtual class value *clone();

        virtual int load(char *&);
	virtual class value *operation(operators_e,class value *);

	value_date();
	~value_date();
};


int
value_date::size()
{
	return( sizeof(class date) );
}

void 
value_date::to_mem(char *mem)
{
#ifdef ASSERTS
	if (dateptr==NULL) {
		cerror << internal_address_not_set << " (value_date)\n";
		exit(1);
	}
#endif
	*((class date *)mem)=*dateptr;
}

void
value_date::from_mem(char *mem)
{
#ifdef ASSERTS
	if (dateptr==NULL) {
		cerror << internal_address_not_set << " (value_date)\n";
		exit(1);
	}
#endif
	*dateptr=*((class date *)mem);
}

void
value_date::assign_storage(char *mem)
{
	if (mem)
		dateptr=(class date *)mem;
	else {
		// date=new class date; 
		//
		// Copy the current value, if one is set
		//
		if (dateptr)
			tmpdate=*dateptr;
		dateptr=&tmpdate;
	}
}

char *
value_date::fetch_storage()
{
#ifdef ASSERTS
	if (dateptr==NULL) {
		cerror << internal_address_not_set << " (value_date)\n";
		exit(1);
	}
#endif
	return((char *)dateptr);
}

char *
value_date::name()
{
	return("date");
}

class value *
value_date::clone()
{
	class value_date *val=new class value_date;

	val->dateptr=&(val->tmpdate);
	val->tmpdate.setdate(*dateptr);
	return((class value *)val);
}

int
value_date::load(char *&str)
{
	if (dateptr==NULL)
		str="No date set";
	else
		str=dateptr->getdate();
	return(0);
}


class value *
value_date::operation(operators_e op,class value *v2)
{
	class date *d1,*d2;
	class value_date *va;
	class value_int *vi;
	char va_alloc=0;
	int days_to_add;
	int comp;

	//
	// If the second argument is an integer then we assume it
	// represents days:
	//
	if (v2->value_flag==VALUE_INT ) {
		int v;

		if (v2->load(days_to_add) == -1) {
			cerror << type_mismatch << "\n";
			return(NULL);
		}
		if (status==temporary) {
			va=this;
		} else {
			va=(class value_date *)clone();
			va_alloc=1;
		}
		switch(op) {
       	 	case op_add:(va->dateptr)->add(days_to_add);
			return(va);
        	case op_sub:(va->dateptr)->sub(days_to_add);
			return(va);
		}
		if (va_alloc)
			delete va;
		cerror << bad_operation ;
		return(NULL);
	}

	//
	// Basic add and subtract is allowed on dates
	//
	if (op==op_add || op==op_sub ) {
		//
		// If the value is temporary, then use it.  Otherwise
		// a temporary node must be created. 
		// 
		if (status==temporary) {
			va=this;
		} else {
			va=(class value_date *)clone();
			va_alloc=1;
		}

		//
		// As we are about to modify the value, point to a temporary
		// workarea.
		//
		va->assign_storage(NULL);
		d1=((class value_date *)va)->dateptr;
		d2=((class value_date *)v2)->dateptr;

		switch(op) {
		case op_add:d1->add(d2);
			return(va);
        	case op_sub:d1->sub(d2);
			return(va);
		}
		if (va_alloc) 
			delete va;
	}
	//
	// Going to return an int because we will be comparing two
	// date values.  Build that int ready for the return.
	//
	vi=new class value_int;
	vi->assign_storage(NULL);
	va=(class value_date *)v2;
	comp=dateptr->compare(va->dateptr);

	switch(op) {
	case comp_eq:
		vi->store( comp==0 );
		return(vi);
	case comp_ne:
		vi->store( comp!=0 );
		return(vi);
	case comp_lt:
		vi->store( comp<0 );
		return(vi);
	case comp_le:
		vi->store( comp<=0 );
		return(vi);
	case comp_gt:
		vi->store( comp>0 );
		return(vi);
	case comp_ge:
		vi->store( comp>=0 );
		return(vi);
	}
	cerror << bad_operation ;

	delete vi;

	return(NULL);
}

value_date::value_date()
{
	dateptr=NULL;
	value_flag=value_date_id;
}

value_date::~value_date()
{
}



//
/////////////////////////////////////////////////////////////////////////////
//
// The value_sec class provides users with a security based type
//
/////////////////////////////////////////////////////////////////////////////
//
int value_sec_id=0;
class value_sec *SEC_VALUE;

class value_sec : public value  {
	class security **secptr;
	class security *sec;
public:
	virtual int size();
	virtual void to_mem(char *mem);
	virtual void from_mem(char *mem);
	virtual void assign_storage(char *mem);
	virtual char *fetch_storage();
	virtual char *name();
	virtual class value *clone();
	virtual class value *operation(operators_e,class value *);

        virtual int load(char *&);

	value_sec();
	~value_sec();
};


int
value_sec::size()
{
	return( sizeof(*secptr) );
}

void 
value_sec::to_mem(char *mem)
{
#ifdef ASSERTS
	if (secptr==NULL) {
		cerror << internal_address_not_set << " (value_sec)\n";
		exit(1);
	}
#endif
	*((class security **)mem)=*secptr;
}

void
value_sec::from_mem(char *mem)
{
#ifdef ASSERTS
	if (secptr==NULL) {
		cerror << internal_address_not_set << " (value_sec)\n";
		exit(1);
	}
#endif
	*secptr=*((class security **)mem);
}

void
value_sec::assign_storage(char *mem)
{
	if (mem)
		secptr=(class security **)mem;
	else {
		// date=new class date; 
		//
		// Copy the current value, if one is set
		//
		if (secptr)
			sec=*secptr;
		secptr=&sec;
	}
}

char *
value_sec::fetch_storage()
{
#ifdef ASSERTS
	if (secptr==NULL) {
		cerror << internal_address_not_set << " (value_sec)\n";
		exit(1);
	}
#endif
	return((char *)secptr);
}

char *
value_sec::name()
{
	return("security");
}

class value *
value_sec::clone()
{
	class value_sec *val=new class value_sec;

	if (secptr) {
		fprintf(stderr,"PANIC: Cannot clone a value_sec once assigned\n");
		exit(1);
	}
	val->secptr=secptr;
	return((class value *)val);
}

int
value_sec::load(char *&str)
{

	if (secptr==NULL || *secptr==NULL)
		str="No security";
	else
		(*secptr)->SecurityExtract(&str,NULL,NULL,NULL,NULL);
	return(0);
}

class value *
value_sec::operation(operators_e op,class value *v2)
{
	class value_int *va;
	class value_sec *vb;

	//
	// Allows a comparison aganst an integer so a user can judge if
	// a security has been defined.
	//
	if (v2->value_flag==VALUE_INT ) {
		int v;

		v2->load(v);
		va=(class value_int *)v2->clone();
		va->store(v ==(int)(*secptr) );
		return(va);
	}

	//
	// Do we have type mismatch problems?
	//
	if (value_flag != v2->value_flag) {
		//
		// Otherwise type mismatch error
		//
		cerror << type_mismatch;
		return(NULL);
	}

	va=new class value_int;
	vb=(class value_sec *)v2;

	//
	// As we are about to modify the value, point to a temporary
	// workarea.
	//
	va->assign_storage(NULL);
	

	switch(op) {
	        case comp_eq:
			va->store( (*(vb->secptr)==*(secptr)) );
			return(va);
		case comp_ne:
			va->store( (*(vb->secptr)!=*(secptr)) );
			return(va);
	}
	cerror << bad_operation ;

	va->dispose();

	return(NULL);
}


value_sec::value_sec()
{
	secptr=NULL;
	value_flag=value_sec_id;
}

value_sec::~value_sec()
{
}

//
/////////////////////////////////////////////////////////////////////////////
//
// Commands based on the security and date types
//
/////////////////////////////////////////////////////////////////////////////
//


//
// Converts date strings to internal date types
//
// getdate(date,"date string");
// getdate(date);
//
static int
getdate(class node_builtin &bptr)
{
	class date *newdate;
	int out;

	if (bptr.number_of_arguments()!=2 && bptr.number_of_arguments()!=1) {
		cerror << incorrect_num_arguments << "\n";
		return(-1);
	}
	newdate=new date;
	
	if (bptr.number_of_arguments()==2) {
		char *datestring;

        	if ( bptr[0]->load(datestring) == -1 ) {
			cerror << not_string << "\n";
			delete newdate;
			return( -1 );
		}
		(void)newdate->setdate(datestring);
		out=1;
	} else
		out=0;
	
	bptr[out]->from_mem((char *)newdate);
	return(0);
}


//
// Returns price data for securities
//
// getprices(security, array, startdate,number)
//
static int
getprices(class node_builtin &bptr)
{
	class security *sec;

	double *real_array;
	double *parray;
        class date date;
	unsigned long number;


	int i;

	if (bptr.number_of_arguments()!=4) {
		cerror << incorrect_num_arguments << "\n";
		return(-1);
	}
        if ( !(bptr[0]->type_equal((class value *)SEC_VALUE)) ) {
		cerror << type_mismatch << " getprices did not locate security";
                return(-1);
        }
	bptr[0]->to_mem((char *)&sec);

	if (sec==NULL) {
		cerror << not_a_valid_security << "\n";
                return(-1);
	}

	sec->prices->extract(&parray,&date,&number);

	//
	// Missing type check TODO TO DO MISSING
	//
	real_array=(double *)bptr[1]->fetch_storage();
	memcpy(real_array,parray,sizeof(double)*number);

	bptr[2]->from_mem((char *)&date);

        if ( bptr[3]->store(number) == -1 ) {
		cerror << not_integer << "\n";
                return(-1);
        }
	return(0);
}

//
// Prompts the user for a security name
//
// security=dialogsecurity(title)
//
static
dialogsecurity(class node_builtin &bptr)
{
	char *str1;
	class security *newsec;
	class value_sec *sec;

	if (bptr.number_of_arguments()!=1) {
		cerror << incorrect_num_arguments << "\n";
		return(-1);
	}
	if ( bptr[0]->load(str1) == - 1) {
		cerror << not_string << "\n";
		return(-1);
	}

	newsec=main_sec_list->select_security(str1);

	sec=new class value_sec;
	sec->assign_storage(NULL);
	sec->from_mem((char *)&newsec);
	bptr.SetReturn(sec);
	return(0);
}



//
// Handles menu callbacks
//
static void
menucall(local_va_dcl)
{
va_list ap;
dialog d;
	struct menu_env *me;

        local_va_start(ap);
        d=va_arg(ap,dialog);
        me=va_arg(ap,struct menu_env *);

	me->node->execute(me->fncname);
}


//
// Creates a menu item on the main xsm window
//
// menu("name of menu","menu item","function to call");
//
static
menuitem(class node_builtin &bptr)
{
	char *barname,*itemname,*fname;
	class menu_item *mi; 
	struct menu_env *me;
	static char maskset=0;
	static class value *argmask[3];

	if (!maskset) {
		argmask[0]=STRING_VALUE;
		argmask[1]=STRING_VALUE;
		argmask[2]=STRING_VALUE;
		maskset=1;
	}

	if (bptr.types_of_arguments(3,argmask)==-1)
		return(-1);

	if ( (bptr[0]->load(barname) == -1) ||
	     (bptr[1]->load(itemname) == -1) || 
	     (bptr[2]->load(fname) == -1) ) {
		cerror << not_string << "\n";
		return(-1);
	}

	me=(struct menu_env *)malloc(sizeof(struct menu_env));
	me->fncname=(char *)malloc(strlen(fname)+1);
	strcpy(me->fncname,fname);
	me->node=bptr.base();

	mi=new menu_item(&mb,barname,itemname,menucall,(int)me);

	//
	// Set up the menu bar..again...
	//
	DialogRootmenu(mb);

	return(0);
}


//
// Displays a date range input dialog
//
// status=dialogdaterange(title,date1string,date1,date2string,date2);
//
static int
dialogrdate( node_builtin &bptr )
{
	char *title;
	char *datetitle1;
	char *datetitle2;
	class date *date1;
	class date *date2;
	int ret;
	class value_int *argr;
 	static char maskset=0;
	static class value *argmask[5];

	if (!maskset) {
		argmask[0]=STRING_VALUE;
		argmask[1]=STRING_VALUE;
		argmask[2]=DATE_VALUE;
		argmask[3]=STRING_VALUE;
		argmask[4]=DATE_VALUE;
		maskset=1;
	}
	if (bptr.types_of_arguments(5,argmask)==-1)
		return(-1);
	if ( (bptr[0]->load(title) == -1) || 
	     (bptr[1]->load(datetitle1) == -1) || 
	     (bptr[3]->load(datetitle2) == -1) ) {
		cerror << not_string << "\n";
		return(-1);
	}
	date1=(class date *) (((class value_date *)(bptr[2]))->fetch_storage());
	date2=(class date *) (((class value_date *)(bptr[4]))->fetch_storage());
	ret=DialogSelect2dates(title,datetitle1,*date1,datetitle2,*date2);
	argr=new class value_int;
	argr->assign_storage(NULL);
	argr->store(ret);
	bptr.SetReturn(argr);
	return(0);
}


//
// Converts a date to a number of days since jan 1 1970
//
// integer=date2days( date )
//
static int
date2days( node_builtin &bptr )
{
	class date *date1;
	int ret;
	class value_int *argr;
 	static char maskset=0;
	static class value *argmask[1];

	if (!maskset) {
		argmask[0]=DATE_VALUE;
		maskset=1;
	}
	if (bptr.types_of_arguments(1,argmask)==-1)
		return(-1);

	date1=(class date *) (((class value_date *)(bptr[0]))->fetch_storage());

	argr=new class value_int;
	argr->assign_storage(NULL);
	argr->store( date1->daysince1970() );
	bptr.SetReturn(argr);
	return(0);
}

//
// Returns whether a real value is a valid price
//
// isaprice(real)
//
static int
isaprice(class node_builtin &bptr)
{
	class value_int *retval;
	double price;

	if (bptr.number_of_arguments()<1) {
		cerror << incorrect_num_arguments << "\n";
		return(-1);
	}

	if (bptr[0]->load(price) == -1) 
		return(-1);

	retval=new class value_int;
	retval->assign_storage(NULL);
	retval->store(IS_PRICE(price) );
	bptr.SetReturn(retval);

        return(0);
}



//
// Provides a system shell out
//
// system(string)
//
static int
systemshell(class node_builtin &bptr)
{
	class value *val;
	class value_int *retval;
	char *string;

	if (bptr.number_of_arguments()<1) {
		cerror << incorrect_num_arguments << "\n";
		return(-1);
	}

	if (bptr[0]->load(string) == -1) 
		return(-1);

	retval=new class value_int;
	retval->assign_storage(NULL);
	retval->store(system(string) );
	bptr.SetReturn(retval);

        return(0);
}
//
// This function will be called each time a program is loaded and
// "compiled".  It should initialize all the necessary entries
// in the program's symbol table.
//
void
xsmbuiltin(class context *cont)
{
	extern int allocate_type_id();
	static char create_types=1;
	class value_real *vr;

	//
	// Create pre-defined variables
	//
	vr=new class value_real;
	vr->assign_storage(NULL);
	vr->store(NOT_A_PRICE);
	cont->ContextAddconstant("NOPRICE",vr);


	new node_builtin(cont,"mainmenu",menuitem);
	new node_builtin(cont,"dialogsecurity",dialogsecurity);
	new node_builtin(cont,"system",systemshell);

	//
	// Date functions
	//
	new node_builtin(cont,"getdate",getdate);
	new node_builtin(cont,"dialogdaterange",dialogrdate);
	new node_builtin(cont,"date2days",date2days);

	//
	// Price handling
	//
	new node_builtin(cont,"getprices",getprices);
	new node_builtin(cont,"isaprice",isaprice);
	

	//
	// New types
	//
	if ( create_types ) {
		value_date_id=allocate_type_id();
		DATE_VALUE=new class value_date;

		value_sec_id=allocate_type_id();
		SEC_VALUE=new class value_sec;
		create_types=0;
	}
	cont->addtype("DATE",new class value_date);
	cont->addtype("SECURITY",new class value_sec);
}
