
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: cmd_window.C,v 2.5 1995/06/17 20:45:44 davidsi Exp $
//
// $Log: cmd_window.C,v $
// Revision 2.5  1995/06/17  20:45:44  davidsi
// - Added dialogyesno, dialogstring and dialogmessage commands.
//
// Revision 2.4  1995/06/02  20:01:46  davidsi
// - BUG FIX.  Some of the sections of code for ASSERTS were incorrect and
//   out of date.
// - BUG FIX.  Many of the functions were still writing strings to the
//   error stream and not using the enumerated type. Also removed remaining
//   printf() calls that were writing error information.
// - BUG FIX. The menu handling functionality would not actually create
//   all the instances of the menu classes resulting in indeterministic
//   behaviour.
//
// Revision 2.3  1995/05/19  18:51:55  davidsi
// - Tidying up and add GNU copyright message.
//
// Revision 2.2  1995/04/25  21:12:45  davidsi
// - Add name() function member to value and all value derived classes. This
//   identifies the type using an ascii string - great for debugging.
// - BUG FIX. In xsmbuiltin() the additional types should only be created
//   once, otherwise they will have multiple ids attached.
//
// Revision 2.1  1995/04/16  20:27:42  davidsi
// - Many changes.  The previous window mechanism had many failings.  The
//   new approach is to use a new Window class that handles the window and
//   the Windowvalue class indirects to that class.
// - Prevent an automatic "Close" menu item.  This is dangerous because
//   the window can close unknown to the user's program.
// - The window command is now windowopen and it returns a window reference
//   rather than using a third parameter. This makes programs much easier to
//   read.
// - Added a windowclose command.
//
// Revision 2.0  1995/04/03  20:44:16  davidsi
// - Initial version.
//
//
//


#include <stdio.h>
#include <varargs.h>
#define _STDARG_H
#include <stdlib.h>
#include <dial.h>

#include "localmacros.h"
#include "support.h"
#include "security.h"
#include "menu.h"
#include "node.h"
#include "builtin.h"
#include "error.h"

#include "date.h"
#include "drawdialog.h"

extern class value_date *DATE_VALUE;

extern errstream cerror;
extern diagstream cinfo;


//
// menu_env is used to bind function names and code modules so that
// the appropriate user function can be triggered at the right time.
//
struct menu_env {
	char *fncname;
	class node_header *node;
};

//
// Variables used throughout this module
//
static int Valuewindow_id=0;
static class Valuewindow *WINDOW_VALUE;
static int Valuemenu_id=0;
static class Valuemenu *MENU_VALUE;

//
/////////////////////////////////////////////////////////////////////////////
//
// The Window and Valuewindow classes provide users with a window type
//
/////////////////////////////////////////////////////////////////////////////
//


//
// Note the window functionality is issolated from the value functionality
// because this needs to inherit from elsewhere and override virtual
// functions.
//
class Window : public Drawdialog 
{
	struct menu_env me;
private:
        virtual void DrawdialogResize();
public:
	void WindowCreate(struct menu_env &mptr,drawd_reg_t &drt);
	Window();
	~Window();
};


//
// Window::DrawdialogResize
//
// On a resize call the user's resize handler
//
void
Window::DrawdialogResize()
{
	if (me.fncname!=NULL)
		me.node->execute(me.fncname);
}


//
// Window::WindowCreate menv drawreg
//
// Create a window 
//
// INPUTS :	menv		The callback info for resizes
//		drawreg		The spec for the window
// RETURNS:	None
//
void
Window::WindowCreate(struct menu_env &mptr,drawd_reg_t &drt)
{
	me=mptr;

	//
	// By setting fncname to NULL we ensure the callback won't 
	// happen.  We cannot let it run because the window we are
	// creating here won't yet be assigned to a variable in the
	// user's program so there is no way the user can actually
	// do anything sensible in the callback.
	//
	me.fncname=NULL;
	DrawdialogCreate(&drt);
	DrawStartpage();
	//
	// Copy the function name and allocate space for it
	//
	me.fncname=(char *)malloc(strlen(mptr.fncname)+1);
	strcpy(me.fncname,mptr.fncname);
}


//
// Constructor for the window class
//
Window::Window()
{
	me.fncname=NULL;
}

//
// Destructor for the window class
//
Window::~Window()
{
	if (me.fncname)
		free(me.fncname);
}


class Valuewindow : public value
{
	Window **dd;
	//
	// The actual member is only used when building a window.  At 
	// that point no heap space is known (or necessarily allocated) 
	// so we need to point dd to somewhere safe until the assignment
	// that usually follows window creation happens.
	//
	Window *actual;
	struct menu_env me;
public:
	class Window *ValuewindowGimmeWindow() {
		if (dd==NULL) 
			DialogPanic("ValuewindowGimmeWindow - no draw dialog");
		return (*dd);
	}
	virtual int size();
	virtual void to_mem(char *mem);
	virtual void from_mem(char *mem);
	virtual void assign_storage(char *mem);
	virtual char *fetch_storage();
	virtual class value *clone();
	virtual char *name();
	//
	// These are somewhat unusual for a value derived class
	//
	void ValuewindowCreate(struct menu_env &mptr,drawd_reg_t &drt);

	Valuewindow();
	~Valuewindow();
};


static class Window *
GetWindow(class value *val)
{
	class Valuewindow *windowval;
	class Window *window;

	if ( !(val->type_equal((class value *)WINDOW_VALUE)) ) {
		cerror << type_mismatch << "\n";
		return(NULL);
	}
        windowval=(class Valuewindow *)val;
	window=windowval->ValuewindowGimmeWindow();
	return(window);
}



int
Valuewindow::size()
{
	return( sizeof(*dd) );
}

void 
Valuewindow::to_mem(char *mem)
{
#ifdef ASSERTS
	if (dd==NULL) {
		cerror << internal_address_not_set << " (Valuewindow)\n";
		exit(1);
	}
#endif
	*((class Window **)mem)=*dd;
}

void
Valuewindow::from_mem(char *mem)
{
#ifdef ASSERTS
	if (dd==NULL) {
		cerror << internal_address_not_set << " (Valuewindow)\n";
		exit(1);
	}
#endif
	*dd=*((class Window **)mem);
}

void
Valuewindow::assign_storage(char *mem)
{
	dd=(class Window **)mem;
}

char *
Valuewindow::fetch_storage()
{
#ifdef ASSERTS
	if (dd==NULL) {
		cerror << internal_address_not_set << " (Valuewindow)\n";
		exit(1);
	}
#endif
	return((char *)dd);
}

char *
Valuewindow::name()
{
	return( "window" );
}

class value *
Valuewindow::clone()
{
	class Valuewindow *val=new class Valuewindow;


	val->dd=dd;

	return((class value *)val);
}

void
Valuewindow::ValuewindowCreate(struct menu_env &mptr,drawd_reg_t &drt)
{
	class Window *w;

	w=*dd=new class Window();
	w->WindowCreate(mptr,drt);
}


Valuewindow::Valuewindow()
{
	//
	// At this point no heap space is known (or necessarily allocated)
	// so we need to point dd to somewhere safe until the assignment
	// that usually follows window creation happens.
	//
	dd=&actual;
	actual=NULL;
	value_flag=Valuewindow_id;
}

Valuewindow::~Valuewindow()
{
}

//
//////////////////////////////////////////////////////////////////////////////
//
// The MenuCmdlist and Valuemenu classes provide users with a menu type
//
//////////////////////////////////////////////////////////////////////////////
//

//
// MenuCmdlist is a container for <node,function name> pairs that allows
// all resources to be freed up when the MenuCmd instance is deleted.
//
class MenuCmdlist : public list_node {
	char *fncname;
	class node_header *node;

	//
	//
	friend class MenuCmd;
public:
	MenuCmdlist(class node_header *,char *fncname);
	virtual ~MenuCmdlist();
};


//
// Constructor
//
MenuCmdlist::MenuCmdlist(class node_header *inode,char *ifncname) 
{
	node=inode;
	fncname=(char *)malloc(strlen(ifncname)+1);
	strcpy(fncname,ifncname);
}


//
// Destructor
//
MenuCmdlist::~MenuCmdlist() 
{
	free(fncname);
}


class MenuCmd : public Menu
{ 
	class list menucmdlist;
private:
        virtual void MenuCallback(unsigned long code, int onflag);
public:
	void MenuCmdCreate(char *bname,char *iname,struct menu_env &mptr);
	MenuCmd();
	virtual ~MenuCmd();
};


void
MenuCmd::MenuCallback(unsigned long code, int onflag)
{
	struct MenuCmdlist *me=(class MenuCmdlist *)code;

	if (me->fncname!=NULL)
		me->node->execute(me->fncname);
}

void
MenuCmd::MenuCmdCreate(char *bname,char *iname,struct menu_env &mptr)
{
	class MenuCmdlist *mcl=new class MenuCmdlist( mptr.node,mptr.fncname);

	menucmdlist.insert_tail(mcl);
	MenuItemaction(bname,iname,(unsigned long)mcl);
}

//
// Constructor
//
MenuCmd::MenuCmd()
{
}

//
// Destructor
//
MenuCmd::~MenuCmd()
{
}



class Valuemenu : public value
{
	MenuCmd **menuptr;
	MenuCmd *actual;
public:
	class MenuCmd *ValuemenuGimmeMenuCmd() {
		return (*menuptr);
	}
	void ValuemenuResetMenuCmd() {
		*menuptr=NULL;
	}
	virtual int size();
	virtual void to_mem(char *mem);
	virtual void from_mem(char *mem);
	virtual void assign_storage(char *mem);
	virtual char *fetch_storage();
	virtual char *name();
	virtual class value *clone();
	//
	// These are somewhat unusual for a value derived class
	//
	void ValuemenuCreate(char *bname,char *iname,struct menu_env &mptr);

	Valuemenu();
	~Valuemenu();
};


int
Valuemenu::size()
{
	return( sizeof(*menuptr) );
}

void 
Valuemenu::to_mem(char *mem)
{
#ifdef ASSERTS
	if (menuptr==NULL) {
		cerror << internal_address_not_set << " (Valuemenu)\n";
		exit(1);
	}
#endif
	*((class MenuCmd **)mem)=*menuptr;
}

void
Valuemenu::from_mem(char *mem)
{
#ifdef ASSERTS
	if (menuptr==NULL) {
		cerror << internal_address_not_set << " (Valuemenu)\n";
		exit(1);
	}
#endif
	*menuptr=*((class MenuCmd **)mem);
}

void
Valuemenu::assign_storage(char *mem)
{
	menuptr=(class MenuCmd **)mem;
}

char *
Valuemenu::fetch_storage()
{
#ifdef ASSERTS
	if (menuptr==NULL) {
		cerror << internal_address_not_set << " (Valuemenu)\n";
		exit(1);
	}
#endif
	return((char *)menuptr);
}

char *
Valuemenu::name()
{
	return( "menu" );
}

class value *
Valuemenu::clone()
{
	class Valuemenu *val=new class Valuemenu;

	val->menuptr=menuptr;

	return((class value *)val);
}

void
Valuemenu::ValuemenuCreate(char *bname,char *iname,struct menu_env &mptr)
{
	class MenuCmd *m;

	m=ValuemenuGimmeMenuCmd();
	if (m==NULL)
		m=*menuptr=new class MenuCmd;

	m->MenuCmdCreate(bname,iname,mptr);
}


Valuemenu::Valuemenu()
{
	//
	// At this point no heap space is known (or necessarily allocated)
	//
	menuptr=&actual;
	actual=NULL;
	value_flag=Valuemenu_id;
}

Valuemenu::~Valuemenu()
{
}



//
//////////////////////////////////////////////////////////////////////////////
//
// The commands
//
//////////////////////////////////////////////////////////////////////////////
//





//
// Plots a graph
//
// plot(window, data,number of element in data,start date, end date);
//
static int
plot(class node_builtin &bptr)
{
	class Window *window;
	double *array;
	int ndata,startd,endd;

	int i;

	if (bptr.number_of_arguments()!=5) {
		cerror << incorrect_num_arguments << "\n";
		return(-1);
	}

	if ( (window=GetWindow(bptr[0]))==NULL )
		return(-1);

	//
	// Missing tyupe check TODO TO DO MISSING
	//
	array=(double *)bptr[1]->fetch_storage();

        if ( bptr[2]->load(ndata) == -1 ) {
		cerror << not_integer << "\n";
                return(-1);
	}
        if ( bptr[3]->load(startd) == -1 ) {
		cerror << not_integer << "\n";
                return(-1);
        }
        if ( bptr[4]->load(endd) == -1 ) {
		cerror << not_integer << "\n";
                return(-1);
        }
	window->DrawGraph(array,ndata,startd,endd);
	return(0);
}


//
// Plots an axis with dates on the x and numbers on the y
//
// axis_date(window,start date, end date, y low, y high)
//
static int
axis_date(class node_builtin &bptr)
{
	class Window *window;
	class date *d1,*d2;
	double min,max;


	if (bptr.number_of_arguments()!=5) {
		cerror << incorrect_num_arguments << "\n";
		return(-1);
	}
	if ( (window=GetWindow(bptr[0]))==NULL )
		return(-1);

        if ( !(bptr[1]->type_equal((class value *)DATE_VALUE)) ) {
		cerror << type_mismatch << "\n";
                return(-1);
        }
        if ( !(bptr[2]->type_equal((class value *)DATE_VALUE)) ) {
		cerror << type_mismatch << "\n";
                return(-1);
        }
	d1=(class date *)bptr[1]->fetch_storage();
	d2=(class date *)bptr[2]->fetch_storage();

        if ( bptr[3]->load(min) == -1 ) {
		cerror << not_real << "\n";
                return(-1);
        }
        if ( bptr[4]->load(max) == -1 ) {
		cerror << not_real << "\n";
                return(-1);
        }
	window->DrawSetfont("Helvetica",12);
	window->DrawAxis(d1,d2,min,max);
	return(0);
}


//
// Specifies a colour
//
// colour(window,"colour name")
//
static int
colour(class node_builtin &bptr)
{
	class Window *window;
	char *col;
	colour_e coltype;
	static char maskset=0;
	static class value *argmask[2];

	if (!maskset) {
		argmask[0]=WINDOW_VALUE;
		argmask[1]=NULL;
		maskset=1;
	}

	if (bptr.types_of_arguments(2,argmask)==-1)
		return(-1);

	if ( (window=GetWindow(bptr[0]))==NULL )
		return(-1);

        if ( bptr[1]->load(col) == -1 ) {
		cerror << not_string << "\n";
                return(-1);
        }

	coltype=string2colour(col);
	
	//
	// If the colour isn't set then default to white
	//
	if ( coltype==COLOUR_NOTSET)
		coltype=WHITE;

	window->DrawSetcolour(coltype);
}


//
// Specifies a line width
//
// linewidth(window,width)
//
static int
linewidth(class node_builtin &bptr)
{
	class Window *window;
	double width;
	static char maskset=0;
	static class value *argmask[2];

	if (!maskset) {
		argmask[0]=WINDOW_VALUE;
		argmask[1]=NULL;
		maskset=1;
	}

	if (bptr.types_of_arguments(2,argmask)==-1)
		return(-1);
	if ( (window=GetWindow(bptr[0]))==NULL )
		return(-1);

        if ( bptr[1]->load(width) == -1 ) {
		cerror << not_real << "\n";
                return(-1);
        }
	window->DrawSetlinewidth(width);
	return(0);
}



//
// Returns the current geometry of a window
//
// size(window,&width,&height)
// 
static int
size(class node_builtin &bptr)
{
	class Window *window;
	double xpos,ypos;
	double width,height;

	if (bptr.number_of_arguments()!=3) {
		cerror << incorrect_num_arguments << "\n";
		return(-1);
	}
        if ( !(bptr[0]->type_equal((class value *)WINDOW_VALUE)) ) {
		cerror << type_mismatch << "\n";
                return(-1);
        }
	if ( (window=GetWindow(bptr[0]))==NULL )
		return(-1);
	window->DrawExtract(width,height);
        if ( bptr[1]->store(width) == -1 ) {
		cerror << not_real << "\n";
                return(-1);
        }
        if ( bptr[2]->store(height) == -1 ) {
		cerror << not_real << "\n";
                return(-1);
        }
	return(0);
}


//
// Moves the current pen location
//
// moveto(window,x,y)
//
static int
moveto(class node_builtin &bptr)
{
	class Window *window;
	double xpos,ypos;

	if (bptr.number_of_arguments()!=3) {
		cerror << incorrect_num_arguments << "\n";
		return(-1);
	}
	if ( (window=GetWindow(bptr[0]))==NULL )
		return(-1);
        if ( bptr[1]->load(xpos) == -1 ) {
		cerror << not_real << "\n";
                return(-1);
        }
        if ( bptr[2]->load(ypos) == -1 ) {
		cerror << not_real << "\n";
                return(-1);
        }
	window->DrawMoveto(xpos,ypos);
}


//
// Draws a line from the current pen location to a new location
//
// drawto(window,x,y)
//
static int
drawto(class node_builtin &bptr)
{
	class Window *window;
	double xpos,ypos;

	if (bptr.number_of_arguments()!=3) {
		cerror << incorrect_num_arguments << "\n";
		return(-1);
	}
	if ( (window=GetWindow(bptr[0]))==NULL )
		return(-1);
        if ( bptr[1]->load(xpos) == -1 ) {
		cerror << not_real << "\n";
                return(-1);
        }
        if ( bptr[2]->load(ypos) == -1 ) {
		cerror << not_real << "\n";
                return(-1);
        }
	window->DrawDrawto(xpos,ypos);
}


//
// Flushes a window
//
// drawpage(window)
//
static int
drawpage(class node_builtin &bptr)
{
	class Window *window;
	static char maskset=0;
	static class value *argmask[1];

	if (!maskset) {
		argmask[0]=WINDOW_VALUE;
		maskset=1;
	}

	if (bptr.types_of_arguments(1,argmask)==-1)
		return(-1);

	if ( (window=GetWindow(bptr[0]))==NULL )
		return(-1);

	window->DrawPage();
}


//
// Clears a window
//
// clearpage(window)
//
static int
clearpage(class node_builtin &bptr)
{
	class Window *window;
	static char maskset=0;
	static class value *argmask[1];

	if (!maskset) {
		argmask[0]=WINDOW_VALUE;
		maskset=1;
	}

	if (bptr.types_of_arguments(1,argmask)==-1)
		return(-1);

	if ( (window=GetWindow(bptr[0]))==NULL )
		return(-1);

	window->DrawClearpage();
}


//
// Opens a window
//
// window=windowopen("title", "resize_function", menu);
//
static int
windowopen(class node_builtin &bptr)
{
	static char maskset=0;
	static class value *argmask[3];
	class Valuewindow *windowv;
	class Window *window;
	drawd_reg_t drt;
	char *title;
	char *funcname;
	struct menu_env me;

	if (!maskset) {
		argmask[0]=STRING_VALUE;
		argmask[1]=STRING_VALUE;
		argmask[2]=MENU_VALUE;
		maskset=1;
	}

	if ( bptr.number_of_arguments()==2 &&
	     bptr.types_of_arguments(2,argmask)!=-1)
		drt.mb=new menu_bar(NULL,NULL,0);
	else if (bptr.number_of_arguments()==3 &&
		 bptr.types_of_arguments(3,argmask)!=-1) {
		class MenuCmd *m;
		class Valuemenu *vm;
		vm=(class Valuemenu *)bptr[2];
		m=vm->ValuemenuGimmeMenuCmd();
		if (m)
			drt.mb=m->MenuExtract();
		else
			drt.mb=new menu_bar(NULL,NULL,0);
	} else
		return(-1);

        if ( ( bptr[0]->load(title) == -1 ) ||
	     ( bptr[1]->load(funcname) == -1 ) ) {
		cerror << not_string << "\n";
                return(-1);
        }

	me.fncname=funcname;
	me.node=bptr.base();

	// drt.flags=ALLOW_SIZE_REDRAW|INITIAL_SIZE_SET;
	drt.flags=ALLOW_SIZE_REDRAW|INITIAL_SIZE_SET|NOSCROLL|NOMENU;
        drt.width=360;
        drt.height=360;
        drt.title=title;

	windowv=new class Valuewindow;
	windowv->ValuewindowCreate(me,drt);

	bptr.SetReturn(windowv);
	return(0);
}

//
// Closes a window
//
// windowclose(window);
//
static int
windowclose(class node_builtin &bptr)
{
        class Window *window;

	if (bptr.number_of_arguments()!=1) {
		cerror << incorrect_num_arguments << "\n";
		return(-1);
	}

        if ( (window=GetWindow(bptr[0]))==NULL )
                return(-1);

	delete window;
	return(0);
}



//
// Creates a menu
//
// menuopen(menu)
//
static
menuopen(class node_builtin &bptr)
{
	class Valuemenu *vm; 
	static char maskset=0;
	static class value *argmask[1];

	if (!maskset) {
		argmask[0]=MENU_VALUE;
		maskset=1;
	}

	if ( bptr.types_of_arguments(1,argmask) == -1) 
		return(-1);

	vm=(class Valuemenu *)bptr[0];
	vm->ValuemenuResetMenuCmd();

	return(0);
}

//
// Closes a menu
//
// menuclose(menu)
//
static
menuclose(class node_builtin &bptr)
{
	class Valuemenu *vm; 
	class MenuCmd *m; 
	static char maskset=0;
	static class value *argmask[1];

	if (!maskset) {
		argmask[0]=MENU_VALUE;
		maskset=1;
	}

	if ( bptr.types_of_arguments(1,argmask) == -1) 
		return(-1);

	vm=(class Valuemenu *)bptr[0];
	m=vm->ValuemenuGimmeMenuCmd();
	if (m)
		delete m;
	vm->ValuemenuResetMenuCmd();
	return(0);
}


//
// Adds an item to a menu
//
// menu(menu,"menu name","item name","function to call");
//
static
menu(class node_builtin &bptr)
{
	char *barname,*itemname,*fname;
	class Valuemenu *vm; 
	struct menu_env me;
	static char maskset=0;
	static class value *argmask[4];

	if (!maskset) {
		argmask[0]=MENU_VALUE;
		argmask[1]=STRING_VALUE;
		argmask[2]=STRING_VALUE;
		argmask[3]=STRING_VALUE;
		maskset=1;
	}

	if (bptr.types_of_arguments(4,argmask)==-1)
		return(-1);

	if ( (bptr[1]->load(barname) == -1) ||
	     (bptr[2]->load(itemname) == -1) || 
	     (bptr[3]->load(fname) == -1) ) {
		cerror << not_string << "\n";
		return(-1);
	}

	me.fncname=fname;
	me.node=bptr.base();

	vm=(class Valuemenu *)bptr[0];
	vm->ValuemenuCreate(barname,itemname,me);
	return(0);
}


//
// Displays a yes/no dialog
//
// int=dialogyesno(message,textyes,textno)
//
static int
dialogyesno( node_builtin &bptr )
{
	char *str,*tstr,*fstr;
	int ret;
	class value_int *argr;
 	static char maskset=0;
	static class value *argmask[3];

	if (!maskset) {
		argmask[0]=STRING_VALUE;
		argmask[1]=STRING_VALUE;
		argmask[2]=STRING_VALUE;
		maskset=1;
	}
	if (bptr.types_of_arguments(3,argmask)==-1)
		return(-1);
	if ( (bptr[0]->load(str) == -1) || 
	     (bptr[1]->load(tstr) == -1) || 
	     (bptr[2]->load(fstr) == -1) ) {
		cerror << not_string << "\n";
		return(-1);
	}
	ret=DialogYesno(str,tstr,fstr);
	argr=new class value_int;
	argr->assign_storage(NULL);
	argr->store(ret);
	bptr.SetReturn(argr);
}


//
// Displays a string input dialog
//
// string=dialogstring(window_title, message, string_to_edit, length)
//
static int
dialogstring( node_builtin &bptr )
{
	char *wtitle,*title,*str;
	int len;
	int ret;
	class value_string *argr;
 	static char maskset=0;
	static class value *argmask[4];
	char buffer[1024];

	if (!maskset) {
		argmask[0]=STRING_VALUE;
		argmask[1]=STRING_VALUE;
		argmask[2]=STRING_VALUE;
		argmask[3]=INT_VALUE;
		maskset=1;
	}
	if (bptr.types_of_arguments(4,argmask)==-1)
		return(-1);
	if ( (bptr[0]->load(wtitle) == -1) || 
	     (bptr[1]->load(title) == -1) || 
	     (bptr[2]->load(str) == -1) ) {
		cerror << not_string << "\n";
		return(-1);
	}
	if (bptr[3]->load(len) == -1) {
		cerror << not_integer << "\n";
		return(-1);
	}
	//
	// Copy the string to allow for modification
	//
	strcpy(buffer,str);
	ret=DialogString(wtitle,title,buffer,len,NULL);
	buffer[len-1]='\0';
	str=(char *)malloc(strlen(buffer)+1);
	strcpy(str,buffer);
	argr=new class value_string;
	argr->assign_storage(str);
	bptr.SetReturn(argr);
	return(0);
}


//
// Displays a single string within a message dialog
//
// dialogmessage(message);
//
static int
dialogmsg( node_builtin &bptr )
{
	char *str;
 	static char maskset=0;
	static class value *argmask[1];

	if (!maskset) {
		argmask[0]=STRING_VALUE;
		maskset=1;
	}
	if (bptr.types_of_arguments(1,argmask)==-1)
		return(-1);
	if (bptr[0]->load(str) == -1) {
		cerror << not_string << "\n";
		return(-1);
	}
	DialogMessage(str);
	return(0);
}


//
// This function will be called each time a program is loaded and
// "compiled".  It should initialize all the necessary entries
// in the program's symbol table.
//
void
cmd_window(class context *cont)
{
	extern int allocate_type_id();
	static char create_types=1;

	new node_builtin(cont,"windowopen",windowopen);
	new node_builtin(cont,"windowclose",windowclose);

	//
	// Drawing functions
	// 
	new node_builtin(cont,"clearpage",clearpage);
	new node_builtin(cont,"drawpage",drawpage);
	new node_builtin(cont,"drawto",drawto);
	new node_builtin(cont,"moveto",moveto);
	new node_builtin(cont,"size",size);
	new node_builtin(cont,"axis_date",axis_date);
	new node_builtin(cont,"plot",plot);
	new node_builtin(cont,"linewidth",linewidth);
	new node_builtin(cont,"colour",colour);
	new node_builtin(cont,"color",colour);


	new node_builtin(cont,"menuopen",menuopen);
	new node_builtin(cont,"menuclose",menuclose);
	new node_builtin(cont,"menu",menu);

	new node_builtin(cont,"dialogyesno",dialogyesno);
	new node_builtin(cont,"dialogstring",dialogstring);
	new node_builtin(cont,"dialogmessage",dialogmsg);
	//
	// New types
	//
	if (create_types) {
		Valuewindow_id=allocate_type_id();
		WINDOW_VALUE=new class Valuewindow;

		Valuemenu_id=allocate_type_id();
		MENU_VALUE=new class Valuemenu;
		create_types=0;
	}
	cont->addtype("WINDOW",new class Valuewindow);
	cont->addtype("MENU",new class Valuemenu);
}
