/*+-------------------------------------------------------------------------
	colormgt.c - u386mon color customization for color curses
	wht@n4hgf.Mt-Park.GA.US
--------------------------------------------------------------------------*/
/*+:EDITS:*/
/*:08-03-1993-15:51-wht@n4hgf-source control point x2.70 */
/*:08-03-1993-12:33-wht@n4hgf-creation */

#include "config.h"
#include <curses.h>

#if defined(COLOR_PAIR)
#include <pwd.h>
#undef timeout
#undef reg		     /* per nba@sysware.dk */
#include "libpanel.h"
#include "u386mon.h"

typedef struct name_num
{
    char *name;
    int num;
}
NAME_NUM;

NAME_NUM colors[] =
{
    {"black", cBLK},
    {"blue", cBLU},
    {"brown", cBRN},
    {"cyan", cCYN},
    {"gray", cGRY},
    {"green", cGRY},
    {"hi_white", cHIW},
    {"lt_blue", cLTB},
    {"lt_cyan", cLTC},
    {"lt_green", cLTG},
    {"lt_magenta", cLTM},
    {"lt_red", cLTR},
    {"magenta", cMAG},
    {"red", cRED},
    {"white", cWHT},
    {"yellow", cYEL},
    {(char *)0, -1}
};

NAME_NUM field_types[] =
{
    {"low", cpLOW},
    {"medium", cpMED},
    {"high", cpHIGH},
    {"literal", cpLIT},
    {"info", cpINFO},
    {"reverse", cpREVERSE},
    {"banner", cpBANNER},
    {"banwarn", cpBANWARN},
};

static char *str_token_static = 0;

char *get_home_dir();
char *arg_token();
void build_arg_array();

/*+-------------------------------------------------------------------------
	name_to_num(cname)
--------------------------------------------------------------------------*/
int
name_to_num(cname)
char *cname;
{
    register NAME_NUM *color = colors;
    register itmp;

    while (color->name)
    {
	if ((itmp = strcmp(color->name, cname)) > 0)
	    return (-1);
	if (!itmp)
	    return (color->num);
	color++;
    }
    return (-1);

}			     /* end of name_to_num */

/*+-------------------------------------------------------------------------
	apply_color_directive(ntokens,tokens)

must be called with at least one token, max three:
field_type, foreground color, background color

the default is hi white foreground, black background
--------------------------------------------------------------------------*/
int
apply_color_directive(ntokens, tokens)
int ntokens;
char **tokens;
{
    int type;
    int fgnd;
    int bgnd;
    int bad_type = 0;
    int printed = 0;

    if (!ntokens)
	return(0);

    if ((type = name_to_num(field_types, tokens[0])) < 0)
    {
	printw("Invalid field type \"%s\" ignored.\n", tokens[0]);
	bad_type = 1;
	printed = 1;
    }

    fgnd = (ntokens > 1) ? name_to_num(colors, tokens[1]) : cHIW;
    bgnd = (ntokens > 2) ? name_to_num(colors, tokens[2]) : cBLK;

    if ((fgnd == -1) || (bgnd == -1))
    {
	printw("Invalid color for \"%s\": ", tokens[0]);
	if (fgnd == -1)
	    printw(" foreground \"%s\",", tokens[1]);
	if (bgnd == -1)
	    printw(" background \"%s\",", tokens[2]);
	printw("\n    defaulting to white on black\n");
	fgnd = cHIW;
	bgnd = cBLK;
	printed = 1;
    }

    if (bad_type)
	return;

    init_pair(type,fgnd,bgnd);

    return(printed);

}			     /* end of apply_color_directive */

/*+-------------------------------------------------------------------------
	read_color_customization() - read color definition if present
--------------------------------------------------------------------------*/
void
read_color_customization()
{
    FILE *fp;
    char s256[256];

#define MAX_COLOR_TOKENS 6
    char *tokens[MAX_COLOR_TOKENS];
    int ntokens;
    char *cp;
    int itmp;
    char *name = "";
    int printed = 0;

    name = get_home_dir();
    strcpy(s256, name);
    strcat(s256, "/.u386monrc");
    name = s256;
    if (!(fp = fopen(s256, "r")))
    {
	if (!(fp = fopen(name = "/usr/local/lib/u386monrc", "r")))
	    return;
    }

    while (fgets(s256, sizeof(s256), fp))
    {
	if (itmp = strlen(s256))	/* itmp = len; if > 0 ... */
	{
	    itmp--;
	    s256[itmp] = 0;  /* ... strip trailing NL */
	}
	if (cp = strchr(s256, '#'))	/* comment? */
	    *cp = 0;
	cp = s256;	     /* first call to str_token, -> buff */
	while ((*cp == 0x20) || (*cp == 0x09))
	    cp++;	     /* strip leading spaces */
	if (*cp == 0)	     /* if line all blank, skip it */
	    continue;
	build_arg_array(s256, tokens, MAX_COLOR_TOKENS, &ntokens);
	if (ntokens < 2)
	    continue;
	printed |= apply_color_directive(ntokens, tokens);
    }			     /* while records left to ready */

    fclose(fp);

    if(printed)
    {
	fprintf(stderr,"\7");
	refresh();
	sleep(3);
	clear();
	refresh();
    }

}			     /* end of read_color_customization */

/*+-----------------------------------------------------------------------
	get_home_dir() - get user home directory
------------------------------------------------------------------------*/
char *
get_home_dir()
{
    static char home_directory[256] = "";
    struct passwd *pwent;
    char *cp;

    if (cp = getenv("HOME"))
    {
	strcpy(home_directory, cp);
	return (home_directory);
    }

    if (!(pwent = getpwuid(getuid())))
    {
	printf("I cannot find a passwd entry for you!!\n");
	exit(1);
    }
    strcpy(home_directory, pwent->pw_dir);
    endpwent();
    return (home_directory);

}			     /* end of get_home_dir */

/*+-----------------------------------------------------------------------
	arg_token(parsestr,termchars)

Get next token from string parsestr ((char *)0 on 2nd, 3rd, etc.
calls), where tokens are nonempty strings separated by runs of chars
from termchars.  Writes nulls into parsestr to end tokens.
termchars need not remain constant from call to call.

Treats multiple occurrences of a termchar as one delimiter (does not
allow null fields).
------------------------------------------------------------------------*/
char *
arg_token(parsestr,termchars)
char *parsestr;
char *termchars;
{
	register char *parseptr;
	char *token;

	if(!parsestr && !str_token_static)
		return((char *)0);

	if(parsestr)
	{
		str_token_static = (char *)0;
		parseptr = parsestr;
	}
	else
       parseptr = str_token_static;

	while(*parseptr)
	{
		if(!strchr(termchars,*parseptr))
			break;
		parseptr++;
	}

	if(!*parseptr)
	{
		str_token_static = (char *)0;
		return((char *)0);
	}

	token = parseptr;

	/*
	 * tokens beginning with apostrophe or quotes kept together
	 */
	if(*token == '\'')
	{
		token++;
		parseptr++;
		while(*parseptr)
		{
			if(*parseptr == '\'')
			{
				str_token_static = parseptr + 1;
				*parseptr = 0;
				return(token);
			}
			parseptr++;
		}
		str_token_static = (char *)0;
		return(token);
	}
	else if(*token == '"')
	{
		token++;
		parseptr++;
		while(*parseptr)
		{
			if(*parseptr == '"')
			{
				str_token_static = parseptr + 1;
				*parseptr = 0;
				return(token);
			}
			parseptr++;
		}
		str_token_static = (char *)0;
		return(token);
	}

	while(*parseptr)
	{
		if(strchr(termchars,*parseptr))
		{
			*parseptr = 0;
			str_token_static = parseptr + 1;
			while(*str_token_static)
			{
				if(!strchr(termchars,*str_token_static))
					break;
				str_token_static++;
			}
			return(token);
		}
		parseptr++;
	}
	str_token_static = (char *)0;
	return(token);
}	/* end of arg_token */

/*+-------------------------------------------------------------------------
	build_arg_array(cmd,arg,arg_max_quan,&narg)
--------------------------------------------------------------------------*/
void
build_arg_array(cmd,arg,arg_max_quan,narg_rtn)
char *cmd;
char **arg;
int arg_max_quan;
int *narg_rtn;
{
	register narg;

	str_token_static = (char *)0;
	memset((char *)arg,0,sizeof(char *) * arg_max_quan);
	if(!(arg[0] = arg_token(cmd," \t\r\n")))
	{
		*narg_rtn = 0;
		return;
	}

	for(narg = 1; narg < arg_max_quan; ++narg)
	{
		if(!(arg[narg] = arg_token((char *)0," \t\r\n"))) 
			break;
	}

	*narg_rtn = narg;

}	/* end of build_arg_array */

#endif /* COLOR_PAIR */

/* vi: set tabstop=4 shiftwidth=4: */
/* end of colormgt.c */
