
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

// $Id: colour.C,v 2.3 1995/06/24 15:01:04 davidsi Exp $
//
// $Log: colour.C,v $
// Revision 2.3  1995/06/24  15:01:04  davidsi
// - Added GNU copyright message.
//
// Revision 2.2  1995/06/02  20:04:49  davidsi
// - Added support for all standard X colours.
//
// Revision 2.1  1995/04/25  21:11:30  davidsi
// - Added support for a busy cursor.
// - Added an owner member to the Colour class. This allows a colour instance
//   to be named and therefore to pernament store its colours.  The name
//   can be set with ColourSetowner(). Constructors and destructors revised
//   to handle this.
// - Colour::ColourSelect() can now (if an owner is set) save colour information
//   to the preferences database.
// - Added Colourfunction2index() to the Colour class to allow function
//   names to be mapped to their index (enumerated) values.
// - Added STORE and CLEAR buttons that allow the user to add or remove
//   their colour preferences.
//
// Revision 2.0  1995/02/27  22:06:19  davidsi
// - Initial version.
//
//
//


//
// Include Files
//
#include <stdio.h>
#include <varargs.h>
#define _STDARG_H
#include <stdlib.h>
#include <string.h>
#include <dial.h>
#include <math.h>

#include"localmacros.h"

#include"colour.h"
#include"dialog.h"
#include"support.h"
#include"button.h"
#include"optionl.h"
#include"prefs.h"

//
// Defines for the database identifiers
//
#define COLOUR_DB_CLASS	"COLOUR"		// main colour key
#define DEFCOL_DB_CLASS	"DEFAULT"		// default secondary key



//
///////////////////////////////////////////////////////////////////////////////
//
// Publicly available functions
//
///////////////////////////////////////////////////////////////////////////////
//

static Colour basecolours;

struct colouremap {
	char *cname;
	colour_e col;
	};

static colouremap mapper[] = {
	{"snow",SNOW},
	{"ghost white",GHOSTWHITE},
	{"GhostWhite",GHOSTWHITE},
	{"white smoke",WHITESMOKE},
	{"WhiteSmoke",WHITESMOKE},
	{"gainsboro",GAINSBORO},
	{"floral white",FLORALWHITE},
	{"FloralWhite",FLORALWHITE},
	{"old lace",OLDLACE},
	{"OldLace",OLDLACE},
	{"linen",LINEN},
	{"antique white",ANTIQUEWHITE},
	{"AntiqueWhite",ANTIQUEWHITE},
	{"papaya whip",PAPAYAWHIP},
	{"PapayaWhip",PAPAYAWHIP},
	{"blanched almond",BLANCHEDALMOND},
	{"BlanchedAlmond",BLANCHEDALMOND},
	{"bisque",BISQUE},
	{"peach puff",PEACHPUFF},
	{"PeachPuff",PEACHPUFF},
	{"navajo white",NAVAJOWHITE},
	{"NavajoWhite",NAVAJOWHITE},
	{"moccasin",MOCCASIN},
	{"cornsilk",CORNSILK},
	{"ivory",IVORY},
	{"lemon chiffon",LEMONCHIFFON},
	{"LemonChiffon",LEMONCHIFFON},
	{"seashell",SEASHELL},
	{"honeydew",HONEYDEW},
	{"mint cream",MINTCREAM},
	{"MintCream",MINTCREAM},
	{"azure",AZURE},
	{"alice blue",ALICEBLUE},
	{"AliceBlue",ALICEBLUE},
	{"lavender",LAVENDER},
	{"lavender blush",LAVENDERBLUSH},
	{"LavenderBlush",LAVENDERBLUSH},
	{"misty rose",MISTYROSE},
	{"MistyRose",MISTYROSE},
	{"white",WHITE},
	{"black",BLACK},
	{"dark slate gray",DARKSLATEGRAY},
	{"DarkSlateGray",DARKSLATEGRAY},
	{"dark slate grey",DARKSLATEGREY},
	{"DarkSlateGrey",DARKSLATEGREY},
	{"dim gray",DIMGRAY},
	{"DimGray",DIMGRAY},
	{"dim grey",DIMGREY},
	{"DimGrey",DIMGREY},
	{"slate gray",SLATEGRAY},
	{"SlateGray",SLATEGRAY},
	{"slate grey",SLATEGREY},
	{"SlateGrey",SLATEGREY},
	{"light slate gray",LIGHTSLATEGRAY},
	{"LightSlateGray",LIGHTSLATEGRAY},
	{"light slate grey",LIGHTSLATEGREY},
	{"LightSlateGrey",LIGHTSLATEGREY},
	{"gray",GRAY},
	{"grey",GREY},
	{"light grey",LIGHTGREY},
	{"LightGrey",LIGHTGREY},
	{"light gray",LIGHTGRAY},
	{"LightGray",LIGHTGRAY},
	{"midnight blue",MIDNIGHTBLUE},
	{"MidnightBlue",MIDNIGHTBLUE},
	{"navy",NAVY},
	{"navy blue",NAVYBLUE},
	{"NavyBlue",NAVYBLUE},
	{"cornflower blue",CORNFLOWERBLUE},
	{"CornflowerBlue",CORNFLOWERBLUE},
	{"dark slate blue",DARKSLATEBLUE},
	{"DarkSlateBlue",DARKSLATEBLUE},
	{"slate blue",SLATEBLUE},
	{"SlateBlue",SLATEBLUE},
	{"medium slate blue",MEDIUMSLATEBLUE},
	{"MediumSlateBlue",MEDIUMSLATEBLUE},
	{"light slate blue",LIGHTSLATEBLUE},
	{"LightSlateBlue",LIGHTSLATEBLUE},
	{"medium blue",MEDIUMBLUE},
	{"MediumBlue",MEDIUMBLUE},
	{"royal blue",ROYALBLUE},
	{"RoyalBlue",ROYALBLUE},
	{"blue",BLUE},
	{"dodger blue",DODGERBLUE},
	{"DodgerBlue",DODGERBLUE},
	{"deep sky blue",DEEPSKYBLUE},
	{"DeepSkyBlue",DEEPSKYBLUE},
	{"sky blue",SKYBLUE},
	{"SkyBlue",SKYBLUE},
	{"light sky blue",LIGHTSKYBLUE},
	{"LightSkyBlue",LIGHTSKYBLUE},
	{"steel blue",STEELBLUE},
	{"SteelBlue",STEELBLUE},
	{"light steel blue",LIGHTSTEELBLUE},
	{"LightSteelBlue",LIGHTSTEELBLUE},
	{"light blue",LIGHTBLUE},
	{"LightBlue",LIGHTBLUE},
	{"powder blue",POWDERBLUE},
	{"PowderBlue",POWDERBLUE},
	{"pale turquoise",PALETURQUOISE},
	{"PaleTurquoise",PALETURQUOISE},
	{"dark turquoise",DARKTURQUOISE},
	{"DarkTurquoise",DARKTURQUOISE},
	{"medium turquoise",MEDIUMTURQUOISE},
	{"MediumTurquoise",MEDIUMTURQUOISE},
	{"turquoise",TURQUOISE},
	{"cyan",CYAN},
	{"light cyan",LIGHTCYAN},
	{"LightCyan",LIGHTCYAN},
	{"cadet blue",CADETBLUE},
	{"CadetBlue",CADETBLUE},
	{"medium aquamarine",MEDIUMAQUAMARINE},
	{"MediumAquamarine",MEDIUMAQUAMARINE},
	{"aquamarine",AQUAMARINE},
	{"dark green",DARKGREEN},
	{"DarkGreen",DARKGREEN},
	{"dark olive green",DARKOLIVEGREEN},
	{"DarkOliveGreen",DARKOLIVEGREEN},
	{"dark sea green",DARKSEAGREEN},
	{"DarkSeaGreen",DARKSEAGREEN},
	{"sea green",SEAGREEN},
	{"SeaGreen",SEAGREEN},
	{"medium sea green",MEDIUMSEAGREEN},
	{"MediumSeaGreen",MEDIUMSEAGREEN},
	{"light sea green",LIGHTSEAGREEN},
	{"LightSeaGreen",LIGHTSEAGREEN},
	{"pale green",PALEGREEN},
	{"PaleGreen",PALEGREEN},
	{"spring green",SPRINGGREEN},
	{"SpringGreen",SPRINGGREEN},
	{"lawn green",LAWNGREEN},
	{"LawnGreen",LAWNGREEN},
	{"green",GREEN},
	{"chartreuse",CHARTREUSE},
	{"medium spring green",MEDIUMSPRINGGREEN},
	{"MediumSpringGreen",MEDIUMSPRINGGREEN},
	{"green yellow",GREENYELLOW},
	{"GreenYellow",GREENYELLOW},
	{"lime green",LIMEGREEN},
	{"LimeGreen",LIMEGREEN},
	{"yellow green",YELLOWGREEN},
	{"YellowGreen",YELLOWGREEN},
	{"forest green",FORESTGREEN},
	{"ForestGreen",FORESTGREEN},
	{"olive drab",OLIVEDRAB},
	{"OliveDrab",OLIVEDRAB},
	{"dark khaki",DARKKHAKI},
	{"DarkKhaki",DARKKHAKI},
	{"khaki",KHAKI},
	{"pale goldenrod",PALEGOLDENROD},
	{"PaleGoldenrod",PALEGOLDENROD},
	{"light goldenrod yellow",LIGHTGOLDENRODYELLOW},
	{"LightGoldenrodYellow",LIGHTGOLDENRODYELLOW},
	{"light yellow",LIGHTYELLOW},
	{"LightYellow",LIGHTYELLOW},
	{"yellow",YELLOW},
	{"gold",GOLD},
	{"light goldenrod",LIGHTGOLDENROD},
	{"LightGoldenrod",LIGHTGOLDENROD},
	{"goldenrod",GOLDENROD},
	{"dark goldenrod",DARKGOLDENROD},
	{"DarkGoldenrod",DARKGOLDENROD},
	{"rosy brown",ROSYBROWN},
	{"RosyBrown",ROSYBROWN},
	{"indian red",INDIANRED},
	{"IndianRed",INDIANRED},
	{"saddle brown",SADDLEBROWN},
	{"SaddleBrown",SADDLEBROWN},
	{"sienna",SIENNA},
	{"peru",PERU},
	{"burlywood",BURLYWOOD},
	{"beige",BEIGE},
	{"wheat",WHEAT},
	{"sandy brown",SANDYBROWN},
	{"SandyBrown",SANDYBROWN},
	{"tan",TAN},
	{"chocolate",CHOCOLATE},
	{"firebrick",FIREBRICK},
	{"brown",BROWN},
	{"dark salmon",DARKSALMON},
	{"DarkSalmon",DARKSALMON},
	{"salmon",SALMON},
	{"light salmon",LIGHTSALMON},
	{"LightSalmon",LIGHTSALMON},
	{"orange",ORANGE},
	{"dark orange",DARKORANGE},
	{"DarkOrange",DARKORANGE},
	{"coral",CORAL},
	{"light coral",LIGHTCORAL},
	{"LightCoral",LIGHTCORAL},
	{"tomato",TOMATO},
	{"orange red",ORANGERED},
	{"OrangeRed",ORANGERED},
	{"red",RED},
	{"hot pink",HOTPINK},
	{"HotPink",HOTPINK},
	{"deep pink",DEEPPINK},
	{"DeepPink",DEEPPINK},
	{"pink",PINK},
	{"light pink",LIGHTPINK},
	{"LightPink",LIGHTPINK},
	{"pale violet red",PALEVIOLETRED},
	{"PaleVioletRed",PALEVIOLETRED},
	{"maroon",MAROON},
	{"medium violet red",MEDIUMVIOLETRED},
	{"MediumVioletRed",MEDIUMVIOLETRED},
	{"violet red",VIOLETRED},
	{"VioletRed",VIOLETRED},
	{"magenta",MAGENTA},
	{"violet",VIOLET},
	{"plum",PLUM},
	{"orchid",ORCHID},
	{"medium orchid",MEDIUMORCHID},
	{"MediumOrchid",MEDIUMORCHID},
	{"dark orchid",DARKORCHID},
	{"DarkOrchid",DARKORCHID},
	{"dark violet",DARKVIOLET},
	{"DarkViolet",DARKVIOLET},
	{"blue violet",BLUEVIOLET},
	{"BlueViolet",BLUEVIOLET},
	{"purple",PURPLE},
	{"medium purple",MEDIUMPURPLE},
	{"MediumPurple",MEDIUMPURPLE},
	{"thistle",THISTLE},
	{"snow1",SNOW1},
	{"snow2",SNOW2},
	{"snow3",SNOW3},
	{"snow4",SNOW4},
	{"seashell1",SEASHELL1},
	{"seashell2",SEASHELL2},
	{"seashell3",SEASHELL3},
	{"seashell4",SEASHELL4},
	{"AntiqueWhite1",ANTIQUEWHITE1},
	{"AntiqueWhite2",ANTIQUEWHITE2},
	{"AntiqueWhite3",ANTIQUEWHITE3},
	{"AntiqueWhite4",ANTIQUEWHITE4},
	{"bisque1",BISQUE1},
	{"bisque2",BISQUE2},
	{"bisque3",BISQUE3},
	{"bisque4",BISQUE4},
	{"PeachPuff1",PEACHPUFF1},
	{"PeachPuff2",PEACHPUFF2},
	{"PeachPuff3",PEACHPUFF3},
	{"PeachPuff4",PEACHPUFF4},
	{"NavajoWhite1",NAVAJOWHITE1},
	{"NavajoWhite2",NAVAJOWHITE2},
	{"NavajoWhite3",NAVAJOWHITE3},
	{"NavajoWhite4",NAVAJOWHITE4},
	{"LemonChiffon1",LEMONCHIFFON1},
	{"LemonChiffon2",LEMONCHIFFON2},
	{"LemonChiffon3",LEMONCHIFFON3},
	{"LemonChiffon4",LEMONCHIFFON4},
	{"cornsilk1",CORNSILK1},
	{"cornsilk2",CORNSILK2},
	{"cornsilk3",CORNSILK3},
	{"cornsilk4",CORNSILK4},
	{"ivory1",IVORY1},
	{"ivory2",IVORY2},
	{"ivory3",IVORY3},
	{"ivory4",IVORY4},
	{"honeydew1",HONEYDEW1},
	{"honeydew2",HONEYDEW2},
	{"honeydew3",HONEYDEW3},
	{"honeydew4",HONEYDEW4},
	{"LavenderBlush1",LAVENDERBLUSH1},
	{"LavenderBlush2",LAVENDERBLUSH2},
	{"LavenderBlush3",LAVENDERBLUSH3},
	{"LavenderBlush4",LAVENDERBLUSH4},
	{"MistyRose1",MISTYROSE1},
	{"MistyRose2",MISTYROSE2},
	{"MistyRose3",MISTYROSE3},
	{"MistyRose4",MISTYROSE4},
	{"azure1",AZURE1},
	{"azure2",AZURE2},
	{"azure3",AZURE3},
	{"azure4",AZURE4},
	{"SlateBlue1",SLATEBLUE1},
	{"SlateBlue2",SLATEBLUE2},
	{"SlateBlue3",SLATEBLUE3},
	{"SlateBlue4",SLATEBLUE4},
	{"RoyalBlue1",ROYALBLUE1},
	{"RoyalBlue2",ROYALBLUE2},
	{"RoyalBlue3",ROYALBLUE3},
	{"RoyalBlue4",ROYALBLUE4},
	{"blue1",BLUE1},
	{"blue2",BLUE2},
	{"blue3",BLUE3},
	{"blue4",BLUE4},
	{"DodgerBlue1",DODGERBLUE1},
	{"DodgerBlue2",DODGERBLUE2},
	{"DodgerBlue3",DODGERBLUE3},
	{"DodgerBlue4",DODGERBLUE4},
	{"SteelBlue1",STEELBLUE1},
	{"SteelBlue2",STEELBLUE2},
	{"SteelBlue3",STEELBLUE3},
	{"SteelBlue4",STEELBLUE4},
	{"DeepSkyBlue1",DEEPSKYBLUE1},
	{"DeepSkyBlue2",DEEPSKYBLUE2},
	{"DeepSkyBlue3",DEEPSKYBLUE3},
	{"DeepSkyBlue4",DEEPSKYBLUE4},
	{"SkyBlue1",SKYBLUE1},
	{"SkyBlue2",SKYBLUE2},
	{"SkyBlue3",SKYBLUE3},
	{"SkyBlue4",SKYBLUE4},
	{"LightSkyBlue1",LIGHTSKYBLUE1},
	{"LightSkyBlue2",LIGHTSKYBLUE2},
	{"LightSkyBlue3",LIGHTSKYBLUE3},
	{"LightSkyBlue4",LIGHTSKYBLUE4},
	{"SlateGray1",SLATEGRAY1},
	{"SlateGray2",SLATEGRAY2},
	{"SlateGray3",SLATEGRAY3},
	{"SlateGray4",SLATEGRAY4},
	{"LightSteelBlue1",LIGHTSTEELBLUE1},
	{"LightSteelBlue2",LIGHTSTEELBLUE2},
	{"LightSteelBlue3",LIGHTSTEELBLUE3},
	{"LightSteelBlue4",LIGHTSTEELBLUE4},
	{"LightBlue1",LIGHTBLUE1},
	{"LightBlue2",LIGHTBLUE2},
	{"LightBlue3",LIGHTBLUE3},
	{"LightBlue4",LIGHTBLUE4},
	{"LightCyan1",LIGHTCYAN1},
	{"LightCyan2",LIGHTCYAN2},
	{"LightCyan3",LIGHTCYAN3},
	{"LightCyan4",LIGHTCYAN4},
	{"PaleTurquoise1",PALETURQUOISE1},
	{"PaleTurquoise2",PALETURQUOISE2},
	{"PaleTurquoise3",PALETURQUOISE3},
	{"PaleTurquoise4",PALETURQUOISE4},
	{"CadetBlue1",CADETBLUE1},
	{"CadetBlue2",CADETBLUE2},
	{"CadetBlue3",CADETBLUE3},
	{"CadetBlue4",CADETBLUE4},
	{"turquoise1",TURQUOISE1},
	{"turquoise2",TURQUOISE2},
	{"turquoise3",TURQUOISE3},
	{"turquoise4",TURQUOISE4},
	{"cyan1",CYAN1},
	{"cyan2",CYAN2},
	{"cyan3",CYAN3},
	{"cyan4",CYAN4},
	{"DarkSlateGray1",DARKSLATEGRAY1},
	{"DarkSlateGray2",DARKSLATEGRAY2},
	{"DarkSlateGray3",DARKSLATEGRAY3},
	{"DarkSlateGray4",DARKSLATEGRAY4},
	{"aquamarine1",AQUAMARINE1},
	{"aquamarine2",AQUAMARINE2},
	{"aquamarine3",AQUAMARINE3},
	{"aquamarine4",AQUAMARINE4},
	{"DarkSeaGreen1",DARKSEAGREEN1},
	{"DarkSeaGreen2",DARKSEAGREEN2},
	{"DarkSeaGreen3",DARKSEAGREEN3},
	{"DarkSeaGreen4",DARKSEAGREEN4},
	{"SeaGreen1",SEAGREEN1},
	{"SeaGreen2",SEAGREEN2},
	{"SeaGreen3",SEAGREEN3},
	{"SeaGreen4",SEAGREEN4},
	{"PaleGreen1",PALEGREEN1},
	{"PaleGreen2",PALEGREEN2},
	{"PaleGreen3",PALEGREEN3},
	{"PaleGreen4",PALEGREEN4},
	{"SpringGreen1",SPRINGGREEN1},
	{"SpringGreen2",SPRINGGREEN2},
	{"SpringGreen3",SPRINGGREEN3},
	{"SpringGreen4",SPRINGGREEN4},
	{"green1",GREEN1},
	{"green2",GREEN2},
	{"green3",GREEN3},
	{"green4",GREEN4},
	{"chartreuse1",CHARTREUSE1},
	{"chartreuse2",CHARTREUSE2},
	{"chartreuse3",CHARTREUSE3},
	{"chartreuse4",CHARTREUSE4},
	{"OliveDrab1",OLIVEDRAB1},
	{"OliveDrab2",OLIVEDRAB2},
	{"OliveDrab3",OLIVEDRAB3},
	{"OliveDrab4",OLIVEDRAB4},
	{"DarkOliveGreen1",DARKOLIVEGREEN1},
	{"DarkOliveGreen2",DARKOLIVEGREEN2},
	{"DarkOliveGreen3",DARKOLIVEGREEN3},
	{"DarkOliveGreen4",DARKOLIVEGREEN4},
	{"khaki1",KHAKI1},
	{"khaki2",KHAKI2},
	{"khaki3",KHAKI3},
	{"khaki4",KHAKI4},
	{"LightGoldenrod1",LIGHTGOLDENROD1},
	{"LightGoldenrod2",LIGHTGOLDENROD2},
	{"LightGoldenrod3",LIGHTGOLDENROD3},
	{"LightGoldenrod4",LIGHTGOLDENROD4},
	{"LightYellow1",LIGHTYELLOW1},
	{"LightYellow2",LIGHTYELLOW2},
	{"LightYellow3",LIGHTYELLOW3},
	{"LightYellow4",LIGHTYELLOW4},
	{"yellow1",YELLOW1},
	{"yellow2",YELLOW2},
	{"yellow3",YELLOW3},
	{"yellow4",YELLOW4},
	{"gold1",GOLD1},
	{"gold2",GOLD2},
	{"gold3",GOLD3},
	{"gold4",GOLD4},
	{"goldenrod1",GOLDENROD1},
	{"goldenrod2",GOLDENROD2},
	{"goldenrod3",GOLDENROD3},
	{"goldenrod4",GOLDENROD4},
	{"DarkGoldenrod1",DARKGOLDENROD1},
	{"DarkGoldenrod2",DARKGOLDENROD2},
	{"DarkGoldenrod3",DARKGOLDENROD3},
	{"DarkGoldenrod4",DARKGOLDENROD4},
	{"RosyBrown1",ROSYBROWN1},
	{"RosyBrown2",ROSYBROWN2},
	{"RosyBrown3",ROSYBROWN3},
	{"RosyBrown4",ROSYBROWN4},
	{"IndianRed1",INDIANRED1},
	{"IndianRed2",INDIANRED2},
	{"IndianRed3",INDIANRED3},
	{"IndianRed4",INDIANRED4},
	{"sienna1",SIENNA1},
	{"sienna2",SIENNA2},
	{"sienna3",SIENNA3},
	{"sienna4",SIENNA4},
	{"burlywood1",BURLYWOOD1},
	{"burlywood2",BURLYWOOD2},
	{"burlywood3",BURLYWOOD3},
	{"burlywood4",BURLYWOOD4},
	{"wheat1",WHEAT1},
	{"wheat2",WHEAT2},
	{"wheat3",WHEAT3},
	{"wheat4",WHEAT4},
	{"tan1",TAN1},
	{"tan2",TAN2},
	{"tan3",TAN3},
	{"tan4",TAN4},
	{"chocolate1",CHOCOLATE1},
	{"chocolate2",CHOCOLATE2},
	{"chocolate3",CHOCOLATE3},
	{"chocolate4",CHOCOLATE4},
	{"firebrick1",FIREBRICK1},
	{"firebrick2",FIREBRICK2},
	{"firebrick3",FIREBRICK3},
	{"firebrick4",FIREBRICK4},
	{"brown1",BROWN1},
	{"brown2",BROWN2},
	{"brown3",BROWN3},
	{"brown4",BROWN4},
	{"salmon1",SALMON1},
	{"salmon2",SALMON2},
	{"salmon3",SALMON3},
	{"salmon4",SALMON4},
	{"LightSalmon1",LIGHTSALMON1},
	{"LightSalmon2",LIGHTSALMON2},
	{"LightSalmon3",LIGHTSALMON3},
	{"LightSalmon4",LIGHTSALMON4},
	{"orange1",ORANGE1},
	{"orange2",ORANGE2},
	{"orange3",ORANGE3},
	{"orange4",ORANGE4},
	{"DarkOrange1",DARKORANGE1},
	{"DarkOrange2",DARKORANGE2},
	{"DarkOrange3",DARKORANGE3},
	{"DarkOrange4",DARKORANGE4},
	{"coral1",CORAL1},
	{"coral2",CORAL2},
	{"coral3",CORAL3},
	{"coral4",CORAL4},
	{"tomato1",TOMATO1},
	{"tomato2",TOMATO2},
	{"tomato3",TOMATO3},
	{"tomato4",TOMATO4},
	{"OrangeRed1",ORANGERED1},
	{"OrangeRed2",ORANGERED2},
	{"OrangeRed3",ORANGERED3},
	{"OrangeRed4",ORANGERED4},
	{"red1",RED1},
	{"red2",RED2},
	{"red3",RED3},
	{"red4",RED4},
	{"DeepPink1",DEEPPINK1},
	{"DeepPink2",DEEPPINK2},
	{"DeepPink3",DEEPPINK3},
	{"DeepPink4",DEEPPINK4},
	{"HotPink1",HOTPINK1},
	{"HotPink2",HOTPINK2},
	{"HotPink3",HOTPINK3},
	{"HotPink4",HOTPINK4},
	{"pink1",PINK1},
	{"pink2",PINK2},
	{"pink3",PINK3},
	{"pink4",PINK4},
	{"LightPink1",LIGHTPINK1},
	{"LightPink2",LIGHTPINK2},
	{"LightPink3",LIGHTPINK3},
	{"LightPink4",LIGHTPINK4},
	{"PaleVioletRed1",PALEVIOLETRED1},
	{"PaleVioletRed2",PALEVIOLETRED2},
	{"PaleVioletRed3",PALEVIOLETRED3},
	{"PaleVioletRed4",PALEVIOLETRED4},
	{"maroon1",MAROON1},
	{"maroon2",MAROON2},
	{"maroon3",MAROON3},
	{"maroon4",MAROON4},
	{"VioletRed1",VIOLETRED1},
	{"VioletRed2",VIOLETRED2},
	{"VioletRed3",VIOLETRED3},
	{"VioletRed4",VIOLETRED4},
	{"magenta1",MAGENTA1},
	{"magenta2",MAGENTA2},
	{"magenta3",MAGENTA3},
	{"magenta4",MAGENTA4},
	{"orchid1",ORCHID1},
	{"orchid2",ORCHID2},
	{"orchid3",ORCHID3},
	{"orchid4",ORCHID4},
	{"plum1",PLUM1},
	{"plum2",PLUM2},
	{"plum3",PLUM3},
	{"plum4",PLUM4},
	{"MediumOrchid1",MEDIUMORCHID1},
	{"MediumOrchid2",MEDIUMORCHID2},
	{"MediumOrchid3",MEDIUMORCHID3},
	{"MediumOrchid4",MEDIUMORCHID4},
	{"DarkOrchid1",DARKORCHID1},
	{"DarkOrchid2",DARKORCHID2},
	{"DarkOrchid3",DARKORCHID3},
	{"DarkOrchid4",DARKORCHID4},
	{"purple1",PURPLE1},
	{"purple2",PURPLE2},
	{"purple3",PURPLE3},
	{"purple4",PURPLE4},
	{"MediumPurple1",MEDIUMPURPLE1},
	{"MediumPurple2",MEDIUMPURPLE2},
	{"MediumPurple3",MEDIUMPURPLE3},
	{"MediumPurple4",MEDIUMPURPLE4},
	{"thistle1",THISTLE1},
	{"thistle2",THISTLE2},
	{"thistle3",THISTLE3},
	{"thistle4",THISTLE4},
	{"gray0",GRAY0},
	{"grey0",GREY0},
	{"gray1",GRAY1},
	{"grey1",GREY1},
	{"gray2",GRAY2},
	{"grey2",GREY2},
	{"gray3",GRAY3},
	{"grey3",GREY3},
	{"gray4",GRAY4},
	{"grey4",GREY4},
	{"gray5",GRAY5},
	{"grey5",GREY5},
	{"gray6",GRAY6},
	{"grey6",GREY6},
	{"gray7",GRAY7},
	{"grey7",GREY7},
	{"gray8",GRAY8},
	{"grey8",GREY8},
	{"gray9",GRAY9},
	{"grey9",GREY9},
	{"gray10",GRAY10},
	{"grey10",GREY10},
	{"gray11",GRAY11},
	{"grey11",GREY11},
	{"gray12",GRAY12},
	{"grey12",GREY12},
	{"gray13",GRAY13},
	{"grey13",GREY13},
	{"gray14",GRAY14},
	{"grey14",GREY14},
	{"gray15",GRAY15},
	{"grey15",GREY15},
	{"gray16",GRAY16},
	{"grey16",GREY16},
	{"gray17",GRAY17},
	{"grey17",GREY17},
	{"gray18",GRAY18},
	{"grey18",GREY18},
	{"gray19",GRAY19},
	{"grey19",GREY19},
	{"gray20",GRAY20},
	{"grey20",GREY20},
	{"gray21",GRAY21},
	{"grey21",GREY21},
	{"gray22",GRAY22},
	{"grey22",GREY22},
	{"gray23",GRAY23},
	{"grey23",GREY23},
	{"gray24",GRAY24},
	{"grey24",GREY24},
	{"gray25",GRAY25},
	{"grey25",GREY25},
	{"gray26",GRAY26},
	{"grey26",GREY26},
	{"gray27",GRAY27},
	{"grey27",GREY27},
	{"gray28",GRAY28},
	{"grey28",GREY28},
	{"gray29",GRAY29},
	{"grey29",GREY29},
	{"gray30",GRAY30},
	{"grey30",GREY30},
	{"gray31",GRAY31},
	{"grey31",GREY31},
	{"gray32",GRAY32},
	{"grey32",GREY32},
	{"gray33",GRAY33},
	{"grey33",GREY33},
	{"gray34",GRAY34},
	{"grey34",GREY34},
	{"gray35",GRAY35},
	{"grey35",GREY35},
	{"gray36",GRAY36},
	{"grey36",GREY36},
	{"gray37",GRAY37},
	{"grey37",GREY37},
	{"gray38",GRAY38},
	{"grey38",GREY38},
	{"gray39",GRAY39},
	{"grey39",GREY39},
	{"gray40",GRAY40},
	{"grey40",GREY40},
	{"gray41",GRAY41},
	{"grey41",GREY41},
	{"gray42",GRAY42},
	{"grey42",GREY42},
	{"gray43",GRAY43},
	{"grey43",GREY43},
	{"gray44",GRAY44},
	{"grey44",GREY44},
	{"gray45",GRAY45},
	{"grey45",GREY45},
	{"gray46",GRAY46},
	{"grey46",GREY46},
	{"gray47",GRAY47},
	{"grey47",GREY47},
	{"gray48",GRAY48},
	{"grey48",GREY48},
	{"gray49",GRAY49},
	{"grey49",GREY49},
	{"gray50",GRAY50},
	{"grey50",GREY50},
	{"gray51",GRAY51},
	{"grey51",GREY51},
	{"gray52",GRAY52},
	{"grey52",GREY52},
	{"gray53",GRAY53},
	{"grey53",GREY53},
	{"gray54",GRAY54},
	{"grey54",GREY54},
	{"gray55",GRAY55},
	{"grey55",GREY55},
	{"gray56",GRAY56},
	{"grey56",GREY56},
	{"gray57",GRAY57},
	{"grey57",GREY57},
	{"gray58",GRAY58},
	{"grey58",GREY58},
	{"gray59",GRAY59},
	{"grey59",GREY59},
	{"gray60",GRAY60},
	{"grey60",GREY60},
	{"gray61",GRAY61},
	{"grey61",GREY61},
	{"gray62",GRAY62},
	{"grey62",GREY62},
	{"gray63",GRAY63},
	{"grey63",GREY63},
	{"gray64",GRAY64},
	{"grey64",GREY64},
	{"gray65",GRAY65},
	{"grey65",GREY65},
	{"gray66",GRAY66},
	{"grey66",GREY66},
	{"gray67",GRAY67},
	{"grey67",GREY67},
	{"gray68",GRAY68},
	{"grey68",GREY68},
	{"gray69",GRAY69},
	{"grey69",GREY69},
	{"gray70",GRAY70},
	{"grey70",GREY70},
	{"gray71",GRAY71},
	{"grey71",GREY71},
	{"gray72",GRAY72},
	{"grey72",GREY72},
	{"gray73",GRAY73},
	{"grey73",GREY73},
	{"gray74",GRAY74},
	{"grey74",GREY74},
	{"gray75",GRAY75},
	{"grey75",GREY75},
	{"gray76",GRAY76},
	{"grey76",GREY76},
	{"gray77",GRAY77},
	{"grey77",GREY77},
	{"gray78",GRAY78},
	{"grey78",GREY78},
	{"gray79",GRAY79},
	{"grey79",GREY79},
	{"gray80",GRAY80},
	{"grey80",GREY80},
	{"gray81",GRAY81},
	{"grey81",GREY81},
	{"gray82",GRAY82},
	{"grey82",GREY82},
	{"gray83",GRAY83},
	{"grey83",GREY83},
	{"gray84",GRAY84},
	{"grey84",GREY84},
	{"gray85",GRAY85},
	{"grey85",GREY85},
	{"gray86",GRAY86},
	{"grey86",GREY86},
	{"gray87",GRAY87},
	{"grey87",GREY87},
	{"gray88",GRAY88},
	{"grey88",GREY88},
	{"gray89",GRAY89},
	{"grey89",GREY89},
	{"gray90",GRAY90},
	{"grey90",GREY90},
	{"gray91",GRAY91},
	{"grey91",GREY91},
	{"gray92",GRAY92},
	{"grey92",GREY92},
	{"gray93",GRAY93},
	{"grey93",GREY93},
	{"gray94",GRAY94},
	{"grey94",GREY94},
	{"gray95",GRAY95},
	{"grey95",GREY95},
	{"gray96",GRAY96},
	{"grey96",GREY96},
	{"gray97",GRAY97},
	{"grey97",GREY97},
	{"gray98",GRAY98},
	{"grey98",GREY98},
	{"gray99",GRAY99},
	{"grey99",GREY99},
	{"gray100",GRAY100},
	{"grey100",GREY100},
	{"scoBackground",SCOBACKGROUND},
	{"scoForeground",SCOFOREGROUND},
	{"scoTopShadow",SCOTOPSHADOW},
	{"scoActiveBackground",SCOACTIVEBACKGROUND},
	{"scoActiveForeground",SCOACTIVEFOREGROUND},
	{"scoActiveTopShadow",SCOACTIVETOPSHADOW},
	{"scoAltBackground",SCOALTBACKGROUND},
	{"scoHighlight",SCOHIGHLIGHT},
	{NULL,COLOUR_NOTSET},
	};

static char **colourmap=NULL;
static int maxcolour;


//
// colour2string colour
//
// Returns an ascii string for the specified colour.
//
// INPUTS :	colour		The colour whose string equivalent is req'd
// RETURNS:	A string containing the name of the given colour. 
//
char *
colour2string(colour_e col)
{
	char *colstr;
	int i;

	switch(col) {
	case BACKGROUND:
		colstr=colour2string(
			     basecolours.ColourGet(COLOUR_FUNCTION_BACKGROUND));
		break;
	case FOREGROUND:
		colstr=colour2string(
			     basecolours.ColourGet(COLOUR_FUNCTION_FOREGROUND));
		break;
	default:
		for(i=0;i<maxcolour;i++) {
			if (mapper[i].col==col)
				return (mapper[i].cname);
		}
		colstr="white";
	}
	return(colstr);
}


char *
background_colour()
{
	return(colour2string(basecolours.ColourGet(COLOUR_FUNCTION_BACKGROUND)));
}

char *
foreground_colour()
{
	return(colour2string(basecolours.ColourGet(COLOUR_FUNCTION_FOREGROUND)));
}

//
// string2colour colourname
//
// Returns an ascii string for the specified colour.
//
// INPUTS :	colour		The colour string whose colour_e equivalent 
//				is req'd. The function can handle NULL
//				being passed as the string address - but
//				don't expect anything to match with it!!
// RETURNS:	The colour_e for the string.
//
colour_e
string2colour(char *colourname)
{
	int i;

	if (!colourname)
		return(COLOUR_NOTSET);

	for(i=0;i<maxcolour;i++) {
		if (!alpha_strcmp(mapper[i].cname,colourname) ) {
			return(mapper[i].col);
		}
	}

	return(COLOUR_NOTSET);
}

//
///////////////////////////////////////////////////////////////////////////////
//
// The Colour selection class
//
///////////////////////////////////////////////////////////////////////////////
//

enum callids_e {
	BUTTON_FORE,
	BUTTON_BACK,
	BUTTON_OK,
	BUTTON_CANCEL,
	BUTTON_STORE,
	BUTTON_CLEAR,
	};

class Colourselect : private Button,public Colour,private Optionlist
{
	int curcolour;
	char finished;		// TRUE when OK or CANCEL hit
	//
	// Dialogs (etc) used when defining the colour
	//
	Dialogshell d_shell;         // top most dialog
	Optionlistinstance *d_optionlist;
	Dialogvertical d_vert;
	Dialoghorizontal d_horiz;
	Dialoghorizontal d_horizok;
	Dialoghorizontal d_horizbut;
private:
	//
	// This overrides the pure virtual button handler in the Button class
	//
	virtual void ButtonCallback(unsigned long id,buttonstate_e flag);
public:
	void select();

	Colourselect(class Colour *);
	~Colourselect();
};

//
// ButtonCallback
//
// Handles button callbacks.
//
void
Colourselect::ButtonCallback(unsigned long id,buttonstate_e flag)
{
	char *str;
	colour_e col;
	int i;

	str=d_optionlist->OptionlistinstanceGet();
	col=string2colour(str);
	if (col!=COLOUR_NOTSET) 
		colours[curcolour]=col; 

	//
	// A button was hit that means we are setting the colour for
	// some other operation
	//
	if (id>=1000) {
		curcolour=id-1000;
		d_optionlist->OptionlistinstanceSet(
					colour2string(colours[curcolour]));
		return;
	}
	
	switch(id) {
	case BUTTON_OK:
		finished=1;
		d_shell.DialogDestroy();
		break;
	case BUTTON_CANCEL:
		finished=1;
		for(i=0;i<MAX_COLOURS;i++)
			colours[i]=COLOUR_NOTSET;
		d_shell.DialogDestroy();
		break;
	case BUTTON_STORE:
		if (owner) {
			for(i=0;i<MAX_COLOURS;i++)
				if (funcnames[i] && colours[i]!=COLOUR_NOTSET)
					db_prefs->DatabasePrefsAssign(
						COLOUR_DB_CLASS,
						owner,funcnames[i],
						colour2string(colours[i]));
		}
		break;
	case BUTTON_CLEAR:
		if (owner) 
			db_prefs->DatabaseDelete(COLOUR_DB_CLASS,owner);

		break;
	}
}


//
// select()
//
// This allows colours for the different colour functions to be selected.
// Be warned, that button types >100 are used to indicate the dynamically
// created colour function section buttons.  The number of these
// buttons will vary with the number of colour functions being used.
//
void
Colourselect::select()
{
	buttonstate_e f1,f2,f3;
	int i;
	buttonstate_e flags;
        static Newoptionlist nolist = {
		30,			// width
		5,			// height
		NULL,			// list of items
		"Available Colours",	// name of list
		"",			// name of item
		0,			// flags
		NULL,			// default font
	};
	dcursortype_e cursor;

	cursor=DialogCursor( DCURSOR_WAIT );

	d_shell.DialogshellCreate("Colour Setup");
	d_vert.DialogverticalCreate(d_shell);

	nolist.listitem=colourmap;
	d_optionlist=OptionlistCreate(d_vert,nolist,0);
	d_optionlist->OptionlistinstanceSet(colour2string(colours[0]));
	curcolour=0;

	d_horiz.DialoghorizontalCreate(d_vert);
	for(i=0;i<MAX_COLOURS;i++)
		if (funcnames[i]!=NULL) {
			if (i==0)
				flags=BUTTON_ENABLED;
			else
				flags=BUTTON_DISABLED;
			(void)ButtonTogglemutex(d_horiz,funcnames[i],1000+i,flags);
		}

	d_horizbut.DialoghorizontalCreate(d_vert);
	(void)ButtonPush(d_horizbut,"STORE SELECTION",BUTTON_STORE);
	(void)ButtonPush(d_horizbut,"CLEAR SELECTION",BUTTON_CLEAR);

	d_horizok.DialoghorizontalCreate(d_vert);
	(void)ButtonPush(d_horizok,"OK",BUTTON_OK);
	(void)ButtonPush(d_horizok,"CANCEL",BUTTON_CANCEL);
	d_shell.DialogRealize();

	DialogCursor( cursor );

	finished=0;
	while (finished==0)
		DialogEvent();
}

//
// Constructor
//
Colourselect::Colourselect(class Colour *og)
{
	ColourCopy(og);
	ColourSetowner(og->owner);
}

//
// Destructor
//
Colourselect::~Colourselect()
{
	d_shell.DialogDestroy();
}

static void
colourselect(local_va_dcl)
{
	basecolours.ColourSelect();
}

//
///////////////////////////////////////////////////////////////////////////////
//
// The Colour class
//
///////////////////////////////////////////////////////////////////////////////
//


//
// ColourGet colourfunction
//
// Returns the colour that has been assigned to a particular function.
//
// INPUTS :	colourfunction	The colour function whose value is to be set
// RETURNS:	The colour id for the specified function
//
colour_e
Colour::ColourGet(colourf_e colid)
{
	return( colours[colid] );
}


//
// ColourSet colourfunction colour
//
// Assigns a colour to a particular function. It is good practice to
// only assign colours to those functions whose names have been 
// specified.
//
// INPUTS :	colourfunction	The colour function whose value is to be set
//		colour		The colour to be assigned
// RETURNS:	None
//
void
Colour::ColourSet(colourf_e colid,colour_e col)
{
	colours[colid]=col;
}

//
// ColourBackground
//
// Simple convenience function to return the background colour in the
// current instace.
//
// INPUTS :	None
// RETURNS:	The colour identifier for the backgound colour
//
colour_e
Colour::ColourBackground()
{
	return( ColourGet(COLOUR_FUNCTION_BACKGROUND) );
}


//
// ColourForeground
//
// Simple convenience function to return the foreground colour in the
// current instace.
//
// INPUTS :	None
// RETURNS:	The colour identifier for the foregound colour
//
colour_e
Colour::ColourForeground()
{
	return( ColourGet(COLOUR_FUNCTION_FOREGROUND) );
}

//
// ColourSelect
// 
// Allows colours within an instance to be defined interactively
// using a dialog.
//
// INPUTS :	None
// RETURNS:	1 if a colour was changed, else 0
//
int
Colour::ColourSelect()
{
	int i;
	char colourchanged=0;

	cselect=new class Colourselect(this);
	cselect->select();
	
	//
	// Copy back the colours which were set.  If a colour changes
	// set a flag so a correct return value can be made.
	// 
	for(i=0;i<MAX_COLOURS;i++)
		if ( (cselect->funcnames[i]!=NULL )
		   && (cselect->colours[i]!=COLOUR_NOTSET)
		   && colours[i]!=cselect->colours[i] )  {
			colours[i]=cselect->colours[i];
			colourchanged=1;
		}

	delete cselect;
	cselect=NULL;

	if (colourchanged)
		return(1);
	
	return(0);
}

//
// function2index functionname
//
// Maps a function name to the correct index.
//
// INPUTS :	functionname 	Name of a colour function
// RETURNS:	The correct colourf_e value for that function which
//		(for at least the moment) better exist!
//
colourf_e
Colour::Colourfunction2index(char *fname)
{
	int i;

	for(i=0;i<MAX_COLOURS;i++) {
		if (!strcmp(funcnames[i],fname))
			//
			// The cast is not strictly correct and its asking
			// for trouble!
			//
			return((colourf_e)i);
	}
	return(COLOUR_FUNCTION_BACKGROUND);
}

//
// ColourString colourid
//
// Returns an ascii representation of a colour identifier, applying
// the appropriate colour function if necessary. 
// 
// INPUTS : The colour identifier
// RETURNS: An ascii representation of a colour identifier.
//
char *
Colour::ColourString(colour_e col)
{
	int index;
	char buf_fname[1024];
	char buf_col[1024];

	//
	// If owner information has been set but no database preferences
	// have yet to be loaded, then attempt to load them now.
	//
	if (owner && dataloaded==0) {
		dataloaded=1;
		db_prefs->DatabaseRetrievereset();
		for(;;) {
			if (db_prefs->DatabasePrefsAssignread(COLOUR_DB_CLASS,
					owner,buf_fname,buf_col)==-1)
				break;
			index=Colourfunction2index(buf_fname);
			colours[index]=string2colour(buf_col);
		}
	}

	switch(col) {
	case BACKGROUND:
		index=COLOUR_FUNCTION_BACKGROUND;
		break;
	case FOREGROUND:
		index=COLOUR_FUNCTION_FOREGROUND;
		break;
	case COLOUR_FUNCTION1:
		index=COLOUR_FUNCTION_EXT1;
		break;
	case COLOUR_FUNCTION2:
		index=COLOUR_FUNCTION_EXT2;
		break;
	case COLOUR_FUNCTION3:
		index=COLOUR_FUNCTION_EXT3;
		break;
	case COLOUR_FUNCTION4:
		index=COLOUR_FUNCTION_EXT4;
		break;
	case COLOUR_FUNCTION5:
		index=COLOUR_FUNCTION_EXT5;
		break;
	case COLOUR_FUNCTION6:
		index=COLOUR_FUNCTION_EXT6;
		break;
	case COLOUR_FUNCTION7:
		index=COLOUR_FUNCTION_EXT7;
		break;
	case COLOUR_FUNCTION8:
		index=COLOUR_FUNCTION_EXT8;
		break;
	case COLOUR_FUNCTION9:
		index=COLOUR_FUNCTION_EXT9;
		break;
	case COLOUR_FUNCTION10:
		index=COLOUR_FUNCTION_EXT10;
		break;
	case COLOUR_FUNCTION11:
		index=COLOUR_FUNCTION_EXT11;
		break;
	case COLOUR_FUNCTION12:
		index=COLOUR_FUNCTION_EXT12;
		break;
	case COLOUR_FUNCTION13:
		index=COLOUR_FUNCTION_EXT13;
		break;
	case COLOUR_FUNCTION14:
		index=COLOUR_FUNCTION_EXT14;
		break;
	case COLOUR_FUNCTION15:
		index=COLOUR_FUNCTION_EXT15;
		break;
	case COLOUR_FUNCTION16:
		index=COLOUR_FUNCTION_EXT16;
		break;
	case COLOUR_FUNCTION17:
		index=COLOUR_FUNCTION_EXT17;
		break;
	case COLOUR_FUNCTION18:
		index=COLOUR_FUNCTION_EXT18;
		break;
	case COLOUR_FUNCTION19:
		index=COLOUR_FUNCTION_EXT19;
		break;
	case COLOUR_FUNCTION20:
		index=COLOUR_FUNCTION_EXT20;
		break;
	case COLOUR_FUNCTION21:
		index=COLOUR_FUNCTION_EXT21;
		break;
	case COLOUR_FUNCTION22:
		index=COLOUR_FUNCTION_EXT22;
		break;
	case COLOUR_FUNCTION23:
		index=COLOUR_FUNCTION_EXT23;
		break;
	case COLOUR_FUNCTION24:
		index=COLOUR_FUNCTION_EXT24;
		break;
	case COLOUR_FUNCTION25:
		index=COLOUR_FUNCTION_EXT25;
		break;
	case COLOUR_FUNCTION26:
		index=COLOUR_FUNCTION_EXT26;
		break;
	case COLOUR_FUNCTION27:
		index=COLOUR_FUNCTION_EXT27;
		break;
	case COLOUR_FUNCTION28:
		index=COLOUR_FUNCTION_EXT28;
		break;
	case COLOUR_FUNCTION29:
		index=COLOUR_FUNCTION_EXT29;
		break;
	case COLOUR_FUNCTION30:
		index=COLOUR_FUNCTION_EXT30;
		break;
	default:
		return ( colour2string(col) );
	}
	return ( colour2string(colours[index]) );
}


//
// ColourSetname colourfunction colorname
//
// Assigns a name to a particular colour function. A colour function
// allows a colour to be defined indirectly, ie. all references to
// one colour function will give rise to the same colour (within
// a colour instance) and changing the colour function mapping
// will result in a consistent change.
// 
// INPUTS : colourfunction	The colour function whose name is to be set
//	    colourname		The name for the function
// RETURNS: None
//
void
Colour::ColourSetname(colourf_e fnc,char *name)
{
	if (funcnames[fnc])
		free(funcnames[fnc]);

	funcnames[fnc]=(char *)malloc(strlen(name)+1);
	strcpy(funcnames[fnc],name);
}

//
// ColourCopy colourinstance
//
// Copies the colours from the specified instance into the current
// instance.
//
// INPUTS :	colourinstance		The instance to copy from
// RETURNS:	None
//
void
Colour::ColourCopy(class Colour *og)
{
	int i;
	for(i=0;i<MAX_COLOURS;i++)
		if (og->funcnames[i]!=NULL) {
			if (funcnames[i]!=NULL) 
				free(funcnames[i]);
			funcnames[i]=(char *)malloc(strlen(og->funcnames[i])+1);
			strcpy(funcnames[i],og->funcnames[i]);
			colours[i]=og->colours[i];
		}
}

//
// ColourSetowner name
//
// Specifies a name that identifies the owner of the colour instance.
//
// INPUTS :	name		The name of the owner
// RETURNS:	None
//
void
Colour::ColourSetowner(char *name)
{
	if (owner)
		free(owner);

	owner=(char *)malloc(strlen(name)+1);
	strcpy(owner,name);
}

	

//
// Constructor for the colour class
//
Colour::Colour()
{
	int i;

	cselect=NULL;
	for(i=0;i<MAX_COLOURS;i++)
		funcnames[i]=NULL; 

	ColourCopy(&basecolours);
	owner=NULL;
	//
	// We can't load database preferences at this point because
	// the owner information won't be set.  We mark this by setting
	// dataloaded to 0 and detect this condition later on.
	//
	dataloaded=0;
}

//
// Destructor for the colour class
//
Colour::~Colour()
{
	int i;

	for(i=0;i<MAX_COLOURS;i++)
		if (funcnames[i])
			free(funcnames[i]);
	if (owner)
		free(owner);
}

//
///////////////////////////////////////////////////////////////////////////////
//
// Initialisation function
//
///////////////////////////////////////////////////////////////////////////////
//

void
colour_init(menu_bar *mb)
{
	int i;

	MENU("Project","Default colours",colourselect,0);

	//
	// Now create a list of strings that specify all the available
	// colours.
	//
	maxcolour=sizeof(mapper)/sizeof(struct colouremap);
	colourmap=(char **)malloc( sizeof(char *)* (1+maxcolour) );
	for(i=0;i<maxcolour;i++) {
		colourmap[i]=mapper[i].cname;
	}
	colourmap[i]=NULL;

	basecolours.ColourSetname(COLOUR_FUNCTION_FOREGROUND,"Foreground");
	basecolours.ColourSetname(COLOUR_FUNCTION_BACKGROUND,"Background");
	basecolours.ColourSet(COLOUR_FUNCTION_FOREGROUND,BLACK);
	basecolours.ColourSet(COLOUR_FUNCTION_BACKGROUND,WHITE);
	basecolours.ColourSetowner(DEFCOL_DB_CLASS);
}

