
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: colour.h,v 2.3 1995/06/09 20:32:04 davidsi Exp $
//
// $Log: colour.h,v $
// Revision 2.3  1995/06/09  20:32:04  davidsi
// - Added GNU copyright message.
//
// Revision 2.2  1995/06/02  20:04:36  davidsi
// - Support for all standard X colours.
//
// Revision 2.1  1995/04/25  21:11:13  davidsi
// - Added an owner member to the Colour class. This allows a colour instance
//   to be named and therefore to pernament store its colours.  The name
//   can be set with ColourSetowner().
// - Added a dataloaded member. This is set to 0 at construction time, and
//   set as soon as it is possible to load colours from the preferences
//   database.
// - Added Colourfunction2index() to the Colour class to allow function
//   names to be mapped to their index (enumerated) values.
//
// Revision 2.0  1995/02/27  22:06:54  davidsi
// - Initial version.
//
//
//
#ifndef __COLOUR_H
#define __COLOUR_H

//
// Supported colours
//
enum colour_e {
	BACKGROUND,
	FOREGROUND,
	COLOUR_FUNCTION1,
	COLOUR_FUNCTION2,
	COLOUR_FUNCTION3,
	COLOUR_FUNCTION4,
	COLOUR_FUNCTION5,
	COLOUR_FUNCTION6,
	COLOUR_FUNCTION7,
	COLOUR_FUNCTION8,
	COLOUR_FUNCTION9,
	COLOUR_FUNCTION10,
	COLOUR_FUNCTION11,
	COLOUR_FUNCTION12,
	COLOUR_FUNCTION13,
	COLOUR_FUNCTION14,
	COLOUR_FUNCTION15,
	COLOUR_FUNCTION16,
	COLOUR_FUNCTION17,
	COLOUR_FUNCTION18,
	COLOUR_FUNCTION19,
	COLOUR_FUNCTION20,
	COLOUR_FUNCTION21,
	COLOUR_FUNCTION22,
	COLOUR_FUNCTION23,
	COLOUR_FUNCTION24,
	COLOUR_FUNCTION25,
	COLOUR_FUNCTION26,
	COLOUR_FUNCTION27,
	COLOUR_FUNCTION28,
	COLOUR_FUNCTION29,
	COLOUR_FUNCTION30,
	SNOW,
	GHOSTWHITE,
	WHITESMOKE,
	GAINSBORO,
	FLORALWHITE,
	OLDLACE,
	LINEN,
	ANTIQUEWHITE,
	PAPAYAWHIP,
	BLANCHEDALMOND,
	BISQUE,
	PEACHPUFF,
	NAVAJOWHITE,
	MOCCASIN,
	CORNSILK,
	IVORY,
	LEMONCHIFFON,
	SEASHELL,
	HONEYDEW,
	MINTCREAM,
	AZURE,
	ALICEBLUE,
	LAVENDER,
	LAVENDERBLUSH,
	MISTYROSE,
	WHITE,
	BLACK,
	DARKSLATEGRAY,
	DARKSLATEGREY,
	DIMGRAY,
	DIMGREY,
	SLATEGRAY,
	SLATEGREY,
	LIGHTSLATEGRAY,
	LIGHTSLATEGREY,
	GRAY,
	GREY,
	LIGHTGREY,
	LIGHTGRAY,
	MIDNIGHTBLUE,
	NAVY,
	NAVYBLUE,
	CORNFLOWERBLUE,
	DARKSLATEBLUE,
	SLATEBLUE,
	MEDIUMSLATEBLUE,
	LIGHTSLATEBLUE,
	MEDIUMBLUE,
	ROYALBLUE,
	BLUE,
	DODGERBLUE,
	DEEPSKYBLUE,
	SKYBLUE,
	LIGHTSKYBLUE,
	STEELBLUE,
	LIGHTSTEELBLUE,
	LIGHTBLUE,
	POWDERBLUE,
	PALETURQUOISE,
	DARKTURQUOISE,
	MEDIUMTURQUOISE,
	TURQUOISE,
	CYAN,
	LIGHTCYAN,
	CADETBLUE,
	MEDIUMAQUAMARINE,
	AQUAMARINE,
	DARKGREEN,
	DARKOLIVEGREEN,
	DARKSEAGREEN,
	SEAGREEN,
	MEDIUMSEAGREEN,
	LIGHTSEAGREEN,
	PALEGREEN,
	SPRINGGREEN,
	LAWNGREEN,
	GREEN,
	CHARTREUSE,
	MEDIUMSPRINGGREEN,
	GREENYELLOW,
	LIMEGREEN,
	YELLOWGREEN,
	FORESTGREEN,
	OLIVEDRAB,
	DARKKHAKI,
	KHAKI,
	PALEGOLDENROD,
	LIGHTGOLDENRODYELLOW,
	LIGHTYELLOW,
	YELLOW,
	GOLD,
	LIGHTGOLDENROD,
	GOLDENROD,
	DARKGOLDENROD,
	ROSYBROWN,
	INDIANRED,
	SADDLEBROWN,
	SIENNA,
	PERU,
	BURLYWOOD,
	BEIGE,
	WHEAT,
	SANDYBROWN,
	TAN,
	CHOCOLATE,
	FIREBRICK,
	BROWN,
	DARKSALMON,
	SALMON,
	LIGHTSALMON,
	ORANGE,
	DARKORANGE,
	CORAL,
	LIGHTCORAL,
	TOMATO,
	ORANGERED,
	RED,
	HOTPINK,
	DEEPPINK,
	PINK,
	LIGHTPINK,
	PALEVIOLETRED,
	MAROON,
	MEDIUMVIOLETRED,
	VIOLETRED,
	MAGENTA,
	VIOLET,
	PLUM,
	ORCHID,
	MEDIUMORCHID,
	DARKORCHID,
	DARKVIOLET,
	BLUEVIOLET,
	PURPLE,
	MEDIUMPURPLE,
	THISTLE,
	SNOW1,
	SNOW2,
	SNOW3,
	SNOW4,
	SEASHELL1,
	SEASHELL2,
	SEASHELL3,
	SEASHELL4,
	ANTIQUEWHITE1,
	ANTIQUEWHITE2,
	ANTIQUEWHITE3,
	ANTIQUEWHITE4,
	BISQUE1,
	BISQUE2,
	BISQUE3,
	BISQUE4,
	PEACHPUFF1,
	PEACHPUFF2,
	PEACHPUFF3,
	PEACHPUFF4,
	NAVAJOWHITE1,
	NAVAJOWHITE2,
	NAVAJOWHITE3,
	NAVAJOWHITE4,
	LEMONCHIFFON1,
	LEMONCHIFFON2,
	LEMONCHIFFON3,
	LEMONCHIFFON4,
	CORNSILK1,
	CORNSILK2,
	CORNSILK3,
	CORNSILK4,
	IVORY1,
	IVORY2,
	IVORY3,
	IVORY4,
	HONEYDEW1,
	HONEYDEW2,
	HONEYDEW3,
	HONEYDEW4,
	LAVENDERBLUSH1,
	LAVENDERBLUSH2,
	LAVENDERBLUSH3,
	LAVENDERBLUSH4,
	MISTYROSE1,
	MISTYROSE2,
	MISTYROSE3,
	MISTYROSE4,
	AZURE1,
	AZURE2,
	AZURE3,
	AZURE4,
	SLATEBLUE1,
	SLATEBLUE2,
	SLATEBLUE3,
	SLATEBLUE4,
	ROYALBLUE1,
	ROYALBLUE2,
	ROYALBLUE3,
	ROYALBLUE4,
	BLUE1,
	BLUE2,
	BLUE3,
	BLUE4,
	DODGERBLUE1,
	DODGERBLUE2,
	DODGERBLUE3,
	DODGERBLUE4,
	STEELBLUE1,
	STEELBLUE2,
	STEELBLUE3,
	STEELBLUE4,
	DEEPSKYBLUE1,
	DEEPSKYBLUE2,
	DEEPSKYBLUE3,
	DEEPSKYBLUE4,
	SKYBLUE1,
	SKYBLUE2,
	SKYBLUE3,
	SKYBLUE4,
	LIGHTSKYBLUE1,
	LIGHTSKYBLUE2,
	LIGHTSKYBLUE3,
	LIGHTSKYBLUE4,
	SLATEGRAY1,
	SLATEGRAY2,
	SLATEGRAY3,
	SLATEGRAY4,
	LIGHTSTEELBLUE1,
	LIGHTSTEELBLUE2,
	LIGHTSTEELBLUE3,
	LIGHTSTEELBLUE4,
	LIGHTBLUE1,
	LIGHTBLUE2,
	LIGHTBLUE3,
	LIGHTBLUE4,
	LIGHTCYAN1,
	LIGHTCYAN2,
	LIGHTCYAN3,
	LIGHTCYAN4,
	PALETURQUOISE1,
	PALETURQUOISE2,
	PALETURQUOISE3,
	PALETURQUOISE4,
	CADETBLUE1,
	CADETBLUE2,
	CADETBLUE3,
	CADETBLUE4,
	TURQUOISE1,
	TURQUOISE2,
	TURQUOISE3,
	TURQUOISE4,
	CYAN1,
	CYAN2,
	CYAN3,
	CYAN4,
	DARKSLATEGRAY1,
	DARKSLATEGRAY2,
	DARKSLATEGRAY3,
	DARKSLATEGRAY4,
	AQUAMARINE1,
	AQUAMARINE2,
	AQUAMARINE3,
	AQUAMARINE4,
	DARKSEAGREEN1,
	DARKSEAGREEN2,
	DARKSEAGREEN3,
	DARKSEAGREEN4,
	SEAGREEN1,
	SEAGREEN2,
	SEAGREEN3,
	SEAGREEN4,
	PALEGREEN1,
	PALEGREEN2,
	PALEGREEN3,
	PALEGREEN4,
	SPRINGGREEN1,
	SPRINGGREEN2,
	SPRINGGREEN3,
	SPRINGGREEN4,
	GREEN1,
	GREEN2,
	GREEN3,
	GREEN4,
	CHARTREUSE1,
	CHARTREUSE2,
	CHARTREUSE3,
	CHARTREUSE4,
	OLIVEDRAB1,
	OLIVEDRAB2,
	OLIVEDRAB3,
	OLIVEDRAB4,
	DARKOLIVEGREEN1,
	DARKOLIVEGREEN2,
	DARKOLIVEGREEN3,
	DARKOLIVEGREEN4,
	KHAKI1,
	KHAKI2,
	KHAKI3,
	KHAKI4,
	LIGHTGOLDENROD1,
	LIGHTGOLDENROD2,
	LIGHTGOLDENROD3,
	LIGHTGOLDENROD4,
	LIGHTYELLOW1,
	LIGHTYELLOW2,
	LIGHTYELLOW3,
	LIGHTYELLOW4,
	YELLOW1,
	YELLOW2,
	YELLOW3,
	YELLOW4,
	GOLD1,
	GOLD2,
	GOLD3,
	GOLD4,
	GOLDENROD1,
	GOLDENROD2,
	GOLDENROD3,
	GOLDENROD4,
	DARKGOLDENROD1,
	DARKGOLDENROD2,
	DARKGOLDENROD3,
	DARKGOLDENROD4,
	ROSYBROWN1,
	ROSYBROWN2,
	ROSYBROWN3,
	ROSYBROWN4,
	INDIANRED1,
	INDIANRED2,
	INDIANRED3,
	INDIANRED4,
	SIENNA1,
	SIENNA2,
	SIENNA3,
	SIENNA4,
	BURLYWOOD1,
	BURLYWOOD2,
	BURLYWOOD3,
	BURLYWOOD4,
	WHEAT1,
	WHEAT2,
	WHEAT3,
	WHEAT4,
	TAN1,
	TAN2,
	TAN3,
	TAN4,
	CHOCOLATE1,
	CHOCOLATE2,
	CHOCOLATE3,
	CHOCOLATE4,
	FIREBRICK1,
	FIREBRICK2,
	FIREBRICK3,
	FIREBRICK4,
	BROWN1,
	BROWN2,
	BROWN3,
	BROWN4,
	SALMON1,
	SALMON2,
	SALMON3,
	SALMON4,
	LIGHTSALMON1,
	LIGHTSALMON2,
	LIGHTSALMON3,
	LIGHTSALMON4,
	ORANGE1,
	ORANGE2,
	ORANGE3,
	ORANGE4,
	DARKORANGE1,
	DARKORANGE2,
	DARKORANGE3,
	DARKORANGE4,
	CORAL1,
	CORAL2,
	CORAL3,
	CORAL4,
	TOMATO1,
	TOMATO2,
	TOMATO3,
	TOMATO4,
	ORANGERED1,
	ORANGERED2,
	ORANGERED3,
	ORANGERED4,
	RED1,
	RED2,
	RED3,
	RED4,
	DEEPPINK1,
	DEEPPINK2,
	DEEPPINK3,
	DEEPPINK4,
	HOTPINK1,
	HOTPINK2,
	HOTPINK3,
	HOTPINK4,
	PINK1,
	PINK2,
	PINK3,
	PINK4,
	LIGHTPINK1,
	LIGHTPINK2,
	LIGHTPINK3,
	LIGHTPINK4,
	PALEVIOLETRED1,
	PALEVIOLETRED2,
	PALEVIOLETRED3,
	PALEVIOLETRED4,
	MAROON1,
	MAROON2,
	MAROON3,
	MAROON4,
	VIOLETRED1,
	VIOLETRED2,
	VIOLETRED3,
	VIOLETRED4,
	MAGENTA1,
	MAGENTA2,
	MAGENTA3,
	MAGENTA4,
	ORCHID1,
	ORCHID2,
	ORCHID3,
	ORCHID4,
	PLUM1,
	PLUM2,
	PLUM3,
	PLUM4,
	MEDIUMORCHID1,
	MEDIUMORCHID2,
	MEDIUMORCHID3,
	MEDIUMORCHID4,
	DARKORCHID1,
	DARKORCHID2,
	DARKORCHID3,
	DARKORCHID4,
	PURPLE1,
	PURPLE2,
	PURPLE3,
	PURPLE4,
	MEDIUMPURPLE1,
	MEDIUMPURPLE2,
	MEDIUMPURPLE3,
	MEDIUMPURPLE4,
	THISTLE1,
	THISTLE2,
	THISTLE3,
	THISTLE4,
	GRAY0,
	GREY0,
	GRAY1,
	GREY1,
	GRAY2,
	GREY2,
	GRAY3,
	GREY3,
	GRAY4,
	GREY4,
	GRAY5,
	GREY5,
	GRAY6,
	GREY6,
	GRAY7,
	GREY7,
	GRAY8,
	GREY8,
	GRAY9,
	GREY9,
	GRAY10,
	GREY10,
	GRAY11,
	GREY11,
	GRAY12,
	GREY12,
	GRAY13,
	GREY13,
	GRAY14,
	GREY14,
	GRAY15,
	GREY15,
	GRAY16,
	GREY16,
	GRAY17,
	GREY17,
	GRAY18,
	GREY18,
	GRAY19,
	GREY19,
	GRAY20,
	GREY20,
	GRAY21,
	GREY21,
	GRAY22,
	GREY22,
	GRAY23,
	GREY23,
	GRAY24,
	GREY24,
	GRAY25,
	GREY25,
	GRAY26,
	GREY26,
	GRAY27,
	GREY27,
	GRAY28,
	GREY28,
	GRAY29,
	GREY29,
	GRAY30,
	GREY30,
	GRAY31,
	GREY31,
	GRAY32,
	GREY32,
	GRAY33,
	GREY33,
	GRAY34,
	GREY34,
	GRAY35,
	GREY35,
	GRAY36,
	GREY36,
	GRAY37,
	GREY37,
	GRAY38,
	GREY38,
	GRAY39,
	GREY39,
	GRAY40,
	GREY40,
	GRAY41,
	GREY41,
	GRAY42,
	GREY42,
	GRAY43,
	GREY43,
	GRAY44,
	GREY44,
	GRAY45,
	GREY45,
	GRAY46,
	GREY46,
	GRAY47,
	GREY47,
	GRAY48,
	GREY48,
	GRAY49,
	GREY49,
	GRAY50,
	GREY50,
	GRAY51,
	GREY51,
	GRAY52,
	GREY52,
	GRAY53,
	GREY53,
	GRAY54,
	GREY54,
	GRAY55,
	GREY55,
	GRAY56,
	GREY56,
	GRAY57,
	GREY57,
	GRAY58,
	GREY58,
	GRAY59,
	GREY59,
	GRAY60,
	GREY60,
	GRAY61,
	GREY61,
	GRAY62,
	GREY62,
	GRAY63,
	GREY63,
	GRAY64,
	GREY64,
	GRAY65,
	GREY65,
	GRAY66,
	GREY66,
	GRAY67,
	GREY67,
	GRAY68,
	GREY68,
	GRAY69,
	GREY69,
	GRAY70,
	GREY70,
	GRAY71,
	GREY71,
	GRAY72,
	GREY72,
	GRAY73,
	GREY73,
	GRAY74,
	GREY74,
	GRAY75,
	GREY75,
	GRAY76,
	GREY76,
	GRAY77,
	GREY77,
	GRAY78,
	GREY78,
	GRAY79,
	GREY79,
	GRAY80,
	GREY80,
	GRAY81,
	GREY81,
	GRAY82,
	GREY82,
	GRAY83,
	GREY83,
	GRAY84,
	GREY84,
	GRAY85,
	GREY85,
	GRAY86,
	GREY86,
	GRAY87,
	GREY87,
	GRAY88,
	GREY88,
	GRAY89,
	GREY89,
	GRAY90,
	GREY90,
	GRAY91,
	GREY91,
	GRAY92,
	GREY92,
	GRAY93,
	GREY93,
	GRAY94,
	GREY94,
	GRAY95,
	GREY95,
	GRAY96,
	GREY96,
	GRAY97,
	GREY97,
	GRAY98,
	GREY98,
	GRAY99,
	GREY99,
	GRAY100,
	GREY100,
	SCOBACKGROUND,
	SCOFOREGROUND,
	SCOTOPSHADOW,
	SCOACTIVEBACKGROUND,
	SCOACTIVEFOREGROUND,
	SCOACTIVETOPSHADOW,
	SCOALTBACKGROUND,
	SCOHIGHLIGHT,
	COLOUR_NOTSET,
	};

//
// The available colour functions
//
enum colourf_e {
	COLOUR_FUNCTION_BACKGROUND,
	COLOUR_FUNCTION_FOREGROUND,
	COLOUR_FUNCTION_EXT1,
	COLOUR_FUNCTION_EXT2,
	COLOUR_FUNCTION_EXT3,
	COLOUR_FUNCTION_EXT4,
	COLOUR_FUNCTION_EXT5,
	COLOUR_FUNCTION_EXT6,
	COLOUR_FUNCTION_EXT7,
	COLOUR_FUNCTION_EXT8,
	COLOUR_FUNCTION_EXT9,
	COLOUR_FUNCTION_EXT10,
	COLOUR_FUNCTION_EXT11,
	COLOUR_FUNCTION_EXT12,
	COLOUR_FUNCTION_EXT13,
	COLOUR_FUNCTION_EXT14,
	COLOUR_FUNCTION_EXT15,
	COLOUR_FUNCTION_EXT16,
	COLOUR_FUNCTION_EXT17,
	COLOUR_FUNCTION_EXT18,
	COLOUR_FUNCTION_EXT19,
	COLOUR_FUNCTION_EXT20,
	COLOUR_FUNCTION_EXT21,
	COLOUR_FUNCTION_EXT22,
	COLOUR_FUNCTION_EXT23,
	COLOUR_FUNCTION_EXT24,
	COLOUR_FUNCTION_EXT25,
	COLOUR_FUNCTION_EXT26,
	COLOUR_FUNCTION_EXT27,
	COLOUR_FUNCTION_EXT28,
	COLOUR_FUNCTION_EXT29,
	COLOUR_FUNCTION_EXT30
	};
//
// These are publicly available functions
//
extern colour_e string2colour(char *);
extern char *colour2string(colour_e);
extern char *background_colour();
extern char *foreground_colour();

#define MAX_COLOURS	32

//
// The colour class allows other classes to inherit colour management
// functionality
//
class Colour {
	colour_e colours[MAX_COLOURS];	// mappings for colour functions
	char *funcnames[MAX_COLOURS];	// names for colour functions

	class Colourselect *cselect;	// colour select class ptr
	char *owner;			// name of colour owner 
	char dataloaded;		// TRUE when database information
					// has been loaded
public:
	colour_e ColourBackground();
	colour_e ColourForeground();
	colour_e ColourGet(colourf_e);
	void ColourSet(colourf_e,colour_e);
	int ColourSelect();
	char *ColourString(colour_e);
	void ColourSetowner(char *);
	void ColourSetname(colourf_e,char *name);
	void ColourCopy(class Colour *);
	colourf_e Colourfunction2index(char *);

	Colour();
	~Colour();

friend class Colourselect;
};

#endif
