
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: context.h,v 2.2 1995/06/17 20:41:01 davidsi Exp $
//
// $Log: context.h,v $
// Revision 2.2  1995/06/17  20:41:01  davidsi
// - Due to the new statement interpreter, the returnvalue members
//   have been cleaned up since they no longer need to mess with
//   error states.
// - Added ContextAddconstant() to allow constants to be added to
//   the symbol table.
//
// Revision 2.1  1995/04/03  20:41:14  davidsi
// - Removed variableaddr() and added ContextSym2addr() and ContextName2sym()
//   functions member so that names can be converted to symbols which in
//   turn can be converted to addresses without a name->symbol lookup.
//   This improves performance.
// - Renamed addvariable() to ContextAddvariable().
//
// Revision 2.0  1995/01/19  21:45:10  davidsi
// - Initial version.
//
//
#ifndef __CONTEXT_H
#define __CONTEXT_H

#include "setjmp.h"

#include "symbol.h"
#include "value.h"

enum contflag_e {
		cf_offset,		// use offset
		cf_function,		// use function
		cf_addr,		// use actual addr
		cf_type			// symbol is a type definition
	};

//
// Class for symbol table entries
//
class psymbol : public symbol_value {
public:
	contflag_e flag;
	union {
	        int offset;			// offset on heap
		class node_funcdecl *nfuncptr;	// ptr to function decl node
		char *addr;			// address
	} value;
	class value *val;
	//
	// Note that each symbol table entry is given a context
	// so that given a symbol, the context can be readily
	// determined.  Without this, it would be necessary 
	// to search on a context by context basis for a contexts' 
	// symbol even if the symbol was known.
	//
	class context *context;			// context
};


class context {
	char *name;
	char heapref;		// 1 if heap being referenced and should
				// not be popped when first complete 
	int localoffset;	// offset for local variables

	symboltable symvar;	// variable symbol table for this context
	symboltable symtype;	// type symbol table for this context
	symboltable symstruct;	// struct symbol table for this context

	char *heap;		// heap space for this context
	class value *retval;	// return value for this context
	class context *parentcontext;	// parent of this context
private:
	int createheap(int size); 
public:
	char *getheap();
	int createheap(); 
	void popheap();
	void unlock_heap(char *heap);
	char *lock_heap();

	class value *returnvalue();
	void returnvalue(class value *);

	void ContextAddvariable(char *name,class value *val);
	void ContextAddconstant(char *name,class value *val);
	char *ContextSym2addr(class psymbol *pysm,class value *&c_type);
	class psymbol *ContextName2sym(char *vname);

	void addfunction(char *vname,class node_funcdecl *);
	class node_funcdecl *findfunction(char *name);

	void addtype(char *vname,class value *);
	class value *findtype(char *name);

	void addstructure(char *vname,class value *);
	class value *findstructure(char *name);

	void assign(char *name,char *addr);

	context(class context *,char *name);
	~context();
};

#endif
