
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: database.C,v 2.3 1995/06/24 15:01:04 davidsi Exp $
//
// $Log: database.C,v $
// Revision 2.3  1995/06/24  15:01:04  davidsi
// - Added GNU copyright message.
//
// Revision 2.2  1995/06/17  20:37:42  davidsi
// - Added an alternative DatabaseInsert() that can operate with integer
//   values.
//
// Revision 2.1  1995/06/07  20:49:59  davidsi
// - BUG FIX. When inserting new items into the database be sure to
//   to reset the current position with DatabaseRetrievereset().
//
// Revision 2.0  1995/04/25  21:23:48  davidsi
// - Intial version.
//
//
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <dial.h>

#include "localmacros.h"

#include "support.h"
#include "database.h"


//
// The Databaseitem class
//
class Databaseitem : public slist_node {
	char *key;
	char *subkey;
	char *value;

	//
	// Friends
	//
	friend class Database;

	//
	// Public parts of the structure
	//
public:
	int DatabaseitemLoad(class record *);
	//
	// Inheiritated from the slist class
	// 
        virtual void save_record(class record *);

	//
	// Save casts
	//
	class Databaseitem *next() {
		return( LISTREMAP(Databaseitem,slist,next,this) );
	}
        void DatabaseitemSetvalue(char *item);


	Databaseitem();
	Databaseitem(char *key,char *subkey,char *item);
	~Databaseitem();
};


//
// DatabaseitemLoad record
//
// Provides a generic record to security class member mapping function.
// In general classes that are based on the security class can call
// this function with an uninitialized member of that class.
//
// INPUTS :	record		The record to be read and processed
// RETURNS:	0 if record mapped correctly otherwise -1.
//
int
Databaseitem::DatabaseitemLoad(class record *rec)
{
	char *string;
	
	//
	// Flags are separated by '|' and there must be a '|' at the end
	// of the line...
	//
	rec->compile("|");
	key=NULL;
	subkey=NULL;
	value=NULL;
	string=rec->uptomatch();
	if (string==NULL) {
		return 0;
	}
	key=(char *)malloc(strlen(string)+1);
	strcpy(key,string);

	string=rec->uptomatch();
	if (string==NULL) {
		free(key);
		return 0;
	}
	subkey=(char *)malloc(strlen(string)+1);
	strcpy(subkey,string);

	string=rec->uptomatch();
	if (string==NULL) {
		free(subkey);
		free(key);
		return 0;
	}
	value=(char *)malloc(strlen(string)+1);
	strcpy(value,string);

	return 0;
}


//
// Databaseload unused_ptr rec_ptr
//
// Call back function for loading a database item
//
// INPUTS :	rec_ptr 	record pointer for where data is to be
//				loaded from
//		db		pointer to the database list instance
//				where the records are to be inserted
// RETURNS:	None
//
static void
Databaseload(char *,va_list ap)
{
	Databaseitem *newdb=new Databaseitem;
	Database *db;
	class record *rec;

        rec=va_arg(ap,class record *);
        db=va_arg(ap,class Database *);

        if (newdb->DatabaseitemLoad(rec)== -1 )
                delete newdb;
	else
		db->insert_order(newdb);
}


//
// save_security record
//
// Provides a generic security class member to record mapping function.
//
// INPUTS :	record		The record to be created from the security
//				data.
// RETURNS:	None
//
void 
Databaseitem::save_record(class record *rec)
{
	rec->create(key);
	rec->append("|");
	rec->append(subkey);
	rec->append("|");
	rec->append(value);
	rec->append("|");
}

void
Databaseitem::DatabaseitemSetvalue(char *ivalue)
{
	if (value)
		free(value);
	value=(char *)malloc(strlen(ivalue)+1);
	strcpy(value,ivalue);
}

//
// Constructor function.
//
Databaseitem::Databaseitem()
{
	key=NULL;
	subkey=NULL;
	value=NULL;
}

Databaseitem::Databaseitem(char *ikey,char *isubkey,char *ivalue)
{
	key=(char *)malloc(strlen(ikey)+1);
	strcpy(key,ikey);
	subkey=(char *)malloc(strlen(isubkey)+1);
	strcpy(subkey,isubkey);
	value=(char *)malloc(strlen(ivalue)+1);
	strcpy(value,ivalue);
}

//
// Destructor function.
//
Databaseitem::~Databaseitem()
{
	if (key)
		free(key);
	if (subkey)
		free(subkey);
	if (value)
		free(value);
}


//
//=============================================================================
//
// Functions for the Database class
//
//=============================================================================
//


//
// Database::DatabaseInsert key subkey item
//
// Inserts an item into the database. If a value for the <key,subkey>
// pair already exists it will be overwritten.  The disk file will
// be automatically be updated so the change is pernament.
//
// INPUTS : 	key		Primary key
//		subkey		Secondary key
//		item		String for the key pair
// RETURNS: 	None
//
void
Database::DatabaseInsert(char *key,char *subkey,char *item)
{
	class Databaseitem *dbi;

	DatabaseRetrievereset();
	dbi=DatabaseSearch(key,subkey);

	if ( (dbi!=NULL) && (strcmp(dbi->value,item)) ) {
		//
		// Value has changed - remove and resinsert
		//
		dbi->remove();
		delete dbi;
	} else if (dbi!=NULL) {
		//
		// Value has not changed so don't do anything
		//
		return;
	}
	
	dbi=new class Databaseitem(key,subkey,item);
	insert_order(dbi);
	save();
}


//
// Database::DatabaseInsert key subkey item
//
// See above. This version is for integers.
//
// INPUTS : 	key		Primary key
//		subkey		Secondary key
//		item		Integer for the key pair
// RETURNS: 	None
//
void
Database::DatabaseInsert(char *key,char *subkey,int item)
{
	char stritem[256];

	sprintf(stritem,"%d",item);
	DatabaseInsert(key,subkey,stritem);
}

//
// Database::DatabaseInsert key subkey item
//
// See above. This version is for doubles.
//
// INPUTS : 	key		Primary key
//		subkey		Secondary key
//		item		Integer for the key pair
// RETURNS: 	None
//
void
Database::DatabaseInsert(char *key,char *subkey,double item)
{
	char stritem[256];

	sprintf(stritem,"%10.4f",item);
	DatabaseInsert(key,subkey,stritem);
}

//
// DatabaseInsertmultiple key subkey item
//
// Inserts an item into the database. If a value for the <key,subkey>
// pair already exists it will NOT be overwritten.  The disk file will
// be automatically be updated so the change is pernament.
//
// INPUTS : 	key		Primary key
//		subkey		Secondary key
//		item		String for the key pair
// RETURNS: 	None
//
void
Database::DatabaseInsertmultiple(char *key,char *subkey,char *item)
{
	class Databaseitem *dbi;
	dbi=new class Databaseitem(key,subkey,item);
	insert_order(dbi);
	save();
}

//
// DatabaseReplace item
//
// Replaces the item at the current search point which MUST be defined.
//
// INPUTS : 	item		String for the key pair
// RETURNS: 	None
//
void
Database::DatabaseReplace(char *item)
{
	class Databaseitem *dbi;
	dbi=lastpos;
	
	if (lastpos==NULL)
		DialogPanic("Database::DatabaseReplace - no item to replace");

	dbi->DatabaseitemSetvalue(item);
	save();
}


//
// DatabaseDelete
//
// Deletes all entries that match the <key,subkey> pair.
//
// INPUTS :	key		The main key to match
//		subkey		The secondary key to match
// RETURNS:	None
//
void
Database::DatabaseDelete(char *key,char *subkey)
{
	class Databaseitem *dbi;
	class Databaseitem *next_dbi;
	
	lastpos=NULL;
	dbi=head();

	while (!eol(dbi) ) {
		next_dbi=dbi->next();
		if ( (!strcmp(dbi->key,key)) && (!strcmp(dbi->subkey,subkey))){
			dbi->remove();
			delete dbi;
		}
		dbi=next_dbi;
	}
	save();
}


//
// DatabaseSearch
//
// Locates an item in the database for the <key,subkey> pair.
//
// INPUTS : 	key		Primary key
//		subkey		Secondary key
//		item		String for the key pair
// RETURNS:	The item located or NULL
//
class Databaseitem *
Database::DatabaseSearch(char *key,char *subkey)
{
	class Databaseitem *dbi;

	dbi=lastpos;

	if (dbi==NULL)
		dbi=head();
	else
		dbi=dbi->next();

	while (!eol(dbi) ) {
		if ( (!strcmp(dbi->key,key)) && (!strcmp(dbi->subkey,subkey))){
			lastpos=dbi;
			return(dbi);
		}
		dbi=dbi->next();
	}
	lastpos=NULL;
	return(NULL);
}


//
// DatabaseRetrieve key subkey
//
// Locates the FIRST item in the database that matches the keys.
//
// INPUTS :	key		The main key to match
//		subkey		The secondary key to match
// RETURNS:	The data for the keys or NULL if not found
//
char *
Database::DatabaseRetrieve(char *key,char *subkey)
{
	class Databaseitem *dbi;

	lastpos=NULL;
	dbi=DatabaseSearch(key,subkey);

	if (dbi==NULL) 
		return(NULL);
	
	return(dbi->value);
}


//
// DatabaseRetrievenext key subkey
//
// Returns the next record that matches the <key,subkey> pair.
//
// INPUTS :	key		The main key to match
//		subkey		The secondary key to match
// RETURNS:	The data for the keys or NULL if not found
//
char *
Database::DatabaseRetrievenext(char *key,char *subkey)
{
	class Databaseitem *dbi;

	dbi=DatabaseSearch(key,subkey);

	if (dbi==NULL) 
		return(NULL);
	
	return(dbi->value);
}


//
// DatabaseRetrievereset
//
// Resets the search position.
//
// INPUTS :	None
// RETURNS:	None
//
void
Database::DatabaseRetrievereset()
{
	lastpos=NULL;
}

//
// sec_compare sec1 sec2
//
// Call back for ordered list node insertions.
//
// INPUTS :	sec1,sec2	The two securities that are to be compared
// RETURNS:	-1,0,1		On basis of the two names of the securities.
//
//
int
Database::compare(list_node *l1,list_node *l2) 
{
	char *name1,*name2;
	int ret;

	ret=strcmp(((Databaseitem *)l1)->key,((Databaseitem *)l2)->key);

	if (ret)
		return(ret);

	ret=strcmp(((Databaseitem *)l1)->subkey,((Databaseitem *)l2)->subkey);

	return(ret);
}


//
// version
//
// Replace the standard slist version function with one which will
// write the version id for security lists that is being currently
// used.
//
// INPUTS :	None
// RETURNS:	Pointer to version id string
//
char *
Database::version()
{
	return("VERSION:1");
}

//
// Constructor function.
//
Database::Database(char *name)
{
	set_filename(name,WRITE_BY_LINE);
	add_load_vect( Databaseload );
	lastpos=NULL;
}

//
// Destructor function.
//
Database::~Database()
{
}
