
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//    This software is copyrighted (c) 1989 1990 by M.J.E. Mol.         
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

// 
// $Id: date.C,v 2.4 1995/06/17 20:35:09 davidsi Exp $
// 
// $Log: date.C,v $
// Revision 2.4  1995/06/17  20:35:09  davidsi
// - Removed debugging messages.
//
// Revision 2.3  1995/05/18  20:24:30  davidsi
// - Modifications to allow for compilation under SCO OpenServer release 5.0.
//
//
//

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <ctype.h>
#include <sys/types.h>

#include "localmacros.h"

#include "support.h"
#include "date.h"

#define SECSPERDAY      86400
#define TIME2DAY(x)     ((x)/SECSPERDAY)
#define DAY2TIME(x)    ((x)*SECSPERDAY)


//
//                                                                     
//  Name    :  weekday                                                 
//  Author  :  Marcel J.E. Mol                                         
//  Date    :  12/28/89              (first release)                   
//  Version :  3.40                                                    
//  Files   :  weekday.c       Main source file                        
//                                                                     
//                                                                     
//         Weekday outputs line on which it informs you about several  
//         thing of the given <date>.                                  
//         The <date> input may be in several formats:                 
//                                                                     
//             mm/dd/yy      e.g.   10/30/89    or 1/2/510             
//             dd-mm-yy      e.g.   30-10-89    or 2-1-510             
//             yyyy-mm-dd    e.g. 1989-10-30    or 0510-1-2            
//             yy.ddd        e.g.   89.303      or 510.2               
//             `date` output e.g.   Mon Oct 30 99:99:99 1989           
//                                                                     
//         If no date argument is given, weekday uses the system date. 
//         The yyyy-mm-dd date format must have 4 digits for the yyyy  
//         part, 2 for the mm part and 2 for the dd part; it's just    
//         the ISO date standard. (undocumented feature: it also works 
//         if the mm and/or dd part is only one digit long.)           
//         So the following trick does not apply to this date format:  
//         When the year number is less than 100 and the string        
//         forming the year does not start with a 0, the year is       
//         converted to be in the current century. So 10/30/89 is      
//         actually 10/30/1989. If the real year 89 is wanted specify  
//         it as 10/30/089. (the current century is a constant in the  
//         source code. If this programs survives 1999 this constant   
//         must be changed, and the program recompiled.)               
//         Note: make sure that `date` output is passed as one         
//         argument, so enclose it with "s.                            
//                                                                     
//         The output, when no options are given, is something like    
//         this:                                                       
//                                                                     
//         30 October 1989 is on a Monday on julian day 303 in week 44 
//                                                                     
//         or:                                                         
//                                                                     
//         2 January 510 is on a Friday on julian day 2 in week 1      
//                                                                     
//   This software is copyrighted (c) 1989 1990 by M.J.E. Mol.         
//   This code or part of this code may be used by freeware,           
//   public domain or any other non-commercial product. I only ask     
//   you to give me credit for the used code in the documentation      
//   and in your source code.                                          
//                                                                     
//   If you want to use this code or part of this code in your         
//   shareware or any other commercial product, I ask you to contact   
//   me so we can setup an arrangement for the use of the code or part 
//   of the code.                                                      
//                                                                     
//   In any case, I can not be held responsible for any damage this    
//   code or part of this code may cause you or anyone or anything     
//   else.                                                             
//                                                                     
//   Mail bugs to: email: marcel@duteca.tudelft.nl                     
//                 pmail: Marcel J.E. Mol                              
//                        Karel Doormanlaan 58                         
//                        2283AT Rijswijk                              
//                        The Netherlands (Holland)                    
//                 Phone: (+31) 070-3932390                            
//                 Bankaccount: 36.89.35.922 (In the Netherlands)      
//                                                                     
//

#define CENTURY     1900                             // prefix for years
#define LEAP(year)  ((year%4 == 0) && (year%100 != 0) || (year%400 == 0)) 

//
// Global variables
//

char *weekdays[8]   = {"", "Monday", "Tuesday", "Wednesday", "Thursday",
                           "Friday", "Saterday", "Sunday"};
char  days[13]      = {0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};
int   juldays[13]   = {0, 0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334};

static char *months[12] = {
	"Jan","Feb","Mar","Apr",
	"May","Jun","Jul","Aug",
	"Sep","Oct","Nov","Dec"
};

static char *monthnames[12] = {
	"January",  "February","March",   "April" ,
	"May",      "June",    "July",    "August" ,
	"September","October", "November","December"
};


//
///////////////////////////////////////////////////////////////////////////////
//
//                    DATE CONVERSION ROUTINES
//
///////////////////////////////////////////////////////////////////////////////
//                                           


//
// daysinyear year
//
// Returns number of days in a year
//
// INPUTS :	year		The year (full year ie 1995)
// RETURNS:	Number of days in the specified year
///
int
daysinyear(int year)
{
	if ( LEAP(year) )
		return(366);
	return(365);
}

//
// checkdate -- checks if the given date is a valid one.
//              if not, it just exits.
//
int 
date::checkdate( int year,int month,int day)
{
	int dim;

	if (year < 0) 
		return(-1);
 
	if ((month < 1) || (month > 12)) 
		return(-1);
 
	dim = days[month];
	if ((month == 2) && LEAP(year)) 
		dim++;
	if ((day > dim) || (day < 1)) 
		return(-1);

	return(0);

}



// 
// julian2day -- converts year and julian day value to the corresponding 
//               month and day of the month values for that year.
//              these values are returned in the ouput variables month and day.
//
int
date::julian2day( int year,int jday, int &month, int &day)
{
	int leap;

	leap = LEAP(year) ? 1 : 0;
	if (jday > 365+leap || jday <= 0) {
		fprintf(stderr, "error in julian date %d.%d\n",year, jday);
		return( -1 );
	}

	if (jday <= 31) {
		month = 1;
		day = jday;
		return(0);
	}

	jday -= leap;
	if (jday <= 59) {
		month = 2;
		day = jday - 31 + leap;
	} else if (jday <= 90) {
		month = 3;
		day = jday - 59;
	} else if (jday <= 120) {
	        month = 4;
		day = jday - 90;
	} else if (jday <= 151) {
		month = 5;
		day = jday - 120;
	} else if (jday <= 181) {
		month = 6;
		day = jday - 151;
	} else if (jday <= 212) {
		month = 7;
		day = jday - 181;
	} else if (jday <= 243) {
		month = 8;
		day = jday - 212;
	} else if (jday <= 273) {
		month = 9;
		day = jday - 243;
	} else if (jday <= 304) {
		month = 10;
		day = jday - 273;
	} else if (jday <= 334) {
		month = 11;
		day = jday - 304;
	} else {
		month = 12;
		day = jday - 334;
	}

	return(0);

}



// 
// julianday -- converts a year, mont and day value the corresponding julian
//              day number of that year.
//              the julian day is passed as a return value.
//
int
date::julianday( int year, int month,int day)
{
	register int doy;

	if (checkdate(year, month, day)== -1)
		return( -1 );
 
	doy = juldays[month];
	if ((month > 2) && LEAP(year)) 
		doy++;
	doy += day;

	return doy;
} 



//
// weekday -- returns the day of the week number based on the given
//            year, month and day of the month value.
//
int
date::weekday( int year,int month,int day)
{
    register int dow;

	if (checkdate(year, month, day)== -1)
		return( -1 );
 
	dow = 6 +				// the magic number
	      year + 				// shift a day each year 
	      year/4 - year/100 + year/400 +	// shift a day each leap year 
	      juldays[month] + day;		// day of year 

	if ((month < 3) && LEAP(year))		// first 2 months of leap don't
		dow--;				// have the extra leap shift

	dow %= 7;
	return (dow == 0) ? 7 : dow;		// convert 0 to 7
} 


//
// weeknr -- returns the weeknumber of the given date (year, month and day of
//           the month) for that year.
//
int
date::weeknr( int year,int month,int day)
{
	register int fday, jul, weeknum;

	if (checkdate(year, month, day)== -1)
		return( -1 );

	fday = weekday(year, 1, 1);

	jul = julianday(year, month, day);
	weeknum = (jul + fday - 2)/7;
	if (fday <= 4)
		weeknum++;

	if (weeknum == 0)		// go into previous year
	        if ((fday == 5) ||	// week 1 has 4 or 5 days and week
		   ((fday == 6) && LEAP(year-1)) ) // 53 has 4 or 5 days in year-1
			return 53;
		else
			return 52;

	if (weeknum == 53) {		// only a week 53 if we have > 4
		fday = weekday(year, 12, 31);     // days in the week
		if (fday < 4)
			weeknum = 1;
	}

	return weeknum;
} 



//
// scanmmddyy --- scan the date variable in the form MM/DD/YY.
//                MM, DD and YY may be variable sized numbers.
//                output is in output variables year, month and day.
//
void 
date::scanmmddyy( char *date, int &year,int &month,int &day)
{
	char *i;

	if ( sscanf(date, "%d/%d/%d", &month, &day, &year) != 3 ||
	     month <= 0 || day <= 0 || year < 0) {
		fprintf(stderr, "error in mm/dd/yy format date %s\n",date);
		year=month=day=-1;
		return;
	}

	if (year < 100) {
		i = strrchr(date, '/') + 1;
		if (*i != '0')
			year += CENTURY;
	}
	return;
}



//
// scanyyyymmdd --- scan the date variable in the form YYYY-MM-DD.
//                YYYY, MM and DD must be 4, 2 and 2 digits wide.
//                output is in output variables year, month and day.
//
void
date::scanyyyymmdd( char *date, int &year, int &month,int &day)
{

	if ( sscanf(date, "%4d-%2d-%2d", &year, &month, &day) != 3 ||
	     month <= 0 || day <= 0 || year < 0) {
		fprintf(stderr, "error in yyyy-mm-dd format date %s\n",date);
		year=month=day=-1;
		return;
	}

	return;
}


//
// scanddmmyy --- scan the date variable in the form DD-MM-YY.
//                Also YYYY-MM-DD is recognized if YYYY > 31 and DD <= 31.
//                DD, MM and YY may be variable sized numbers.
//                output is in output variables year, month and day.
//
void 
date::scanddmmyy( char *date, int &year, int &month,int &day)
{
	char *i;

	if ( sscanf(date, "%d-%d-%d", &day, &month, &year) != 3 ||
	     month <= 0 || day <= 0 || year < 0) {
		fprintf(stderr, "error in dd-mm-yy format date %s\n",date);
		year=month=day=-1;
		return;
	}

	if (year < 100) {
		i = strrchr(date, '-') + 1;
		if (*i != '0')
			year += CENTURY;
	}

	return;
} 



//
// scanjulian --- scan the date variable in the form YY.DDD.
//                YY and DDD may be variable sized numbers.
//                output is in output variables year, month and day.
//                the return value is tha actual julian day DDD.
//
int
date::scanjulian( char *date, int &year, int &month,int &day)
{
	int jday;

	if (sscanf(date, "%d.%d", &year, &jday) != 2 || year < 0 || jday <= 0) {
		fprintf(stderr, "error in julian date %s\n", date);
		year=month=day= -1;
		return(-1);
	}

	if ((year < 100) && (*date != '0'))
		year += CENTURY;
	if (julian2day(year, jday, month, day)== -1)
		return( -1 );

	return jday;
}



//
// scandate --- scan the date variable in the Unix `date` form.
//              This can look like: "Mon Jan 8 20:10:10 MET 1990"
//              or:                 "Mon Jan 8 20:10:10 1990"
//              the number of spaces between the fields is not fixed
//              (although date output is always in  the same with, see ctime(3).
//              output is in output variables year, month and day.
//              the return value is the day of the week number.
//              If the date input is the NULL pointer, the xurrent system date
//              is taken.
//
int
date::scandate( char *date, int &year, int &month,int &day)
{
	char *i;
	char dayname[4];
	char monname[10];
	int j, k;
	long tim;

	if (date == NULL) {
		tim = time((long *) 0);
		date= ctime(&tim);
	}

	if (sscanf(date, "%3s %3s %d %*s %d",      // date without time zone
			dayname, monname, &day, &year) != 4 &&
		sscanf(date, "%3s %3s %d %*s %*s %d", // date with time zone
                     dayname, monname, &day, &year) != 4) {
		fprintf(stderr, "error in unix `date` format date %s\n",date);
		year=month=day= -1;
		return(-1);
	}

	for(j=1; j<12; j++) {
		if (strncmp(monname, monthnames[j-1], 3) == 0)
			break;
	}
	if (strncmp(monname, monthnames[j-1], 3) != 0) {
		fprintf(stderr, "error in unix `date` format date %s\n",date);
		year=month=day= -1;
		return(-1);
	}
	month = j;
	for(j=1; j<7; j++) {
		if (strncmp(dayname, weekdays[j], 3) == 0)
			break;
	}
	if (strncmp(dayname, weekdays[j], 3) != 0) {
		fprintf(stderr, "error in unix `date` format date %s\n",date);
		year=month=day= -1;
		return(-1);
	}

	return j;
}




//
// 
//
// Converts a date string monthname dd yy or dd monthname year  to 
// seconds from 1970
//
// INPUTS :	str		A pointer to the string to convert
// RETURNS:	A long value for number of seconds or -1 for an error
//
int
date::scanenglish( char *str,int &year, int &month, int &day )
{
char b[10],c[10],d[10];
int iter;

	extern char * strcpy_2_space(char *,char *,int);
	extern int  alpha_strcmp(char *,char *);

	str=strcpy_2_space(b,str,10);
	str=strcpy_2_space(c,str,10);
	str=strcpy_2_space(d,str,10);

	for(iter=0;iter<12;iter++) {
		if ( !alpha_strcmp(b,months[iter]) 
		   || !alpha_strcmp(b,monthnames[iter]) )
		break;
	}

	if (iter!=12)  {		/* date was in MONTH DAY YEAR form */
		year=atoi(d);
		if (year<100)
			year += CENTURY;
		day=atoi(c);
		month=iter+1;
		return(0);
	}

	for(iter=0;iter<12;iter++) {
		if ( !alpha_strcmp(c,months[iter])
		   ||!alpha_strcmp(c,monthnames[iter]) )
			break;
	}
	if (iter!=12)  {	/* date was in DAY MONTH YEAR form */
		month=iter+1;
		year=atoi(d);
		if (year<100)
			year += CENTURY;
		day=atoi(b);
		return(0);
	}
	//
	// Not a date in simple english form
	//
	year=month=day= -1;
	return(-1);
}






#ifdef NEVER
//
// day_number year month day
//
// Returns the day of the week for the given date.
//
// INPUTS :	year			year of the date in question
//		month			month of the date in question
//		day			day of the date in question
// RETURNS:	Day of week
//
int
date::day_number( int Year, int month, int Day)
{
	long days_ctr;

	Year -= 1900;
	days_ctr = ((long)Year * 365L) + ((Year + 3) / 4);
	days_ctr += juldays[month] + Day + 6;
	if (month > 1 && (Year % 4 == 0))
		days_ctr++;
	return (int)(days_ctr % 7L);
}
int
date::time_to_year( unsigned long *time )
{
	struct tm *tm;       /* time struct */

	tm=localtime((long *)time);
	return(tm->tm_year+1900);
}




//
// month_name month
// 
// Returns a month given a month number
//
// INPUTS :	month		Month number 0-11
// RETURNS:	ascii string of the req'd month.
//
char *
date::month_name(int month)
{
    return(monthnames[month]);
}


#endif


//
// setdate year month day
//
// Sets the date to the specified value using day, month, year components.
// 
// INPUTS :	year		Year part of date
//		month		Month part of date
// 		day		Day part of date
// RETURNS:	If an error is detected -1 is returned, else 0.
//
int
date::setdate(
	int day,		// day of month 1-... 
	int month,		// month 1-12
	int year		// year from 1970 
	)
{
register int iter;		// iteration value for loops
register long t_days=0;		// number of days so far computed
int dinyear;			// number of days in target year 


	//
	// Make jan 0 and 1st of a month zero.
	//
	day--;
	month--;

	dinyear=daysinyear(year);
	if ( (year<1970 || year>2030 ||  month<0 || month>11) ||
		//
		// Make some use of of the law of implication!
		// - i.e  A=>B = !A|B
		// So there is an error if day>size of month AND
		// if leap year then month!=1 or day>28.
 		//
	    (day>=days[month+1]&&(dinyear!=366 || (month!=1 || day>28)))) {
		return(-1); 
	}
			
	for(iter=1970;iter<year;iter++)	// get num of days to start of year
		t_days+=daysinyear(iter);

	for(iter=0;iter<month;iter++)	// get num of days to start of month
		t_days+=days[iter+1];

	if (dinyear==366 && month>1)	// adjust for leap years
		t_days++;

	timestamp=(t_days+day)*86400;	// add days and convert to seconds

	return(0);
}

//
// setdate string
//
// Sets the date to the specified value using a string that can take
// several formats:
//
//	mm/dd/yy		e.g. 10/30/89 or 1/2/510
//	dd-mm-yy		e.g. 30-10-89  or 2-1-510
//	yyyy-mm-dd		e.g. 1989-10-30 or 0510-1-2 
//	yy.ddd			e.g. 89.303 or 510.2 
//	`date` output	 	e.g. Mon Oct 30 99:99:99 1989
//	monthname dd yy		e.g. Jan 20 1994 or January 20 1994
//	dd monthname yy		e.g. 20 Jan 1994 or 20 January 1994
// 
// INPUTS :	String		See above
// RETURNS:	If an error is detected -1 is returned, else 0.
//
int
date::setdate(char *date)
{
	int year, month, day, week;
	char *ptr;

	for(ptr=date;*ptr!='\0';ptr++)
		if ( !numeric(*ptr) )
			break;

	if (*ptr=='\0')	{			// seconds since 1970
		timestamp=atol(date);
		return(0);
	} else if (strchr(date, '/') != NULL)	// mm/dd/yy format date
		scanmmddyy(date, year, month, day);
	else if (strchr(date, '-') == date+4)	// yyyy-mm-dd format
		scanyyyymmdd(date, year, month, day);
	else if (strchr(date, '-') != NULL)	// dd-mm-yy format date
		scanddmmyy(date, year, month, day);
	else if (strchr(date, '.') != NULL) 	// yyyy.ddd format date
		scanjulian(date, year, month, day);
	else if (scanenglish(date,year, month, day)==-1)
		scandate(date, year, month, day);

	if (year==-1 || month==-1 || day==-1)
		return(-1);

	return(setdate(day,month,year));

}

//
// setdate 
//
// Sets the date to todays date
//
// INPUTS :	None
// RETURNS:	If an error is detected -1 is returned, else 0.
// 
int
date::setdate()
{
	int year, month, day;

        if ( scandate(NULL, year, month, day) == -1 )
                return( -1);
	return(setdate(day,month,year));
}

//
// setdate 
//
// Sets the date to the given date
//
// INPUTS :	None
// RETURNS:	Always 0 - cannot be an error!
// 
int
date::setdate(class date &indate)
{
	timestamp=indate.timestamp;
	return(0);
}

//
// Returns a date in ascii form
//
// INPUTS:	None
// RETURNS:	A string in Day Monthname Year format
char *
date::getdate()
{
	char *timebuf; 
	struct tm *tm;
	int i;

	tm=localtime((long *)&timestamp);
	timebuf=asctime(tm);
	timebuf[i=strlen(timebuf)-1]='\0';
	strcpy(&timebuf[i-13],&timebuf[i-4]);

	return(&timebuf[4]);
}


//
// inrange startdate enddate
//
// Determines if a date lies within a range of dates.
// 
// INPUTS:	startdate	The first date that is in the range
//		enddate		The final date that is in the range
// RETURNS:	TRUE if the date is within the specified range, else FALSE.
//
int 
date::inrange(class date *startd,class date *endd)
{

	if (startd->timestamp<=timestamp && timestamp<=endd->timestamp)
		return(1);
	else
		return(0);
}

int
date::daysince1970()
{
	return (TIME2DAY(timestamp));
}

void
date::daysince1970(int day)
{
	timestamp=DAY2TIME(day);
}


void
date::getdate(int &d,int &m,int &y)
{
struct tm *tm;       // time struct

        tm=localtime((long *)&timestamp);

	d=tm->tm_mday;
	m=tm->tm_mon+1;
	y=tm->tm_year+1900;
}

void
date::add(int days)
{
	unsigned long otimestamp=timestamp;

	timestamp += DAY2TIME(days);
	//
	// Don't allow wraparound
	//
	if (timestamp<otimestamp) 
		timestamp=(unsigned long) -1;
}

void
date::add(class date *operand)
{
	timestamp += operand->timestamp;
}

void
date::sub(int days)
{
	//
	// Don't allow wraparound
	//
	if (timestamp>=DAY2TIME(days))
		timestamp -= DAY2TIME(days);
	else
		timestamp = 0;
}

void
date::sub(class date *operand)
{
	timestamp -= operand->timestamp;
}


int
date::days_in_month()
{
	int d,month,year;
	int day;

	getdate(d,month,year);
	day = days[month];
	if (month == 2 && year%4 == 0 && (year%100 != 0 || year%400 == 0))
		day++;
	return day;
	
}

int
date::compare(class date *d2)
{
	if (timestamp==d2->timestamp)
		return(0);
	if (timestamp>d2->timestamp)
		return(1);
	if (timestamp<d2->timestamp)
		return(-1);
}

date::date()
{
	timestamp=0;
}

date::~date()
{
}
