
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: dchange.h,v 2.4 1995/05/18 20:54:13 davidsi Exp $
//
// $Log: dchange.h,v $
// Revision 2.4  1995/05/18  20:54:13  davidsi
// - Callbacks are now stored using opaque handles rather than pointers to
//   the actual nodes (see callback.C for the whole story).
//
// Revision 2.3  1995/04/16  20:33:50  davidsi
// - Added callback_sec member to allow price change displays to be updated
//   when securities are added, removed or modified.
//
// Revision 2.2  1995/02/27  22:01:00  davidsi
// - Added a Semaphore member, that is used to protect against multiple
//   updates occurring.
// - Added further flags used to specify which types of change are to be
//   displayed.
//
// Revision 2.1  1995/01/31  21:53:34  davidsi
// - Revised to use the textdialog class.  This results in much of
//   the required functionality now being inherited.
// - Added a callback to handle transactions - if a transaction
//   occurs it could potentially require a change to the displayed
//   information.
//
// Revision 2.0  1995/01/19  22:00:08  davidsi
// - Initial version.
//
//

#ifndef __DCHANGE_H
#define __DCHANGE_H

#include <stdio.h>
#include <string.h>

#include "textdialog.h"
#include "date.h"
#include "semaphore.h"


class change_display : public Textdialog {
	unsigned long flags;			// see DCHG_FLAGS_XXX
	unsigned long lastflags;		// DCHG_FLAGS_XXX for last update
	Callbackid callback_price;		// callback when prices
						// are changed
	Callbackid callback_trans;		// callback when a transaction
						// occurs
	Callbackid callback_sec;		// callback for security changes

	int setmenuflag;			// TRUE when menu defined
	Semaphore update;			// Update underway semaphore 
	Semaphore colourdialog;			// colour dialog display sem.

private:
	virtual void MenuCallback(unsigned long code, int onflag);
	static void redisplay(char *realthis,char *);
	void list_change();
public:
	change_display();
	~change_display();
};

#define DCHG_FLAGS_PORT		0x1	// display only securities in portfolio
#define DCHG_FLAGS_ALL		0x2	// display all securities
#define DCHG_FLAGS_ALPHA	0x4	// use alpha ordering for the display
#define DCHG_FLAGS_NUML		0x8	// use losers first order on display
#define DCHG_FLAGS_NUMG		0x10	// use gainers first order on display
#define DCHG_FLAGS_SHOWG	0x20	// show gainers
#define DCHG_FLAGS_SHOWL	0x40	// show losers
#define DCHG_FLAGS_SHOWN	0x80	// show non movers
#define DCHG_FLAGS_EXIT		0x100	// exit pending

#endif
