
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: ddiv.h,v 2.6 1995/05/18 21:04:20 davidsi Exp $
//
// $Log: ddiv.h,v $
// Revision 2.6  1995/05/18  21:04:20  davidsi
// - Callbacks are now stored using opaque handles rather than pointers to
//   the actual nodes (see callback.C for the whole story).
//
// Revision 2.5  1995/04/16  20:30:29  davidsi
// - Added callback_sec member to allow dividend displays to be updated when
//   securities are added, removed or modified.
//
// Revision 2.4  1995/04/03  20:40:59  davidsi
// - BUG FIX. The #define flag protecting this file from multiple inclusion
//   was incorrect.
//
// Revision 2.3  1995/03/08  21:23:15  davidsi
// - BUG FIX. Added redisplay() to handle callbacks for when dividends are
//   added.
//
// Revision 2.2  1995/02/27  22:27:20  davidsi
// - Added a semaphore for the colour dialog that is now supported.
//
// Revision 2.1  1995/01/31  21:41:58  davidsi
// - Revised to use the textdialog class.  This results in much of
//   the required functionality now being inherited.
//
// Revision 2.0  1995/01/19  21:53:52  davidsi
// - Removed the time members of the dividends_display class.  Dates are now
//   managed using the date class.
//
// Revision 1.2  1994/08/06  16:41:15  davidsi
// - Changed times from long to unsigned long.
//
// Revision 1.1  1994/06/28  21:00:30  davidsi
// Initial revision
//
//
#ifndef __DDIV_H
#define __DDIV_H

#include <stdio.h>
#include <string.h>

#include "div.h"
#include "semaphore.h"
#include "callback.h"
#include "textdialog.h"

class dividends_display : public Textdialog {
	class date start_date;			// Start date for dividends
	class date end_date;			// End date for dividends
	unsigned long optionflags;		// Options for display
	char setmenuflag;			// has menu/title been set
	Semaphore colourdialog;			// colour dialog display sem.
	Callbackid callback_div;		// callback for dividend changes
	Callbackid callback_sec;		// callback for security changes
	
	//
private:
	virtual void MenuCallback(unsigned long code, int onflag);
	void create_display();
	static void redisplay(char *tmpthis,char *);

	//
public:
	dividends_display();
	~dividends_display();
};

#endif
