/* CHK=0x7B68 */
/*+-------------------------------------------------------------------------
	det_table.c - UNIX V/386 system monitor table usage detail
	martin@hppcmart.grenoble.hp.com

00000000001111111111222222222233333333334444444444555555555566666666667777777777
01234567890123456789012345678901234567890123456789012345678901234567890123456789
Resource     actual total ovfl %used -------------------------------------------
#################### ###### ##### #### #########################################
  Defined functions:
files_in_use()
inodes_in_use()
update_table_line(win, y, x, width, num_item, max_item, conf_item, percent_item)
init_table()
draw_table_literals(win,y,x)

--------------------------------------------------------------------------*/
/*+:EDITS:*/
/*:06-30-1995-19:50-guym@sco.com-SCO open server 5 support added */
/*:08-03-1993-15:50-wht@n4hgf-source control point x2.70 */
/*:07-15-1992-14:31-wht@n4hgf-2.60 release - u386mon+siotools merge */
/*:08-20-1991-12:44-root@n4hgf-nba@sysware.dk S5R31 and config reorg */
/*:08-01-1991-23:34-wht@n4hgf-release 2.40 source control point */
/*:05-15-1991-17:22-wht@n4hgf-2.3 patches for SVR31 from nba@sysware.dk */
/*:02-07-1991-13:30-martin@hppcmart-creation */

#include "config.h"

#if defined(DPT_TABLE)

#include <curses.h>
#include "libpanel.h"
#include <string.h>
#include <nlist.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/var.h>
#undef NGROUPS_MAX
#undef NULL
#include <sys/param.h>
#include <sys/inode.h>
#include <sys/file.h>
#include <sys/fcntl.h>
#include <sys/flock.h>
#include <sys/sysinfo.h>
#include <sys/immu.h>
#include <sys/region.h>
#include <sys/callo.h>

#include "nlsym.h"
#include "libkmem.h"
#include "libmem.h"
#include "libswap.h"
#include "libnlsym.h"
#include "u386mon.h"

extern struct var v;
extern int procs_alive;
struct file *fpbase;
struct inode *ipbase;
struct flckinfo flckinfo;
struct syserr syserr;
struct region *rgbase;
struct callo *callbase;
int max_files = 0;
int max_inodes = 0;
int max_locks = 0;
int max_regions = 0;
int max_callouts = 0;
int max_procs = 0;

/*+-------------------------------------------------------------------------
	files_in_use()

	Returns the number of files in use or -1 if the free list
	pointers were munged when read which happens every now and then.
--------------------------------------------------------------------------*/
#ifdef SCO325
static int
files_in_use()
{
    unsigned num_free;

    kread((caddr_t) &num_free, filefreecountaddr, sizeof(unsigned) );
    return (v.v_file - num_free);
}
#else
static int
files_in_use()
{
    struct file *fp, *fflist;
    register int cnt = 0;
    unsigned int ndx;

    kread((caddr_t) fpbase, fileaddr, sizeof(struct file) * v.v_file);
    kread((caddr_t) & fflist, ffreelistaddr, sizeof(struct file *));

    fp = fpbase + (fflist - (struct file *)fileaddr);
    while (1)
    {
	cnt++;
	if (fp->f_next == 0)
	    break;
	ndx = fp->f_next - (struct file *)fileaddr;
	if ((ndx >= v.v_file) || (ndx < 0) || (cnt > v.v_file))
	    return (-1);
	fp = fpbase + ndx;

    }
    return (v.v_file - cnt);
}
#endif
/*+-------------------------------------------------------------------------
	inodes_in_use()

	Returns the number of inodes used or -1 if the free list
	pointers were munged when read which happens every now and then.
--------------------------------------------------------------------------*/
#ifdef SCO325
static int
inodes_in_use()
{
    unsigned num_free;

    kread((caddr_t) &num_free, inodefreecountaddr, sizeof(unsigned) );
    return (v.v_inode - num_free);

}
#else
static int
inodes_in_use()
{
    struct inode *ip, *iflistend;
#ifdef SCO325
    struct inode iflist;				/* LXXX */
#else
    struct ifreelist iflist;
#endif
    register int cnt = 0;
    int ndx;

    kread((caddr_t) ipbase, inodeaddr, sizeof(struct inode) * v.v_inode);
#ifdef SCO325
    kread((caddr_t) & iflist, ifreelistaddr, sizeof(struct inode));
#else
    kread((caddr_t) & iflist, ifreelistaddr, sizeof(struct ifreelist));
#endif

    ip = ipbase + (iflist.av_forw - (struct inode *)inodeaddr);
    iflistend = ipbase + (iflist.av_back - (struct inode *)inodeaddr);
    while (1)
    {
	cnt++;
	if (ip == iflistend)
	    break;
	ndx = ip->av_forw - (struct inode *)inodeaddr;
	if ((ndx >= v.v_inode) || (ndx < 0) || (cnt > v.v_inode))
	    return (-1);
	ip = ipbase + ndx;
    }
    return (v.v_inode - cnt);
}
#endif
/*+-------------------------------------------------------------------------
	regions_in_use()

	Returns the number of regions used or -1 if the free list
	pointers were munged when read which happens every now and then.
--------------------------------------------------------------------------*/
#ifdef SCO325
static int
regions_in_use()
{
     unsigned num_free;

    kread((caddr_t) &num_free, regionfreecountaddr, sizeof(unsigned) );
    return (v.v_region - num_free);
}
#else
static int
regions_in_use()
{
    struct region *rg, ractive;
    register int cnt = 0;
    int ndx;

    kread((caddr_t) rgbase, regionaddr, sizeof(struct region) * v.v_region);
    kread((caddr_t) & ractive, ractiveaddr, sizeof(struct region));

    rg = rgbase + (ractive.r_forw - (struct region *)regionaddr);
    while (1)
    {
	cnt++;
	if (rg->r_forw == (struct region *)ractiveaddr)
	    break;
	ndx = rg->r_forw - (struct region *)regionaddr;
	if ((ndx >= v.v_region) || (ndx < 0) || (cnt > v.v_region))
	    return (-1);
	rg = rgbase + ndx;
    }
    return (cnt);
}
#endif
/*+-------------------------------------------------------------------------
	callouts_in_use()

	Returns the number of callouts used.
--------------------------------------------------------------------------*/
#ifdef SCO325
static int
callouts_in_use(int *num_free)
{
    unsigned num_callouts;

    kread((caddr_t) &num_callouts, numcalloutaddr, sizeof(unsigned) );
    kread((caddr_t) num_free, numfreecalloutaddr, sizeof(unsigned) );
    return (num_callouts);
}
#else
static int
callouts_in_use()
{
	struct callo *call;
	register int cnt = 0;
	register int ndx = 0;

	kread((caddr_t) callbase, calloutaddr, sizeof(struct callo) * v.v_call);

	call = callbase;
	while (ndx < v.v_call)
	{
		/* This is a guess - I'm checking it with SCO */
		if (call->c_id > 0)
			cnt++;
		ndx++;
		call++;
	}
	return (cnt);
}
#endif /* DSCO325 */

/*+-------------------------------------------------------------------------
	update_table_line(win)
00000000001111111111222222222233333333334444444444555555555566666666667777777777
01234567890123456789012345678901234567890123456789012345678901234567890123456789
tual total ovfl %used ------------------------------------------------
####  ####  ###   ### xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
--------------------------------------------------------------------------*/

#define _TABLESCALE_AC	0
#define _TABLESCALE_TO	6
#define _TABLESCALE_OV	12
#define _TABLESCALE_PE	18
#define _TABLESCALE_SX	22

static void
update_table_line(win, y, x, width, num_item, max_item, conf_item, percent_item, ovfl_item)
WINDOW *win;
int y;
int x;
int width;
int num_item;
int max_item;
int conf_item;
int percent_item;
int ovfl_item;

{
    int itmp, max;

    if (num_item < 0)
    {

	wmove(win, y, x + _TABLESCALE_AC);
	use_cp(win, cpHIGH);
	waddstr(win, "OVER");

    }
    else
    {

	wmove(win, y, x + _TABLESCALE_AC);
	use_cp(win, cpINFO);
	wprintw(win, "%4ld", num_item);

	wmove(win, y, x + _TABLESCALE_TO);
	use_cp(win, cpINFO);
	wprintw(win, "%4ld", conf_item);

	if (ovfl_item != -1)
	{
	    wmove(win, y, x + _TABLESCALE_OV);
	    use_cp(win, cpINFO);
	    wprintw(win, "%3ld", ovfl_item);
	}

	wmove(win, y, x + _TABLESCALE_PE);
	if (percent_item < 70)
	    use_cp(win, cpLOW);
	else if (percent_item < 90)
	    use_cp(win, cpMED);
	else
	    use_cp(win, cpHIGH);
	wprintw(win, "%3ld", percent_item);

	wmove(win, y, x + _TABLESCALE_SX);

	itmp = (width * percent_item) / 100;
	max = (width * max_item) / 100;
	while (-1)
	{
	    itmp--;
	    max--;
	    if (max < 1)
	    {
		use_cp(win, cpHIGH);
		waddch(win, (chtype) 'M');
		break;
	    }
	    if (itmp > -1)
		waddch(win, (chtype) '-');
	    else
		waddch(win, (chtype) ' ');
	}
	wclrtoeol(win);
    }
}

/*+-------------------------------------------------------------------------
	init_table()

--------------------------------------------------------------------------*/

void
init_table()

{

#ifndef SCO325
    if (!(rgbase = (struct region *)malloc(sizeof(struct region) * v.v_region)))
	    leave_text("cannot alloc memory for region table", 1);
    if (!(fpbase = (struct file *)malloc(sizeof(struct file) * v.v_file)))
	    leave_text("cannot alloc memory for file table", 1);
    if (!(ipbase = (struct inode *)malloc(sizeof(struct inode) * v.v_inode)))
	    leave_text("cannot alloc memory for inode table", 1);
    if (!(callbase = (struct callo *)malloc(sizeof(struct callo) * v.v_call)))
	    leave_text("cannot alloc memory for region table", 1);
#endif
}

/*+-------------------------------------------------------------------------
	update_table(win,y,x)

--------------------------------------------------------------------------*/

void
update_table(win, y, x)
WINDOW *win;
int y;
int x;
{
    int num_files;
    int num_inodes;
    int num_regions;
    int num_callouts;
    int num_free_callouts;
    int percent_files;
    int percent_inodes;
    int percent_locks;
    int percent_regions;
    int percent_callouts;
    int percent_procs;
#ifdef SCO325    
    int v_call;
#endif
    register width;

    x += 15;		     /* skip literals */
    width = COLS - _TABLESCALE_SX - x;


    num_files = files_in_use();
    num_inodes = inodes_in_use();
    num_regions = regions_in_use();
#ifdef SCO325
    num_callouts = callouts_in_use(&num_free_callouts);
    v_call = num_callouts + num_free_callouts;
#else
    num_callouts = callouts_in_use();
    v_call = v.v_call;
#endif
    kread((caddr_t) & flckinfo, flckinfoaddr, sizeof(struct flckinfo));

    grok_proc();
    kread((caddr_t) & syserr, syserraddr, sizeof(struct syserr));

    percent_files = (num_files * 100 / v.v_file);
    percent_inodes = (num_inodes * 100 / v.v_inode);
    percent_regions = (num_regions * 100 / v.v_region);
    percent_locks = flckinfo.recs ? (flckinfo.reccnt * 100 / flckinfo.recs) : 0;
    percent_callouts = (num_callouts * 100 / v_call);
    percent_procs = (procs_alive * 100 / v.v_proc);
    if (percent_files > max_files)
	max_files = percent_files;
    if (percent_inodes > max_inodes)
	max_inodes = percent_inodes;
    if (percent_locks > max_locks)
	max_locks = percent_locks;
    if (percent_regions > max_regions)
	max_regions = percent_regions;
    if (percent_callouts > max_callouts)
	max_callouts = percent_callouts;
    if (percent_procs > max_procs)
	max_procs = percent_procs;

    update_table_line(win, y, x, width,
	  num_files, max_files, v.v_file, percent_files, syserr.fileovf);
    update_table_line(win, y + 1, x, width,
     num_inodes, max_inodes, v.v_inode, percent_inodes, syserr.inodeovf);
    update_table_line(win, y + 2, x, width,
		      flckinfo.reccnt, max_locks, flckinfo.recs, percent_locks, flckinfo.recovf);
    update_table_line(win, y + 3, x, width,
	      num_regions, max_regions, v.v_region, percent_regions, -1);
    update_table_line(win, y + 4, x, width,
	     num_callouts, max_callouts, v_call, percent_callouts, -1);
    update_table_line(win, y + 5, x, width,
	procs_alive, max_procs, v.v_proc, percent_procs, syserr.procovf);
    return;
}			     /* end of update_table */

/*+-------------------------------------------------------------------------
	draw_table_literals(win)
00000000001111111111222222222233333333334444444444555555555566666666667777777777
01234567890123456789012345678901234567890123456789012345678901234567890123456789
Resource     actual total ovfl %used -------------------------------------------
############   ####  ####  ###   ### xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
--------------------------------------------------------------------------*/
void
draw_table_literals(win, y, x)
WINDOW *win;
int y;
int x;
{
    int x2 = x;

    wmove(win, y, x);
    use_cp(win, cpBANNER);
    waddstr(win, "Resource     actual total ovfl %used ");
    getyx(win, y, x2);
    while (x2 < COLS)
	waddch(win, (chtype) '-'), x2++;
    use_cp(win, cpLIT);
    wmove(win, y + 1, x);
    waddstr(win, "File table ");
    wmove(win, y + 2, x);
    waddstr(win, "Inode table");
    wmove(win, y + 3, x);
    waddstr(win, "Lock table ");
    wmove(win, y + 4, x);
    waddstr(win, "Region table ");
    wmove(win, y + 5, x);
    waddstr(win, "Callout table ");
    wmove(win, y + 6, x);
    waddstr(win, "Proc table ");

}			     /* end of draw_table_literals */

#endif /* DPT_TABLE */
