/* CHK=0x7811 */
/*+-------------------------------------------------------------------------
	det_wd.c - UNIX V/386 system monitor WD disk controller usage detail
	martin@hppcmart.grenoble.hp.com

00000000001111111111222222222233333333334444444444555555555566666666667777777777
01234567890123456789012345678901234567890123456789012345678901234567890123456789
Device  Av Q  r+w/s Blks/s -- (% Utilization) ----------------------------------
######  ##### ##### #####  #####################################################
  Defined functions:

--------------------------------------------------------------------------*/
/*+:EDITS:*/
/*:08-03-1993-15:50-wht@n4hgf-source control point x2.70 */
/*:07-15-1992-14:31-wht@n4hgf-2.60 release - u386mon+siotools merge */
/*:08-20-1991-12:44-root@n4hgf-nba@sysware.dk S5R31 and config reorg */
/*:08-01-1991-23:34-wht@n4hgf-release 2.40 source control point */
/*:02-07-1991-13:30-martin@hppcmart-creation */

#include "config.h"

#if defined(DPT_WD)

#include <curses.h>
#include "libpanel.h"
#include <string.h>
#include <nlist.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ascii.h>
#include <sys/var.h>
#undef NGROUPS_MAX
#undef NULL
#include <sys/param.h>
#include <sys/elog.h>

#include "nlsym.h"
#include "libkmem.h"
#include "libmem.h"
#include "libswap.h"
#include "libnlsym.h"
#include "u386mon.h"

extern struct var v;

/* WD Driver */
#define NCTLR		2
#define NWD			2
#define NWDTOTAL	NCTLR*NWD
/* Floppy drives */
#define NDRIVES		4

static int max_util[NWDTOTAL + NDRIVES];
static struct iotime ia[NWDTOTAL + NDRIVES], ialast[NWDTOTAL + NDRIVES];
static time_t timelast, now;

/*+----------------------------------------------------------------------------
	init_wd();
	Get initial drive stats.
-----------------------------------------------------------------------------*/
void
init_wd()
{
    register i;

    (void)time(&timelast);
    for (i = 0; i < (NWDTOTAL + NDRIVES); i++)
	max_util[i] = 0;
    kread((caddr_t) ialast, wd1010statsaddr, sizeof(struct iotime) * NCTLR * NWD);
    kread((caddr_t) & ialast[NWDTOTAL], flstatsaddr, sizeof(struct iotime) * NDRIVES);
}

/*+----------------------------------------------------------------------------
	draw_wd_literals(win, y, x);

00000000001111111111222222222233333333334444444444555555555566666666667777777777
01234567890123456789012345678901234567890123456789012345678901234567890123456789
Device  Jobs/s r+w/s Blks/s -- (% Utilization) ---------------------------------
######   ##### ##### #####  ####################################################
%wd##
%fd##
-----------------------------------------------------------------------------*/

void
draw_wd_literals(win, y, x)
WINDOW *win;
int y, x;

{
    static char *header =
    "Device  Jobs/s r+w/s Blks/s -- (% Utilization) ";
    int x2, ctlr, drive;

    wmove(win, y, x);
    use_cp(win, cpBANNER);
    waddstr(win, header);
    getyx(win, y, x2);
    while (x2++ <= COLS)
	waddch(win, (chtype) '-');
    use_cp(win, cpLIT);
    clear_area(win, getmaxy(win) - 1, 0, getmaxx(win));
    for (ctlr = 0; ctlr < NCTLR; ctlr++)
    {
	for (drive = 0; drive < NWD; drive++)
	{
	    mvwprintw(win, (y + (ctlr * NWD) + drive + 1), x, "wd%1d%1d", ctlr, drive);
	}
    }
    getyx(win, y, x2);
    for (drive = 0; drive < NDRIVES; drive++)
	mvwprintw(win, (y + drive + 1), x, "fd%1d", drive);
}

/*+----------------------------------------------------------------------------
	update_wd(win, y, x, width);

00000000001111111111222222222233333333334444444444555555555566666666667777777777
01234567890123456789012345678901234567890123456789012345678901234567890123456789
obs/s r+w/s Blks/s -- (% Utilization) ---------------------------------
##### ##### #####  ####################################################
-----------------------------------------------------------------------------*/
#define _WD_AQ	0
#define _WD_RW	6
#define _WD_BK	12
#define _WD_UT	19

void
update_wd(win, y, x, width)
WINDOW *win;
int y, x;
int width;

{
    register i;
    time_t elapsed_secs;
    int itmp;
    int percent_used;

/*
	int max;
*/

    (void)time(&now);
    elapsed_secs = now - timelast;
    kread((caddr_t) ia, wd1010statsaddr, sizeof(struct iotime) * NCTLR * NWD);

    kread((caddr_t) & ia[NWDTOTAL], flstatsaddr,
	  sizeof(struct iotime) * NDRIVES);

    for (i = 0; i < (NWDTOTAL + NDRIVES); i++)
    {
	use_cp(win, cpINFO);
	mvwprintw(win, y + i, _WD_AQ + x, "%5i",
		  (ia[i].io_qc - ialast[i].io_qc));
	mvwprintw(win, y + i, _WD_RW + x, "%5i",
		  (ia[i].io_cnt - ialast[i].io_cnt));
	itmp = ia[i].io_bcnt - ialast[i].io_bcnt;
	mvwprintw(win, y + i, _WD_BK + x, "%5i",
		  (itmp > 99999) ? 99999 : itmp);	/* protect transient
							 * overflow */
	wmove(win, y + i, _WD_UT + x);
	percent_used = (ia[i].io_act - ialast[i].io_act);
	wprintw(win, "%5i", (percent_used > 99999) ? 99999 : percent_used);
/*
		max_util[i]=max_util[i]<percent_used ? percent_used : max_util[i];
		if(percent_used < 70)
			use_cp(win, cpLOW);
		else if (percent_used < 90)
			use_cp(win, cpMED);
		else
			use_cp(win, cpHIGH);

		itmp = (width * percent_used) / 100;
		max = (width * max_util[i]) / 100;
		while(-1) {
			itmp--;max--;
			if (max<1) {
				use_cp(win, cpHIGH);
				waddch(win, (chtype)'M');
				break;
			}
			if (itmp>-1)
				waddch(win, (chtype)'-');
			else
				waddch(win, (chtype)' ');
		}

		wclrtoeol(win);
*/
    }
    memcpy((char *)ialast, (char *)ia,
	   sizeof(struct iotime) * (NWDTOTAL + NDRIVES));

    timelast = now;
}

#endif /* DPT_WD */
