
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: dialog.C,v 2.8 1995/06/24 14:33:09 davidsi Exp $
//
// $Log: dialog.C,v $
// Revision 2.8  1995/06/24  14:33:09  davidsi
// - BUG FIX. When retrieving information from the database, it may not
//   be found and a NULL pntr may be returned.  This must be checked for,
//   and sensible defaults chosen if appropriate.
//
// Revision 2.7  1995/06/17  20:31:55  davidsi
// - Added support for different measurement units in the preferences dialog.
//
// Revision 2.6  1995/06/07  20:56:31  davidsi
// - Removed DialogshellUnmanage() function member from the dialog
//   class - the underlying graphics support now automatically does the
//   "right" thing.
// - Added preferences entry for the path specifying where the price
//   files exist.
//
// Revision 2.5  1995/06/02  20:00:04  davidsi
// - BUG FIX. DialogPrintf() was not overwriting the old strings when the
//   line pointer wrapped.
//
// Revision 2.4  1995/05/19  18:51:22  davidsi
// - The DialogYesno() can now cope with multi line strings.  It also sets modal
//   behaviour.
// - Extended the Dialog class with DialogSetmodal() and DialogClearmodal()
//   function members so that all dialogs can set whether they require
//   modal behaviour.
// - DialogentryCreate() now requires an argument specifying total number
//   of visible columns.
// - Added a new preferences dialog.  Fonts can now also be set on the fly
//   using the new font_specifier structure and dialog_setfont().
// - Fixes to compile under SCO OpenServer release 5.0.
//
// Revision 2.3  1995/04/25  21:13:54  davidsi
// - Added support for multiple cursors.  The dcursortype_e describes the
//   available cursors and DialogCursor() sets the active cursor for all
//   dialogs.
// - Added support for preventing output being flushed on every call to
//   DialogPrintf() as this can result in a lot of flicker if much is being
//   output in different calls.  Also saves loads of CPU cycles.
// - Added DialogentrySet(int,char) function member that allows the input
//   value for a given field in a dialog entry to be modified once set.
//
// Revision 2.2  1995/04/16  20:35:07  davidsi
// - Added DialogPrintf() which will allow xsm to printf to the main
//   window.
//
// Revision 2.1  1995/04/03  20:40:48  davidsi
// - Added new Dialoglist class.
//
// Revision 2.0  1995/03/02  22:02:43  davidsi
// - Initial version.
//
//
//
//




//
// Include Files
//
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <dial.h>
#include <math.h>
#include <sys/types.h>
#include <sys/times.h>


#define FULL_INCLUDE

#include "localmacros.h"
#include "dialog.h"
#include "button.h"
#include "optionl.h"
#include "prefs.h"
#include "support.h"


#define MAX_TEXTS 	32		// max number of texts for input dialogs
#define FONT_NUM        6		// arbitary limit for font mappings

//
// Defines the maximum width (in characters) a DialogOptionlist will have
//
#define MAX_OPTIONLIST_WIDTH 	40

//
// Physical font names
//
static char *pfontnames[]= {
#ifdef DEF_FONT
        "-adobe-courier-bold-r-normal--14-140-75-75-m-90-iso8859-1",
        "-adobe-helvetica-bold-r-normal--14-100-100-100-p-82-iso8859-1",
        "-adobe-helvetica-bold-r-normal--11-80-100-100-p-60-iso8859-1",
        "-adobe-helvetica-bold-r-normal--17-120-100-100-p-92-iso8859-1",
#else
        "6x10",
        "6x10",
        "6x10",
        "6x10",
#endif
        "6x10",
        "6x10",
    };

static char *fontnames[]= {
        "prop1",
        "prop2",
        "prop3",
        "prop4",
        "fixed1",
        "fixed2",
    };

//
//
//
dialog d_root;			// main root dialog


static dialog d_text1;		// upper output window
Dialogvertical *d_mainvert;	// lower output window
//
// Note that lines are held in a FIFO so it will wrap around
//
static char **line_map;		// pointer to text for this line
static int *line_size;		// size of line (only grows)
static int max_lines;		// total number of lines in line_map/size arrys
static int line_index;		// line where text is to be inserted
static int textsize;		// total number of characters in output buffer
static char noflush;		// 0 if DialogPrintf() output not to be flushed
static int current_state;	// State of main dialog hierarchy.




//
///////////////////////////////////////////////////////////////////////////////
//
// The Dialog class
//
///////////////////////////////////////////////////////////////////////////////
//
void
Dialogshell::DialogshellCreate(char *title)
{
	d=dialog_shell(d_root,title,SHELL_POPUP);
	dialog_type=DIALOG_SHELL;
}


Dialogshell::Dialogshell()
{
}

Dialogshell::~Dialogshell()
{
}


//
///////////////////////////////////////////////////////////////////////////////
//
// The DialogverticalCreate class
//
///////////////////////////////////////////////////////////////////////////////
//
void
Dialogvertical::DialogverticalCreate(class Dialog &pd)
{
	if (pd.d==NULL)
		DialogPanic("Dialog::DialogVertical - no parent Dialog");
	d=dialog_containv(pd.d,NULL);
	dialog_type=DIALOG_VERT;
}

Dialogvertical::Dialogvertical()
{
}

Dialogvertical::~Dialogvertical()
{
}

//
///////////////////////////////////////////////////////////////////////////////
//
// The DialoghorizontalCreate class
//
///////////////////////////////////////////////////////////////////////////////
void
Dialoghorizontal::DialoghorizontalCreate(class Dialog &pd)
{
	if (pd.d==NULL)
		DialogPanic("Dialog::DialogHorizontal - no parent Dialog");
	d=dialog_containh(pd.d,NULL);
	dialog_type=DIALOG_HORIZ;
}

Dialoghorizontal::Dialoghorizontal()
{
}

Dialoghorizontal::~Dialoghorizontal()
{
}


//
///////////////////////////////////////////////////////////////////////////////
//
// The Dialogentry class
//
///////////////////////////////////////////////////////////////////////////////
//
void
Dialogentry::DialogentryCreate(class Dialog &pd,int cols)
{
	new_d_entrylist nde;

	if (pd.d==NULL)
		DialogPanic("Dialog::Dialogentry - no parent Dialog");
	nde.cols=cols;
	d=dialog_entrylist(pd.d,&nde,NULL);
	dialog_type=DIALOG_ENTRY;
}

void
Dialogentry::DialogentrySet(char *name,char *data,int id)
{
	if (d==NULL)
		DialogPanic("Dialog::DialogentrySet - no parent Dialog");
	dialog_labelentry(d,name,data,id);
}

void
Dialogentry::DialogentrySet(char *name,int data,int id)
{
	char buffer[256];

	sprintf(buffer,"%d",data);
	DialogentrySet(name,buffer,id);
}

void
Dialogentry::DialogentrySet(char *name,double data,int id)
{
	char buffer[256];

	sprintf(buffer,"%6.2f",data);
	DialogentrySet(name,buffer,id);
}

char *
Dialogentry::DialogentryGet(int id)
{
	if (d==NULL)
		DialogPanic("Dialog::DialogentryGet - no dialog");
	return(dialog_entrylist_value(d,id));
}

void
Dialogentry::DialogentrySet(int id,char *val)
{
	if (d==NULL)
		DialogPanic("Dialog::DialogentryGet - no dialog");
	return(dialog_entrylist_setvalue(d,id,val));
}

Dialogentry::Dialogentry()
{
}

Dialogentry::~Dialogentry()
{
}


//
///////////////////////////////////////////////////////////////////////////////
//
// The Dialogframe class
//
///////////////////////////////////////////////////////////////////////////////
//
void
Dialogframe::DialogframeCreate(class Dialog &pd)
{
	static new_d_frame nf = {
		0,                              // no flags
	};

	if (pd.d==NULL)
		DialogPanic("Dialog::DialogFrame- no parent Dialog");
	d=dialog_frame(pd.d,&nf,NULL);
	dialog_type=DIALOG_FRAME;
}

Dialogframe::Dialogframe()
{
}

Dialogframe::~Dialogframe()
{
}



//
///////////////////////////////////////////////////////////////////////////////
//
// The Dialogform class
//
///////////////////////////////////////////////////////////////////////////////
//
void
Dialogform::DialogformCreate(class Dialog &pd)
{
	if (pd.d==NULL)
		DialogPanic("Dialog::DialogForm - no parent Dialog");
	d=dialog_form(pd.d);
	dialog_type=DIALOG_FORM;
}
Dialogform::Dialogform()
{
}

Dialogform::~Dialogform()
{
}


//
///////////////////////////////////////////////////////////////////////////////
//
// The Dialogverticalform class
//
///////////////////////////////////////////////////////////////////////////////
//
void
Dialogverticalform::DialogverticalformCreate(class Dialogform &pd,
		int left,
		int right,
		int top,
		int bottom
		)
{
	position p;

	p.left=left;
	p.right=right;
	p.top=top;
	p.bottom=bottom;
	p.flags=POSITION_PERC;

	if (pd.d==NULL)
		DialogPanic("Dialog::DialogVerticalform - no parent Dialog");
	d=dialog_containv(pd.d,&p);
	dialog_type=DIALOG_VFORM;
}

Dialogverticalform::Dialogverticalform()
{
}

Dialogverticalform::~Dialogverticalform()
{
}


//
///////////////////////////////////////////////////////////////////////////////
//
// The Dialoghorizontalform class
//
///////////////////////////////////////////////////////////////////////////////
//
void
Dialoghorizontalform::DialoghorizontalformCreate(class Dialogform &pd,
		int left,
		int right,
		int top,
		int bottom
		)
{
	position p;

	p.left=left;
	p.right=right;
	p.top=top;
	p.bottom=bottom;
	p.flags=POSITION_PERC;

	if (pd.d==NULL)
		DialogPanic("Dialog::DialogHorizform - no parent Dialog");
	d=dialog_containh(pd.d,&p);
	dialog_type=DIALOG_HFORM;
}

Dialoghorizontalform::Dialoghorizontalform()
{
}

Dialoghorizontalform::~Dialoghorizontalform()
{
}


//
///////////////////////////////////////////////////////////////////////////////
//
// The Dialoglabel class
//
///////////////////////////////////////////////////////////////////////////////
//
void
Dialoglabel::DialoglabelCreate(class Dialog &pd,char *text)
{
	static new_label nl1 = {
		"Name Goes Here",
		NULL,
	};

	if (pd.d==NULL)
		DialogPanic("Dialog::DialogLabel - no parent Dialog");

	nl1.name=text;
	d=dialog_label(pd.d,&nl1,NULL);
	dialog_type=DIALOG_LABEL;
}

void
Dialoglabel::DialoglabelSet(char *text)
{
	if (d==NULL)
		DialogPanic("Dialog::DialogLabel - no parent Dialog");
	dialog_label_change(d,text);
}


Dialoglabel::Dialoglabel()
{
}

Dialoglabel::~Dialoglabel()
{
}


//
///////////////////////////////////////////////////////////////////////////////
//
// The Dialoginput class
//
///////////////////////////////////////////////////////////////////////////////
//
void
Dialoginput::DialoginputCreate(class Dialog &pd,int wi,char *text)
{
	static new_d_entry ne;

	if (pd.d==NULL)
		DialogPanic("Dialog::DialogInput - no parent Dialog");

	ne.cols=wi;
	ne.ivalue=text;
	d=dialog_entry(pd.d,&ne,NULL);
	dialog_type=DIALOG_INPUT;
}

char *
Dialoginput::DialoginputGet()
{
	if (dialog_type!=DIALOG_INPUT)
		DialogPanic("Dialog::DialogInputGet - Not a DialogLabel");
	if (d==NULL)
		DialogPanic("Dialog::DialogInputGet - no parent Dialog");
	return(dialog_entry_value(d));
}

Dialoginput::Dialoginput()
{
}

Dialoginput::~Dialoginput()
{
}


//
///////////////////////////////////////////////////////////////////////////////
//
// The Dialoglist class
//
///////////////////////////////////////////////////////////////////////////////
//
void
Dialoglist::DialoglistCreate(class Dialog &pd,Newdialoglist &nd)
{
	static new_list ne;

	if (pd.d==NULL)
		DialogPanic("Dialog::Dialoglist - no parent Dialog");

	ne.width=nd.width;
	ne.height=nd.height;
	ne.list_item=nd.listitem;
	ne.listname=nd.listname;
	ne.fname=nd.font;
	d=dialog_list(pd.d,&ne);
}

char **
Dialoglist::DialoglistGet(int &length)
{
	if (d==NULL)
		DialogPanic("Dialog::DialoglistGet - no parent Dialog");
	return(dialog_list_items(d,&length));
}

Dialoglist::Dialoglist()
{
}

Dialoglist::~Dialoglist()
{
}



//
///////////////////////////////////////////////////////////////////////////////
//
// Extra functions available for Dialog derivedclasses
//
///////////////////////////////////////////////////////////////////////////////
//
void
Dialog::DialogRealize()
{
	if (d)
		dialog_realise(d);
}

void
Dialog::DialogDestroy()
{
	if (d && (dialog_type==DIALOG_SHELL)) {
		dialog_destroy(d);
		d=NULL;
	}
}

void
Dialog::DialogSetmodal()
{
	if (d)
		dialog_setmodal(d,TRUE);
}

void
Dialog::DialogClearmodal()
{
	if (d)
		dialog_setmodal(d,FALSE);
}


Dialog::Dialog()
{
	d=NULL;
}

Dialog::~Dialog()
{
	if (d && (dialog_type==DIALOG_SHELL)) {
		dialog_destroy(d);
		d=NULL;
	}
}

//
///////////////////////////////////////////////////////////////////////////////
//
// Publicly available functions
//
///////////////////////////////////////////////////////////////////////////////
//

enum buttone {
	BUTTON_NONE,
	BUTTON_OK,
	BUTTON_CANCEL,
	};
	

//
// The DialogYesno class provides a message and ok/cancel style buttons.
// Generally this class need not be directly accessed - use DialogYesno()
//
class DialogYesno : private Button {

	Dialogshell d_shell;			// top most dialog
	Dialogvertical  d_mainvert;
	Dialoglabel  d_label[MAX_TEXTS];
	Dialoghorizontal d_buttonhoriz;		// button horizontal form
	Buttoninstance *d_ok; 
	Buttoninstance *d_cancel; 

	int buttonid;
	
private:
	//
	// This overrides the pure virtual button handler in the Button class
	//
	virtual void ButtonCallback(unsigned long id,buttonstate_e flag);

public:
	int DialogYesnoGet();
	DialogYesno(char *wstr,char *str,char *truemsg,char *falsemsg);
	~DialogYesno();
};

void
DialogYesno::ButtonCallback(unsigned long id,buttonstate_e flag)
{
	switch(id) {
	case BUTTON_OK:
		buttonid=BUTTON_OK;
		break;
	case BUTTON_CANCEL: 
		buttonid=BUTTON_CANCEL;
		break;
	}
}

int
DialogYesno::DialogYesnoGet()
{
	while (buttonid==BUTTON_NONE)
		DialogEvent();

	if (buttonid!=BUTTON_OK) {
		return(0);
	}
	return (1);
}

DialogYesno::DialogYesno(char *wstr,char *msg, char *tmsg,char *fmsg)
{
	int i;
	char *buf;
	char *ebuf,*sbuf;
	char partstring;

	buttonid=BUTTON_NONE;

        d_shell.DialogshellCreate(wstr);

        d_mainvert.DialogverticalCreate(d_shell);

	buf=(char *)malloc(strlen(msg)+1);
	strcpy(buf,msg);
	sbuf=buf;
	for(i=0;i<MAX_TEXTS;i++) {
		ebuf=sbuf;
		while (*ebuf!='\0' && *ebuf!='\n')
			ebuf++;
		partstring=(*ebuf=='\n');
		*ebuf='\0';
	       	d_label[i].DialoglabelCreate(d_mainvert,sbuf);
		if (!partstring)
			break;
		sbuf=ebuf+1;
	}
	free(buf);

        d_buttonhoriz.DialoghorizontalCreate(d_mainvert);

	if (tmsg)
	        (void)ButtonPush(d_buttonhoriz,tmsg,BUTTON_OK);
	else
	        (void)ButtonPush(d_buttonhoriz,"OK",BUTTON_OK);

	if (fmsg)
	        (void)ButtonPush(d_buttonhoriz,fmsg,BUTTON_CANCEL);
	d_shell.DialogSetmodal();
	d_shell.DialogRealize();
}

DialogYesno::~DialogYesno()
{
}

//
// DialogYesno message true false
//
// Displays a message and allows the user to select via buttons
// one of two options.
//
// INPUTS :	message		The message to be displayed
//		true		The TRUE button message
//		false		the FALSE button message
// RETURNS:	1 if OK button pressed, else 0.
//
int
DialogYesno(char *str,char *truestr,char *falsestr)
{
	int r;
	class DialogYesno *yesno;

	yesno=new class DialogYesno("Message",str,truestr,falsestr);
	r=yesno->DialogYesnoGet();
	delete yesno;

	DialogUpdatedisplay();

	return(r);
}




//
// DialogMessage message
//
// Displays a message and expects the user to press a button.
//
// INPUTS :	message		The message to be displayed
// RETURNS:	None
//
void
DialogMessage(char *errstr)
{
	class DialogYesno *yesno;

	yesno=new class DialogYesno("Message",errstr,NULL,NULL);

	(void)(yesno->DialogYesnoGet());
	delete yesno;

	DialogUpdatedisplay();
}

//
// DialogPanic message
//
// Displays a message and expects the user to press a button. Once
// pressed the program quits.
//
// INPUTS :	message		The message to be displayed
// RETURNS :	NEVER DOES!
//
void
DialogPanic(char *errstr)
{
	class DialogYesno yesno("Major system failure",errstr,"EXIT",NULL);

	(void)yesno.DialogYesnoGet();

	exit(1);
}


//
// The DialogString class displays multiple strings and allows for
// user entry of another string.
//
// Generally this class need not be directly accessed - use DialogString()
//
class DialogString : private Button {

	Dialogshell d_shell;			// top most dialog
	Dialogvertical  d_mainvert;
	Dialoglabel  d_label[MAX_TEXTS];
	Dialoghorizontal d_inputhoriz;		// input horizontal form
	Dialoglabel  d_namelabel;		// name label
	Dialoginput  d_nameinput;		// name label
	Dialoghorizontal d_buttonhoriz;		// button horizontal form
	Buttoninstance *d_ok; 
	Buttoninstance *d_cancel; 

	int buttonid;
	
private:
	//
	// This overrides the pure virtual button handler in the Button class
	//
	virtual void ButtonCallback(unsigned long id,buttonstate_e flag);

public:
	char *DialogStringGet(int &);
	DialogString(char *wt,char *t, char *istring, int length, char **);
	~DialogString();
};

void
DialogString::ButtonCallback(unsigned long id,buttonstate_e flag)
{
	switch(id) {
	case BUTTON_OK:
		buttonid=BUTTON_OK;
		break;
	case BUTTON_CANCEL: 
		buttonid=BUTTON_CANCEL;
		break;
	}
}

char *
DialogString::DialogStringGet(int &flag)
{
	while (buttonid==BUTTON_NONE)
		DialogEvent();
	if (buttonid!=BUTTON_OK) {
		flag= -1;
		return(NULL);
	}
	flag= 0;
	return( d_nameinput.DialoginputGet() );
}

DialogString::DialogString(char *wt,
		char *t, 
		char *istring,
		int length, 
		char **texts)
{
	int i;

	buttonid=BUTTON_NONE;

        d_shell.DialogshellCreate(wt);

        d_mainvert.DialogverticalCreate(d_shell);

	i=0;
	while (*texts) {
		if (i >= MAX_TEXTS)
			break;
        	d_label[i].DialoglabelCreate(d_mainvert,*texts);
		i++;
		texts++;
	}

        d_inputhoriz.DialoghorizontalCreate(d_mainvert);
        d_namelabel.DialoglabelCreate(d_inputhoriz,t);
        d_nameinput.DialoginputCreate(d_inputhoriz,length,istring);

        d_buttonhoriz.DialoghorizontalCreate(d_mainvert);
        (void)ButtonPush(d_buttonhoriz,"OK",BUTTON_OK);
        (void)ButtonPush(d_buttonhoriz,"CANCEL",BUTTON_CANCEL);
	d_shell.DialogSetmodal();
	d_shell.DialogRealize();
}

DialogString::~DialogString()
{
}

//
// DialogString windowtitle title string length more_strings
//
// INPUTS :	windowtitle	The title for the window
//		title		The main title within the window
//		string		The initial string to be edited by
//				the user.
//		length		The max length of the user entered string.
//		more_strings	List of NULL terminated string ptrs.
// RETURNS:	0 if string entered, else -1
//
int
DialogString(char *wtitle,char *title,char *string,int length, ...)
{
	char *str;
	va_list ap;
	char *texts[MAX_TEXTS];
	int i;
	class DialogString *dsptr;
	int flag;

        // local_va_start(ap);
        local_va_start2(ap,length);

	(void)va_arg(ap, char *);
	for (i=0;i<MAX_TEXTS-1;i++) {
		texts[i]=va_arg(ap, char *);
		if (texts[i]==NULL)
			break;
	}
	texts[i]=NULL;
	va_end(ap);
	dsptr=new class DialogString(wtitle,title,string,length,texts);
	str=dsptr->DialogStringGet(flag);
	if (flag==-1) {
		delete dsptr;
		return( -1 );
	}
	strncpy(string,str,length);
	delete dsptr;
	return(0);
}


//
// The DialogOptionlist class displays optionlists and buttons for
// user entry.
//
// Generally this class need not be directly accessed - use DialogOptionlist()
//
class DialogOptionlist : private Button, private Optionlist {

	Dialogshell d_shell;			// top most dialog
	Dialogvertical  d_mainvert;
	Optionlistinstance *d_optionlist;
	Dialoghorizontal d_buttonhoriz;		// button horizontal form
	Buttoninstance *d_ok; 
	Buttoninstance *d_cancel; 

	int buttonid;
	
private:
	//
	// This overrides the pure virtual button handler in the Button class
	//
	virtual void ButtonCallback(unsigned long id,buttonstate_e flag);
	virtual void OptionlistCallback(unsigned long id);

public:
	char *DialogOptionlistGet();
	DialogOptionlist(char *wtitle,char *ltitle,char **items);
	~DialogOptionlist();
};

void
DialogOptionlist::OptionlistCallback(unsigned long id)
{
	buttonid=BUTTON_OK;
}

void
DialogOptionlist::ButtonCallback(unsigned long id,buttonstate_e flag)
{
	switch(id) {
	case BUTTON_CANCEL: 
		buttonid=BUTTON_CANCEL;
		break;
	}
}

char *
DialogOptionlist::DialogOptionlistGet()
{
	while (buttonid==BUTTON_NONE)
		DialogEvent();
	d_mainvert.DialogClearmodal();

	if (buttonid!=BUTTON_OK) 
		return(NULL);

	return ( d_optionlist->OptionlistinstanceGet() );
}

DialogOptionlist::DialogOptionlist(char *wtitle,char *ltitle,char **items)
{
        static Newoptionlist nolist = {
                5,			// width
                10,			// height
                NULL,			// list of items
                "",			// name of list
                "",			// name of item
                OLIST_FLAGS_SEARCH,	// flags
		NULL,			// default font
        };
	int width=10;
	char **scan;
	cursortype_e cursor;

	cursor=DialogCursor( DCURSOR_WAIT );

	//
	// Determine suitable width for the dialog
	//
	scan=items;
	while (*scan) {
		int cols=strlen(*scan);

		if (cols>width) {
			width=cols; 
			if (width >= MAX_OPTIONLIST_WIDTH)
				break;
		}
		scan++;
	}

	//
	// Add an extra characters to provide a more pleasing spacing
	//
	width += 1;
	if (width>MAX_OPTIONLIST_WIDTH)
		width=MAX_OPTIONLIST_WIDTH;

	buttonid=BUTTON_NONE;

        d_shell.DialogshellCreate(wtitle);

        d_mainvert.DialogverticalCreate(d_shell);
	
	nolist.listitem=items;
	nolist.listname=ltitle;
	nolist.width=width;
	d_optionlist=OptionlistCreate(d_mainvert,nolist,0);

        d_buttonhoriz.DialoghorizontalCreate(d_mainvert);
        (void)ButtonPush(d_buttonhoriz," CANCEL ",BUTTON_CANCEL);

	d_mainvert.DialogSetmodal();
	d_shell.DialogRealize();
	DialogCursor( cursor );
}

DialogOptionlist::~DialogOptionlist()
{
}


//
// DialogOptionlist windowtitle title items
//
// Provides a list of items that can be selected by the user
//
// INPUTS :	windowtittle	The title of the window
// 		title		The title within the window
//		items		A list of items (ascii strings) to chose
//				from
// RETURNS:	The string that was selected or NULL if none chosen. 
//
char *
DialogOptionlist(char *wtitle,char *title,char **items)
{
        class DialogOptionlist dol(wtitle,title,items);
        return( dol.DialogOptionlistGet() );
}



//
// DialogUpdatedisplay
//
// Rethinks the display and refreshes as necessary - mileage varies!
//
// INPUTS :	None
// RETURNS:	None
//
void
DialogUpdatedisplay()
{
	dialog_updatedisplay(d_root);
}


//
// DialogEvent
//
// Processes the next event in the input queue.
//
// INPUTS :	None
// RETURNS:	None
//
void
DialogEvent()
{
	handle_event();
}



//
// DialogRootmenu menubar
//
// Sets or resets the menu on the main window.
//
// INPUTS :	menubar		The menu bar to be displayed
// RETURNS	None
//
void
DialogRootmenu(menu_bar &mb)
{
	static int alreadyset=0;

	if (alreadyset)
		mb.redisplay(d_root);
	else {
		mb.display(d_root);
		alreadyset=1;
	}
}


class Dialogverticalhack : public Dialogvertical {
public:
	DialogverticalCreate(dialog nd,position &p) {
		d=dialog_containv(nd,&p);
		dialog_type=DIALOG_VERT;
	}
};


static void
DialogSetfonts()
{
	char *str;
	int iter;
	static font_specifier nf;

	for( iter=0 ; iter<FONT_NUM ; iter++ ) {
		nf.fname[iter]=fontnames[iter];
		nf.pfont[iter]=pfontnames[iter];
	}
	nf.npfont=FONT_NUM;

	str=db_prefs->DatabaseRetrieve(FONT_DB_CLASS,FONT_PROP1_CLASS);
	if (str)
		nf.pfont[0]=str;

	str=db_prefs->DatabaseRetrieve(FONT_DB_CLASS,FONT_PROP2_CLASS);
	if (str)
		nf.pfont[1]=str;

	str=db_prefs->DatabaseRetrieve(FONT_DB_CLASS,FONT_FIXED_CLASS);
	if (str)
		nf.pfont[4]=str;

	dialog_setfonts(&nf);
}

//
// DialogInitialize argc argv menubar
//
// Main dialog setup - not as generic as it should be!
//
// INPUTS :	argc,argv	The usual
// 		menubar		The menubar to be displayed
// RETURNS:	None
//
void
DialogInitialize(unsigned int argc,
	char **argv,
	menu_bar &mb
	)
{
	dialog d_form;
	position p;
	int i;
	static new_prefs np={
		"xsm",
		0,
		0,
		0
	};

	static new_d_edit new_high = {
		40,				// columns
		10,				// rows
		0,				// flags
		NULL,				// initial value 
		NULL,				// no call back
	};


	class Dialogverticalhack *d_mainvert_tmp=new class Dialogverticalhack;

	np.argc=argc;
	np.argv=argv;

	d_root=dialog_root(&np);

	DialogSetfonts();



        //
        // Set up the menu bar
        //
        DialogRootmenu(mb);

	d_form=dialog_form(d_root);

	p.left=0;
	p.right=100;
	p.top=0;
	p.bottom=30;
	p.flags=POSITION_PERC;

	d_mainvert_tmp->DialogverticalCreate(d_form,p);
	d_mainvert=d_mainvert_tmp;
	p.top=31;
	p.bottom=100;
	d_text1=dialog_edit(d_form,&new_high,&p);

	//
	//
	//
	max_lines=200;
	line_map=(char **)malloc(sizeof(char *)*max_lines);
	line_size=(int *)malloc(sizeof(int)*max_lines);
	line_index=0;
	textsize = 0;
	for(i=0;i<max_lines;i++ ) {
		line_map[i]=NULL;
		line_size[i]=0;
	}

	dialog_realise(d_root);
}



//
// DialogPrintf fonrmat string
//	
// Writes a string to the main output window
//
// INPUTS :	format		The format specifier
//		string		The string(s) to write
// RETURNS:	None
//
int
DialogPrintf(char *fmt ...)
{
	va_list ap;
	char string[1024];
	int len;

	va_start(ap,fmt);
	vsprintf(string,fmt,ap);

	len=strlen(string)+1;

	//
	// If new line won't fit in the space that is already allocated
	// or there is no space allocated then allocate now.  Maintain
	// textsize which is total (possible) space that might be needed
	// to combine lines. Possibly less space will be needed because
	// some lines may now have less data than when they were first
	// allocated (memory used for line storage only increases).
	//
	if (line_map[line_index]==NULL || len>line_size[line_index] ) {
		if (line_map[line_index]) {
			free(line_map[line_index]);
			textsize -= line_size[line_index];
		}
		line_map[line_index]=(char *)malloc(len);
		strcpy(line_map[line_index],string);
		line_size[line_index]=len;
		textsize += line_size[line_index];
	} else
		strcpy(line_map[line_index],string);

	//
	// Maintain FIFO
	//
	line_index++;

	if (line_index>=max_lines) {
		line_index=0;
	}
	DialogFlushoutput();
}


// 
// DialogNoflush
//
// Stops DialogPrintf() from flushing the internal buffer to the display.
// This can save flicker if multiple DialogPrintf() statements are being
// executed - not to say CPU cycles.
// 
// INPUTS :	None
// RETURNS:	None - but sets a global variable that can only be reset
//		by calling DialogFlush(). If DialogFlush() is never called
//		again no output will be displayed!
//
void
DialogNoflush()
{
	noflush=1;
}




// 
// DialogFlush
//
// Allows DialogPrintf() to flush the internal buffer to the display and
// initiates a flush.
// 
// INPUTS :	None
// RETURNS:	None
//
void
DialogFlush()
{
	noflush=0;
	DialogFlushoutput();
}




// 
// DialogFlushoutput
//
// Flushes the DialogPrintf() buffer to the display.
// 
// INPUTS :	None
// RETURNS:	None
//
void
DialogFlushoutput()
{
	char *build;
	int i;
	int pos;

	if (noflush)
		return;

	build=(char *)malloc(textsize);
	*build='\0';

	pos=line_index+1;

	for(i=0;i<max_lines;i++ ) {
		if (line_map[pos])
			strcat(build,line_map[pos]);
		pos = (pos+1) % max_lines;
	}
	dialog_edit_set(d_text1,build);
	free(build);
}

//
// DialogMainloop
//
// Grabs and processes events as they occur.
//
// INPUTS :	None
// RETURNS:	None
//
void
DialogMainloop()
{
	dialog_mainloop(d_root);
	//
	// NOT REACHED
	//
}


//
// DialogCursor cursortype
//
// Set the cursor for ALL known windows
//
// INPUTS :	cursortype		The type of cursor to be displayed
// RETURNS:	None
//
dcursortype_e
DialogCursor(dcursortype_e cursortype)
{
	cursortype_e ctype;
	static int busycursorcount=0;		// counter for busy cursor
	static int busycursordelay=0;		// current length busy counter
	static unsigned long lastticks=0;	// timer for timing between cursor updates
	static dcursortype_e current_cursor=DCURSOR_NORMAL;
	dcursortype_e retval;
	struct tms foo; 
	unsigned long newticks;

	switch(cursortype) {
        case DCURSOR_BUSY:
		if (busycursorcount++<busycursordelay) 
			return(current_cursor);
		busycursorcount=0; 
		newticks=times(&foo);
		if ((newticks-lastticks)<20)
			busycursordelay *= 2;
		else if ( ((newticks-lastticks)>30) &&
		          busycursordelay>2)
			busycursordelay /= 2;
		lastticks=newticks;
		ctype=CURSOR_BUSY;
		break;
        case DCURSOR_WAIT:
		busycursordelay=1;
		busycursorcount=busycursordelay-1;
		ctype=CURSOR_WAIT;
		break;
        case DCURSOR_NORMAL:
		busycursordelay=1;
		busycursorcount=busycursordelay-1;
		ctype=CURSOR_NORMAL;
		break;
	}
	dialog_setcursor(d_root,ctype);
	retval=current_cursor;
	current_cursor=cursortype;
	return(retval);
}



//
// The Preferences class
//
class DialogPreferences : private Button, private Optionlist {
	//
	// Dialogs (etc) used when setting the preferences
	//
	Dialogshell d_shell;		// top most dialog
	Dialogvertical d_vert;
        Dialogframe     d_typeframe1;
	Dialoghorizontal d_horiz;
	Dialogentry d_entrylist;
	Dialoghorizontal d_buttonh;
	//
	// Type of price input
	// 
        Dialogframe     d_typeframe2;
	Dialogvertical  d_typevert;
	Dialoghorizontal  d_typehoriz;
	Dialoglabel     d_typelabel;
	Buttoninstance *but_types[3];
	//
	// Price input  
	//
        Dialogframe     d_typeframe3;
	Dialoghorizontal d_horiz2;
	Dialogentry d_entrylist2;
	//
	// Printer measurement units
	//
        Dialogframe     d_typeframe4;
	Dialogvertical d_vert4;
	Dialoglabel     d_typelabel4;
	Dialoghorizontal d_horiz4;
	Buttoninstance *but_types4[3];

private:
	//
	// Overrides the function in the Button class
	//
	virtual void ButtonCallback(unsigned long id,buttonstate_e flag);

public:
	DialogPreferences();
	virtual ~DialogPreferences();
};

enum dpref_e {
	ID_FONTMAIN,
	ID_FONTTEXT,
	ID_FONTFIXD,
	BUTTON_ASCII,
	BUTTON_TTEXT,
	MENU_PREFERENCES,		// Not truly DialogPreferences related
	ID_PRICES,
	};

void
DialogPreferences::ButtonCallback(unsigned long id,buttonstate_e flag)
{
	char *str;
	char *str1,*str2,*str3,*str4;

	switch(id) {
	case BUTTON_OK:
		if (but_types[0]->ButtoninstanceGet() == BUTTON_ENABLED)
			str=FEED_ASCII;
		else if (but_types[1]->ButtoninstanceGet() == BUTTON_ENABLED)
			str=FEED_TELETEXT;
		db_prefs->DatabaseInsert(INPUT_DB_CLASS,INPUTFEED_DB_CLASS,str);

		//
		// Measurement units
		//
		if (but_types4[0]->ButtoninstanceGet() == BUTTON_ENABLED)
			str=MUNITS_INCHES;
		else if (but_types4[1]->ButtoninstanceGet() == BUTTON_ENABLED)
			str=MUNITS_CM;
		else if (but_types4[2]->ButtoninstanceGet() == BUTTON_ENABLED)
			str=MUNITS_POINTS;
		db_prefs->DatabaseInsert(UNITS_DB_CLASS,
						UNITSDEFINED_DB_CLASS,str);

		str1=d_entrylist.DialogentryGet(ID_FONTMAIN);
		str2=d_entrylist.DialogentryGet(ID_FONTTEXT);
		str3=d_entrylist.DialogentryGet(ID_FONTFIXD);
		str4=d_entrylist.DialogentryGet(ID_PRICES);
		if (!str1 || !str2 || !str3 )
			return;

		db_prefs->DatabaseDelete(FONT_DB_CLASS,FONT_PROP1_CLASS);
		db_prefs->DatabaseDelete(FONT_DB_CLASS,FONT_PROP2_CLASS);
		db_prefs->DatabaseDelete(FONT_DB_CLASS,FONT_FIXED_CLASS);

		db_prefs->DatabaseInsert(FONT_DB_CLASS,FONT_PROP1_CLASS,str1);
		db_prefs->DatabaseInsert(FONT_DB_CLASS,FONT_PROP2_CLASS,str2);
		db_prefs->DatabaseInsert(FONT_DB_CLASS,FONT_FIXED_CLASS,str3);

		//
		// Only if there is a string with characters in it, set
		// the price path
		//
		if (str4) {
			str1=remove_lt_spaces(str4);
			if (*str1!='\0')
				db_prefs->DatabaseInsert(PREFS_DB_ENTRY,
							VAR_PRICEPATH,str4);
		}
		delete this;
		//
		// Bring internal state up to date
		//
		DialogSetfonts();
		break;
	case BUTTON_CANCEL: 
		delete this;
		break;
	}
}

DialogPreferences::DialogPreferences()
{
	buttonstate_e bflags[3];
	char *str;

	d_shell.DialogshellCreate("Preferences");

	d_vert.DialogverticalCreate(d_shell);
	d_typeframe1.DialogframeCreate(d_vert);
	d_horiz.DialoghorizontalCreate(d_typeframe1);

	d_entrylist.DialogentryCreate(d_horiz,30);

	str=db_prefs->DatabaseRetrieve(FONT_DB_CLASS,FONT_PROP1_CLASS);
	if (!str)
		str=pfontnames[0];
	d_entrylist.DialogentrySet("Main font",str,ID_FONTMAIN);

	str=db_prefs->DatabaseRetrieve(FONT_DB_CLASS,FONT_PROP2_CLASS);
	if (!str)
		str=pfontnames[1];
	d_entrylist.DialogentrySet("Text font",str,ID_FONTTEXT);

	str=db_prefs->DatabaseRetrieve(FONT_DB_CLASS,FONT_FIXED_CLASS);
	if (!str)
		str=pfontnames[4];
	d_entrylist.DialogentrySet("Fixed font",str,ID_FONTFIXD);

	
	//
	// Type of price input
	//
	d_typeframe2.DialogframeCreate(d_vert);
	d_typevert.DialogverticalCreate(d_typeframe2);
	d_typelabel.DialoglabelCreate(d_typevert,"Update Method");


	bflags[0]=BUTTON_DISABLED;
	bflags[1]=BUTTON_DISABLED;

	str=db_prefs->DatabaseRetrieve(INPUT_DB_CLASS,INPUTFEED_DB_CLASS);
	if (str==NULL || !strcmp(str,FEED_ASCII))
		bflags[0]=BUTTON_ENABLED;
	else
		bflags[1]=BUTTON_ENABLED;

	d_typehoriz.DialoghorizontalCreate(d_typevert);
	but_types[0]=ButtonTogglemutex(d_typehoriz,"Ascii",
						BUTTON_ASCII,bflags[0]);
	but_types[1]=ButtonTogglemutex(d_typehoriz,"Teletext",
						BUTTON_TTEXT,bflags[1]);

	//
	// Price location
	//
	d_typeframe3.DialogframeCreate(d_vert);
	d_horiz2.DialoghorizontalCreate(d_typeframe3);
	d_entrylist2.DialogentryCreate(d_horiz2,30);

	str=db_prefs->DatabaseRetrieve(PREFS_DB_ENTRY,VAR_PRICEPATH);
	if (!str)
		str="";
	d_entrylist2.DialogentrySet("Path for prices",str,ID_PRICES);

	
	//
	// Printer Measurement units
	//
	d_typeframe4.DialogframeCreate(d_vert);
	d_vert4.DialogverticalCreate(d_typeframe4);
	d_typelabel4.DialoglabelCreate(d_vert4,"Measurement Units");
	d_horiz4.DialoghorizontalCreate(d_vert4);

	str=db_prefs->DatabaseRetrieve(UNITS_DB_CLASS,UNITSDEFINED_DB_CLASS);
	bflags[0]=bflags[1]=bflags[2]=BUTTON_DISABLED;
	if (str==NULL || !strcmp(str,MUNITS_INCHES))
		bflags[0]=BUTTON_ENABLED;
	else if (!strcmp(str,MUNITS_CM))
		bflags[1]=BUTTON_ENABLED;
	else if (!strcmp(str,MUNITS_POINTS))
		bflags[2]=BUTTON_ENABLED;
	else
		//
		// Default is inches
		//
		bflags[0]=BUTTON_ENABLED;

	but_types4[0]=ButtonTogglemutex(d_horiz4,"Inches",
						BUTTON_ASCII,bflags[0]);
	but_types4[1]=ButtonTogglemutex(d_horiz4,"cm",
						BUTTON_ASCII,bflags[1]);
	but_types4[2]=ButtonTogglemutex(d_horiz4,"Points",
						BUTTON_ASCII,bflags[2]);

	//
	// OK and CANCEL button
	//
	d_buttonh.DialoghorizontalCreate(d_vert);
	(void)ButtonPush(d_buttonh," OK ",BUTTON_OK);
	(void)ButtonPush(d_buttonh,"CANCEL",BUTTON_CANCEL);

	d_shell.DialogRealize();
}

DialogPreferences::~DialogPreferences()
{
}


static void
decode(local_va_dcl)
{
va_list ap;
dialog d;
unsigned long code;

	local_va_start(ap);
	d=va_arg(ap,dialog);
	code=va_arg(ap,unsigned long);

	switch(code) {
	case MENU_PREFERENCES:
		DialogPreferences *ds=new class DialogPreferences;
		break;
	}
}
//
// Initialization function for this module
//
dialog_init(menu_bar *mb)
{
	char *home;

	//
	// Create menu bar
	//
        MENU("Project","Preferences",decode, MENU_PREFERENCES);
}
