
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: dialog.h,v 2.6 1995/06/17 20:32:06 davidsi Exp $
//
// $Log: dialog.h,v $
// Revision 2.6  1995/06/17  20:32:06  davidsi
// - Added database identifiers for measurement units.
//
// Revision 2.5  1995/06/07  20:55:08  davidsi
// - Removed DialogshellUnmanage() function member from the dialog
//   class - the underlying graphics support now automatically does the
//   "right" thing.
//
// Revision 2.4  1995/05/18  21:04:56  davidsi
// - Added a dialog (d_mainvert) in the main window that can be added to by
//   other modules.
// - Extended the Dialog class with DialogSetmodal() and DialogClearmodal()
//   function members so that all dialogs can set whether they require
//   modal behaviour.
// - DialogentryCreate() now requires an argument specifying total number
//   of visible columns.
//
// Revision 2.3  1995/04/25  21:14:13  davidsi
// - Added support for multiple cursors.  The dcursortype_e describes the
//   available cursors and DialogCursor() sets the active cursor for all
//   dialogs.
// - Added support for preventing output being flushed on every call to
//   DialogPrintf() as this can result in a lot of flicker if much is being
//   output in different calls.  Also saves loads of CPU cycles.
// - Added DialogentrySet(int,char) function member that allows the input
//   value for a given field in a dialog entry to be modified once set.
//
// Revision 2.2  1995/04/16  20:34:07  davidsi
// - Added DialogPrintf() declaration.
//
// Revision 2.1  1995/04/03  20:41:54  davidsi
// - Added new Dialoglist class.
//
// Revision 2.0  1995/03/02  22:02:43  davidsi
// - Initial version.
//
//
#ifndef __DIALOG_H
#define __DIALOG_H

#include "localmacros.h"
#include "suplib.h"
#include "menu.h"

enum dialog_e {
	DIALOG_SHELL,
	DIALOG_VERT,
	DIALOG_HORIZ,
	DIALOG_ENTRY,
	DIALOG_FORM,
	DIALOG_VFORM,
	DIALOG_HFORM,
	DIALOG_FRAME,
	DIALOG_INPUT,
	DIALOG_LABEL,
	};
//
// The Dialog class
//
class Dialog 
{
protected:
	dialog d;
	dialog_e dialog_type;
	
public:
	void DialogRealize();
	void DialogDestroy();
	void DialogSetmodal();
	void DialogClearmodal();

	Dialog();
	virtual ~Dialog();

friend class Button;
friend class Optionlist;
};


//
// Dialogshell
//
class Dialogshell : public Dialog 
{
protected:
public:
	void DialogshellCreate(char *title);
	Dialogshell();
	virtual ~Dialogshell();
};




//
// Dialogvertical
//
class Dialogvertical : public Dialog {
protected:
public:
	void DialogverticalCreate(class Dialog &d);
	Dialogvertical();
	virtual ~Dialogvertical();
};




//
// Dialoghorizontal
//
class Dialoghorizontal : public Dialog {
protected:
public:
	void DialoghorizontalCreate(class Dialog &d);

	Dialoghorizontal();
	virtual ~Dialoghorizontal();
};




//
// Dialogentry
//
class Dialogentry : public Dialog {
protected:
public:
	void DialogentrySet(char *name,char *data,int id);
	void DialogentrySet(char *name,int data,int id);
	void DialogentrySet(char *name,double data,int id);
	void DialogentrySet(int id,char *val);
	char *DialogentryGet(int id);
	void DialogentryCreate(class Dialog &d,int cols);

	Dialogentry();
	virtual ~Dialogentry();
};




//
// Dialogframe
//
class Dialogframe : public Dialog {
protected:
public:
	void DialogframeCreate(class Dialog &d);

	Dialogframe();
	virtual ~Dialogframe();
};





//
// Dialogform
//
class Dialogform : public Dialog {
protected:
public:
	void DialogformCreate(class Dialog &d);

	Dialogform();
	virtual ~Dialogform();
};




//
// Dialogverticalform
//
class Dialogverticalform : public Dialogform {
protected:
public:
	void DialogverticalformCreate(class Dialogform &d,
				int l,int r,int top,int bottom);

	Dialogverticalform();
	virtual ~Dialogverticalform();
};




//
// Dialoghorizontalform
//
class Dialoghorizontalform : public Dialogform {
protected:
public:
	void DialoghorizontalformCreate(class Dialogform &d,
				int l,int r,int top,int bottom);

	Dialoghorizontalform();
	virtual ~Dialoghorizontalform();
};





//
// Dialoglabel
//
class Dialoglabel : public Dialog {
protected:
public:
	void DialoglabelSet(char *text);
	void DialoglabelCreate(class Dialog &,char *text);

	Dialoglabel();
	virtual ~Dialoglabel();
};




//
// Dialoginput
//
class Dialoginput : public Dialog {
protected:
public:
	char *DialoginputGet();
	void DialoginputCreate(class Dialog &,int width,char *initialtext); 

	Dialoginput();
	virtual ~Dialoginput();
};

//
// Struct to specify a new dialog list
//
typedef struct Newdialoglist {
	int width;                      // width in characters
	int height;                     // height in visible items
	char **listitem;                // items to display
	char *listname;                 // name of items in list
	char *font;                     // font (can be NULL)
	} Newdialoglist;

//
// Dialoglist
//
class Dialoglist : public Dialog {
	dialog d_list;
protected:
public:
	char **DialoglistGet(int &length);
	void DialoglistCreate(class Dialog &,Newdialoglist &);

	Dialoglist();
	virtual ~Dialoglist();
};


//
// Supported cursors
//
enum dcursortype_e {
	DCURSOR_BUSY,
	DCURSOR_PROP,
	DCURSOR_WAIT,
	DCURSOR_NORMAL,
	};


//
// Defines for the database identifiers that are used by the dialogs
//
#define INPUT_DB_CLASS "INPUT"		// main input key
#define INPUTFEED_DB_CLASS "PRICEFEED"	// default secondary key

#define FEED_TELETEXT	"TELETEXT"	// feed via teletext
#define FEED_ASCII	"ASCII"		// feed via ascii

#define FONT_DB_CLASS "FONT"		// main font
#define FONT_PROP1_CLASS "PROP1"		
#define FONT_PROP2_CLASS "PROP2"		
#define FONT_FIXED_CLASS "FIXED"		

#define UNITS_DB_CLASS "MEASUREUNITS"	// units for measurements
#define UNITSDEFINED_DB_CLASS "DEFINED"	// secondary key
#define MUNITS_INCHES	"INCHES"
#define MUNITS_CM	"CM"
#define MUNITS_POINTS	"POINTS"

//
// Public variables
//
extern Dialogvertical *d_mainvert;	// main vertical dialog

//
// Public functions
//
extern int DialogYesno(char *str,char *truestr,char *falsestr);
extern void DialogMessage(char *errstr);
extern void DialogPanic(char *errstr);
extern void DialogUpdatedisplay();
extern void DialogEvent();
extern int DialogString(char *,char *,char *,int ,...);
extern void DialogRootmenu(menu_bar &mb);
extern void DialogInitialize(unsigned int argc, char **argv,menu_bar &mb);
extern char *DialogOptionlist(char *title,char *ltitle,char **items);
extern void DialogMainloop();
extern void DialogNoflush();
extern void DialogFlushoutput();
extern void DialogFlush();
extern int DialogPrintf(char *,...);
extern dcursortype_e DialogCursor(dcursortype_e cursortype);

#endif
