
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: div.h,v 2.3 1995/06/09 20:53:44 davidsi Exp $
//
// $Log: div.h,v $
// Revision 2.3  1995/06/09  20:53:44  davidsi
// - Added GNU copyright message.
//
// Revision 2.2  1995/03/08  21:23:29  davidsi
// - Factored the dividend class into two classes, the dividend class provides
//   all the main dividend handling functionality and the Dividendedit class
//   allows dividends to be edited.  This reduces the amount of memory
//   req'd to represent a diviend.
// - The extract() member has been renamed to DividendExtract.
// - main_div_list is now declared extern.
//
// Revision 2.1  1995/03/02  21:01:56  davidsi
// - Removed all dialog references - uses the Dialog class instead.
//
// Revision 2.0  1995/01/19  22:01:27  davidsi
// - Removed the time member of the dividend class.  Dates are now managed
//   using the date class.
// - Added add_vect to the div_list class so that callbacks are handled
//   when dividends are added.
//
// Revision 1.1  1994/06/28  21:01:04  davidsi
// Initial revision
//
//
#ifndef __DIV_H
#define __DIV_H

#include <stdio.h>
#include <string.h>

#include "security.h"
#include "dslist.h"
#include "record.h"
#include "date.h"
#include "button.h"
#include "dialog.h"
#include "optionl.h"


//
// The class that provides a list of dividends on a per security basis
//
class secdiv_list : public dslist_head {
public:
	//
	// These are to prevent needing casts
	//
        class dividend *dual_head();
};


//
// The class that lists ALL dividends.
//
class div_list : public dslist {
public:
	class callback add_vect;	// callbacks for when a dividend
					// is added.
	trans_list();
	div_list();
};

//
// Defines for the type field
//
enum divtype_e {
	DIV_TYPE_CASH,
	DIV_TYPE_STOCK
	};


//
// The Dividendedit class
//
class Dividendedit : private Button, private Optionlist {
	divtype_e type;			// dividend types
	//
	// Dialogs (etc) used when adding a dividend
	//
	Dialogshell d_shell;		// top most dialog
	Dialogvertical d_vert;
	Dialoghorizontal d_horiz;
	Dialogentry d_entrylist;
	Dialoghorizontal d_buttonh;
	Optionlistinstance *d_secname;

private:
        void addw_figures();
        void addw_secname();
	//
	// Overrides the function in the Button class
	//
	virtual void ButtonCallback(unsigned long id,buttonstate_e flag);

public:
	void DividendeditEdit(char *title,divtype_e);

	Dividendedit();
	virtual ~Dividendedit();
};



//
// The dividend class
//
class dividend : public dslist_node, private Button, private Optionlist {
	class date date;	// date dividend was recieved
	double 	number;		// number of securities that dividend applies to
	double 	amount;		// amount of dividend received 
	double 	tax;		// amount of tax paid 
	divtype_e type;		// dividend types
	class security *sec;	// security to which this applies

public:
	dividend();
	~dividend();

        void DividendCreate(class security *,
			class date &,
			double n,
			double amt,
			double tx,
			divtype_e type);
	void DividendExtract(class date *,
			double *n,
			double *amt,
			double *tx,
			divtype_e *type);

	//
	// Inheiritated from the slist class
	// 
        virtual void save_record(class record *);
	void DividendLoad( class record *rec );

	//
	// These are to prevent needing casts
	//
        class dividend *dual_next();
};

//
// Externally visible variables
// 
extern div_list *main_div_list;


#endif
