
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: dplot.C,v 2.9 1995/06/24 14:34:43 davidsi Exp $
//
// $Log: dplot.C,v $
// Revision 2.9  1995/06/24  14:34:43  davidsi
// - When no price information is available don't quietly ignore the user's
//   request - display a warning.
//
// Revision 2.8  1995/06/23  20:58:08  davidsi
// - Before plotting anything check a price database has been specified.
//   Putting up a dialog during an expose callback, with some X server/
//   manager combinations, can cause problems.
//
// Revision 2.7  1995/06/07  20:53:07  davidsi
// - BUG FIX. Fixed many memory leaks.
// - BUG FIX. Removed unecessary delete calls that could lead to serious
//   system failures - the class members were being deleted automatically.
// - BUG FIX. Skip over the initial tab when displaying transactions.
// - BUG FIX. Specify DTEXT_FLAG_NODRAWPAGE when drawing the different
//   Drawtext components, otherwise page throws may appear before all of
//   the page has been produced.
//
// Revision 2.6  1995/06/02  19:58:48  davidsi
// - Removed unnecessary call to DrawExtract().
//
// Revision 2.5  1995/05/19  18:55:47  davidsi
// - Updated the summary display so that complete information is provided
//   for the selected share.
//
// Revision 2.4  1995/03/08  21:25:29  davidsi
// - Factored the dividend class into two classes, the dividend class provides
//   all the main dividend handling functionality and the Dividendedit class
//   allows dividends to be edited.  This reduces the amount of memory
//   req'd to represent a diviend.
// - The extract() member has been renamed to DividendExtract.
// - The extract() member of security was renamed to SecurityExtract(), so
//   so this module modified to reflect that.
// - BUG FIX. dividend::save_record() was incorrectly writing the address
//   of the date string to the file, rather than the date string itself.
//
// Revision 2.3  1995/03/02  21:39:07  davidsi
// - The dplot class is now also derived from the menu class.  This allows
//   a much cleaner menu() function to implemented.
// - Added full printer support via the Printer class.
//
// Revision 2.2  1995/02/05  18:00:10  davidsi
// - References to drawdialog are now Drawdialog (straight name change).
//
// Revision 2.1  1995/01/31  22:21:03  davidsi
// - Changes to support name changes in other classes.
//
// Revision 2.0  1995/01/19  22:02:55  davidsi
// - Revised to use the new date class.
//
// Revision 1.1  1994/12/11  20:47:38  davidsi
// Initial revision
//
//

#include <stdio.h>
#include <varargs.h>
#define _STDARG_H
#include <stdlib.h>
#include <string.h>
#include <dial.h>
#include <math.h>

#include "localmacros.h"

#include "dplot.h"
#include "drawtext.h"
#include "price.h"
#include "callback.h"
#include "printer.h"



//
// The Dplotprint class
//
// The sole function of this class is to support the textdialog class.
// It provides an issolated class that can inherit from the Printer 
// class without any interactions (ie. from Menu or Button classes).
//
class Dplotprint : public Printer 
{
	class dplot *dplotptr;
private:
	//
	// Overrides the virtual functions in the Printer class
	//
	virtual void PrinterPage(int pno);
public:
	Dplotprint(class dplot *);
	virtual ~Dplotprint();
};

void
Dplotprint::PrinterPage(int)
{
	Draw *old_draw;

	old_draw=dplotptr->dd;
	dplotptr->dd=dplotptr->outputdraw;

        dplotptr->outputdraw->DrawClearpage();
	dplotptr->outputdraw->DrawSetfont("Helvetica",12);
	dplotptr->plot();
	dplotptr->dd=old_draw;
}

Dplotprint::Dplotprint(class dplot *dp)
{
	dplotptr=dp;
}

Dplotprint::~Dplotprint()
{
	if (dplotptr->outputdraw)
		delete dplotptr->outputdraw;
	delete dplotptr;
}


double ptrs[]={5.0,5.0};


void
dplot::plot()
{
	dd->DrawSetfont("Helvetica",12);
	
	dd->DrawSetcolour(FOREGROUND);
 	dd->DrawAxis(&start_date,&end_date, min,max);
	//
	//	dd->DrawSetlinestyle(2,ptrs);
	//
 	dd->DrawGraph(array,number,0,number-1);
        dd->DrawPage();
	dd->DrawSetlinestyle(0,NULL);
}

void
dplot::setsecurity(security *sec)
{
	int count;
	char notset=1;
	double *parray;
        class date date;

	sec->prices->extract(&parray,&date,&number);

	array=parray;
        for(count=0;count<number;parray++,count++) {
		if (*parray<0)
			continue;
                if (notset) {
                        max=min=*parray;
			notset=0;
		}
                if (max<*parray)
                        max=*parray;
                if (min>*parray)
                        min=*parray;
        }

	start_date=date;
	end_date=start_date;
	end_date.add(number);
}



enum menuids {
	MENU_PRINT,
	MENU_SELECT,
	MENU_PRICEG,
	MENU_SUMMARY,
};

void
dplot::MenuCallback(unsigned long id,int flag)
{
	security *newsec;
	double w,h;
	Printer *printer;

	switch(id) {
	case MENU_PRINT:
		//
		// Get output drawing area
		//
		printer=new class Dplotprint(this);
		if (printer->PrinterSetup() == -1)
			return;
		outputdraw=printer->GimmeDraw();

		if (outputdraw==NULL) {
			DialogMessage("No printer device specified");
			return;
		}
        	outputdraw->DrawStartpage();

		outputdraw->DrawExtract(w,h);

		outputdraw->DrawSetarea(0.0,h/2,w,h/4);
		if (printer->PrinterOutput(1,1) == 0)
			printer=NULL;
		break;
	case MENU_SELECT:
	        newsec=main_sec_list->select_security("Select New Security");
		if (newsec==NULL)
			return;
		setsecurity(newsec);
		dd->DrawClearpage();
		plot();
		break;
	default:
		;
	}
}


void
dplot::DrawdialogResize()
{
	dd->DrawClearpage();
	plot();
}



//
// dplot
//
// Constructor for the dplot display class.
//
// INPUTS : None
// RETURNS: None
//
dplot::dplot()
{
	outputdraw=NULL;
}

//
// ~dplot
//
// Destructor for the dplot display class.
//
// INPUTS : None
// RETURNS: None
//
dplot::~dplot()
{
}

void
dplot::setoutput(Draw *drw)
{
	dd=drw;
}



//
// ipriceplot
//
// Main entry point for creating the graph window
//
// INPUTS :	None
// RETURNS:	None.
//
void
dplot::ipriceplot()
{
	class Drawdialog *drawdialog;
	drawd_reg_t drt;

	drt.flags=ALLOW_SIZE_REDRAW|INITIAL_SIZE_SET;
        drt.width=360;
        drt.height=360;
        drt.title="Security Price";

        sec=main_sec_list->select_security("Plot Security Prices");
        if (sec==NULL)
                return;

	setsecurity(sec);

	MenuItemaction("Project","Print",MENU_PRINT);
	MenuItemaction("Security","Select",MENU_SELECT);
	drt.mb=MenuExtract();

	DrawdialogCreate(&drt);
	dd=this;
        dd->DrawStartpage();
}


//
// The Summaryprint class
//
// The sole function of this class is to support the textdialog class.
// It provides an issolated class that can inherit from the Printer 
// class without any interactions (ie. from Menu or Button classes).
//
class Summaryprint : public Printer 
{
	class dplot *dplotptr;
        class Drawtext *dps;
	Holding *holding;
private:
	//
	// Overrides the virtual functions in the Printer class
	//
	virtual void PrinterPage(int pno);
public:
	Summaryprint(class dplot *,Holding *holding);
	virtual ~Summaryprint();
};


void
Summaryprint::PrinterPage(int)
{
	double number,cost,received;
	double w,h;
	double ix,iy;
	char *name;
	double price;
	char buffer[1024];
	class attr *attr;
	class security *sec;
	dividend *div;
	dividend *head_div;
	class date date;
	char *datebuf;
	double amount;
	double tax;
	char str[256];
	int day,month,year;
	class date start_date,end_date;
	divtype_e type;
	int i;
	transaction *trans;		// current transaction
	transaction *head_trans;	// head of transaction list


	sec=holding->HoldingSecurity();
	dplotptr->dd->DrawClearpage();

	// if (dps)
	//	delete dps;

       	dps=new Drawtext(dplotptr->dd);		// max 2 fields for moment!
	attr=new class attr(FIELDATTR_RIGHT);
	dps->TextSetdefaultattr(2,attr);
	//
	// set font for text output
	//
	dps->DrawSetfont("Helvetica",16);
	dps->TextClear();

	//
	// Output name
	//
	sec->SecurityExtract(&name,NULL,NULL,NULL,NULL);
	dps->TextAdd(name);

	//
	// Output todays price
	//
	end_date.setdate();		    // until the date now
	price=sec->prices->PriceFetch(&end_date,TRUE);
	sprintf(buffer,"Price Today\t%6.2f",price);
	dps->TextAdd(buffer);

	//
	// Output holding, total effective cost of those held
	//
	holding->HoldingTransactionhead()->extract(0,&end_date,&number,
					&cost,&received,NULL);

	sprintf(buffer,"Holding\t%6.2f",number);
	dps->TextAdd(buffer);
	sprintf(buffer,"Value\t%6.2f",price*number);
	dps->TextAdd(buffer);
	sprintf(buffer,"Cost\t%6.2f",cost-received);
	dps->TextAdd(buffer);

	dps->TextComplete();
	dps->DrawtextPrepare();

	//
	// Plot it
	//
	dps->DrawtextOutput(1,DTEXT_FLAG_NODRAWPAGE);


	//
	// Now get the dividends for the last year.
	//
	dps->DrawPush();
	dps->DrawGetarea(ix,iy,w,h);
	dps->DrawSetarea(ix,iy+h/4,w,h/4);


	dps->TextClear();

	for(i=1;i<2;i++) {
		class attr *attr;
		attr=new class attr(FIELDATTR_LEFT,FOREGROUND);
		dps->TextSetdefaultattr(i,attr);
	}

	for(i=2;i<5;i++) {
		class attr *attr;
		attr=new class attr(FIELDATTR_RIGHT,FOREGROUND);
		dps->TextSetdefaultattr(i,attr);
	}

	dps->DrawtextTitle("Dividends In The Last Year");
	dps->TextAdd("Date\tNumber\tTax\tAmount");

	start_date.setdate();
	//
	// This would give from start of current year
	//
	// start_date.getdate(day,month,year);
	// start_date.setdate(1,1,year);
	//
	start_date.sub(365);		// dividends in the last 365 days
	end_date.setdate();             // 


	head_div=div=holding->HoldingDividend();
	while (div!=NULL ) {
		div->DividendExtract(&date,&number,&amount,&tax,&type);

		if ( date.inrange(&start_date,&end_date) ) {
			datebuf=date.getdate();
			sprintf(str,"%s\t%.4f\t%.2f\t%.2f",
				  datebuf,number,tax,amount);
			dps->TextAdd(str);
		}
		div=div->dual_next();
		if (head_div==div)
			break;
	}

	dps->TextComplete();
	dps->DrawtextPrepare();
	dps->DrawtextOutput(1,DTEXT_FLAG_NODRAWPAGE);
	dps->DrawPop();

	//
	// Now get the transactions for the last year.
	//
	dps->DrawPush();
	dps->DrawGetarea(ix,iy,w,h);
	dps->DrawSetarea(ix,iy+h/2,w,h/4);
	dps->TextClear();

	dps->DrawtextTitle("Transactions In The Last Year");
	dps->TextAdd("Type\tDate\tNumber\tPrice\tTotal");

	head_trans=holding->HoldingTransaction();
	trans=head_trans;
	while ( trans!=NULL ) {
		//
		// Check date of transaction
		//
		trans->extract(NULL, &date,NULL,NULL,NULL);

		if ( date.inrange(&start_date,&end_date) ) {
			name=trans->trans2string(NAMEFIRST,TRANS_ALL,"");
			//
			// The [1] allows for the tab which will 
			// appear in the first character position of the
			// string that is returned - we didn't pass a 
			// security name, so name is included in the returned
			// string but the tab delimiter is.
			//
			dps->TextAdd(&name[1]);
		}
		//
		// Next transaction
		//
		trans=trans->dual_next();
		if (trans==head_trans)
			break;
	}
	dps->TextComplete();
	dps->DrawtextPrepare();
	dps->DrawtextOutput(1,DTEXT_FLAG_NODRAWPAGE);
	dps->DrawPop();



	dplotptr->dd->DrawPush();
	dplotptr->dd->DrawGetarea(ix,iy,w,h);
	dplotptr->dd->DrawSetarea(ix,iy+h*.75,w,h/4);
	dplotptr->plot();
	dplotptr->dd->DrawPop();
}


Summaryprint::Summaryprint(class dplot *dp,Holding *iholding)
{
	dplotptr=dp;
	holding=iholding;
	dps=NULL;
}

Summaryprint::~Summaryprint()
{
	delete dplotptr;
}
//
//
//
void
summary()
{
	Holding *holding;
        dplot *dp;
        Draw *outputdraw;
	Summaryprint *printer;

	holding=portfolio->HoldinglistSelect("Select Security");
        if (holding==NULL)
		return;
       	dp=new dplot;
       	dp->setsecurity(holding->HoldingSecurity());
	//
	// Get output drawing area
	//
	printer=new class Summaryprint(dp,holding);
	if (printer->PrinterSetup() == -1)
		return;
       	outputdraw=printer->GimmeDraw();
	if (outputdraw==NULL) {
		DialogMessage("No printer device specified");
		return;
	}
	dp->setoutput(outputdraw);

       	outputdraw->DrawStartpage();

	if (printer->PrinterOutput(1,1)==0)
		printer=NULL;
}

static void
display_graph(local_va_dcl)
{
        dplot *dp;
	va_list ap;
	dialog d;
	unsigned long code;
	class price *ptr;

	//
	// Before go any further check that there are some prices
	// out there - if we put up a dialog during a expose 
	// callback some X server/manager combinations have 
	// problems..
	//
	ptr=new class price(NULL);
	if (!strcmp("not a security",ptr->PriceFilename("not a security"))) {
		DialogMessage("Please define a path for prices before using this operation");
		delete ptr;
		return;
	}
	 
	delete ptr;

        local_va_start(ap);
        d=va_arg(ap,dialog);
        code=va_arg(ap,unsigned long);

        switch(code) {
	case MENU_PRICEG:
        	dp=new dplot;
		dp->ipriceplot();
		break;
	case MENU_SUMMARY:
		summary();
		break;
	}
}


//
// dval_init parent menu_bar
//
// Initialises the display valuations module.
// 
// INPUTS :	None
// RETURNS:	None
//
void
dplot_init(menu_bar *mb)
{
        MENU("Analysis","Plot Security Price",display_graph,MENU_PRICEG);
        MENU("Analysis","Show Summary",display_graph,MENU_SUMMARY);
}
