
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: dprice.h,v 2.4 1995/05/18 21:03:34 davidsi Exp $
//
// $Log: dprice.h,v $
// Revision 2.4  1995/05/18  21:03:34  davidsi
// - Callbacks are now stored using opaque handles rather than pointers to
//   the actual nodes (see callback.C for the whole story).
//
// Revision 2.3  1995/04/16  20:31:45  davidsi
// - The date_changed() callback now must use class date pointers to
//   match changes to the Setdate class.
//
// Revision 2.2  1995/03/01  22:21:58  davidsi
// - The function member price_display::editprice() no longer requires
//   a dialog to be passed.
// - Removed all dialog references - use the Dialog class instead.
//
// Revision 2.1  1995/02/05  18:02:26  davidsi
// - Added sdate member to handle the setting of dates.
// - The date_changed() member function is now static to allow it to be
//   used in callbacks.
//
// Revision 2.0  1995/01/19  22:04:55  davidsi
// - Revised to use the new date class.
// - Added callback_price node for when prices change and a window is on
//   display.
// - Added redisplay() member to handle the price change callback.
//
// Revision 1.1  1994/06/29  21:04:26  davidsi
// Initial revision
//
//
#ifndef _DPRICEH
#define _DPRICEH


#include "date.h"
#include "setdate.h"
#include "price.h"
#include "dialog.h"
#include "button.h"
#include "optionl.h"

class price_display : private Button, private Optionlist 
{
	class date date_start;	// date for start of prices being displayed
	class date date_end;	// as date_start but end of date range
	dialog d_pricelist;	// list of prices for a security -
				// this is not the base of the tree
				// but rather a dialog within a much
				// larger tree.
	class security *sec;	// security whose prices are being
				// displayed

	Callbackid callback_price; // callback when prices are changed
	class Setdate *sdate;

	Dialogshell d_shell;
	Dialogvertical d_mainv;
	Dialoghorizontal d_pricemain_h;
	Dialogvertical d_secmain_v;
	Dialogvertical d_pricemain_v;
	Dialoghorizontal d_buttonh;
	Optionlistinstance *d_securities;
	Optionlistinstance *d_prices;

	int editenabled;

private:
	char **list_prices();
	static void redisplay(char *realthis,char *);
	virtual void OptionlistCallback(unsigned long id);
	virtual void ButtonCallback(unsigned long id,buttonstate_e flag);
public:
        price_display();
        ~price_display();

	void create_display(int mode);
	void price_changed(security *);
	static void date_changed(char *rthis,class date *);
	void editprice();
};


#endif
