
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: draw.h,v 2.4 1995/05/18 21:03:12 davidsi Exp $
//
// $Log: draw.h,v $
// Revision 2.4  1995/05/18  21:03:12  davidsi
// - Added a DrawGetarea() that returns the information previously set by a
//   DrawSetarea() or the defaults if DrawSetarea() has not yet been called.
//
// Revision 2.3  1995/03/01  22:33:27  davidsi
// - All colour handling has been moved to colour.h
// - Added support for the Printerdraw class.  This inherits all
//   draw functionality but provides for specifying a file or printer
//   based output device.
// - A gfx_context structure now defines the state that DrawPush() and
//   DrawPop() handle.  A stack of these is supported by the Draw class.
// - Added DrawRef() and DrawUnref() function members to allow multiple
//   references to the same Draw instance.
//
// Revision 2.2  1995/02/05  18:01:32  davidsi
// - Added DrawRectangle() and drv_rect() member functions.
//
// Revision 2.1  1995/02/02  21:19:53  davidsi
// - Name change.  All public function members of this class are now prefixed
//   with Draw.  Some names have been further simplified.
// - Added colour handling.
//
// Revision 2.0  1995/01/19  21:56:01  davidsi
// - Added drv_setlinewidth() function to allow line widths to be set.
// - Added drv_setcolour.
// - Modified to use the new date class.
// - Added new setcolour function and enumerated type.
//
// Revision 1.1  1994/12/11  21:17:42  davidsi
// Initial revision
//
//
//
#ifndef __DRAW_H
#define __DRAW_H

#include "colour.h"

#include "paper.h"


//
// Basic graphics context information
//
struct gfx_context {
	double width;		// width of drawing area in points
	double height;		// height of drawing area in points

	double xbase;		// where drawing area starts on paper
	double ybase;

	double x_conv;		// converts points to physical units (in x)
	double y_conv;		// converts points to physical units (in y)

	double xpos,ypos;	// current coorindate in device indept. space.
	struct gfx_context *next;
};

//
//
//
class Draw : public Colour {
	int refcount;

	//
	// Set by axis() used by any graph plot function
	//
	double g_width;
	double g_height;
	double xb;
	double yb;
	double xmax,xmin;
	double ymax,ymin;

	int greylevel;		// grey level (percent)
	struct gfx_context *stack;

protected:
	double width;		// width of drawing area in points
	double height;		// height of drawing area in points

	double xbase;		// where drawing area starts on paper
	double ybase;

	double x_conv;		// converts points to physical units (in x)
	double y_conv;		// converts points to physical units (in y)

	double xpos,ypos;	// current coorindate in device indept. space.


protected:
	virtual void drv_startpage();
	virtual void drv_clearpage();
	virtual void drv_moveto(double x,double y);
	virtual void drv_drawto(double x,double y);
	virtual void drv_rect(double x1,double y1,double x2,double y2);
	virtual void drv_text(char *string);
	virtual void drv_drawpage();
	virtual void drv_setfont(char *name,double size);
	virtual double drv_textlength(char *text);
	virtual double drv_textheight();
	virtual void drv_setgrey(int percentage);
	virtual void drv_setlinestyle(int count,double *lens);
	virtual void drv_setlinewidth(double width);
	virtual void drv_setcolour(colour_e);

public:
	//
	// Front end function members for driver calls
	//
	void DrawStartpage();
	void DrawClearpage();
	void DrawMoveto(double x,double y);
	void DrawDrawto(double x,double y);
	void DrawRectangle(double x,double y,double w,double h);
	void DrawText(char *string);
	void DrawPage();
	void DrawSetfont(char *name,double size);
	double DrawTextlength(char *text);
	double DrawTextheight();
	void DrawSetgrey(int percentage);
	void DrawSetlinestyle(int count,double *lens);
	void DrawSetlinewidth(double width);
	void DrawSetcolour(colour_e);
	//
	// Other goodies..!
	// 
	void DrawExtract(double &width, double &height);
	void DrawSetarea(double xpos,double ypos,double width,double height);
	void DrawGetarea(double &xpos,double &ypos,double &w,double &h);
	void DrawPush();
	void DrawPop();

	void DrawUnref();
	void DrawRef();

	//
	// Graphing plotting functions
	//
	void DrawAxis(class date *xmn,class date *xmx,double ymin,double ymax);
	void DrawGraph(double *data,int ndata,int startd,int endd);

	Draw();
	virtual ~Draw();
};

//
// The Printerdraw class is used by classes wishing to inherit
// draw functionality but that also control printer output.  The
// Printerdraw class adds printer management facilities to the
// base Draw class.
//
enum printermode_e {
	PRTMDE_FILE,
	PRTMDE_PRINTER
	};

class Printerdraw : public Draw 
{
protected:
	printermode_e currentmode;
public:
	virtual void PrinterdrawSetup();
	void PrinterdrawMode(printermode_e);
	Printerdraw();
	virtual ~Printerdraw();
};

#endif
