
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: drawdialog.h,v 2.4 1995/06/09 20:53:44 davidsi Exp $
//
// $Log: drawdialog.h,v $
// Revision 2.4  1995/06/09  20:53:44  davidsi
// - Added GNU copyright message.
//
// Revision 2.3  1995/03/02  20:51:45  davidsi
// - Added DrawdialogCreate() so that creation is not linked with
//   specifying how the dialog will be displayed.
// - Support new POPUP window.  This is a simple window with no menus or
//   scroll bars.  Ideal for simple dialogs.
//
// Revision 2.2  1995/02/05  18:02:05  davidsi
// - Keep a copy of the original flags in a class data member.
// - Provide NOSCROLL and NOMENU options.
// - Added the drv_rect() member function.
//
// Revision 2.1  1995/02/02  21:22:00  davidsi
// - The drv_drawto() and drv_moveto() member functions now handle lx,ly
//   variables as the draw class does not.
// - The drv_setfont() function now takes a fontname.
// - Modifications to reflect name changes in the Draw class.
// - This class now called Drawdialog rather than drawdialog to be
//   consistent with other name changes.
//
// Revision 2.0  1995/01/19  22:03:58  davidsi
// - Added the drv_setlinewidth() functions.
// - Added drv_setcolour.
//
// Revision 1.1  1994/12/11  21:37:31  davidsi
// Initial revision
//
//
//
#ifndef __DRAWDIALOG_H
#define __DRAWDIALOG_H

#include "suplib.h"
#include "menu.h"
#include "draw.h"
#include "callback.h"

//
// The drawd_reg defines how the draw dialog will behave when it is 
// realized.
//
typedef struct drawd_reg {
        unsigned long flags;		// see below
        double width;			// initial width
        double height;			// initial height
	menu_bar *mb;			// menu bar to use or NULL
	char *title;			// title of window
} drawd_reg_t;

#define ALLOW_SIZE_REDRAW       0x1     // make size_change callback
                                        // when size changes or to perform
                                        // initial rendering, else prevent
                                        // size from changing.
#define INITIAL_SIZE_SET	0x2	// use specified width/height else
					// assume default
#define NOSCROLL		0x4	// don't render any scrollbars
#define NOMENU			0x8	// don't render any menubars
#define POPUP			0x10	// simple pop up window - no menus
					// or scroll bars



class Drawdialog : public Draw {

private:
	int oflags;			// flags from creation information
	double lx,ly;			// last coordinate handled by device
	double p_width,p_height;	// paper width and height

        dialog d_shell;
        dialog d_area;
	char *title;			// name of window
	menu_bar *add_mb;		// menu bar to use or NULL
	//
	// Font information
	//
	double fontsize;		// current size of font in points
	char *fontname;			// current font name
	double font_height;		// current font height in points

	class mcallback menudata;	// for encoding menu callbacks 

private:

	//
	// These handle callbacks
	//
	static void drv_sizechange(local_va_dcl);
	static void drv_menu(local_va_dcl);

	//
	// These are the actual drawing functions for a dialog
	// based drawing.
	//
	virtual void drv_startpage();
	virtual void drv_clearpage();
	virtual void drv_moveto(double x,double y);
	virtual void drv_drawto(double x,double y);
	virtual void drv_rect(double x1,double y1,double x2,double y2);
	virtual void drv_text(char *string);
	virtual void drv_setfont(char *name,double fsize);
	virtual void drv_drawpage();
	virtual double drv_textlength(char *text);
	virtual double drv_textheight();
        virtual void drv_setgrey(int percentage);
        virtual void drv_setlinestyle(int count,double *len);
	virtual void drv_setlinewidth(double width);
	virtual void drv_setcolour(colour_e);

protected:
	virtual void DrawdialogResize();

public:
	void DrawdialogCreate(drawd_reg_t *);
	Drawdialog();
	virtual ~Drawdialog();
};

#endif
