
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//    Font widths derived from troff-to-lj by Sverre Froyen & David MacKenzie
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: drawlj.C,v 2.3 1995/06/24 14:35:31 davidsi Exp $
//
// $Log: drawlj.C,v $
// Revision 2.3  1995/06/24  14:35:31  davidsi
// - Store the user's preferred output file and the print spooler
//   command line when specified.
//
// Revision 2.2  1995/06/07  21:02:50  davidsi
// - Support for fonts is now builtin - rather than using the postscript
//   class.
//
// Revision 2.1  1995/05/19  18:54:18  davidsi
// - DialogentryCreate() now requires an argument specifying total number
//   of visible columns.
//
// Revision 2.0  1995/03/02  21:37:35  davidsi
// - Initial version.
//
//
//



//
// Include Files
//
#include <stdio.h>
#include <varargs.h>
#define _STDARG_H
#include <stdlib.h>
#include <string.h>
#include <dial.h>
#include <math.h>

#include "localmacros.h"
#include "paper.h"


#include "drawlj.h"
#include "menu.h"
#include "printer.h"
#include "prefs.h"

//
// This appears to be missing from the standard include files
//
extern "C" {
	extern int unlink(char *);
}

//
// All postscript file output will go to this file unless specified 
// otherwise
//
#define DEFAULT_OUTPUTFILE "output.hpgl"
#define DEFAULT_OUTPUTCMD "lp -oraw -c"
#define PRTLJ_CLASS     "PRINTER_LJ"
#define PRTLJ_CMD       "CMD"
#define PRTLJ_FILE      "FILE"


//
// This is the current (real) filename for postscript output
//
static char *filename;
//
// This is the command needed to output to the printer
//
static char *printercmd;

//
///////////////////////////////////////////////////////////////////////////////
//
// Font widths start here
//
///////////////////////////////////////////////////////////////////////////////
//

char w_troman[256-32] = {
12,		/* space */
10,		/* ! */
0,		/* " */
0,		/* # */
20,		/* $ */
30,		/* % */
30,		/* & */
10,		/* ' */
13,		/* ( */
14,		/* ) */
19,		/* * */
20,		/* + */
12,		/* , */
20,		/* - hyphen */
10,		/* . */
18,		/* / */
20,		/* 0 */
20,		/* 1 */
20,		/* 2 */
20,		/* 3 */
20,		/* 4 */
20,		/* 5 */
20,		/* 6 */
20,		/* 7 */
20,		/* 8 */
20,		/* 9 */
9,		/* : */
10,		/* ; */
0,		/* < */
36,		/* = */
0,		/* > */
19,		/* ? */
0,		/* @ */
30,		/* A */
25,		/* B */
27,		/* C */
29,		/* D */
26,		/* E */
24,		/* F */
31,		/* G */
30,		/* H */
14,		/* I */
17,		/* J */
33,		/* K */
26,		/* L */
36,		/* M */
30,		/* N */
29,		/* O */
22,		/* P */
29,		/* Q */
29,		/* R */
24,		/* S */
26,		/* T */
32,		/* U */
29,		/* V */
33,		/* W */
31,		/* X */
29,		/* Y */
27,		/* Z */
11,		/* [ */
0,		/* \ */
12,		/* ] */
0,		/* ^ */
0,		/* _ */
10,		/* ` */
18,		/* a */
22,		/* b */
18,		/* c */
22,		/* d */
18,		/* e */
16,		/* f */
20,		/* g */
22,		/* h */
12,		/* i */
15,		/* j */
22,		/* k */
11,		/* l */
32,		/* m */
22,		/* n */
20,		/* o */
22,		/* p */
21,		/* q */
16,		/* r */
15,		/* s */
13,		/* t */
22,		/* u */
21,		/* v */
29,		/* w */
21,		/* x */
21,		/* y */
17,		/* z */
0,		/* { */
7,		/* | */
0,		/* } */
0,		/* ~ */
6,		/* narrow space */
20,		/* hyphen */
12,		/* bullet */
22,		/* square */
33,		/* 3/4 em dash */
30,		/* rule */
18,		/* 1/4 */
18,		/* 1/2 */
0,		/* 3/4 */
20,		/* minus */
0,		/* fi */
0,		/* fl */
0,		/* ff */
0,		/* ffi */
0,		/* ffl */
14,		/* degree */
16,		/* dagger */
0,		/* section (unimplem) */
8,		/* foot mark */
0,		/* acute acc (unimplem) */
0,		/* grave acc (unimplem) */
0,		/* underrule (unimplem) */
0,		/* slash (unimplem) */
3,		/* half narrow space */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
22,		/* registered */
22,		/* copyright */
0,		/* null */
21,		/* cent */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0,		/* null */
0		/* null */
};


static double 
compute_width( char *text , double pointsize)
{
	char load;
	char *ptr;
	double wi=0.0;
	
	ptr=text;
	while ( (load=*ptr) != '\0' )	{
		if (load>=32)
			wi += (double)(w_troman[load-32]);
		ptr++;
	}
	return( (wi*pointsize)/(4.0*10.0) );
}



//
///////////////////////////////////////////////////////////////////////////////
//
// The Setlj class provides a private mechanism for controlling
// the laserjet specific options.
//
///////////////////////////////////////////////////////////////////////////////
//

//
// Identifiers for the different inputs within the dialog
//
enum dialogids {
	BUTTON_CANCEL,		// The cancel button
	BUTTON_OK,		// The ok button
	ID_FNAME,		// The filename text
	ID_CMD			// the cmd text
	};


//
// Setlj simply provides an issolate class that inherits the Button
// capabilities.
//
class Setlj : private Button
{
	Dialogshell d_shell;		// main shell
	Dialogentry d_entrylist;	// entry list for filename etc
	Dialogvertical d_vert_main;	// main form 
	Dialoghorizontal d_horiz_ok;	// form for buttons

	int preferences_set;		// TRUE when ok/cancel hit
public:
	//
	// This overrides the pure virtual button handler in the Button class
	//
	virtual void ButtonCallback(unsigned long id,buttonstate_e flag);

	Setlj();
};

//
// Handles the button callbacks
//
void
Setlj::ButtonCallback(unsigned long id,buttonstate_e flag)
{
	char *str;

	switch(id) {
	case BUTTON_CANCEL:
		//
		// Nothing to do for a cancel
		//
		preferences_set=1;
		return;
	case BUTTON_OK:
		//
		// For the OK case, before we can set 
		// "preferences_set" we must load the
		// information in the dialog into the
		// main variables - ready for use.
		//
		str=d_entrylist.DialogentryGet(ID_FNAME);
		if (str) {
			if (filename)
				free(filename);
			filename=(char *)malloc(strlen(str)+1);
			strcpy(filename,str);
		}
		str=d_entrylist.DialogentryGet(ID_CMD);
		if (str) {
			if (printercmd)
				free(printercmd);
			printercmd=(char *)malloc(strlen(str)+1);
			strcpy(printercmd,str);
		}
		preferences_set=1;

		//
		// Update the database
		//
		db_prefs->DatabaseInsert(PRTLJ_CLASS,PRTLJ_CMD,printercmd);
		db_prefs->DatabaseInsert(PRTLJ_CLASS,PRTLJ_FILE,filename);
		break;
	}
}


//
// The Setlj constructor does all the work needed to build the
// entire dialog.
//
Setlj::Setlj()
{
        d_shell.DialogshellCreate("Printer Setup");
	d_shell.DialogSetmodal();
	d_vert_main.DialogverticalCreate(d_shell);

	d_entrylist.DialogentryCreate(d_vert_main,10);
	d_entrylist.DialogentrySet("Filename",filename,ID_FNAME);
	d_entrylist.DialogentrySet("Output Command",printercmd,ID_CMD);

	d_horiz_ok.DialoghorizontalCreate(d_vert_main);
	(void)ButtonPush(d_horiz_ok,"OK",BUTTON_OK);
	(void)ButtonPush(d_horiz_ok,"CANCEL",BUTTON_CANCEL);

        d_shell.DialogRealize();

	preferences_set=0;
	while (preferences_set==0)
		DialogEvent();
}

//
///////////////////////////////////////////////////////////////////////////////
//
// Main laser jet functions
//
///////////////////////////////////////////////////////////////////////////////
//

//
// If you change this macro, please also check the function member
// Drawlj::drv_setfont()
//
#define FORCEHPGL					\
	if ( !(ljflags & LJFLAGS_HPGLMODE) && file ) {	\
		fprintf(file,"%cE%c%%0BIN;",27,27);	\
		ljflags|=LJFLAGS_HPGLMODE;		\
		drv_setfont(current_font,fontsize);	\
	}

#define PLU_PER_INCH	1016
	
	

//
// drv_moveto
//
// INPUTS :
// RETURNS: 
void
Drawlj::drv_moveto(double x,double y)
{
	if (file==NULL)
		return;
	FORCEHPGL
	fprintf(file,"PU%d,%d;",(int)(pxbase+x),(int)(pybase-y));
}

//
// drv_drawto
//
// INPUTS :
// RETURNS: 
//
void
Drawlj::drv_drawto(double x,double y)
{
	if (file==NULL)
		return;

	FORCEHPGL
	fprintf(file,"PD%d,%d;",(int)(pxbase+x),(int)(pybase-y));
}

//
// drv_rect
//
// INPUTS :
// RETURNS: 
//
void
Drawlj::drv_rect(double x1,double y1,double x2,double y2)
{
	if (file==NULL)
		return;

	FORCEHPGL
	drv_moveto(x1,y1);
	fprintf(file,"PD%d,%d;",(int)(x2),(int)(y1));
	fprintf(file,"PD%d,%d;",(int)(x2),(int)(y2));
	fprintf(file,"PD%d,%d;",(int)(x1),(int)(y2));
	fprintf(file,"PD%d,%d;",(int)(x1),(int)(y1));
}



//
// drv_text
//
// INPUTS :
// RETURNS: 
//
void
Drawlj::drv_text(char *text)
{
	if (file==NULL)
		return;

	FORCEHPGL
	fprintf(file,"DT\n;");
	fprintf(file,"LB%s\n;",text);
}


struct HPGL2_font_str {
    char           *compare,
                   *name;
    int             symbol_set,
                    spacing;
    double          pitch,
                    height;
    int             posture,
                    stroke_weight,
                    typeface;
};

#define HPGL2_DEF_POINT 14      /* Height of font */
#define HPGL2_DEF_PITCH (3 * 72 / (HPGL2_DEF_POINT * 2))

static struct HPGL2_font_str HPGL2_font_table[] = {
	{"u$nivers", "univers", 277, 1, 0.0, HPGL2_DEF_POINT, 0, 0, 4148},
	{"s$tick", "stick", 277, 0, HPGL2_DEF_PITCH, 0.0, 0, 0, 48},
	{"c$g_times", "cg_times", 277, 1, 0.0, HPGL2_DEF_POINT, 0, 0, 4101}
};

//
// drv_setfont
//
// INPUTS :
// RETURNS: 
//
void
Drawlj::drv_setfont(char *name,double fsize)
{
	int x=2;

	//
	// If the request being passed is using the same (actual instance)
	// string for the font name and the size is unchanged then it
	// is a request to reprime the printer's current font due to
	// a page being written.
	//
	if (name==current_font && fsize==fontsize)
		goto reprime;

	name="Times-Roman";
	if (current_font)
		free(current_font);
	current_font=(char *)malloc(strlen(name)+1);
	strcpy(current_font,name);
	fontsize=fsize;
	font_height=fsize;

	// ps.set(current_font,(int)fsize);

reprime:
	if (file==NULL)
		return;

	//
	// Don't use the FORCEHPGL macro as that may call this
	// function and we don't want to recurse!
	//
	if ( !(ljflags & LJFLAGS_HPGLMODE) ) {
		fprintf(file,"%cE%c%%0BIN;",27,27);
		ljflags|=LJFLAGS_HPGLMODE;
	}
	//
	// Switch into PCL mode 
	//
	fprintf(file,"SP1SD1,%d,2,%d,",HPGL2_font_table[x].symbol_set,
				HPGL2_font_table[x].spacing);
	fprintf(file,"4,%lf,",font_height);
	fprintf(file,"5,%d,6,%d,7,%d;",HPGL2_font_table[x].posture,
			HPGL2_font_table[x].stroke_weight, 
			HPGL2_font_table[x].typeface);
}


//
// drv_textlength
//
// INPUTS :
// RETURNS: 
//
double
Drawlj::drv_textlength(char *txt)
{
	//
	// We multiply by x_conv because ps.textwidth() returns
	// point based information - but this driver expects
	// to use plus (plotter units).
	//
	return ( x_conv*compute_width(txt,fontsize) );
}


//
// drv_clearpage
//
// INPUTS :
// RETURNS: 
//
void
Drawlj::drv_clearpage()
{
}


//
// drv_drawpage
//
// INPUTS :
// RETURNS: 
//
void
Drawlj::drv_drawpage()
{
	if (file==NULL)
		return;

	FORCEHPGL
	//
	// Back to PCL, just in case this is the end
	//
	fprintf(file,"%c%%0A%cE",27,27);
	ljflags&=~LJFLAGS_HPGLMODE;
	fflush(file);
}

//
// drv_setgrey
//
// INPUTS :
// RETURNS: 
//
void
Drawlj::drv_setgrey(int perc)
{
}

//
// drv_setlinestyle
//
// INPUTS :
// RETURNS: 
//
void
Drawlj::drv_setlinestyle(int count,double *len)
{
}

//
// drv_setlinewidth
//
// INPUTS :
// RETURNS: 
//
void
Drawlj::drv_setlinewidth(double width)
{
}

void
Drawlj::drv_setcolour(colour_e col)
{
}

//
// drv_startpage
//
// INPUTS :
// RETURNS: 
//
void
Drawlj::drv_startpage()
{
	if (currentmode==PRTMDE_FILE)
		file=fopen(filename,"w");
	else {		// assume it will be PRTMDE_PRINTER
		char *str;

		str=tmpnam(NULL);

		if (tmpfilename)
			free(tmpfilename);
		tmpfilename=(char *)malloc(strlen(str)+1);
		strcpy(tmpfilename,str);
		file=fopen(tmpfilename,"w");
	}
		
	if (file==NULL) {
		DialogMessage("Cannot open output file");
		return;
	}
	FORCEHPGL
}

//
//
//
double
Drawlj::drv_textheight()
{
        return(font_height);
}

void
Drawlj::PrinterdrawSetup()
{
	//
	// This does all the work, when created!
	//
	class Setlj lj;
}




//
// Constructor function
//
Drawlj::Drawlj()
{
	//
	// The x_conv and y_conv values cannot be reset at any other point
	// as this can confuse classes such as the drawtext class, which
	// assumes they are constant for all draw derived classes.
	//
	x_conv=y_conv=PLU_PER_INCH/INCHES2POINTS(1.0);
	ljflags=0;

	width=INCHES2POINTS(A4_PAPER_WIDTH-0.5);
	height=INCHES2POINTS(A4_PAPER_HEIGHT-0.5);
	pybase=(A4_PAPER_HEIGHT-0.5)*PLU_PER_INCH;
	pxbase=0.0;

	file=NULL;
	current_font=NULL;
	tmpfilename=NULL;
}

//
// Destructor function
//
Drawlj::~Drawlj()
{
	if (file)
		fclose(file);

	if (currentmode==PRTMDE_PRINTER && tmpfilename) {
		char buffer[1024];

		sprintf(buffer,"%s %s",printercmd,tmpfilename);
		system(buffer);
		unlink(tmpfilename);
	}

	if (tmpfilename)
		free(tmpfilename);
}


static class Printerdraw *
ljcreate()
{
	return( new class Drawlj );
}

void
drawlj_init(menu_bar *mb)
{
	char *str;

	printerdevices.PrinterdeviceAdd("Epson EPL-5200",ljcreate);
	printerdevices.PrinterdeviceAdd("HP LaserJet",ljcreate);
	printerdevices.PrinterdeviceAdd("HP LaserJet II",ljcreate);
	printerdevices.PrinterdeviceAdd("HP LaserJet 4Si",ljcreate);

	str=db_prefs->DatabaseRetrieve(PRTLJ_CLASS,PRTLJ_FILE);
	if (str==NULL)
		str=DEFAULT_OUTPUTFILE;
	filename=(char *)malloc(strlen(str)+1);
	strcpy(filename,str);

	str=db_prefs->DatabaseRetrieve(PRTLJ_CLASS,PRTLJ_CMD);
	if (str==NULL)
		str=DEFAULT_OUTPUTCMD;
	printercmd=(char *)malloc(strlen(str)+1);
	strcpy(printercmd,str);
}
