
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: drawps.h,v 2.4 1995/06/09 20:53:44 davidsi Exp $
//
// $Log: drawps.h,v $
// Revision 2.4  1995/06/09  20:53:44  davidsi
// - Added GNU copyright message.
//
// Revision 2.3  1995/03/01  22:36:05  davidsi
// - Changed name from drawps to Drawps for consistancy.
// - Added member to store name of temporary file.
//
// Revision 2.2  1995/02/05  18:00:41  davidsi
// - Added the drv_rect() member function.
//
// Revision 2.1  1995/01/31  22:15:28  davidsi
// - Now uses the postscript class.
// - Many new members added which were previously draw members, but did not
//   really belong in the base class.
// - Modifications to reflect name changes in the draw class.
//
// Revision 2.0  1995/01/19  21:50:54  davidsi
// - Added drv_setlinewidth() function to allow line widths to be set.
// - Added drv_setcolour.
//
// Revision 1.1  1994/12/11  21:37:46  davidsi
// Initial revision
//
//
//
#ifndef __DRAWPS_H
#define __DRAWPS_H

#include "draw.h"
#include "postscript.h"

class Drawps : public Printerdraw {

private:
	unsigned long psflags;	// see below PSFLAGS_XXX
	FILE *file;		// output file
	double pxbase;		// conversion for moving (0,0) on paper
	double pybase;

	class postscript ps;

	//
	// Font information
	//
	double fontsize;	// current size of font in points
	char *fontname;		// current font name
	double font_height;	// current font height in points
	char *current_font;	// current name of postscript font 
	char *tmpfilename;	// name of temporary file

protected:

private:
	//
	// These are the actual drawing functions for a dialog
	// based drawing.
	//
	virtual void drv_clearpage();
	virtual void drv_drawpage();
	virtual void drv_drawto(double x,double y);
	virtual void drv_moveto(double x,double y);
	virtual void drv_rect(double x1,double y1,double x2,double y2);
	virtual void drv_setfont(char *name,double fsize);
        virtual void drv_setgrey(int percentage);
        virtual void drv_setlinestyle(int count,double *len);
	virtual void drv_startpage();
	virtual void drv_text(char *string);
	virtual double drv_textlength(char *text);
	virtual double drv_textheight();
	virtual void drv_setlinewidth(double width);
	virtual void drv_setcolour(colour_e);
	//
	// This overides the inherited setup function.
	//
	virtual void PrinterdrawSetup();

public:
        Drawps();
        ~Drawps();
};

#define PSFLAGS_STROKEPENDING	0x1

#endif
