
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: drawtext.h,v 2.5 1995/06/07 20:54:12 davidsi Exp $
//
// $Log: drawtext.h,v $
// Revision 2.5  1995/06/07  20:54:12  davidsi
// - Added DTEXT_FLAG_NODRAWPAGE flag that prevents page throws at the
//   end of writing a text page.
//
// Revision 2.4  1995/05/18  21:06:11  davidsi
// - Added a new variant for the DrawtextOutput() function member, where
//   control is provided over exactly what gets drawn.
//
// Revision 2.3  1995/04/16  20:32:01  davidsi
// - Added a title mechanism.  The title can be set with DrawtextTitle()
//   and is held in the title member.
// - Added text_lines_per_page so that the total number of lines per page
//   that text will be draw to can be varied dynamically.
//
// Revision 2.2  1995/03/02  21:36:57  davidsi
// - Added DrawtextOutputline() function member.
//
// Revision 2.1  1995/02/05  18:01:53  davidsi
// - Added the drv_rect() member function.
//
// Revision 2.0  1995/01/31  22:24:33  davidsi
// - First version as drawtext.h (was drawps.h).
//
//
//
#ifndef __DRAWTEXT_H
#define __DRAWTEXT_H

#include "text.h"
#include "draw.h"

class Drawtext : public Text, public Draw {
private:
	class Draw *dptr;		// ptr to drawing class member

	//
	// These get set by DrawtextPrepare and used by DrawtextOutput
	//
        char **strings;
        int *maxwidths;
        double *positions;
	double charheight;
	
	int lines_per_page;		// total number of lines per page
					// including those used for page 
					// numbers (etc)
	int text_lines_per_page;	// number of lines containing
					// text - not inc. page number (etc)
	int totalpages;

	char *title;		// Title or NULL if no title
	
protected:
	//
	// These are the actual drawing functions
	//
	virtual void drv_startpage();
	virtual void drv_clearpage();
	virtual void drv_moveto(double x,double y);
	virtual void drv_drawto(double x,double y);
	virtual void drv_rect(double x,double y,double x2,double y2);
	virtual void drv_text(char *string);
	virtual void drv_setfont(char *name,double fsize);
	virtual void drv_drawpage();
	virtual double drv_textlength(char *text);
	virtual double drv_textheight();
        virtual void drv_setgrey(int percentage);
        virtual void drv_setlinestyle(int count,double *len);
	virtual void drv_setlinewidth(double width);
	virtual void drv_setcolour(colour_e);
public:

	int DrawtextNpages();
	void DrawtextPrepare();
	void DrawtextOutputline(double &sx,double &sy,int lineno,int fields);
	void DrawtextOutput(int pageno);
	void DrawtextOutput(int pageno,unsigned long flags);
	void DrawtextComplete();
	void DrawtextTitle(char *ititle);
	Drawtext(class Draw *dr);
	~Drawtext();
};

//
// Flags for the DrawtextOutput() function member. Select any combination
// an or them together.
//
#define DTEXT_FLAG_NONE		0x0	// do nothing
#define DTEXT_FLAG_PAGECLEAR	0x1	// clear page prior to drawing
#define DTEXT_FLAG_PAGENUMBER	0x2	// render a page number
#define DTEXT_FLAG_NODRAWPAGE	0x4	// don't issue a draw page at end

#endif
