
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: dsec.C,v 2.5 1995/06/17 20:42:13 davidsi Exp $
//
// $Log: dsec.C,v $
// Revision 2.5  1995/06/17  20:42:13  davidsi
// - BUG FIX.  Unit trust handling was incorrect and unit trusts would
//   not be correctly displayed.
//
// Revision 2.4  1995/06/07  20:43:31  davidsi
// - Added support for specifying unit trusts.
//
// Revision 2.3  1995/05/19  18:55:36  davidsi
// - Callbacks are now stored using opaque handles rather than pointers to
//   the actual nodes (see callback.C for the whole story).
//
// Revision 2.2  1995/04/25  21:11:01  davidsi
// - Added support for a busy cursor.
// - Updated to use ColourSetowner() so that the user's colour prefences for
//   the securities display can be pernamently stored in the preference
//   database.
//
// Revision 2.1  1995/04/16  20:28:39  davidsi
// - Added callback_sec member to allow security displays to be updated when
//   securities are added, removed or modified.  This gets registered and
//   removed as appropriate.
//
// Revision 2.0  1995/04/03  20:44:53  davidsi
// - Initial version.
//
//
//

//
// Include Files
//
#include <stdio.h>
#include <varargs.h>
#define _STDARG_H
#include <stdlib.h>
#include <string.h>
#include <dial.h>

#include "support.h"
#include "localmacros.h"

#include "dsec.h"

//
// Menu identifiers
//
enum menu_e {
	MENU_CLASS,
	MENU_PRINT,
	MENU_EXIT,
	MENU_COLOUR,
	MENU_SHARE,
	MENU_WARRANT,
	MENU_UNITTRUST,
	MENU_TRADED,
	MENU_NOTTRADED,
	MENU_AUTO,
	MENU_MANUAL,
	};

//
// Defines for the database identifiers
//
#define SECDISP_DB_CLASS "SECDISP"


//
// Securitiesdisplay::MenuCallback client_data onflag
//
// This is the call back used when the user selects something on the 
// menu.  The function overrides the standard function in the
// inheritated menu class.  
//
// INPUTS :	client_data	The id for the menu item thats been selected.
//		onflag		Whether the menu item is on/off.
// RETURNS: 	None
//
void
Securitiesdisplay::MenuCallback(unsigned long client_data, int onflag)
{
    	switch (client_data) {
	case MENU_EXIT:
		if (!colourdialog.SemaphoreIslocked())
			delete this;
		return;
	case MENU_COLOUR:
		if (colourdialog.SemaphoreClock()!= -1) {
			ColourSelect();
			colourdialog.SemaphoreUnlock();
			TextdialogRedraw();
		}
		break;
	case MENU_PRINT:
		TextdialogPrint();
		break;
        case MENU_SHARE:
		if (onflag)
			optionflags|=DSEC_OPTION_SHARE;
		else
			optionflags&=~DSEC_OPTION_SHARE;
		create_display();
		break;
	case MENU_WARRANT:
		if (onflag)
			optionflags|=DSEC_OPTION_WARRANT;
		else
			optionflags&=~DSEC_OPTION_WARRANT;
		create_display();
		break;
	case MENU_UNITTRUST:
		if (onflag)
			optionflags|=DSEC_OPTION_UNITTRUST;
		else
			optionflags&=~DSEC_OPTION_UNITTRUST;
		create_display();
		break;
	case MENU_TRADED:
		if (onflag)
			optionflags|=DSEC_OPTION_TRADED;
		else
			optionflags&=~DSEC_OPTION_TRADED;
		create_display();
		break;
	case MENU_NOTTRADED:
		if (onflag)
			optionflags|=DSEC_OPTION_NOTTRADED;
		else
			optionflags&=~DSEC_OPTION_NOTTRADED;
		create_display();
		break;
	case MENU_AUTO:
		if (onflag)
			optionflags|=DSEC_OPTION_AUTO;
		else
			optionflags&=~DSEC_OPTION_AUTO;
		create_display();
		break;
	case MENU_MANUAL:
		if (onflag)
			optionflags|=DSEC_OPTION_MANUAL;
		else
			optionflags&=~DSEC_OPTION_MANUAL;
		create_display();
		break;
	default:
		DialogPanic("Bad switch: Securitiesdisplay::MenuCallback");
	}
}

//
// Securitiesdisplay::redisplay this 
//
// This callback is registered to execute whenever securities are added(etc)
// It converts the first argument to a class ptr.
//
// INPUTS :	tmpthis		The this ptr defined when the callback
//				was registered.
// RETURNS:	None
void
Securitiesdisplay::redisplay(char *tmpthis,char *)
{
        class Securitiesdisplay *rthis=(class Securitiesdisplay *)tmpthis;
        rthis->create_display();
}


//
// Securitiesdisplay::create_display
//
// Generates text giving security information in alphanumeric order.
// This function is private to the class.
//
// INPUTS : None
// RETURNS: None
//
void
Securitiesdisplay::create_display()
{
	char str[256];			// build area
	class security *sec;
	int i;

	if (!setmenuflag) {
		MenuItemaction("Project","Close",MENU_EXIT);
		MenuItemaction("Project","Print",MENU_PRINT);
		MenuItemaction("Project","Colours",MENU_COLOUR);

		MenuItemnofn("Type","Share",
				MENU_SHARE,MENUSTATE_ON);
		MenuItemnofn("Type","Warrant",
				MENU_WARRANT,MENUSTATE_ON);
		MenuItemnofn("Type","Unit Trust",
				MENU_UNITTRUST,MENUSTATE_ON);

		MenuItemnofn("Requirements","Traded",
				MENU_TRADED,MENUSTATE_ON);
		MenuItemnofn("Requirements","Not Traded",
				MENU_NOTTRADED,MENUSTATE_ON);
		MenuItemnofn("Requirements","Automatic",
				MENU_AUTO,MENUSTATE_ON);
		MenuItemnofn("Requirements","Manual",
				MENU_MANUAL,MENUSTATE_ON);
		for(i=1;i<8;i++) {
			class attr *attr;
			attr=new class attr(FIELDATTR_LEFT,FOREGROUND);
			TextSetdefaultattr(i,attr);
		}
		TextdialogSettitle("Summary of all securities");
		ColourSetowner(SECDISP_DB_CLASS);
		setmenuflag=1;
	}

	TextClear();

	TextAdd("Security name\tType\tState\tAuto\tManual");

	for(sec=main_sec_list->head();!main_sec_list->eol(sec);sec=sec->next()){
		unsigned long flags;
		int count;
		sectype_e typef;
		char *name;
		char *typename=NULL;

		sec->SecurityExtract(&name,&flags,NULL,&count,&typef);

		//
		// Check flags are a match using an OR
		//
		if ( !( ( (optionflags&DSEC_OPTION_TRADED) 
				&& (flags&SEC_FLAGS_TRADED) ) 
			||( (optionflags&DSEC_OPTION_NOTTRADED) 
				&& (flags&SEC_FLAGS_NOTTRADED) )
			||( (optionflags&DSEC_OPTION_AUTO) 
				&& (flags&SEC_FLAGS_AUTO) )
			||( (optionflags&DSEC_OPTION_MANUAL) 
				&& (flags&SEC_FLAGS_MANUAL) ) ) )
			continue;
#ifdef NEVER
		//
		// Check flags are a match using an AND
		//
		if ( (optionflags&DSEC_OPTION_TRADED) 
				&& !(flags&SEC_FLAGS_TRADED) ) 
			continue;
		if ( (optionflags&DSEC_OPTION_NOTTRADED) 
				&& !(flags&SEC_FLAGS_NOTTRADED) )
			continue;
		if ( (optionflags&DSEC_OPTION_AUTO) 
				&& !(flags&SEC_FLAGS_AUTO) )
			continue;
		if ( (optionflags&DSEC_OPTION_MANUAL) 
				&& !(flags&SEC_FLAGS_MANUAL) )
			continue;
#endif

		//
		// If not looking for both shares and warrants, check
		// more carefully (this saves a strcmp() when we don't
		// care what type of security we are searching for.
		//
		if ( (optionflags&DSEC_OPTION_SHARE|DSEC_OPTION_WARRANT|
					DSEC_OPTION_UNITTRUST)!=
			DSEC_OPTION_SHARE|DSEC_OPTION_WARRANT|
					DSEC_OPTION_UNITTRUST) {
			typename=sec->typename();
			if ( (optionflags&DSEC_OPTION_SHARE) &&
			     !alpha_strcmp("share",typename) )
				;
			else if ( (optionflags&DSEC_OPTION_WARRANT) &&
			     !alpha_strcmp("warrant",typename) )
				;
			else if ( (optionflags&DSEC_OPTION_UNITTRUST) &&
			     !alpha_strcmp("Unit Trust",typename) )
				;
			else
				continue;
		}
		strncpy(str,name,15);
		str[15]='\0';
		strcat(str,"\t");

		if (typename==NULL)
			typename=sec->typename();
		strcat(str,typename);
		strcat(str,"\t");

		if (flags & SEC_FLAGS_TRADED)
			strcat(str,"Traded");
		else if (flags&SEC_FLAGS_NOTTRADED)
			strcat(str,"Not traded");
		else 
			strcat(str,"Unknown");
		strcat(str,"\t");
	
		if (flags & SEC_FLAGS_AUTO)
			strcat(str,"Yes");
		else 
			strcat(str,"No");
		strcat(str,"\t");

		if (flags & SEC_FLAGS_MANUAL)
			strcat(str,"Yes");
		else 
			strcat(str,"No");
		TextAdd(str);
		DialogCursor(DCURSOR_BUSY);
    	}
	TextComplete();
	TextdialogOutput();

	DialogCursor(DCURSOR_NORMAL);
}



//
// Securitiesdisplay::Securitiesdisplay
//
// Constructor for Securitiesdisplay class.
// 
// Creates & manages the window which displays the summary status for all
// (or selected) securities.
//
// INPUTS :	None
// RETURNS:	None
//
Securitiesdisplay::Securitiesdisplay()
{
	setmenuflag=0;			// menu not yet attached

	optionflags=DSEC_OPTION_SHARE|DSEC_OPTION_WARRANT|
			SEC_TYPE_UNITTRUST|
			DSEC_OPTION_TRADED|DSEC_OPTION_NOTTRADED|
			DSEC_OPTION_AUTO|DSEC_OPTION_MANUAL;

	//
	// Set up call backs for security changes
	//
	callback_sec=main_sec_list->changed_vect.CallbackAdd(
			&Securitiesdisplay::redisplay,
			(char *)this);
	
	create_display();
}

//
// Securitiesdisplay::~Securitiesdisplay
// 
// Destructor for Securitiesdisplay class.
//
Securitiesdisplay::~Securitiesdisplay()
{
	//
	// Remove the callback node, from the list of callbacks
	// that are made when a security is added/modified/deleted.
	//
	main_sec_list->changed_vect.CallbackRemove(callback_sec);
}



//
//=============================================================================
//
// The C "Glue"
//
//=============================================================================
//

//
// display_secs d a args
//
// C wrapper function that creates a securities display
//
// INPUTS :	dialog,a,args	All place holders to satify type checking
// RETURNS:	None
//
static void
display_secs(local_va_dcl)
{
	va_alist=va_alist;	// stop compiler warnings
	//
	// Simply display the securities by creating a Securitiesdisplay
	// member
	//
        Securitiesdisplay *newsec=new Securitiesdisplay;
}


//
// dsec_init menu_bar
//
// Initialises the display securities module:
//
// INPUTS :	Menu bar where option to display securities is to be inserted
// RETURNS:	None
//
void
dsec_init(menu_bar *mb)
{
	MENU("Securities","Display Summary",display_secs,0);
}
