
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: dslist.C,v 1.3 1995/06/24 15:01:04 davidsi Exp $
//
// $Log: dslist.C,v $
// Revision 1.3  1995/06/24  15:01:04  davidsi
// - Added GNU copyright message.
//
// Revision 1.2  1994/10/17  20:43:43  davidsi
// - Added dual_remove() to allow entries to be taken off the second list.
//   In order to implement this, the second (dual) list is also a back/forw
//   maintained list.
//
// Revision 1.1  1994/06/29  20:14:58  davidsi
// Initial revision
//
//
//



//
// Include Files
//
#include "dslist.h"


//
// dual_add dsptr
//
// Adds a node to the secondary chain
//
// INPUTS : 	dsptr	The node that is to be added to the second chain
//			of nodes.
// RETURNS: 	None
void
dslist_head::dual_add(class dslist_node *dsptr)
{
	//
	// Note that the pointer in the head structure does not
	// sit on the linked list.
	//
	if (first==NULL) {
		first=dsptr;
		dsptr->next_node=dsptr;
		dsptr->last_node=dsptr;
	} else {
		dsptr->next_node=first;
		dsptr->last_node=first->last_node;

		first->last_node->next_node=dsptr;
		first->last_node=dsptr;

		first=dsptr;
	}
}

//
// dual_head
//
// Given the start of the secondary chain of nodes, returns the
// first node on that chain.
//
// INPUTS : 	None
// RETURNS: 	The first node on the secondary list
//
dslist_node *
dslist_head::dual_head()
{
	return(first);
}


//
// dual_next
//
// Given a node that is on a dual stored list, this function will
// return the next node that is on the secondary chain of nodes
//
// INPUTS : 	None
// RETURNS: 	The next node on the secondary list
//
dslist_node *
dslist_node::dual_next()
{
	return(next_node);
}

//
// dual_remove
//
// Given a node that is on a dual stored list, this function will
// remove that node.
//
// INPUTS :	None
// RETURNS:	None. The structure is removed from the dual list.
//
void
dslist_node::dual_remove()
{
        last_node->next_node=next_node;
        next_node->last_node=last_node;
}


//
// dslist_head
//
// Constructor function for heads of double stored lists
//
// INPUTS :	None
// RETURNS:	None
//
dslist_head::dslist_head()
{
	first=NULL;
}

//
// ~dslist_head
// 
// Denstructor function for heads of double stored lists
//
// INPUTS :	None
// RETURNS:	None
//
dslist_head::~dslist_head()
{
}
