
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: dtrans.h,v 2.4 1995/05/18 21:01:32 davidsi Exp $
//
// $Log: dtrans.h,v $
// Revision 2.4  1995/05/18  21:01:32  davidsi
// - Callbacks are now stored using opaque handles rather than pointers to
//   the actual nodes (see callback.C for the whole story).
//
// Revision 2.3  1995/04/16  20:34:32  davidsi
// - Added list_trans_dates() function member for date ordered output.
// - Added list_trans_value() function member for value ordered output.
// - Added callback_sec member to allow transaction displays to be updated
//   when securities are added, removed or modified.
//
// Revision 2.2  1995/03/02  20:47:17  davidsi
// - Added a semaphore for controlling access to the colour dialog.
//
// Revision 2.1  1995/01/31  22:09:43  davidsi
// - Revised to use the textdialog class.  This results in much of
//   the required functionality now being inherited.
//
// Revision 2.0  1995/01/19  21:48:00  davidsi
// - Removed the time members of the trans_list class.  Dates are now
//   managed using the date class.
// - Removed trans2date() member.
// - Added transactions_display::trans_change for transaction change
//   callbacks.
//
// Revision 1.1  1994/07/03  15:55:37  davidsi
// Initial revision
//
//
#ifndef __DTRANS_H
#define __DTRANS_H

#include <stdio.h>
#include <string.h>

#include "trans.h"
#include "date.h"
#include "semaphore.h"
#include "callback.h"
#include "textdialog.h"

class transactions_display : public Textdialog {
	int trans_displayed;	// transactions being displayed, this uses 
				// the TRANS_XX values ored to as single
				// value
	int ordering;		// how the display is ordered
	class date start_date;	// Start date for transactions
	class date end_date;	// End date for transactions
	char setmenuflag;	// has menu/title been set
	Semaphore colourdialog;	// colour dialog display sem.

	Callbackid callback;	// callback 
	Callbackid callback_sec;// callback for security changes


private:
	//
	// This overrides the menu function member in the Menu class
	//
	virtual void MenuCallback(unsigned long code, int onflag);
	void static trans_change(char *realthis,char *);

	void create_header( int mode );
	int append(class transaction *trans,int mode, char *secname);
	void list_trans_shares();
	void list_trans_dates();
	void list_trans_value();
	void create_display();
	void reselect_types(int,int);

public:
	transactions_display();
	~transactions_display();
	void menu(unsigned long code);

};

#endif
