
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: dval.h,v 2.4 1995/05/18 20:45:29 davidsi Exp $
//
// $Log: dval.h,v $
// Revision 2.4  1995/05/18  20:45:29  davidsi
// - Callbacks are now stored using opaque handles rather than pointers to
//   the actual nodes (see callback.C for the whole story).
//
// Revision 2.3  1995/04/16  20:31:22  davidsi
// - Added callback_sec member to allow valuation displays to be updated when
//   securities are added, removed or modified.
//
// Revision 2.2  1995/03/02  21:10:58  davidsi
// - Added colourdialog semaphore to allow synchronized access to the
//   colour dialog.
//
// Revision 2.1  1995/01/31  21:37:08  davidsi
// - Revised to use the textdialog class.  This results in much of
//   the required functionality now being inherited.
//
// Revision 2.0  1995/01/19  21:47:21  davidsi
// - Removed the time members of the valuation_display class.  Dates are now
//   managed using the date class.
// - Added redisplay() function which is now used by price and transaction
//   change callbacks.
//
// Revision 1.3  1994/12/11  20:44:25  davidsi
// - The menu() function is now declared a static member of the
//   valuation_display class, so it can be used (directly) in a callback.
//
// Revision 1.2  1994/10/17  20:53:22  davidsi
// - Added flags field so as to support current and historic valuations.
//
// Revision 1.1  1994/07/03  16:23:23  davidsi
// Initial revision
//
//

#ifndef __DVAL_H
#define __DVAL_H

#include <stdio.h>
#include <string.h>

#include "textdialog.h"
#include "date.h"
#include "semaphore.h"
#include "callback.h"


class valuation_display : public Textdialog {
	class date start_date;			// Start date for the valuation
	class date end_date;			// End date for the valuation
	unsigned long flags;			// see DVAL_FLAGS_XXX
	Semaphore colourdialog;			// colour dialog display sem.

	Callbackid callback_trans;		// callback for transactions
	Callbackid callback_price;		// callback for price changes
	Callbackid callback_div;		// callback for dividend changes
	Callbackid callback_sec;		// callback for security changes

	char setmenuflag;			// has menu/title been set

private:
	virtual void MenuCallback(unsigned long code, int onflag);
	static void redisplay(char *realthis,char *);
	void list_valuation();
public:
	valuation_display();
	~valuation_display();
};

#define DVAL_FLAGS_EVAL		1	// current valuation
#define DVAL_FLAGS_HISTORIC	2	// historic valuation

#endif
