/* code to manage the stuff on the "earthmap" menu.
 */

#include <stdio.h>
#include <ctype.h>
#include <math.h>
#if defined(__STDC__)
#include <stdlib.h>
#endif
#include <X11/Xlib.h>
#include <Xm/Xm.h>
#include <Xm/Form.h>
#include <Xm/Frame.h>
#include <Xm/Label.h>
#include <Xm/PushB.h>
#include <Xm/RowColumn.h>
#include <Xm/DrawingA.h>
#include <Xm/Separator.h>
#include <Xm/ToggleB.h>
#include "astro.h"
#include "circum.h"
#include "map.h"

#if defined(__STDC__) || defined(__cplusplus)
#define P_(s) s
#else
#define P_(s) ()
#endif

extern double mjd_day P_((double jd));
extern double mjd_hr P_((double jd));
extern int any_ison P_((void));
extern int obj_cir P_((Now *np, Obj *op));
extern void f_gangle P_((Widget w, double a));
extern void get_something P_((Widget w, char *resource, char *value));
extern void hlp_dialog P_((char *tag, char *deflt[], int ndeflt));
extern void precess P_((double mjd1, double mjd2, double *ra, double *dec));
extern void range P_((double *v, double r));
extern void register_selection P_((char *name));
extern void timestamp P_((Now *np, Widget w));
extern void utc_gst P_((double Mjd, double utc, double *gst));

void e_manage P_((void));
void e_update P_((Now *np, int force));
int e_ison P_((void));
void e_selection_mode P_((int whether));
void e_cursor P_((Cursor c));
static void e_create_form P_((void));
static void e_ll_cb P_((Widget w, XtPointer client, XtPointer call));
static void e_close_cb P_((Widget w, XtPointer client, XtPointer call));
static void e_help_cb P_((Widget w, XtPointer client, XtPointer call));
static void e_exp_cb P_((Widget w, XtPointer client, XtPointer call));
static void e_set_buttons P_((int whether));
static void e_subsolar P_((Now *np, double *latp, double *longp));
static void e_map P_((Now *np, double slat, double slong));

#undef P_

/* these are from earthmap.c */
extern MRegion ereg[];
extern int nereg;

extern Widget toplevel_w;
extern Now *mm_get_now();
#define	XtD	XtDisplay(toplevel_w)

static int e_selecting;	/* set while our fields are being selected */

static Widget eform_w;	/* main form dialog */
static Widget e_long_w;	/* solar longitude numberic display widget */
static Widget e_lat_w;	/* solar latitude numberic display widget */
static Widget e_da_w;	/* map DrawingArea widget */
static Widget e_dt_w;	/* data/time stamp label widget */

#define	DSIN(x)		sin(degrad(x))
#define	DCOS(x)		cos(degrad(x))

void
e_manage()
{
	if (!eform_w)
	    e_create_form();

	if (XtIsManaged(eform_w))
	    XtUnmanageChild(eform_w);
	else {
	    XtManageChild(eform_w);
	    e_set_buttons(e_selecting);
	}
}

/* display the map and lat/long info for circumstances described by *np..
 * don't bother with the computation of lat/long unless we are managed or
 *   someone cares.
 * don't bother with the map unless we are managed.
 */
void
e_update(np, force)
Now *np;
int force;
{
	double slat, slong;

	if (!eform_w)
	    return;
	if (!XtIsManaged(eform_w) && !any_ison() && !force)
	    return;

	e_subsolar (np, &slat, &slong);

	if (XtIsManaged(eform_w))
	    e_map (np, slat, slong);

	f_gangle (e_lat_w, slat);

	/* we want to display +W in range of -PI..+PI */
	slong = -slong;
	range (&slong, 2*PI);
	if (slong >= PI)
	    slong -= 2*PI;
	f_gangle (e_long_w, slong);
}

e_ison()
{
	return (eform_w && XtIsManaged(eform_w));
}

/* called by other menus as they want to hear from our buttons or not.
 * the ons and offs stack.
 */
void
e_selection_mode (whether)
int whether;
{
	e_selecting += whether ? 1 : -1;

	if (e_ison())
	    if (whether && e_selecting == 1	   /* first one to want on */
		|| !whether && e_selecting == 0 /* first one to want off */)
		e_set_buttons (whether);
}

void
e_cursor(c)
Cursor c;
{
	Window win;

	if (eform_w && (win = XtWindow(eform_w))) {
	    Display *dsp = XtDisplay(eform_w);
	    if (c)
		XDefineCursor(dsp, win, c);
	    else
		XUndefineCursor(dsp, win);
	}
}

static void
e_create_form()
{
	static struct {
	    char *name;
	    void (*cb)();
	} ctls[] = {
	    {"Close", e_close_cb},
	    {"Help", e_help_cb}
	};
	Widget w;
	Widget fr_w;
	Widget f_w;
	Widget sep_w;
	XmString str;
	Arg args[20];
	int n;
	int i;

	/* create form */
	n = 0;
	XtSetArg (args[n], XmNautoUnmanage, False); n++;
	XtSetArg (args[n], XmNdefaultPosition, False); n++;
	XtSetArg (args[n], XmNresizePolicy, XmRESIZE_NONE); n++;
	XtSetArg (args[n], XmNhorizontalSpacing, 4); n++;
	XtSetArg (args[n], XmNverticalSpacing, 4); n++;
	eform_w = XmCreateFormDialog (toplevel_w, "Earth", args,n);

	/* set some stuff in the parent DialogShell.
	 * setting XmNdialogTitle in the Form didn't work..
	 */
	n = 0;
	XtSetArg (args[n], XmNtitle, "xephem Earth subsolar view"); n++;
	XtSetValues (XtParent(eform_w), args, n);

	/* make the bottom control panel form */

	n = 0;
	XtSetArg (args[n], XmNbottomAttachment, XmATTACH_FORM); n++;
	XtSetArg (args[n], XmNleftAttachment, XmATTACH_FORM); n++;
	XtSetArg (args[n], XmNrightAttachment, XmATTACH_FORM); n++;
	XtSetArg (args[n], XmNfractionBase, 9); n++;
	XtSetArg (args[n], XmNverticalSpacing, 5); n++;
	f_w = XmCreateForm (eform_w, "CtlF", args, n);
	XtManageChild (f_w);

	    for (i = 0; i < XtNumber(ctls); i++) {
		n = 0;
		XtSetArg (args[n], XmNtopAttachment, XmATTACH_FORM); n++;
		XtSetArg (args[n], XmNbottomAttachment, XmATTACH_FORM); n++;
		XtSetArg (args[n], XmNleftAttachment, XmATTACH_POSITION); n++;
		XtSetArg (args[n], XmNleftPosition, 1+i*4); n++;
		XtSetArg (args[n], XmNrightAttachment, XmATTACH_POSITION); n++;
		XtSetArg (args[n], XmNrightPosition, 4+i*4); n++;
		w = XmCreatePushButton (f_w, ctls[i].name, args, n);
		XtManageChild (w);
		XtAddCallback (w, XmNactivateCallback, ctls[i].cb, NULL);
	    }

	/* make a separator */

	n = 0;
	XtSetArg (args[n], XmNbottomAttachment, XmATTACH_WIDGET); n++;
	XtSetArg (args[n], XmNbottomWidget, f_w); n++;
	XtSetArg (args[n], XmNleftAttachment, XmATTACH_FORM); n++;
	XtSetArg (args[n], XmNrightAttachment, XmATTACH_FORM); n++;
	sep_w = XmCreateSeparator (eform_w, "Sep", args, n);
	XtManageChild(sep_w);

	/* make the "latitude" label and push button pair */

	str = XmStringCreateLtoR("Latitude (degs, +N): ",
						    XmSTRING_DEFAULT_CHARSET);
	n = 0;
	XtSetArg (args[n], XmNbottomAttachment, XmATTACH_WIDGET); n++;
	XtSetArg (args[n], XmNbottomWidget, sep_w); n++;
	XtSetArg (args[n], XmNleftAttachment, XmATTACH_FORM); n++;
	XtSetArg (args[n], XmNalignment, XmALIGNMENT_BEGINNING); n++;
	XtSetArg (args[n], XmNlabelString, str); n++;
	w = XmCreateLabel (eform_w, "LatLbl", args, n);
	XtManageChild (w);
	XmStringFree(str);

	str = XmStringCreateLtoR("-88:88:88",XmSTRING_DEFAULT_CHARSET); /*size*/
	n = 0;
	XtSetArg (args[n], XmNbottomAttachment, XmATTACH_WIDGET); n++;
	XtSetArg (args[n], XmNbottomWidget, sep_w); n++;
	XtSetArg (args[n], XmNrightAttachment, XmATTACH_FORM); n++;
	XtSetArg (args[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
	XtSetArg (args[n], XmNleftWidget, w); n++;
	XtSetArg (args[n], XmNalignment, XmALIGNMENT_END); n++;
	XtSetArg (args[n], XmNlabelString, str); n++;
	XtSetArg (args[n], XmNuserData, "Earth.Lat"); n++;
	e_lat_w = XmCreatePushButton (eform_w, "LatVal", args, n);
	XtAddCallback (e_lat_w, XmNactivateCallback, e_ll_cb, 0);
	XtManageChild (e_lat_w);
	XmStringFree(str);

	/* make the "Longitude" label and push button pair */

	str = XmStringCreateLtoR("Longitude (degs, +W): ",
						    XmSTRING_DEFAULT_CHARSET);
	n = 0;
	XtSetArg (args[n], XmNbottomAttachment, XmATTACH_WIDGET); n++;
	XtSetArg (args[n], XmNbottomWidget, e_lat_w); n++;
	XtSetArg (args[n], XmNleftAttachment, XmATTACH_FORM); n++;
	XtSetArg (args[n], XmNalignment, XmALIGNMENT_BEGINNING); n++;
	XtSetArg (args[n], XmNlabelString, str); n++;
	w = XmCreateLabel (eform_w, "LongLbl", args, n);
	XtManageChild (w);
	XmStringFree(str);

	str = XmStringCreateLtoR("-188:88:88",XmSTRING_DEFAULT_CHARSET);/*size*/
	n = 0;
	XtSetArg (args[n], XmNbottomAttachment, XmATTACH_WIDGET); n++;
	XtSetArg (args[n], XmNbottomWidget, e_lat_w); n++;
	XtSetArg (args[n], XmNrightAttachment, XmATTACH_FORM); n++;
	XtSetArg (args[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
	XtSetArg (args[n], XmNleftWidget, w); n++;
	XtSetArg (args[n], XmNalignment, XmALIGNMENT_END); n++;
	XtSetArg (args[n], XmNlabelString, str); n++;
	XtSetArg (args[n], XmNuserData, "Earth.Long"); n++;
	e_long_w = XmCreatePushButton (eform_w, "LongVal", args, n);
	XtAddCallback (e_long_w, XmNactivateCallback, e_ll_cb, 0);
	XtManageChild (e_long_w);
	XmStringFree(str);

	/* make the date/time indicator label */

	n = 0;
	XtSetArg (args[n], XmNbottomAttachment, XmATTACH_WIDGET); n++;
	XtSetArg (args[n], XmNbottomWidget, e_long_w); n++;
	XtSetArg (args[n], XmNleftAttachment, XmATTACH_FORM); n++;
	XtSetArg (args[n], XmNrightAttachment, XmATTACH_FORM); n++;
	XtSetArg (args[n], XmNalignment, XmALIGNMENT_CENTER); n++;
	e_dt_w = XmCreateLabel (eform_w, "DTstamp", args, n);
	XtManageChild (e_dt_w);

	/* make a drawing area in a frame on top */

	n = 0;
	XtSetArg (args[n], XmNtopAttachment, XmATTACH_FORM); n++;
	XtSetArg (args[n], XmNbottomAttachment, XmATTACH_WIDGET); n++;
	XtSetArg (args[n], XmNbottomWidget, e_dt_w); n++;
	XtSetArg (args[n], XmNleftAttachment, XmATTACH_FORM); n++;
	XtSetArg (args[n], XmNrightAttachment, XmATTACH_FORM); n++;
	XtSetArg (args[n], XmNshadowType, XmSHADOW_ETCHED_OUT); n++;
	fr_w = XmCreateFrame (eform_w, "EarthFrame", args, n);
	XtManageChild (fr_w);

	    n = 0;
	    e_da_w = XmCreateDrawingArea (fr_w, "Map", args, n);
	    XtAddCallback (e_da_w, XmNexposeCallback, e_exp_cb, NULL);
	    XtManageChild (e_da_w);
}

/* callback for when the lat/long buttons are activated. */
/* ARGSUSED */
static void
e_ll_cb (w, client, call)
Widget w;
XtPointer client;
XtPointer call;
{
	if (e_selecting) {
	    char *userD;
	    get_something (w, XmNuserData, (char *)&userD);
	    register_selection (userD);
	}
}

/* callback for when the Close button is activated. */
/* ARGSUSED */
static void
e_close_cb (w, client, call)
Widget w;
XtPointer client;
XtPointer call;
{
	XtUnmanageChild (eform_w);
}

/* callback from the Help button
 */
/* ARGSUSED */
static void
e_help_cb (w, client, call)
Widget w;
XtPointer client;
XtPointer call;
{
	static char *msg[] = {
"This is a simple schematic depiction of the Earth surface at the given time.",
"The view is from directly over the spot where the Sun is currently over head.",
};

	hlp_dialog ("Earth", msg, XtNumber(msg));
}

/* called whenever the earth drawing area gets an expose.
 */
/* ARGSUSED */
static void
e_exp_cb (w, client, call)
Widget w;
XtPointer client;
XtPointer call;
{
	XmDrawingAreaCallbackStruct *c = (XmDrawingAreaCallbackStruct *)call;

	/* filter out a few oddball cases */
	switch (c->reason) {
	case XmCR_EXPOSE: {
	    /* turn off gravity so we get expose events for either shrink or
	     * expand.
	     */
	    static before;
	    XExposeEvent *e = &c->event->xexpose;

	    if (!before) {
		XSetWindowAttributes swa;
		swa.bit_gravity = ForgetGravity;
		XChangeWindowAttributes (e->display, e->window, 
							    CWBitGravity, &swa);
		before = 1;
	    }
	    /* wait for the last in the series */
	    if (e->count != 0)
		return;
	    break;
	    }
	default:
	    printf ("Unexpected eform_w event. type=%d\n", c->reason);
	    exit(1);
	}

	e_update (mm_get_now(), 1);
}

/* go through all the buttons and set whether they
 * should appear to look like buttons or just flat labels.
 */
static void
e_set_buttons (whether)
int whether;	/* whether setting up for plotting or for not plotting */
{
	static Arg look_like_button[] = {
	    {XmNtopShadowColor, (XtArgVal) 0},
	    {XmNbottomShadowColor, (XtArgVal) 0},
	    {XmNfillOnArm, (XtArgVal) True},
	};
	static Arg look_like_label[] = {
	    {XmNtopShadowColor, (XtArgVal) 0},
	    {XmNbottomShadowColor, (XtArgVal) 0},
	    {XmNfillOnArm, (XtArgVal) False},
	};
	static int called;
	Arg *ap;
	int na;

	if (!called) {
	    /* get baseline label and shadow appearances.
	     */
	    Pixel topshad, botshad, bgcol;
	    Arg args[20];
	    int n;

	    n = 0;
	    XtSetArg (args[n], XmNtopShadowColor, &topshad); n++;
	    XtSetArg (args[n], XmNbottomShadowColor, &botshad); n++;
	    XtSetArg (args[n], XmNbackground, &bgcol); n++;
	    XtGetValues (e_lat_w, args, n);

	    look_like_button[0].value = topshad;
	    look_like_button[1].value = botshad;
	    look_like_label[0].value = bgcol;
	    look_like_label[1].value = bgcol;

	    called = 1;
	}

	if (whether) {
	    ap = look_like_button;
	    na = XtNumber(look_like_button);
	} else {
	    ap = look_like_label;
	    na = XtNumber(look_like_label);
	}

	XtSetValues (e_lat_w, ap, na);
	XtSetValues (e_long_w, ap, na);
}

/* return the long and lat of the subsolar point (where the sun is directly
 * overhead) at np.
 * both are in rads, lat is +N, long is +E.
 */
static void
e_subsolar(np, latp, longp)
Now *np;
double *latp, *longp;
{
	Obj o;
	double gst;

	o.type = PLANET;
	o.pl.code = SUN;

	if (obj_cir (np, &o) < 0) {
	    printf ("e_subsolar(): can't get sun loc\n");
	    exit (1);
	}
	if (epoch != EOD) {
	    double tra = o.s_ra, tdec = o.s_dec;
	    precess (epoch, mjd, &tra, &tdec);
	    o.s_ra = tra;
	    o.s_dec = tdec;
	}

	*latp = o.s_dec;

	utc_gst (mjd_day(mjd), mjd_hr(mjd), &gst);
	*longp =  o.s_ra - hrrad(gst);	/* remember: we want +E */
}

/* draw map centered at slat/slong.
 * each in rads, slat is +N, slong is +E.
 */
static void
e_map (np, slat, slong)
Now *np;
double slat, slong;
{
#define	CRAD		20		/* cross radius, pixels */
#define	COAST_LW	2		/* line_width for drawing coast lines */
#define	CROSS_LW	0		/* line_width for drawing the cross */
#define	LINE_ST		LineSolid
#define	CAP_ST		CapRound
#define	JOIN_ST		JoinRound
	static unsigned long last_w, last_h;
	static Pixmap e_pm;
	static GC e_gc;
	static Pixel e_fg, e_bg;
	Display *dsp = XtDisplay(e_da_w);
	Window win = XtWindow(e_da_w);
	int x, y;
	unsigned w, h, b, d;
	Window root;
	int wb, hb, r;
	MRegion *rp;
	MCoord *cp;

	/* make the GC if this is our first time */
	if (!e_gc) {
	    XGCValues gcv;
	    unsigned long gcm;

	    gcm = 0;
	    e_gc = XCreateGC (dsp, win, gcm, &gcv);

	    get_something (e_da_w, XmNforeground, (char *)&e_fg);
	    get_something (e_da_w, XmNbackground, (char *)&e_bg);
	}

	/* set up width and height borders and the radius so the
	 * planetary sphere is a circle centered within the window.
	 */
	XGetGeometry (dsp, win, &root, &x, &y, &w, &h, &b, &d);
	if (w > h) {
	    wb = (w - h)/2;
	    hb = 0;
	    r = h/2;
	} else {
	    wb = 0;
	    hb = (h - w)/2;
	    r = w/2;
	}

	/* make the pixmap if this is our first time or the size has changed */
	if (!e_pm || w != last_w || h != last_h) {
	    last_w = w;
	    last_h = h;
	    if (e_pm)
		XFreePixmap (dsp, e_pm);
	    e_pm = XCreatePixmap (dsp, win, w, h, d);
	}

	/* clear the scene, draw filled circle in foreground then fill in
	 * each region in background.
	 */
	XSetLineAttributes (dsp, e_gc, COAST_LW, LINE_ST, CAP_ST, JOIN_ST);
	XSetForeground (dsp, e_gc, e_bg);
	XFillRectangle (dsp, e_pm, e_gc, 0, 0, w, h);
	XSetForeground (dsp, e_gc, e_fg);
	XFillArc (dsp, e_pm, e_gc, wb, hb, 2*r, 2*r, 0, 360*64);
	XSetForeground (dsp, e_gc, e_bg);
	for (rp = ereg; rp < ereg + nereg; rp++) {
	    XPoint xp[1024];
	    int nxp = 0;
	    MCoord *lastcp = rp->mcp + rp->nmcp - 1;
	    for (cp = rp->mcp; cp <= lastcp; cp++) {
		/* rotate to (lat,long)=(0,0) then project onto x/y plane */
		double longp = degrad(cp->lg) - slong;
		double z = r*cos(longp);
		if (z > 0 && nxp < XtNumber(xp)) {
		    double latp = degrad(cp->lt) - slat;
		    double lx = r*sin(longp)*cos(latp);
		    double ly = -r*sin(latp);
		    xp[nxp].x = (int)floor(lx+0.5) + wb + r;
		    xp[nxp].y = (int)floor(ly+0.5) + hb + r;
		    nxp++;
		}
		if (z < 0 || cp == lastcp) {
		    if (nxp > 1)
			XDrawLines (dsp, e_pm, e_gc, xp, nxp, CoordModeOrigin);
		    nxp = 0;
		}
	    }
	}

	/* draw a cross-hair in the center to mark the subsolar point */
	XSetLineAttributes (dsp, e_gc, CROSS_LW, LINE_ST, CAP_ST, JOIN_ST);
	XDrawLine (dsp, e_pm, e_gc, wb+r-CRAD, hb+r-CRAD, wb+r+CRAD, hb+r+CRAD);
	XDrawLine (dsp, e_pm, e_gc, wb+r-CRAD, hb+r+CRAD, wb+r+CRAD, hb+r-CRAD);

	XSetFunction (dsp, e_gc, GXcopy);
	XCopyArea (dsp, e_pm, win, e_gc, 0, 0, w, h, 0, 0);

	timestamp (np, e_dt_w);
}
