
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: error.C,v 2.4 1995/06/17 20:33:57 davidsi Exp $
//
// $Log: error.C,v $
// Revision 2.4  1995/06/17  20:33:57  davidsi
// - Added index_too_large error message.
//
// Revision 2.3  1995/06/07  20:42:40  davidsi
// - Added knowledge of the value class.
//
// Revision 2.2  1995/06/02  20:04:00  davidsi
// - Added an errstream operator that can handle node pointer types.  The idea
//   being to allow line numbers (etc) to be dumped out with errors.
// - Added not_string, not_real and incorrect_num_arguments.
// - Added file_not_open.
// - Added field_lookup_failed.
//
// Revision 2.1  1995/05/18  20:43:38  davidsi
// - Added not_a_valid_security error message.
//
// Revision 2.0  1995/01/19  21:46:24  davidsi
// - Initial version.
//
//
//

#include "error.h"
#include "node.h"
#include "value.h"

errstream cerror = stderr;
diagstream cinfo = stdout;

struct errno_2_errmsg {
	errmsg_e no;
	char *msg;
};

static errno_2_errmsg map[]={
	symbol_not_found,"Symbol not found",
	function_not_found,"Function not found",
	type_not_defined,"Type not defined",
	bad_operation,"Operation not valid on type",
	type_mismatch,"Type mismatch",
	not_evaluated,"Expression could not be evaluated",
	no_expression,"No expression to evaluate",
	not_boolean,"Expression did not yield a boolean result",
	not_integer,"Expression did not yield a integer result",
	not_string,"Expression did not yield a string result",
	not_real,"Expression did not yield a real result",
	empty_body,"While loop with no body",
	cannot_deref,"Cannot derefence",
	cannot_execute,"Cannot execute instruction",
	bad_arg_decl,"Bad argument declaration - cannot resolve",
	bad_var_decl,"Bad variable declaration - cannot resolve",
	bad_type,"Cannot determine type of expression",
	internal_address_not_set,"Attempt to set value when no address assigned",
	not_a_valid_security,"Not a valid security identifier",
	incorrect_num_arguments,"Incorrect number of arguments",
	file_not_open,"Attempt to access file that is not open",
	field_lookup_failed,"Field lookup failed",
        index_too_large,"Array index too large",
	no_message,NULL,
};



errstream &
errstream::operator << (char *s) {
        fprintf(fp,"%s",s);
        return *this;
}

errstream &
errstream::operator << (int s) {
        fprintf(fp,"%d",s);
        return *this;
}

errstream &
errstream::operator << (double d) {
        fprintf(fp,"%6.2f",d);
        return *this;
}

errstream &
errstream::operator << (long s) {
        fprintf(fp,"%d",s);
        return *this;
}

errstream &
errstream::operator << (errmsg_e s) {
	struct errno_2_errmsg *ptr=map;
	
	while (ptr->msg!=NULL && ptr->no!=s)
		ptr++;

	if (ptr->msg==NULL) 
		fprintf(fp,"Unknown error message");
	else
		fprintf(fp,"%s",ptr->msg);
	return *this;
}

errstream &
errstream::operator << (class node *nptr) {
	fprintf(fp,"line %d:",nptr->NodeGetline() );
	return *this;
}

errstream &
errstream::operator << (class value *vptr) {
	fprintf(fp,"value 0x%x type=%d:",vptr,vptr->id() );
	return *this;
}



diagstream &
diagstream::operator << (char *s) {
	if (dodump)
		fprintf(fp,"%s",s);
	return *this;
}

diagstream &
diagstream::operator << (int s) {
	if (dodump)
		fprintf(fp,"%d",s);
	return *this;
}

diagstream &
diagstream::operator << (double d) {
	if (dodump)
		fprintf(fp,"%6.2f",d);
	return *this;
}

diagstream &
diagstream::operator << (long s) {
	if (dodump)
		fprintf(fp,"%d",s);
	return *this;
}

diagstream &
diagstream::operator << (address s) {
	if (dodump)
		fprintf(fp,"0x%x",s);
	return *this;
}
