
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: error.h,v 2.4 1995/06/17 20:33:40 davidsi Exp $
//
// $Log: error.h,v $
// Revision 2.4  1995/06/17  20:33:40  davidsi
// - Added index_too_large error message.
//
// Revision 2.3  1995/06/07  20:43:02  davidsi
// - Added knowledge of the value class.
//
// Revision 2.2  1995/06/02  20:03:49  davidsi
// - Added not_string, not_real and incorrect_num_arguments.
// - Added file_not_open.
// - Added field_lookup_failed.
// - Added an errstream operator that can handle node pointer types.  The idea
//   being to allow line numbers (etc) to be dumped out with errors.
//
// Revision 2.1  1995/05/18  20:43:16  davidsi
// - Added not_a_valid_security error message.
//
// Revision 2.0  1995/01/19  21:46:50  davidsi
// - Initial version.
//
//

#ifndef __ERROR_H
#define __ERROR_H

#include "stdio.h"
#include "stdlib.h"

typedef FILE *FP;

enum errmsg_e {
	symbol_not_found,
	function_not_found,
	type_not_defined,
	bad_operation,
	type_mismatch,
	not_evaluated,
	no_expression,
	not_boolean,
	not_integer,
	not_real,
	not_string,
	empty_body,
	cannot_deref,
	cannot_execute,
        bad_arg_decl,
        bad_var_decl,
	bad_type,
	internal_address_not_set,
	not_a_valid_security,
	incorrect_num_arguments,
	file_not_open,
	field_lookup_failed,
	index_too_large,
	no_message			// should always be last  
};

typedef long *address;

class errstream {
	FP fp;
public:
	errstream (FP fl): fp( fl ) {}
	operator FP () { return fp; }
	errstream &operator << (char *);
	errstream &operator << (int);
	errstream &operator << (long);
	errstream &operator << (double);
	errstream &operator << (errmsg_e); 
	errstream &operator << (class node *); 
	errstream &operator << (class value *); 
};

class diagstream {
	FP fp;
	int dodump;
public:
	diagstream (FP fl): fp( fl ) {
 char *x;
                        x=getenv("XSMDIAG");
                        if (x==NULL || strcmp(x,"ENABLED") )
                                dodump=0;
                        else
                                dodump=1;
                        }
	diagstream &operator << (char *);
	diagstream &operator << (int);
	diagstream &operator << (long);
	diagstream &operator << (double);
	diagstream &operator << (address);
};

extern errstream cerror;
extern diagstream cinfo;

//
// Use info to surround all << operations on cinfo.  This allows
// the extra information to be compiled in/out.
//
#ifdef ASSERTS
#define INFO(x) x
#else
#define INFO(x)
#endif

#endif
