
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: list.h,v 2.2 1995/06/09 20:53:44 davidsi Exp $
//
// $Log: list.h,v $
// Revision 2.2  1995/06/09  20:53:44  davidsi
// - Added GNU copyright message.
//
// Revision 2.1  1995/02/27  22:02:57  davidsi
// - Removed set_compare() and made compare() a virtual function that
//   can be overidden in derived classes.  The default compare()
//   function member inserts into the list with minimal list traversal.
//
// Revision 2.0  1995/01/30  21:06:48  davidsi
// - Made the destructors for both list and list_node classes virtual.
//
// Revision 1.2  1994/10/17  20:37:07  davidsi
// - Enhanced the list class to allow for ordered lists.  A node can be
//   inserted in order onto the list using insert_order().  The comparison
//   is set by set_compare().
//
// Revision 1.1  1994/06/27  20:55:51  davidsi
// Initial revision
//
//
#ifndef __LIST_H
#define __LIST_H

//
// This list_node class provides the basic class that can sit on 
// a linked list 
//
class list_node {
	list_node *forw;
	list_node *back;
	class list *list_head;

	friend class list;

public:
	list_node();
	virtual ~list_node();
	list_node *next();
	list_node *previous();
	void remove();
	void add(list_node *);
};


//
// The list class is used to link list_node class members.
//
class list : private list_node {
	int count;			// number of nodes in list

	friend class list_node;

protected:
	virtual int compare(list_node *,list_node *);

public:
	list();
	virtual ~list();

	void insert_head(list_node *);
	void insert_tail(list_node *);
	void insert_order(list_node *);
	list_node *head();
	list_node *tail();
	int eol(list_node *);
	int length();
};


#endif
