
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: main.C,v 2.8 1995/06/24 14:36:17 davidsi Exp $
//
// $Log: main.C,v $
// Revision 2.8  1995/06/24  14:36:17  davidsi
// - Removed 'beta' message in preparation for distribution.
//
// Revision 2.7  1995/06/07  20:45:11  davidsi
// - More credits!
//
// Revision 2.6  1995/05/17  20:40:54  davidsi
// - Added support for multiple portfolios.  Initialize the portfolio support
//   and leave portfolio_load() to load the necessary dividends, transactions
//   and security information.
// - Leave the initialization of transaction, security and dividend lists
//   to the portfolio module. main() simply needs, now, to set them to NULL.
// - The call to val_init() has also been moved to the portfolio module.
// - Change module initialization. The project module is now the last
//   to be initialized so as to guarantee that the Quit item is at the
//   bottom of the Project menu.
//
// Revision 2.5  1995/04/25  21:08:18  davidsi
// - Added support for loading the preferences database.
// - Added delays to force the main window to be displayed in full before
//   initializing the other modules.
// - Added more copyright notices.
//
// Revision 2.4  1995/04/16  20:36:15  davidsi
// - Added calls to DialogPrintf() to print copyright information and
//   startup status.
//
// Revision 2.3  1995/04/03  20:42:31  davidsi
// - Added dsec module.
//
// Revision 2.2  1995/03/08  21:20:52  davidsi
// - Added callout to val_init().
//
// Revision 2.1  1995/03/01  22:13:25  davidsi
// - Added initialization calls for drawps, drawlj and colour classes.
// - Removed all dialog usage.  This has resulted in a MUCH cleaner module.
//
// Revision 2.0  1995/01/19  21:03:24  davidsi
// - Move the *->load() calls so as to get a complete window displayed
//   sooner rather than later.
// - Added module entries for the interpret module.
// - Added dchange module declaration.
//
// Revision 1.4  1994/12/11  21:18:07  davidsi
// - Added dplot module.
//
// Revision 1.3  1994/11/05  21:07:20  davidsi
// - Various modifications to allow this file to compile using g++.
//
// Revision 1.2  1994/10/17  20:49:18  davidsi
// - Removed warrants module definition.
//
// Revision 1.1  1994/07/03  20:37:41  davidsi
// Initial revision
//
//

#include <stdio.h>
#include <varargs.h>
#define _STDARG_H
#include <stdlib.h>
#include <dial.h>

#include "main.h"
#include "security.h"
#include "val.h"
#include "prefs.h"
#include "portfolio.h"
#include "interpret.h"


//
// Variables
//
extern void project_init(menu_bar *);
extern void share_init(menu_bar *);
extern void dsec_init(menu_bar *);
extern void trans_init(menu_bar *);
extern void dtrans_init(menu_bar *);
extern void div_init(menu_bar *);
extern void ddiv_init(menu_bar *);
extern void price_init(menu_bar *);
extern void dprice_init(menu_bar *);
extern void dval_init(menu_bar *);
extern void dplot_init(menu_bar *);
extern void interpret_init(menu_bar *);
extern void dchange_init(menu_bar *);
extern void colour_init(menu_bar *);
extern void drawps_init(menu_bar *);
extern void drawlj_init(menu_bar *);
extern void prefs_init(menu_bar *);
extern void portfolio_init(menu_bar *);
extern void dialog_init(menu_bar *);

menu_bar mb(NULL,NULL,0);

module::module()
{
	initialize=NULL;
}

module::module( void (*function)(menu_bar *) ) 
{
	initialize=function;
}

void
module::init( menu_bar *mb)
{
	if (initialize)
		initialize(mb);	
}


//
// main
//
// The main start of the program
//
//
int
main(unsigned int argc,
	char **argv
	)
{
	int i;

	main_sec_list=NULL;
	main_trans_list=NULL;
	main_div_list=NULL;
	main_price_list=new price_list;

	module project=project_init;
	module shares=share_init;
	module dsec=dsec_init;
	module dtrans=dtrans_init;
	module trans=trans_init;
	module div=div_init;
	module ddiv=ddiv_init;
	module dprice=dprice_init;
	module price=price_init;
	module dval=dval_init;
	module dplot=dplot_init;
	module interpret=interpret_init;
	module dchange=dchange_init;
	module colour=colour_init;
	module drawps=drawps_init;
	module drawlj=drawlj_init;
	module prefs=prefs_init;
	module portfolio=portfolio_init;
	module dialog=dialog_init;

	prefs.init(&mb);
	dialog.init(&mb);
	shares.init(&mb);
	dsec.init(&mb);
	dtrans.init(&mb);
	trans.init(&mb);
	ddiv.init(&mb);
	div.init(&mb);
	dprice.init(&mb);
	price.init(&mb);
	dval.init(&mb);
	dplot.init(&mb);
	interpret.init(&mb);
	dchange.init(&mb);
	colour.init(&mb);
	drawps.init(&mb);
	drawlj.init(&mb);
	portfolio.init(&mb);
	project.init(&mb);

	DialogInitialize(argc,argv,mb);

	DialogCursor(DCURSOR_WAIT);

	DialogNoflush();
	DialogPrintf("XSM (V2.0)\n");
	DialogPrintf("(C) 1995 D.Simons (davidsi@sco.com)\n");
	DialogPrintf("Portions (C) 1993 David Koblas (koblas@netcom.com)\n");
	DialogPrintf("Portions (C) 1989-1990 M.J.E. Mol (marcel@duteca.tudelft.nl)\n");
	DialogPrintf("Portions (C) 1994 Jeffry H. Kingston (jeff@cs.su.oz.au)\n");
	DialogPrintf("Portions (C) 1984-1991 Adobe Systems Incorporated\n");
	DialogPrintf("Portions by David Brooks of the OSF\n");
	DialogPrintf("Thanks to Sverre Froyen & David MacKenzie for ljet font widths\n");
	DialogPrintf("\n");
	DialogFlush();

	DialogUpdatedisplay();
	


	//
	// Load all portfolio information.
	//
	portfolio_load();

	//
	// Load all binaries/source specified by user
	//
	InterpretLoad();

	//
	// Force a menu redisplay so as to ensure the geometry on
	// the main window corrects itself (HACK!)
	//
	DialogRootmenu(mb);
	
	DialogCursor(DCURSOR_NORMAL);

	if (!DialogYesno("*CAUTION*
This program is distributed in the hope that it will be useful, but\n
WITHOUT ANY WARRANTY; without even the implied warranty of\n
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n
See the GNU General Public License for more details.\n
\n
By selecting 'ACCEPT' you are accepting that there is no\n
warranty. If you are unable to use this program with these\n
conditions then you must select 'ABORT'.","ACCEPT","ABORT"))
		exit(0);


	DialogMainloop();
	//
	// NOT REACHED
	//
	return(0);
}
