
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: menu.C,v 2.3 1995/06/24 15:01:04 davidsi Exp $
//
// $Log: menu.C,v $
// Revision 2.3  1995/06/24  15:01:04  davidsi
// - Added GNU copyright message.
//
// Revision 2.2  1995/03/02  21:19:59  davidsi
// - Added support for the dialog class.
// - A new simpler Menuxxx() set of function members are now used to
//   build menus.
// - Added enumerated type for specification of initial menu state.
// - Added a menu_item() function member to build simple action menus.
//
// Revision 2.1  1995/02/02  21:17:10  davidsi
// - Added a new Menu class. This class is a convenience class that allows
//   classes to inherit menu functionality.  One of the key advantages of
//   this class, is that menu callbacks in derived classes do not require
//   static functions (or messy casting).
//
// Revision 2.0  1995/01/19  21:26:07  davidsi
// - Added menu_bar::redisplay() so that menus can be dynamically updated.
//
// Revision 1.3  1994/11/05  21:07:50  davidsi
// - Various modifications to allow this file to compile using g++.
//
// Revision 1.2  1994/10/17  20:31:42  davidsi
// - Tidied up the handling of bar flags so that they are correctly masked.
//
// Revision 1.1  1994/07/03  20:53:51  davidsi
// Initial revision
//
//
//
//

#include <stdio.h>
#include <varargs.h>
#define _STDARG_H
#include <stdlib.h>
#include <string.h>

#define FULL_INCLUDE

#include "localmacros.h"

#include "dialog.h"

//
///////////////////////////////////////////////////////////////////////////////
//
// Simple menu handling.  The menu_bar and menu_item classes allow a menu
// to be built prior to being displayed.
//
///////////////////////////////////////////////////////////////////////////////
//

//
// menu_bar menubar name flags
//
// Adds or creates a menu bar entry.
//
// INPUTS :	menubar		A previous menubar that is bar is to be
//				attached to OR NULL if no previous bar
//				exists
// 		name		The menu name
//		flags		Please refer to flags supported by menus
// RETURNS:	None
//
menu_bar::menu_bar(menu_bar *mb,
		char *iname,
		unsigned long iflags
		)
{

	//
	// If a name is specified then we need to obtain a copy in our
	// own memory
	//
	if (iname) {
		name=new char [strlen(iname) +1 ];
		strcpy(name,iname);
	} else
		name=NULL;

	//
	// Set up remainder of fields
	//
	flags=iflags;
	next=NULL;
	first_item=NULL;

	//
	// Insert new entry at end of list for this menu
	//
	if (mb) {
		while (mb->next!=NULL)
			mb=mb->next;
		mb->next=this;
	}
}


//
// display dialog
//
// Displays a menu bar
//
// INPUTS :	dialog		The dialog in which the menu is to exist
// RETURNS:	None
//
void
menu_bar::display( dialog d_main )
{
	dialog d_menu;
	menu_bar *mb=this;

        d_menubar=menu_createbar(d_main);
	while (mb!=NULL) {
		if (mb->name!=NULL) {
			d_menu=menu_addtobar(d_menubar,mb->name,mb->flags);
			mb->first_item->display(d_menu);
		}
		mb=mb->next;
	}
}


//
// redisplay dialog
//
// Redisplays a menu bar.  This used when a menu bar is already being
// displayed but the entries have been changed since the last display()
// or redisplay() call.
//
// INPUTS :	dialog		The dialog in which the menu is to exist
// RETURNS:	None
//
void
menu_bar::redisplay(dialog d_main )
{
        dialog d_oldmenubar=d_menubar;

	display(d_main);
	menu_refresh(d_oldmenubar,d_menubar);

	// TODO: really need to free () on d_oldmenubar but need to 
	// walk the menu struct (in slib) to do this!
}

//
// builditem menubar barname barflags itemname callback callbackv flags
//
// Adds or creates a menu bar entry.
//
// INPUTS :	menubar		The menubar to use
//		barname		The menu name under which this item will
//				exist
//		barflags	The flags for the bar (if it has to be
//				created)
//		itemname	The name of the menu item to create
//		callback	The callback function address
//		callbackv	The value to pass to the callback
//		flags		The menu item flags
// RETURNS:	None
//
void
menu_item::builditem(
	menu_bar *mb,
	char *bar_name,
	unsigned long bar_flags,
	char *item_name,
	void  (*icallback)(va_list),
	unsigned long icallback_val,
	unsigned long iflags)
{
	menu_bar *smb=mb;
	menu_item *mi;

	name=new char [strlen(item_name) +1 ];
	strcpy(name,item_name);
	callback=icallback;
	callback_val=icallback_val;
	flags=iflags;
	next=NULL;

	//
	// See if the named bar already exists
	//
	while ( smb != NULL ) {
		if ( ( smb->flags==(bar_flags&MENUTYPE_MASK) ) &&
		     ( smb->name != NULL ) &&
		     ( strcmp(smb->name, bar_name)==0 ) )
			break;
		smb = smb->next;
	}

	//
	// Bar does not exist so create one now
	//
	if (smb == NULL )
		smb=new menu_bar (mb,bar_name,bar_flags&MENUTYPE_MASK);

	mi=smb->first_item;

	//
	// Append this item to the bottom of the bar
	//
	if ( mi!=NULL )  {
		while (mi->next!=NULL)
			mi=mi->next;
		mi->next=this;
	} else {
		smb->first_item=this;
	}
}
//
// menu_item menubar barname barflags itemname callback callbackv flags
//
// Adds or creates a menu bar entry.
//
// INPUTS :	menubar		The menubar to use
//		barname		The menu name under which this item will
//				exist
//		barflags	The flags for the bar (if it has to be
//				created)
//		itemname	The name of the menu item to create
//		callback	The callback function address
//		callbackv	The value to pass to the callback
//		flags		The menu item flags
// RETURNS:	None
//
menu_item::menu_item(
	menu_bar *mb,
	char *bar_name,
	unsigned long bar_flags,
	char *item_name,
	void  (*icallback)(va_list),
	unsigned long icallback_val,
	unsigned long iflags)
{
	builditem(mb,bar_name,bar_flags,item_name,icallback,
			icallback_val,iflags);
}

menu_item::menu_item(
	menu_bar *mb,
	char *bar_name,
	char *item_name,
	void  (*icallback)(va_list),
	unsigned long icallback_val
	)
{
	builditem(mb,bar_name,MENUTYPE_ACTION,item_name,
				icallback,icallback_val,0);
}

//
// display dialog
//
// Displays a menu item.  This is intended to be used when creating
// an entire menu bar and not solely a single item.
//
// INPUTS :	dialog		The dialog in which the item is to exist
// RETURNS:	None
//
void
menu_item::display( dialog d)
{
	menu_item *mi=this;

	while (mi!=NULL) {
		menu_additem(d,mi->name,( void (*)(... ) )(mi->callback),mi->callback_val,mi->flags);
		mi=mi->next;
	}
}



//
///////////////////////////////////////////////////////////////////////////////
//
// The Menu convenience class.  This class allows other classes to inherit 
// a simple menu handling capability.  It allows a menu to be defined and
// displayed.  The callback handler is a member of the derived class
// minimizing the number of static functions and casts. 
//
///////////////////////////////////////////////////////////////////////////////
//

//
// The val_cb class is a convenience class used for providing wrappers
// between C callbacks and C++ classes.  Construct members of this
// class and use them to wrap a "this" pointer and an identifier. The
// values are attached to a linked list so that they can be deleted
// at some (appropriate) point in the future.
//
class val_cb 
{
public:
	class Menu *faddr;
        unsigned long cbvalue;
        val_cb( class Menu *faddr , unsigned long);
};

val_cb::val_cb( class Menu *ifaddr,unsigned long icbvalue)
{
        faddr=ifaddr;
        cbvalue=icbvalue;
}


//
// mainmenu varargs
//
// The mainmenu function is static to allow it to be specified as the
// callback function for menu handling.  It unwraps the arguments and
// calls the MenuCallback() member.
//
// INPUTS :	vargs		Expected to contain dialog, wrapper and
//				menu state information.
// RETURNS:	None
//
void
Menu::mainmenu(local_va_dcl)
{
	va_list ap;
	dialog d;
	class val_cb *ptr;
	int flag;

        local_va_start(ap);
        d=va_arg(ap,dialog);
        ptr=va_arg(ap,class val_cb *);
        flag=va_arg(ap,int);

        ptr->faddr->MenuCallback(ptr->cbvalue,flag);
}

//
// MenuCallback
//
// Dummy menu callback handler.  This traps classes that don't 
// override the MenuCallback() function member.
//
void
Menu::MenuCallback(unsigned long id,int flag)
{
	DialogPanic("Menu::MenuCallback() called-should be overridden\n");
}



//
// MenuItem name flags itemname callbackid iflags
//
// Creates a menu item (and bar if named bar does not already exist).
//
// INPUTS :	name		The name of the bar to place the item on
//		flags		The flags for the bar
//		itemname	The name of the menu item
//		callbackid	The id for the callback
//		iflags		The flags for the menu item
// RETURNS:	None
//
void
Menu::MenuItemaction(
	char *bar_name,
	char *item_name,
	unsigned long icallback_val)
{
	class val_cb *cb;

	if (!menubar) 
		menubar=new menu_bar(NULL,NULL,0);

	cb=new class val_cb(this,icallback_val);
	new menu_item(menubar,bar_name,MENUTYPE_ACTION,item_name,
			&Menu::mainmenu,(unsigned long)cb,0);
}

//
// MenuItem name flags itemname callbackid iflags
//
// Creates a menu item (and bar if named bar does not already exist).
//
// INPUTS :	name		The name of the bar to place the item on
//		flags		The flags for the bar
//		itemname	The name of the menu item
//		callbackid	The id for the callback
//		iflags		The flags for the menu item
// RETURNS:	None
//
void
Menu::MenuItem1ofn(
	char *bar_name,
	char *item_name,
	unsigned long icallback_val,
	menustate_e mstate)
{
	class val_cb *cb;
	unsigned long iflags; 

	if (!menubar) 
		menubar=new menu_bar(NULL,NULL,0);

	if (mstate==MENUSTATE_ON)
		iflags=MENUTYPE_ENABLED;
	else
		iflags=0; 

	cb=new class val_cb(this,icallback_val);
	new menu_item(menubar,bar_name,MENUTYPE_ONEOFMANY,item_name,
			&Menu::mainmenu,(unsigned long)cb,iflags);
}

//
// MenuItem name flags itemname callbackid iflags
//
// Creates a menu item (and bar if named bar does not already exist).
//
// INPUTS :	name		The name of the bar to place the item on
//		flags		The flags for the bar
//		itemname	The name of the menu item
//		callbackid	The id for the callback
//		iflags		The flags for the menu item
// RETURNS:	None
//
void
Menu::MenuItemnofn(
	char *bar_name,
	char *item_name,
	unsigned long icallback_val,
	menustate_e mstate)
{
	class val_cb *cb;
	unsigned long iflags; 

	if (!menubar) 
		menubar=new menu_bar(NULL,NULL,0);

	if (mstate==MENUSTATE_ON)
		iflags=MENUTYPE_ENABLED;
	else
		iflags=0; 

	cb=new class val_cb(this,icallback_val);
	new menu_item(menubar,bar_name,MENUTYPE_NOFMANY,item_name,
			&Menu::mainmenu,(unsigned long)cb,iflags);
}

//
// MenuDisplay dialog
//
// Displays the menu (built by MenuItem) within the specified dialog.
// 
// INPUTS :	dialog		The dialog to contain the menu
// RETURNS:	None
//
void
Menu::MenuDisplay(dialog d)
{
	menubar->display(d);
}

//
// Constructor function
//
Menu::Menu()
{
	menubar=NULL; 
}

//
// Destructor function
//
Menu::~Menu()
{
}
