
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: menu.h,v 2.3 1995/06/09 20:53:44 davidsi Exp $
//
// $Log: menu.h,v $
// Revision 2.3  1995/06/09  20:53:44  davidsi
// - Added GNU copyright message.
//
// Revision 2.2  1995/03/02  21:18:58  davidsi
// - Added support for the dialog class.
// - A new simpler Menuxxx() set of function members are now used to
//   build menus.
// - Added enumerated type for specification of initial menu state.
// - Added a menu_item() function member to build simple action menus.
//
// Revision 2.1  1995/02/02  21:17:23  davidsi
// - Added a new Menu class. This class is a convenience class that allows
//   classes to inherit menu functionality.  One of the key advantages of
//   this class, is that menu callbacks in derived classes do not require
//   static functions (or messy casting).
//
//
//
#ifndef __MENU_H
#define __MENU_H

#include "suplib.h"

//
// The menu_item class
//
class menu_item {
	menu_item 	*next;			// next item
	char 		*name;			// name
	void		(*callback)(va_list);	// call back
	unsigned long	callback_val;		// call back value
	unsigned long 	flags;			// flags

private:
	void builditem(class menu_bar *,
		  char *,
		  unsigned long,char *,
		  void (*)(va_list ),
		  unsigned long,unsigned long);
	
public:
	menu_item(class menu_bar *,
		  char *,
		  unsigned long,char *,
		  void (*)(va_list ),
		  unsigned long,unsigned long);
	//
	// Creates simple action menu items
	//
	menu_item(class menu_bar *,
		  char *,
		  char *,
		  void (*)(va_list ),
		  unsigned long);
	void display(dialog);
};


//
// The menu_bar class
//
class menu_bar {
	menu_bar 	*next;			// next on bar
	menu_item	*first_item;		// first item
	char 		*name;			// name of this item
	unsigned long 	flags;			// flags 

	dialog d_menubar;

	friend menu_item;

public:
	menu_bar(class menu_bar *, char *, unsigned long);
	void display(dialog);
	void redisplay(dialog);
};


enum menustate_e {
	MENUSTATE_ON,
	MENUSTATE_OFF,
	};

// 
// The Menu class
//
class Menu 
{
	class menu_bar  *menubar;

private:
	static void mainmenu(va_list);
	virtual void MenuCallback(unsigned long id,int flag);

public:
	void MenuItemaction(char *, char *, unsigned long);
	void MenuItem1ofn(char *, char *, unsigned long, menustate_e);
	void MenuItemnofn(char *, char *, unsigned long, menustate_e);
	void MenuDisplay(dialog d);
	class menu_bar *MenuExtract() { return menubar; };
	Menu(); 
	virtual ~Menu(); 
};

#endif
