/* CHK=0x24D1 */
/*+-------------------------------------------------------------------------
	nap.c - nap() support
	wht@n4hgf.atl.ga.us

  Defined functions:
	Nap(msec)
	ftime(tptr)
	init_Nap()

--------------------------------------------------------------------------*/
/*+:EDITS:*/
/*:05-11-1995-15:55-wht@n4hgf-ck_sigint fools optimizing compilers */
/*:04-11-1995-15:48-wht@n4hgf-pretty work: memset 0 ftime timeb before plug */
/*:05-04-1994-04:39-wht@n4hgf-ECU release 3.30 */
/*:01-12-1994-07:17-wht@fep-move ftime() from ecutime.c */
/*:09-10-1992-14:00-wht@n4hgf-ECU release 3.20 */
/*:08-22-1992-15:39-wht@n4hgf-ECU release 3.20 BETA */
/*:08-10-1992-03:04-wht@n4hgf-add init_Nap */
/*:07-17-1992-18:19-wht@n4hgf-creation of common module for all binaries */

#include "ecu.h"

#undef NULL					 /* some stdio and param.h define these
							  * differently */
#include <sys/param.h>
#ifndef NULL				 /* fake usual sys/param.h value */
#define NULL 0
#endif

#if defined(BSD)
#if !defined(HZ)
#define	HZ	100				 /* I don't want to talk about this */
#endif
#endif

int hz;						 /* HZ from environ or sys/param.h */
ulong hzmsec;				 /* clock period in msec rounded up */

#ifdef CFG_SelectNap
#ifdef CFG_HasFdSet
#define SelBitmask struct fd_set
#else
#define SelBitmask int
#endif /* CFG_HasFdSet */
#endif /* CFG_SelectNap */

/*+-------------------------------------------------------------------------
	ftime(tptr) - SVR3 ftime() surrogate

  If ftime() not available, fake one with gettimeofday() ftime()
  will become less common as SVR3 becomes the less common port;
  this usage should keep the memory of it alive a bit longer <smile>
  See ecu_time.h for a fake struct timeb definition.

--------------------------------------------------------------------------*/
#if defined(CFG_GettodFtime)
void
ftime(tptr)
struct timeb *tptr;
{
	struct timeval tval;
	struct timezone tzval;

	memset((char *)&tval, 0, sizeof(tval));
	gettimeofday(&tval, &tzval);
	memset((char *)tptr, 0, sizeof(*tptr));
	tptr->time = tval.tv_sec;
	tptr->millitm = tval.tv_usec / 1000;

}							 /* end of ftime */
#endif /* CFG_GettodFtime */

/*+-------------------------------------------------------------------------
	Nap(msec) - system-dependent accurate short sleeps
--------------------------------------------------------------------------*/
long
Nap(msec)
long msec;
{
#if defined(M_XENIX) || defined(SCO32v4) || defined(WORKING_NAP)
	return (nap(msec));
#else
#if defined(CFG_SelectNap)
/* precision guard */
#define SECDELTA 684300000L	 /* sometime in 9/91 */
/*
 * Compute  A -= B for timeb structs A, B (old method thanks to ping.c)
 */
#if !defined(OLD_TBSUB)		 /* new tbsub is better number mgt */
#define tbsub(t, t0) \
    if(1) \
    { \
		register int millitm_delta = (t)->millitm; \
		(t)->millitm = (millitm_delta -= (t0)->millitm); \
        (t)->time -= (t0)->time; \
        if(millitm_delta < 0) \
        { \
            (t)->time--; \
            (t)->millitm += 1000; \
        } \
    } else					 /* ; supplied by invoker */
#else /* old tvsub */
#define tbsub(t, t0) \
	if(1) { \
		register long delta_msec = ((((t)->time * 1000L) + ( t)->millitm) - \
                                   (((t0)->time * 1000L) + (t0)->millitm)); \
		(t)->time = delta_msec / 1000; \
		(t)->millitm = delta_msec % 1000; \
	} else					 /* ; supplied by invoker */
#endif /* OLD_TBSUB */

	struct timeb timer;
	struct timeb start;
	struct timeb now;

	ftime(&start);
	start.time -= SECDELTA;
	timer.time = msec / 1000;
	timer.millitm = msec % 1000;
	while (((timer.time * 1000) + timer.millitm) > 0)
	{
		struct timeval tval;

		tval.tv_sec = timer.time;
		tval.tv_usec = timer.millitm * 1000;
		errno = 0;
		if (!select(0, (SelBitmask *) 0, (SelBitmask *) 0,
				(SelBitmask *) 0, &tval))
		{
			break;
		}
		if (errno != EINTR)
			break;
		if (ck_sigint())	 /* if SIGINT posted, exit now */
			return (-1);
		ftime(&now);
		now.time -= SECDELTA;
		tbsub(&now, &start);
		timer.time = msec / 1000;
		timer.millitm = msec % 1000;
		tbsub(&timer, &now);
	}
	ftime(&now);
	now.time -= SECDELTA;
	tbsub(&now, &start);
	msec = (now.time * 1000) + now.millitm;
	return (msec);
#else
#if defined(M_UNIX) || defined(ISC)
	if (msec < hzmsec)
		msec = hzmsec;
	if (nap(msec) < 0)
		return (-1);
	return (msec);
#else
	porting_attention_needed_here;
#endif /* use hacked nap */
#endif /* CFG_SelectNap */
#endif /* working nap */

}							 /* end of Nap */

/*+-------------------------------------------------------------------------
	init_Nap()
--------------------------------------------------------------------------*/
void
init_Nap()
{

	/*
	 * learn tick rate for various timers
	 */
	if (getenv("HZ"))
		hz = (ulong) atoi(getenv("HZ"));
	else
		hz = HZ;
	hzmsec = (ulong) (1000 / hz) + 2;

}							 /* end of init_Nap */

/* vi: set tabstop=4 shiftwidth=4: */
/* end of nap.c */
