
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: node.h,v 2.3 1995/06/17 20:48:12 davidsi Exp $
//
// $Log: node.h,v $
// Revision 2.3  1995/06/17  20:48:12  davidsi
// - The new node_command class provides a class that is able to
//   execute all command type nodes without shifting execution from
//   one function to another.
//
// Revision 2.2  1995/05/29  19:10:24  davidsi
// - Added sourceline member which stores the line number in the source
//   where the node was parsed from. NodeGetline() allows the line number
//   to be read.
//
// Revision 2.1  1995/04/03  20:40:03  davidsi
// - Hold symbols as symbol pntrs rather than names that have to be looked
//   up. Both the node_vardecl and node_var have been modified to achieve
//   this - should be a performance win.
//
// Revision 2.0  1995/01/19  21:34:54  davidsi
// - Initial version.
//
//
//
#ifndef __NODE_H
#define __NODE_H
#include "context.h"
#include "value.h"

#define NODE_ARITH_ADD  1
#define NODE_ARITH_SUB  2
#define NODE_ARITH_MUL  3
#define NODE_ARITH_DIV  4

//
//
class node {
	int sourceline; 
				// Don't remove this!
public:
	int NodeGetline() { return(sourceline); }
	class node_header *base();
	virtual char *nodename();
	virtual class value *evaluate(class context *);
	virtual int lvalue(class context *,char *&,class value *&);
	virtual int resolve(class context *);
	node();
	virtual ~node();
};




class node_header : public node {
	class node *decl;

	//
	// context for this node
	//
	class context *prog_context;
	char heap_built;		// TRUE when prog_context has had
					// its heap allocated.  It is only
					// ever allocated once!

public:
	virtual char *nodename();
	virtual int resolve(class context *);
	int execute();
	int execute(char *funcname);

	node_header(class node *decl);
	virtual ~node_header();
};

//
//
// Becausse many different nodes can be used in a declation list,
// they all form part of the node_declaration family.  This
// is another fundament node type.
//
class node_declaration : public node {

protected:
	class node_declaration *next;
public:
	void setnext(class node_declaration *);
	class node_declaration *getnext();
	node_declaration();
	virtual ~node_declaration();
	
};

//
//
//
class node_funcdecl : public node_declaration {
	char *name;
	class node_typedecl *typedecl;
	class node_vardecl *argdecl;
	class node *vardecl;
	class node_command *actions;

protected:
	//
	// context for this node
	//
	class context *func_context;
public:
	virtual char *nodename();
	//
	// Because func_decl nodes have information needed at function
	// call time, there are some extra functions available!
	//
	class context *getcontext();

	virtual int resolve(class context *);
	virtual class value *evaluate(class context *,class context *,class node_arg *);

	//node_funcdecl(char *,class node *arg,class node *vd,class node *actns);
	node_funcdecl(class node *typedecl, char *fname);
	node_funcdecl(class node *cvardecl, class node_command *cactions);
	node_funcdecl(class node_funcdecl *c_typesym, class node_funcdecl *c_varact, class node *cargdecl);

	virtual ~node_funcdecl();
};


//
// Supported types of variable 
//

//
//
enum argmode_e {arg_ref,		// argument passed by reference
		arg_value,		// argument passed by value
		arg_not			// not an argument
	};

class node_vardecl : public node_declaration {
	char *name;
	class psymbol *psym;
	class node_typedecl *type;
	argmode_e amode;
public:
	virtual char *nodename();
	virtual int resolve(class context *);
	virtual int lvalue(class context *,char *&,class value *&);

	int setaddr(class context *,char *addr);
	argmode_e argmode();

	//
	// NB. next may be either funcdecl or vardecl node!
	//
	node_vardecl(char *,class node_typedecl *,argmode_e m);
	virtual ~node_vardecl();
};

//
// Command type nodes
//
enum node_e {
	NODE_COMPOUND,
	NODE_IF,
	NODE_RETURN,
	NODE_WHILE,
	NODE_DOWHILE,
	NODE_NONE,
	NODE_EVAL,
	NODE_BREAK,
	NODE_CONTINUE,
	};

class node_command : public node {
protected:
	node_e nodeid;
	class node_command *nextnode; 
public:
	int sequence(class context *);
	virtual class node_command* resolve(class context *,
						class node_command *next,
						class node_command *in,
						class node_command *out);

	node_command() {nextnode=NULL;nodeid=NODE_NONE;} ;
	virtual ~node_command() {} ;
};
	

//
//
//
class node_compound : public node_command {
	class node_command *action;
	class node_command *nextfnc;

friend class node_command;

public:
	virtual char *nodename();
	virtual class node_command* resolve(class context *,
						class node_command *next,
						class node_command *in,
						class node_command *out);

	node_compound(class node_command *caction,class node_command *cnext);
	virtual ~node_compound();
};

class node_if : public node_command {
	class node *bexpr;
	class node_command *truecommand;
	class node_command *falsecommand;

friend class node_command;

public:
	virtual char *nodename();
	virtual class node_command* resolve(class context *,
						class node_command *next,
						class node_command *in,
						class node_command *out);

	node_if(class node *bexpr,class node_command *tc,class node_command *fc);
	virtual ~node_if();
};


class node_while : public node_command {
	class node *bexpr;
	class node_command *command;

friend class node_command;

public:
	virtual char *nodename();
	virtual class node_command* resolve(class context *,
						class node_command *next,
						class node_command *in,
						class node_command *out);

	node_while(class node *bexpr,class node_command *cmd);
	virtual ~node_while();
};

class node_dowhile : public node_command {
	class node *bexpr;
	class node_command *command;

friend class node_command;

public:
	virtual char *nodename();
	virtual class node_command* resolve(class context *,
						class node_command *next,
						class node_command *in,
						class node_command *out);

	node_dowhile(class node *bexpr,class node_command *cmd);
	virtual ~node_dowhile();
};


class node_return : public node_command {
	class node *expr;

friend class node_command;

public:
	virtual class node_command* resolve(class context *,
						class node_command *next,
						class node_command *in,
						class node_command *out);
	virtual char *nodename();
	node_return(class node *expr);
	virtual ~node_return();
};

class node_break : public node_command {

friend class node_command;

public:
	virtual char *nodename();
	virtual class node_command* resolve(class context *,
						class node_command *next,
						class node_command *in,
						class node_command *out);

	node_break();
	virtual ~node_break();
};


class node_continue : public node_command {

friend class node_command;

public:
	virtual char *nodename();
	virtual class node_command* resolve(class context *,
						class node_command *next,
						class node_command *in,
						class node_command *out);

	node_continue();
	virtual ~node_continue();
};

//
// node_eval is used as a wrapper around expr nodes.  It is used by
// the execution sequence to evaluate calls for the expr_node.
//
class node_eval : public node_command {
	class node *expr;

friend class node_command;

public:
	virtual char *nodename();
	virtual class node_command* resolve(class context *,
						class node_command *next,
						class node_command *in,
						class node_command *out);

	node_eval(class node *expr);
	virtual ~node_eval();
};




//
// Expression type nodes
//

class node_assign : public node {
	class node *lptr;
	class node *expr;
public:
	virtual char *nodename();
	virtual class value *evaluate(class context *);

	node_assign(class node *lptr,class node *expr);
	virtual ~node_assign();
};

class node_relop : public node {
	class node *bexpr_left;
	class node *bexpr_right;
	operators_e operation;
public:
	virtual char *nodename();
	virtual class value *evaluate(class context *);

	node_relop(operators_e,class node *bexpr,class node *bexpr);
	virtual ~node_relop();
};

class node_expr : public node {
	class node *expr_left;
	class node *expr_right;
	operators_e operation;
public:
	virtual char *nodename();
	virtual class value *evaluate(class context *);

	node_expr(operators_e,class node *,class node *);
	virtual ~node_expr();
};

class node_int : public node {
	int	number;
	class value_int *val;
public:
	virtual char *nodename();
	virtual class value *evaluate(class context *);

	node_int(int num);
	virtual ~node_int();
};

class node_real : public node {
	double	number;
	class value_real *val;
public:
	virtual char *nodename();
	virtual class value *evaluate(class context *);

	node_real(double num);
	virtual ~node_real();
};

class node_string : public node {
	char *str;
	class value_string *val;
public:
	virtual char *nodename();
	virtual class value *evaluate(class context *);

	node_string(char *s);
	virtual ~node_string();
};

class node_call : public node {
	class node_expr *funcaddr;
	class node_arg *arglist;
public:
	virtual char *nodename();
	virtual class value *evaluate(class context *);

	node_call(class node *,class node *argl);
	virtual ~node_call();
};


class node_arg : public node {
	class node *expr;
	class node_arg *nextarg;
public:
	virtual char *nodename();
	virtual class value *evaluate(class context *);
	virtual int lvalue(class context *,char *&,class value *&);

	class node_arg *next();

	node_arg(class node *arg,class node *next);
	virtual ~node_arg();
};

class node_var : public node {
	char *name;
	class psymbol *psym;
public:
	virtual char *nodename();
	virtual int lvalue(class context *,char *&,class value *&);
	virtual class value *evaluate(class context *);
	virtual int resolve(class context *);

	node_var(char *str);
	virtual ~node_var();
};


class node_typedef : public node_declaration {
	class node_typedecl *basetype;
	char *newname;
	int size;
public:
	virtual char *nodename();
	virtual int resolve(class context *);

	node_typedef(class node_typedecl *basetype,char *newname,int size);
	virtual ~node_typedef();
};

class node_deref : public node {
	class node *postfix;
	class node *index;
public:
	virtual char *nodename();
	virtual class value *evaluate(class context *);
	virtual int lvalue(class context *,char *&,class value *&);

	node_deref(class node *post,class node *ind);
	virtual ~node_deref();
};

enum typedecl_e {
	type_symbol,			// symbol is type
	type_struct,			// symbol is a named structure
	type_struct_defn		// symbol is defining structurey
	};

class node_typedecl : public node_declaration {
	typedecl_e type;
	char *identifier; 
	class node_structdefn  *structdef;
public:
	virtual char *nodename();
	virtual int resolve(class context *);
	virtual class value *evaluate(class context *);

	void setstructure(class node_structdefn  *);

	node_typedecl(typedecl_e,char *);
	virtual ~node_typedecl();
};

class node_structdefn : public node_declaration {
	char *fieldname; 
	class node_typedecl *type;
public:
	virtual char *nodename();
	virtual class value *evaluate(class context *);

	void build_struct(class context *cont,class value_struct *val);

	node_structdefn(char *,class node_typedecl *t);
	virtual ~node_structdefn();
};

//
// Used in expressions to access a field
class node_field : public node_declaration {
	class node *expr;
	char *fieldname;
public:
	virtual class value *evaluate(class context *);
	virtual int lvalue(class context *,char *&,class value *&);
	virtual char *nodename();

	node_field(class node *expr,char *);
	virtual ~node_field();
};


#endif
