
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: optionl.C,v 2.1 1995/06/24 15:01:04 davidsi Exp $
//
// $Log: optionl.C,v $
// Revision 2.1  1995/06/24  15:01:04  davidsi
// - Added GNU copyright message.
//
// Revision 2.0  1995/03/02  21:35:52  davidsi
// - Initial version.
//
//
//

#include <stdio.h>
#include <varargs.h>
#define _STDARG_H
#include <stdlib.h>
#include <string.h>

#define FULL_INCLUDE

#include "localmacros.h"

#include "optionl.h"


//
// OptionlistinstanceSet itemname
//
// Sets the current selected item in an optionlist.
//
// INPUTS :	itemname 	The ascii string of the item that is to be
//				selected - the item should ideally exist.
//				If the item does not exist the results are
//				undefined.
// RETURNS:	None
//
void
Optionlistinstance::OptionlistinstanceSet(char *listitem)
{
	dialog_optionlist_setvalue(d_list,listitem);
}


//
// OptionlistinstanceGet
//
// Gets the current selected item in an optionlist.
//
// INPUTS :	None
// RETURNS:	The ascii string of the item that is to be
//		selected, or NULL if no item has been selected.
//
char *
Optionlistinstance::OptionlistinstanceGet()
{
	return(dialog_optionlist_textvalue(d_list));
}



//
// OptionlistinstanceRevise listitems
//
// Resets the items to be used by the optionlist.
//
// INPUTS :	An array of ptrs to the strings that make up the items.
// RETURNS:	None
//
void
Optionlistinstance::OptionlistinstanceRevise(char **newlist)
{
	dialog_optionlist_revise(d_list,newlist);
}




//
// Optionlistinstance ret_count
//
// Returns the current items that form the list.
//
// INPUTS :	None
// RETURNS:	An array of ptrs to strings that form the items
//		ret_count	The number of items in the list
//
char **
Optionlistinstance::OptionlistinstanceGetlist(int &count)
{
	return(dialog_optionlist_value(d_list,&(count)));
}



//
// OptionlistinstanceDestroy
//
// Destroys a specific instance of an option list. If an optionlist
// instance is not destroyed it will remain (chewing up resources)
// until the entire Optionlist instance goes - such long lived
// lists may not be desirable, hence this function.
// 
// INPUTS :	None
// RETURNS:	None
//
void
Optionlistinstance::OptionlistinstanceDestroy()
{
	remove();
	delete this;
}



//
// Connstructor
//
// Generally only of use to the Optionlist class!
//
Optionlistinstance::Optionlistinstance( class Optionlist *ifaddr,
		unsigned long icbvalue
		)
{
        faddr=ifaddr;
        cbvalue=icbvalue;
	d_list=NULL;
}

//
// Destructor
//
Optionlistinstance::~Optionlistinstance()
{
}


//
// mainlist
//
// This is the static function that list callbacks first come to. It
// unwraps the pointer to determine the actual instance of the class
// and an id that can be passed to the OptionlistCallback() member.
// 
// INPUTS :	d 	The dialog causing the callback
//		iptr	The wrapped up values for this and the id.
// RETURNS:	None
//
void
Optionlist::mainlist(dialog d, char *iptr)
{
	class Optionlistinstance *ptr;
	
	ptr=(class Optionlistinstance *)iptr;

        ptr->faddr->OptionlistCallback(ptr->cbvalue);
}



//
// OptionlistCallback id 
//
// Standard optionlist callback handler.  This function will normally always 
// be overriden.
//
// INPUTS :	id 	The id for the optionlist that was "touched".
// RETURNS:	None
//
void
Optionlist::OptionlistCallback(unsigned long id)
{
}

//
// OptionlistCreate Dialog Newoptionlist callbackid
//
// Creates an optionlist. When the optionlist is "touched" the 
// OptionlistCallback function member will be called with the
// callbackid argument.
//
// INPUTS :	dialog		The Dialog in which the list is to be created
//		Newoptionlist	The optionlist specification.
//		callbackid	The id for the callback
// RETURNS:	None
//
class Optionlistinstance *
Optionlist::OptionlistCreate(
		Dialog &d,
		Newoptionlist &nl,
		unsigned long icallback_val
	)
{
	class Optionlistinstance *cb;
	dialog d_list;
	new_optionlist nolist;

	cb=new class Optionlistinstance(this,icallback_val);

	nolist.width=nl.width;
	nolist.height=nl.height;
	nolist.list_item=nl.listitem;
	nolist.listname=nl.listname;
	nolist.selname=nl.selname;
	
	//
	// Convert flags from C++ values to underlying values
	//
	nolist.flags=0;
	if (nl.flags & OLIST_FLAGS_SEARCH)
		nolist.flags |=OPTION_LIST_SEARCH;
	if (nl.flags & OLIST_FLAGS_NOLABEL)
		nolist.flags |=OPTION_LIST_NOLABEL;
	if (nl.flags & OLIST_FLAGS_AUTOP)
		nolist.flags |=OPTION_LIST_AUTOP;
	if (nl.flags & OLIST_FLAGS_NOTEXT)
		nolist.flags |=OPTION_LIST_NOTEXT;

	nolist.fname=nl.font;
	nolist.select=&Optionlist::mainlist;
	nolist.callvalue=(char *)cb;

	d_list=dialog_optionlist(d.d,&nolist);

	cb->d_list=d_list;

	//
	// Add to list, so it can be deleted later on.
	//
	callbacks->insert_head(cb);

	return(cb);
}

// 
// Constructor for the Optionlist class.
//
Optionlist::Optionlist()
{
	callbacks=new class list;
}

// 
// Destructor for the Optionlist class.
//
Optionlist::~Optionlist()
{
	//
	// Delete all remaining callbacks
	//
	delete callbacks;
}
