.\" pcal.man - documentation for Pcal v4.3
.\"
.TH PCAL 1
.SH NAME
pcal \- generate PostScript calendars
.SH SYNOPSIS
.na
.in +5n
.ti -5n
.B pcal
[\fB\-e\fP|\fB\-f\fP\ \fIcal\fP\|]
[\fB\-o\fP\ \fIfile\fP\|]
[\fB\-j\fP\ |\ \fB\-J\fP]
[\fB\-l\fP\ |\ \fB\-p\fP]
[\fB\-m\fP\ |\ \fB\-M\fP]
[\fB\-b\fP\ \fIday\fP|all]
[\fB\-g\fP\ \fIday\fP|all]
[\fB\-O\fP]
[\fB\-G\fP]
[\fB\-s\fP\ \fP[\fIdate_shade\fP][/\fIfill_shade\fP\|]\fP]
[\fB\-F\fP\ \fIday\fP\|]
[\fB\-A\fP|\fB\-E\fP]
[\fB\-t\fP\ \fItitle_font\fP\|]
[\fB\-d\fP\ \fIday_font\fP\|]
[\fB\-n\fP\ \fItext_font\fP[/\fIsize\fP\|]]
[\fB\-L\fP\ \fIfooter_str\fP\|]
[\fB\-C\fP\ \fIfooter_str\fP\|]
[\fB\-R\fP\ \fIfooter_str\fP\|]
[\fB\-N\fP\ \fInotes_str\fP\|]
[\fB\-D\fP\ \fIsymbol\fP\|]
[\fB\-U\fP\ \fIsymbol\fP\|]
[\fB\-x\fP\ \fIxscale\fP\|]
[\fB\-y\fP\ \fIyscale\fP\|]
[\fB\-X\fP\ \fIxtrans\fP\|]
[\fB\-Y\fP\ \fIytrans\fP\|]
[\fB\-I\fP]
[\fB\-B\fP]
[\fB\-S\fP\ |\ \fB\-k\fP\ |\ \fB\-K\fP]
[\fB\-w\fP]
[\fB\-h\fP\ |\ \fB\-u\fP\ |\ \fB\-v\fP]
[month]
[year]
[nmonths]
.SH DESCRIPTION
.I Pcal
generates PostScript to produce landscape or portrait calendars for any 
month and year.  The arguments
.BR month ,
.BR year ,
and
.BR nmonths ,
if provided, should be numeric.  The 
.BR month
value should be in the range 1 \- 12, and the 
.BR year
value should be specified as 1 or 2 digits or as the full 4 digit year.
If no numeric arguments are provided, the calendar for the current month
and year will be generated.
.PP
If one numeric argument is provided, it is interpreted as the
.BR year
value, and calendars for the entire year will be generated.  Otherwise, 
.BR nmonths
months, starting with
.BR month
and
.BR year ,
will be generated.
.PP
For whole-year calendars (i.e. the
.B \-w
option is given), the command line arguments are interpreted somewhat
differently.  By default, all months in the current year are printed, starting
with January.  If the
.B month
argument alone is given, it is expected to be the desired
.B year
to print, and prints all of the months in the given year.  If both
.BR month " and " year
are given, then 12 consecutive months are printed starting at the given
month and year.  If the
.BR month ", " year ", and " nmonths
arguments are all present, printing begins with the given month and year and
.B nmonths
months are printed, rounded up to the nearest multiple of 12.
.PP
.B The Date File
.PP
By default,
.I pcal
simply prints an empty calendar.  Its real power is in its ability to place
``events'' in appropriate days on the calendar, thus allowing the user to
create personalized calendars.  This is achieved through the use of the
``date file''.
.PP
The date file is named
.IR \.calendar " (" pcal.dat
under MS-DOS),
or
.I calendar
for compatibility with older versions.
.I Pcal
will look in several places for such a file.  First, if the environment variable
.BR PCAL_DIR
is defined,
.I pcal
searches the directory indicated by that variable.
Next,
.I pcal 
searches the user's home directory (as specified by the
.BR HOME
environment variable).
If neither
.BR PCAL_DIR
nor
.BR HOME
is defined,
.I pcal
searches the current directory instead.  Finally, the directory where the
.I pcal
executable resides will be checked.  If no date file is found, an empty
calendar is printed; no error is generated.
.PP
If a date file is found, it will be searched for lines with
leading dates matching the requested month and year.
Any text following the dates found will be printed on the calendar under the
appropriate day of the month.  Dates in the
.I \.calendar
file may be expressed in any of several formats:
.PP
.ft CW
.nf
	<ordinal> <day_spec> in <month_spec>{*} {<text>}
	<day_spec> <prep> <date_spec>{*} {<text>}
	<date_spec>{*} {<text>}
.fi
.ft
.PP
Where:
.PP
.nf
	<month_name>	:= first 3+ characters of name of month, or ``all''
	<month_spec>	:= <month_name>, or ``year''
	<day_spec>	:= first 3+ characters of name of weekday, ``day'',
				``weekday'', ``workday'', ``holiday'', ``nonweekday'',
				``nonworkday'', ``nonholiday'', ``new_moon'',
				``first_quarter'', ``full_moon'', or ``last_quarter''
	<ordinal>	:= any ordinal number (``1st'', ``2nd'', etc.), ``first'' ... ``fifth'',
				``last'', ``odd'', ``even'', or ``all''
	<prep>		:= ``before'', ``preceding'', ``after'', ``following'',
				``on_or_before'' (``oob''), or ``on_or_after'' (``ooa'')
	<sep>		:= one or more non-numeric, non-space, non-`*' characters
	<month>	:= a numeric month (1-12)
	<day>		:= day of month (1-31)
	<year>		:= a numeric year
.fi
.ft
.PP
If the
.B \-A
option (American date formats, the default) is given:
.PP
.nf
	<date_spec>	:= <month_name> <day> | <month><sep><day>{<sep><year>}
.fi
.PP
If the
.B \-E
option (European date formats) is given:
.PP
.nf
	<date_spec>	:= <month_name> <day> | <month><sep><day>{<sep><year>}
.fi
.PP
.I Pcal
also allows format specifiers in both the text and foot strings (see the
.BR \-L ,
.BR \-C ,
.BR \-R ,
and
.B \-N
options below); each will be replaced by a corresponding string as outlined
in the table below.  Most of these are derived from the ANSI C
.ft CW
strftime()
.ft
function; the 
.B %[louwMD]
and
.B %[o0+\-]
format specifiers are specific to
.IR pcal :
.PP
.nf
	%a	abbreviated weekday
	%A	full weekday
	%b	abbreviated month name
	%B	full month name
	%d	day of month (1-31)
	%j	day of year (1-366)
	%l	days left in year (0-365)
	%m	month (1-12)
	%U	week number (0-53)
	%W	week number (0-53)
	%u	week number (1-54)
	%w	week number (1-54)
	%y	year w/o century (00-99)
	%Y	year w/century
	%%	`%' character

	%o	print number as ordinal
	%0	print number with leading zeroes
	%+	use following month or year
	%\-	use previous month or year
	%{+N}[DWMY]	adjust date by +N days/weeks/months/years
	%{\-N}[DWMY]	adjust date by \-N days/weeks/months/years
.fi
.PP
The
.B %u
specifier considers the week containing 1/1 as week 1 and the following
logical Sunday (the first day of the week as printed; cf. the
.B \-F
option below) as the start of week 2; 
.B %U
considers the first logical Sunday as the first day of week 1.  
.B %w
and 
.B %W
behave like 
.B %u
and 
.B %U
respectively, but use the first logical Monday instead.  Note that
.B %w
.has a different meaning to
.ft CW
strftime().
.ft
.PP
The 
.B %o
format specifier prints a number as an ordinal, with the appropriate suffix
(``st'', ``nd'', ``rd'', or ``th'' in English) appended.  For example,
.B %od
prints the day of the month as ``1st'', ``2nd'', ``3rd'', etc.
.PP
Unlike 
.ft CW
strftime(),
.ft
.I pcal
defaults to printing numbers (except
.BR %y )
without leading zeroes.  If leading zeroes are desired, the `0'
prefix may be used.  For example,
.B %0j
prints the first day of year as ``001''.
.PP
The
.B %+
and
.B %\-
format specifiers direct
.I pcal
to substitute the following/previous month/year in the following
.B [bBmyY]
specifier.  For example,
.B %+B
prints the name of the next month.
.PP
The
.B %{[+\-]N}[DWMY] 
format specifiers do not print anything, but instead adjust the working
date by \(+-
.BR N days
.RB ( D ),
weeks
.RB ( W ),
months
.RB ( M ),
or years
.RB ( Y ).
Subsequent format specifiers use the adjusted date instead of the
current date.  For example, 
.B %+1M %B %Y
adjusts the date forward by one month and then prints the resulting
month and year (``January 1992'' in December, 1991);
.B %\-2W %b %d
adjusts the date backward by two weeks and prints the resulting
month and day (``Jul 26'' on August 9).
.PP
Such date adjustments are normally cumulative; for example,
.B %+1Y%\-1D
adjusts the date forward by one year and then backward by one day.  If
.B %D
or
.B %M
is specified alone (or if
.B N
is zero),
.I pcal
restores the original date.  Note that
.B %M
has a different meaning to the
.ft CW
strftime()
.ft
function.
.PP
The ``Notes'' box (see below) uses the first of the current month as the
default date.  All foot strings use the first of the current month in
single-month mode and the first of the starting month in whole-year mode.
.PP
Examples:
.PP
.ft CW
.nf
	last Monday in May*         Memorial Day Holiday
.sp
	all Fridays in Oct          Status Meeting, 11 AM
	first workday in all        %-B progress report due
	all Fri in all              Time card due, 3 PM
	all Monday in all           Fiscal week %0W
	-2nd workday in all         Schedule for %+B due %+2D
	2nd full_moon in all        Blue Moon
	Fri on_or_before all 15     Pay Day
	even Fridays in year        Pay Day
	183rd day of year           Mid-year (%l days left)
.sp
	Tue after first Mon in Nov  Election Day (USA)
.sp
	4th Thu in Nov*             Thanksgiving
	Fri after 4th Thu in Nov*   Day after Thanksgiving
.sp
	12/25/90*                   Christmas     # American
	25.12.90*                   Christmas     # European
.sp
	Dec 25*                     Christmas     # American
	25 Dec*                     Christmas     # European
.fi
.ft
.PP
Any non-numeric character may separate numeric dates.  Holidays may
be flagged by following the date immediately with `*'
as in the examples above; this will cause the
date to be printed in gray.
``Each'' and ``every'' are accepted as synonyms for ``all'', and any word may
be used in place of ``in''.  The abbreviations ``oob'' and ``ooa'' may be used
in place of the keywords ``on_or_before'' and ``on_or_after'', respectively.
``Nearest'' attempts to match the specified date; if that fails, it tries
the day after, then the day before,
then two days after, two days before, and so forth until a match occurs.
.PP
Wildcard day names are also provided.  The keyword ``weekday'' applies to any
days which are normally printed in black on the calendar.  The keyword
``workday'' is the same, but does not include any holidays.  The keyword
``holiday'' includes only those days flagged as holidays.  The keywords
``nonweekday'', ``nonworkday'', and ``nonholiday'' are also recognized as
negations of the above.  See the
.B CAVEATS
below for important notes on using these keywords.
Moon phases may also appear as wildcards; ``nm'' is accepted as a
synonym for ``new_moon'', ``1q'' and ``fq'' for ``first_quarter'', ``fm'' for
``full_moon'', ``3q'' for ``third_quarter'', and ``lq'' for ``last_quarter''. 
.PP
Ordinal day numbers may be used to specify dates, either relative to the
month or to the year.  Either words or numeric abbreviations may be used for
``first'' through ``fifth''; higher numbers must be given using the
numeric equivalent (e.g. 100th).  Negative ordinal numbers may even be used.
For example, ``\-2nd'' means ``next to last''.
.PP
``Odd'' and ``even'' do not refer to the actual date; instead, ``odd''
means ``alternate, starting with the first'', and ``even'' means ``alternate,
starting with the second''.  Thus, ``odd Fridays in March'' refers to
the first, third, and (if present) fifth Fridays in March - not to
those Fridays falling on odd dates.
.PP
``All'' refers to each individual month; ``year'' refers to the year
as an entity.  Thus ``odd Fridays in all'' refers to the first, third, and fifth
Friday of each month, while ``odd Fridays in year'' refers to
the first Friday of January and every other Friday thereafter.
.PP
Text in the date file may use C-like escape sequences (i.e. a `\\' followed by
a character, 1 \- 3 octal digits, or `x' followed by 1 \- 2 hexadecimal digits).
Escaped whitespace (including
.B newline
) and the standard ANSI character escapes (`\\a', `\\b', `\\f', `\\n', `\\r',
`\\t', `\\v') are all replaced by a single blank.
.PP
Lines in the
.I \.calendar
file consisting of 
.B year ####
(where
.B ####
is a numeric year) can be used
to set the year for following entries.  This assumes that the following
entries do not contain a year; any date entries containing year information
will set the remembered year to that year.
.PP
Lines in the
.I \.calendar
file consisting of 
.B opt <options>
can be used to override the defaults for
any command-line options except
.BR \-c ,
.BR \-e ,
.BR \-f ,
.BR \-h ,
.BR \-u ,
.BR \-v ,
.BR \-D ", and"
.BR \-U .
Any options specified in this manner
are, in turn, overridden by those specified explicitly on the command line.
.PP
Lines in the
.I \.calendar
file consisting of
.B note{/<number>} <month>
can be used to place notes regarding the
entire month in one of the unused blocks of the calendar.  The
.B <month>
indicator may be either a number 1 through 12 or an alphabetic month name
as described above;  ``note all'' will place the associated text in the
notes block for each month in the current year.
.B <number>
is an optional positive or negative number specifying the
empty box where the associated text is to be placed.  If positive,
.I pcal
counts forward from the first empty box; if negative,
.I pcal
counts backward from the last empty box.  Thus,
.BR ``note/1''
places the associated text in the first empty box;
.BR "note/-3"
in the third-to-last.  The default is -1 if no <number> is given
(last empty box, immediately preceding the small
calendars on the bottom row; cf.
.BR \-S ,
.BR \-k ,
and
.BR \-K ,
below).
.PP
Comments are supported in the
.I \.calendar
file.  Any characters following a `#' character through the end of the
line are ignored.
.PP
.I Pcal
supports rudimentary
.IR cpp -like
functionality in the
date file, allowing the following constructs:
.BR "define | undef" ,
.B if{{n}def} ... {elif ...}* {else ...} endif,
and
.BR include .
Note that these are not preceded by `#' as they are in C.
Symbol names defined using these keywords (or via the 
.B -D
option) are case-insensitive.
It is not an error to
.BR undef
an undefined symbol, nor to
.BR define
a previously-defined one.  An
.BR ifdef
alone is always
.BR false "; an"
.BR ifndef
alone is always
.BR true .
.BR if
is accepted as a synonym for
.BR ifdef .
.PP
The name of the file in the 
.BR include
directive may optionally be
surrounded by either "" or <>, both of which are ignored.  If the
name is not an absolute path, it is taken to be relative to the
directory where the file containing the directive is located.
.I Pcal
is smart enough to translate
.B ~/
to the user's home directory.
.PP
In addition to pre-processing keywords,
.I pcal
also accepts boolean expressions in
.B if{{n}def}
and
.B elif
directives.  These expressions consist of symbol names joined by the boolean
operators
.BR ! ", " & ", " ^ ", and "
.BR | ,
in order of precedence, high to low.  Parentheses may be used to alter the
precedence.  The synonyms
.BR && " and " ||
are accepted for
.BR & " and " | .
A symbol name evaluates to 
.B true
if currently defined,
.B false
if not; thus:
.PP
.ft CW
.nf
	ifdef A | B | C
.fi
.ft
.PP
\&...is
.B true
if any of the symbols A, B, and C is defined, and:
.PP
.ft CW
.nf
	ifdef A & B & C
.fi
.ft
.PP
\&...is
.B true
if they all are.  Note that 
.B ifndef <expr>
is equivalent to 
.B ifdef !( <expr> ).
.PP
.B The Moon File
.PP
If a file of the name
.IR .moon## " (" moon##.dat
under MS-DOS),
where
.B ##
is the last two digits of the calendar year, exists in the same directory
as the date file (or in the directory where 
.I pcal
resides),
.I pcal
uses the information contained within to calculate the phase of the
moon.  If no such file exists,
.I pcal
uses an approximate algorithm.
.PP
Entries in the moon file must conform to the following syntax:
.PP
If the
.B \-A
option (American date formats, the default) is given:
.PP
.ft CW
.nf
	<quarter> <month><sep><day> {<hour><sep><min>}
.fi
.ft
.PP
If the
.B \-E
option (European date formats) is given:
.PP
.ft CW
.nf
	<quarter> <day><sep><month> {<hour><sep><min>}
.fi
.ft
.PP
Where:
.PP
.nf
	<quarter>	:= ``nm'', ``fq'' or ``1q'', ``fm'', ``3q'' or ``lq'' (new moon,
				first quarter, full moon, last quarter)
	<hour>		:= number 0-23 (24-hour clock)
	<min>		:= number 0-59
.fi
.PP
This file must contain entries for all quarter moons in the year,
in chronological order; if any errors are encountered,
.I pcal
will revert to using its default algorithm.
.PP
As in the date file, comments start with `#' and run through the
end of the given line.  
.PP
.B Options
.PP
.I Pcal
has many options:
.P
.TP 12
.B \-e
Prints an empty calendar.  Do not print entries from a
.I \.calendar
file.
.TP
.BI \-f " cal"
Directs
.I pcal
to use the file name
.I cal
as the input file in place of the default
.I \.calendar
file.  Note that the search rules are different when
.B \-f
is used.  If
.I cal
is an absolute file name (i.e., starting with a `/'), then
.I pcal
attempts to open only that file.  Otherwise,
.I pcal
looks for
.I cal
in the current directory, then in the directory indicated by the environment
variable
.B PCAL_DIR 
(if defined), and finally in the directory where the
.I pcal
executable resides.  If the given
.I cal
file is not found, an error results.
.TP
.BI \-o " file"
Directs
.I pcal
to write the output to
.I file
instead of to stdout.
.TP
.B \-l
Causes the output to be in landscape mode (default).  This also resets the
x- and y-axis scaling and translation factors to the defaults for
landscape mode.
.TP
.B \-p
Causes the output to be in portrait mode.  This also resets the
x- and y-axis scaling and translation factors to the defaults for portrait
mode.
.TP
.B \-j
Causes the Julian date (day of year) to be printed in each calendar box.
.TP
.B \-J
Causes the Julian date and the number of days remaining in the year to
be printed in each calendar box.
.TP
.B \-m
Causes moon icons to be printed on dates corresponding to
new, half, and full moons (the default is that no moons are printed).
.TP
.B \-M
Causes moon icons to be printed on all dates (the default is
that no moons are printed).
.TP
.BI \-b " day" " | all"
Causes all dates falling on weekday
.I day
to be printed in black;
.B \-b all
causes all weekdays to be printed in black.
.TP
.BI \-g " day" " | all"
Causes all dates falling on weekday
.I day
to be printed in gray;
.B \-g all
causes all weekdays to be printed in gray.
.IP
The default for the
.BR \-b " and " \-g
options is to print Saturdays and Sundays in gray and other days, unless
flagged as holidays, in black.
.TP
.BI \-O
Causes ``gray'' dates to be printed as outlined characters.
.TP
.BI \-G
Causes ``gray'' dates to be printed as outlined characters filled with gray.
.TP
.BI \-s "{date}{/fill}"
Overrides the default values for date and/or fill box shading.  These values
must be in the range 0.0 (black) through 1.0 (white); they may be set
independently of each other.  The default values are 0.8 for dates and 0.9
for empty boxes.
.TP
.BI \-F " day"
Selects weekday
.I day
as the first day of the week.  The given day will appear in the left-most
column of the calendar.
.TP
.B \-A
Directs
.I pcal
to use American date conventions
.B mm/dd{/yy}
and
.B month dd
) when parsing the date file (default).
.TP
.B \-E
Directs
.I pcal
to use European date conventions
.B dd/mm{/yy}
and
.B dd month
) when parsing the date file.
.TP
.BI \-X " xtrans"
Specifies the x-axis translation value for positioning the output on the page.
.TP
.BI \-Y " ytrans"
Specifies the y-axis translation value for positioning the output on the page.
.TP
.BI \-x " xscale"
Specifies the x-axis scaling factor for the calendar size.
.TP
.BI \-y " yscale"
Specifies the y-axis scaling factor for the calendar size.
.TP
.BI \-t " title_font"
Specifies the name of a font to use to print the
month name and year at the top of the calendar, the foot strings, and the
notes box heading.  For example,
.BR "pcal \-t Times-Roman" .
.TP
.BI \-d " day_font"
Similar to the
.B \-t
option, but selects the font used to print the day numbers and weekday names.
.TP
.BI \-n " text_font\fR[\fB/\fIsize\fR]"
Similar to the
.B \-t
option, but selects the font used to print the text inside each day and in
the notes block.  The user may also select the font size;
.BR "pcal \-n Helvetica/8"
sets the font to 8-point Helvetica.
.TP
.BI \-D " symbol"
Defines the named symbol prior to reading the date file.
.TP
.BI \-U " symbol"
Un-defines the named symbol prior to reading the date file.
.TP
.BI \-L " string"
Causes the accompanying string to be printed as a left-justified footer.
Format specifiers denoting the month and/or year may appear in the string;
the appropriate values will be substituted upon printing.
.TP
.BI \-C " string"
Similar to
.BR \-L ,
but causes the accompanying string to be printed as a centered footer.
.TP
.BI \-R " string"
Similar to
.BR \-L ,
but causes the accompanying string to be printed as a right-justified footer.
.TP
.BI \-N " string"
Causes the accompanying string to be printed as the heading for the "Notes"
box.  Note, however, that 
.I pcal
makes no attempt to ensure that it fits.
.TP
.BI \-B
Causes
.I pcal
to leave unused calendar boxes blank (default is gray).
.TP
.BI \-S
Causes
.I pcal
to suppress printing the small calendars.  See the
.B CAVEATS
section for further details.
.TP
.BI \-k
Causes
.I pcal
to print the small calendars in the upper left corner (the default is to print
them at the lower right). 
.TP
.BI \-K
Causes
.I pcal
to print the small calendar for the previous month in the upper left corner
and the next month in the lower right (the default is to print both at
the lower right).
.TP
.BI \-w
Causes
.I pcal
to print a calendar for 12 consecutive months: 3 rows / 4
columns in landscape mode, 4 rows / 3 columns in portrait mode.  See the
.B CAVEATS
section for details on the use of this option with other options.
.TP
.BI \-c
Causes
.I pcal
to generate a date file suitable for use as input to the Un*x
.I "calendar(1)"
utility.  The normal PostScript output is suppressed.
.TP
.B \-I
Resets all parameters to the program defaults.
.TP
.B \-h
Causes
.I pcal
to write version information, parameter usage message, and full explanation of options
and file formats (to
.IR stdout )
and terminate.
.TP
.B \-u
Causes
.I pcal
to write version information and parameter usage message (to
.IR stdout )
and terminate.
.TP
.B \-v
Causes
.I pcal
to write version information only (to
.IR stdout )
and terminate.
.PP
Any option which normally takes an argument may be specified without
the argument in order to reset the value to the program default.  Note that
while the
.B \-D
option alone clears all the defined symbols, the
.B \-U
option alone has no effect.  The
.B \-
(or
.BR "\-\|\-" " as per"
System V) argument may be used to disambiguate command lines such as:
.IP
.B pcal \-t 9 90
.PP
This could be written instead as one of the following:
.IP
.B pcal \-t \- 9 90
.br
.B pcal \-t \-\|\- 9 90
.PP
If the environment variable
.BR PCAL_OPTS
is defined, its contents are parsed as a command line.  Flags set via
.BR PCAL_OPTS
override the program defaults, but are overridden by options set via 
.B opt
lines in the
.I \.calendar
file or explicitly on the command line.
.SH CAVEATS
The ``workday'' and ``holiday'' keywords are aware of only those holidays
which have already been flagged at the point where they appear.  For example,
consider January 1990:
.PP
.ft CW
.nf
.ne 7
			    January 1990
			 S  M Tu  W Th  F  S
			    1  2  3  4  5  6
			 7  8  9 10 11 12 13
			14 15 16 17 18 19 20
			21 22 23 24 25 26 27
			28 29 30 31
.fi
.ft
.PP
If the
.I \.calendar
file looked like this:
.PP
.ft CW
.nf
	workday on_or_before all 15	payday
	3rd Mon in Jan*			MLK day
.fi
.ft
.PP
\&\.\.\. then
.I pcal
would mark the 15th as ``payday'' since at that point in the
.I \.calendar
file it has no way of knowing that January 15th will later be flagged as a
holiday.  If the two lines were reversed, such that the holiday preceded the
``workday'' wildcard, then
.I pcal
would work as intended, marking instead the 12th as ``payday''.  Also, beware
of year boundaries which affect the handling of all of the day wildcard
keywords.  In general, it is best to place monthly wildcards such as the
example above at the end of each year to achieve the desired effect.
.PP
When the
.B \-w
and
.B \-p
options are used together,
.I pcal
revises the y-scale factor in order to use the entire portrait page; therefore,
the user should avoid using use the
.B \-y
option when using both the
.B \-w
and
.B \-p
options.  Use of the
.B \-w
option in any case effectively disables the
.BR \-m ", " \-M ", " \-j ", and "
.B \-J
options.
.PP
The output of the
.B \-c
option may be used as input to subsequent runs of
.IR pcal .
Note, however, that
.B opt
lines (except for an automatic
.BR "opt \-[A|E]" ),
comments, ``note'' text, and
.BR ifdef 'd-out
source will be lost.
.PP
The
.B \-S
option interacts with
.BR note{/<number>} ;
if used, it should be specified either on the command line or prior to
the first
.B note
line in the date file.
.SH SEE ALSO
cal(1),
calendar(1).
.SH AUTHORS
The original PostScript code to generate the calendars was written by
Patrick Wood (Copyright (c) 1987 by Patrick Wood of Pipeline Associates,
Inc.), and authorized for modification and redistribution.  The calendar
file inclusion code was originally written in
.IR bs (1)
by Bill Vogel of AT&T.  Patrick's original PostScript was modified and
enhanced several times by others whose names have regrettably been lost.
Ken Keirnan of Pacific Bell assembled the original ``C'' version upon which
this is based; additional modifications and enhancements are the work of
Joseph P. Larson, Ed Hand, Andrew W. Rogers, Mark Kantrowitz, Joe Brownlee,
Jamie Zawinski, Richard L. Dyson, Bill Hogsett, Floyd Miller, Andy Fyfe,
and Geoff Kuenning.
