
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: portfolio.h,v 2.1 1995/06/07 20:46:16 davidsi Exp $
//
// $Log: portfolio.h,v $
// Revision 2.1  1995/06/07  20:46:16  davidsi
// - More support for handling multiple portfolios. Added the functions
//   portfolio_checkremove() and portfolio_remove() to handle the removal
//   of a security - they scan through all known portfolios.  Also added
//   portfolio_rename() to handle security renaming across all portfolios.
//
// Revision 2.0  1995/05/19  18:56:04  davidsi
// - Initial version
//
#ifndef __PORTFOLIO_H
#define __PORTFOLIO_H

#include <stdio.h>
#include <string.h>


//
// The Holding class
//
class Holding : public list_node {
	class security *security;
	class sectrans_list *transactions;	// transaction list
	class secdiv_list *dividends;		// dividends list

	//
	// Friends
	//
	friend class Holdinglist;
public:
	class security *HoldingSecurity();
	class dividend *HoldingDividend();
	class secdiv_list *HoldingDividendhead();
	class transaction *HoldingTransaction();
	class sectrans_list *HoldingTransactionhead();

	void HoldingDividendadd(class dividend *);
	void HoldingTransactionadd(class transaction *);
	Holding();
	~Holding();

	//
	// These save casts...
	//
	class Holding *next();
};


//
// The holding list class
//
class Holdinglist : public list {
	unsigned long flags;
public:
	char **HoldinglistExtract();
	class Holding *HoldinglistSelect(char *title);
	class Holding *HoldinglistFind(char *);
	void HoldinglistRevise(class security *dest_sec);
	virtual int compare(list_node *,list_node *);
	//
	// These save casting..
	//
	class Holding *head();
	
	Holdinglist();
	virtual ~Holdinglist();
};

//
// Public functions and structures
//
extern void portfolio_load();
extern void portfolio_validate(class Busy &busy);
extern int portfolio_checkremove(char *name);
extern int portfolio_remove(char *name);
extern int portfolio_rename(char *oldname,char *newname);



extern Holdinglist *portfolio;


#endif
