
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    Significant portions of this module have been derived from the
//    lout formatting system (v3.02). Such portions are:
//
//    COPYRIGHT (C) 1994 Jeffry H. Kingston
// 
//    Jeffry H. Kingston (jeff@cs.su.oz.au)
//    Basser Department Of Computer Science
//    The Universirty of Sydney 2006
//    Australia
//
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: postscript.C,v 2.2.1.1 1995/06/21 20:57:06 davidsi Exp $
//
// $Log: postscript.C,v $
// Revision 2.2.1.1  1995/06/21  20:57:06  davidsi
// - Use the FONTDIR define rather than assuming ./font/ gives the path
//   for font information.
//
// Revision 2.2  1995/05/19  18:55:09  davidsi
// - Fixes to compile under SCO OpenServer release 5.0.
//
// Revision 2.1  1995/03/02  20:58:26  davidsi
// - Added POSTSCRIPT_DEBUG debug option.
// - BUG FIX. EvRetrieve() was incorrectly decoding named character
//   values - assuming the first character was the character of
//   interest.
//
// Revision 2.0  1995/01/31  22:20:33  davidsi
// - First version.
//
//



#include "stdio.h"
#include "stdlib.h"
#include "bool.h"
#include "malloc.h"
#include "ctype.h"

#include "postscript.h"
#include "localmacros.h"

int
font::set(char *,int)
{
}

double
font::textwidth(char *)
{
}

font::font()
{
}

font::~font()
{
}



#define DEFAULT_XHEIGHT 500     // the default XHeight if font has none
#define MAX_CHARS       256     // maximum number of chars in a font
// #define SZ_DFT         1000     // default lout size is 50p
#define MAX_BUFF        512


int Kern=1;		/* Kerning required */

int
StringBeginsWith( char *str,char *pattern )
{
	char *sp, *pp;

	sp = str;
	pp = pattern;
	while( *sp != '\0' && *pp != '\0' ) {
		 if( *sp++ != *pp++ )
			return 0;
	}
	return (*pp == '\0');
} 

/*@::EvRetrieve(), EvName(), EvPrintEncodings()@******************************/
/*                                                                           */
/*  FULL_CHAR EvRetrieve(str, enc)                                           */
/*                                                                           */
/*  Returns the character code corresponding to character name str in        */
/*  ENCODING enc, or 0 if not found.                                         */
/*                                                                           */
/*****************************************************************************/

struct nametochar {
	char *name;
	char *val;
};

/*
   ###################################################################
   #                                                                 #
   #   Full names                                                    #
   #                                                                 #
   #   These symbols and their names are taken directly from the     #
   #   Adobe Systems Inc. Symbol font (see PostScript Language       #
   #   Reference Manual, pp. 256-257).  The only differences are:    #
   #                                                                 #
   #      Adobe:  theta1        Eq:  thetaone                        #
   #              sigma1             sigmaone                        #
   #              phi1               phione                          #
   #              omega1             omegaone                        #
   #                                                                 #
   #   These were needed since Lout identifiers do not have digits.  #
   #                                                                 #
   ###################################################################
*/

struct nametochar mapping[] = {
   "space","\040"  ,
   "exclam","\041"  ,
   "universal","\042"  ,
   "numbersign","\043"  ,
   "existential","\044"  ,
   "percent","\045"  ,
   "ampersand","\046"  ,
   "suchthat","\047"  ,
   "parenleft","\050"  ,
   "parenright","\051"  ,
   "asteriskmath","\052"  ,
   "plus","+"     ,
   "comma","\054"  ,
   "minus","-"     ,
   "period","\056"  ,
   "slash","\057"  ,
   "zero","\060"  ,
   "one","\061"  ,
   "two","\062"  ,
   "three","\063"  ,
   "four","\064"  ,
   "five","\065"  ,
   "six","\066"  ,
   "seven","\067"  ,
   "eight","\070"  ,
   "nine","\071"  ,
   "colon","\072"  ,
   "semicolon","\073"  ,
   "less","\074"  ,
   "equal","\075"  ,
   "greater","\076"  ,
   "question","\077"  ,
   "congruent","\0100"  ,
   "Alpha","\0101"  ,
   "Beta","\0102"  ,
   "Chi","\0103"  ,
   "Delta","\0104"  ,
   "Epsilon","\0105"  ,
   "Phi","\0106"  ,
   "Gamma","\0107"  ,
   "Eta","\0110"  ,
   "Iota","\0111"  ,
   "thetaone","\0112"  ,
   "Kappa","\0113"  ,

   "Lambda","\0114"  ,
   "Mu","\0115"  ,
   "Nu","\0116"  ,
   "Omicron","\0117"  ,
   "Pi","\0120"  ,
   "Theta","\0121"  ,
   "Rho","\0122"  ,
   "Sigma","\0123"  ,
   "Tau","\0124"  ,
   "Upsilon","\0125"  ,
   "sigmaone","\0126"  ,
   "Omega","\0127"  ,
   "Xi","\0130"  ,
   "Psi","\0131"  ,
   "Zeta","\0132"  ,
   "bracketleft","\0133"  ,
   "therefore","\0134"  ,
   "bracketright","\0135"  ,
   "perpendicular","\0136"  ,
   "underscore","\0137"  ,
   "radicalex","\0140"  ,
   "alpha","\0141"  ,
   "beta","\0142"  ,
   "chi","\0143"  ,
   "delta","\0144"  ,
   "epsilon","\0145"  ,
   "phi","\0146"  ,
   "gamma","\0147"  ,
   "eta","\0150"  ,
   "iota","\0151"  ,
   "phione","\0152"  ,
   "kappa","\0153"  ,
   "lambda","\0154"  ,
   "mu","\0155"  ,
   "nu","\0156"  ,

   "omicron","\0157"  ,
   "pi","\0160"  ,
   "theta","\0161"  ,
   "rho","\0162"  ,
   "sigma","\0163"  ,
   "tau","\0164"  ,
   "upsilon","\0165"  ,
   "omegaone","\0166"  ,
   "omega","\0167"  ,
   "xi","\0170"  ,
   "psi","\0171"  ,
   "zeta","\0172"  ,
   "braceleft","\0173"  ,
   "bar","\0174"  ,
   "braceright","\0175"  ,
   "similar","\0176"  ,
   "Upsilonone","\0241"  ,
   "minute","\0242"  ,
   "lessequal","\0243"  ,
   "fraction","\0244"  ,
   "infinity","\0245"  ,
   "florin","\0246"  ,
   "club","\0247"  ,
   "diamond","\0250"  ,
   "heart","\0251"  ,
   "spade","\0252"  ,
   "arrowboth","\0253"  ,
   "arrowleft","\0254"  ,
   "arrowup","\0255"  ,
   "arrowright","\0256"  ,
   "arrowdown","\0257"  ,
   "degree","\0260"  ,
   "plusminus","\0261"  ,
   "second","\0262"  ,
   "greaterequal","\0263"  ,

   "multiply","\0264"  ,
   "proportional","\0265"  ,
   "partialdiff","\0266"  ,
   "bullet","\0267"  ,
   "divide","\0270"  ,
   "notequal","\0271"  ,
   "equivalence","\0272"  ,
   "approxequal","\0273"  ,
   "ellipsis","\0274"  ,
   "arrowvertex","\0275"  ,
   "arrowhorizex","\0276"  ,
   "carriagereturn","\0277"  ,
   "aleph","\0300"  ,
   "Ifraktur","\0301"  ,
   "Rfraktur","\0302"  ,
   "weierstrass","\0303"  ,
   "circlemultiply","\0304"  ,
   "circleplus","\0305"  ,
   "emptyset","\0306"  ,
   "intersection","\0307"  ,
   "union","\0310"  ,
   "propersuperset","\0311"  ,
   "reflexsuperset","\0312"  ,
   "notsubset","\0313"  ,
   "propersubset","\0314"  ,
   "reflexsubset","\0315"  ,
   "element","\0316"  ,
   "notelement","\0317"  ,
   "angle","\0320"  ,
   "gradient","\0321"  ,
   "registerserif","\0322"  ,
   "copyrightserif","\0323"  ,
   "trademarkserif","\0324"  ,
   "product","\0325"  ,
   "radical","\0326"  ,
   "dotmath","\0327"  ,

   "logicalnot","\0330"  ,
   "logicaland","\0331"  ,
   "logicalor","\0332"  ,
   "arrowdblboth","\0333"  ,
   "arrowdblleft","\0334"  ,
   "arrowdblup","\0335"  ,
   "arrowdblright","\0336"  ,
   "arrowdbldown","\0337"  ,
   "lozenge","\0340"  ,
   "angleleft","\0341"  ,
   "registersans","\0342"  ,
   "copyrightsans","\0343"  ,
   "trademarksans","\0344"  ,
   "summation","\0345"  ,
   "parenlefttp","\0346"  ,
   "parenleftex","\0347"  ,
   "parenleftbt","\0350"  ,
   "bracketlefttp","\0351"  ,
   "bracketleftex","\0352"  ,
   "bracketleftbt","\0353"  ,
   "bracelefttp","\0354"  ,
   "braceleftmid","\0355"  ,
   "braceleftbt","\0356"  ,
   "braceex","\0357"  ,
   "angleright","\0361"  ,
   "integral","\0362"  ,
   "integraltp","\0363"  ,
   "integralex","\0364"  ,
   "integralbt","\0365"  ,
   "parenrighttp","\0366"  ,
   "parenrightex","\0367"  ,
   "parenrightbt","\0370"  ,
   "bracketrighttp","\0371"  ,
   "bracketrightex","\0372"  ,
   "bracketrightbt","\0373"  ,
   "bracerighttp","\0374"  ,
   "bracerightmid","\0375"  ,
   "bracerightbt","\0376"  ,
	NULL,0
};
 
static short
EvRetrieve( char *str )
{ 
struct nametochar *ptr;

	ptr=mapping;
	while (ptr->name) {
		if ( !strcmp(ptr->name,str) )
			return(*(ptr->val));
		ptr++;
	}
	if (isprint(*str) && *(str+1)=='\0')
		return(*str);
	return('\0');

} /* end EvRetrieve */


int
postscript::set(char *fontname,int size)
{ 
	char buff[MAX_BUFF], command[MAX_BUFF];
	short ch;
	int wx, llx, lly, urx, ury, xheight2, i, lnum, ligtop;
	int xhfound, wxfound, bfound;
	//
	// Free all previous information
	//
	freeresources();

	pointsize=(double)size;

	filename=(char *)malloc(strlen(FONTDIR)+strlen(fontname)+2);
	sprintf(filename,"%s/%s",FONTDIR,fontname);

	/*
	 * open the Adobe font metrics (AFM) file of the font 
	 */
	fp = fopen(filename,"r");
	if ( fp == NULL ) {
		printf("cannot open font file %s", filename);
		exit(1);
	}

	/*
	 * check that the AFM file begins, as it should, with
	 * "StartFontMetrics" 
	 */
	if ( fgets(buff, MAX_BUFF, fp) == NULL ||
	    sscanf( (char *) buff, "%s", command) != 1 ||
	    strcmp(command, "StartFontMetrics")  ) { 
		printf("font file %s does not begin with StartFontMetrics",
			 filename);
		exit(1);
	}

	/*
	 * initialise font metrics table for the new font 
	 */
	size_table = (struct metrics *) malloc(MAX_CHARS * sizeof(struct metrics));
	if( size_table == (struct metrics *) NULL ) {
		printf("run out of memory while reading font file %s",
			 filename);
		exit(1);
	}
	lig_table= (char *) malloc(2*MAX_CHARS*sizeof(char));
	/*
	 * initialise ligature table for the new font 
	 */
	if( lig_table == (char *) NULL ) {
		printf("run out of memory while reading font file %s",
			filename);
		exit(1);
	}
	for( i = 0;  i < MAX_CHARS;  i++ )  
		lig_table[i] = 1;	/* i.e. char unknown */

	ligtop = MAX_CHARS+2;	/* must avoid ligtop - MAX_CHARS == 0 or 1 */

	/*
	 * initialise kerning table for the new font 
	 */
	kern_table=(unsigned short *) malloc(MAX_CHARS * sizeof(unsigned short));
	if( kern_table == (unsigned short *) NULL ) {
		printf( "run out of memory while reading font file %s",
			 filename);
		exit(1);
	}

	for( i = 0;  i < MAX_CHARS;  i++ )
		kern_table[i] = 0;  /* i.e. no kerns */
	kern_sizes = (int *) NULL;		  /* i.e. no kern sizes */

	/* 
	 * read font metrics file 
	 */
	xhfound = FALSE;  font_table = NULL;  lnum = 1;
	
nextline:
	if ( ( fgets(buff, MAX_BUFF, fp) ) == NULL ) {
		printf( "EndFontMetrics missing from font file %s",
			filename);
		exit(1);
	}
	lnum++;
	sscanf( (char *) buff, "%s", command);
	switch( command[0] ) {

	case 'X':if( !strcmp(command, "XHeight") ) { 
			if( xhfound ) {
				 printf("XHeight found twice in font file (line %d)",lnum);
				exit(1);
			}
			sscanf( (char *) buff, "XHeight %d", &xheight2);
			xheight2 = xheight2 / 2;
			xhfound = TRUE;
		}
		break;

      case 'F': if( !strcmp(command, "FontName") ) {
			if( font_table != NULL ) {
				 printf("FontName found twice in font file %s (line %d)", lnum);
				exit(1);
			}
			sscanf( (char *) buff, "FontName %s", command);
			if( !strcmp(command, "") ) {
				 printf(
					 "FontName empty in file %s (line %d)",
					 lnum);
			}
			font_table=(char *)malloc(strlen(command)+1);
			strcpy(font_table,command);
		}
		break;

	case 'S': if( !strcmp(command, "StartCharMetrics") ) {
		if( font_table == NULL ) {
			printf("FontName missing in file %s",
					 filename);
		}
		if( !xhfound )	
			xheight2 = DEFAULT_XHEIGHT / 2;
		while( fgets(buff, MAX_BUFF, fp) != NULL &&
		         !StringBeginsWith(buff, "EndCharMetrics") ) {

			/*	
			 * read one line containing metric info
			 * for one character 
			 */
			lnum++;  ch = '\0';  
			wxfound = bfound = FALSE;
			i = 0;  
			while( buff[i] == ' ' )  
				i++;
			while( buff[i] != '\n' ) {
				sscanf( (char *) &buff[i], "%s", command);
				if ( !strcmp(command, "N") ) {
					sscanf( (char *) &buff[i], "N %s", command);
					ch = EvRetrieve(command);
				} else if( !strcmp(command, "WX") ) {
					sscanf( (char *) &buff[i], "WX %d", &wx);
					wxfound = TRUE;
				} else if( !strcmp(command, "B") ) {
					sscanf((char *)&buff[i],"B %d %d %d %d",
						&llx, &lly, &urx, &ury);
					bfound = TRUE;
				} else if(!strcmp(command,"L")&&ch!='\0') {
					if( lig_table[ch] == 1 )
						lig_table[ch] = ligtop - MAX_CHARS;
					lig_table[ligtop++] = ch;
					i++;  /* skip L */
					while( buff[i] == ' ' )	
						i++;
			while( buff[i] != ';' && buff[i] != '\n' ) {
				char ligchar;

				sscanf( (char *) &buff[i], "%s", command);
				ligchar = EvRetrieve(command);
				if( ligchar != '\0' )
					lig_table[ligtop++] = ligchar;
				else { 
#ifdef POSTSCRIPT_DEBUG
					printf("ignoring unencoded ligature\ncharacter %s in font file %s (line %d)",
					command, filename,lnum);
#endif
					lig_table[ch] = 1;
				}
#ifdef POSTSCRIPT_DEBUG
				if( ligtop > 2*MAX_CHARS - 5 )
					printf(
					"too many ligature characters\nin font file %s (line %d)",
	 					filename, lnum);
#endif
				while( buff[i] != ' '&&buff[i] != ';' )
					i++;
				while( buff[i] == ' ' )
					i++;
			}
			lig_table[ligtop++] = '\0';
		}
		while( buff[i] != ';' && buff[i] != '\n' )  
			i++;
		if( buff[i] == ';' ) {
			i++;
			while( buff[i] == ' ' )
				i++;
		}
	}
	if ( ch > '\0' && (ch<127) ) { 
		if( !wxfound ) {
#ifdef POSTSCRIPT_DEBUG
			printf(
			        "WX missing in font file %s (line %d)",
				filename, lnum);
#endif
		}
		if ( !bfound ) { 
#ifdef POSTSCRIPT_DEBUG
			printf( "B missing in font file %s (line %d)",
 				filename, lnum);
#endif
		} 
		if( lig_table[ch] == 1 )
			lig_table[ch] = 0;	/* set to known if unknown */
		else if( lig_table[ch] > 1 )		/* add '\0' to end of ligs */
			lig_table[ligtop++] = '\0';
		size_table[ch].left  = llx;
		size_table[ch].down  = lly - xheight2;
		size_table[ch].right = wx;
		size_table[ch].up    = ury - xheight2;
	}
	}
	} else if( Kern && !strcmp(command, "StartKernPairs") ) {
		char ch1, ch2, last_ch1;
	  	char name1[30], name2[30];
		int kc_top, ks_top, pos, num_pairs, ksize;

		if( sscanf((char *)buff,"StartKernPairs %d",&num_pairs)!=1) {
			printf(
				"syntax error on StartKernPairs line in font file %s (line %d)",
			filename, lnum);
			exit(1);
		}
		kc_top = 1;  ks_top = 1;
		kern_chars =(char *) malloc(2 * num_pairs * sizeof(char));
		kern_value =(unsigned char *)malloc(2*num_pairs*sizeof(unsigned char));
		kern_sizes = (int *) malloc(num_pairs * sizeof(int));
		last_ch1 = '\0';
		while( fgets(buff, MAX_BUFF, fp) == (char *) buff &&
			!StringBeginsWith(buff, "EndKernPairs") ) {
			lnum++;
			if( StringBeginsWith(buff, "KPX") ) {
	      			/*
				 * Get the two character names and
				 * kern size from buff 
				 */
				if( sscanf((char *)buff, "KPX %s %s %d",name1,name2,&ksize)!=3 )
					printf("syntax error in font file %s (line %d): %s",
						filename, lnum, buff);

				/* ignore size 0 kern pairs (they are frequent, why?) */
				if( ksize == 0 ) 
					continue;

				/* check that both characters are encoded */
				ch1 = EvRetrieve(name1);
				if( ch1 == '\0' ) {
#ifdef POSTSCRIPT_DEBUG
					printf(
						"unencoded kern character\n%s in font file %s (line %d)",
						name1, filename, lnum);
#endif
					continue;
				}
				ch2 = EvRetrieve(name2);
				if( ch2 == '\0' ) {
#ifdef POSTSCRIPT_DEBUG
					printf(
						"unencoded kern character\n%s in font file %s (line %d)",
						name2, filename, lnum);
#endif
					continue;
				}

				/* check that ch1 is contiguous with previous occurrences */
				if( ch1 != last_ch1 && kern_table[ch1] != 0 ) {
#ifdef POSTSCRIPT_DEBUG
					printf(
						"non-contiguous kerning pair %s %s in font file %s (line %d)\n",
						name1, name2, filename, lnum);
#endif
					continue;
				}
				last_ch1 = ch1;

				// if ch1 never seen before, make new entry
				// in kern_table[] and kern_chars[] 

				if( kern_table[ch1] == 0 ) {
					kern_table[ch1] = kc_top;
					kern_chars[kc_top] = '\0';
					kern_value[kc_top] = 0;
					kc_top++;
				}
				//	
				// find kerning size in kern_sizes[] or
				// else add it to the end 
				//
				for( pos = 1;  pos < ks_top;  pos++ ) {
					if( kern_sizes[pos] == ksize )
						break;
				}
				if( pos == ks_top ) {
					if( ks_top == num_pairs )
						printf(
							"too many kerning pairs in font file %s (line %d)\n",
		 				filename, lnum);
					kern_sizes[pos] = ksize;
					ks_top++;
				}

		     		// insert ch2 into the kc entries
				// (sorted decreasing) for ch1
				for( i = kc_top-1; i >= kern_table[ch1] && kern_chars[i] < ch2;  i-- ) {
					kern_chars[i+1] = kern_chars[i];
					kern_value[i+1] = kern_value[i];
				}
#ifdef POSTSCRIPT_DEBUG
				if( i >= kern_table[ch1] && kern_chars[i] == ch2 )
				printf(
					"kerning pair %s(%d) %s(%d) appears twice in font file %s (line %d)\n",
					 name1,ch1, name2,ch2, filename, lnum);
#endif
				kern_chars[i+1] = ch2;
				kern_value[i+1] = pos;
				kc_top++;
			}
		}
		kern_sizes[0] = ks_top;
	}
	break;


      case 'E': if( !strcmp(command, "EndFontMetrics") ) {
		/*
		 * make a new font record and insert into font tree 
		 */
#ifdef POSTSCRIPT_DEBUG
		font_size(font_table) =
			(BackEnd == POSTSCRIPT) ? SZ_DFT : PlainCharHeight;
		font_xheight2(font_table) =
			(BackEnd == POSTSCRIPT)?xheight2 : PlainCharHeight / 4;
#endif
		//
		// close file
		//
		free(filename);
		fclose(fp);
		return(0);
	}
	break;
	default:
		break;
	}
	goto nextline;
} 


/*@::FontWordSize()@**********************************************************/
/*                                                                           */
/*  FontWordSize(x)                                                          */
/*                                                                           */
/*  Calculate the horizontal and vertical size of WORD or QWORD x, replacing */
/*  ligature sequences by ligature characters wherever they occur.           */
/*                                                                           */
/*****************************************************************************/

double
postscript::textwidth( char *text)
{
#define	MAX_LEN		32767
#define COL             0
#define ROW             1
int  oback[2];
int  ofwd[2];


	char *p, *q, *a, *b;
	int r, u, d, ksize; 
	char *newtext=(char *)malloc(strlen(text)+1);
	strcpy(newtext,text);

	if (pointsize<0) {
		fprintf(stderr,"PANIC:postscript::textwidth - no font set\n");
		exit(1);
	}

	p = q = newtext;

	//
	// An empty string has no length
	//
	if( !(*p) ) 
		return(0);

	d = u = r = 0;
	do { 
		//
		// check for missing glyph (lig_table[] == 1) 
		// or ligatures(lig[] > 1)
		//
		*q = *p++;
		if( lig_table[*q] ) {
			if( lig_table[*q] == 1 ) {
				printf("character %d omitted (it has no glyph in font %s\n)",*q,font_table);
				continue;
			} 
			a = &lig_table[ lig_table[*(p-1)] + MAX_CHARS ];
			while( *a++ == *(p-1) ) {
				b = p;
		    		while( *a==*b && *(a+1) != '\0' && *b != '\0') 
					a++, b++;
				if( *(a+1) == '\0' ) {
					*q = *a;
					p = b;
					break;
				} else {
					 while( *++a );
					      a++;
		    		}
			}
		}
		//
		// Accumulate size of *q 
		//
		if( size_table[*q].up   > u )	
			u = size_table[*q].up;
		if( size_table[*q].down < d )
			d = size_table[*q].down;
		r += size_table[*q++].right;

	} while( *p );

	*q++ = '\0';

    	/* add kern lengths to r */
	for( p = newtext, q = p+1;  *q;  p++, q++ ) {
		int i;
		int j;

 		i = kern_table[*p];
		if( i == 0 )   {
			ksize = 0;
		} else { 	
			for( j = i;  kern_chars[j] > *q;  j++ );
			ksize = (kern_chars[j]==*q) ?
			kern_sizes[kern_value[j]] : 0;
  		}
		r += ksize;
	}

	/* set sizes of x */
	oback[COL] = 0;
	ofwd[COL]  = min(MAX_LEN, r);
	oback[ROW] = min(MAX_LEN, u);
	ofwd[ROW]  = min(MAX_LEN, -d);
	free ( newtext);
	return( ((double)(ofwd[COL]))/1000.0*pointsize );
}


postscript::postscript()
{
	size_table=NULL;
	lig_table=NULL;
	kern_table=NULL;
	font_table=NULL;
	kern_chars=NULL;
	kern_value=NULL;
	kern_sizes=NULL;

	//
	// Indicates that no font is loaded
	//
	pointsize= -1.0;

}

void
postscript::freeresources()
{
	if (size_table)
		free(size_table);
	if (lig_table)
		free(lig_table);
	if (kern_table)
		free(kern_table);
	if (font_table)
		free(font_table);
	if (kern_chars)
		free(kern_chars);
	if (kern_value)
		free(kern_value);
	if (kern_sizes)
		free(kern_sizes);

	//
	// Indicates that no font is loaded
	//
	pointsize= -1.0;
}

postscript::~postscript()
{
	freeresources();
}
