
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: prefs.C,v 2.2 1995/06/07 20:47:37 davidsi Exp $
//
// $Log: prefs.C,v $
// Revision 2.2  1995/06/07  20:47:37  davidsi
// - The Getdefault() function now also check the database for
//   PREFS_DB_ENTRY entries that might specify a value for the
//   indicated "environment" variable.
//
// Revision 2.1  1995/05/19  18:52:34  davidsi
// - Added support for defaults held in the file /etc/defaults/smwb.  This
//   gives system wide information (set by the administrator) rather than
//   user specific preferences.
// - BUG FIX. When opening the preferences database also look at $HOME first.
//
// Revision 2.0  1995/04/25  21:28:44  davidsi
// - Initial version.
//
//
//

#include <stdio.h>
#include <varargs.h>
#define _STDARG_H
#include <stdlib.h>

#include "menu.h"
#include "prefs.h"

class DatabasePrefs *db_prefs;
class Defaultlist *defaults;


//
// DatabasePrefsAssignread key subkey lhs rhs
//
// Reads the next assignment in the database for the <key,subkey> pair.
// The function can be called multiple times to locate all possible
// assignments.
// 
// INPUTS :	key,subkey	The key pair
// RETURNS:	0 if found, else -1.
//		lhs,rhs		The next assignment in the database.
// 
int
DatabasePrefs::DatabasePrefsAssignread(char *key,
		char *subkey,	
		char *lhs,
		char *rhs)
{
	char buffer[1024];
	char tmp[1024];
	char *ptr;
	char *ass;

	for(;;) {
		ass=DatabaseRetrievenext(key,subkey);
		//
		// Never did find an appropriate assignment, so bail
		//
		if (ass==NULL) 
			return(-1);

		strcpy(tmp,ass);
		ptr=tmp;
		while(*ptr!='\0' && *ptr!='=')
			ptr++;
		if (ptr=='\0')
			continue;
		strcpy(rhs,ptr+1);
		*ptr='\0';
		strcpy(lhs,tmp);
		return(0);
	}
}

//
// DatabasePrefsAssign key subkey lhs rhs
//
// Creates an assignment in the database.
//
// INPUTS :	key,subkey	The key pair
// 		lhs,rhs		The assignment to create in the database.
// RETURNS:	None
void
DatabasePrefs::DatabasePrefsAssign(char *key,char *subkey,char *lhs,char *rhs)
{
	char buffer[1024];
	char tmp[1024];
	char *ptr;
	char *ass;

	sprintf(buffer,"%s=%s",lhs,rhs);
	DatabaseRetrievereset();

	for(;;) {
		ass=DatabaseRetrievenext(key,subkey);
		//
		// Never did find an appropriate assignment, so lets
		// create one!
		//
		if (ass==NULL) {
			DatabaseInsertmultiple(key,subkey,buffer);
			return;
		}
		strcpy(tmp,ass);
		ptr=tmp;
		while(*ptr!='\0' && *ptr!='=')
			ptr++;
		if (ptr=='\0')
			continue;
		*ptr='\0';
		//
		// Now check for a matching lhs
		//
		if (!strcmp(tmp,lhs)) {
			DatabaseReplace(buffer);
			return;
		}
	}
}

//
// Constructor
//
DatabasePrefs::DatabasePrefs(char *name) : Database(name)
{
}


//
// Destructor
//
DatabasePrefs::~DatabasePrefs()
{
}



//
// Defaultnode::DefaultnodeLoad
//
// Attempts to read a default record (ie from /etc/default/xsm)
//
// INPUTS :	rec		The record to parse
// RETURNS:	None, but the instance will be updated if the
//		record could be parsed.
//
void
Defaultnode::DefaultnodeLoad(class record *rec)
{
	char *buf;
	char *input;
	char *ptr;

	input=rec->line();
	buf=(char *)malloc(strlen(input)+1);
	strcpy(buf,input);

	for(ptr=buf;*ptr!='\0';ptr++) {
		if (*ptr=='=')
			break;
	}
	//
	// Did not look like an assign so ignore it
	//
	if (*ptr!='=')  {
		free(buf);
		delete this;
		return;
	}
	*ptr='\0';
	ptr++;
	lhs=(char *)malloc(strlen(buf)+1);
	strcpy(lhs,buf);
	rhs=(char *)malloc(strlen(ptr)+1);
	strcpy(rhs,ptr);

	//
	// Onto main list of defaults
	//
	defaults->insert_head(this);
	free(buf);
}


//
// Defaultnode::DefaultnodeMatch vname
//
// Attempts to locate a default variable and, if found, returns its
// current value.
//
// INPUTS :	vname		The name of the variable to search for
// RETURNS:	The current value, or NULL if not found.
//
char *
Defaultnode::DefaultnodeMatch(char *vname)
{
	if (!strcmp(lhs,vname))
		return(rhs);
	
	return(NULL);
}


//
// Constructor for Defaultnode
// 
Defaultnode::Defaultnode()
{
	lhs=rhs=NULL;
}

//
// Destructor for Defaultnode
// 
Defaultnode::~Defaultnode()
{
	if (lhs)
		free(lhs);
	if (rhs)
		free(rhs);
}

//
// default_load args
//
// C wrapper to ensure correct C++ function is called on a callback
//
// INPUTS :	args	The arguments to pass to the C++ function
// RETURNS:	None
//
void
default_load(char *,va_list ap)
{
	class record *rec;
	Defaultnode *newdef=new Defaultnode;

        rec=va_arg(ap,class record *);
        newdef->DefaultnodeLoad(rec);
}


//
// Constructor for Defaultlist
//
Defaultlist::Defaultlist()
{
	set_filename("xsm",WRITE_BY_LINE);
	SlistSetpath("/etc/default");
	//
	// Register our decoder for building the list of defaults
	//
        add_load_vect(default_load);
}

//
// Destructor for Defaultlist
//
Defaultlist::~Defaultlist()
{
}



//
// Getdefault
//
// Locates either a environment variable (preference to) or a default
// setting for the named variable.  If found, it returns its current
// value.
// 
// INPUTS :	vname		The name to search for
// RETURNS:	The value of vname, or NULL if not found
//
char *
Getdefault(char *vname)
{
	Defaultnode *dp;
	char *name;
	char *ret;

	//
	// See if this user has an override.
	//
	name=getenv(vname);
	if (name)
		//
		// Yes, so return the environment value not the default
		// value.
		//
		return(name);


	//
	// Now check for any preferences in the preferences database
	//
	name=db_prefs->DatabaseRetrieve(PREFS_DB_ENTRY,vname);
	if (name)
		return(name);

	//
	// Start looking in the defaults
	//
	for ( dp=defaults->head(); !defaults->eol(dp); dp=dp->next() ) {
		ret=dp->DefaultnodeMatch(vname);
		if (ret)
			return(ret);
	}

	return(NULL);
}



//
// Initialization function for this module
//
prefs_init(menu_bar *)
{
	char *home;

	//
	// Open configuration file
	//
	db_prefs=new class DatabasePrefs(".xsmrc");
	
	//
	// Set up pathname for home directory
	//
	home=getenv("HOME");
	if (home==NULL)
		home="./";
	db_prefs->SlistSetpath(home);
	db_prefs->load();
	
	//
	// Load system wide defaults
	//
	defaults=new class Defaultlist;
	defaults->load();
}
