
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: printer.C,v 2.5 1995/06/17 20:36:57 davidsi Exp $
//
// $Log: printer.C,v $
// Revision 2.5  1995/06/17  20:36:57  davidsi
// - Added support for user selection of printer units.
// - BUG FIX. All the margin values were not being read back from the dialog,
//   so even if the user modified the values they would not be used by the
//   program.
//
// Revision 2.4  1995/05/18  20:42:12  davidsi
// - DialogentryCreate() now requires an argument specifying total number
//   of visible columns.
//
// Revision 2.3  1995/04/25  21:12:25  davidsi
// - Added support for pernamently storing the user's printer preferences
//   in the preference database.
// - Added support for a busy cursor.
//
// Revision 2.2  1995/03/01  22:09:40  davidsi
// - Printer::GimmeDraw() now displays an error if the printer device
//   cannot be correctly determined.
// - Added new SETUP button that allows printer specific information to
//   be setup.
// - Modified to use the Dialog class rather than direct calls to the
//   dialog_xxx functions.
// - Removed the drawps.h include.  This file is printer independant!
// - BUG FIX. For continuous output call PrinterFlush() to ensure
//   information is pumped out to the printer.
//
// Revision 2.1  1995/02/05  17:59:09  davidsi
// - References to drawdialog are now Drawdialog (straight name change).
// - A declaration for "str" in ButtonCallback() was lost when version 2.0
//   was checked in (don't ask!).
//
// Revision 2.0  1995/02/02  21:15:27  davidsi
// Initial revision
//
//
//
//

#include <stdio.h>
#include <varargs.h>
#define _STDARG_H
#include <stdlib.h>
#include <dial.h>

#include "localmacros.h"

#include "dialog.h"
#include "list.h"
#include "paper.h"
#include "menu.h"
#include "printer.h"
#include "drawdialog.h"
#include "prefs.h"

// 
// Defines for entry fields in printer dialog
//
enum PRINTER_DIALOGS {
	ID_WIDTH,
	ID_HEIGHT,
	ID_TYPE,
	ID_LMARGIN,
	ID_RMARGIN,
	ID_TMARGIN,
	ID_BMARGIN,
	//
	// Button identifiers
	//
	BUTTON_CANCEL,
	BUTTON_OK,
	BUTTON_LPT,
	BUTTON_FILE,
	BUTTON_PREVIEW,
	BUTTON_SETUP,
	BUTTON_SETDEFAULTS,
	//
	// Menu identifiers
	//
	MENU_NEXTPAGE,
	MENU_LASTPAGE,
	MENU_COLOURS,
	};


//
// Defines for the database identifiers
//
#define PRT_DB_CLASS		"PRINTER"
#define PRT_DEVICE		"DEVICE"	// device
#define PRT_DESTINATION		"DEST"		// destination (file etc)
#define PRT_DESTVAL_PRT		"PRT"		// destination value
#define PRT_DESTVAL_FILE	"FILE"		// destination value
#define PRT_DESTVAL_PREV	"PREV"		// destination value
#define PRT_WIDTH		"WIDTH"		//
#define PRT_HEIGHT		"HEIGHT"	//
#define PRT_LMARGIN		"LMARGIN"	//
#define PRT_RMARGIN		"RMARGIN"	//
#define PRT_TMARGIN		"TMARGIN"	//
#define PRT_BMARGIN		"BMARGIN"	//

//
// Defines for default values
//
#define DEFAULT_PRINTER "HP LaserJet"
#define DEFAULT_WIDTH INCHES2POINTS(A4_PAPER_WIDTH)
#define DEFAULT_HEIGHT INCHES2POINTS(A4_PAPER_HEIGHT)

//
// Map internal units (points actually) to the units being used by the 
// user
//
#define INT_2_UNITS(x) \
	if (units==UNITS_INCHES)		\
		x=POINTS2INCHES(x);		\
	else if (units==UNITS_CM)		\
		x=POINTS2CM(x);			\
	else if (units==UNITS_POINTS)		\
		;				\

//
// Map internal units (points actually) to the units being used by the 
// user
//
#define UNITS_2_INT(x) \
	if (units==UNITS_INCHES)		\
		x=INCHES2POINTS(x);		\
	else if (units==UNITS_CM)		\
		x=CM2POINTS(x);			\
	else if (units==UNITS_INCHES)		\
		;				\



//
// Global variables
//
class Printerdevice printerdevices;


//
// The device_list class is a holding list for the device class. Simply
// gives a list of all available output devices.
//
class device_list : public list {
	//
	// It really is nothing more and nothing less than a list!
	//
};

class device : public list_node {
	char *name; 
	class Printerdraw * (*create)();
public:
	device(const char *dname,class Printerdraw * (*f)());
	virtual ~device();

friend class Printerdevice;
};

//
// Constructor for device class
//
device::device(const char *istr,class Printerdraw * (*f)())
{
	name=(char *)malloc(strlen(istr)+1);
	strcpy(name,istr);
	create=f;
}

//
// Destructor for device class
//
device::~device()
{
	free(name);
}


//
// Printerdevice::PrinterdeviceAdd devname creatorfunction
//
// Adds a new printer device to the list of available devices
//
// INPUTS :	devname			The name of the device
//		creatorfunction		A function that creates the draw
//					instance that accessing the device  
//
void
Printerdevice::PrinterdeviceAdd(const char *name, class Printerdraw * (*f)() )
{
	class device *chg=new class device(name,f);
	dev.insert_order(chg);
	if (!stalelist)
		free(namelist);
	stalelist=1;
}



//
// Printerdevice::PrinterdeviceExtract
//
// Extracts a list of names of available devices
//
// INPUTS :	None
// RETURNS:	A null terminated array of pointers to names of devices
//
char **
Printerdevice::PrinterdeviceExtract()
{
	char **nameptr;
	int number;
	class list_node *ptr;

	//
	// May already have a valid name list
	//
	if (!stalelist)
		return(namelist);
	
	number=dev.length();

	namelist=(char **)malloc(sizeof(char *)*(number+1));
	nameptr=namelist;
	ptr=dev.head();
	while (!dev.eol(ptr)) {
		*nameptr=((class device *)ptr)->name;
		nameptr++;
		ptr=ptr->next();
	}
	*nameptr=NULL;
	return(namelist);
}


//
// Printerdevice::PrinterdeviceCreate name
//
// Given the name of a device, creates a drawing area that will access
// the device.
//
// INPUTS :	name		The name of the device
// RETURNS:	The drawing area if the device exists, else NULL
//
class Printerdraw *
Printerdevice::PrinterdeviceCreate(const char *name)
{
	class list_node *ptr;

	ptr=dev.head();
	while (!dev.eol(ptr)) {
		if ( !strcmp(name,((class device *)ptr)->name) ) {
			return( ((class device *)ptr)->create() );
		}
		ptr=ptr->next();
	}
	return(NULL);
}


//
// Constructor for the Printerdevice class
//
Printerdevice::Printerdevice()
{
	stalelist=1;
	dev=new class device_list;
}


//
// Destructor for the Printerdevice class
//
Printerdevice::~Printerdevice()
{
	if (!stalelist)
		free(namelist);
	delete dev;
}




//
// Printer::MenuCallback id flag
//
// Handles menu callbacks.
//
// INPUTS :	id		The id of the selected menu item
// 		flag		On/Off flag for toggle items
// RETURNS:	None
//
void
Printer::MenuCallback(unsigned long id,int flag)
{
	switch (id) {
	case MENU_NEXTPAGE:
		if (currentpage<endpage) {
			currentpage++;
			PrinterPage(currentpage);
		}
		break;
	case MENU_LASTPAGE:
		if (currentpage>1) {
			currentpage--;
			PrinterPage(currentpage);
		}
		break;
	case MENU_COLOURS:
		if (drawptr)
			drawptr->ColourSelect();
		PrinterPage(currentpage);
		break;
	}
}


//
// Printer::ButtonCallback id state
//
// Handles all button callbacks.
// 
// INPUTS :	id		The id of the button
// 		flag		On/Off flag for toggle buttons
// RETURNS:	None
//
void 
Printer::ButtonCallback(unsigned long id,buttonstate_e flag)
{
	char *str;
	class Printerdraw *pd;
	char buffer[20];

	//
	// If the printer specific dialog is on display don't accept
	// any input for this dialog.
	//
	if ( printerblk )
		return;

	//
	// Check for the simple on/off buttons and change the
	// state as appropriate.
	//
	switch(id) {
	case BUTTON_LPT:
		dest=PRINTER;
		return;
	case BUTTON_FILE:
		dest=FILE;
		return;
	case BUTTON_PREVIEW:
		dest=PREVIEW;
		return;
	case BUTTON_CANCEL:
		//
		// Save lots of identing by using a GOTO!
		//
		preferences_set= -1;
		goto done;
	case BUTTON_OK:
		break;
	case BUTTON_SETUP:
		str=d_optionlist->OptionlistinstanceGet();
		pd=printerdevices.PrinterdeviceCreate(str);
		if (pd==NULL) {
			DialogMessage("No such printer device");
			return;
                }
		//
		// Block button presses on the main dialog
		//
		printerblk=1;
		pd->PrinterdrawSetup();
		delete pd;
		printerblk=0;
		return;
	case BUTTON_SETDEFAULTS:
                width=DEFAULT_WIDTH;
		INT_2_UNITS(width);
                height=DEFAULT_HEIGHT;
		INT_2_UNITS(height);
		sprintf(buffer,"%6.2f",width);
		d_entrylist1.DialogentrySet(ID_WIDTH,buffer);
		sprintf(buffer,"%6.2f",height);
		d_entrylist2.DialogentrySet(ID_HEIGHT,buffer);
		strcpy(device,DEFAULT_PRINTER);
		d_optionlist->OptionlistinstanceSet(device);
	default:
		return;
	}
	
	//
	// Got an "OK" so pernamently update the user's preferences
	//
	switch(dest) {
	case PRINTER:db_prefs->DatabaseInsert(PRT_DB_CLASS,PRT_DESTINATION,
						PRT_DESTVAL_PRT);
		break;
	case FILE:
		db_prefs->DatabaseInsert(PRT_DB_CLASS,PRT_DESTINATION,
						PRT_DESTVAL_FILE);
		break;
	case PREVIEW:
		db_prefs->DatabaseInsert(PRT_DB_CLASS,PRT_DESTINATION,
						PRT_DESTVAL_PREV);
		break;
	}

	//
	// Must be the OK button (BUTTON_OK) so read all other
 	// fields and initialize ready for output
	//
	str=d_entrylist1.DialogentryGet(ID_WIDTH);
	if (str)
		width=atof(str);
	else
		width=0.0;
	UNITS_2_INT(width);
	db_prefs->DatabaseInsert(PRT_DB_CLASS,PRT_WIDTH,width);

	str=d_entrylist2.DialogentryGet(ID_HEIGHT);
	if (str)
		height=atof(str);
	else
		height=0.0;
	UNITS_2_INT(height);
	db_prefs->DatabaseInsert(PRT_DB_CLASS,PRT_HEIGHT,height);

	//
	// Store the margin values
	//
	str=d_entrylist2.DialogentryGet(ID_LMARGIN);
	if (str)
		left_margin=atof(str);
	else
		left_margin=0.5;
	UNITS_2_INT(left_margin);
	db_prefs->DatabaseInsert(PRT_DB_CLASS,PRT_LMARGIN,left_margin);

	str=d_entrylist2.DialogentryGet(ID_RMARGIN);
	if (str)
		right_margin=atof(str);
	else
		right_margin=0.5;
	UNITS_2_INT(right_margin);
	db_prefs->DatabaseInsert(PRT_DB_CLASS,PRT_LMARGIN,right_margin);

	str=d_entrylist2.DialogentryGet(ID_TMARGIN);
	if (str)
		top_margin=atof(str);
	else
		top_margin=0.5;
	UNITS_2_INT(top_margin);
	db_prefs->DatabaseInsert(PRT_DB_CLASS,PRT_LMARGIN,top_margin);

	str=d_entrylist2.DialogentryGet(ID_RMARGIN);
	if (str)
		right_margin=atof(str);
	else
		right_margin=0.5;
	UNITS_2_INT(right_margin);
	db_prefs->DatabaseInsert(PRT_DB_CLASS,PRT_LMARGIN,right_margin);

	//
	// Get the device
	str=d_optionlist->OptionlistinstanceGet();
	if (str) {
		strcpy(device,str);
		db_prefs->DatabaseInsert(PRT_DB_CLASS,PRT_DEVICE,str);
	}
	preferences_set=1;

done:
	d_shell.DialogDestroy();

#undef _L_BUFSIZE
}

//
// Printer::OptionlistCallback id
//
// Handles all optionlist callbacks.
// 
// INPUTS :	id		The id of the optionlist
// RETURNS:	None
void 
Printer::OptionlistCallback(unsigned long id)
{
	char *str;

	switch(id) {
	case 0:
		break;
	default:
		DialogPanic("Printer::OptionlistCallback - unknown id");
	}
	
}


//
// Printer::PrinterPapersize parent
//
// Adds the dialogs to a parent dialog that allow the user to specify
// the date, the number of securities, etc.
// 
// INPUTS :	parent		The parent Dialog
// RETURNS:	None
///
void 
Printer::PrinterPapersize(class Dialog &d_parent)
{
	char buffer[256];

	d_horiz_paper.DialoghorizontalCreate(d_parent);

	d_entrylist1.DialogentryCreate(d_horiz_paper,10);
	d_entrylist2.DialogentryCreate(d_horiz_paper,10);

	d_entrylist1.DialogentrySet("Paper Width",width,ID_WIDTH);
	d_entrylist2.DialogentrySet("Paper Height",height,ID_HEIGHT);

	d_entrylist1.DialogentrySet("Left Margin",left_margin,ID_LMARGIN);
	d_entrylist2.DialogentrySet("Right Margin",right_margin,ID_RMARGIN);
	d_entrylist1.DialogentrySet("Top Margin",top_margin,ID_TMARGIN);
	d_entrylist2.DialogentrySet("Bottom Margin",bottom_margin,ID_BMARGIN);

}




//
// Printer::GimmeDraw
// 
// Returns a Draw instance that the inheriting/owning class should 
// use when writing to the output device.
//
// INPUTS :	None
// RETURNS:	A drawing area for accessing the output device
// 
class Draw *
Printer::GimmeDraw()
{
	class Drawdialog *dd;
	class Printerdraw *pd;

	if (dest==PREVIEW) {
		drawd_reg_t drt;

		drt.flags=ALLOW_SIZE_REDRAW|INITIAL_SIZE_SET;
		drt.width=width;
		drt.height=height;
		drt.title="Preview";
		MenuItemaction("Project","Next Page",MENU_NEXTPAGE);
		MenuItemaction("Project","Last Page",MENU_LASTPAGE);
		MenuItemaction("Project","Colours",MENU_COLOURS);

		drt.mb=MenuExtract();
		printmode=PAGE;
		//
		// We have to use dd (class Drawdialog) so that we
		// can call the DrawdialogCreate() function which
		// does not exist for the base Draw class. 
		//
		dd=new class Drawdialog();
		dd->DrawdialogCreate(&drt);
		drawptr=dd;
	} else {
		printmode=CONTINUOUS;
		pd=printerdevices.PrinterdeviceCreate(device);
		if (dest==FILE)
			pd->PrinterdrawMode(PRTMDE_FILE);
		else
			pd->PrinterdrawMode(PRTMDE_PRINTER);
		drawptr=pd;
		if (drawptr==NULL) {
			DialogMessage("No such printer device");
			return(NULL);
		}
	}
	//
	// Set up actual drawing area.
	//
	drawptr->DrawSetarea(left_margin,top_margin,
			width-left_margin-right_margin,
			height-top_margin-bottom_margin);

	return(drawptr);
}


//
// Printer::PrinterOutput startpage endpage 
// 
// Outputs a sequence of pages. If a CONTINUOUS print mode has been
// selected then all pages are produced, otherwise a single page
// is output and the system will wait for menu (etc) input.
//
//
// INPUTS :	startpage,endpage	The page range to output
// RETURNS:	0 if the printer instance has be deleted, else 1.
//
int
Printer::PrinterOutput(int sp, int ep )
{
	startpage=sp;
	endpage=ep;

	switch( printmode) {
	case CONTINUOUS:
		for(currentpage=sp;currentpage<=endpage;currentpage++)
			PrinterPage(currentpage);
		delete this;
		return(0);
	case PAGE:
		currentpage=sp;
		PrinterPage(currentpage);
		return(1);
	}
}

//
// Printer::PrinterSetup
//
// Displays a printer setup Dialog so the user can modify how
// output will be produced.
//
// INPUTS :	None
// RETURNS:	None
//
int
Printer::PrinterSetup()
{
	buttonstate_e f1,f2,f3;
        static Newoptionlist nolist = {
                30,			// width
                5,			// height
                NULL,			// list of items
                "Available Devices",	// name of list
                "",			// name of item
                0,			// flags
		NULL,			// default font
        };
	dcursortype_e cursor;


	cursor=DialogCursor( DCURSOR_WAIT );

	d_shell.DialogshellCreate("Page Setup");

	d_vert_main.DialogverticalCreate(d_shell);

	PrinterPapersize(d_vert_main);

	nolist.listitem=printerdevices.PrinterdeviceExtract();
	d_optionlist=OptionlistCreate(d_vert_main,nolist,0);
	d_optionlist->OptionlistinstanceSet(device);
	//
	// NB: Don't free the array of pntrs! returned from 
	// printerdevices.PrinterdeviceExtract() - it was
	// not ours to free!
	//

	d_horiz_prt.DialoghorizontalCreate(d_vert_main);
	f1=BUTTON_DISABLED;
	f2=BUTTON_DISABLED;
	f3=BUTTON_DISABLED;
	switch(dest) {
	case PRINTER:
		f1=BUTTON_ENABLED;
		break;
	case FILE:
		f2=BUTTON_ENABLED;
		break;
	case PREVIEW:
		f3=BUTTON_ENABLED;
		break;
	}
	(void)ButtonTogglemutex(d_horiz_prt,"To lpt:",BUTTON_LPT,f1);
	(void)ButtonTogglemutex(d_horiz_prt,"To File",BUTTON_FILE,f2);
	(void)ButtonTogglemutex(d_horiz_prt,"Preview",BUTTON_PREVIEW,f3);

	d_horiz_extra.DialoghorizontalCreate(d_vert_main);
	(void)ButtonPush(d_horiz_extra,"PRINTER SETUP",BUTTON_SETUP);
	(void)ButtonPush(d_horiz_extra,"SET DEFAULTS",BUTTON_SETDEFAULTS);

	d_horiz_ok.DialoghorizontalCreate(d_vert_main);
	(void)ButtonPush(d_horiz_ok,"OK",BUTTON_OK);
	(void)ButtonPush(d_horiz_ok,"CANCEL",BUTTON_CANCEL);

	d_shell.DialogRealize();
	DialogCursor( cursor );

	preferences_set=0;
	while (preferences_set==0)
		DialogEvent();

	if (preferences_set<0)
		return(-1);

	return(0);
}


//
// Constructor function for the Printer class.
//
Printer::Printer()
{
	char *str;

	//
	// Set up measurement units
	//
	str=db_prefs->DatabaseRetrieve(UNITS_DB_CLASS,UNITSDEFINED_DB_CLASS);
	if (!strcmp(str,MUNITS_INCHES))
		units=UNITS_INCHES;
	else if (!strcmp(str,MUNITS_CM))
		units=UNITS_CM;
	else if (!strcmp(str,MUNITS_POINTS))
		units=UNITS_POINTS;
        else
		//
		// Default is inches
		//
		units=UNITS_INCHES;

	
	str=db_prefs->DatabaseRetrieve(PRT_DB_CLASS,PRT_WIDTH);
	if (str) 
		width=atof(str);
	else
		width=DEFAULT_WIDTH;
	INT_2_UNITS(width);

	str=db_prefs->DatabaseRetrieve(PRT_DB_CLASS,PRT_HEIGHT);
	if (str) 
		height=atof(str);
	else
	        height=DEFAULT_HEIGHT;
	INT_2_UNITS(height);

	//
	// Fetch the margin values
	//
	str=db_prefs->DatabaseRetrieve(PRT_DB_CLASS,PRT_LMARGIN);
	if (str) 
		left_margin=atof(str);
	else
		left_margin=INCHES2POINTS(0.5);
	INT_2_UNITS(left_margin);

	str=db_prefs->DatabaseRetrieve(PRT_DB_CLASS,PRT_RMARGIN);
	if (str) 
		right_margin=atof(str);
	else
		right_margin=INCHES2POINTS(0.5);
	INT_2_UNITS(right_margin);

	str=db_prefs->DatabaseRetrieve(PRT_DB_CLASS,PRT_TMARGIN);
	if (str) 
		top_margin=atof(str);
	else
		top_margin=INCHES2POINTS(0.5);
	INT_2_UNITS(top_margin);

	str=db_prefs->DatabaseRetrieve(PRT_DB_CLASS,PRT_BMARGIN);
	if (str) 
		bottom_margin=atof(str);
	else
		bottom_margin=INCHES2POINTS(0.5);
	INT_2_UNITS(bottom_margin);

	//
	// Fetch the device
	//
	str=db_prefs->DatabaseRetrieve(PRT_DB_CLASS,PRT_DEVICE);
	if (str)
		strcpy(device,str);
	else
		strcpy(device,DEFAULT_PRINTER);

	str=db_prefs->DatabaseRetrieve(PRT_DB_CLASS,PRT_DESTINATION);

	if (!strcmp(str,PRT_DESTVAL_PRT))
		dest=PRINTER;
	else if (!strcmp(str,PRT_DESTVAL_FILE))
		dest=FILE;
	else if (!strcmp(str,PRT_DESTVAL_PREV))
		dest=PREVIEW;
	else 
		dest=PRINTER;


	printerblk=0;
	drawptr=NULL;
}

//
// Destructor function for the Printer class.
//
Printer::~Printer()
{
}
