
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: printer.h,v 2.4 1995/06/24 15:06:13 davidsi Exp $
//
// $Log: printer.h,v $
// Revision 2.4  1995/06/24  15:06:13  davidsi
// - Added GNU copyright message.
//
// Revision 2.3  1995/06/17  20:36:48  davidsi
// - Added support for user selection of printer units.
//
// Revision 2.2  1995/04/25  21:12:10  davidsi
// - Added a further horizontal dialog for the setup and default buttons.
// - Expose the OptionlistCallback() function so printer selections cause
//   callbacks.
//
// Revision 2.1  1995/03/01  22:11:14  davidsi
// - The printer destructor needs to be virtual to allow derived classes to
//   override it.
// - Added PrinterFlush() which gets overriden in derived classes to ensure
//   information is pumped out to the printer.
// - Added a Printerdevice class. This allows the managment of different
//   printer drivers.
// - The removed of dialog references has required the appropriate Dialog
//   entries to be added to the Printer class.
//
// Revision 2.0  1995/02/02  21:15:42  davidsi
// - Initial version.
//
//
//
//
#ifndef __PRINTER_H
#define __PRINTER_H

#include "dialog.h"
#include "button.h"
#include "optionl.h"
#include "draw.h"

//
// Destinations for output
//
enum destination_e {
	PRINTER,
	PREVIEW,
	FILE
	};

//
// Print modes
//
enum prtmode_e {
	CONTINUOUS,
	PAGE
	};

//
// Measurement units
//
enum prtunits_e {
	UNITS_CM,
	UNITS_INCHES,
	UNITS_POINTS,
	};

class Printer : private Button, private Menu, private Optionlist {
	double width;			// width of output paper
	double height;			// height of output paper
	double left_margin;		// margin descriptors
        double right_margin;
        double top_margin;
        double bottom_margin;
	char device[256];		// ascii representation of output dev
	int preferences_set;		// 0 while printer preferences are being
					// chosen.
	destination_e dest;		// destination selected
	prtmode_e printmode;		// mode for printing
	prtunits_e units;		// measurement units
	int startpage,endpage;		// range of pages to print
	int currentpage;		// current pages
	//
	// Dialogs (etc) used when defining the output device
	//
	Dialogshell d_shell;				// top most dialog
	Dialogentry d_entrylist1;
	Dialogentry d_entrylist2;
	Optionlistinstance *d_optionlist;
	Dialoghorizontal d_horiz_paper;
	Dialogvertical d_vert_main;
	Dialoghorizontal d_horiz_prt;
	Dialoghorizontal d_horiz_ok;
	Dialoghorizontal d_horiz_extra;

	char printerblk;		// TRUE when printer specific dialog
					// is on display and all buttons
					// should be ignored in the main
					// printer dialog.
	//
	// This is only a pointer to a possible Draw instance.  The
	// printer instance does not own, allocate, or free this. The
	// pointer allows the Printer class to affect the underling
	// drawing where applicable
	//
	class Draw *drawptr;

private:

	//
	// This overrides the menu handler in the menu class
	// 
	virtual void MenuCallback(unsigned long id,int flag);

	//
	// This overrides the pure virtual button handler in the Button class
	// 
	virtual void ButtonCallback(unsigned long id,buttonstate_e flag);
	
	//
	// This overrides the virtual optionlist callback handler
	//
	virtual void OptionlistCallback(unsigned long id);


	void PrinterPapersize(Dialog &d_parent);
	virtual void PrinterPage(int pageno) = 0;   
public:
	class Draw *GimmeDraw();
	int PrinterSetup();
	int PrinterOutput(int startpage,int endpage);
	Printer();
	virtual ~Printer();
};

class Printerdevice {
	char stalelist;
	class device_list *dev;
	char **namelist;
public:
	void PrinterdeviceAdd(const char *name, class Printerdraw *(*fname)() );
	char **PrinterdeviceExtract();
	class Printerdraw *PrinterdeviceCreate(const char *name);

	Printerdevice();
	~Printerdevice();
};

extern class Printerdevice printerdevices;

#endif
