
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: project.C,v 2.6.1.1 1995/06/23 20:55:16 davidsi Exp $
//
// $Log: project.C,v $
// Revision 2.6.1.1  1995/06/23  20:55:16  davidsi
// - Added help system.
//
// Revision 2.6  1995/05/19  18:50:34  davidsi
// - Switch order of menu items so that 'Quit' is the last menu item created
//   by this module. main() now calls this module's init function last, so
//   Quit is guaranteed to be the last menu item on the project menu.
//
// Revision 2.5  1995/04/25  21:09:34  davidsi
// - Added support for a busy cursor.
// - BUG FIX. Removed Getinput() as it is better to remove the dialog on the
//   callback as that will ensure it goes away as soon as the button is pressed.
// - BUG FIX. In DialogSummary(), don't count non-traded securities as
//   having been updated.
//
// Revision 2.4  1995/04/16  20:28:23  davidsi
// - BUG FIX. name_fetch() was using named files and not removing them.
//   Now obtains temporary filenames and unlinks the files.
//
// Revision 2.3  1995/04/03  20:39:07  davidsi
// - In name_fetch() "." must now also be considered as valid when scanning
//   for security names - the teletext folks have changed all the names
//   (again!).
// - So must parenthesis - to make this easier added a allowchar macro.
// - Use the new Dialoglist class so as to re-enable the bulk create
//   feature.
// - BUG FIX. Getinput() was not releasing the string list.
//
// Revision 2.2  1995/03/08  21:22:58  davidsi
// - create_securities() must now use the Securityedit class rather than
//   the security class when allowing the user to create securities.
// - The extract() member of security was renamed to SecurityExtract(), so
//   so this module modified to reflect that. The set() member also became
//   SecurityCreate().
//
// Revision 2.1  1995/03/01  22:23:46  davidsi
// - Removed all dialog references - use the Dialog class instead.
//
// Revision 2.0  1995/01/19  21:27:27  davidsi
// - Removed time.h include.
//
// Revision 1.3  1994/11/05  21:08:26  davidsi
// - Various modifications to allow this file to compile using g++.
//
// Revision 1.2  1994/10/18  20:10:01  davidsi
// - Updated to use teletext (ttext class) to allow security names to
//   be loaded.
// - Still much work needed.
//
// Revision 1.1  1994/07/03  16:24:52  davidsi
// Initial revision
//
//

#include <stdio.h>
#include <varargs.h>
#define _STDARG_H
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <dial.h>
#include <ctype.h>


#include "localmacros.h"

#include "dialog.h"
#include "button.h"
#include "support.h"
#include "list.h"
#include "security.h"
#include "ttext.h"

#define DEFAULT_NS     "nameldr"

//
// The allowchar() macro is used to determine which characters
// are valid in a security name when scanning the teletext input
//
#define allowchar(c) \
	(  (isalnum((c))) || 	\
	((c)=='&') ||		\
	((c)==')') ||		\
	((c)=='(') ||		\
	((c)=='\"') )

//
// project_exit
//
// Terminates execution of the program
//
// INPUTS : 	None
// RETURNS: 	None
//
static void
project_exit(local_va_dcl)
{
	va_alist=va_alist;	// stop compiler warnings
	exit(0);
}

class name_list : public list {
	class security *basesec;
public:
	char **datalist;
	char **extract();
	void create_securities(char **list,int);

	//	name_list();
	//	~name_list();
};

class name_node : public list_node {
public:
	char *name;
};



char **
name_list::extract()
{
	class name_node *ptr;
	char **namelist;
	char **nameptr;
	int number;
	class security *sec;
	char **al_names;
	int al_count;
	int iter;
	char *secname;
	unsigned long secflags;

	number=length();
	nameptr=namelist=(char **)malloc(sizeof(char *)*(number+1));

	ptr=(name_node *)head();
	while (!eol(ptr)) {
		//
		// See if the name is new..
		//
		sec=main_sec_list->head();
		while ( !main_sec_list->eol(sec) ) {
			sec->SecurityExtract(&secname,&secflags,
						&al_names,&al_count,NULL);
			if ( !(strcmp(ptr->name,secname)) ) {
				goto next_iter;
			}
			for(iter=0;iter<al_count;iter++)
				if ( !(strcmp(ptr->name,al_names[iter])) ) {
					goto next_iter;
				}
			sec=sec->next();
		}
		*nameptr =(char *)malloc(strlen(ptr->name)+1);
		strcpy(*nameptr,ptr->name);
		nameptr++;
next_iter:
		ptr=(name_node *)ptr->next();
		DialogCursor(DCURSOR_BUSY);
	}
	DialogCursor(DCURSOR_NORMAL);
	*nameptr = NULL;
	return(namelist);
}



/*
 * name_fetch parent
 * 
 * INPUTS :	parent 		Parent for new dialog.
 * RETURNS:	None.
 */
static int
name_fetch(
		class name_list *r_list)
{
#define _L_BUFFERSIZE	256		// size of work buffer - price server
					// should use a string out larger
					// than this 
	FILE *ptr;			// pointer to file
	char buffer[_L_BUFFERSIZE];	// input area
	char *option;			// ptr to price server option string
	class name_node *item;
	class ttext tt;
	int starts[10];
	int ends[10];
	char fname1[1024];
	char fname2[1024];

	starts[0]=1; ends[0]=15;
	starts[1]=21; ends[1]=35;

	DialogCursor(DCURSOR_WAIT);
 	tt.set(24,2,6,20,starts,ends);

#ifdef NEVER
	option=XGetDefault(display,PROGRAM_NAME,XDEFAULT_NS);
	if (!option)
#endif
	option=DEFAULT_NS;
	strcpy(fname1,tmpnam(NULL));
	strcpy(fname2,tmpnam(NULL));

	tt.read_pagelist(option,fname1);

	sprintf(buffer,"sort %s | uniq > %s",fname1,fname2);

	if (system(buffer)) {
		DialogMessage("Failed to sort name list");
		unlink(fname1);
		unlink(fname2);
		return(-1);
	}

	if ((ptr=fopen(fname2,"r"))==NULL) {
		DialogMessage("Failed to load name list");
		unlink(fname1);
		unlink(fname2);
		return(-1);
	}
	unlink(fname1);
	unlink(fname2);
	while(fgets(buffer,_L_BUFFERSIZE,ptr)!=NULL) {
		char *c=buffer;
		int count=0;

		/*
		 * Share names must start with alphabetic characters
		 */
		if  ( !(isalpha(*c)) )
			continue;

		/*
		 * Assume all share names are less than 10 characters long
		 */
		while (   count<9 &&
			  (allowchar(*c) || 
			  (*c=='.') ||
			  ((*c==' ') && (allowchar(*(c+1)))) ||
			  (*c=='/') || (*c=='-') || (*c=='&') ) ) {
			c++;
			count++;
		}
		*c='\0';

		/*
		 * Ignore any name that appears to be less than
		 * 3 characters long
		 */
		if (count<3)
			continue;

		if (*buffer=='\0')
			continue;
		item=new class name_node;
		item->name=(char *)malloc(strlen(buffer)+1);
		strcpy(item->name,buffer);
		r_list->insert_tail(item);
		DialogCursor(DCURSOR_BUSY);
	}
	fclose(ptr);
	DialogCursor(DCURSOR_NORMAL);
	return(0);
#undef _L_BUFFERSIZE
}

void
name_list::create_securities( char **inlist,
	int number)
{
	int iter;
	char **al_names;
	int al_count;
	unsigned long secflags;
	sectype_e type_flags;
	class security *basesec=new security(SEC_FLAGS_DUMMY|SEC_FLAGS_AUTO|
						SEC_FLAGS_TRADED);
	class security *newsec;
	class Securityedit *editptr=new class Securityedit;
	editptr->SecurityeditEdit(basesec);

	//
	// Wait for security edit to complete
	//
	while (basesec->SecurityIslocked() )
		DialogEvent();

	basesec->SecurityExtract(NULL,&secflags,&al_names,&al_count,&type_flags);
	for(iter=0;iter<number;iter++)  {
		newsec=new security(SEC_FLAGS_DUMMY);
		newsec->SecurityCreate(inlist[iter],secflags,NULL,0,type_flags);
		newsec->add();
	}
	//
	// Save the revised list
	//
	main_sec_list->save();

}

enum buttone {
	BUTTON_NONE,
	BUTTON_OK,
	BUTTON_CANCEL,
	};
	

class Dialogsecnames : private Button, private Optionlist {

	Dialogshell d_shell;			// top most dialog
	Dialogvertical  d_mainvert;
	Dialoghorizontal d_buttonhoriz;		// button horizontal form
	Buttoninstance *d_ok; 
	Buttoninstance *d_cancel; 
	Dialoglist d_namelist;

	int buttonid;

	class name_list *share_name_list;
private:
	//
	// This overrides the pure virtual button handler in the Button class
	//
	virtual void ButtonCallback(unsigned long id,buttonstate_e flag);

public:
	Dialogsecnames();
	~Dialogsecnames();
};

void
Dialogsecnames::ButtonCallback(unsigned long id,buttonstate_e flag)
{
	char **list;
	int number;

	switch(id) {
	case BUTTON_OK:
		buttonid=BUTTON_OK;
		list=d_namelist.DialoglistGet(number);
		share_name_list->create_securities(list,number);
		freestringlist(list);
		break;
	case BUTTON_CANCEL: 
		buttonid=BUTTON_CANCEL;
		break;
	}
	delete this;
}

Dialogsecnames::Dialogsecnames()
{
	static Newdialoglist listdefn = {
		20,			// width
		10,			// height
		NULL,			// list of items
		"Securities",		// name of list
		NULL,			// default font
	};

	share_name_list=new class name_list;

	if (name_fetch(share_name_list)==-1)
		return;

	listdefn.listitem=share_name_list->extract();

	buttonid=BUTTON_NONE;

        d_shell.DialogshellCreate("ADD SECURITY NAMES");
        d_mainvert.DialogverticalCreate(d_shell);
	d_namelist.DialoglistCreate(d_mainvert,listdefn);

        d_buttonhoriz.DialoghorizontalCreate(d_mainvert);
        (void)ButtonPush(d_buttonhoriz,"OK - BULK CREATE",BUTTON_OK);
        (void)ButtonPush(d_buttonhoriz,"CANCEL",BUTTON_CANCEL);
	d_shell.DialogRealize();
}

Dialogsecnames::~Dialogsecnames()
{
}



static void
project_loadsec(local_va_dcl)
{
	va_alist=va_alist;	// stop compiler warnings
	class Dialogsecnames *secnames=new class Dialogsecnames;
}

enum menu_e {
	MENU_HELPXSM,
	MENU_HELPMENUS,
	};

static void
project_help(local_va_dcl)
{
	va_list ap;
	dialog d;
	unsigned long code;
	char *name;
	char buffer[2048];

	local_va_start(ap);
	d=va_arg(ap,dialog);
	code=va_arg(ap,unsigned long);

	switch(code) {
	case MENU_HELPXSM:
		name="xsm.html";
		break;
	case MENU_HELPMENUS:
		name="overview.html";
		break;
	default:
		return;
	}
	sprintf(buffer,"xsm_loadpage file://localhost%s/%s",DOCDIR,name);
	system(buffer);
}


void
project_init( menu_bar *mb)
{
	MENU("Project","Find known securities",project_loadsec,1);
	MENU("Project","Quit",project_exit,1);
	MENU("Help","On XSM",project_help,MENU_HELPXSM);
	MENU("Help","On Menus",project_help,MENU_HELPMENUS);
}
