
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: record.h,v 2.0 1995/06/24 15:05:47 davidsi Exp $
//
// $Log: record.h,v $
// Revision 2.0  1995/06/24  15:05:47  davidsi
// - Added GNU copyright message.
//
// Revision 1.1  1994/07/03  20:31:47  davidsi
// Initial revision
//
//
#ifndef __RECORD_H
#define __RECORD_H

#include <stdio.h>

//
// Types of record format
//
#define FORMAT_BY_LINE	1	// record are line orientated
#define FORMAT_BY_SIZE	2	// records are determine by a byte count
#define FORMAT_BY_PERC	3	// records terminate with %

//
// Flags for open()
//
#define OPEN_WRITE_LINE	1	// open file for writing by lines
#define OPEN_WRITE_SIZE	2	// open file for writing by size
#define OPEN_WRITE_PERC	3	// open file for writing by % delimeters
#define OPEN_READ	4	// open file for reads

//
// The record class
//
class record {
	char *string;		// ptr to start of actual input record
	char *text_ptr;		// current ptr to string in input record
	char *buffer;		// start of output buffer (copy of some
				// string in string)
	int length;		// number of bytes allocated for each string
	char *cpattern;		// compiled match pattern
	char *versionstring;	// version string read from file
	unsigned long format;	// file format (FORMAT_BY_XX)
	FILE *file;		// input/output file
private:
	void check_length(int len);
	int readbytes(char terminating_character,int flag);
public:
	record();
	~record();

	int open(char *,unsigned long);
	int close();
	int read();
	void write();

	char *version();

	void reset();
	char *line();
	int skipover(char *pattern);
	void create(char *);
	void append(char *);
	char *match();
	char *uptomatch();
	void compile(char *);
};

#endif
