
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: security.h,v 2.5 1995/06/07 20:43:49 davidsi Exp $
//
// $Log: security.h,v $
// Revision 2.5  1995/06/07  20:43:49  davidsi
// - More support for handling multiple portfolios. Use functions
//   in portfolio.C to handle security renaming and removal.
//
// Revision 2.4  1995/05/18  21:02:08  davidsi
// - Moved the declaration of portfolio to portfolio.h and now include
//   portfolio.h file.
// - The definitions of Holdinglist and Holding classes have been moved
//   to portfolio.C.
//
// Revision 2.3  1995/04/16  20:30:53  davidsi
// - Added a unit trust security type (SEC_TYPE_UNITTRUST).  This
//   required the but_types[] member of Securityedit to be increased in
//   size.
// - Added a changed_vect member, that provides callbacks when securities
//   are added, removed or modified.
// - Added HoldinglistRevise(), so that securities can be reposition on
//   the holding list when(if) their name changes.
//
// Revision 2.2  1995/04/03  20:38:05  davidsi
// - Added HoldingDividendhead() to the holding class.
// - Updated security::remove() to correctly location the appropriate
//   holding information.
// - Added a SecuritylistFindname() member to the Securitylist class. This
//   allows alias names to be used when searching for securities as well
//   as the key (prime) names.
// - BUG FIX. Added a new callback semaphore to the Securityedit to prevent
//   multiple callbacks - these could result in serious system failures as
//   many "are you sure" dialogs could be created.
//
// Revision 2.1  1995/03/08  21:24:05  davidsi
// - Factored the security class into two classes, the security class provides
//   all the main security handling functionality and the Securityedit class
//   allows securities to be edited.  This reduces the amount of memory
//   req'd to represent a security.
// - The SEC_TYPE_XX defines are now a enumerated types - function members
//   changed to reflect this.
// - Renamed extract() to SecurityExtract().
//
// Revision 2.0  1995/02/27  22:31:46  davidsi
// - The function member sec_compare() has become compare() overriding
//   the compare() in the inherited list class.
// - Many changes to support the new Dialog class, rather than issuing
//   direct calls to the c "dialog" functions.
//
// Revision 1.3  1994/11/05  21:11:01  davidsi
// - Added set() function to the security list class.  This allows
//   the flags to be modified.
// - Renamed typeof() to typename() so that no errors are generated
//   when compiling with g++.
// - Modified ~security_list() to be a virtual function so that g++
//   does not issue warnings.
//
// Revision 1.2  1994/10/18  20:01:01  davidsi
// - Added SEC_TYPE_XXX flags now that securitie types are handled solely
//   by the security class. This means typeof() no longer is a virtual
//   function - making dynamic type changes possible.
// - extract() can now extract security type information - therefore
//   has one further argument.
// - Securities are maintained using a sorted list, but to allow for
//   caching of names a security specific insert_order() function is
//   needed that invalidates the cache.
// - Added extract_action(), remove(), add() and set().
//
// Revision 1.1  1994/08/06  16:38:38  davidsi
// Initial revision
//
//
#ifndef __SECURITY_H
#define __SECURITY_H

#include <stdio.h>
#include <string.h>

#include "slist.h"
#include "record.h"
#include "trans.h"
#include "div.h"
#include "price.h"
#include "button.h"
#include "optionl.h"
#include "semaphore.h"
#include "callback.h"
#include "portfolio.h"


#define SECNAMELEN	256

//
// The security list class
//
class security_list : public slist {
	unsigned long flags;
public:
	class callback changed_vect;	// callbacks for when a security
                                        // is added, modified or deleted.

	class security *select_security(char *title);
	class security *find_security(char *);
	class security *SecuritylistFindname(char *secname);
	char **extract();
	void set(unsigned long newflags);
        virtual char *version();
	void insert_order(security *);
	virtual int compare(list_node *,list_node *);
	//
	// These save casting..
	//
	class security *head();

	security_list();
	virtual ~security_list();
	
};

#define SECLST_FLAGS_STALE	1	// Any cached names are now stale


//
// The securityedit class
//
class Securityedit : public Button,public Optionlist {
	//
	// The next group of fields specifies widget information
	// when securities are being created/modified etc
	//
	unsigned long action;			// see SEC_ACTION_XX 
	Optionlistinstance *d_namealias;	// the name alias dialog

	Dialogshell d_shell;			// main shell dialog 
	Dialogvertical d_mainvert;		// main vertical form
	Dialogform	d_form;
	Dialogverticalform d_vert1;
	Dialogverticalform d_vert2;
	Dialoghorizontal d_buttonhoriz;		// button horizontal form
	Dialogvertical d_namevert;		// vertical name form
	Dialoglabel  d_namelabel;		// name label
	Dialoginput  d_nameinput;		// name label
	//
	Dialogframe 	d_typeframe;	
	Dialogvertical 	d_typevert;	
	Dialoglabel 	d_typelabel;	
	Buttoninstance *but_types[3];
	//
	Dialogframe 	d_statusframe;	
	Dialogvertical 	d_statusvert;	
	Dialoglabel 	d_statuslabel;	
	Buttoninstance *but_status[2];
	//
	Dialogframe 	d_upframe;	
	Dialogvertical 	d_upvert;	
	Dialoglabel 	d_uplabel;	
	Buttoninstance *but_up[2];

	class security *secptr;		// the real security being edited
	//
	// Security locking
	//
	Semaphore callback;		// callback is handling security

	void addw_buttons();
	void addw_namealias();
	void addw_name();
	void addw_update_mode();
	void addw_status();
	void addw_type();

	//
	// Overrides the function in the Button class
	//
	virtual void ButtonCallback(unsigned long id,buttonstate_e flag);
	//
	// Overrides the function in the Optionlist class
	//
	virtual void OptionlistCallback(unsigned long id);
	//
	// Public parts of the structure
	//
public:
	void SecurityeditEdit(class security *);
	Securityedit();
	virtual ~Securityedit();
};

//
// Defines for action
//
#define SEC_ACTION_NONE		0	// user adding
#define SEC_ACTION_ADD		1	// user adding 
#define SEC_ACTION_ALT		2	// user altering
#define SEC_ACTION_DEL		3	// user deleting 
#define SEC_ACTION_SEL		4	// user selecting 

enum sectype_e {
	SEC_TYPE_SECURITY,
	SEC_TYPE_SHARE,
	SEC_TYPE_WARRANT,
	SEC_TYPE_GROUP,			// e.g a valuation or other
					// derived value
	SEC_TYPE_UNITTRUST,		// A unit trust
	};


//
// The security class
//
class security : public slist_node {
	unsigned long flags;		// see below defines
	char	name[SECNAMELEN];	// name of share 
	int	alias_count;		// how many alises 
	char	**aliases;		// position of alias array 
	sectype_e type_flags;	// type of security SEC_TYPE_XX

	//
	// Security locking
	//
	Semaphore edited;		// security is being edited

	//
	// Friends
	//

	friend class security_list;
	friend class Securityedit;
	friend class Holdinglist;

	//
	// Public parts of the structure
	//
public:
	class price *prices;			// price information

	security();
	~security();
	security(unsigned long);
	int SecurityIslocked();
	void add();
	void remove();
	int SecurityLoad(class record *);
	char *typename();
	void SecurityExtract(char **name,
		unsigned long *flgs,
		char ***alnms, 
		int *alcnt,
		sectype_e *);
	void SecurityCreate(char *c_name,
		unsigned long c_flags,
		char **c_names,
		int c_count,
		sectype_e c_typef);
	void SecuritySetname(char *);


	//
	// Inheiritated from the slist class
	// 
        virtual void save_record(class record *);

	//
	// These save casts...
	//
	class security *next();
};

//
// Defines for flags
//
#define SEC_FLAGS_TRADED	1	// security is traded
#define SEC_FLAGS_NOTTRADED	2	// security is not traded
#define SEC_FLAGS_MANUAL	4	// security price updated manually
#define SEC_FLAGS_AUTO		8	// security updated electronically
#define SEC_FLAGS_INSERTED	16	// on main securities list
#define SEC_FLAGS_DUMMY		32	// dummy security-will not be
					// saved OR placed on the securities
					// list


//
// Externally visible variables
//
extern security_list *main_sec_list;

#endif
