
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: semaphore.C,v 2.1 1995/06/24 15:01:04 davidsi Exp $
//
// $Log: semaphore.C,v $
// Revision 2.1  1995/06/24  15:01:04  davidsi
// - Added GNU copyright message.
//
// Revision 2.0  1995/03/02  21:16:36  davidsi
// - Initial version.
//
//

#include <stdio.h>
#include <varargs.h>
#define _STDARG_H
#include <stdlib.h>
#include <string.h>
#include <dial.h>
#include <math.h>

#include "localmacros.h"

#include "semaphore.h"
#include "dialog.h"


//
// SemaphoreLock
// 
// Attempts to obtain a semaphore. Blocks until access becomes possible.
// Note there is NO deadlock protection. Use this at your own risk!
//
// INPUTS :	None
// RETURNS:	None
//
void
Semaphore::SemaphoreLock()
{
	while (status==SEM_LOCKED) {
		DialogEvent();
	}
	// owner= getpid();
	status=SEM_LOCKED;
}


//
// SemaphoreClock
//
// Conditionally attempts to get a semaphore. 
//
// INPUTS :	None
// RETURNS:	-1 the semaphore is already locked. A call to
//		SemaphoreLock() will result in a block.
//		0 the semaphore was unlocked but is now locked
// 		on behalf of the caller.
//
int
Semaphore::SemaphoreClock()
{
	if (status==SEM_LOCKED)
		return(-1);

	SemaphoreLock();
	return(0);
}

//
// SemaphoreUnlock
//
// Makes a previously locked semaphore available.  Note NO checks are
// made on the current ownership of the semaphore, so you can unlock it
// even if you did not lock it.
//
// INPUTS :	None
// RETURNS:	None
//
void
Semaphore::SemaphoreUnlock()
{
	if (status!=SEM_LOCKED)
		DialogPanic("Semaphore::SemaphoreUnlock - Attempt to unlock an unlocked semaphore");
	//if ( owner != getpid() )
	//DialogPanic("Semaphore::SemaphoreUnlock - Attempt to unlock an unowned semamphore");
	
	//owner= -1;
	status=SEM_UNLOCKED;
}

//
// SemaphoreIslocked
//
// Returns the current state of a semaphore
//
// INPUTS :	None
// RETURNS:	TRUE if the semaphore is locked
//
int
Semaphore::SemaphoreIslocked()
{
	return( status==SEM_LOCKED );
}


//
// Constructor
//
Semaphore::Semaphore()
{
	status=SEM_UNLOCKED;
	owner= -1;
}

//
// Destructor
//
Semaphore::~Semaphore()
{
}
