
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: shares.C,v 2.3 1995/06/24 15:01:04 davidsi Exp $
//
// $Log: shares.C,v $
// Revision 2.3  1995/06/24  15:01:04  davidsi
// - Added GNU copyright message.
//
// Revision 2.2  1995/06/24  14:36:55  davidsi
// - Don't allow security groupings to be either deleted or modified.
//
// Revision 2.1  1995/06/02  19:59:33  davidsi
// - Change "Create Share" menu item to "Create Security".
//
// Revision 2.0  1995/03/08  21:22:22  davidsi
// - share_create() and share_modify() must now use the Securityedit class
//   rather than the security class when allowing the user to create securities.
//
// Revision 1.4  1995/03/02  21:08:22  davidsi
// - Replaced the menu_item declarations with the cleaner MENU macro.
//
// Revision 1.3  1994/11/05  21:10:05  davidsi
// - Various modifications to allow this file to compile using g++.
//
// Revision 1.2  1994/10/17  20:54:22  davidsi
// - Now that security type is handled by the security class remove the
//   virtual typeof() function. Also remove the share specific load
//   callback function - the security class now does this all for us.
//
// Revision 1.1  1994/07/03  20:30:16  davidsi
// Initial revision
//
//

#include <varargs.h>
#define _STDARG_H
#include <stdlib.h>
#include <stdarg.h>

#include "shares.h"
#include "menu.h"
#include "record.h"

#include "localmacros.h"

//
// share_create dialog args
//
// Creates a share
//
// INPUTS :     dialog          The parent dialog
//              args            The arguments needed to satisfy type checking
// RETURNS:     None
//
static void
share_create(local_va_dcl)
{
	share *newsec=new share;
	class Securityedit *editptr;

	va_alist=va_alist;	// stop compiler warnings
	editptr=new class Securityedit;
	editptr->SecurityeditEdit(newsec);
}

//
// share_modify dialog args
//
// Allows the details of ANY security (not just shares) to be modified.
//
// INPUTS :	dialog		The parent dialog
//		args		The arguments needed to satisfy type checking
// RETURNS:	None
//
static void
share_modify(local_va_dcl)
{
	security *sec;
	class Securityedit *editptr;
	sectype_e sectype;

	va_alist=va_alist;	// stop compiler warnings
	sec=main_sec_list->select_security("Modify Security");

	if (sec==NULL)
		return;

	//
	// Don't allow group type securities to be modified
	//
	sec->SecurityExtract(NULL,NULL,NULL,NULL,&sectype);
        if (sectype==SEC_TYPE_GROUP) {
                DialogMessage("This security cannot be modified");
                return;
        }
	editptr=new class Securityedit;
	editptr->SecurityeditEdit(sec);
}

//
// share_delete dialog args
//
// Allows the details of ANY security (not just shares) to be modified.
//
// INPUTS :	dialog		The parent dialog
//		args		The arguments needed to satisfy type checking
// RETURNS:	None
//
static void
share_delete(local_va_dcl)
{
	security *sec;
	sectype_e sectype;

	va_alist=va_alist;	// stop compiler warnings
	sec=main_sec_list->select_security("Delete Security");

	if (sec==NULL)
		return;

	//
	// Don't allow group type securities to be deleted
	//
	sec->SecurityExtract(NULL,NULL,NULL,NULL,&sectype);
        if (sectype==SEC_TYPE_GROUP) {
                DialogMessage("This security cannot be deleted");
                return;
        }

	sec->remove();
        main_sec_list->save();
}


//
// dshare_init
//
// Initialization function. Adds various menu options and the
// the share decode function for when securities are loaded. Note that
// this function adds general security handling menu options.
//
// INPUTS :     mb      The menu bar to be used
// RETURNS:     None
//
void
share_init( menu_bar *mb)
{
	MENU("Securities","Create Security",share_create,0);
	MENU("Securities","Modify Security",share_modify,0);
	MENU("Securities","Delete Security",share_delete,0);
}
