/* CHK=0xE646 */
/*+-------------------------------------------------------------------------
	kmemsio.c
	...!gatech!emory!tridom!wht

  Defined functions:
	kmem_init_tty()
	kmem_read_tty(ttfirst,ttcount)

--------------------------------------------------------------------------*/
/*+:EDITS:*/
/*:08-03-1993-15:51-wht@n4hgf-source control point x2.70 */
/*:07-15-1992-14:33-wht@n4hgf-2.60 release - u386mon+siotools merge */
/*:06-24-1992-20:19-wht@n4hgf-use nlsym in cahorts with u386mon */
/*:12-04-1989-16:05-wht-creation */

#include <sys/types.h>
#include <sys/stat.h>
#define TERMIO_HACK /* for XENIX termio.h multiple inclusion problem */
#include <sys/tty.h>

#include "kmemsio.h"

#if defined(M_UNIX) || defined(ISC)
char *kernel_file = "/unix";

#include <nlist.h>
#define XLIST nlist
#define XL_name n_name
#define XL_value n_value
#define XL_type n_type
#else
char *kernel_file = "/xenix";

#include <sys/a.out.h>
#define XLIST xlist
#define XL_name xl_name
#define XL_value xl_value
#define XL_type xl_type
#endif

#ifdef STANDALONE_NLIST
#if defined(M_UNIX)
struct XLIST kernel_symbols[] = {
#define KS_SIO_TTY	0
	{ "sio_tty" },
	{ (char *)0 }
};
#else /* !M_UNIX */

struct XLIST kernel_symbols[] = {
#define KS_SIO_TTY	0
	{0,0,0, "_sio_tty" },
	{0,0,0, (char *)0 }
};
#endif /* M_UNIX */

#define	sio_ttyaddr kernel_symbols[KS_SIO_TTY].XL_value

#else  /* !STANDALONE_NLIST */
#undef M_TERMINFO
#include "../config.h"
#include "../nlsym.h"
#include "../libnlsym.h"
#endif /* STANDALONE_NLIST */

char *kmem_file = "/dev/kmem";

int kmemfd;

struct tty sio[SIO_NTTY];

static char errmsg[80];

extern int errno;
extern char *sys_errlist[];

/*+-------------------------------------------------------------------------
	kmem_read_tty(ttfirst,ttcount)
--------------------------------------------------------------------------*/
char *
kmem_read_tty(ttfirst,ttcount)
int ttfirst;
int ttcount;
{
int len = ttcount * sizeof(struct tty);

	if(((unsigned)ttfirst >= SIO_NTTY) ||
		(((unsigned)ttfirst + (unsigned)ttcount) > SIO_NTTY))
		return(" illegal tty ");

	if(lseek(kmemfd,sio_ttyaddr + (ttfirst * sizeof(struct tty)),0) == -1)
	{
		sprintf(errmsg,
			" kmem lseek failure: %s ",sys_errlist[errno]);
		return(errmsg);
	}
	errno = 0;
	if(read(kmemfd,(char *)sio,len) != len)
	{
		sprintf(errmsg,
			" kmem read failure: %s ",sys_errlist[errno]);
		return(errmsg);
	}

	return((char *)0);

}	/* end of kmem_read_tty */

/*+-------------------------------------------------------------------------
	kmem_init_tty()
--------------------------------------------------------------------------*/
char *
kmem_init_tty()
{
	char sio_ttyname[64];

	strncpy(sio_ttyname,nlsym[Nindex_SIO_TTY].XL_name,
		sizeof(sio_ttyname));
	sio_ttyname[sizeof(sio_ttyname) - 1] = 0;


	if((kmemfd = open(kmem_file,0)) < 0)
	{
		sprintf(errmsg," %s: %s ",kmem_file,sys_errlist[errno]);
		return(errmsg);
	}

#ifdef STANDALONE_NLIST
	if(XLIST(kernel_file,kernel_symbols))
	{
		sprintf(errmsg," xlist/nlist failure: %s ",kernel_file);
		return(errmsg);
	}

#else
	nlsym_read();
	if(!sio_ttyaddr)
	{
		sprintf(errmsg," %s not found in this kernel", sio_ttyname);
		return(errmsg);
	}
#endif

	return((char *)0);

}	/* end of kmem_init_tty */

/*+-------------------------------------------------------------------------
	leave_text(text,code)
--------------------------------------------------------------------------*/
#ifndef STANDALONE_NLIST
void
leave_text(text,code)
char *text;
int code;
{
	disp_cmd_legend(text);
	bye(1);
}	/* end of leave_text */

#endif

/* vi: set tabstop=4 shiftwidth=4: */
/* end of kmemsio.c */
