/* CHK=0xBED5 */
char *revision = "2.10";
/*+-------------------------------------------------------------------------
	siomon.c -- watch XENIX/UNIX serial I/O
	...!gatech!kd4nc!n4hgf!wht
	XENIX debugging help by Jim Stratton (stratton@hpda.da.com)

    Sample output:
   000000000011111111112222222222333333333344444444445555555555666666666677777
   012345678901234567890123456789012345678901234567890123456789012345678901234
00  siomon 1.34 (UNIX V.3.2/i386) n4hgf           interval:  1        21:54:54
01 
02  tty  raw  can   out   speed  state iflag  oflag  cflag  lflag  pgrp
03  ---  ---  ---  -----  -----  ----- ------ ------ ------ ------ -----
04  1a     0    0      0   9600  OC     10045      0   6655      0     0
05  1b
06  1c
07  1d
08  1e
09  1f
10  1g
11  1h
12  2a
13  2b
14  2c
15  2d     0    0      0   9600  W          0      0   2275      0     0
16  2e
17  2f     0    0      0   4800  OC     10040      0   2374      0     0
18  2g     0    0      0   4800  OC     10005      0   6374      0     0
19  2h
20 
21 State: W waiting for open to complete  O open  C carrier on
22        S stopped by XOFF  T interval timeout in progress
23 Commands: + inc interval  - dec interval  ^L refresh  q quit  d detail

Under XENIX, termio.h is included twice (once by curses.h/tcap.h and
again by sys/tty.h.  You need to bracket the termio.h under XENIX with
  #if !defined(TERMIO_HACK) || (defined(TERMIO_HACK) && !defined(IOCTYPE))
  ...
  #endif

  Defined functions:
	bye(sig)
	detail()
	disp_cmd_legend(text)
	disp_interval()
	disp_state_legend()
	disp_tod()
	disp_tty(y,sionum,tty)
	main(argc,argv,envp)
	main_template()
	my_resetty()
	my_savetty()
	strcmpi(s1,s2)
	to_lower(ch)
	to_upper(ch)

  Kudos to pyrnj.pyramid.com!romain (Romain Kang) for many suggestions
  regarding screen and CPU efficiency in the original itpmon.c for Pyramid
  Thanks to mpxenix!news (Randy Bush) for tireless help with a curses
  bug hopefully now retired, even if it crockish.
--------------------------------------------------------------------------*/
/*+:EDITS:*/
/*:08-03-1993-15:51-wht@n4hgf-source control point x2.70 */
/*:08-10-1992-07:32-wht@n4hgf-put state legend on detail screen too */
/*:07-15-1992-14:33-wht@n4hgf-2.60 release - u386mon+siotools merge */
/*:06-25-1992-02:37-root@n4hgf-integrate with u386mon's nlsym */
/*:06-22-1992-13:28-wht@n4hgf-revisit for 3.2v4 */
/*:02-23-1990-17:10-wht-more keyboard problems with UNIX: try typeahead(-1) */
/*:02-18-1990-03:29-wht-bug in curses tty mode handling - force a fix */
/*:12-28-1989-13:35-wht-change rev - B600 added to wintty */
/*:12-21-1989-15:20-wht-version number chg - CTSFLOW/RTSFLOW */
/*:12-04-1989-16:45-wht-add more to detail */
/*:11-28-1989-19:49-wht-no more sprintf and allow nap/rdchk in lieu of select */
/*:11-22-1989-12:33-wht-rehost my Pyramid itpmon.c */

/* -------- select(S) availability --------- */
/* define HAVE_SELECT */ /* uncomment for XENIX/386 2.3.2 and later */
#if (defined(M_UNIX) && (!defined(HAVE_SELECT))) /* automatic for UNIX */
#define HAVE_SELECT
#endif

#ifndef SYSV
#define SYSV
#endif

#include <curses.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/utsname.h>
#include <time.h>

#if defined(HAVE_SELECT)
#if defined(ISC)
#include <sys/bsdtypes.h>
#else
#include <sys/select.h>
#endif
#endif

#include <sys/tty.h>

#include "wintty.h"
#define DEFINE_TTYNM
#include "kmemsio.h"
#include "utoa.h"

#define HY 2	/* header line */
#define TX 1
#define RX 6
#define CX 11
#define OX 16
#define SX 23
#define FX 30

extern char _sobuf[];
extern int errno;
extern char *sys_errlist[];

int no_tod = 0;
int interval_secs = 1;
struct utsname me;
char bell = 7;

/*+-------------------------------------------------------------------------
	workaround curses tty handling difficulties among different revs
--------------------------------------------------------------------------*/
struct termio my_termio_crock;
void my_savetty()	/* used in lieu of savetty() */
{
	ioctl(0,TCGETA,&my_termio_crock);
}
void my_resetty()	/* used in lieu of resetty() and endwin() */
{
	ioctl(0,TCSETA,&my_termio_crock);
}

/*+-------------------------------------------------------------------------
	bye() - terminate curses, reset terminal and exit
--------------------------------------------------------------------------*/
void
bye(sig)
int sig;
{
	if(stdscr)
	{
		standend();
		move(LINES - 1,0);
		refresh();
		my_resetty();
	}
	exit(0);
}	/* end of bye */

/*+-------------------------------------------------------------------------
	disp_tty(y,sionum,tty) - display info for 1 tty
--------------------------------------------------------------------------*/
void
disp_tty(y,sionum,tty)
int y;
int sionum;
register struct tty *tty;
{
	register unsigned int itmp;
	register opened = tty->t_state & (ISOPEN | WOPEN);
	char s8[8];

	move(y,TX + 0);
	printw("%-4.4s",ttynm[sionum]);

	if(!opened)
	{
		clrtoeol();
		return;
	}

	if((itmp = (unsigned)tty->t_rawq.c_cc) > 999)
		itmp = 999;
	utoda(s8,3,itmp);
	move(y,RX + 0);
	if(itmp > 10)
		standout();
	addstr(s8);
	if(itmp > 10)
		standend();

	if((itmp = (unsigned)tty->t_canq.c_cc) > 999)
		itmp = 999;
	utoda(s8,3,itmp);
	move(y,CX + 0);
	addstr(s8);

	if((itmp = (unsigned)tty->t_outq.c_cc + tty->t_tbuf.c_count) > 99999)
		itmp = 99999;
	utoda(s8,5,itmp);
	move(y,OX + 0);
	addstr(s8);

	move(y,SX + 0);
	addstr(B_to_baud_rate(tty->t_cflag & CBAUD));

	strcpy(s8,"     ");
	if(tty->t_state & WOPEN)
		s8[0] = 'W';
	else if(tty->t_state & ISOPEN)
		s8[0] = 'O';
	if(tty->t_state & CARR_ON)
		s8[1] = 'C';
	if(tty->t_state & BUSY)
		s8[2] = 'B';
	if(tty->t_state & TTSTOP)
		s8[3] = 'S';
	if(tty->t_state & TIMEOUT)
		s8[4] = 'T';
	move(y,FX);
	addstr(s8);

	utooa(s8,7,tty->t_iflag);
	addstr(s8);

	utooa(s8,7,tty->t_oflag);
	addstr(s8);

	utooa(s8,7,tty->t_cflag);
	addstr(s8);

	utooa(s8,7,tty->t_lflag);
	addstr(s8);

	utoda(s8,6,tty->t_pgrp);
	addstr(s8);

}	/* end of disp_tty */

/*+-------------------------------------------------------------------------
	disp_interval() - display current display interval
--------------------------------------------------------------------------*/
void
disp_interval()
{
	char dmsg[12];

	move(0,47);
	addstr("interval: ");
	utoda(dmsg,2,interval_secs);
	addstr(dmsg);
}	/* end of disp_interval */

/*+-------------------------------------------------------------------------
	disp_state_legend() - display "State: W ......"
--------------------------------------------------------------------------*/
void
disp_state_legend()
{

	move(LINES - 4,0);
	addstr(
"State: W waiting for open to complete  O open  C carrier on");
	move(LINES - 3,0);
	addstr(
"       S stopped by XOFF  T interval timeout in progress  B transmitter busy");
}	/* end of disp_state_legend */

/*+-------------------------------------------------------------------------
	disp_tod() - display time of day
--------------------------------------------------------------------------*/
void
disp_tod()
{
	register struct tm *lt;		/* local time */
	struct tm *localtime();
	long now;
	char buf[10];

	if(no_tod)
		return;

	time(&now);
	lt = localtime(&now);
	utoda_lz(buf,2,lt->tm_hour);
	buf[2] = ':';
	utoda_lz(buf + 3,2,lt->tm_min);
	buf[5] = ':';
	utoda_lz(buf + 6,2,lt->tm_sec);
	move(0,COLS - 13);
	addstr(buf);
}	/* end of disp_tod */

/*+-------------------------------------------------------------------------
	disp_cmd_legend(text) - display command legend
--------------------------------------------------------------------------*/
void
disp_cmd_legend(text)
char *text;
{
	move(LINES - 2,0);
	clrtoeol();

	if(!text)
	{
		move(LINES - 2,0);
		addstr("Commands: ");
		standout();
		addstr("ESC");
		standend();
		addstr("/");
		standout();
		addstr("q");
		standend();
		addstr(" quit   ");
		standout();
		addstr("d");
		standend();
		addstr(" detail on tty   ");
		standout();
		addstr("+");
		standend();
		addstr("/");
		standout();
		addstr("-");
		standend();
		addstr(" change interval   ");
		standout();
		addstr("^L");
		standend();
		addstr(" refresh");
	}
	else
		addstr(text);

}	/* end of disp_cmd_legend */

/*+-------------------------------------------------------------------------
	main_template() - draw main screen template
--------------------------------------------------------------------------*/
void
main_template()
{
	static char *header  = 
		" tty  raw  can   out   speed  state iflag  oflag  cflag  lflag  pgrp";
	static char *hyphens =
		" ---  ---  ---  -----  -----  ----- ------ ------ ------ ------ -----";
	FILE *fp = fopen("/etc/systemid","r");
	char sysid[64];

	wclear(stdscr);
	move(0,0);
	standout();
	printw(" siomon %s (%s V.%s/%s)",
		revision,
#if defined(M_UNIX) || defined(ISC)
		"UNIX",
#else
		"XENIX",
#endif
		me.release,me.machine);
	if(fp)
	{
		sysid[0] = 0;
		fgets(sysid,sizeof(sysid),fp);
		if(sysid[0])
		{
			sysid[strlen(sysid) - 1] = 0;
			printw(" %s ",sysid);
		}
		fclose(fp);
	}
	standend();
	move(HY,0);
	addstr(header);
	move(HY + 1,0);
	addstr(hyphens);
	disp_state_legend();
	disp_cmd_legend((char *)0);
	move(LINES - 1,0);
	disp_interval();
	refresh();
}	/* end of main_template */

#ifdef SIO_IS_FAS
#define  STRCMP strcmp
#else
#define  STRCMP strcmpi
/*+-------------------------------------------------------------------------
    to_upper() / to_lower()

One would think that these were relatively standard types of
thing, but System V specifies toupper() to convert to upper case
if not already and BSD says to adjust without testing, so, two
stupid little routines here.  ASCII only -- no EBCDIC gradoo here please.
--------------------------------------------------------------------------*/
char to_upper(ch)
register char ch;
{ return( ((ch >= 'a') && (ch <= 'z')) ? ch - 0x20 : ch);
}   /* end of to_upper() */

char to_lower(ch)
register char ch;
{ return( ((ch >= 'A') && (ch <= 'Z')) ? ch + 0x20 : ch);
}   /* end of to_lower() */

/*+-------------------------------------------------------------------------
	strcmpi(s1,s2) - case-insensitive strcmp

This version of strcmp() is case-insensitive and works like a sane one
should, per strcmp(3), not per  the K&R1 example or POSIX/ANSI.

In here rather than ecuutil.c since other executables besides ecu
uses this module and strcmpi needed there too 
--------------------------------------------------------------------------*/
int
strcmpi(s1,s2)
register char *s1;
register char *s2;
{

    while(*s1)
	{
        if(to_upper(*s1++) != to_upper(*s2++))
		{
			s1--;
			s2--;
            break;
		}
	}
	return(to_upper(*s1) - to_upper(*s2));

}	/* end of strcmpi */
#endif /* SIO_IS_FAS */

/*+-------------------------------------------------------------------------
	detail() - detail screen handler
--------------------------------------------------------------------------*/
void
detail()
{
	int y,x;
	int sionum;
	char tty_name[32];
	char cmd;
	char **cpptr;
	char delim;
	int cmd_available;
#if defined(HAVE_SELECT)
	struct timeval timeout;
	int readfds;
#else
	long ltimeout;
	long nap();
#endif

	disp_cmd_legend("display detail on /dev/tty");
	getyx(stdscr,y,x);

	if(wingets(stdscr,y,x,tty_name,5,&delim,0,(int *)0) < 0)
		return;
	for(sionum = 0,cpptr = ttynm; sionum < SIO_NTTY; sionum++,cpptr++)
	{
		if(!STRCMP(*cpptr,tty_name))
			break;
	}
	if(sionum == SIO_NTTY)
	{
		write(2,&bell,1);
		return;
	}

	for(y = 0; y < (LINES - 4); y++)
	{
		move(HY + 2 + y,0);
		clrtoeol();
	}
	disp_state_legend();
	disp_cmd_legend("");
	move(LINES - 2,0);
	addstr("Commands: ");
	standout();
	addstr("ESC");
	standend();
	addstr("/");
	standout();
	addstr("q");
	standend();
	addstr(" main display   ");
	standout();
	addstr("+");
	standend();
	addstr("/");
	standout();
	addstr("-");
	standend();
	addstr(" change interval");

	move(HY + 2,0);
	clrtoeol();
	wintty_template(stdscr,HY + 2 + 2,0,1);
	while(1)
	{
		disp_tod();
		kmem_read_tty(sionum,1);
		disp_tty(HY + 2,sionum,&sio[0]);
		wintty(stdscr,HY + 2 + 2,0,&sio[0]);
		move(HY + 2 + 11,0);
		refresh();
#if defined(HAVE_SELECT)
		readfds = 1;			/* for standard input */
		timeout.tv_sec  = interval_secs;
		timeout.tv_usec = (interval_secs)
								? 0
								: ((sio[0].t_state & ISOPEN)
									? 20*1000L
									: 200*1000L);
		cmd_available = (select(32,&readfds,0,0,&timeout) > 0);
#else
		ltimeout = (interval_secs)
						? (interval_secs * 1000L)
						: ((sio[0].t_state & ISOPEN) ? 20L : 100L);
		while(ltimeout > 0)
		{
			ltimeout -= nap((interval_secs) ? 20L : 100L);
			if(cmd_available = rdchk(0))
				break;
		}
#endif
		if(rdchk(0))
		{
			cmd = getch();
			if((cmd == 0x1B) || (cmd == 'q'))
				break;
			switch(cmd)
			{
				case '+':
					interval_secs++;
					disp_interval();
					break;

				case '-':
					if(!interval_secs)
						break;
					interval_secs--;
					disp_interval();
					break;
			}
		}
	}
}	/* end of detail */

/*+-------------------------------------------------------------------------
	main(argc,argv,envp)
--------------------------------------------------------------------------*/
main(argc,argv,envp)
int argc;
char **argv;
char **envp;
{
	register sionum;
	char *cptr;
	int cmd_available;
#if defined(HAVE_SELECT)
	struct timeval timeout;
	int readfds;
#else
	long ltimeout;
	long nap();
#endif

	if((argc > 1) && !strcmp(argv[1],"-n"))
		no_tod = 1;

	signal(SIGINT,bye);
	signal(SIGTERM,bye);
	uname(&me);

	setbuf(stdout,_sobuf);
	my_savetty();
	if(!initscr())
	{
		fputs("curses init failed ... check terminal type\n",stderr);
		exit(1);
	}
	typeahead(-1);
	cbreak();
	noecho();
	leaveok(stdscr,1);

	main_template();

	move(4,0);
	standout();
	addstr(" Reading kernel symbol table ");
	standend();
	refresh();

	if(cptr = kmem_init_tty())
	{
		move(5,0);
		standout();
		addstr(cptr);
		standend();
		bye(0);
	}

	move(4,0);
	clrtoeol();
	refresh();

	while(1)
	{
		kmem_read_tty(0,SIO_NTTY);
#if defined(SIO_IS_FAS)
		for(sionum = 0; sionum < (SIO_NTTY / 2); sionum++)
		{
			int yy = HY + 2 + sionum;
			if(yy == (LINES - 5))
			{
				move(yy,0);
				clrtoeol();
				addstr("not enough room for all ttys");
				break;
			}
			if (sio[sionum].t_state & ISOPEN)
				disp_tty(yy,sionum,&sio[sionum]);
			else
				disp_tty(yy,sionum + (SIO_NTTY / 2),
						&sio[sionum + (SIO_NTTY / 2)]);
		}
#else
		for(sionum = 0; sionum < SIO_NTTY; sionum++)
		{
			int yy = HY + 2 + sionum;
			if(yy == (LINES - 5))
			{
				move(yy,0);
				clrtoeol();
				addstr("not enough room for all ttys");
				break;
			}
			disp_tty(yy,sionum,&sio[sionum]);
		}
#endif
		disp_tod();
		move(LINES - 1,0);
		refresh();

#if defined(HAVE_SELECT)
		readfds = 1;			/* for standard input */
		timeout.tv_sec  = interval_secs;
		timeout.tv_usec = (interval_secs) ? 0 : 100*1000L;
		cmd_available = (select(32,&readfds,0,0,&timeout) > 0);
#else
		ltimeout = (interval_secs) ? (interval_secs * 1000L) : 100L;
		while(ltimeout > 0)
		{
			ltimeout -= nap(100L);
			if(cmd_available = rdchk(0))
				break;
		}
#endif

		if(cmd_available)
		{
		char ch;
			ch = getch() & 0x7F;
			switch(ch)
			{
			case 'L' & 0x1F:		/* redraw screen */
			case 'R' & 0x1F:		/* redraw screen */
				main_template();
				break;

			case '+':
				interval_secs++;
				disp_interval();
				break;

			case '-':
				if(!interval_secs)
					break;
				interval_secs--;
				disp_interval();
				break;

			case 'q':		/* quit */
			case 0x1B:
				bye(0);
				break;

			case 'd':		/* detail */
				detail();
				main_template();
				refresh();
				break;

			}
		}
	}
}	/* end of main */

/* vi: set tabstop=4 shiftwidth=4: */
/* end of siomon.c */
