/* CHK=0x2504 */
char *revision = "2.11";
/*+-------------------------------------------------------------------------
	uusnap.c - display UUCP communications status for HDB systems
    ...gatech!kd4nc!n4hgf!wht

  Many ideas in this program came from uustatus.c by Ed Carp

  Defined functions:
	basename(fullname)
	bye(sig)
	datetime5(secs)
	detail()
	disp_cmd_legend(text)
	display_status(sysnum,system_name,y,status_info)
	display_tod()
	display_tty(sionum,tty)
	get_status_info(system_name,buf,bufsize)
	lockpid_to_tty(lockpid)
	main(argc,argv)
	my_resetty()
	my_savetty()
	lockdirname(fname)
	statdirname(fname)
	strcmpi(s1,s2)
	to_lower(ch)
	to_upper(ch)

Under XENIX, termio.h is included twice (once by curses.h/tcap.h and
again by sys/tty.h.  You need to bracket the termio.h under XENIX with
  #if !defined(TERMIO_HACK) || (defined(TERMIO_HACK) && !defined(IOCTYPE))
  ...
  #endif

Sample output:

uusnap 1.70 (UNIX V.3.2/i386) n4hgf                          19:28:26

00000000001111111111222222222233333333334444444444555555555566666
01234567890123456789012345678901234567890123456789012345678901234
 # SYSTEM    CALLS OUTQ INPQ LAST  NEXT  PID   STATUS
-- ------    ----- ---- ---- ----- ----- ----- ------------------------
 1 kd4nc        2    3       15:22 15:22       successful
 2 n4vu                      19:07 19:07       successful
 3 tridom                 1  19:17 19:17       TALKING (tty2e)
 4 ki4xo                     01/25 01/25       successful


 Thanks to mpxenix!news (Randy Bush) for tireless help with a curses
 bug hopefully now retired, even if it crockish.
--------------------------------------------------------------------------*/
/*+:EDITS:*/
/*:08-03-1993-15:51-wht@n4hgf-source control point x2.70 */
/*:12-07-1992-15:49-wht@n4hgf-explicit standend after error in kmem_init_tty */
/*:07-15-1992-14:33-wht@n4hgf-2.60 release - u386mon+siotools merge */
/*:06-25-1992-02:38-root@n4hgf-integrate with u386mon's nlsym */
/*:06-22-1992-13:28-wht@n4hgf-revisit for 3.2v4 */
/*:02-23-1990-17:10-wht-more keyboard problems with UNIX: try typeahead(-1) */
/*:02-18-1990-03:29-wht-bug in curses tty mode handling - force a fix */
/*:02-07-1990-16:34-wht-use crmode and nocrmode rather than raw and noraw */
/*:01-27-1990-19:11-wht-show input and output queues + cosmetics */
/*:12-28-1989-13:35-wht-change rev - B600 added to wintty */
/*:12-26-1989-14:53-wht-datetime5 month off by 1 */
/*:12-07-1989-03:10-wht-make changes per email suggestions */
/*:12-04-1989-16:45-wht-add detail in full */
/*:12-01-1989-16:53-wht-XENIX 286 work: no beep()! */
/*:11-30-1989-17:48-wht-complete rewrite to add detail status */

#ifndef SYSV
#define SYSV
#endif

#include <curses.h>
#include <ctype.h>
#include <memory.h>
#include <signal.h>
#include <time.h>
#include <sys/types.h>
#include <dirent.h>
#include <sys/stat.h>
#include <sys/tty.h>
#include <sys/utsname.h>
#if defined(HAVE_SELECT)
#if defined(M_SYSV)
#include <sys/select.h>
#else
#if defined(ISC)
#include <sys/bsdtypes.h>
#else
#include <sys/time.h>
#endif
#endif
#endif

#include "wintty.h"
#define DEFINE_TTYNM
#include "kmemsio.h"
#include "utoa.h"

#define FIRSTSYS_Y	4
#define NAME_X		4
#define CALLS_X		14
#define OUTQ_X		19
#define INPQ_X		24
#define DTAQ_X		29
#define LAST_X		34
#define NEXT_X		40
#define PID_X		46
#define STATUS_X	52

/* display_tty x offsets */
#define TX 0
#define RX 5
#define CX 10
#define OX 15
#define SX 22
#define FX 29

extern char _sobuf[];
extern int errno;
extern char *sys_errlist[];

DIR *stat_dp;
DIR *work_dp;
DIR *uspool_dp;

long secs_now;
int systems;
int no_tod = 0;
char sysnames[64 * 32];	/* sorta hack, but ka-plenty name space */
char locked_tty[16];
char bell = 7;

char *STATUSDIR = "/usr/spool/uucp/.Status";
#if defined(ISC)
char *LOCKDIR = "/usr/spool/locks";
#else
char *LOCKDIR = "/usr/spool/uucp";
#endif
char *WORKDIR = "/usr/spool/uucp/";

#define MAX_SYSTEMS (LINES - FIRSTSYS_Y - 4)
#define WAITSECS_ACTIVE 1 /* sleep secs between samples when talking */
#define WAITSECS_IDLE   7 /* sleep secs between samples when not talking */

#define SS_OK                    0 	/* successful */
#define SS_NO_DEVICE             1 	/* no device */
#define SS_TIME_WRONG            2 	/* wrong time to call */
#define SS_TALKING               3 	/* TALKING */
#define SS_CONVERSATION          4 	/* conversation failed */
#define SS_SEQBAD                5 	/* bad sequence check */
#define SS_LOGIN_FAILED          6 	/* login failed */
#define SS_DIAL_FAILED           7 	/* dial failed */
#define SS_BAD_LOG_MCH           8 	/* bad login/machine */
#define SS_LOCKED_DEVICE         9 	/* DEVICE LOCKED */
#define SS_ASSERT_ERROR          10	/* assert error */
#define SS_BADSYSTEM             11	/* system not in Systems */
#define SS_CANT_ACCESS_DEVICE    12	/* can't access device */
#define SS_DEVICE_FAILED         13	/* device failed */
#define SS_WRONG_MCH             14	/* wrong machine name */
#define SS_CALLBACK              15	/* callback required */
#define SS_RLOCKED               16	/* remote has lock for me */
#define SS_RUNKNOWN              17	/* remote does not know me */
#define SS_RLOGIN                18	/* remote reject after login */
#define SS_UNKNOWN_RESPONSE      19	/* remote reject, unknown msg */
#define SS_STARTUP               20	/* startup failed */
#define SS_CHAT_FAILED           21	/* caller script failed */
#if defined(M_SYSV) || defined(ISC)
#define SS_CALL_IN_PROGRESS	     22	/* CALL IN PROGRESS */
#define SS_CALL_FAILED           23	/* call failed (busy?) */
#else
#define SS_CALL_IN_PROGRESS	     22	/* CALL IN PROGRESS */
#endif

struct utsname me;
int uucico_active;

char *errortext[] = {
	/*       00000000001111111111222222 */
	/*       01234567890123456789012346 */
	/* 0 */ "successful",
	/* 1 */ "no device",
	/* 2 */ "wrong time to call",
	/* 3 */ "TALKING",
	/* 4 */ "conversation failed",
	/* 5 */ "bad sequence check",
	/* 6 */ "login failed",
	/* 7 */ "dial failed",
	/* 8 */ "bad login/machine",
	/* 9 */ "DEVICE LOCKED",
	/* 10*/ "assert error",
	/* 11*/ "system not in Systems",
	/* 12*/ "can't access device",
	/* 13*/ "device failed",
	/* 14*/ "wrong machine name",
	/* 15*/ "callback required",
	/* 16*/ "remote has lock for me",
	/* 17*/ "remote does not know me",
	/* 18*/ "remote reject after login",
	/* 19*/ "remote reject, unknown msg",
	/* 20*/ "startup failed",
	/* 21*/ "caller script failed",
#if defined(M_SYSV) || defined(ISC)
	/* 22*/ "CALL IN PROGRESS",
	/* 23*/ "call failed (busy?)",
#else
	/* 22*/ "CALL IN PROGRESS",
#endif
};

#if defined(M_SYSV) || defined(ISC)
#define SS_MSG_MAX 23
#else
#define SS_MSG_MAX 22
#endif

/*+-------------------------------------------------------------------------
	workaround curses tty handling difficulties among different revs
--------------------------------------------------------------------------*/
struct termio my_termio_crock;
void my_savetty()	/* used in lieu of savetty() */
{
	ioctl(0,TCGETA,&my_termio_crock);
}
void my_resetty()	/* used in lieu of resetty() and endwin() */
{
	ioctl(0,TCSETA,&my_termio_crock);
}

/*+-------------------------------------------------------------------------
	datetime5(secs) - return 5 char date or time string

needs global 'secs_now' to reflect current time in "seconds since epoch"
(done by display_tod()); returns 5 character 'mm/dd' if time now is more
than 12 hours before now or 12 hours from now
--------------------------------------------------------------------------*/
char *
datetime5(secs)
long secs;
{
	register long delta = secs_now - secs;
	struct tm *lt = localtime(&secs);
	static char dt5[8];

	if(delta < 0)
		delta = -delta;

	if(delta > (12L * 3600))
	{
		utoda_lz(dt5,2,lt->tm_mon + 1);
		dt5[2] = '/';
		utoda_lz(dt5 + 3,2,lt->tm_mday);
	}
	else
	{
		utoda_lz(dt5,2,lt->tm_hour);
		dt5[2] = ':';
		utoda_lz(dt5 + 3,2,lt->tm_min);
	}

	return(dt5);

}	/* end of datetime5 */

/*+-------------------------------------------------------------------------
	basename(fullname) - strip directory name from filename

returns address of static string
--------------------------------------------------------------------------*/
char *
basename(fullname)
char *fullname;
{
	register char *start;
	static char outstr[256];
	char *strrchr();

	start = strrchr(fullname,'/'); /* find last slash */
	if(!start)
		return(fullname);
	start++;
	strcpy(outstr,start);
	return(outstr);
}	/* end of basename */

/*+-------------------------------------------------------------------------
	statdirname(fname) - return base name prepended with status dir

returns address of static string
--------------------------------------------------------------------------*/
char *
statdirname(fname)
char *fname;
{
	static char fullname[128];
	static int fullname_cat_point = 0;

	if(!fullname_cat_point)
	{
		strcpy(fullname,STATUSDIR);
		strcat(fullname,"/");
		fullname_cat_point = strlen(fullname);
	}

	strcpy(fullname + fullname_cat_point,fname);
	return(fullname);

}	/* end of statdirname */

/*+-------------------------------------------------------------------------
	lockdirname(fname) - return base name prepended with lock dir

returns address of static string
--------------------------------------------------------------------------*/
char *
lockdirname(fname)
char *fname;
{
	static char fullname[128];
	static int fullname_cat_point = 0;

	if(!fullname_cat_point)
	{
		strcpy(fullname,LOCKDIR);
		strcat(fullname,"/");
		fullname_cat_point = strlen(fullname);
	}

	strcpy(fullname + fullname_cat_point,fname);
	return(fullname);

}	/* end of lockdirname */

/*+-------------------------------------------------------------------------
	display_tod()
--------------------------------------------------------------------------*/
void
display_tod()
{
	register struct tm *lt;		/* local time */
	struct tm *localtime();
	char buf[20];

	(void)time(&secs_now);

	if(no_tod)
		return;

	lt = localtime(&secs_now);
	utoda_lz(buf,2,lt->tm_mon + 1);
	buf[2] = '/';
	utoda_lz(buf + 3,2,lt->tm_mday);
	buf[5] = '/';
	utoda_lz(buf + 6,2,lt->tm_year);
	buf[8] = ' ';
	utoda_lz(buf + 9,2,lt->tm_hour);
	buf[11] = ':';
	utoda_lz(buf + 12,2,lt->tm_min);
	buf[14] = ':';
	utoda_lz(buf + 15,2,lt->tm_sec);
	move(0,COLS - 26);
	addstr(buf);

}	/* end of display_tod */

/*+-------------------------------------------------------------------------
	bye(sig) - exit cleanly
--------------------------------------------------------------------------*/
void
bye(sig)
int sig;
{
	if(stdscr)
	{
		standend();
		move(LINES - 1,0);
		clrtobot();
		refresh();
		echo();
		nocbreak();
		move(LINES - 1,0);
		refresh();
		my_resetty();
	}
	exit(0);
}	/* end of bye */

/*+-------------------------------------------------------------------------
	lockpid_to_tty(lockpid) - given pid, find first ttyname it has locked
--------------------------------------------------------------------------*/
char *
lockpid_to_tty(lockpid)
int lockpid;
{
	struct dirent *dp;
	FILE *fp;
	int testpid;
	static char rtnname[32];

	seekdir(uspool_dp,0);
	while(dp = readdir(uspool_dp))
	{
		if(strncmp(dp->d_name,"LCK..tty",8))
			continue;
		if(fp = fopen(lockdirname(dp->d_name),"r"))
		{
			fscanf(fp,"%d",&testpid);
			fclose(fp);
			if(testpid == lockpid)
			{
				strcpy(rtnname,dp->d_name + 5);
				return(rtnname);
			}
		}
	}
	return((char *)0);

}	/* end of lockpid_to_tty */

/*+-------------------------------------------------------------------------
	disp_cmd_legend(text)
--------------------------------------------------------------------------*/
disp_cmd_legend(text)
char *text;
{
	move(LINES - 2,0);
	clrtoeol();

	if(!text)
	{
		addstr("Commands: ");
		standout();
		addstr("ESC");
		standend();
		addstr("/");
		standout();
		addstr("q");
		standend();
		addstr(" quit   ");
		standout();
		addstr("d");
		standend();
		addstr(" detail    ");
		standout();
		addstr("SPACE");
		standend();
		addstr(" immediate update");
	}
	else
		addstr(text);

}	/* end of disp_cmd_legend */

/*+-------------------------------------------------------------------------
	get_status_info(system_name,buf,bufsize)
--------------------------------------------------------------------------*/
int
get_status_info(system_name,buf,bufsize)
char *system_name;
char *buf;
int bufsize;
{
	FILE *fp;

	if(!(fp = fopen(statdirname(system_name),"r")))
		return(-1);
	buf[0] = 0;
	fgets(buf,bufsize,fp);
	fclose(fp);
	if(!buf[0])
		return(-1);
	buf[strlen(buf) - 1] = 0;
	return(0);

}	/* end of get_status_info */

/*+-------------------------------------------------------------------------
	display_status(sysnum,system_name,y,status_info)

called with status_info either .Status/<system_name> line or null
if a tty is locked on the line, global locked_tty has "ttyxx"
returns -1 if status_info null and can't get it, else 0
--------------------------------------------------------------------------*/
display_status(sysnum,system_name,y,status_info)
int sysnum;
char *system_name;
int y;
char *status_info;
{
	int itmp;
	int retry_count;
	int status;
	int input_queue_count;
	int output_queue_count;
	int data_queue_count;
	int poll_seen;
	char *stattxt;
	int locking_pid;
	char *ttyname;
	char lock_name[64];
	char s32[32];
	char work_dir[64];
	char linebuf[512];
	long secs_last_try;
	long secs_next_try;
	struct dirent *dp;
	FILE *fp;

	if(!status_info)
	{
		status_info = linebuf;
		if(get_status_info(system_name,status_info,sizeof(linebuf)))
			return(-1);
	}

	sscanf(status_info,"%d %d %ld %ld",
	    &status,&retry_count,&secs_last_try,&secs_next_try);
	secs_next_try += secs_last_try;	/* files has secs til next retry */

	strcpy(work_dir,WORKDIR);
	strcat(work_dir,system_name);

	input_queue_count = 0;
	output_queue_count = 0;
	data_queue_count = 0;
	poll_seen = 0;

	if(work_dp = opendir(work_dir))
	{
		/* count the work files */
		while(dp = readdir(work_dp))
		{
			if(!strncmp(dp->d_name,"X.",2) && (input_queue_count <= 99))
				input_queue_count++;
			if(!strncmp(dp->d_name,"C.",2) && (output_queue_count <= 99))
			{
				output_queue_count++;
				itmp = strlen(dp->d_name);
				if((itmp > 5) && !strcmp(dp->d_name + itmp - 4,"0000"))
					poll_seen = 1;
			}
			if(!strncmp(dp->d_name,"D.",2) && (data_queue_count <= 99))
				data_queue_count++;
			if((input_queue_count > 99) && (output_queue_count > 99))
				break;
		}
		closedir(work_dp);
	}

	/* "kludges for screwy status stuff with HDB" thanks to Ed Carp */
	strcpy(lock_name,lockdirname("LCK.."));
	strcat(lock_name,system_name);
	if(!access(lock_name,0) && (status != SS_TALKING))
		status = SS_CALL_IN_PROGRESS;
#if 0/* not so sure about this one any more */
	if(status == SS_BADSYSTEM)
		status = SS_DEVICE_FAILED;
#endif

	move(y,0);
	clrtoeol();
	utoda(s32,2,sysnum);
	addstr(s32);
	addch(' ');

	strcpy(s32,system_name);
	if(strlen(s32) > 10)
		s32[10] = 0;
	addstr(s32);

	move(y,CALLS_X);
	if(!retry_count)
		addstr("   ");
	else
	{
		if(retry_count > 999)
			retry_count = 999;
		utoda(s32,3,retry_count);
		addstr(s32);
	}

	move(y,INPQ_X);
	if(!input_queue_count)
		addstr("   ");
	else
	{
		if(input_queue_count > 99)
			addstr(">99");
		else
		{
			utoda(s32,3,input_queue_count);
			addstr(s32);
		}
	}

	move(y,OUTQ_X);
	if(!output_queue_count)
		addstr("   ");
	else
	{
		if((output_queue_count == 1) && poll_seen)
			addstr("Poll");
		else
		{
			if(output_queue_count > 99)
				strcpy(s32,">99");
			else
				utoda(s32,3,output_queue_count);
			if(poll_seen)
				strcat(s32,"P");
			addstr(s32);
		}
	}

	move(y,DTAQ_X);
	if(!data_queue_count)
		addstr("   ");
	else
	{
		if(data_queue_count > 99)
			addstr(">99");
		else
		{
			utoda(s32,3,data_queue_count);
			addstr(s32);
		}
	}

	move(y,LAST_X);
	addstr(datetime5(secs_last_try));

	move(y,NEXT_X);
	addstr(datetime5(secs_next_try));

	locking_pid = 0;
	memset(locked_tty,0,sizeof(locked_tty));
	move(y,PID_X);
	addstr("     ");
	ttyname = (char *)0;
	if( (status == SS_TALKING) ||
		(status == SS_LOCKED_DEVICE ) ||
		(status == SS_CALL_IN_PROGRESS))
	{
		uucico_active = 1;
		if(fp = fopen(lock_name,"r"))
		{
			fscanf(fp,"%d",&locking_pid);
			fclose(fp);
		}
		if(locking_pid)
		{
			strcpy(locked_tty,ttyname = lockpid_to_tty(locking_pid));
			move(y,PID_X);
			utoda(s32,5,locking_pid);
			addstr(s32);
		}
		standout();
	}

	move(y,STATUS_X);
	if(status > SS_MSG_MAX)
	{
		stattxt = s32;
		strcpy(stattxt,"status ");
		utoda_lz(stattxt + 7,3,status);
	}
	else
		stattxt = errortext[status];
	addstr(stattxt);
	if(ttyname)
	{
		addstr(" (");
		addstr(ttyname);
		addch(')');
	}
	standend();
	itmp = 27 - strlen(stattxt);
	while(itmp--)
		addch(' ');
	return(0);
}	/* end of display_status */

/*+-------------------------------------------------------------------------
	display_tty(sionum,tty)
--------------------------------------------------------------------------*/
void
display_tty(sionum,tty)
int sionum;
register struct tty *tty;
{
	register int x;
	register int y;
	register unsigned itmp;
	char s8[8];

	clrtoeol();
	getyx(stdscr,y,x);
	move(y,TX + x);
	printw("%-4.4s",ttynm[sionum]);
	if((itmp = (unsigned)tty->t_rawq.c_cc) > 999)
		itmp = 999;
	utoda(s8,3,itmp);
	move(y,RX + x);
	if(itmp > 10)
		standout();
	addstr(s8);
	if(itmp > 10)
		standend();

	if((itmp = (unsigned)tty->t_canq.c_cc) > 999)
		itmp = 999;
	utoda(s8,3,itmp);
	move(y,CX + x);
	addstr(s8);

	if((itmp = (unsigned)tty->t_outq.c_cc + tty->t_tbuf.c_count) > 99999)
		itmp = 99999;
	utoda(s8,5,itmp);
	move(y,OX + x);
	addstr(s8);

	move(y,SX + x);
	addstr(B_to_baud_rate(tty->t_cflag & CBAUD));

	strcpy(s8,"     ");
	if(tty->t_state & WOPEN)
		s8[0] = 'W';
	else if(tty->t_state & ISOPEN)
		s8[0] = 'O';
	if(tty->t_state & CARR_ON)
		s8[1] = 'C';
	if(tty->t_state & BUSY)
		s8[2] = 'B';
	if(tty->t_state & TTSTOP)
		s8[3] = 'S';
	if(tty->t_state & TIMEOUT)
		s8[3] = 'T';
	move(y,FX + x);
	addstr(s8);

	utooa(s8,7,tty->t_iflag);
	addstr(s8);

	utooa(s8,7,tty->t_oflag);
	addstr(s8);

	utooa(s8,7,tty->t_cflag);
	addstr(s8);

	utooa(s8,7,tty->t_lflag);
	addstr(s8);

	utoda(s8,6,tty->t_pgrp);
	addstr(s8);

}	/* end of display_tty */

#ifdef SIO_IS_FAS
#define  STRCMP strcmp
#else
#define  STRCMP strcmpi
/*+-------------------------------------------------------------------------
    to_upper() / to_lower()

One would think that these were relatively standard types of
thing, but System V specifies toupper() to convert to upper case
if not already and BSD says to adjust without testing, so, two
stupid little routines here.  ASCII only -- no EBCDIC gradoo here please.
--------------------------------------------------------------------------*/
char to_upper(ch)
register char ch;
{ return( ((ch >= 'a') && (ch <= 'z')) ? ch - 0x20 : ch);
}   /* end of to_upper() */

char to_lower(ch)
register char ch;
{ return( ((ch >= 'A') && (ch <= 'Z')) ? ch + 0x20 : ch);
}   /* end of to_lower() */

/*+-------------------------------------------------------------------------
	strcmpi(s1,s2) - case-insensitive strcmp

This version of strcmp() is case-insensitive and works like a sane one
should, per strcmp(3), not per  the K&R1 example or POSIX/ANSI.

In here rather than ecuutil.c since other executables besides ecu
uses this module and strcmpi needed there too 
--------------------------------------------------------------------------*/
int
strcmpi(s1,s2)
register char *s1;
register char *s2;
{

    while(*s1)
	{
        if(to_upper(*s1++) != to_upper(*s2++))
		{
			s1--;
			s2--;
            break;
		}
	}
	return(to_upper(*s1) - to_upper(*s2));

}	/* end of strcmpi */
#endif /* SIO_IS_FAS */

/*+-------------------------------------------------------------------------
	detail()
--------------------------------------------------------------------------*/
void
detail()
{
	int y,x;
	int sysdetail = 0;
	int ttnum;
	int have_tty = 0;
	char *sysname;
	char *cptr;
	char **cpptr;
	char delim;
	char s8[8];
#if defined(HAVE_SELECT)
	struct timeval timeout;
	int fdmask;
#endif

	static char *header  =
	"tty  raw  can   out   speed  state iflag  oflag  cflag  lflag  pgrp";
	static char *hyphens =
	"---  ---  ---  -----  -----  ----- ------ ------ ------ ------ -----";
	static char *press_any_key =
	"press any key to return to main display  ";
	static char *state1 =
	"State: W waiting for open to complete  O open  C carrier on";
	static char *state2 =
	"       S stopped by XOFF  T delay timeout in progress  B transmitter busy";
	static kmem_initialized = 0;

	disp_cmd_legend("display detail on system # (ESC to cancel): ");
	getyx(stdscr,y,x);
	if(wingets(stdscr,y,x,s8,3,&delim,0,(int *)0)< 0)
		goto SKIP_OUT;
	sysdetail = atoi(s8);
	sysdetail--;
	if(!x || ((unsigned)sysdetail > (unsigned)systems))
	{
		write(2,&bell,1);
SKIP_OUT:
		disp_cmd_legend((char *)0);
		refresh();
		return;
	}

	if(!kmem_initialized)
	{
		standout();
		disp_cmd_legend(" Reading kernel symbol table ");
		standend();
		refresh();
		if(cptr = kmem_init_tty())
		{
			move(3,0);
			clrtoeol();
			standout();
			disp_cmd_legend(cptr);
			standend();
			refresh();
			bye(255);
		}
		kmem_initialized = 1;
	}

	sysname = sysnames + (sysdetail << 4);
	y = FIRSTSYS_Y;
	move(y,0);
	clrtobot();
	disp_cmd_legend(press_any_key);

	while(1)
	{
		display_tod();	/* display time and get 'secs_now' */
		if(display_status(sysdetail + 1,sysname,FIRSTSYS_Y,(char *)0))
		{
			move(y,0);
			standout();
			printw("cannot get status for %s",sysname);
			standend();
			clrtobot();
			disp_cmd_legend(press_any_key);
			goto DETAIL_SLEEP;
		}

							/* 00000 */
							/* 01234 */
							/* ttyxx */
		move(y+5,0);

		if(locked_tty[0])
		{
			for(ttnum = 0,cpptr = ttynm; ttnum < SIO_NTTY; ttnum++,cpptr++)
			{
				if(!STRCMP(*cpptr,locked_tty + 3))
					break;
			}
			if(ttnum == SIO_NTTY)
				goto NO_TTY;
		
			if(cptr = kmem_read_tty(ttnum,1))
			{
#if defined(SIO_IS_FAS) && !defined(ISC)
BAD_NEWS:
#endif
				standout();
				addstr(cptr);
				addstr(": ");
				addstr(locked_tty);
				standend();
				refresh();
				getch();
				goto DETAIL_EXIT;
			}
#if defined(SIO_IS_FAS) && !defined(ISC)
			if(!(sio[0].t_state & (WOPEN | ISOPEN)))
			{
				char other_tty[64];
				strcpy(other_tty,locked_tty);
				cptr = other_tty + strlen(other_tty) - 1;
				if(isupper(*cptr))
					*cptr = tolower(*cptr);
				else if(islower(*cptr))
					*cptr = toupper(*cptr);
				else
					goto NO_TTY;
				for(ttnum = 0,cpptr = ttynm; ttnum < SIO_NTTY; ttnum++,cpptr++)
				{
					if(!strcmp(*cpptr,other_tty + 3))
						break;
				}
				if(ttnum == SIO_NTTY)
					goto NO_TTY;
			
				if(cptr = kmem_read_tty(ttnum,1))
					goto BAD_NEWS;
			}
#endif /* SIO_IS_FAS && !ISC */
			if(!have_tty)
			{
				move(y+3,0);
				addstr(header);
				move(y+4,0);
				addstr(hyphens);
				move(y+7,0);
				addstr(state1);
				move(y+8,0);
				addstr(state2);
			}
			have_tty = 1;
			move(y+5,0);
			display_tty(ttnum,&sio[0]);
			wintty_template(stdscr,y+10,0,1);
			wintty(stdscr,y+10,0,&sio[0]);
			move(y+10,0);
		}
		else
		{
NO_TTY:
			move(y+3,0);
			clrtobot();
			move(y+5,0);
			printw(" no tty or non-async connection");
			disp_cmd_legend(press_any_key);
			have_tty = 0;
			wintty_template(stdscr,y+10,0,0);
			move(LINES - 1,0);
		}

DETAIL_SLEEP:
		refresh();
		if(rdchk(0))
		{
			getch();
			break;
		}
#if defined(HAVE_SELECT)
		timeout.tv_sec = 0;
		timeout.tv_usec = ((have_tty) ? 20L : 500L) * 1000L;
		fdmask = 1 << 0;
		select(32,&fdmask,(int *)0,(int *)0,&timeout);
#else
		nap((have_tty) ? 20L : 500L);
#endif
	}

DETAIL_EXIT:
	move(y = FIRSTSYS_Y,0);
	clrtobot();
	disp_cmd_legend((char *)0);
	refresh();

}	/* end of detail */

/*+-------------------------------------------------------------------------
	main(argc,argv)
--------------------------------------------------------------------------*/
main(argc,argv)
int argc;
char **argv;
{
	int display_delta_y;
	char s32[32];
	char linebuf[512];
	char *sysname = "";
	char *dirname = "";
	char cmd;
	FILE *fp;
	struct dirent *dp;
	int initializing = 1;
#if defined(HAVE_SELECT)
	struct timeval timeout;
	int fdmask;
#else
	long timeout;
#endif

	if((argc > 1) && !strcmp(argv[1],"-n"))
		no_tod = 1;

	setbuf(stdout,_sobuf);

	my_savetty();
	if(!initscr())
	{
		fprintf(stderr,"curses init failed... check terminal type\n");
		exit(0);
	}
#if defined(M_TERMINFO)
	typeahead(-1);
#endif
	noecho();
	crmode();
	leaveok(stdscr,1);

	signal(SIGTERM,bye);
	signal(SIGINT,bye);
	signal(SIGHUP,bye);

	uname(&me);
	disp_cmd_legend((char *)0);

	move(0,0);
	standout();
	printw(" uusnap %s (%s V.%s/%s)",
	    revision,
#if defined(M_UNIX) || defined(ISC)
	    "UNIX",
#else
	    "XENIX",
#endif
	    me.release,me.machine);
	if(fp = fopen("/etc/systemid","r"))
	{
		s32[0] = 0;
		fgets(s32,sizeof(s32),fp);
		if(s32[0])
		{
			s32[strlen(s32) - 1] = 0;
			printw(" %s ",s32);
		}
		fclose(fp);
	}
	standend();


	move(2,0);
	addstr(
" # SYSTEM    CALLS OUTQ INPQ DATA LAST  NEXT  PID   STATUS");
	move(3,0);
	addstr(
"-- ------    ----- ---- ---- ---- ----- ----- ----- ------------------------");
	move(4,0);
	standout();
	addstr(" initializing ");
	standend();
	refresh();

	dirname = LOCKDIR;
	if(!(uspool_dp = opendir(dirname)))
		goto BOO_HISS;

	dirname = statdirname(".");
	if(!(stat_dp = opendir(dirname)))
	{
BOO_HISS:
		move(6,0);
		printw("%s: %s",dirname,sys_errlist[errno]);
		move(7,0);
		addstr("I don't think HoneyDanBer UUCP is installed.");
		bye(255);
	}

	while(1)
	{
		uucico_active = 0;		/* assume no uucico_active connections */
		display_tod();	/* display time and get 'secs_now' */

		seekdir(stat_dp,0);
		display_delta_y = systems = 0;
		move(FIRSTSYS_Y,0);
		while(dp = readdir(stat_dp))
		{
			if(display_delta_y > MAX_SYSTEMS)
				break;
			if(*dp->d_name == '.')
				continue;
			if(get_status_info(dp->d_name,linebuf,sizeof(linebuf)))
				continue;

			sysname = sysnames + (systems++ << 4);
			strcpy(sysname,dp->d_name);

			if(initializing)
			{
				move(4,0);
				addstr("              ");
				initializing = 0;
			}
			display_status(systems,sysname,display_delta_y + FIRSTSYS_Y,
				linebuf);
			display_delta_y++;
		}
		while(display_delta_y < MAX_SYSTEMS)
		{
			move(display_delta_y++ + FIRSTSYS_Y,0);
			clrtoeol();
		}

		if(initializing)
		{
			move(3,0);
			addstr("              ");
			initializing = 0;
		}
		move(LINES - 1,0);
		refresh();

#if defined(HAVE_SELECT)
		timeout.tv_sec = (uucico_active) ? WAITSECS_ACTIVE : WAITSECS_IDLE;
		timeout.tv_usec = 0;
		fdmask = 1 << 0;
		select(32,&fdmask,(int *)0,(int *)0,&timeout);
#else
		timeout = ((uucico_active) ? WAITSECS_ACTIVE : WAITSECS_IDLE) * 1000L;
		while(timeout > 0)
		{
			timeout -= nap(200L);
			if(rdchk(0))
				break;
		}
#endif

		if(rdchk(0))
		{
			cmd = getch() & 0x7f;
			switch(cmd)
			{
				case 'q':
				case 0x1B:
					bye(0);
				case 'd':
					detail();
					break;
				case ' ':
					break;
				default:
					write(2,&bell,1);
					break;
			}
		}
	}
}	/* end of main */

/* vi: set tabstop=4 shiftwidth=4: */
/* end of uusnap.c */
