/* CHK=0xF82A */
/*+-------------------------------------------------------------------------
	wingets.c - curses get string
	wht@n4hgf.Mt-Park.GA.US

  Defined functions:
	clear_area_char(win,y,x,len,fillchar)
	wingets(win,y,x,buf,bufsize,delim,edit,pwgpos)

--------------------------------------------------------------------------*/
/*+:EDITS:*/
/*:08-03-1993-15:51-wht@n4hgf-source control point x2.70 */
/*:07-15-1992-14:33-wht@n4hgf-2.60 release - u386mon+siotools merge */
/*:06-22-1992-16:00-wht@n4hgf-creation */
#ifndef SYSV
#define SYSV
#endif

#include <curses.h>
#include <string.h>
#include <memory.h>

#define mem_cpy memcpy	/* if ever executed, needs to avoid overlap */

/*+-------------------------------------------------------------------------
	clear_area_char(win,y,x,len,fillchar)
--------------------------------------------------------------------------*/
void
clear_area_char(win,y,x,len,fillchar)
WINDOW *win;
int y;
int x;
int len;
char fillchar;
{
	wmove(win,y,x);
	while(len-- > 0)
		waddch(win,fillchar & 0xFF);
	wmove(win,y,x);

}	/* end of clear_area_char */

/*+-------------------------------------------------------------------------
	wingets(win,y,x,buf,bufsize,delim,edit,pwgpos)

This procedure reads a string from win and returns the number
of characters read.

If edit is non-zero and pwgpos is not null, the inital string
position is set by dereferencing the pointer.

The terminating delim is returned in 'delim'.

If pwgpos is not null, the ending string position is returned in
the integer pointed to.

-1 is returned if an ESCape is typed by the keyboard user,
otherwise the count of characters in the string.

The entire line must be contained on one line (no line wrap supported).
--------------------------------------------------------------------------*/
int
wingets(win,y,x,buf,bufsize,delim,edit,pwgpos)
WINDOW *win;
int y;
register x;
register char *buf;
int bufsize;	/* includes room for null..field is 1 less */
char *delim;
int edit;
int *pwgpos;
{
	register count = 0;
	register pos = 0;
	int insert_mode = 0;
	int rtn_val = 0;
	char bell = 7;

	leaveok(win,0);
	bufsize--;
	clear_area_char(win,y,x,bufsize,'_');
	if(edit)
	{
		waddstr(win,buf);
		count = pos = strlen(buf);
		if(pwgpos)
		{
			pos = *pwgpos;
			if((pos < 0) || (pos > count))
				pos = count;
		}
	}
	else
		*buf = 0;
	wmove(win,y,x + pos);

	while(1)
	{
		wrefresh(win);
		*delim = wgetch(win) & 0x7F;
		if((*delim < 0x20) || (*delim >= 0x7F))
		{
			switch(*delim)
			{
				case 0x0D:
					*delim = 0x0A;;
				case 0x0A:
					wrefresh(win);
					rtn_val = count;
					goto RETURN;

				case 0x08:
					if(count)
					{
						if(count == pos)
						{
							*(buf + --count) = 0;
							wmove(win,y,x + count);
							waddch(win,'_');
							wmove(win,y,x + count);
							pos--;
						}
						else
						{
							if(!pos)
								continue;
							mem_cpy(buf + pos - 1,buf + pos,count - pos);
							*(buf + --count) = 0;
							wmove(win,y,x + --pos);
							waddstr(win,buf + pos);
							waddch(win,'_');
							wmove(win,y,x + pos);
						}
					}
					continue;

				case 0x1B:
					rtn_val = -1;
					goto RETURN;

				case 'U' & 0x1F:
					clear_area_char(win,y,x,bufsize,'_');
					count = 0;
					pos = 0;
					*buf = 0;
					continue;

				default:
					write(2,&bell,1);
					continue;

			}	/* end of switch(*delim) */
			/*NOTREACHED*/
		}		/* end of if read delimiter */

		if(count == bufsize)
		{
			write(2,&bell,1);
			continue;
		}

		if(insert_mode && (pos != count))
		{
			waddch(win,*delim);
			waddstr(win,buf + pos);
			mem_cpy(buf + pos + 1,buf + pos,count - pos);
			*(buf + pos++) = *delim;
			*(buf + ++count) = 0;
			wmove(win,y,x + pos);
		}
		else
		{
			waddch(win,*delim);
			*(buf + pos) = *delim;
			if(pos == count)
				*(buf + ++count) = 0;
			pos++;
		}
	}	/* end of while can get character */

RETURN:
	if(pwgpos)
		*pwgpos = pos;
	leaveok(win,1);
	return(rtn_val);

}	/* end of wingets */

/* vi: set tabstop=4 shiftwidth=4: */
/* end of wingets.c */
