
/*
 *    X SHARE MANAGER (XSM)
 *
 *    Copyright (C) 1995  David Simons
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * $Id: d_button.c,v 2.1 1995/06/06 21:46:51 davidsi Exp $
 *
 * $Log: d_button.c,v $
 * Revision 2.1  1995/06/06  21:46:51  davidsi
 * - Added GNU copyright message.
 *
 * Revision 2.0  1995/01/31  21:56:31  davidsi
 * - BUG FIX. In dialog_tbutton() the toggle button dialog could be referenced
 *   before it had been created if mutual exclusion was being requested.
 *
 * Revision 1.3  1994/07/14  20:17:32  davidsi
 * - BUG FIX. Many of the globally visible functions were incorrectly
 *   typed.
 *
 * Revision 1.2  1994/07/03  21:06:45  davidsi
 * - BUG FIX.  In dialog_button_value set the integer value that is
 *   returned to zero prior to reading the button. The Motif libraries
 *   may only set single bits so if the value is not zero strange values
 *   are returned back to callers of this function.
 *
 * Revision 1.1  1993/01/17  16:27:01  davidsi
 * Initial revision
 *
 *
 */

/*
 * Include Files
 */
#include <X11/Intrinsic.h>
#include <X11/Shell.h>

#include <Xm/Xm.h>
#include <Xm/Frame.h>
#include <Xm/PushB.h>
#include <Xm/ToggleB.h>

#define IN_DIALOGS

#include "d_base.h"
#include "d_button.h"
#include "d_form.h"

/*
 * Variable Declarations 
 */
static char rcsid[]="$Id: d_button.c,v 2.1 1995/06/06 21:46:51 davidsi Exp $";


/*
 * dialog_pbutton_c widget pb_dialog_ptr
 *
 * Handles the callback from a push button
 *
 * INPUTS :	widget		The initiating widget
 *		pb_dialog_ptr	Pointer to the related push button dialog
 *				structure (comes in as the client data!)
 * RETURNS:	None
 */
static void
dialog_pbutton_c(w,pbutton)
Widget w;				/* widget id */
d_pbutton *pbutton;			/* data from application */
{
	if (pbutton->callback)
		pbutton->callback(pbutton,pbutton->rvalue);
}



/*
 * dialog_pbutton parent, newbutton, position
 *
 * Creates a push button dialog.
 *
 * INPUTS :	parent		The parent dialog
 *		newbutton	The new button structure
 *		position	Describes position of button in parent dialog
 * RETURNS:	A pointer to the newly created dialog.	
 */
dialog
dialog_pbutton(parent,nb,p)
dialog parent;
new_d_pbutton *nb;
position *p;
{
	d_pbutton *pbutton;
	Arg al[10]; 		      /* arg list */
	int ac;

	pbutton=(d_pbutton *)dialog_init(parent,sizeof(d_pbutton));
	ac=0;
	if (p)
		dialog_position(p,al,&ac);
	pbutton->widget=XmCreatePushButton(parent->widget,nb->name,al,ac);
	XtAddCallback(pbutton->widget,XmNarmCallback,dialog_pbutton_c,pbutton);
	pbutton->callback=nb->callback;
	pbutton->rvalue=nb->value;
	return((dialog)pbutton);
}

/*
 * dialog_tbutton_c widget tb_dialog_ptr
 *
 * Handles the callback from a toggle button
 *
 * INPUTS :	widget		The initiating widget
 *		pb_dialog_ptr	Pointer to the related toggle button dialog
 *				structure (comes in as the client data!)
 * RETURNS:	None
 */
static void
dialog_tbutton_c(w,tbutton)
Widget w;			/* widget id */
d_tbutton *tbutton;		/* data from application */
{
	if (tbutton->callback)
		tbutton->callback(tbutton,tbutton->rvalue);
}



/*
 * dialog_button_set tbutton value
 *
 * Sets a toggle button to a given state.
 *
 * INPUTS :	tbutton		The toggle button whose state is to be set
 *		value		TRUE/FALSE value to trun button on/off
 * RETURNS:	-1 if error.
 */
int
dialog_button_set(d,value)
dialog d;
{
	d_tbutton *tbutton=(d_tbutton *)d;	/* satisfy type checking */
	Arg al[2];
	int val;

	if (value)
		XtSetArg(al[0],XmNset,TRUE);
	else
		XtSetArg(al[0],XmNset,FALSE);

	XtSetValues(tbutton->widget,al,1);
	return(0);
}



/*
 * dialog_button_value tbutton
 *
 * Returns the state of a toggle button.
 *
 * INPUTS :	tbutton		The toggle button in question
 * RETURNS:	TRUE/FALSE if button is on/off.
 */
int
dialog_button_value(d)
dialog d;
{
	d_tbutton *tbutton=(d_tbutton *)d;
	Arg al[2];				/* arg list */
	int val=0;

	XtSetArg(al[0],XmNset,&val);	/* default is Active */
	XtGetValues(tbutton->widget,al,1);
	return(val);
}


/*
 * dialog_tbutton parent, newtbutton, position
 *
 * Creates a toggle button dialog
 *
 * INPUTS :	parent		The parent dialog
 *		nf		The new toggle button structure
 *		position	Describe position of button in parent dialog
 * RETURNS:	A pointer to the newly created dialog.	
 */
dialog
dialog_tbutton(parent,nb,p)
dialog parent;
new_d_tbutton *nb;
position *p;
{
	d_tbutton *tbutton;
	Arg al[10]; 				      /* arg list */
	int i;

	/*
	 * Pray parent is a row column!
	 */
	if (nb->flags & DIALOG_TBUTTON_MUTEX) {
		XtSetArg(al[0],XmNradioBehavior,1);
		XtSetValues(parent->widget,al,1);
	}


	tbutton=(d_tbutton *)dialog_init(parent,sizeof(d_tbutton));
	if (nb->ivalue) {
		XtSetArg(al[0],XmNset,TRUE);        /* default is Active */
		i=1;
	} else
		i=0;
	if (p)
		dialog_position(p,al,&i);
	tbutton->widget=XmCreateToggleButton(parent->widget,nb->name,al,i);
	XtAddCallback(tbutton->widget,XmNarmCallback,dialog_tbutton_c,tbutton);
	tbutton->callback=nb->callback;
	tbutton->rvalue=nb->value;

	return((dialog)tbutton);
}
