
/*
 *    X SHARE MANAGER (XSM)
 *
 *    Copyright (C) 1995  David Simons
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * $Id: d_colour.c,v 2.2 1995/06/06 21:55:17 davidsi Exp $
 *
 * $Log: d_colour.c,v $
 * Revision 2.2  1995/06/06  21:55:17  davidsi
 * - Added GNU copyright message.
 *
 * Revision 2.1  1995/02/25  21:10:15  davidsi
 * - fetchcolour now also returns the bitmask, so it may be used by the
 *   caller.
 *
 * Revision 2.0  1995/01/31  21:57:29  davidsi
 * - All previous code disposed of.  This module now provides a
 *   colour cache to speed up performance.
 *
 * Revision 1.2  1994/07/03  21:11:17  davidsi
 * - Reset string to empty string when prompting for program name
 *
 * Revision 1.1  1994/02/09  22:09:30  davidsi
 * Initial revision
 *
 *
 */
#include <stdio.h>
#include <X11/Intrinsic.h>
#include <X11/Shell.h>

#include <Xm/Xm.h>
#include <Xm/DialogS.h>
#include <Xm/MainW.h>

#define IN_DIALOGS  1

#include "d_text.h"
#include "d_form.h"
#include "d_base.h"

#include "d_base.p"

#ifdef ALLVISIBLE

extern struct d_root  *base_node;

/*
 * Private mapping information for colours
 */
struct colmap {
	struct colmap *next;
	char *name;
	long pixelvalues;
};

static struct colmap *cm=NULL; 


/*
 * fetchcolour colourname gc
 *
 * Looks up a colour for use by a particular graphics context
 *
 * INPUTS :	colourname		The name of the colour
 *		gc			The graphics context
 * RETURNS:	The pixel values for the colour  - the foreground
 *		information in the gc will also be set as a side
 *		effect.
 */ 
int
fetchcolour(name,gc)
char *name;
GC gc;
{
	Colormap cmap;
	XColor exact;
	XColor colour;
	long pixelvalues;
	Display *display=base_node->display;
	struct colmap *newcm;

	/*
	 * See if the colour is already known
	 */
	newcm=cm;
	while (newcm!=NULL) {
		if ( !strcmp(name,newcm->name) )
			goto found;
		newcm=newcm->next;
	}

	/*
	 * Colour is new, so go and allocate it
	 */
	cmap=DefaultColormap(display,DefaultScreen(display));
	if ( XAllocNamedColor(display, cmap, name, &colour, &exact) ) 
	        pixelvalues=colour.pixel;
	else
	        pixelvalues=BlackPixel( display,DefaultScreen(display));
	
	newcm=(struct colmap *)malloc(sizeof(struct colmap));
	newcm->next=cm;
	newcm->name=(char *)malloc(strlen(name)+1);
	strcpy(newcm->name,name);
	newcm->pixelvalues=pixelvalues;

found:
	XSetForeground(display,gc,newcm->pixelvalues);
	return(newcm->pixelvalues);
}
#endif
