
/*
 *    X SHARE MANAGER (XSM)
 *
 *    Copyright (C) 1995  David Simons
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * $Id: d_draw.c,v 2.6 1995/06/17 20:27:19 davidsi Exp $
 *
 * $Log: d_draw.c,v $
 * Revision 2.6  1995/06/17  20:27:19  davidsi
 * - Remove debug printfs.
 *
 * Revision 2.5  1995/05/29  19:00:14  davidsi
 * - Added pixmapmade field to d_draw. This is TRUE only when the pixmaps, gcs,
 *   etc have all been correctly made for the drawing area.  The new function
 *   dialog_drawexposed() allows this value to be read.
 *
 * Revision 2.4  1995/05/19  19:00:07  davidsi
 * - Removed dialog_waitformap() which is now in d_base.c
 *
 * Revision 2.3  1995/04/25  21:02:24  davidsi
 * - Removed dialog_updatedisplay() - now in d_base.c where it belongs.
 *
 * Revision 2.2  1995/02/26  21:52:35  davidsi
 * - BUG FIX. In dialog_draw_lstyle() when there are either no elements
 *   in the line dash array, or no array is specified, only attempt to
 *   set the line attributes, don't reference the array.
 *
 * Revision 2.1  1995/01/31  22:02:19  davidsi
 * - Fixed to support proper font selection.
 *
 * Revision 2.0  1995/01/19  22:15:01  davidsi
 * - Added dialog_draw_lwidth() to allow line widths to be set.
 * - Added dialog_draw_colour() to allow colours to be set.
 * - Added dialog_updatedisplay() to allow one to wait for a dialog to
 *   be realised.
 *
 * Revision 1.6  1994/12/11  22:34:06  davidsi
 * - Removed dependancy of dialog_waitformapped() on draw dialogs.
 *
 * Revision 1.5  1994/12/11  17:57:52  davidsi
 * - Added dialog_waitformap() that waits for a dialog's window to be
 *   realized before returning.
 * - Added dialog_draw_lstyle to allow line dashing to be set.
 * - Initial width and height can now be specified for the dialog.  A flags
 *   value can also be given. The DRAW_CLEAR_ON_RESIZE flag specifies an
 *   automatic clear to be performed when the window changes size.
 *
 * Revision 1.4  1994/07/14  20:16:50  davidsi
 * - BUG FIX. Many of the globally visible functions were incorrectly
 *   typed.
 *
 * Revision 1.3  1993/01/31  18:46:45  davidsi
 * - When doing a redraw callback pass the value from the dialog structure
 *
 * Revision 1.2  1993/01/23  19:53:56  davidsi
 * - Use indirect font names for font manipulation rather than expecting
 *   X11 font names to be passed in.
 *
 * Revision 1.1  1993/01/20  20:54:36  davidsi
 * Initial revision
 *
 *
 */


/*
 * Include Files
 */
#include "stdio.h"
#include <math.h>
#include <malloc.h>

#include <X11/Intrinsic.h>
#include <X11/Shell.h>
#include <X11/Xatom.h>

#include <Xm/Xm.h>
#include <Xm/DrawingA.h>

#define IN_DIALOGS

#include "dialogcmn.h"
#include "d_base.h"
#include "d_form.h"
#include "d_draw.h"

#include "d_base.p"

extern d_root  *base_node;



/*
 * Variable Declarations 
 */
static char rcsid[]="$Id: d_draw.c,v 2.6 1995/06/17 20:27:19 davidsi Exp $";


extern d_root  *base_node;

#define POINTS_PER_INCH	72.0		/* points to inches */
#define SCREEN_WIDTH    10.0		/* width of screen in inches */
#define SCREEN_HEIGHT   7.0		/* height of screen in inches */
#define X_PIXELS        1024.0		/* width of screen in pixels */
#define Y_PIXELS        768.0		/* height of screen in pixels */
#define FIDDLE_FACTOR	1.4		/* "correction" factor for font size */


/*
 * Convenience macro for point/pixel conversion
 */
#define PIXEL2POINT ( (SCREEN_WIDTH*POINTS_PER_INCH) / X_PIXELS ) 


/*
 * fontmap maps application font names to X fontnames
 */
typedef struct fontmap {
	char *appname;
	char *x11name;
} fontmap_t;

fontmap_t fonts[10];


/*
 * init
 *
 * Set up font maps
 *
 * INPUTS :	None
 * RETURNS:	Non
 */
static
init()
{
	fonts[0].appname="Helvetica";
	fonts[0].x11name="-adobe-helvetica-medium-r-normal";
}
 

/*
 * dialog_draw_setfont dg id fname fsize
 *
 * Sets the current font for a draw dialog and context.
 *
 * INPUTS :	dg		The draw dialog
 *		id		The id to be used
 *		fname		The name of the font
 *		fsize		The size of the font
 * RETURNS:	Non
 */
void
dialog_draw_setfont(di,id,fname,fsize)
dialog di;
int id;
char *fname;
int fsize;
{
	static char ** fontnamelist=NULL;
	static XFontStruct **fontlist = NULL;
	static int count;
	static char 	*namecache=NULL;
	static XFontStruct     *fontcache = NULL;
	static int sizecache;
	int cursize=0;

	int i;
	int len;
	XFontStruct     *font = NULL;
	XFontStruct     *selected_font = NULL;
	unsigned long retval;
	char *xfont;
	int dobest;
	FILE *f;
	Display *display=base_node->display;
	char dofree;
	double wi;
	char *search_name;
	d_draw *dg=(d_draw *)di;

	fsize=((int) (0.5+(double)fsize*FIDDLE_FACTOR));

	if ( namecache && !(strcmp(fname,namecache)) && sizecache==fsize){
		font=fontcache;
		goto done;
	}

	init();

	/*
	 * Look for the font
	 */
	for(i=0;i<1;i++) {
		if (strcmp(fname,fonts[i].appname)==0)
			break;
	}
	xfont=fonts[i].x11name;

	if (i==sizeof(fonts) ) {
		fprintf(stderr,"Failed to find font %s\n",fname);
		exit(1);
	}
	if (fontnamelist) 
		goto scan;

	search_name=malloc(3+strlen(xfont));
	sprintf(search_name,"*%s*",xfont);
       	fontnamelist = XListFonts(display,search_name,2048, &count);

        if (!fontnamelist) {
		fprintf(stderr,"No font list on your server, exit\n");
		exit(1);
        }

	fontlist=(XFontStruct **)malloc(sizeof(XFontStruct *)*count);
	bzero(fontlist,sizeof(XFontStruct *)*count);

scan:
	len=strlen(xfont);
	dobest=-1;

	for(i=0;i<count;i++) {
		if (fontlist[i]==NULL) {
			fontlist[i] = XLoadQueryFont (display,fontnamelist[i]);
			if (fontlist[i]==NULL) {
				printf("Cannot load font %s\n",fontnamelist[i]);
				continue;
			}
			dofree=1;
		} else
			dofree=0;
		if (strncmp(xfont,fontnamelist[i],len)!=0) {
			if (dofree) {
				XFreeFont(display,fontlist[i]);
				fontlist[i]=NULL;
			}
			continue;
		}
		font = fontlist[i];
		XGetFontProperty(font,XA_POINT_SIZE,&retval);
		if (retval/10==fsize) {
			dobest=0;
			selected_font=font;
			cursize=fsize;
			break;
		} else if ( (retval/10<fsize) && (retval/10>cursize) ) {
			dobest=0;
			selected_font=font;
			cursize=retval/10;
		} 
	}

	if (dobest==-1) {
		fprintf(stderr,"Can't even guess at a good font!\n");
		exit(1);
	}
	font=selected_font;

	if (namecache) 
		free(namecache);

	namecache=malloc(strlen(fname)+1);
	strcpy(namecache,fname);
	fontcache=font;
	sizecache=fsize;
	/*
	 * Don't display a warning when font size is not exactly
	 * what was asked for...
	 *
	 * printf("Wanted %d but got %d\n",fsize,cursize);
	 */

done:
	dg->modes[id].font_info=font;
	XSetFont(base_node->display,dg->gc[id],dg->modes[id].font_info->fid);
}


/*
 * dialog_draw_lstyle dialog gcid number widths
 *
 * Sets line attrbutes to allow dashes to be displayed.
 *
 * INPUTS :	dg		The draw dialog
 *		id		The id to be used
 *		number		Number of dash widths to be used 
 *				(0 = turn of dashing)
 *		widths		Width of each dask
 *				(NULL = turn of dashing)
 * RETURNS:	None
 */
dialog_draw_lstyle(d,id,count,ptrs)
dialog d;
int id;
int count;
unsigned char *ptrs;
{
	d_draw *dg=(d_draw *)d;

	if (count==0 || ptrs==NULL) {
		XSetLineAttributes(base_node->display,dg->gc[id],
				dg->line_width[id],LineSolid,CapButt,JoinMiter);
		return;
	}
	XSetLineAttributes(base_node->display,dg->gc[id],0,
				LineOnOffDash,CapButt,JoinMiter);
	XSetDashes(base_node->display,dg->gc[id],dg->line_width[id],ptrs,count);
}


/*
 * dialog_draw_lwidth dialog gcid width
 *
 * Sets line attrbutes to allow dashes to be displayed.
 *
 * INPUTS :	dg		The draw dialog
 *		id		The id to be used
 *		width		Width of line
 * RETURNS:	None
 */
dialog_draw_lwidth(d,id,width)
dialog d;
int id;
int width;
{
	d_draw *dg=(d_draw *)d;
	XGCValues values;

	dg->line_width[id]=width;
	values.line_width=width;
	XChangeGC(base_node->display,dg->gc[id],GCLineWidth,&values);
}


/*
 * dialog_draw_colour d d col
 *
 * Set the colour for a draw dialog.
 * 
 * INPUTS :	dg		The draw dialog
 *		id		The id to be used
 *		col		The name of the colour
 * RETURNS:	None
 */
void
dialog_draw_colour(d,id,col)
dialog d;
int id;
char *col;
{
	d_draw *dg=(d_draw *)d;

	fetchcolour(col,dg->gc[id]);
}


/*
 * dialog_draw_fontheight 
 *
 * Returns the height of the current font.
 * 
 * INPUTS :	dg		The draw dialog
 *		id		The id to be used
 * RETURNS:	The height of the current font
 */
int
dialog_draw_fontheight(d,id)
dialog d;
int id;
{
	XFontStruct *font_info;
	d_draw *dg=(d_draw *)d;

	font_info=dg->modes[id].font_info;
	return(font_info->descent+font_info->ascent);
}

/*
 * dialog_draw_lines dialog id points number
 *
 * Draws lines between the specified points in the given draw dialog 
 * 
 * INPUTS :	dialog		The dialog to draw into
 *		id		The mode id for the draw
 *		points		An array of points (x,y) elements
 *		number		The number of points in the array
 * RETURNS:	None
 */
void
dialog_draw_lines(d,id,points,number)
dialog d;
int id;
coords *points;
int number;
{
	d_draw *dg=(d_draw *)d;

	XDrawLines(base_node->display,dg->pixmap,dg->gc[id],
			(XPoint *)points,number,CoordModeOrigin);
}



/*
 * dialog_draw_lines dialog id x1,y2 x1,y2
 *
 * Draws a line between the specified points in the given draw dialog 
 * 
 * INPUTS :	dialog		The dialog to draw into
 *		id		The mode id for the draw
 *		x1,y1		start coordinate
 *		x2,y2		start coordinate
 * RETURNS:	None
 */
void
dialog_draw_line(d,id,x1,y1,x2,y2)
dialog d;
{
	d_draw *dg=(d_draw *)d;

	XDrawLine(base_node->display,dg->pixmap,dg->gc[id],x1,y1,x2,y2);
}



/*
 * dialog_draw_lines dialog id x1,y1 text
 *
 * Writs a sting at the specified point in the given draw dialog 
 * 
 * INPUTS :	dialog		The dialog to draw into
 *		id		The mode id for the draw
 *		x1,y1		Start coordinate
 *		text		The ascii string to write
 * RETURNS:	None
 */
dialog_draw_string(d,id,x1,y1,text)
dialog d;
char *text;
{
	d_draw *dg=(d_draw *)d;

	XDrawString(base_node->display,dg->pixmap,dg->gc[id],
				x1,y1,text,strlen(text));
}


/*
 * dialog_string_length dialog id text
 *
 * Returns the length of a string in pixels.
 *
 * INPUTS :	dialog		The dialog to draw into
 *		id		The mode id for the draw
 *		text		The text
 * RETURNS:	The length of the string in pixels.
 */
int
dialog_string_length(d,id,text)
dialog d;
int id;
char *text;
{
	int twi;
	d_draw *dg=(d_draw *)d;

	twi=XTextWidth(dg->modes[id].font_info,text,strlen(text));
	return(twi);
}


/*
 * dialog_draw_rect dialog id x1,y2 width height
 *
 * Draws a rectange from the specified points in the given draw dialog.
 * The size is specified by the width and height of the box.
 * 
 * INPUTS :	dialog		The dialog to draw into
 *		id		The mode id for the draw
 *		x1,y1		Start coordinate
 *		width,height	Size of rectangle to draw
 * RETURNS:	None
 */
dialog_draw_rect(d,id,x1,y1,w,h)
dialog d;
{
	d_draw *dg=(d_draw *)d;

	XFillRectangle(base_node->display,dg->pixmap,dg->gc[id],x1,y1,w,h);
}


/*
 * dialog_draw_refresh dialog 
 *
 * Refreshes the on screen image of the dialog 
 * 
 * INPUTS :	dialog		The dialog to refresh
 * RETURNS:	None
 */
void
dialog_draw_refresh(d)
dialog d;
{
	d_draw *dg=(d_draw *)d;

	XCopyArea(base_node->display,dg->pixmap,dg->window,dg->gc_back,
			0,0,dg->width,dg->height,0,0);
}



/*
 * dialog_draw_value dialog width height 
 *
 * Gets the size of the on screen dialog
 * 
 * INPUTS :	dialog		The dialog to get the size of
 * RETURNS:	width		The width of the dialog
 * 		height		The height of the dialog
 */
void
dialog_draw_value(d,w,h)
dialog d;
int *w,*h;
{
	d_draw *dg=(d_draw *)d;

	*w=dg->width;
	*h=dg->height;
}


/*
 * GetPixmap dialog window_attr
 *
 * Ensures the pixmap is large enough for the window size, if it is found
 * not to be then is reallocated for the new size.  In this way the pixmap
 * grows with the window, but it should be noted that it will NOT shrink
 * with the window.
 *
 * INPUTS :	dialog		The draw dialog in question
 * 		window_attr	A XWindowAttributes ptr for the window
 *				in question
 * RETURNS:	None.
 */
static void
GetPixmap(dg,wa)
XWindowAttributes *wa;
d_draw *dg;
{
	int i;

	/*
	 * Check to see if pixmap has grown
	 */
	if (dg->p_height>=wa->height && dg->p_width>=wa->width && dg->pixmap )
		return;


	if (dg->pixmap)
		XFreePixmap(base_node->display,dg->pixmap);
	dg->pixmap=XCreatePixmap(base_node->display,dg->window,
					wa->width,wa->height,wa->depth);
	if ( !dg->pixmap ) {
		printf("SANITY:Failed to get pixmap\n");
		exit(1);
	}
	dg->p_height=wa->height;
	dg->p_width=wa->width;

	/*
	 * Set up the graphics contexts
	 */
	if (dg->gc_back)
		XFreeGC(base_node->display,dg->gc_back);
	dg->gc_back=XCreateGC(base_node->display,dg->pixmap,0,NULL);
	if (dg->gc_back==NULL) {
		printf("Failed to obtain graphics context\n");
		exit(1);
	}
	XSetFunction(base_node->display,dg->gc_back,GXcopy);
	for(i=0;i<MAXMODES;i++) {
		dg->gc[i]=XCreateGC(base_node->display,dg->pixmap,0,NULL);
		if (dg->modes[i].font_info)
			XSetFont(base_node->display,
					dg->gc[i],dg->modes[i].font_info->fid);
		dg->line_width[i]=0;
	}
	XSetForeground(base_node->display,dg->gc[0],
	    WhitePixel(base_node->display,DefaultScreen(base_node->display)));
	XSetBackground(base_node->display,dg->gc[1],
	    BlackPixel(base_node->display,DefaultScreen(base_node->display)));

	/*
	 * From this point, the pixmaps and graphic contexts are made
	 */
	dg->pixmapmade=TRUE;
}


/*
 * dialog_resize widget padding callback
 *
 * Handles expose events for the drawing area.  If the callback is because
 * the window is larger than it was then it gets fully redrawn and (for
 * the moment) we continue to process events related to this resize.  For
 * non-window increasing events we simply render the area specified by
 * the callback information.
 *
 * INPUTS : 	widget		The widget involved
 *          	callback	The callback info
 * RETURNS:	None.
 */
static void
dialog_resize(w,dg,b)
Widget w;
d_draw *dg;
XmDrawingAreaCallbackStruct *b;
{
	XExposeEvent *e;
	XWindowAttributes wa;
	Region region;
	XRectangle rect;
	XEvent report;
	XGCValues v;
	int ac;
	Arg al[2];

	dg->window=XtWindow(w);
	if ( dg->window ) {
		XGetWindowAttributes(base_node->display,dg->window,&wa);
		e=(XExposeEvent *)(b->event);
		/*
		 * If window is growing, check pixmap size and redraw if
		 * necessary
		 */
		if (wa.height!=dg->height || wa.width!=dg->width) {
			dg->height=wa.height;
			dg->width=wa.width;
			GetPixmap(dg,&wa);
    			if (dg->dflags & DRAW_CLEAR_ON_RESIZE)
				dialog_draw_rect(dg,1,0,0,wa.width,wa.height);
			dg->redraw_callback(dg,dg->value);
			return;
		}
   	} 
}


/*
 * dialog_expose widget padding callback
 *
 * Handles expose events for the drawing area.  If the callback is because
 * the window is larger than it was then it gets fully redrawn and (for
 * the moment) we continue to process events related to this resize.  For
 * non-window increasing events we simply render the area specified by
 * the callback information.
 *
 * INPUTS :	widget		The widget involved
 *		callback	The callback info
 * RETURNS: None.
 */
static void
dialog_expose(w,dg,b)
Widget w;
d_draw *dg;
XmDrawingAreaCallbackStruct *b;
{
	XExposeEvent *e;
	XWindowAttributes wa;
	Region region;
	XRectangle rect;
	XEvent report;
	XGCValues v;
	int ac;
	Arg al[2];				/* arg list	*/

	dg->window=XtWindow(w);
	if ( dg->window ) {
		XGetWindowAttributes(base_node->display,dg->window,&wa);
		e=(XExposeEvent *)(b->event);
		/*
		 * If window is growing, check pixmap size and redraw if
		 * necessary
		 */
		if (wa.height!=dg->height || wa.width!=dg->width) {
			dg->height=wa.height;
			dg->width=wa.width;
  			GetPixmap(dg,&wa);
    			if (dg->dflags & DRAW_CLEAR_ON_RESIZE)
				dialog_draw_rect(dg,1,0,0,wa.width,wa.height);
			dg->redraw_callback(dg,dg->value);
			return;
      		} else {
        	 	XCopyArea(base_node->display,dg->pixmap,dg->window,
			    dg->gc_back,e->x,e->y,e->width,e->height,e->x,e->y);
		}
	}
}


/*
 * dialog_drawexposed dialog
 *
 * Returns whether a drawing area is ready
 *
 * INPUTS :	dialog		The drawing dialog
 * RETURNS:	TRUE if the drawing area is ready to accept all rendering
 *		calls, else FALSE
 */
int
dialog_drawexposed(d)
dialog d;
{
	d_draw *dg=(d_draw *)d;

	return(dg->pixmapmade);
}

/*
 * dialog_draw parent new_d position
 *
 * Creates a draw dialog.
 *
 * INPUTS :	parent		The parent dialog
 *		new_d		The new draw structure
 *		position	The position within the parent dialog
 * RETURNS:	A pointer to the newly created dialog.	
 */
dialog
dialog_draw(d,new_d,p)
dialog d;
new_d_draw *new_d;
position *p;
{
	Arg al[20];
	int ac;
	d_draw *dg;

	dg=(d_draw *)dialog_init(d,sizeof(d_draw));
	XtSetArg(al[0],XmNwidth,new_d->width);
	XtSetArg(al[1],XmNheight,new_d->height);

	ac=2;
	if (p) 
		dialog_position(p,al,&ac);
	dg->widget=XmCreateDrawingArea(d->widget, "", al, ac);
	dg->redraw_callback=new_d->redraw_callback;
	dg->value=new_d->value;
	dg->dflags=new_d->flags;
	dg->pixmapmade=FALSE;
	XtAddCallback(dg->widget,XmNexposeCallback,dialog_expose,dg);
	XtAddCallback(dg->widget,XmNresizeCallback,dialog_resize,dg);

	return((dialog)dg);
}
