
/*
 *    X SHARE MANAGER (XSM)
 *
 *    Copyright (C) 1995  David Simons
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef _DIALOGDRAW_H
#define _DIALOGDRAW_H
/*
 * $Id: d_draw.h,v 2.1 1995/05/29 18:59:49 davidsi Exp $
 *
 * $Log: d_draw.h,v $
 * Revision 2.1  1995/05/29  18:59:49  davidsi
 * - Added GNU copyright message.
 * - Removed the exposed field from d_draw - it wasn't used. Also removed
 *   l_width and l_height.
 * - Added pixmapmade field to d_draw. This is TRUE only when the pixmaps, gcs,
 *   etc have all been correctly made for the drawing area.
 *
 * Revision 2.0  1995/01/19  22:14:08  davidsi
 * - Added line_width array to the d_draw dialog so as to support
 *   different line widths for each GC.
 *
 * Revision 1.4  1994/12/11  17:36:03  davidsi
 * - The new_d_draw structure now allows an initial width and height
 *   value.  Flags can also be specified.
 * - Add to d_draw an exposed field and a copy of the flags that
 *   were specified at creation time.
 *
 * Revision 1.3  1993/01/31  18:46:28  davidsi
 * - Added comment for callback value
 *
 * Revision 1.2  1993/01/20  22:21:58  davidsi
 * - Removed all function prototypes, now stored separately.
 *
 * Revision 1.1  1993/01/20  21:08:10  davidsi
 * Initial revision
 *
 *
 */


typedef struct new_d_draw {
	void (*redraw_callback)();
	long value;
	int width;
	int height;
	unsigned long flags;
} new_d_draw;

#define DRAW_CLEAR_ON_RESIZE	0x1	/* clear all drawing area when it */ 
					/* is resized */

typedef struct {
	short x, y;
} coords;

typedef struct {
	int ascent;
	int descent;
	int width;
	int height;
} fontsize;



#ifdef IN_DIALOGS

#include "dialogcmn.h"

#define MAXMODES 16

typedef struct {
    XFontStruct *font_info;
} modes_t;

typedef struct d_draw {
	DIALOGHEADER
	Pixmap pixmap;
	Window window;
	GC gc_back;
	GC gc[MAXMODES];
	int line_width[MAXMODES];
	modes_t modes[MAXMODES];
	void (*redraw_callback)();
	long value;			/* callback value */
	int width;
	int height;
	int p_width;			/* size of pixmap */
	int p_height;
	char pixmapmade;		/* TRUE when pixmaps/gcs complete */
	unsigned long dflags;		/* as flags above */
} d_draw;
#endif

#endif /* _DIALOGDRAW_H */
