
/*
 *    X SHARE MANAGER (XSM)
 *
 *    Copyright (C) 1995  David Simons
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * $Id: d_entry.c,v 2.1 1995/05/19 19:00:23 davidsi Exp $
 *
 * $Log: d_entry.c,v $
 * Revision 2.1  1995/05/19  19:00:23  davidsi
 * - Tidying up and added GNU copyright.
 *
 * Revision 2.0  1995/04/25  21:01:54  davidsi
 * - Added appropriate casts to get rid of compiler warnings.
 *
 * Revision 1.3  1994/07/14  20:18:30  davidsi
 * - BUG FIX. Many of the globally visible functions were incorrectly
 *   typed.
 *
 * Revision 1.2  1994/07/03  21:24:42  davidsi
 * - Added support for a new entry list dialog part. The function
 *   dialog_labelentry_cb() allows a label to be created that will
 *   cause a callback when the user modifies the information.
 *
 * Revision 1.1  1993/01/20  21:05:01  davidsi
 * Initial revision
 *
 *
 */


/*
 * Include Files
 */
#include <X11/Intrinsic.h>
#include <X11/Shell.h>

#include <Xm/Xm.h>
#include <Xm/RowColumn.h>
#include <Xm/Text.h>
#include <Xm/Label.h>
#include <Xm/Form.h>

#define IN_DIALOGS

#include "d_base.h"
#include "d_entry.h"
#include "d_form.h"


/*
 * Variable Declarations 
 */
static char rcsid[]="$Id: d_entry.c,v 2.1 1995/05/19 19:00:23 davidsi Exp $";



/*
 * dialog_labelentry_cb dialog label initial id
 *
 * Adds an entry field to an entry list.
 *
 * INPUTS :	dialog		The parent dialog (an entrylist)
 *		label		The name of the entry field (displayed)
 *		initial		The initial entry of the entry field
 *		func		The function to call back to
 *		id		The reference id
 * RETURNS:	The newly created dialog
 */
dialog
dialog_labelentry_cb(d,labelstr,ivalue,func,id)
dialog d;
char *labelstr;
char *ivalue;
void (*func)();
int id;
{
	d_entrylist *parent=(d_entrylist *)d;
	Widget c1=parent->widget;
	Arg al[10];		/* arg list */
	Widget label;		/* the label widget that gets created */
	Widget rc;		/* the rowcol widget to bind title & entry to */
	Widget new_widget;	/* the new widget */
	int textheight;
	int lblheight;
	d_entry *d_e;
	int ac;
	Widget form;

	d_e=(d_entry *)dialog_init(parent,sizeof(d_entry));
	d_e->id=id;
	
	form=XmCreateForm(c1,"",al,0);
	XtManageChild(form);
	XtSetArg(al[0],XmNcolumns,10);	/* entry field size is 10 chars */
	XtSetArg(al[1],XmNeditMode,XmSINGLE_LINE_EDIT);
	XtSetArg(al[2],XmNrightAttachment,XmATTACH_POSITION);
	XtSetArg(al[3],XmNrightPosition,100);

	if (ivalue) {
		XtSetArg(al[4],XmNvalue,ivalue);
		ac=5;
	} else
		ac=4;

	d_e->widget=new_widget=XmCreateText(form,"rc-frame",al,ac);
	XtAddCallback(new_widget, XmNactivateCallback,
			(XtCallbackProc)XmProcessTraversal, 
			(XtPointer)XmTRAVERSE_NEXT_TAB_GROUP);
	XtManageChild(new_widget);

	XtSetArg(al[0],XmNleftAttachment,XmATTACH_POSITION);
	XtSetArg(al[1],XmNleftPosition,0);
	XtSetArg(al[2],XmNrightAttachment,XmATTACH_WIDGET);
	XtSetArg(al[3],XmNrightWidget,new_widget);
	XtSetArg(al[4],XmNalignment,XmALIGNMENT_BEGINNING);
	label= XmCreateLabel(form, labelstr, al, 5);
	XtManageChild(label);

	return((dialog)d_e);
}

/*
 * dialog_entrylist_value dialog id
 *
 * Returns the value of an item in an entrylist as given by the id passed
 *
 * INPUTS :	dialog		The entrylist dialog
 * 		id		The id of the item in the list req'd
 * RETURNS:	An ascii string giving the data entered, or NULL if
 *		the item with the given id cannot be found.
 */
char *
dialog_entrylist_value(ol_ptr,id)
dialog ol_ptr;
int id;
{
	Arg al[10];
	d_entry *tree;
	char *str;

	tree=(d_entry *)ol_ptr->child;
	do {
		if (tree->id==id) {
			XtSetArg(al[0],XmNvalue,&str);
			XtGetValues(tree->widget,al,1);
			return(str);
		}
		tree=(d_entry*)tree->sibling_f;
	} while (tree!=(d_entry *)ol_ptr->child);
	return(NULL);
}


/*
 * dialog_entrylist_setvalue dialog id val
 *
 * Sets the value of an item in an entrylist as given by the id passed
 *
 * INPUTS :	dialog		The entrylist dialog
 * 		id		The id of the item in the list req'd
 *		val		The value to set in the entry field
 * RETURNS:	None
 */
void
dialog_entrylist_setvalue(ol_ptr,id,val)
dialog ol_ptr;
int id;
char *val;
{
	Arg al[10];
	d_entry *tree;

	tree=(d_entry *)ol_ptr->child;
	do {
		if (tree->id==id) {
			XtSetArg(al[0],XmNvalue,val);
			XtSetValues(tree->widget,al,1);
			return;
		}
		tree=(d_entry*)tree->sibling_f;
	} while (tree!=(d_entry *)ol_ptr->child);
	return;
}




/*
 * dialog_labelentry dialog label initial id
 *
 * Adds an entry field to an entry list.
 *
 * INPUTS :	dialog		The parent dialog (an entrylist)
 *		label		The name of the entry field (displayed)
 *		initial		The initial entry of the entry field
 *		id		The reference id
 * RETURNS:	The newly created dialog
 */
dialog
dialog_labelentry(parent,labelstr,ivalue,id)
dialog parent;
char *labelstr;
char *ivalue;
int id;
{
	d_entrylist *d_el=(d_entrylist *)parent;
	Widget c1=d_el->widget;
	Arg al[10];		/* arg list */
	Widget label;		/* the label widget that gets created */
	Widget rc;		/* the rowcol widget to bind title & entry to */
	Widget new_widget;	/* the new widget */
	int textheight;
	int lblheight;
	d_entry *d_e;
	int ac;
	Widget form;

	d_e=(d_entry *)dialog_init(parent,sizeof(d_entry));
	d_e->id=id;
	
	form=XmCreateForm(c1,"",al,0);
	XtManageChild(form);
	XtSetArg(al[0],XmNcolumns,d_el->cols);	/* entry field size in chars */
	XtSetArg(al[1],XmNeditMode,XmSINGLE_LINE_EDIT);
	XtSetArg(al[2],XmNrightAttachment,XmATTACH_POSITION);
	XtSetArg(al[3],XmNrightPosition,100);

	if (ivalue) {
		XtSetArg(al[4],XmNvalue,ivalue);
		ac=5;
	} else
		ac=4;

	d_e->widget=new_widget=XmCreateText(form,"rc-frame",al,ac);
	XtAddCallback(new_widget, XmNactivateCallback,
			(XtCallbackProc)XmProcessTraversal, 
			(XtPointer)XmTRAVERSE_NEXT_TAB_GROUP);
	XtManageChild(new_widget);

	XtSetArg(al[0],XmNleftAttachment,XmATTACH_POSITION);
	XtSetArg(al[1],XmNleftPosition,0);
	XtSetArg(al[2],XmNrightAttachment,XmATTACH_WIDGET);
	XtSetArg(al[3],XmNrightWidget,new_widget);
	XtSetArg(al[4],XmNalignment,XmALIGNMENT_BEGINNING);
	label= XmCreateLabel(form, labelstr, al, 5);
	XtManageChild(label);

	return((dialog)d_e);
}




/*
 * dialog_entrylist dialog
 *
 * Creates an entrylist dialog
 *
 * INPUTS :	dialog		The parent dialog
 *		position	The position information
 * RETURNS:	The newly created dialog
 */
dialog
dialog_entrylist(parent,ndl,p)
dialog parent;
new_d_entrylist *ndl;
position *p;
{
	Arg al[10];
	int ac;
	d_entrylist *d_el;

	d_el=(d_entrylist *)dialog_init(parent,sizeof(d_entrylist));
	XtSetArg(al[0],XmNnumColumns,1);
	XtSetArg(al[1],XmNpacking,XmPACK_TIGHT);
	XtSetArg(al[2],XmNorientation,XmVERTICAL);
	ac=3;
	if (p)
		dialog_position(p,al,&ac);
	d_el->widget=XmCreateRowColumn(parent->widget,"",al,ac);
	d_el->cols=ndl->cols;
	return((dialog)d_el);
}

/*
 * dialog_entry_value dialog
 *
 * Returns the information that has been entered into a entry dialog
 * 
 * INPUTS :	dialog		The entry dialog
 * RETURNS:	An ascii string containing whats been entered.
 */
char *
dialog_entry_value(parent)
dialog parent;
{
	Arg al[1];
	char *str;

	XtSetArg(al[0],XmNvalue,&str);
	XtGetValues(parent->widget,al,1);
	return(str);
}


/*
 * dialog_entry parent new_entry position
 *
 * Creates a form dialog, for use when newly create dialogs specify
 * a position.
 *
 * INPUTS :	parent		The parent dialog
 *		new_entry	The new entry structure
 *		position	The structure describing positioning
 * RETURNS:	A pointer to the newly created dialog.	
 */
dialog
dialog_entry(parent,ne,p)
dialog parent;
new_d_entry *ne;
position *p;
{
	Arg al[10];
	d_entry *d_e;
	int ac;

	d_e=(d_entry *)dialog_init(parent,sizeof(d_entry));
	
	XtSetArg(al[0],XmNcolumns,ne->cols);
	XtSetArg(al[1],XmNeditMode,XmSINGLE_LINE_EDIT);
	ac=2;

	if (ne->ivalue) {
		XtSetArg(al[ac],XmNvalue,ne->ivalue);
		ac++;
	} 
	if (p)
		dialog_position(p,al,&ac);

	d_e->widget=XmCreateText(parent->widget,"rc-frame",al,ac);
	XtAddCallback(d_e->widget, XmNactivateCallback,
			(XtCallbackProc)XmProcessTraversal, 
			(XtPointer)XmTRAVERSE_NEXT_TAB_GROUP);
	return((dialog)d_e);
}
