
/*
 *    X SHARE MANAGER (XSM)
 *
 *    Copyright (C) 1995  David Simons
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * $Id: d_menu.c,v 2.1 1995/05/19 18:59:12 davidsi Exp $
 *
 * $Log: d_menu.c,v $
 * Revision 2.1  1995/05/19  18:59:12  davidsi
 * - BUG FIX. In menu_refresh() re-establish the new scrollbar as part of
 *   the main window.
 *
 * Revision 2.0  1995/01/19  22:12:49  davidsi
 * - Added menu_refresh() to allow menus to be dynamically updated.
 *
 * Revision 1.4  1994/10/18  20:11:53  davidsi
 * - Modified MENUI_ENABLED to MENUTYPE_ENABLED which although
 *   numerically the same value are different.  MENUI_ENABLED is
 *   the internal flag and MENUTYPE_ENABLED is the external flag.
 *
 * Revision 1.3  1994/08/06  16:47:00  davidsi
 * - Modified definitions for flags so that C++ does not issue warnings.
 *
 * Revision 1.2  1994/07/14  20:19:03  davidsi
 * - BUG FIX. Many of the globally visible functions were incorrectly
 *   typed.
 *
 * Revision 1.1  1993/01/20  20:57:52  davidsi
 * Initial revision
 *
 *
 */

static char rcsid[]="$Id: d_menu.c,v 2.1 1995/05/19 18:59:12 davidsi Exp $";

/*
 * Include Files
 */
#define IN_DIALOGS	1

#include <stdio.h>

#include <X11/Intrinsic.h>
#include <X11/Shell.h>

#include <Xm/Xm.h>
#include <Xm/RowColumn.h>
#include <Xm/SelectioB.h>
#include <Xm/Label.h>
#include <Xm/PushB.h>
#include <Xm/CascadeB.h>
#include <Xm/ToggleB.h>
#include <Xm/Text.h>

#include "d_menu.h"
#include "d_base.h"

extern d_root *base_node;

/*
 * menu_docallback widget menuptr
 *
 * Handles the callbacks from the menu widgets
 *
 * INPUTS :	parent		The parent widget of the menu bar to be set up
 *		head_wmb_ptr	The menu that the callback relates to
 * RETURNS:	None
 */
static void
menu_docallback(w,wm_ptr)
Widget w;
d_menu *wm_ptr;
{
	int onflag=0;
	Boolean on;
	Arg al[2];

	if (wm_ptr->flags&(MENUTYPE_NOFMANY|MENUTYPE_ONEOFMANY)) {
		XtSetArg(al[0],XmNset,&on);
		XtGetValues(w,al,1);
		if (on)
			onflag=1;
	}
	wm_ptr->callback(wm_ptr,wm_ptr->value,onflag);
}


/*
 * menu_displaybar dialog
 *
 * Displays a previously set up menu bar.
 *
 * INPUTS :	dialog		The dialog for the menu bar.
 * RETURNS:	None
 */
static void
menu_displaybar(d)
dialog d;
{
	d_menub *head_wmb_ptr=(d_menub *)d;
	Widget menu_bar,button;
	Widget p_menu;             /* Dates now changed via a menu */
	Widget p_button;           /* Button for changing dates */
	d_menub *wmb_ptr;
	d_menu *wm_ptr;
	Arg al[10];
	Widget w1,w2,w3,w4;
	Widget parent;

	if (head_wmb_ptr->privateflags & DIALOG_REALISED) 
		return;
	head_wmb_ptr->privateflags |= DIALOG_REALISED;
	parent=head_wmb_ptr->parent->widget;
	menu_bar = XmCreateMenuBar(parent, "menu_bar", NULL, 0);
	head_wmb_ptr->menu_bar=menu_bar;
	for(wmb_ptr=head_wmb_ptr->next;wmb_ptr!=NULL;wmb_ptr=wmb_ptr->next) {
		switch (wmb_ptr->flags & MENUTYPE_MASK ) {
		case MENUTYPE_ACTION:
			p_menu=XmCreatePulldownMenu(menu_bar,"",NULL,0);
			wmb_ptr->widget=p_menu;
			for(wm_ptr=wmb_ptr->mitem;
			     wm_ptr!=NULL;wm_ptr=wm_ptr->next) {
				p_button=XmCreatePushButton(p_menu,
						   wm_ptr->name,NULL,0);
				wm_ptr->widget=p_button;
				XtAddCallback(p_button, XmNactivateCallback, 
						menu_docallback,wm_ptr);
				XtManageChild(p_button);
			}
			XtSetArg(al[0],XmNsubMenuId,p_menu);
			button=XmCreateCascadeButton(menu_bar,
							wmb_ptr->name,al,1);
			XtManageChild(button);
			break;
		case MENUTYPE_NOFMANY:
			p_menu=XmCreatePulldownMenu(menu_bar,"",NULL,0);
			wmb_ptr->widget=p_menu;
			XtSetArg(al[0],XmNset,TRUE);
			for(wm_ptr=wmb_ptr->mitem;wm_ptr!=NULL;
			     wm_ptr=wm_ptr->next) {
				if (wm_ptr->flags & MENUTYPE_ENABLED)
	  				p_button=XmCreateToggleButton(p_menu,
							wm_ptr->name,al,1);
				else
	  				p_button=XmCreateToggleButton(p_menu,
							wm_ptr->name,NULL,0);
				wm_ptr->widget=p_button;
				XtAddCallback(p_button,XmNvalueChangedCallback,
						menu_docallback,wm_ptr);
				XtManageChild(p_button);
			}
			XtSetArg(al[0],XmNsubMenuId,p_menu);
			button=XmCreateCascadeButton(menu_bar,
							wmb_ptr->name,al,1);
			XtManageChild(button);
			break;
		case MENUTYPE_ONEOFMANY:
			XtSetArg(al[0],XmNradioBehavior,1);
			p_menu=XmCreatePulldownMenu(menu_bar,"",al,1);
			wmb_ptr->widget=p_menu;
			XtSetArg(al[0],XmNindicatorType,XmONE_OF_MANY);
			XtSetArg(al[1],XmNset,TRUE);
			for(wm_ptr=wmb_ptr->mitem;
			    wm_ptr!=NULL;wm_ptr=wm_ptr->next) {
				if (wm_ptr->flags & MENUTYPE_ENABLED)
	  				p_button=XmCreateToggleButton(p_menu,
							wm_ptr->name,al,2);
				else
  					p_button=XmCreateToggleButton(p_menu,
							wm_ptr->name,al,1);
				wm_ptr->widget=p_button;
				XtAddCallback(p_button,XmNvalueChangedCallback,
						menu_docallback,wm_ptr);
				XtManageChild(p_button);
			}
			XtSetArg(al[0],XmNsubMenuId,p_menu);
			button=XmCreateCascadeButton(menu_bar,
							wmb_ptr->name,al,1);
			XtManageChild(button);
			break;
		default:
			fprintf(stderr,"menu_displaybar: Bad menu type\n");
			exit(1);
		}
	}
	XtManageChild(menu_bar);

}


/*
 * menu_refresh oldd,newd
 *
 * Updates a menu (thats already on display) with new items (as necessary).
 *
 * INPUTS :	oldd		The old menu dialog
 *		newd		The new menu dialog
 * RETURNS:	None
 */
void
menu_refresh(oldd,newd)
dialog oldd;
dialog newd;
{
	d_menub *head_wmb_ptr=(d_menub *)oldd;
	d_menub *newhead_wmb_ptr=(d_menub *)newd;
	Arg al[10];

	XtUnmanageChild(head_wmb_ptr->menu_bar);
	XtDestroyWidget(head_wmb_ptr->menu_bar);
	menu_displaybar(newd);
	XtSetArg(al[0],XmNmenuBar,newhead_wmb_ptr->menu_bar);
	XtSetValues(base_node->widget,al,1);
}


/*
 * menu_additem wmb_ptr name callback flags
 *
 * Adds a new menu item to a menu.
 *
 * INPUTS :	wmb_ptr		The menu structure to add the item to
 *		name		The name of the new item
 *		callback	The callback when the item is selected
 *		value		The value to pass to the callback
 *		flags		see MENUI_XXX
 * RETURNS: None
 */
void
menu_additem(d,name,callback,value,flags)
dialog d;
char *name;
void (*callback)();
long value;
unsigned long flags;
{
	d_menub *wmb_ptr=(d_menub *)d;
	d_menu *new_wm_ptr;

	new_wm_ptr=(d_menu *)malloc(sizeof(d_menu));
	if (wmb_ptr->last_mitem==NULL) {
		wmb_ptr->last_mitem=wmb_ptr->mitem=new_wm_ptr;
	} else {
		wmb_ptr->last_mitem->next=new_wm_ptr;
		wmb_ptr->last_mitem=new_wm_ptr;
	}
	new_wm_ptr->next=NULL;
	new_wm_ptr->name=(char *)malloc(strlen(name)+1);
	strcpy(new_wm_ptr->name,name);
	new_wm_ptr->callback=callback;
	new_wm_ptr->value=value;
	new_wm_ptr->flags=flags|(wmb_ptr->flags & MENUTYPE_MASK);
}



/* 
 * menu_createbar
 *
 * Creates a menu bar.  Note that a menu bar is simply a menu to which
 * no items will be attached.
 *
 * INPUTS :	parent		The parent dialog for the menu bar
 * RETURNS:	A pointer to the newly created menu bar
 */
dialog
menu_createbar(parent)
dialog parent;
{
	d_menub *new_wmb_ptr;

	new_wmb_ptr=(d_menub *)dialog_init(parent,sizeof(d_menub));
	new_wmb_ptr->next=NULL;
	new_wmb_ptr->last=NULL;
	new_wmb_ptr->last_mitem=NULL;
	new_wmb_ptr->mitem=NULL;
	new_wmb_ptr->name=NULL;
	new_wmb_ptr->realise=menu_displaybar;
	new_wmb_ptr->destroy=NULL;
	return((dialog)new_wmb_ptr);
}


/*
 * menu_addtobar wmb_ptr name
 *
 * Adds a new menu structure to a menu bar structure.
 *
 * INPUTS :	wmb_ptr			Pointer to the menu bar
 *					(head menu structure)
 *		name			Name of the new menu
 *	 	flags			See MENUTYPE_XX
 * RETURNS:	The new menu structure.
 */
dialog
menu_addtobar(d,name,flags)
dialog d;
char *name;
unsigned long flags;
{
	d_menub *new_wmb_ptr;
	d_menub *wmb_ptr=(d_menub *)d;

	for( new_wmb_ptr=wmb_ptr->next;new_wmb_ptr!=NULL;
	     new_wmb_ptr=new_wmb_ptr->next) {
		if (!strcmp(new_wmb_ptr->name,name) )
			return((dialog)new_wmb_ptr);
	}
	new_wmb_ptr=(d_menub *)malloc(sizeof(d_menub));
	if (wmb_ptr->last==NULL) {
		wmb_ptr->last=wmb_ptr->next=new_wmb_ptr;
	} else {
		wmb_ptr->last->next=new_wmb_ptr;
		wmb_ptr->last=new_wmb_ptr;
	}
	new_wmb_ptr->flags=flags;
	new_wmb_ptr->next=NULL;
	new_wmb_ptr->mitem=NULL;
	new_wmb_ptr->last_mitem=NULL;
	new_wmb_ptr->name=(char *)malloc(strlen(name)+1);
	strcpy(new_wmb_ptr->name,name);
	return((dialog)new_wmb_ptr);
}
