
/*
 *    X SHARE MANAGER (XSM)
 *
 *    Copyright (C) 1995  David Simons
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


/*
 * $Id: d_optionl.c,v 2.1 1995/05/19 19:00:44 davidsi Exp $
 *
 * $Log: d_optionl.c,v $
 * Revision 2.1  1995/05/19  19:00:44  davidsi
 * - Tidying up and added GNU copyright.
 *
 * Revision 2.0  1995/02/25  21:09:29  davidsi
 * - Added dialog_optionlist_setvalue() such that an item in the
 *   list can be pre-selected.
 *
 * Revision 1.5  1994/10/18  20:13:26  davidsi
 * - Removed the XtManageChild() from dialog_optionlist() since the
 *   widgets should only get managed when the dialog hierarchy is
 *   realized.
 *
 * Revision 1.4  1994/07/14  20:19:20  davidsi
 * - BUG FIX. Many of the globally visible functions were incorrectly
 *   typed.
 *
 * Revision 1.3  1994/07/03  21:19:27  davidsi
 * - Support for NULL terminated lists - rather than requiring list
 *   sizes to be passed from the outside world.
 *
 * Revision 1.2  1993/01/20  21:47:03  davidsi
 * - Added many casts to remove compiler warnings.
 *
 * Revision 1.1  1993/01/20  21:06:25  davidsi
 * Initial revision
 *
 *
 */


/*
 * Include Files
 */
#include <X11/Intrinsic.h>
#include <X11/Shell.h>

#include <Xm/Xm.h>
#include <Xm/SelectioB.h>
#include <Xm/List.h>

#define IN_DIALOGS

#include "d_base.h"
#include "d_optionl.h"
#include "d_form.h"


/*
 * Variable Declarations 
 */
static char rcsid[]="$Id: d_optionl.c,v 2.1 1995/05/19 19:00:44 davidsi Exp $";


/*
 * findname widget
 *
 * When the text value changes findname is used to see if the new
 * name matches some item in the list.  If it does then callbacks(etc)
 * take place if req'd by the option list preferences.
 * 
 * INPUTS :	widget 		The initiating widget
 * RETURNS:	None
 */
static void
findname(w)
Widget w;
{
	Arg al[10];				/* arg list */
	Widget w_list;
	char *textin;
	char **list;
	int len;
	int pos;
	int located=0;
	int oldpos;
	char *oldname;
	char *text;
	XmString *items;
	d_optionl *ol_ptr;

	XtSetArg(al[0],XmNuserData,&ol_ptr);
	XtGetValues(w,al,1);
	
	w=XmSelectionBoxGetChild (ol_ptr->widget, XmDIALOG_TEXT);
	w_list=XmSelectionBoxGetChild (ol_ptr->widget, XmDIALOG_LIST);
	XtSetArg(al[0],XmNvalue,&textin);
	XtGetValues(w,al,1);
	len=strlen(textin);

	pos=1 ;
	list=ol_ptr->list;
	while (*list!=NULL) {
		if (!(strncmp(*list,textin,len))) {
		       /*
			* Found the start of all matching names, so hyperspace
			* the display to this point
			*/
			if (located==0) {
				if (ol_ptr->user.flags & OPTION_LIST_SEARCH)
					XmListSetPos(w_list,pos);
				oldpos=pos;
				oldname=*list;
			}
			located++;
		}
		pos++;
		list++;
	}

	/*
	 * If only one name looks close pick it - but only if it is not
	 * already selected - otherwise one has infinite recursion.
	 */
	if ( (ol_ptr->user.flags & OPTION_LIST_AUTOP) && (located==1) )  {
		XtSetArg(al[0],XmNselectedItems,&items);
		XtGetValues(w_list,al,1);
		if (items!=NULL) {
		XmStringGetLtoR(*items,
					XmSTRING_DEFAULT_CHARSET,&text);
			if (!strcmp(text,oldname))
				return;
		}
		XmListSelectPos(w_list,oldpos,TRUE);
	}
}


/*
 * dialog_optionlist_setvalue dialog textin
 * 
 * Sets the currently selected item within the displayed list.
 *
 * INPUTS :	dialog		The dialog 
 *		textint		The next text that's to be selected
 * RETURNS:	None
 */
void
dialog_optionlist_setvalue(d,textin)
dialog d;
char *textin;
{
	Arg al[10];
	Widget w_list;
	char **list;
	int pos;
	d_optionl *ol_ptr=(d_optionl *)d;

	w_list=XmSelectionBoxGetChild (ol_ptr->widget, XmDIALOG_LIST);

	pos=1 ;
	list=ol_ptr->list;
	while (*list!=NULL) {
		if (!(strcmp(*list,textin))) {
			/*XmListSetPos(w_list,pos);*/
			XmListSelectPos(w_list,pos,TRUE);
			return;
		}
		pos++;
		list++;
	}
	XmListSelectPos(w_list,0,TRUE);
}

/*
 * dialog_optionlist_value dialog number
 * 
 * Returns the current items in the option list
 *
 * INPUTS :	dialog		The dialog 
 * RETURNS:	number		Points to the size of the returned array
 *		Pointer to the array.  The array can be freed with 
 *		freestringlist(array,number)
 */
char **
dialog_optionlist_value(ol_ptr,number)
dialog ol_ptr;
int *number;
{
	Arg al[10];			/* arg list */
	char *namestr;
	XmString *list;
	char **newlist;
	int iter;
	char *text;


	XtSetArg(al[0],XmNlistItemCount,number);
	XtSetArg(al[1],XmNlistItems,&list);
	XtGetValues(ol_ptr->widget,al,2);
	if (*number!=0) {
		newlist=(char **)malloc( (*number+1) * sizeof(char *));
		for(iter=0;iter<*number;iter++) {
			XmStringGetLtoR(list[iter],
						XmSTRING_DEFAULT_CHARSET,&text);
			newlist[iter]=(char *)malloc(strlen(text)+1);
			strcpy(newlist[iter],text);
		}
		newlist[iter]=NULL;
		return(newlist);
	} else
		return(NULL);
}


/*
 * dialog_optionlist_textvalue dialog
 *
 * Returns the current text entered by the user
 *
 * INPUTS :	dialog		The optionlist dialog
 * RETURNS:	The ascii string of the text entered.
 */
char *
dialog_optionlist_textvalue(ol_ptr)
dialog ol_ptr;
{
	Arg al[10];			/* arg list */
	char *namestr;

	XtSetArg(al[0],XmNvalue,&namestr);
	XtGetValues(XmSelectionBoxGetChild(ol_ptr->widget,XmDIALOG_TEXT),al,1);
	return(namestr);
}




/*
 * CB_textentered widget
 *
 * Handles the callback when data is entered by the user
 *
 * INPUTS :	widget		The initiating widget
 * RETURNS:	None
 */
static void
CB_textentered(w)
Widget w;		/* widget id */
{
	Arg al[10];		/* arg list */
	d_optionl *ol_ptr;
	char *name;

	XtSetArg(al[0],XmNuserData,&ol_ptr);
	XtGetValues(w,al,1);

	if (ol_ptr->user.select) {
		ol_ptr->user.select( (dialog)ol_ptr,ol_ptr->user.callvalue);
		/*
		 * Don't look at ol_ptr after this point - select() may have
		 * deleted it!
		 */
	}
	XmProcessTraversal(w,XmTRAVERSE_NEXT_TAB_GROUP);
}




/*
 * CB_list_item_selected widget
 *
 * Handles the callback when an item is selected from the list
 *
 * INPUTS :	widget		The initiating widget
 * RETURNS:	None
 */
static void
CB_list_item_selected(w)
Widget w;		/* widget id */
{
	Arg al[10];		/* arg list */
	d_optionl *ol_ptr;

	XtSetArg(al[0],XmNuserData,&ol_ptr);
	XtGetValues(w,al,1);

	if (ol_ptr->user.select) {
		ol_ptr->user.select( (dialog)ol_ptr,ol_ptr->user.callvalue);
	       /*
		* Don't look at ol_ptr after this point - select() may have
		* deleted it!
		*/
	}
	XmProcessTraversal(w,XmTRAVERSE_NEXT_TAB_GROUP);
}



/*
 * dialog_optionlist_d dialog
 *
 * Destroys an optionlist, freeing all the necessary resources.
 *
 * INPUTS :	dialog		The dialog to destroy
 * RETURNS:	None
 */
static void
dialog_optionlist_d(ol_ptr)
d_optionl *ol_ptr;
{
	int iter;

	for(iter=0;iter<ol_ptr->nlist;iter++)  {
		free(ol_ptr->list[iter]);
		XmStringFree(ol_ptr->xmlist[iter]);
	}
	free(ol_ptr->list);
	free(ol_ptr->xmlist);
}



/*
 * dialog_optionlist dialog new_option
 *
 * Creates a option list dialog
 *
 * INPUTS :	dialog		The parent dialog 
 *		nopt		The new option list structure
 * RETURNS:	A pointer to the newly created dialog
 */
dialog
dialog_optionlist(parent,nopt)
dialog parent;
new_optionlist *nopt;
{
	Arg al[10];	/* arg list */
	XmString text1;	/* the text we create (since we need to free it) */
	XmString text2;	/* the text we create (since we need to free it) */
	int iter;
	d_optionl *ol_ptr;
	int n;
	int num;
	char **str;


	ol_ptr=(d_optionl *)dialog_init(parent,sizeof(d_optionl));

	str=nopt->list_item;
	num=0;
	while (*str) {
		str++;
		num++;
	    }
	bcopy(nopt,&(ol_ptr->user),sizeof(new_optionlist));
	
	ol_ptr->list=(char **)malloc(sizeof(char *)*(num+1));
	ol_ptr->xmlist=(XmString *)malloc(sizeof(XmString)*(num+1));
	ol_ptr->nlist=num;

	ol_ptr->fname=nopt->fname;
	for(iter=0;iter<num;iter++) {
		ol_ptr->list[iter]=(char *)malloc(
					strlen(nopt->list_item[iter])+1);
		strcpy(ol_ptr->list[iter],nopt->list_item[iter]);
		ol_ptr->xmlist[iter]=(XmString)XmStringCreateLtoR(
						nopt->list_item[iter],
						XmSTRING_DEFAULT_CHARSET);
	}
	ol_ptr->list[iter]=NULL;

	n=0;
	XtSetArg(al[n],XmNselectionLabelString,
		text1=XmStringCreate(nopt->selname,
					XmSTRING_DEFAULT_CHARSET)); n++;
	XtSetArg(al[n],XmNlistLabelString,	
		text2=XmStringCreate(nopt->listname,
					XmSTRING_DEFAULT_CHARSET));n++;
	XtSetArg(al[n],XmNtextColumns,nopt->width);n++;
	if (num!=0) {
		XtSetArg(al[n],XmNlistItems, ol_ptr->xmlist);n++;
	} else {
		XtSetArg(al[n],XmNlistItems, NULL);n++;
	}
	XtSetArg(al[n],XmNlistItemCount,num);n++;
	XtSetArg(al[n],XmNlistVisibleItemCount,nopt->height);n++;
	XtSetArg(al[n],XmNresizePolicy,XmRESIZE_NONE);n++;

	ol_ptr->widget=XmCreateSelectionBox(parent->widget,"name",al,n);
	XtAddCallback(XmSelectionBoxGetChild (ol_ptr->widget, XmDIALOG_LIST),
		 XmNbrowseSelectionCallback,CB_list_item_selected,0);
	XtAddCallback(XmSelectionBoxGetChild (ol_ptr->widget, XmDIALOG_TEXT),
		 XmNactivateCallback,CB_textentered,0);
	XtAddCallback(XmSelectionBoxGetChild (ol_ptr->widget, XmDIALOG_TEXT),
		 XmNvalueChangedCallback,findname, NULL);
	/*
	 * User data of text & list widgets points to the d_optionl structure
	 */
	XtSetArg(al[0],XmNuserData,ol_ptr);
	XtSetValues(XmSelectionBoxGetChild (ol_ptr->widget,XmDIALOG_TEXT),al,1);
	XtSetValues(XmSelectionBoxGetChild (ol_ptr->widget,XmDIALOG_LIST),al,1);
	wid_selbox_remove(ol_ptr->widget,5,XmDIALOG_SEPARATOR,
		XmDIALOG_OK_BUTTON,
		XmDIALOG_CANCEL_BUTTON,
		XmDIALOG_APPLY_BUTTON,
		XmDIALOG_HELP_BUTTON);
	if (ol_ptr->user.flags & OPTION_LIST_NOLABEL)
		wid_selbox_remove(ol_ptr->widget,2,
				XmDIALOG_TEXT,XmDIALOG_SELECTION_LABEL);
	if (ol_ptr->user.flags & OPTION_LIST_NOTEXT)
		wid_selbox_remove(ol_ptr->widget,2,
				XmDIALOG_TEXT,XmDIALOG_TEXT);

	/*
	 * If a font was specified for the list set it now
	 */
	if (ol_ptr->fname) {
		XtSetArg(al[0],XmNfontList,fname_2_fontlist(ol_ptr->fname));
		XtSetValues(XmSelectionBoxGetChild(ol_ptr->widget,
						XmDIALOG_LIST),al,1);
	}

	XmStringFree(text1);
	XmStringFree(text2);
	ol_ptr->destroy=dialog_optionlist_d;
	return((dialog)ol_ptr);
}



/*
 * dialog_optionlist_revise dialog items
 *
 * Sets a new list of options for a option list dialog
 *
 * INPUTS :	dialog		The dialog to modify
 *		items		Array of ascii strings (NULL terminated)
 * RETURNS:	None
 */
void
dialog_optionlist_revise(d,items)
dialog d;
char **items;
{
	d_optionl *ol_ptr=(d_optionl *)d;
	Arg al[10];				/* arg list */
	int ac;
	int iter;
	int num;
	char **str;


	/*
	 * Compute number of items in array
	 */
	num=0;
	str=items;
	while (*str) {
		str++;
		num++;
	}

	/*
	 * Free off resources used by current list
	 */
	for(iter=0;iter<ol_ptr->nlist;iter++)  {
		free(ol_ptr->list[iter]);
		XmStringFree(ol_ptr->xmlist[iter]);
	}
	free(ol_ptr->list);
	free(ol_ptr->xmlist);
	
	/*
	 * Create new list
	 */
	ol_ptr->list=(char **)malloc(sizeof(char *)*(num+1));
	ol_ptr->xmlist=(XmString *)malloc(sizeof(XmString)*(num+1));
	ol_ptr->nlist=num;

	for(iter=0;iter<num;iter++) {
		ol_ptr->list[iter]=(char *)malloc(strlen(items[iter])+1);
		strcpy(ol_ptr->list[iter],items[iter]);
		ol_ptr->xmlist[iter]=(XmString)XmStringCreateLtoR(items[iter],
						 XmSTRING_DEFAULT_CHARSET);
	}
	ol_ptr->list[iter]=NULL;
	XmListDeselectAllItems(ol_ptr->widget);
	XtSetArg(al[0],XmNitems,ol_ptr->xmlist);
	XtSetArg(al[1],XmNitemCount,num);
	ac=2;
	if (ol_ptr->fname) {
		XtSetArg(al[ac],XmNfontList,fname_2_fontlist(ol_ptr->fname));
		ac++;
	}
	XtSetValues(XmSelectionBoxGetChild(ol_ptr->widget,XmDIALOG_LIST),al,ac);
}
