/*
 *    X SHARE MANAGER (XSM)
 *
 *    Copyright (C) 1995  David Simons
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * $Id: d_shell.c,v 2.4 1995/06/06 21:49:12 davidsi Exp $
 *
 * $Log: d_shell.c,v $
 * Revision 2.4  1995/06/06  21:49:12  davidsi
 * - Removed dialog_shell_unmanage(). This is no longer necessary as
 *   the XtUnmapWidget() call in dialog_shell_d() will remove the
 *   shell hierachy from the display and not generate warning messages
 *   even if the modal information has been changed.
 *
 * Revision 2.3  1995/05/18  21:09:46  davidsi
 * - Added dialog_shell_unmanage() which allows a shell dialog to be marked
 *   as requiring a XtUnmanageChild() call when being destroyed. Essentially
 *   this is a work around, because the widget state is set such that
 *   the XtDestroyWidget() won't until the main loop is executed and that's
 *   not always convenient!  Universally calling XtUnmanageChild() also won't
 *   work, because interacts with XtAdd/RemoveGrab().
 * - BUG FIX. dialog_shell() now uses XtAppCreateShell() to create shells for
 *   main windows, so that they can be fully featured.
 *
 * Revision 2.2  1995/02/25  21:08:33  davidsi
 * - Shells are now initially positioned at 100,100 rather than 0,0.  This
 *   looks much tidier.
 *
 * Revision 2.1  1995/02/05  17:50:55  davidsi
 * - BUG FIX. When destroying a shell dialog, first unmanage it.  This
 *   ensures it disappears from the display regardless of whether
 *   the XtDestroyWidget stalls.
 *
 * Revision 2.0  1995/01/19  22:12:21  davidsi
 * - Added SHELL_SCROLL to allow shells with scrollbars
 *
 * Revision 1.2  1993/01/20  21:48:06  davidsi
 * - Added missing includes for Xm/MainW.h and Xm/BulletinB.h.
 *
 * Revision 1.1  1993/01/20  21:02:01  davidsi
 * Initial revision
 *
 *
 */


/*
 * Include Files
 */
#include <X11/Intrinsic.h>
#include <X11/Shell.h>

#include <Xm/Xm.h>
#include <Xm/RowColumn.h>
#include <Xm/DialogS.h>
#include <Xm/MainW.h>
#include <Xm/BulletinB.h>



#define IN_DIALOGS

#include "d_base.h"
#include "d_shell.h"
#include "d_form.h"


/*
 * Variable Declarations 
 */
static char rcsid[]="$Id: d_shell.c,v 2.4 1995/06/06 21:49:12 davidsi Exp $";

extern d_root *base_node;

/*
 * dialog_shell_d shellw
 *
 * Destroys a shell dialog
 *
 * INPUTS :	shellw		The shell widget to destroy
 * RETURNS:	None
 */
static void
dialog_shell_d(parent)
dialog parent;
{
	d_shell *d=(d_shell *)parent;
	/*
	 * We unmap the widgets to make sure the disappear from the
	 * display -  there are some instances where the destroy
	 * won't happen immediately.
	 */
	XtUnmapWidget(d->widget);
	XtUnmapWidget(d->toplevel);
	XtDestroyWidget(d->toplevel);
}

/*
 * dialog_dshell_r shellw 
 *
 * Realizes a shell dialog
 *
 * INPUTS :	shellw		The shell widget to realize
 * RETURNS:	None
 */
static void
dialog_dshell_r(d)
d_shell *d;
{
	XtManageChild(d->widget);
        XtRealizeWidget(d->toplevel);
}




/*
 * dialog_shell parent name flags
 *
 * Creates a shell dialog.
 *
 * INPUTS :	parent		The parent dialog
 *		name		The name of the shell (displayed)
 *		flags		see SHELL_XXXX
 * RETURNS:	A pointer to the newly created dialog.	
 */
dialog
dialog_shell(parent,name,flags)
dialog parent;
char *name;
int flags;
{
	d_shell *newshell;
	Arg al[20];
	int ac;

	newshell=(d_shell *)dialog_init(parent,sizeof(d_shell));
	/*
    	 * Mark this as a base node
	 */
	newshell->privateflags=DIALOG_BASENODE;
	ac=0;
	XtSetArg(al[ac],XmNx,100); ac++;
        XtSetArg(al[ac],XmNy,100); ac++;

	if (flags & SHELL_MAINWINDOW) {
		newshell->toplevel = XtAppCreateShell (name, "name1",
                        applicationShellWidgetClass, base_node->display,NULL,0);
		ac=0;
		if (flags & SHELL_SCROLL) {
    			XtSetArg(al[0],XmNscrollingPolicy,XmAUTOMATIC);
			ac++;
		}
	 	newshell->widget=XmCreateMainWindow(newshell->toplevel,"",
					al,ac);
		newshell->realise=dialog_dshell_r;
	} else if (flags & SHELL_POPUP) {
		newshell->toplevel=XmCreateDialogShell(parent->widget,name,
								al,ac);
		newshell->widget=XmCreateBulletinBoard(newshell->toplevel,"",
					NULL,0);
	} else {
		printf("bad flag to main window\n");
		exit(1);
	}
	/*
	 * A shell destroy must kill the BulletinBoard not the shell
	 */
	newshell->destroy=dialog_shell_d;

	return((dialog)newshell);
}
