
/*
 *    X SHARE MANAGER (XSM)
 *
 *    Copyright (C) 1995  David Simons
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * $Id: d_text.c,v 2.6 1995/06/23 20:48:08 davidsi Exp $
 *
 * $Log: d_text.c,v $
 * Revision 2.6  1995/06/23  20:48:08  davidsi
 * - Increase the safety margin when computing the number of columns.
 *
 * Revision 2.5  1995/06/19  18:09:31  davidsi
 * - BUG FIX. With some X servers it is possible that the window is not
 *   correctly rendered when first exposed. Fixed by adding a new flag
 *   neverdrawn to the d_text dialog and by ensuring a complete
 *   update will be performed on the first expose if the flag is not
 *   cleared.
 *
 * Revision 2.4  1995/05/29  18:24:32  davidsi
 * - Revision 2.3 went too far.  text_manage() needs to know when its being
 *   called once the widget hierachy is realized since this can and does
 *   occur when the text has been updated.
 *
 * Revision 2.3  1995/05/19  18:59:24  davidsi
 * - Much tidying up and removal of unnecessary variables.
 * - BUG FIX. dialog_text() now uses XtAppCreateShell() to create shells for
 *   text windows, so that they can be fully featured.
 *
 * Revision 2.2  1995/04/25  21:00:45  davidsi
 * - BUG FIX.  When changing the number of columns to be displayed
 *   make sure that they never exceed the maximum size assigned to
 *   the horizontal scrollbar.
 *
 * Revision 2.1  1995/02/25  21:06:34  davidsi
 * - Added support for centering text.
 * - Tided up and switched to using tab identation.
 *
 * Revision 2.0  1995/01/31  22:00:40  davidsi
 * - Revised user interface. Field contents and their attributes are now
 *   stored outside of this dialog.  A callback is made to retrieve the
 *   appropriate information.
 * - Now supports colour!
 * - MTEXT_LOCK support removed. Any locking must occur in the calling
 *   code.
 *
 * Revision 1.4  1994/11/05  21:18:11  davidsi
 * - Must include d_form.h before using any references to the position
 *   structure, otherwise gcc will fail with a parse error.
 *
 * Revision 1.3  1994/07/14  20:19:33  davidsi
 * - BUG FIX. Many of the globally visible functions were incorrectly
 *   typed.
 *
 * Revision 1.2  1993/01/20  21:47:43  davidsi
 * - Need to include "d_base.p"
 *
 * Revision 1.1  1993/01/20  21:03:34  davidsi
 * Initial revision
 *
 *
 */

/*
 * Include files
 */
#define IN_DIALOGS  1

#include <stdio.h>
#include <macros.h>

#include <X11/Intrinsic.h>
#include <X11/Shell.h>

#include <Xm/Xm.h>
#include <Xm/DialogS.h>
#include <Xm/MainW.h>
#include <Xm/ScrollBar.h>
#include <Xm/DrawingA.h>

#include "d_text.h"
#include "d_form.h"
#include "d_base.h"

#include "d_base.p"

extern d_root  *base_node;
static char rcsid[]="$Id: d_text.c,v 2.6 1995/06/23 20:48:08 davidsi Exp $";


/*
 * text_redraw textdialog mode
 *
 * Redraws, from scratch, the text for the specified dialog.
 *
 * INPUTS :	textdialog	The text dialog to redraw
 *		mode		TRUE if area to be erased first, FALSE
 *				is an optimisation if area is known to
 *				be already clear.
 * RETURNS:	None
 */
static void
text_redraw(d_ptr,mode)
d_text *d_ptr;
int mode;
{
	int dummy;		/* used for unreq'd values from scroll bar */
	int iter;		/* count of position */
	int ypos;		/* current vertical position within window */
	int field;		/* current field number */
	int vpos;		/* position of vertical scroll bar  */
	char firsttime;
	char *stringptr;	/* current string for field */
	int stringlen;
	char *col;
	char *lastcol=NULL;
	int fattr;
	int pixbits;
	Arg al[10];
	int status;

	if (!d_ptr->onscreen)		/* must wait till we're on */
		return;

	d_ptr->accessfield(d_ptr->accessfieldv,-1,-1,NULL,NULL,&col,&status);
	/*
	 * If status is -1 then don't attempt anything more.
	 */
	if (status== -1)
		return;
	pixbits=fetchcolour(col,d_ptr->gc2);
	XtSetArg(al[0],XmNbackground,pixbits);
	XtSetValues(d_ptr->w_text2,al,1);

	XmScrollBarGetValues(d_ptr->w_vbar,&vpos,&dummy,&dummy,&dummy);
	ypos=0;
        /*
         * Erase the old image if necessary
         */
        if (mode)
                XFillRectangle(d_ptr->display,XtWindow(d_ptr->w_text2),
                                d_ptr->gc2,0,0,d_ptr->maxwidth,d_ptr->height);
	/*
	 * 'iter' is the pointer to the line of text we are going to show,
	 * if a header is desired then set the value initially to zero so
	 * as to display the first line, then warp speed to the text req'd
	 * as given by vpos.
	 */
	if (d_ptr->flags & MTEXT_HEADER) {
		iter=0;
		firsttime=1;
	} else {
		iter=vpos;
		firsttime=0;
	}
	for(;iter<min(vpos+d_ptr->nvisiblel,d_ptr->n_lines);iter++)  {
		ypos += d_ptr->charheight;
		for(field=0;field<d_ptr->fields;field++) {
			d_ptr->accessfield(d_ptr->accessfieldv,iter,field,
					&stringptr,&stringlen,&col,&fattr);
			if (lastcol!=col) {
				fetchcolour(col,d_ptr->gc);
				lastcol=col;
			}

			if (!stringptr)
				goto nodraw;
			if (fattr & FIELD_ATTR_LEFT)
				XDrawString(d_ptr->display,
					XtWindow(d_ptr->w_text2),
					d_ptr->gc,
					d_ptr->field_pos[field],
					ypos,stringptr,stringlen);
			else if (fattr & FIELD_ATTR_RIGHT) {
				int twidth=d_ptr->widths[iter*d_ptr->fields+field];
				XDrawString(d_ptr->display,
					XtWindow(d_ptr->w_text2),
					d_ptr->gc,
					d_ptr->field_pos[field+1]-d_ptr->spacing-twidth,
					ypos,stringptr,stringlen);
			} else { /* FIELD_ATTR_CENTRE */
				int twidth=d_ptr->widths[iter*d_ptr->fields+field];
				XDrawString(d_ptr->display,
					XtWindow(d_ptr->w_text2),
					d_ptr->gc,
					(d_ptr->field_pos[field+1]-d_ptr->spacing-d_ptr->field_pos[field]-twidth)/2,
					ypos,stringptr,stringlen);
			}
nodraw:
		}
		/* 
		 * Warp to real text after dumping header line
		 */
		if (firsttime) {
			iter=vpos;
			firsttime=0;
		}
	}
}




/*
 * text_move d_ptr
 *
 * Handles the redrawing of the onscreen image when the view area
 * moves.  Also needed for the first initial draw of the onscreen
 * image
 *
 * INPUTS :	d_ptr	The text widget to process
 * RETURNS:	None
 */
static void
text_move(pad,d_ptr)
char *pad;
d_text *d_ptr;
{
	int dummy;
	int hpos;		/* position of horizontal scroll bar */
	int vpos;		/* position of vertical scroll bar  */
	Arg al[10];

 	if (!d_ptr->onscreen)
		return;

	XmScrollBarGetValues(d_ptr->w_vbar,&vpos,&dummy,&dummy,&dummy);
	XmScrollBarGetValues(d_ptr->w_hbar,&hpos,&dummy,&dummy,&dummy);
	/*
	 * If window has moved horizontally move the underlying window, that
	 * will cause an expose event - at which point we update it
	 */
	if (d_ptr->neverdrawn) {
		d_ptr->neverdrawn=0;
		/*
		 * Note that we clear the window here using the width of
		 * the physical window rather than relying on the 
		 * text_redraw() function. This is because the text_redraw()
		 * function clears the "logical" window, which may 
		 * be smaller than the physical window.
		 */
                XFillRectangle(d_ptr->display,XtWindow(d_ptr->w_text2),
                                d_ptr->gc2,0,0,d_ptr->width,d_ptr->height);
		text_redraw(d_ptr,FALSE);
		XMoveWindow(d_ptr->display,XtWindow(d_ptr->w_text2), 0,0);
	} else if (d_ptr->lasthpos!=hpos) {
		XMoveWindow(d_ptr->display,XtWindow(d_ptr->w_text2),
				-hpos*d_ptr->charwidth,0);
		d_ptr->lasthpos=hpos;
	} else if (d_ptr->lastvpos!=vpos)  {
		text_redraw(d_ptr,TRUE);
		d_ptr->lastvpos=vpos;
	}
}



/*
 * text_recompute
 *
 * Recomputes the internal state if the number of lines to be displayed
 * changes, or the number of lines visible on the screen changes.
 *
 * INPUTS :	d_ptr		The text dialog
 * RETURNS:	None
 */
static void
text_recompute(d_ptr)
d_text *d_ptr;
{
	XWindowAttributes wa;
	Window window;
	int position;
	int dummy;
	int scroll_size;
	Arg al[10];
	char cleararea=0;	/* TRUE when area cleared & must be redrawn */
	int currentmax;

	XGetWindowAttributes(d_ptr->display,XtWindow(d_ptr->w_text),&wa);
	/*
	 * If window shrinks clear the entire window otherwise
	 * partial lines of text can get left 
	 */
	if (wa.width < d_ptr->width || wa.height<d_ptr->height) {
		XFillRectangle(d_ptr->display,XtWindow(d_ptr->w_text2),
				d_ptr->gc2,0,0,d_ptr->width,d_ptr->height);
		cleararea=1;
	}
	d_ptr->width=wa.width;
	d_ptr->height=wa.height;
	d_ptr->nvisiblel=wa.height/d_ptr->charheight;
	d_ptr->nvisiblec=wa.width/d_ptr->charwidth;
	XmScrollBarGetValues(d_ptr->w_vbar,&position,&dummy,&dummy,&dummy);
	/*
	 * NB: d_ptr->n_lines = max of scroll bar
	 */
	if (position>d_ptr->n_lines-d_ptr->nvisiblel)
		position=d_ptr->n_lines-d_ptr->nvisiblel;
	if (position<0)
		position=0;
	scroll_size=d_ptr->nvisiblel;
	if (scroll_size>d_ptr->n_lines)
		scroll_size=d_ptr->n_lines;
	XmScrollBarSetValues(d_ptr->w_vbar,position,scroll_size,1,
				d_ptr->nvisiblel-1,TRUE);
	XmScrollBarGetValues(d_ptr->w_hbar,&position,&dummy,&dummy,&dummy);
	/*
	 * NB: d_ptr->n_col = max of scroll bar
	 */
	if (position>d_ptr->n_cols-d_ptr->nvisiblec)
		position=d_ptr->n_cols-d_ptr->nvisiblec;
	if (position<0)
		position=0;
	scroll_size=d_ptr->nvisiblec;
	if (scroll_size>d_ptr->n_cols)
		scroll_size=d_ptr->n_cols;

	/*
	 * Ensure we don't get any silly messages!  If the new value
	 * is larger than the max, then set the max first, otherwise
	 * set the max afterwards.
	 */
	XtSetArg(al[0],XmNmaximum,&currentmax);
	XtGetValues(d_ptr->w_hbar,al,1);

	if (scroll_size>currentmax) {	
		currentmax=0;
		XtSetArg(al[0],XmNmaximum,d_ptr->n_cols);
		XtSetValues(d_ptr->w_hbar,al,1);
	}

	XmScrollBarSetValues(d_ptr->w_hbar,position,scroll_size,1,
				d_ptr->nvisiblec,TRUE);

	if (currentmax)  {
		XtSetArg(al[0],XmNmaximum,d_ptr->n_cols);
		XtSetValues(d_ptr->w_hbar,al,1);
	}

	/*
	 * if area shrinks we don't get an expose event, so redraw now
	 */
	if (cleararea)
		text_redraw(d_ptr,FALSE);		/* update the display */

}


/*
 * text_resize widget 
 *
 * Handles resize callbacks. Provides a wrapper to the other functions,
 * since the text dialog structure has to be located from the 
 * widget information that is passed back by Motif.
 *
 * INPUTS :	w 	The widget that resized
 * RETURNS:	None
 */
static void
text_resize(w)
Widget w;
{
	Window window;
	Arg al[10];
	d_text *d_ptr;

 	XtSetArg(al[0],XmNuserData,&d_ptr);
	XtGetValues(w,al,1);
	window=XtWindow(w);
	if (window==(Window)NULL || !d_ptr->onscreen)
		return;
	text_recompute(d_ptr);
}


/*
 * text_setup dialog
 *
 * Performs the initialisation of a text dialog, when it is realised.
 *
 * INPUTS :	dialog 		The dialog being realised
 * RETURNS:	None
 */
static void
text_setup(d_ptr)
d_text *d_ptr;
{
	int iter;
	int field;
	int dir,ascent,descent;
	XCharStruct overall;
	Arg al[10];
	int runwidth;
	int totwidth;
	int totcharlen;		/* max number of characters on current line */
	int vmax;		/* max on vertical scroll bar */
	char *stringptr;
	int stringlen;
	char *colr;

	if (d_ptr->lens)
		free(d_ptr->lens);
	if (d_ptr->widths)
		free(d_ptr->widths);

	d_ptr->lens=(int *)malloc(d_ptr->fields*d_ptr->n_lines*sizeof(int));
	d_ptr->widths=(int *)malloc(d_ptr->fields*d_ptr->n_lines*sizeof(int));
	d_ptr->onscreen=TRUE;
	for(field=0;field<d_ptr->fields;field++) {
		d_ptr->field_pos[field]=0;
		d_ptr->field_max[field]=0;
	}
	d_ptr->maxwidth=0;
	d_ptr->n_cols=0;
	for(iter=0;iter<d_ptr->n_lines;iter++) {
		totwidth=0;
		totcharlen=0;
		for(field=0;field<d_ptr->fields;field++) {
			d_ptr->accessfield(d_ptr->accessfieldv,iter,field,
					&stringptr,&stringlen,&colr,NULL);
			d_ptr->field_max[field]=max(d_ptr->field_max[field],
							stringlen);
			totcharlen += stringlen;
			d_ptr->lens[iter*d_ptr->fields+field]=stringlen;
			XTextExtents(d_ptr->font_info,stringptr,stringlen,&dir,
					&ascent,&descent,&overall);
			d_ptr->widths[iter*d_ptr->fields+field]=overall.width;
			runwidth=max(overall.width,d_ptr->field_pos[field]);
			totwidth += runwidth+d_ptr->spacing;
			d_ptr->field_pos[field]=runwidth;
		}
		d_ptr->n_cols=max(d_ptr->n_cols,totcharlen);
		d_ptr->maxwidth=max(d_ptr->maxwidth,totwidth);
	}
	d_ptr->n_cols += 5;		/* bump to give some safety */
	totwidth=0;
	for(field=0;field<d_ptr->fields;field++) {
		runwidth = totwidth;
		totwidth += d_ptr->field_pos[field] + d_ptr->spacing;
		d_ptr->field_pos[field]=runwidth;
	}
	/*
	 * Go one field too far so that one can use
	 * field_pos[nextfield]-d_ptr->spacing to find the rightmost
	 * position of the current field 
	 */
	d_ptr->field_pos[field+1]=totwidth;
	d_ptr->charheight=ascent+descent;
	d_ptr->ascent=ascent;
	d_ptr->charwidth=d_ptr->maxwidth/d_ptr->n_cols;
	/*
	 * Motif insists that the scroll bar size always lies within
	 * the scrollbar maximum value minus the scrollbar minimum value,
	 * this means when the number of lines increases the max value
 	 * must change first, but when the number of lines drop the
	 * scroll bar size must be changed first...Grhh..
	 */
	XtSetArg(al[0],XmNmaximum,&vmax);
	XtGetValues(d_ptr->w_vbar,al,1);
	if (d_ptr->n_lines < vmax)		/* lines shrinking */
		text_recompute(d_ptr);

	XtSetArg(al[0],XmNminimum,0);
	XtSetArg(al[1],XmNmaximum,d_ptr->n_lines);
	XtSetValues(d_ptr->w_vbar,al,2);

	if (d_ptr->n_lines >= vmax)		/* lines increasing */
		text_recompute(d_ptr);

	XtSetArg(al[0],XmNminimum,0);
	XtSetArg(al[1],XmNmaximum,d_ptr->n_cols);
	XtSetValues(d_ptr->w_hbar,al,2);
	/*
	 * Pass any widget to text_move() - it does not actually use the
	 * the widget but takes it as a parameter because it can be used
	 * as a callback.
	 */
	text_move(NULL,d_ptr);
}


/*
 * text_expose widget padding callbackstr
 *
 * Handles exposures on the text dialog. 
 *
 * INPUTS :	widget		The widget  initiating the callback
 *		callbackstr	The drawing area callback structure
 * RETURNS:	None
 */
static void
text_expose(w,a,b)
Widget w;
char *a;
XmDrawingAreaCallbackStruct *b;
{
	Window window;
	XExposeEvent *xevent=(XExposeEvent *)(b->event);
	Arg al[10];
	d_text *d_ptr;
	Pixel backgrnd;
	Pixel foregrnd;

	XtSetArg(al[0],XmNuserData,&d_ptr);
	XtGetValues(w,al,1);
	window=XtWindow(d_ptr->w_text);
	if (window!=(Window)NULL) {
		/*
		 * Is this the first expose event? If so set all the text up
		 * and draw it
		 */
		if (!d_ptr->onscreen) {
			XtSetArg(al[0],XmNbackground,&backgrnd);
			XtSetArg(al[1],XmNforeground,&foregrnd);
			XtGetValues(d_ptr->w_text2,al,2);

			d_ptr->gc=XCreateGC(d_ptr->display,
						XtWindow(d_ptr->w_text2),0,0);
			XSetBackground(d_ptr->display,d_ptr->gc,backgrnd);
			XSetForeground(d_ptr->display,d_ptr->gc,foregrnd);
			d_ptr->gc2=XCreateGC(d_ptr->display,
						XtWindow(d_ptr->w_text2),0,0);
			/*XSetForeground(d_ptr->display,d_ptr->gc2,backgrnd);*/

			d_ptr->font_info=fname_2_fontinfo(d_ptr->fname);
			XSetFont(d_ptr->display,d_ptr->gc,d_ptr->font_info->fid);
			text_setup(d_ptr);
			/*
			 * Otherwise if there are no other expose events pending
			 * redraw the display
			 */
		} else {
			if (xevent->count==0) 
				text_redraw(d_ptr,FALSE);/* update display */
		}
	}
}


/*
 * text_display dialog
 *
 * Handles the text dialog when it is realised
 *
 * INPUTS :	d_ptr	The text dialog 
 * RETURNS:	None
 */
static void
text_display(d_ptr)
d_text *d_ptr;
{
	Widget w_shell;
	Widget parent;
	
	/*
	 * If this is not the first time, simply set the text up again
	 */
	if (d_ptr->w_shell != NULL ) {
		XFillRectangle(d_ptr->display,XtWindow(d_ptr->w_text2),
			d_ptr->gc2,0,0,d_ptr->width,d_ptr->height);
		text_setup(d_ptr);
		text_redraw(d_ptr,FALSE);	/* update the display */
	}
}



/*
 * text_destroy dialog
 *
 * Frees all resources related to text dialogs
 *
 * INPUTS :	dialog		The text dialog
 * RETURNS:	None
 */
static void
text_destroy(d_ptr)
d_text *d_ptr;
{
 	XtDestroyWidget(d_ptr->w_shell);
	if (d_ptr->lens)
		free(d_ptr->lens);
	if (d_ptr->widths)
		free(d_ptr->widths);
}




/*
 * text_manage dialog
 *
 * Realises the text dialog
 *
 * INPUTS :	dialog		The dialog to realise
 * RETURNS:	None
 */
static void
text_manage(d_ptr)
d_text *d_ptr;
{
	/*
	 * This function does get called multiple times, because if the
	 * text being displayed is updated the realisation call will
	 * occur again.  However, on the second the current display needs
	 * to be updated rather than managed/realized so detect callbacks
	 * that are not the first and simply redraw the display.
	 */
	if (!d_ptr->firsttime) {
		text_display(d_ptr);
		return;
	}
	d_ptr->firsttime=0;

        XtManageChild(d_ptr->widget);
	XtManageChild(d_ptr->w_vbar);
	XtManageChild(d_ptr->w_hbar);
	XtManageChild(d_ptr->w_scroll);
	XtManageChild(d_ptr->w_text);
	XtManageChild(d_ptr->w_text2);
        XtRealizeWidget(d_ptr->w_shell);
}

/*
 * dialog_text_redraw dialog
 *
 * Redraws a text dialog
 *
 * INPUTS :	dialog		The text dialog which is to be redrawn
 * RETURNS:	None
 * 
 */
dialog_text_redraw(d)
dialog d;
{
	d_text *d_ptr=(d_text *)d;

	text_redraw(d_ptr,TRUE);	/* update the display */
}

void
dialog_text_setlines(d,li )
dialog d;
int li;
{
d_text *d_ptr=(d_text *)d;

	d_ptr->n_lines=li;
}


/*
 * dialog_text parent
 *
 * Creates a text dialog.
 *
 * INPUTS :	parent		The parent dialog
 *		nt		Pointer to new text structure
 * RETURNS:	A pointer to the newly created dialog.	
 */
dialog
dialog_text(parent,nt,p)
dialog parent;
new_d_text *nt;
char *p;
{
	d_text *d_ptr;
	int iter;
	Arg al[10];

	d_ptr=(d_text *)dialog_init(parent,sizeof(d_text));
	d_ptr->realise=text_manage;
	d_ptr->destroy=text_destroy;
	d_ptr->width=nt->width;
	d_ptr->height=nt->height;
	d_ptr->name=(char *)malloc(strlen(nt->name)+1);
	strcpy(d_ptr->name,nt->name);
	d_ptr->fname=(char *)malloc(strlen(nt->fname)+1);
	strcpy(d_ptr->fname,nt->fname);
	d_ptr->fields=nt->fields;
	d_ptr->spacing=nt->spacing;
	d_ptr->flags=nt->flags;
	d_ptr->fieldend='\t';
	d_ptr->max_n_lines=0;
	d_ptr->accessfield=nt->accessfield;
	d_ptr->accessfieldv=nt->accessfieldv;
	d_ptr->charheight=0;
	d_ptr->lasthpos=-1;
	d_ptr->lastvpos=-1;
	d_ptr->neverdrawn=1;
	d_ptr->display=XtDisplay(parent->widget);
	d_ptr->firsttime=1;
	XtSetArg(al[0],XmNx,100);
	XtSetArg(al[1],XmNy,100);
	d_ptr->w_shell=XtAppCreateShell(d_ptr->name, "name1",
			applicationShellWidgetClass, base_node->display,NULL,0);

	XtSetArg(al[0],XmNscrollBarPlacement,XmBOTTOM_RIGHT);
	XtSetArg(al[1],XmNvisualPolicy,XmVARIABLE);
	XtSetArg(al[2],XmNscrollingPolicy,XmAPPLICATION_DEFINED);
	XtSetArg(al[3],XmNwidth,d_ptr->width);
	XtSetArg(al[4],XmNheight,d_ptr->height);
	d_ptr->widget=d_ptr->w_scroll=
	XmCreateMainWindow(d_ptr->w_shell,"swindow",al,5);

	XtSetArg(al[0],XmNorientation,XmVERTICAL);
	XtSetArg(al[1],XmNshowArrows,TRUE);
	XtSetArg(al[2],XmNuserData,d_ptr);
	d_ptr->w_vbar=XmCreateScrollBar(d_ptr->w_scroll,"vscroll",al,3);

	XtSetArg(al[0],XmNorientation,XmHORIZONTAL);
	XtSetArg(al[1],XmNshowArrows,TRUE);
	XtSetArg(al[2],XmNuserData,d_ptr);
	d_ptr->w_hbar=XmCreateScrollBar(d_ptr->w_scroll,"hscroll",al,3);

	XtSetArg(al[0],XmNhorizontalScrollBar,d_ptr->w_hbar);
	XtSetArg(al[1],XmNverticalScrollBar,d_ptr->w_vbar);
	XtSetValues(d_ptr->w_scroll,al,2);

	XtSetArg(al[0],XmNuserData,d_ptr);
	d_ptr->w_text=XmCreateDrawingArea(d_ptr->w_scroll,"drawingarea",al,1);
	XtAddCallback(d_ptr->w_text,XmNresizeCallback,text_resize,0);

	XtSetArg(al[0],XmNwidth,1000);
	XtSetArg(al[1],XmNheight,768);
	XtSetArg(al[2],XmNuserData,d_ptr);
	d_ptr->w_text2=XmCreateDrawingArea(d_ptr->w_text,"drawingarea2",al,3);
	XtAddCallback(d_ptr->w_text2,XmNexposeCallback,text_expose,0);

	XmMainWindowSetAreas(d_ptr->w_scroll,NULL,NULL,d_ptr->w_hbar,
			d_ptr->w_vbar,d_ptr->w_text);
	XtAddCallback(d_ptr->w_vbar,XmNvalueChangedCallback,text_move,d_ptr);
	XtAddCallback(d_ptr->w_vbar,XmNdragCallback, text_move,d_ptr);
	XtAddCallback(d_ptr->w_hbar,XmNvalueChangedCallback,text_move,d_ptr);
	XtAddCallback(d_ptr->w_hbar, XmNdragCallback, text_move, d_ptr);
	
	return((dialog) d_ptr);
}
