
/*
 *    X SHARE MANAGER (XSM)
 *
 *    Copyright (C) 1995  David Simons
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


/*
 * $Id: d_text.h,v 2.4 1995/06/19 18:11:04 davidsi Exp $
 *
 * $Log: d_text.h,v $
 * Revision 2.4  1995/06/19  18:11:04  davidsi
 * - BUG FIX. With some X servers it is possible that the window is not
 *   correctly rendered when first exposed. Fixed by adding a new flag
 *   neverdrawn to the d_text dialog and by ensuring a complete
 *   update will be performed on the first expose if the flag is not
 *   cleared.
 *
 * Revision 2.3  1995/05/29  18:25:42  davidsi
 * - Revision 2.3 went too far.  text_manage() needs to know when its being
 *   called once the widget hierachy is realized since this can and does
 *   occur when the text has been updated.
 *
 * Revision 2.2  1995/05/18  21:13:18  davidsi
 * - Removed the ondisplay field from the d_text structure as it is now
 *   surplus to requirements.
 *
 * Revision 2.1  1995/02/25  21:07:19  davidsi
 * - Added FIELD_ATTR_CENTRE to allow text to be centered.
 *
 * Revision 2.0  1995/01/31  21:58:16  davidsi
 * - Removed the MTEXT_LOCK flag. It no longer serves any purpose.
 * - The field_attr[] array is no longer used. Instead a new function
 *   accessfield() returns the information about the various fields.
 * - Various tidying up.
 *
 * Revision 1.2  1993/01/20  22:21:58  davidsi
 * - Removed all function prototypes, now stored separately.
 *
 * Revision 1.1  1993/01/20  21:08:10  davidsi
 * Initial revision
 *
 *
 */

#ifndef _DIALOGTEXT_H
#define _DIALOGTEXT_H

#define MAX_FIELDS 20

typedef struct new_d_text {
	int fields;
	int spacing;
	int flags;
	char *name;
	char *fname;
	int width,height;
	void (*accessfield)(char *,int l,int f,char **s,int *sl,char **c,int *atr); 
	char *accessfieldv;
} new_d_text;

#ifdef IN_DIALOGS
#include "dialogcmn.h"




typedef struct d_text {
	DIALOGHEADER
	int fields;
	int field_pos[MAX_FIELDS+1];/* x pos for each field */
	int field_max[MAX_FIELDS];/* max number of chars in each used */
	int max_n_lines;	/* total number of lines that can be */
				/* held in 'lines' */
	int *lens;		/* lengths of each field per line */
	int *widths;		/* widths of text in each field per line */
	int spacing;		/* number of pixels for spacing between cols */
	int flags;		/* see above TEXTW_XXX defines */
	char *name;		/* name for outer shell */

	char fieldend;		/* character to terminate fields with */
	int n_lines;		/* number of lines in 'lines' array */
	int n_cols;		/* on longest line this is how */
				/* many chars there are */
	int charheight;		/* height of each line */
	int charwidth;		/* width of a char on longest line */
	int nvisiblel;		/* number of visible lines */
	int nvisiblec;		/* number of visible columns */
	int ascent;		/* ascent of character set */

	int width;		/* width of physical window */
	int height;		/* width of physical window */
	int maxwidth;		/* width in pixels of longest line */
	int lasthpos;		/* stop moving windows when not req'd */
	int lastvpos;		/* stop redrawing window when not req'd */

	char onscreen;
	char neverdrawn;
	Display *display;	/* display where target windows are */
				/* to be found */
	char firsttime;		/* TRUE if realization has not yet happened */
	XFontStruct *font_info;
	char *fname;
	Widget w_shell;
	Widget w_text;
	Widget w_text2;
	Widget w_scroll;
	Widget w_hbar,w_vbar;
	GC gc;
	GC gc2;
	void (*accessfield)(char *,int l,int f,char **s,int *sl,char **col,int *attr); 
	char *accessfieldv;
} d_text;

#else

#endif

/*
 * Defines for flags field  and for mtext_create()
 */
#define MTEXT_HEADER	1	/* first entry in lines array is a heading */


/*
 * values for field_attr field
 */
#define FIELD_ATTR_LEFT		1
#define FIELD_ATTR_RIGHT	2
#define FIELD_ATTR_CENTRE	2

#endif /* _DIALOGTEXT_H */
