%{

/*
 *    X SHARE MANAGER (XSM)
 *
 *    Copyright (C) 1995  David Simons
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * $Id: smwblex.l,v 2.3 1995/06/17 20:30:26 davidsi Exp $
 *
 * $Log: smwblex.l,v $
 * Revision 2.3  1995/06/17  20:30:26  davidsi
 * - Added TK_NOT token.
 * - Added the // comment introducer so comments can be specified on a line
 *   by line basis.
 * - Added TK_BREAK and TK_CONTINUE tokens.
 *
 * Revision 2.2  1995/06/02  19:54:27  davidsi
 * - Added FOR keyword.
 * - BUG FIX. When a comment was being skipped the count of source lines
 *   was not kept up to date.
 *
 * Revision 2.1  1995/05/29  19:07:34  davidsi
 * - Rule for '\n' now increments the global variable source_line.  This allows
 *   better information to be dumped with errors.
 *
 * Revision 2.0  1995/01/19  21:28:41  davidsi
 * - Initial version.
 *
 * Revision 3.0  1993/01/23  20:27:17  davidsi
 * - First version of lexical analyser
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "node.h"

#include "smwbyacc.tab.h"

char *yytext;
FILE *yyin;
int source_line;

void skip_comment();
void skip_line();
char *malloc_symbol(char *);
extern int yyparse();
extern "C" int _filbuf(...);


%}

D	[0-9]
A	[a-z]
L	[a-zA-Z]
DL	[a-zA-Z0-9_]

%%
\n		{source_line++;}
\t		{}
" "		{}
auto		return(TK_AUTO);
static		return(TK_STATIC);
begin		return(TK_BEGIN);
byref		return(TK_BYREF);
byvalue		return(TK_BYVALUE);
typedef		return(TK_TYPEDEF);
struct		return(TK_STRUCT);
\{		return(TK_LBRACE);
\}		return(TK_RBRACE);
\(		return(TK_BRA);
\)		return(TK_KET);
\[		return(TK_SBRA);
\]		return(TK_SKET);
\;		return(TK_SEMICOLON);
\:		return(TK_COLON);
\,		return(TK_COMMA);
\&\&		return(TK_LAND);
\&		return(TK_AND);
\|\|		return(TK_LOR);
\|		return(TK_OR);
\^		return(TK_XOR);
\+		return(TK_PLUS);
\-		return(TK_MINUS);
\*		return(TK_MULTIPLY);
\/		return(TK_DIVIDE);
\>		return(TK_GT);
\<		return(TK_LT);
\>=		return(TK_GE);
\<=		return(TK_LE);
\=\=		return(TK_EQ);
!=		return(TK_NE);
\=		return(TK_ASSIGN);
\.		return(TK_PERIOD);
\!		return(TK_NOT);
\"		{
		char string[1024];
		int iter;

			for(iter=0;iter<1024;iter++) {
				string[iter]=yyinput();
				if (string[iter]=='\\') {
					string[iter+1]=yyinput();
					if (string[iter+1]=='n') {
						string[iter]='\n';
						source_line++;
					} else 
						iter++;
				}
				if (string[iter]=='\"') {
					string[iter]='\0';
					yylval.strval=malloc_symbol(string);
					return(TK_STRING);
				}
			}
			fprintf(stderr,"String length too long\n");
			exit(1);
		}
if		return(TK_IF);
else		return(TK_ELSE);
return		return(TK_RETURN);
for		return(TK_FOR);
while		return(TK_WHILE);
do		return(TK_DO);
continue	return(TK_CONTINUE);
break		return(TK_BREAK);
{D}+\.{D}+	{
			yylval.realval=atof(yytext);
			return(TK_REAL);
		}
{D}+		{
			yylval.intval=atoi(yytext);
			return(TK_INTEGER);
		}
{L}{DL}*	{
			yylval.strval=malloc_symbol(yytext);
			return(TK_SYMBOL);
		}
"/*"		skip_comment();
"//"		skip_line();
%%

/*
 * Allocates a symbol value ready for passing back
 */
char *malloc_symbol(char *s)
{
char *temp;

	if ((temp=(char *)malloc(strlen(s)+1))==NULL) {
		fprintf(stderr,"Cannot allocate memory (malloc_symbol)\n");
		exit(1);
	}
	strcpy(temp,s);
	return(temp);
}

/*
 * skip_comment
 *
 * Skips past comments
 */
void
skip_comment()	
{
	char a;

	for(;;) {
		while ( (a=yyinput()) !='*')
			if (a=='\n')
				source_line++;
		if ( (a=yyinput())!='/') {
			unput(a);
		} else
			return;
	}
}

/*
 * skip_lines
 *
 * Skips lines
 */
void
skip_line()	
{
	char a;

	while ( (a=yyinput()) !='\n')
		;
	source_line++;
}
