%{

/*
 *    X SHARE MANAGER (XSM)
 *
 *    Copyright (C) 1995  David Simons
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* 
 * $Id: smwbyacc.y,v 2.3 1995/06/17 20:39:10 davidsi Exp $
 *
 * $Log: smwbyacc.y,v $
 * Revision 2.3  1995/06/17  20:39:10  davidsi
 * - Added rule for the TK_NOT token using the already existent node_expr
 *   class.
 * - Added rules for the TK_CONTINUE token (continue) and TH_BREAK
 *   (break).
 * - The node_while class no longer supports do..while type constructs
 *   so change the construction calls to use the new node_dowhile class.
 *
 * Revision 2.2  1995/06/02  19:55:37  davidsi
 * - Added support for a "for(..;..;..)" construct based on the already
 *   existing while and compound node types.
 *
 * Revision 2.1  1995/05/29  19:05:32  davidsi
 * - Removed code body for yyerror(). Let the caller handle the error
 *   situation in a more user friendly way.
 *
 * Revision 2.0  1995/01/19  21:29:03  davidsi
 * - Initial version.
 *
 * 
 */
#include <stdio.h>

#include "node.h"
#include "context.h"
#include "error.h"

//
// This is the node from which the parse tree hangs
//
class node_header *basenode;


yyerror(char *foo)
{
}

extern int yylex();

extern errstream cerror;
extern diagstream cinfo;

#ifdef DEBUG_TOKEN_STREAM
/*
 * Not got a clue why we have to subratract one of the input stream
 */
int yylex2()
{
	int a;
	a=yylex();
	printf("TOKEN = %d\n",a);
	return( a );
}

#define yylex	yylex2
#else
#define yylex	yylex
#endif



%}

%union {
        long    intval;
        double	realval;
        char *strval;
	class node *nodeptr;
	class node_declaration *nodedecptr;
	class node_typedecl *nodetypptr;	// type defn ptr 
	class node_structdefn *nodestcptr;	// structure defn ptr
	class node_funcdecl *nodefncptr;	// function decl node
	class node_command *nodecmdptr;		// command node
	argmode_e argm;
	operators_e op;
}




/*
 * The real tokens..
 */
%token	<intval>	TK_BYREF	/* BYREF */
%token	<intval>	TK_BYVALUE	/* BYVALUE */
%token	<intval>	TK_TYPEDEF	/* typedef */
%token	<intval>	TK_STRUCT	/* struct */
%token	<intval>	TK_BEGIN	/* begin */
%token	<intval>	TK_INTEGER	/* integer */
%token	<realval>	TK_REAL		/* real */
%token	<strval>	TK_SYMBOL	/* symbol */
%token	<intval>	TK_LBRACE	/* { */
%token	<intval>	TK_RBRACE	/* } */
%token	<intval>	TK_BRA		/* ( */
%token	<intval>	TK_KET		/* ) */
%token	<intval>	TK_SBRA		/* [ */
%token	<intval>	TK_SKET		/* ] */
%token 	<intval>	TK_SEMICOLON	/* ; */
%token 	<intval>	TK_COLON	/* : */
%token 	<intval>	TK_COMMA	/* , */
%token	<intval>	TK_IF		/* if */
%token	<intval>	TK_ELSE		/* else */
%token	<intval>	TK_WHILE	/* while */
%token	<intval>	TK_DO		/* do */
%token	<intval>	TK_FOR		/* for */
%token	<intval>	TK_BREAK	/* break */
%token	<intval>	TK_CONTINUE	/* continue */
%token	<intval>	TK_RETURN	/* return */
%token	<strval>	TK_STRING	/* quoted string */
%token	<intval>	TK_LAND		/* && */
%token	<intval>	TK_AND		/* & */
%token	<intval>	TK_LOR		/* || */
%token	<intval>	TK_OR		/* | */
%token	<intval>	TK_XOR		/* ^ */
%token	<intval>	TK_PLUS		/* + */
%token	<intval>	TK_MINUS	/* - */
%token	<intval>	TK_DIVIDE	/* / */
%token	<intval>	TK_MULTIPLY	/* * */
%token	<intval>	TK_GT		/* > */
%token	<intval>	TK_LT		/* < */
%token	<intval>	TK_GE		/* >= */
%token	<intval>	TK_LE		/* <= */
%token	<intval>	TK_EQ		/* == */
%token	<intval>	TK_NE		/* != */
%token	<intval>	TK_ASSIGN	/* = */
%token	<intval>	TK_AUTO		/* auto */
%token	<intval>	TK_STATIC	/* static */
%token	<intval>	TK_PERIOD	/* . */
%token	<intval>	TK_NOT		/* ! */

%type	<nodedecptr>	funcdecl
%type	<nodedecptr>	declaration
%type	<nodedecptr>	argdecl
%type	<nodedecptr>	vardecl
%type	<nodefncptr>	funcbody
%type	<nodefncptr>	functype
%type	<nodetypptr>	type_specifier
%type	<nodetypptr>	struct_or_union_specifier
%type	<nodestcptr>	struct_declaration_list
%type	<nodestcptr>	struct_declaration
%type	<strval>	struct_declarator_list
%type	<strval>	struct_declarator
%type	<strval>	declarator
%type	<nodedecptr>	vartype
%type	<argm>		aclass
%type	<nodecmdptr>	actions
%type	<nodecmdptr>	command
%type	<nodeptr>	expr
%type   <nodeptr>	cond_expr
%type   <nodeptr>	log_or_expr
%type   <nodeptr>	log_and_expr
%type   <nodeptr>	inc_or_expr
%type   <nodeptr>	exc_or_expr
%type   <nodeptr>	and_expr
%type   <nodeptr>	equality_expr
%type   <nodeptr>	relational_expr
%type   <nodeptr>	shift_expr
%type   <nodeptr>	additive_expr 
%type   <nodeptr>	multiplicative_expr
%type   <nodeptr>	cast_expr
%type   <nodeptr>	unary_expr
%type   <nodeptr>	postfix_expr
%type   <nodeptr>	primary_expr
%type	<op>		relop
%type	<nodeptr>	arglist


%start header
%%
header	: funcdecl 
	{
		class node_header *nptr=new node_header($1);
		cinfo << "header processed\n";
		nptr->resolve(NULL);
		basenode=nptr;
		yyerrok; 
	}
	;

/*
// Handles the parsing of function declarations.  These are taken to have
// the form:
//
// function_name {
// actions statements
// }
//
*/
funcdecl: funcdecl declaration 
	{
		cinfo << "funcdecl (1 - recursive) processed\n";
		$2->setnext($1);
		$$=$2;	
	}
	| declaration
	{
		cinfo << "funcdecl (1 - termination) processed\n";
		$$=$1;	
	}
	;

declaration: functype TK_BRA TK_KET funcbody
	{
		class node_funcdecl *nptr=new node_funcdecl($1,$4,NULL);
		cinfo << "declaration (1 - func no args) processed\n";
		$$=nptr;
	}
	| functype TK_BRA argdecl TK_KET funcbody
	{
		class node_funcdecl *nptr=new node_funcdecl($1,$5,$3);
		cinfo << "declaration (2 - func with args) processed\n";
		$$=nptr;
	}
	| vartype
	{
		cinfo << "declaration (3 - vardecl) processed\n";
                $$=$1;
	}
	| struct_or_union_specifier TK_SEMICOLON
 	{
		$$=$1;
	}
	;	

functype : TK_SYMBOL
	{
                class node_funcdecl *nptr=new node_funcdecl(NULL,$1);
                cinfo << "declaration (2 - func with args) processed\n";
                $$=nptr;
	}
	| type_specifier TK_SYMBOL
	{
                class node_funcdecl *nptr=new node_funcdecl($1,$2);
                cinfo << "declaration (2 - func with args) processed\n";
                $$=nptr;
	}
	; 

funcbody: TK_LBRACE vardecl actions TK_RBRACE
	{
                class node_funcdecl *nptr=new node_funcdecl($2,$3);
                cinfo << "declaration (2 - func with args) processed\n";
                $$=nptr;
	}
	| TK_LBRACE actions TK_RBRACE
	{
                class node_funcdecl *nptr=new node_funcdecl(NULL,$2);
                cinfo << "declaration (2 - func with args) processed\n";
                $$=nptr;
	}
	;
	

/* * argdecl - arguments are separated by commas but we still build 
 * a vardecl node! */ 
argdecl:aclass type_specifier TK_SYMBOL
        {
                class node_vardecl *nptr=new node_vardecl($3,$2,$1);
                cinfo << "argdecl (1) processed\n";
                $$=nptr;
	}
        | aclass type_specifier TK_SYMBOL TK_COMMA argdecl
        {
                class node_vardecl *nptr=new node_vardecl($3,$2,$1);
                cinfo << "argdecl(2) processed\n";
		nptr->setnext($5);
                $$=nptr;
        }
        ;

aclass:	{
		cinfo << "aclass (1 - default) processed\n";
		$$=arg_value;
	}
	| TK_BYREF
	{
		cinfo << "aclass(2 - ref declaration) processed\n";
		$$=arg_ref;
	}
	| TK_BYVALUE
	{
		cinfo << "aclass(3 - value declaration) processed\n";
		$$=arg_value;
	}
	;


/*
 *
// Note that funcdecl has to handle vardecl type rules to prevent
// specific reduce/reduce errors. If you add rules here
// it is likely similar rules are needed for funcdecl.
 */

vardecl : vartype
	{
		cinfo << "vardecl(1 - termination) processed\n";
		$$=$1;	
	}
	| vardecl vartype
	{
		cinfo << "vardecl(1 - recursive) processed\n";
		$2->setnext($1);
		$$=$2;
	}
	;
	

vartype: TK_TYPEDEF type_specifier TK_SYMBOL TK_SEMICOLON 
	{
		class node_typedef *nptr=new node_typedef($2,$3,0);
		cinfo << "vardecl(1 - typedef) processed\n";
                $$=nptr;
	}
	| TK_TYPEDEF type_specifier TK_SYMBOL TK_SBRA TK_INTEGER TK_SKET TK_SEMICOLON
	{
		class node_typedef *nptr=new node_typedef($2,$3,$5);
		cinfo << "vardecl(2 - typedef for arrays) processed\n";
                $$=nptr;
	}
        | type_specifier TK_SYMBOL TK_SEMICOLON
        {
		class node_vardecl *nptr=new node_vardecl($2,$1,arg_not);
		cinfo << "vartype(3 - variable) processed\n";
                $$=nptr;
	}
	;

type_specifier: TK_SYMBOL
	{
		class node_typedecl *nptr=new node_typedecl(type_symbol,$1);
		cinfo << "type_specifier(1 - type name) processed\n";
                $$=nptr;
	}
	| struct_or_union_specifier
	{
		cinfo << "type_specifier(1 - struct/union) processed\n";
                $$=$1;
	}
	;

struct_or_union_specifier: TK_STRUCT TK_SYMBOL TK_LBRACE struct_declaration_list TK_RBRACE
	{
		class node_typedecl *nptr=new node_typedecl(type_struct_defn,$2);
		cinfo << "struct_or_union_specifier(1 - struct defn)\n";
		nptr->setstructure($4);
		$$=nptr;

	}
	| TK_STRUCT TK_SYMBOL 
	{
		class node_typedecl *nptr=new node_typedecl(type_struct,$2);
		cinfo << "struct_or_union_specifier(2 - struct name)\n";
                $$=nptr;
	}
	;

struct_declaration_list: struct_declaration
	{
		$$=$1
	}
	| struct_declaration_list struct_declaration
	{
		$2->setnext($1);
                $$=$2;
	}
	;

struct_declaration: type_specifier struct_declarator_list TK_SEMICOLON
	{
		class node_structdefn *nptr=new node_structdefn($2,$1);
		$$=nptr;
	}

struct_declarator_list: struct_declarator
	{
		$$=$1;
	}
	;

struct_declarator: declarator
	{
		$$=$1;
	}
	;

declarator: TK_SYMBOL
	{
		$$=$1;
	}
	;


actions : actions command
	{
		class node_compound *nptr=new node_compound($1,$2);
		cinfo << "actions(1 - iteration) processed\n";
		$$=nptr;	
	}
	| command
        {
		cinfo << "actions(2 - command) processed\n";
		$$=$1;
        }
	;

command	: TK_LBRACE actions TK_RBRACE
	{
		cinfo << "command(1 - compound) processed\n";
		$$=$2;
	}	
	| TK_IF TK_BRA expr TK_KET command TK_ELSE command
	{
		class node_if *nptr=new node_if($3,$5,$7);
		cinfo << "command(2 - if..else) processed\n";
		$$=nptr;
	}
	| TK_IF TK_BRA expr TK_KET command
	{
		class node_if *nptr=new node_if($3,$5,NULL);
		cinfo << "command(3 - if) processed\n";
		$$=nptr;
	} 
	| TK_WHILE TK_BRA expr TK_KET command
	{
		class node_while *nptr=new node_while($3,$5);
		cinfo << "command(while) processed\n";
		$$=nptr;
	} 
	| TK_FOR TK_BRA expr TK_SEMICOLON expr TK_SEMICOLON expr TK_KET command
	{
		/*
		 * For the moment we must wrap "expr" nodes before 
		 * enclosing them in a compound node because we require
		 * a node than can convert an execute call to an
		 * evaluate call
		 */
		class node_eval *etoc1=new node_eval($7);
		class node_compound *nptr2=new node_compound($9,etoc1);

		class node_while *nptr1=new node_while($5,nptr2);

		class node_eval *etoc2=new node_eval($3);
		class node_compound *nptr=new node_compound(etoc2,nptr1);
		cinfo << "command(for) processed\n";
		$$=nptr;
	}
	| TK_DO command TK_WHILE TK_BRA expr TK_KET TK_SEMICOLON
	{
		class node_dowhile *nptr=new node_dowhile($5,$2);
		cinfo << "command(while) processed\n";
		$$=nptr;
	} 
	|  expr TK_SEMICOLON
	{
		class node_eval *nptr=new node_eval($1);
		$$=nptr;
		cinfo << "command(3 - expr) processed\n";
	}
	|  TK_RETURN TK_BRA expr TK_KET TK_SEMICOLON
	{
		class node_return *nptr=new node_return($3);
		cinfo << "command(return()) processed\n";
		$$=nptr;
	}
	|  TK_RETURN TK_SEMICOLON
	{
		class node_return *nptr=new node_return(NULL);
		cinfo << "command(return) processed\n";
		$$=nptr;
	} 
	| TK_CONTINUE TK_SEMICOLON
	{
		class node_continue *nptr=new node_continue();
		cinfo << "command(continue) processed\n";
		$$=nptr;
	}
	| TK_BREAK TK_SEMICOLON
	{
		class node_break *nptr=new node_break();
		cinfo << "command(break) processed\n";
		$$=nptr;
	}
	;

expr:	cond_expr
	{
		$$=$1
	}
	|
	unary_expr TK_ASSIGN  expr
	{
		class node_assign *nptr=new node_assign($1,$3);
		cinfo << "command(3 - assign) processed\n";
		$$=nptr;
	}
	;

cond_expr : log_or_expr
	{
		$$=$1;
	}

log_or_expr : log_and_expr
	{
		$$=$1
	}
	| log_or_expr TK_LOR log_and_expr
	{
		class node_expr *nptr=new node_expr(op_lor,$1,$3);
                cinfo << "bfactor(3 - relop) processed\n";
                $$=nptr;
	}
	;

log_and_expr : inc_or_expr
	{
		$$=$1
	}
	| log_and_expr TK_LAND inc_or_expr
	{
		class node_expr *nptr=new node_expr(op_land,$1,$3);
                cinfo << "bfactor(3 - relop) processed\n";
                $$=nptr;
	}
	;

inc_or_expr: exc_or_expr
	{
		$$=$1;
	}
	|
	inc_or_expr TK_OR exc_or_expr
	{
		class node_expr *nptr=new node_expr(op_or,$1,$3);
                cinfo << "bfactor(3 - relop) processed\n";
                $$=nptr;
	}
	;

exc_or_expr: and_expr
	{
		$$=$1;
	}
	| exc_or_expr TK_XOR and_expr
	{
		class node_expr *nptr=new node_expr(op_xor,$1,$3);
                cinfo << "bfactor(3 - relop) processed\n";
                $$=nptr;
	}
	;

and_expr : equality_expr
	{
		$$=$1;
	}
	|
	and_expr TK_AND equality_expr 
	{
		class node_expr *nptr=new node_expr(op_and,$1,$3);
                cinfo << "bfactor(3 - relop) processed\n";
                $$=nptr;
	}
	;

equality_expr: relational_expr
	{
		$$=$1
	}
	| equality_expr TK_EQ relational_expr
	{
		class node_relop *nptr=new node_relop(comp_eq,$1,$3);
                cinfo << "bfactor(3 - relop) processed\n";
                $$=nptr;
	}
	| equality_expr TK_NE relational_expr
	{
		class node_relop *nptr=new node_relop(comp_ne,$1,$3);
                cinfo << "bfactor(3 - relop) processed\n";
                $$=nptr;
	}
	;

relational_expr: shift_expr 
	{
		$$=$1;
	}
	| 
	relational_expr relop shift_expr
	{
		class node_relop *nptr=new node_relop($2,$1,$3);
                cinfo << "bfactor(3 - relop) processed\n";
                $$=nptr;
	}
	;
	

relop	: TK_GT
	{
		$$=comp_gt;
	}
	| TK_GE
	{
		$$=comp_ge;
	}
	| TK_LT
	{
		$$=comp_lt;
	}
	| TK_LE
	{
		$$=comp_le;
	}
	;
	
shift_expr: additive_expr 
	{
		$$=$1;
	}
	;

additive_expr: multiplicative_expr
	{
		$$=$1;
	}
	| additive_expr TK_PLUS multiplicative_expr
	{
		class node_expr *nptr=new node_expr(op_add,$1,$3);
		cinfo << "expr(2 - plus) processed\n";
		$$=nptr;
	}
	| additive_expr TK_MINUS multiplicative_expr
	{
		class node_expr *nptr=new node_expr(op_sub,$1,$3);
		cinfo << "expr(2 - minus) processed\n";
		$$=nptr;
	}
	;

multiplicative_expr	: cast_expr
	{
		$$=$1;
	}
	| multiplicative_expr TK_MULTIPLY cast_expr
	{
		class node_expr *nptr=new node_expr(op_mul,$1,$3);
		cinfo << "term(1 - multiply) processed\n";
		$$=nptr;
	}
	| multiplicative_expr TK_DIVIDE cast_expr
	{
		class node_expr *nptr=new node_expr(op_div,$1,$3);
		cinfo << "term(2 - divide) processed\n";
		$$=nptr;
	}
	;

cast_expr: unary_expr 
	{
		$$=$1;
	}
	;

unary_expr: postfix_expr
	{
		$$=$1;
	}
	| TK_NOT cast_expr
	{
		class node_expr *nptr=new node_expr(op_not,$2,NULL);
		cinfo << "unary_expr(1 - not) processed\n";
		$$=nptr;
	}
	;


postfix_expr: postfix_expr TK_SBRA expr TK_SKET
	{
		class node_deref *nptr=new node_deref($1,$3);
		cinfo << "postfix_expr (1 - deref)\n";
		$$=nptr;
	}
	| postfix_expr TK_PERIOD TK_SYMBOL
	{
		class node_field *nptr=new node_field($1,$3);
		cinfo << "postfix_expr (2 - field access)\n";
		$$=nptr;
	}
	| postfix_expr TK_BRA arglist TK_KET
	{
		class node_call *nptr=new node_call($1,$3);
		cinfo << "postfix_expr (3 - call)\n";
		$$=nptr;
	}
	| primary_expr
	{
		cinfo << "postfix_expr (4 - nop)\n";
		$$=$1
	}
	;
	

				
primary_expr :  TK_BRA expr TK_KET
	{
		cinfo << "primary_expr:(expr)\n";
		$$=$2;
	}
	| TK_INTEGER
	{
		class node_int *nptr=new node_int($1);
		cinfo << "factor(1 - integer) processed\n";
		$$=nptr;
	}
	| TK_MINUS TK_INTEGER
	{
		class node_int *nptr=new node_int(-$2);
		cinfo << "factor(2 - neg integer) processed\n";
		$$=nptr;
	}
	| TK_REAL
	{
		class node_real *nptr=new node_real($1);
		cinfo << "factor(1 - integer) processed\n";
		$$=nptr;
	}
	| TK_MINUS TK_REAL
	{
		class node_real *nptr=new node_real(-$2);
		cinfo << "factor(2 - neg integer) processed\n";
		$$=nptr;
	}
	| TK_STRING
	{
		class node_string *nptr=new node_string($1);
		cinfo << "factor(3 - string) processed\n";
		$$=nptr;
	}
	| TK_SYMBOL
	{
		class node_var *nptr=new node_var($1);
		cinfo << "factor(4 - variable) processed\n";
		$$=nptr;
	}
	;
			
arglist :
	{
		cinfo << "arglist(1 - no list) processed\n";
		$$=NULL;
	}
	| expr TK_COMMA arglist
	{
		class node_arg *nptr=new node_arg($1,$3);
		cinfo << "arglist(2 - list) processed\n";
		$$=nptr;
	}
	| expr
	{
		class node_arg *nptr=new node_arg($1,NULL);
		cinfo << "arglist(3 - 1 arg) processed\n";
		$$=nptr;
	}
	;
