/***********************************************************************
 *                                                                     *
 *              WILLOW(tm) Information Retrieval Software              *
 *     (Washington Information Looker-upper Layered Over Windows)      *
 *             Copyright 1992-1994 University of Washington            *
 *                                                                     *
 *   Willow is a trademark of the University of Washington.  No        *
 *   commercial use of this trademark or software may be made without  *
 *   prior written permission of the University of Washington.         *
 *                                                                     *
 * Permission to use, copy, modify, and distribute Willow(tm) software *
 * and its documentation for any noncommercial purpose is hereby       *
 * granted, provided that the above copyright notice appears in all    *
 * copies and that both the above copyright notice and this permission *
 * notice appear in supporting documentation, and that the name of the *
 * University of Washington not be used in advertising or publicity    *
 * pertaining to distribution of the software without specific,        *
 * written prior permission.  This software is made available "as is", *
 * and THE UNIVERSITY OF WASHINGTON DISCLAIMS ALL WARRANTIES, EXPRESS  *
 * OR IMPLIED, WITH REGARD TO THIS SOFTWARE, INCLUDING WITHOUT         *
 * LIMITATION ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS    *
 * FOR A PARTICULAR PURPOSE, AND IN NO EVENT SHALL THE UNIVERSITY OF   *
 * WASHINGTON BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL     *
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA  *
 * OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, TORT (INCLUDING       *
 * NEGLIGENCE) OR STRICT LIABILITY, ARISING OUT OF OR IN CONNECTION    *
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.                       *
 *                                                                     *
 * For further information please contact:                             *
 * Willow Project                                                      *
 * Information Systems                                                 *
 * University of Washington                                            *
 * 4545 15th NE, 1st floor                                             *
 * Seattle, Washington, 98105, USA                                     *
 * Internet: willow@cac.washington.edu                                 *
 ***********************************************************************/

#include <config.h>

/* license.c
   This module combines the C routines together to produce the
   license interface.
 */

#ifdef DEBUG
/* debugging in the bison parsers */
#define YYDEBUG 1
int yydebug = 1;
#endif

#include <stdio.h>
#include <ctype.h>
#ifdef SOLARIS
#include <string.h>
#else /* SOLARIS */
#include <strings.h>
#endif /* SOLARIS */
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#ifdef _AIX
#include <sys/select.h>
#endif

#include <errno.h>
#include <netdb.h>

#ifdef LICENSE_SERVER
#  include <search.h>
#endif

/* define prototypes for license calls */
#include <license.h>

/* define prototypes for network functions */
#include <net.h>

/* define database globals. */
#ifdef LICENSE_SERVER
#  include <database.h>
#endif

/* define error codes */
#include <license_status.h>

/* define prototypes for parsing functions  */
#include <lex.h>

#ifdef LICENSE_SERVER

/* define prototypes for local db query routines */
#  include <local.h>

#endif

/*****  include C Routines  **********************************/

/* the interface routines */
#include "interface.c"

/* the network interface routines */
#include "net.c"

#ifdef LICENSE_SERVER

/* the local database query routines */
#include "db.c"
#include "local.c"

/* the database parser generated from database.y */
/* the static parsing functions called by the database and command parsers */
#include "database.c"
#include "lex.c"

#endif
