/***********************************************************************
 *                                                                     *
 *              WILLOW(tm) Information Retrieval Software              *
 *     (Washington Information Looker-upper Layered Over Windows)      *
 *             Copyright 1992-1994 University of Washington            *
 *                                                                     *
 *   Willow is a trademark of the University of Washington.  No        *
 *   commercial use of this trademark or software may be made without  *
 *   prior written permission of the University of Washington.         *
 *                                                                     *
 * Permission to use, copy, modify, and distribute Willow(tm) software *
 * and its documentation for any noncommercial purpose is hereby       *
 * granted, provided that the above copyright notice appears in all    *
 * copies and that both the above copyright notice and this permission *
 * notice appear in supporting documentation, and that the name of the *
 * University of Washington not be used in advertising or publicity    *
 * pertaining to distribution of the software without specific,        *
 * written prior permission.  This software is made available "as is", *
 * and THE UNIVERSITY OF WASHINGTON DISCLAIMS ALL WARRANTIES, EXPRESS  *
 * OR IMPLIED, WITH REGARD TO THIS SOFTWARE, INCLUDING WITHOUT         *
 * LIMITATION ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS    *
 * FOR A PARTICULAR PURPOSE, AND IN NO EVENT SHALL THE UNIVERSITY OF   *
 * WASHINGTON BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL     *
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA  *
 * OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, TORT (INCLUDING       *
 * NEGLIGENCE) OR STRICT LIABILITY, ARISING OUT OF OR IN CONNECTION    *
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.                       *
 *                                                                     *
 * For further information please contact:                             *
 * Willow Project                                                      *
 * Information Systems                                                 *
 * University of Washington                                            *
 * 4545 15th NE, 1st floor                                             *
 * Seattle, Washington, 98105, USA                                     *
 * Internet: willow@cac.washington.edu                                 *
 ***********************************************************************/

#ifndef _LICENSE_H_
#define _LICENSE_H_

#include <sys/types.h>
#include <time.h>

/* Maximum number of strings in ticket profile */
#define LICENSE_PROFILE      10
/* This is to avoid problems with buggy malloc() */
#define MIN_MALLOC            4
/* Token maximum size used in database parsing */
/* For instance, this limits the length of a profile string */
#define LICENSE_MAX_TOKEN 20480


struct ticket_str {
  long int    seq;
  time_t      connect;
  time_t      idle;
  time_t      update;
  time_t      reclaim;
  char        *license;
  char        *assoc;
  char        *profile[LICENSE_PROFILE +1];
};

int get_ticket(struct ticket_str **ticket, char *service, char *class,
               char *assoc);
int update_ticket(struct ticket_str *ticket);
int return_ticket(struct ticket_str *ticket);
int validate_ticket(char *key,char *assoc);
char *license_message(int code);
char **license_server(char **list);

struct ticket_str *str2ticket(char *buffer, int buflen);
char *ticket2str(struct ticket_str *ticket);

#ifdef LICENSE_SERVER
int license_init_db(char *db);
int freeze_tickets(char *fname);
int thaw_tickets(char *fname);
int reclaim_tickets(time_t now);
#endif

#endif /* _LICENSE_H_ */
