/***********************************************************************
 *                                                                     *
 *              WILLOW(tm) Information Retrieval Software              *
 *     (Washington Information Looker-upper Layered Over Windows)      *
 *             Copyright 1992-1994 University of Washington            *
 *                                                                     *
 *   Willow is a trademark of the University of Washington.  No        *
 *   commercial use of this trademark or software may be made without  *
 *   prior written permission of the University of Washington.         *
 *                                                                     *
 * Permission to use, copy, modify, and distribute Willow(tm) software *
 * and its documentation for any noncommercial purpose is hereby       *
 * granted, provided that the above copyright notice appears in all    *
 * copies and that both the above copyright notice and this permission *
 * notice appear in supporting documentation, and that the name of the *
 * University of Washington not be used in advertising or publicity    *
 * pertaining to distribution of the software without specific,        *
 * written prior permission.  This software is made available "as is", *
 * and THE UNIVERSITY OF WASHINGTON DISCLAIMS ALL WARRANTIES, EXPRESS  *
 * OR IMPLIED, WITH REGARD TO THIS SOFTWARE, INCLUDING WITHOUT         *
 * LIMITATION ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS    *
 * FOR A PARTICULAR PURPOSE, AND IN NO EVENT SHALL THE UNIVERSITY OF   *
 * WASHINGTON BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL     *
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA  *
 * OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, TORT (INCLUDING       *
 * NEGLIGENCE) OR STRICT LIABILITY, ARISING OUT OF OR IN CONNECTION    *
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.                       *
 *                                                                     *
 * For further information please contact:                             *
 * Willow Project                                                      *
 * Information Systems                                                 *
 * University of Washington                                            *
 * 4545 15th NE, 1st floor                                             *
 * Seattle, Washington, 98105, USA                                     *
 * Internet: willow@cac.washington.edu                                 *
 ***********************************************************************/

#define _NET_C_
/* net.c

   This file contains auxilliary routines which are
   called by the LICENSE applications interface routines.
   These routines are not intended to be called by the
   applications programmer.

   Globals are: license_servers[]
                license_open()
                license_close()
                license_send()
                license_recv()
                license_net_opt()

   These routines are called from lib/interface.c and server/licensed.c
 */
static int License_Socket = -1;
static int License_Timeout[4] = {1,3,6,-1};

#include <license_servers.h>

#ifdef __alpha
typedef int netint; /* 32 bit network integer */
#else
typedef long int netint; /* 32 bit network integer */
#endif

/* LICENSE_OPEN
   Open connection to LICENSE network server
 */
int license_open(primary_flag)
int primary_flag;
{

  struct sockaddr_in sa_in;  /* Socket address in internet style */
  register struct hostent *host = 0;
  char **server;

  int     len, sock;
  char    buf[80];
  struct  servent *sp;
  struct  timeval timeout;
  fd_set  fds;
  char   *name_in = LICENSE_SERVICE;
  int     rc;
  int     cycle;

  /*
  **  Check to see if already open.
  */
  if ( License_Socket != -1 ) {
    return(LICENSE_SUCCESS);
  }

  /*
  **  Get service number
  */
  if ((sp = getservbyname(name_in,"tcp"))!= NULL) {
    sa_in.sin_port = sp->s_port;
#ifdef HESIOD
  } else if ((sp = hes_getservbyname(name_in, "tcp")) != NULL) {
    sa_in.sin_port = sp->s_port;
#endif /* HESIOD */
  } else {
#ifdef LICENSE_PORT_DEFAULT
    sa_in.sin_port = htons(LICENSE_PORT_DEFAULT);
#else /* LICENSE_PORT_DEFAULT */
    return(LICENSE_NETWORK_ERROR);
#endif /* LICENSE_PORT_DEFAULT */
  }


  /*
  **  Attempt to connect to a server
  */

  /* get socket */
  if ((sock = socket(AF_INET,SOCK_DGRAM,0)) < 0) {
    return(LICENSE_NETWORK_ERROR);
  }

  /* Repeat datagram according to Timeout schedule */
  for ( cycle = 0; ; cycle++ ) {

    if ( License_Timeout[cycle] < 0 ) {  /* no response from server */
      close(sock);
      return(LICENSE_SERVER_FAILURE);
    }

    /* send to each server */
    for (server = license_servers; *server; server++) {
      sa_in.sin_addr.s_addr = inet_addr(*server);
      if (sa_in.sin_addr.s_addr != -1) {
        sa_in.sin_family = AF_INET;
      } else {
        if (host = gethostbyname(*server)) {
          sa_in.sin_family = host->h_addrtype;
#ifndef BCOPY    
          memcpy((caddr_t)&sa_in.sin_addr, host->h_addr, host->h_length);
#else
          bcopy(host->h_addr, (caddr_t)&sa_in.sin_addr, host->h_length);
#endif
        } else {
          close(sock);
          return(LICENSE_NETWORK_ERROR);
        }
      }
      strcpy(buf,"p\n");
      len = strlen(buf);
#ifdef XA
      cmxlate(buf,ebcdictoascii,len);
#endif
      if (sendto(sock,buf,len,0,&sa_in,sizeof(sa_in)) != len) {
        close(sock);
        return(LICENSE_NETWORK_ERROR);
      }
    }

    /*
    **  Now check to see if we got any "bytes".
    */

    len = sizeof(sa_in);
    do {
      FD_ZERO(&fds);
      FD_SET(sock,&fds);
      timeout.tv_sec = License_Timeout[cycle];
      timeout.tv_usec = 0;
      rc = select(sock+1,&fds,0,0,&timeout);
      if ( rc == 0 ) goto nextcycle;  /* timed out */
      recvfrom(sock,buf,sizeof(buf),0,&sa_in,&len);
#ifdef XA
      asciitoebcdic[10] = '\n';    /*SJ    yecch */
      cmxlate(buf,asciitoebcdic,len);
#endif
    } while (primary_flag && *buf != 'p' && *buf != 'P');
    break;       /* received good response */
  nextcycle:     /* did not get good response */
    continue;
  } 

  close(sock);
  if ( primary_flag == 2 ) return(LICENSE_SUCCESS);

  if ((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
    return(LICENSE_NETWORK_ERROR);
  }
  if (connect(sock, (struct sockaddr *)&sa_in, sizeof (sa_in)) < 0) {
    return(LICENSE_NETWORK_ERROR);
  }

  License_Socket = sock;
  return(LICENSE_SUCCESS);
}

/* LICENSE_SEND
   Send query to LICENSE network server
 */
static int license_send(qstring)
char *qstring;
{
  int rc, flags;
  netint qlen, querylen;

  qlen = strlen(qstring);
  querylen = htonl(qlen);
  flags = 0;
  rc = send(License_Socket,(char *)&querylen,sizeof(netint), flags);
  if ( rc != sizeof(netint) ) {
    perror("license_send");
    license_close();
    return(LICENSE_NETWORK_ERROR);
  }
#ifdef XA
  {
    char *qbuf;
    qbuf = (char *)malloc(qlen < MIN_MALLOC ? MIN_MALLOC : qlen+1);
    if (qbuf==NULL) return(LICENSE_ERROR_ALLOC);
    strcpy(qbuf,qstring);
    cmxlate(qbuf,ebcdictoascii,qlen);
    rc = send(License_Socket, qbuf, qlen, flags);
    free(qbuf);
  }
#else
  rc = send(License_Socket, qstring, qlen, flags);
#endif
  if ( rc != qlen ) {
    perror("license_send");
    license_close();
    return(LICENSE_NETWORK_ERROR);
  }
  return(LICENSE_SUCCESS);
}

/* LICENSE_RECV
   Receive message from LICENSE network server
 */
static int license_recv(buf,bufsiz)
char **buf;
int *bufsiz;
{
  netint buflen;
  int rc, flags, len;
  char *buffer, *bp;

  /* test validity of args */
  *bufsiz = 0;
  *buf = NULL;

  /* get length */
  len = sizeof(buflen);

  flags = 0;
  rc = recv(License_Socket,(char *)&buflen, len, flags);
  if ( rc != len ) {
    license_close();
    return(LICENSE_ERROR_NREAD);
  }
  buflen = ntohl(buflen);

  if ( buflen == 0 ) return(LICENSE_FAILURE);

  buffer = (char *) malloc(buflen < MIN_MALLOC ? MIN_MALLOC : buflen+1);
  if ( buffer == NULL ) {
    return(LICENSE_ERROR_ALLOC);
  }

  for ( bp = buffer, len = buflen;  0 < len; bp += rc, len -= rc ) {
    rc = recv(License_Socket, bp,len, flags);
    if ( rc < 0) {
      license_close();
      free(buffer);
      return(LICENSE_ERROR_NREAD);
    }
  }
  *bp = '\0';
  *bufsiz = buflen;
  *buf = buffer;
#ifdef XA
  cmxlate(buffer,asciitoebcdic,buflen);
#endif
  return(LICENSE_SUCCESS);
}

/* LICENSE_CLOSE
   Close the connection to the network server
 */
int license_close()
{
  if ( License_Socket != -1 ) {
    close(License_Socket);
    License_Socket = -1;
  }
  return(LICENSE_SUCCESS);
}


/* LICENSE_NET_OPT
   Set network options
 */
int license_net_opt(option,value)
long int option,value;
{
  switch(option) {
    case LICENSE_OPTION_TIMEOUT :
        License_Timeout[0] = value;
        License_Timeout[1] = -1;
        return(LICENSE_SUCCESS);
    default:
        return(LICENSE_FAILURE);
  }
}
