/***********************************************************************
 *                                                                     *
 *              WILLOW(tm) Information Retrieval Software              *
 *     (Washington Information Looker-upper Layered Over Windows)      *
 *             Copyright 1992-1994 University of Washington            *
 *                                                                     *
 *   Willow is a trademark of the University of Washington.  No        *
 *   commercial use of this trademark or software may be made without  *
 *   prior written permission of the University of Washington.         *
 *                                                                     *
 * Permission to use, copy, modify, and distribute Willow(tm) software *
 * and its documentation for any noncommercial purpose is hereby       *
 * granted, provided that the above copyright notice appears in all    *
 * copies and that both the above copyright notice and this permission *
 * notice appear in supporting documentation, and that the name of the *
 * University of Washington not be used in advertising or publicity    *
 * pertaining to distribution of the software without specific,        *
 * written prior permission.  This software is made available "as is", *
 * and THE UNIVERSITY OF WASHINGTON DISCLAIMS ALL WARRANTIES, EXPRESS  *
 * OR IMPLIED, WITH REGARD TO THIS SOFTWARE, INCLUDING WITHOUT         *
 * LIMITATION ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS    *
 * FOR A PARTICULAR PURPOSE, AND IN NO EVENT SHALL THE UNIVERSITY OF   *
 * WASHINGTON BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL     *
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA  *
 * OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, TORT (INCLUDING       *
 * NEGLIGENCE) OR STRICT LIABILITY, ARISING OUT OF OR IN CONNECTION    *
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.                       *
 *                                                                     *
 * For further information please contact:                             *
 * Willow Project                                                      *
 * Information Systems                                                 *
 * University of Washington                                            *
 * 4545 15th NE, 1st floor                                             *
 * Seattle, Washington, 98105, USA                                     *
 * Internet: willow@cac.washington.edu                                 *
 ***********************************************************************/

/***************************************************************************
 * host.c
 * This file contains the routines that are specific to the UW's BRS
 * interface. 
 ***************************************************************************/

#include <stdio.h>
#ifdef SOLARIS
#include <string.h>
#else /* SOLARIS */
#include <strings.h>
#endif /* SOLARIS */
#include <ctype.h>
#ifdef vax
typedef long size_t;
#endif
#include "driver.h"
#include "host.h"
#include "protocol.h"
#include "hash.h"

#ifdef ibm032
#define strchr index
#define strrchr rindex
#endif

extern char *parse_buf;
extern char *db_name;
extern state_table states;
extern int state;
extern char logging;
extern char *last_sent_req;

/* Global variables needed in this file for keeping track of searches */
int num_results = 0;         /* How many summaries expected */
int search_num = 0;          /* reference number for this result set */

char *summary_fields = 0;    /* list of fields we use for summaries */
hash_tbl *summ_form_tab;     /* hash table of field format records */

/* keep vectors of field format record pointers around for packaging of
   result packets */
format_rec *title_form_vec[128];
format_rec *summ_form_vec[128];
format_rec *ext_form_vec[128];

int  summary_long_len = 0;   /* How wide are the expanded field names */
int  field_cnt = 0;          /* How many unique fields in summary record */
int  title_field_cnt = 0;    /* How many fields in the summ title */
int  summ_field_cnt = 0;     /* How many fields in the summ body */
int  ext_field_cnt = 0;      /* How many fields in the summ extension */
int  cur_record = 0;         /* Which record we are displaying */
int  summ_total = 0;         /* How many summaries have been retrieved */
int  num_summ_allocs = 0;    /* How many times have we allocated */

/* Are we returning expanded or short fields for full records? */
char expanded_to_willow;

/* These buffers must be out here, cuz they get copied to table */
char get_full_cmd[MAX_LINE]; /* Buffer for full-command */
char next_summ_cmd[MAX_LINE];/* buffer for sprinting command to get summ.*/
char search[MAX_QUERY_LEN];  /* search buffer */

/* A table for holding search commands from the boolean fields, which
   are in queue to be executed */
search_desc_rec *limit_queue = NULL;

/* Flag for whether circulation is done for this database */
extern char do_circ;

/* Table to hold parsed summaries. We start counting at 1. Hold
   all of them that we retrieve until a new search is launched.
   Necessary to support III circulation querying */
summ_tab_entry *summ_tab;

extern state_table states;           /* The state table */
extern char *circ_host;              /* III machine for circ queries */

/* External routines */
extern char *append_cr (char *s);
extern int atoi ( char *);
extern int cleanup (void);
extern int close_willow (void);
extern int got_info (char *info);
extern int error (char *msg);
extern char *get_circ(int num);
extern int hl_end (int point);
extern int hl_start (int point);
extern int host_send (char *str);
extern char *list_lookup(char *db_name,
			 char *list_name, 
			 char search_type, 
			 char *prefix_st, 
			 int lines_wanted, 
			 int *line_pos, 
			 int line_offset);
extern int send_browse(int ref_num, char *list_data, int line_pos);
extern int log_trans (int type, char *comment);
extern int misc_warning_hcb (char *msg);
extern int send_full_record (int num, char *text);
extern int send_stats_cmd (char *stats, int num_results, int search_num);
extern int send_title (int num, char *title);
extern void send_ext(int num, char *title);
extern int set_alarm (int time_out);
extern void *malloc (size_t);
extern void *realloc (void *, size_t);
extern void bad_search(char *msg);
extern void close_circ (void);
extern void force_transition (int which);
extern void free (void *);
extern send_fields_info(int info_cnt, char *info, int len);
extern void send_summ_record (int num, char *text);
extern void send_user_req(int type, char *query_text, char *query_tag,
		   char *yes_label, char *no_label);

/* forward decls. */
char *build_title(int num);
char *build_ext(int num);
char *build_summ(int num, char abbreviated);
char *make_range(search_desc_rec *search);
char  strain(char ch, char *command);
int alloc_summ_tab (void);
int bad_search_hcb (void);
int end_of_field (char *str);
int extract_num (char *str, int sdex);
int fix_periods (char *word, char any_field);
int free_limit_queue (void);
int init_summ_tab (void);
int lookup_field(char *str);
int new_record (char *str);
int queue_cmd (char *cmd, int next_state);
int start_of_field (char *str);
void parse_info_hcb(void);
void lcat_titlefy (char *title);
void lcat_callnumify (char *cnum);
int  hash_compare(char *key, format_rec *element);

/* Do any host-specific initialization. Just zero our summary table */
init_host(void)
{
  alloc_summ_tab();
}

/* Get the first block. We may have to realloc it later. */
alloc_summ_tab(void)
{
  num_summ_allocs++;
  if (num_summ_allocs == 1) {
    /* +1 cuz we start counting at 1, not 0. */
    summ_tab = (summ_tab_entry *) 
      malloc(sizeof(summ_tab_entry) * (SUMM_ALLOC_SIZE + 1));
#if defined(POSIX) || defined (_POSIX_SOURCE)
    memset(summ_tab, 0, sizeof(summ_tab_entry) * (SUMM_ALLOC_SIZE + 1));
#else
    bzero(summ_tab, sizeof(summ_tab_entry) * (SUMM_ALLOC_SIZE + 1));
#endif
  }
  else {
    summ_tab = (summ_tab_entry *) 
      realloc(summ_tab, sizeof(summ_tab_entry) * (num_summ_allocs *
	      SUMM_ALLOC_SIZE + 1));
#if defined(POSIX) || defined (_POSIX_SOURCE)
    memset(summ_tab + SUMM_ALLOC_SIZE * (num_summ_allocs - 1) + 1, 0,
	  sizeof(summ_tab_entry) * SUMM_ALLOC_SIZE);
#else
    bzero(summ_tab + SUMM_ALLOC_SIZE * (num_summ_allocs - 1) + 1,
	  sizeof(summ_tab_entry) * SUMM_ALLOC_SIZE);
#endif
  }
}

/* Save login id, password, and database name into appropriate slots 
   in the state table Save summary fields as global. 
 */
save_login_info(char *uid, char *pw, char *db_name, char *title_fields, char *summ_fields, char *ext_fields)
{
  char *cmd;
  int sdex;
  format_rec *summ_form;
  char field[8];

  if (uid) {
    cmd = malloc(strlen(UID_CMD) + strlen(uid) + 1);
    sprintf(cmd, UID_CMD, uid);
    free(uid);
    states[AWAIT_LOGIN][0].out_string = cmd;
  }
  else {
    /* We are going to have to hang at AWAIT_LOGIN while user sends back uid */
    states[AWAIT_LOGIN][0].next_state = AWAIT_LOGIN;
  }

  if (pw)
    states[AWAIT_PW][0].out_string = append_cr(pw);
  else {
    /* We are going to have to hang at AWAIT_PW while user sends back pw */
    states[AWAIT_PW][0].next_state = AWAIT_PW;
  }

  if (db_name) {
    cmd = malloc(strlen(DB_CMD) + strlen(db_name) + 1);
    sprintf(cmd, DB_CMD, db_name);
    states[AWAIT_DB][0].out_string = cmd;

    if (logging)
      states[AWAIT_DB][3].out_string = CPR_LOG_CMD;
    else
      states[AWAIT_DB][3].out_string = CPR_CMD;

#ifdef ndef
    if (logging) {
      cmd = malloc(strlen(CPR_LOG_CMD) + strlen(db_name) + 1);
      sprintf(cmd, CPR_LOG_CMD, db_name);
    }
    else {
      cmd = malloc(strlen(CPR_CMD) + strlen(db_name) + 1);
      sprintf(cmd, CPR_CMD, db_name);
    }
    states[AWAIT_DB][3].out_string = cmd;
#endif
  }

  /* Parse out the summary fields line */
  if (!title_fields)
    error("mising title_fields");

  /* initialize a hash table to hold our field format records */
  summ_form_tab = hash_Init(HASH_TAB_SIZE);

  field_cnt = 0;
  for (sdex = 0; title_fields[sdex]; ) {
    /* Get a new format record */
    title_form_vec[title_field_cnt++] = summ_form = 
      (format_rec *) malloc(sizeof(format_rec));
#if defined(POSIX) || defined (_POSIX_SOURCE)
    memset(summ_form, 0, sizeof(format_rec));
#else
    bzero(summ_form, sizeof(format_rec));
#endif
    summ_form->idex = field_cnt++;
    /* Parse out the two-char field abbrev */
    while (!isupper(title_fields[sdex]))
      sdex++;
    strncpy(summ_form->field, title_fields + sdex, 2);
    summ_form->field[2] = 0;
    sdex += 2;
    summ_form->title_field = TRUE;
    
    /* extract the width */
    while (!isdigit(title_fields[sdex]))
      sdex++;
    summ_form->width =  atoi(title_fields + sdex);
    
    /* Insert this field into our hash table */
    hash_Insert(summ_form_tab, hash_HashFunction(summ_form->field, 2),
		hash_compare, summ_form->field, summ_form);
    
    /* Scan to next field, or end */
    while (title_fields[sdex] && (title_fields[sdex] != SUMM_FIELD_SEP))
      sdex++;
  }

  /* parse the summary body fields */
  if (!summ_fields)
    error("missing summary fields");
  for (sdex = 0; summ_fields[sdex]; ) {
    /* Parse out the two-char field abbrev */
    while (!isupper(summ_fields[sdex]))
      sdex++;
    strncpy(field, summ_fields + sdex, 2);
    field[2] = 0;
    sdex += 2;

    /* Has a format record for this field already been created? */
    if (!(summ_form = (format_rec *) hash_Lookup(summ_form_tab, 
				        hash_HashFunction(field, 2),
				        hash_compare, field))) {
      summ_form = (format_rec *) malloc(sizeof(format_rec));
#if defined(POSIX) || defined (_POSIX_SOURCE)
      memset(summ_form, 0, sizeof(format_rec));
#else
      bzero(summ_form, sizeof(format_rec));
#endif
      summ_form->idex = field_cnt++;
      strcpy(summ_form->field, field);
      /* Insert this field into our hash table */
      hash_Insert(summ_form_tab, hash_HashFunction(summ_form->field, 2),
		  hash_compare, summ_form->field, summ_form);
    }
    summ_form->summ_field = TRUE;
    summ_form_vec[summ_field_cnt++] = summ_form;

    /* Scan to next field, or end */
    while (summ_fields[sdex] && (summ_fields[sdex] != SUMM_FIELD_SEP))
      sdex++;
  }

  /* parse the summary extension fields */
  if (ext_fields) {
    for (sdex = 0; ext_fields[sdex]; ) {
      /* Parse out the two-char field abbrev */
      while (!isupper(ext_fields[sdex]))
	sdex++;
      strncpy(field, ext_fields + sdex, 2);
      field[2] = 0;
      sdex += 2;
      
      /* Has a format record for this field already been created? */
      if (!(summ_form = (format_rec *) hash_Lookup(summ_form_tab, 
                                         hash_HashFunction(field, 2),
				         hash_compare, field))) {
	summ_form = (format_rec *) malloc(sizeof(format_rec));
#if defined(POSIX) || defined (_POSIX_SOURCE)
	memset(summ_form, 0, sizeof(format_rec));
#else
	bzero(summ_form, sizeof(format_rec));
#endif
	summ_form->idex = field_cnt++;
	strcpy(summ_form->field, field);
	/* Insert this field into our hash table */
	hash_Insert(summ_form_tab, hash_HashFunction(summ_form->field, 2),
		    hash_compare, summ_form->field, summ_form);
      }
      summ_form->ext_field = TRUE;
      ext_form_vec[ext_field_cnt++] = summ_form;
      
      /* Scan to next field, or end */
      while (ext_fields[sdex] && (ext_fields[sdex] != SUMM_FIELD_SEP))
	sdex++;
    }
  }
  
  /* create a list of all fields, for use in summary request command */
  summary_fields = malloc(4 * field_cnt);
  sdex = 0;
  for (summ_form = (format_rec *) hash_FirstEntry(summ_form_tab); summ_form;
       summ_form = (format_rec *) hash_NextEntry(summ_form_tab)) {
    strncpy(summary_fields + sdex, summ_form->field, 2);
    sdex += 2;
    summary_fields[sdex++] = SUMM_FIELD_SEP;
  }
  summary_fields[sdex] = 0;
}

/* This routine records the abbreviated to long name mappings and
   puts any fields which have not already been added into the 
   hash table.
 */
void parse_fields(void)
{
  int cmp_len;
  int sdex = 0, tdex = 0;
  char field[3];
  char lname[MAX_SUMM_FIELD_NAME_LEN];
  format_rec *summ_form;
  char *info_list;  /* list of fields for reporting back to Willow */
  int   info_dex = 0;   /* pointer for constructing info_list */
  int   info_cnt = 0;
  char  is_default;

  /* Fields must be smaller than this */
  info_list = malloc(strlen(parse_buf));

  /* Find the beginning */
  cmp_len = strlen(EXPANDED_START);
  while (strncmp(parse_buf+sdex, EXPANDED_START, cmp_len) != 0)
      sdex++;
  sdex += cmp_len + 1;
  
  /* Find the long-len */
  summary_long_len = atoi(parse_buf + sdex);
  if (summary_long_len < 1)
    summary_long_len = DFLT_LONG_LEN;
  else if (summary_long_len > MAX_SUMM_FIELD_NAME_LEN)
    summary_long_len = MAX_SUMM_FIELD_NAME_LEN;

  while (parse_buf[sdex] != '\n')
    sdex++;
  sdex++;

  /* For each field in the list, look it up in our hash table and record 
     the long name */
  while (parse_buf[sdex]) {
    is_default = (parse_buf[sdex++] != ' ');
    strncpy(field, parse_buf + sdex, 2);
    field[2] = 0;
    sdex += 2;

    if (!(summ_form = (format_rec *) 
	  hash_Lookup(summ_form_tab, 
		      hash_HashFunction(field, 2),
		      hash_compare, field))) {
      summ_form = (format_rec *) malloc(sizeof(format_rec));
#if defined(POSIX) || defined (_POSIX_SOURCE)
      memset(summ_form, 0, sizeof(format_rec));
#else
      bzero(summ_form, sizeof(format_rec));
#endif
      summ_form->idex = -1; /* Not used in summary record */
      strcpy(summ_form->field, field);
      /* Insert this field into our hash table */
      hash_Insert(summ_form_tab, hash_HashFunction(summ_form->field, 2),
		  hash_compare, summ_form->field, summ_form);
    }
    summ_form->no_wrap = parse_buf[sdex] != ' ';
    sdex++;

    for (tdex = 0; (parse_buf[sdex] != '\n') && (tdex < summary_long_len);)
      lname[tdex++] = parse_buf[sdex++];
    lname[tdex++] = 0;
    strcpy(summ_form->lname, lname);
    
    /* stick copy of short and long name into list for return to willow. 
       Screen out IT/OR/PR Status fields from LCAT/GCAT */
    if (!(((strcmp(db_name, "LCAT") == 0) && (strcmp(field, "IT") == 0)) ||
	  ((strcmp(db_name, "GCAT") == 0) &&
	   ((strcmp(field, "OR") == 0) || (strcmp(field, "PR") == 0))))) {
      info_list[info_dex++] = is_default;
      strcpy(info_list + info_dex, field);
      info_dex += strlen(field) + 1;
      strcpy(info_list + info_dex, lname);
      info_dex += strlen(lname) + 1;
      info_cnt++;
    }

    while (parse_buf[sdex] != '\n')
      sdex++;
    sdex++;
    if (!(isupper(parse_buf[sdex+1]) && (isupper(parse_buf[sdex+2]) ||
				       isdigit(parse_buf[sdex+2]))))
      break;
  }

  /* If we are doing circ, put in a cheat record, so that later we
     can do an expand_summ() on it */
  if (do_circ) {
    strcpy(field, CIRC_ABBREV);
    summ_form = (format_rec *) malloc(sizeof(format_rec));
#if defined(POSIX) || defined (_POSIX_SOURCE)
    memset(summ_form, 0, sizeof(format_rec));
#else
    bzero(summ_form, sizeof(format_rec));
#endif
    summ_form->idex = -1;
    strcpy(summ_form->field, field);
    summ_form->no_wrap = TRUE;
    strcpy(summ_form->lname, CIRC_LONG_NAME);
    hash_Insert(summ_form_tab, hash_HashFunction(summ_form->field, 2),
		hash_compare, summ_form->field, summ_form);
  }
  send_fields_info(info_cnt, info_list, info_dex);
}

/* Do the field expansion for a summary field */
char *expand_summ(char *summary)
{
  int sdex, tdex, sblank, tblank, scr;
  int s_len;
  char *exp_buf;
  char *wrap_buf;
  int width, i;
  format_rec *summ_form;
  char field[3];

  if (!summary)
    return(NULL);
  s_len = strlen(summary);
  sdex = tdex = 0;
  exp_buf = malloc(MAX(PARSE_BUF_SIZE, 3*s_len));
  wrap_buf = malloc(MAX(PARSE_BUF_SIZE, 3*s_len));

  /* Find out what field we are dealing with */
  strncpy(field, summary, 2);
  field[2] = 0;
  if (!(summ_form = (format_rec *) hash_Lookup(summ_form_tab, 
                      hash_HashFunction(field, 2), hash_compare, field)))
    error("Trying to expand an unknown field?");

  /* Expand the field name */
  if (summ_form->lname[0]) {
    for (i = 0; i < summary_long_len; i++)
      if (summ_form->lname[i])
	exp_buf[tdex++] = summ_form->lname[i];
      else
	exp_buf[tdex++] = ' ';
  }
  else {
    strncpy(exp_buf, summary, 2);
    tdex = 2;
    for (i = 2; i < summary_long_len; i++)
      exp_buf[tdex++] = ' ';
  }
  sdex = ABBREV_INDENT;

  /* Go through expanded field names, turning \n to ' ' for later wrapping. */
  while (sdex < s_len) {
    if (summary[sdex] != '\n')
      exp_buf[tdex++] = summary[sdex++];
    else {
      if (summ_form->no_wrap) {
	if (sdex == (s_len - 1))
	  sdex++;
	else {
	  exp_buf[tdex++] = summary[sdex++];
	  for (i = 0; i < summary_long_len; i++)
	    exp_buf[tdex++] = ' ';
	}
      }
      else if (summary[sdex+1] == '\n') {
	/* If there are two newlines in a row, keep both */
	exp_buf[tdex++] = summary[sdex++];
	exp_buf[tdex++] = summary[sdex++];
	for (i = 0; i < summary_long_len; i++)
	  exp_buf[tdex++] = ' ';
      }
      else {
	exp_buf[tdex++] = ' ';
	sdex++;
      }
      sdex += ABBREV_INDENT;
    }
  }
  exp_buf[tdex] = 0;

  /* Go through and word-wrap everything */
  sdex = tdex = sblank = tblank = width = scr = 0;
  s_len = strlen(exp_buf);
  while (sdex < s_len) {
    if (((exp_buf[sdex] == ' ') || (exp_buf[sdex] == '-')) && 
	((sdex - scr) > summary_long_len)) {
      if (!((exp_buf[sdex] == '-') && (width > (MAX_DISP_LINE - 1)))) {
	sblank = sdex;
	tblank = tdex;
      }
    }
    if (exp_buf[sdex] == '\n') {
      scr = sdex;
      width = sblank = tblank = 0;
    }
    wrap_buf[tdex++] = exp_buf[sdex++];
    width++;
    if (width > MAX_DISP_LINE) {
      if (sblank) {
	if (exp_buf[sblank] == '-')
	  tblank++;
	wrap_buf[tblank] = '\n';
	sdex = sblank + 1;
	tdex = tblank + 1;
	width = sblank = tblank = 0;
      }
      else {
	wrap_buf[tdex - 1] = '\n';
	sdex--;
	width = sblank = tblank = 0;
      }
      for (i = 0; i < summary_long_len; i++) {
	wrap_buf[tdex++] = ' ';
	width++;
      }
    }
  }
  wrap_buf[tdex++] = '\n';
  wrap_buf[tdex] = 0;
  free(exp_buf);
  return(wrap_buf);
}

/* The necessary information for constructing the searches has
   been extracted from the packet, and passed in in two linked
   lists.  Scan through the passed in strings, pass
   through punctuation characters, keep track of each word. 
   Wrap each section in parens, and append command suffix. Join with 
   'and' We are also responsible for freeing the passed in text list.
   The list of limit fields is saved as queue, for later execution.
 */
make_search(search_desc_rec *text_fields, search_desc_rec *limit_fields)
{
  search_desc_rec *cur, *prev;  /* Used to traverse lists */
  char  word[MAX_LINE];
  int   len;
  int   tdex, sdex, wdex;	/* indices for target, source, word strings */
  int   markdex;                /* index of end of last search */
  int   wcount;                 /* how many words in search */
  char  ch;			/* current character */
  char  quoting;                /* are we currently in a quote */

  summ_total = 0;
  search[0] = 0;
  
  prev = NULL;
  for (cur = text_fields; 
       cur !=NULL; 
       cur = cur->next){

    if (prev) 
      free(prev);
    prev = cur;
    
    markdex = strlen(search);
    if (search[0])
      strcat(search," and ");
    
    tdex = strlen(search);
    sdex = wdex = 0;
    len = strlen(cur->text);
    wcount = 0;
    search[tdex++] = '(';
      
    /* loop through the source (cur->text), copy each character to
       target (search). Build a current word. Notice that we
       loop one extra, so we see the null at the end. This serves
       as a terminator for last word. Text between two double-quotes
       is treated as a single word */
    quoting = FALSE;
    for (sdex = 0; sdex <= len; sdex++){
      ch = strain(cur->text[sdex], cur->command);
      if (ch == -1)
	continue;
      if (ch == '"') 
	quoting = !quoting;
      /* Word ends at end of string or whitespace/parens unless we are in
	 quoted text */
      if ((ch == '\0') ||
	  (!quoting && ((ch == ' ') || (ch == '\t') ||  
			(ch == '(') || (ch == ')')))) {
	/* just ended a word */
	if (wdex) {	
	  word[wdex] = 0;
	  wdex = 0;
	  wcount++;

	  /* Can't strip periods out of Call Numbers */
	  if (strncmp(cur->command, LCAT_CALL_NUM,
		      strlen(LCAT_CALL_NUM)) != 0)
	    fix_periods(word, cur->command[0] == 0);

	  search[tdex] = 0;
	  strcat(search, word);
	  tdex = strlen(search);
	}
	/* Add in the separator character */
	search[tdex++] = ch;
      }
      else
	word[wdex++] = ch;
    }  
    /* last char is a '\0' */
    search[--tdex] = ')';
    search[++tdex] = 0;
      
    /* check if there were no actual words in search line (just blanks) */
    if (!wcount)
      search[markdex] = 0;
    /* Check for special case */
    else if (strncmp(cur->command, LCAT_EXACT_TITLE, 
		     strlen(LCAT_EXACT_TITLE)) == 0)
      lcat_titlefy(&search[markdex]);
    else if (strncmp(cur->command, LCAT_CALL_NUM,
		     strlen(LCAT_CALL_NUM)) == 0)
      lcat_callnumify(&search[markdex]);
    else {
      /* append field suffix */
      strcat(search, cur->command);
      tdex = strlen(search);
    }
    free(cur->text);
    free(cur->command);
  }

  if (strlen(search) > MAX_SEARCH) {
    bad_search("Search statement is too long");
    return;
  }
    
  /* Save the limit fields as queue to be executed as limits to the
     main search. 
   */
  limit_queue = limit_fields;
  
  if (search[0]) {
    strcat(search, "\n");
    queue_cmd(search, AWAIT_STATS);
  }
  else 
    bad_search("Invalid search");
  if (logging) {
    char comment[1024];
    packet_rec packet;
    search_desc_rec *lq;

    packet.type = SEARCH;
    strcpy(comment, search);
    comment[strlen(comment) - 1] = 0;
    for (lq = limit_queue; lq; lq = lq->next) {
      strcat(comment, " + LIMIT:");
      strcat(comment, lq->command);
    }
    log_trans(packet.type, comment);
  }
}

/* If this is a one word LCAT title, use .TS., otherwise .TI. */
void lcat_titlefy(char *title)
{
  if (strchr(title, ' ') || strchr(title, '-'))
    strcat(title, LCAT_TITLE_CMD);
  else
    strcat(title, LCAT_SINGLE_TITLE_CMD);
}

/* If this is a LCAT call number, replace whitespace-period and
   period-whitespace with blank. Append search fields to end.
 */
void lcat_callnumify(char *cnum)
{
  int sdex, tdex;
  char prev_ch = 0; /* Last char we saw, 0:normal, 1:blank; 2:period */

  sdex = tdex = 0;
  while (cnum[sdex]) {
    if (isspace(cnum[sdex])) {
      if (prev_ch == 2)
	cnum[tdex - 1] = cnum[sdex];
      else
	cnum[tdex++] = cnum[sdex];
      prev_ch = 1;
    }
    else if (cnum[sdex] == '.') { 
      if (prev_ch != 1)
	cnum[tdex++] = cnum[sdex];
      prev_ch = 2;
    }
    else {
      prev_ch = 0;
      cnum[tdex++] = cnum[sdex];
    }
    sdex++;
  }
  cnum[tdex] = 0;
  strcat(cnum, LCAT_CNUM_CMD);
}

/* Strain out unwelcome characters */
char strain(char ch, char *command)
{
  int i;

  if (isalpha(ch) || isdigit(ch))
    return(ch);
  
  if (ch == '*')
    return('$');

  for (i = 0; i < sizeof(valid_extra_chars); i++)
    if (ch == valid_extra_chars[i])
      return(ch);

  /* Special treatment of & and , in call numbers */
  if (strcmp(command, LCAT_CALL_NUM) == 0)
    for (i = 0; i < sizeof(callnum_extra_chars); i++)
      if (ch == callnum_extra_chars[i])
	return('-');
  
  return((char) -1);
}

/* Do yucky special case processing to allow people to use .'s properly.
   They are only allowed inside quoted words, and for Any Field searches,
   when they appear to be a valid field qualifier.
   This routine can have one level of recursion, just for fun.
  */
fix_periods(char *word, char any_field)
{
  int len,          /* Length of word */
      pcount = 0,   /* How many periods in the word */
      sdex = 0, tdex = 0;   /* source and target indices */

  len = strlen(word);

  /* If word is quoted, just send it back, periods and all */
  if ((word[0] == '\"') && (word[len-1] == '\"'))
    return;

  /* If we are in a specific field, simply delete all periods */
  if (!any_field) {
    for (sdex = 0; sdex <= len; ) {
      if (word[sdex] == '.')
	sdex++;
      else
	word[tdex++] = word[sdex++];
    }
  }

  /* If we are doing an any field search, allow
     periods for field qualification */
  else {
    for (sdex = 0; sdex < len; sdex++)
      if (word[sdex] == '.')
	pcount++;
    if ((pcount == 2) && (word[len-1] == '.'))
      return;
    else if (pcount)
      fix_periods(word, FALSE);
  }
}

/* Handle a request for list-browse search */
browse_wcb(packet_rec *packet)
{
  char *ptr;
  char *list_name;
  char *prefix_st;
  char *list_data;
  int   line_pos;

  ptr = packet->data;
  list_name = strdup(ptr);
  ptr += strlen(ptr) + 1;
  
  prefix_st = strdup(ptr);
  list_data = list_lookup(db_name, list_name, 
			  packet->var.browse_req.search_type,
			  prefix_st,
			  packet->var.browse_req.lines_wanted,
			  &line_pos,
			  packet->var.browse_req.line_offset);
  send_browse(packet->var.browse_req.ref_num, list_data, line_pos);
}

/* Start the process of disconnecting from host database, and quitting. */
disconnect_wcb(void)
{
  if (do_circ)
    close_circ();
  if (state == NO_CON)
    cleanup();
  else
    close_willow();

  queue_cmd(QUIT_CMD, NO_CON);
  
  /* Set an alarm, to force a quit */
  set_alarm(QUIT_TIMEOUT);
}

/* Called if we are supposed to try and call off this search */
search_cancel_wcb(packet_rec *packet)
{
  host_send(INTERRUPT_CMD);
}

/* Go through the parse buffer, which at the point this routine is called
   should contain everything between the "Enter search request" prompt,
   and the "display or search" prompt. We are looking for
   the BRS-assigned number of the search we just entered.
   We also need the total number of results.
   Also throw all the stats directly into summary window. Maybe do
   something fancier later. 
 */
parse_stats_hcb(void)
{
  int sdex = 0;
  static char out_string[MAX_LINE];
  int cmp_len, parse_buf_len;
  search_desc_rec *prev_limit;

  cmp_len = strlen(RESULTS_STR);
  parse_buf_len = strlen(parse_buf);

  num_results = search_num = 0;

  /* scan for start of stats */
  while ((strncmp(parse_buf+sdex, RESULTS_STR, cmp_len) != 0) &&
	 (sdex < parse_buf_len))
    sdex++;
  
  if (sdex < parse_buf_len) { /* found something */
    sdex += cmp_len;
    
    /* Scan through, and extract the two numbers */
    search_num  = extract_num(parse_buf, sdex);
    
    /* This only works if sdex is pointing right at first # */
    while (isdigit(parse_buf[sdex]))
      sdex++;
    num_results = extract_num(parse_buf, sdex);
    
    /* Look for the end of current line, which marks end of stats */
    while (parse_buf[sdex] != '\n')
      sdex++;

    /* Check for truncation overflow */
    cmp_len = strlen(OVERFLOW_STR);
    while ((strncmp(parse_buf+sdex, OVERFLOW_STR, cmp_len) != 0) &&
	   (sdex < parse_buf_len))
      sdex++;
    if (sdex < parse_buf_len)
      misc_warning_hcb("\
Truncation limit reached.  You may wish to:\n\
        1) use \"or\" instead of truncation\n\
                (train or trains or training instead of train$)\n\
        2) press \"Help\" for search help\n");
  }

  if (num_results == 0){  /* Didn't find anything */
    free_limit_queue();
  }

  /* Set up state table, depending on whether we need to do a limit. 
     Note that we are not *always* called from [1], so the changes
     to state table might be totally ignored */
  if (limit_queue) {
    if (limit_queue->range_type) {
      sprintf(out_string, LIMIT_CMD, make_range(limit_queue));
      free(limit_queue->text);
    }
    else
      sprintf(out_string, LIMIT_CMD, limit_queue->command);
    free(limit_queue->command);
    prev_limit = limit_queue;
    limit_queue = limit_queue->next;
    free(prev_limit);
      
    states[AWAIT_STATS][1].out_string = out_string;
    states[AWAIT_STATS][1].next_state = AWAIT_STATS;
  }
  else {
    states[AWAIT_STATS][1].out_string = "";
    states[AWAIT_STATS][1].next_state = READY;

    /* No extraordinary status string to send */
    send_stats_cmd(NULL, num_results, search_num);
  }
}

/* Called when BRS rejects the search statement */
bad_search_hcb(void)
{
  int cmp_len;
  int sdex = 0;
  int err_start;

  cmp_len = strlen(ERROR_STR);

  /* scan for start of error msg */
  while (strncmp(parse_buf+sdex, ERROR_STR, cmp_len) != 0)
    sdex++;
  sdex += cmp_len;

  /* This is error text from here to end of line */
  err_start = sdex;
  while (parse_buf[sdex] != '\n')
    sdex++;
  parse_buf[sdex] = 0;
  bad_search(parse_buf + err_start);
}


/* Create a range limit command */
char *make_range(search_desc_rec *search)
{
  int   month, day, year;
  int   ptr = 0;
  int   i;
  char *operator = ">=";
  static char buf[MAX_LINE];

  buf[0] = 0;
  for (i = 0; i < 2; i++) {
    month = atoi(search->text + ptr);
    while (isdigit(search->text[ptr])) ptr++;
    ptr++;
    
    day = atoi(search->text + ptr);
    while (isdigit(search->text[ptr])) ptr++;
    ptr++;
    
    year = atoi(search->text + ptr);
    while (isdigit(search->text[ptr])) ptr++;
    ptr++;
    
    if (year) {
      if (buf[0])
	strcat(buf, " AND ");

      switch (search->range_type) {
      case MONTH:
        sprintf(buf + strlen(buf), "%s%s%c%c%c%c", search->command, operator,
		((year  % 100) / 10) + '0', (year % 10) + '0', 
		(month / 10) + '0', (month % 10) + '0');
	break;
      case DAY:
        sprintf(buf + strlen(buf), "%s%s%c%c%c%c%c%c", search->command,
		operator, ((year  % 100) / 10) + '0', (year % 10) + '0', 
		(month / 10) + '0', (month % 10) + '0',
		(day / 10) + '0', (day % 10) + '0');
	break;
      case NUMERIC:
      case YEAR:
        sprintf(buf + strlen(buf), "%s%s%d", search->command, operator, year);
	break;
      }
    }
    operator = "<=";
  }
  return(buf);
}

/* Free up the limit queue. Called either from parse_stats_hcb() or 
   no_limit_hcb()
 */
free_limit_queue(void)
{
  search_desc_rec *prev_limit;
     
  while (limit_queue) {
    free(limit_queue->command);
    if (limit_queue->range_type)
      free(limit_queue->text);
    prev_limit = limit_queue;
    limit_queue = limit_queue->next;
    free(prev_limit);
  }
}

/* Extract a number from a string. Passed int pointer, index.
 */
int extract_num(char *str, int sdex)
{
  char  numeral[20];
  int   numdex = 0;

  while (!isdigit(str[sdex]))
    sdex++;

  while (isdigit(str[sdex]))
    numeral[numdex++] = str[sdex++];

  numeral[numdex] = 0;
  return(atoi(numeral));
}

/* Called when willow asks us for the next summary */
title_req_wcb(packet_rec *packet)
{
  int req;
  int idex;

  req = packet->var.title_req.title_num;

  if (req <= summ_total) {
    send_title(req, build_title(req));
  }
  else {
    if (logging)
      log_trans(packet->type, "");
    if ((req + SUMM_BLOCK_SIZE - 1) <= num_results)
      idex = req + SUMM_BLOCK_SIZE -1;
    else
      idex = num_results;
    sprintf(next_summ_cmd, NEXT_SUMM_CMD, summary_fields, req, idex);
    queue_cmd(next_summ_cmd, DISP_SHRT);
  }
}

/* Called when willow asks us for an extensions record */
ext_req_wcb(packet_rec *packet) 
{
  send_ext(packet->var.record_req.ref_num, 
	   build_ext(packet->var.record_req.ref_num));
}

/* Parse out the copyright information. Don't take the \n's at the
   beginning and end.
 */
void parse_info_hcb(void)
{
  int sdex = 0;
  int len;
  char *info_str;

  len = strlen(INFO_START);
  while (strncmp(parse_buf + sdex, INFO_START, len) != 0)
    sdex++;
  sdex += len + 1;
  info_str = strdup(parse_buf + sdex);

  sdex = 0;
  len = strlen(INFO_END);
  while (strncmp(info_str + sdex, INFO_END, len) != 0)
    sdex++;
  info_str[sdex - 1] = 0;
  got_info(info_str);
  free(info_str);
}

/*
  Parse a block of summaries, and store them for later delivery to
  Willow as requested.
  */
parse_summary_hcb(void)
{
  char *summ_buf;
  int   sdex = 0,			/* source index */
        tdex = 0;			/* target index */
  int   buf_len;
  int   howmany = 0, i;
  char  field[3];
  format_rec *summ_form;
  int   field_start;

  summ_buf =parse_buf;    
  buf_len = strlen(summ_buf);

  /* Do one pass through to skip non-printing characters */
  for (; sdex <= buf_len; sdex++) {
    if ((isprint(summ_buf[sdex])) || 
	(summ_buf[sdex] == '\n') || 
	(summ_buf[sdex] == 0))
      summ_buf[tdex++] = summ_buf[sdex];
  }

  /* Skip heading junk */
  sdex = 0;
  while (strncmp(summ_buf+sdex, 
		 SUMM_START_STR, 
		 SUMM_START_STR_LEN) != 0)
    sdex++;

  if (summ_total == (num_summ_allocs * SUMM_ALLOC_SIZE))
    alloc_summ_tab();

  for (howmany = summ_total + 1; howmany <= summ_total + SUMM_BLOCK_SIZE; 
       howmany++) {
    /* Free up memory from previous search if applicable */
    if (summ_tab[howmany].fields){
      for (i = 0; i < field_cnt; i++)
	if (summ_tab[howmany].fields[i])
	  free(summ_tab[howmany].fields[i]);

      if (summ_tab[howmany].circ) {
	free(summ_tab[howmany].circ);
	summ_tab[howmany].circ = NULL;
      }
    }
    /* Allocate a fresh fields block. This is big enough to hold the
       total number of fields we know about. */
    else
      summ_tab[howmany].fields = (char **) malloc(sizeof(char *) * field_cnt);
      
    /* clear the fields block */
#if defined(POSIX) || defined (_POSIX_SOURCE)
    memset(summ_tab[howmany].fields, 0, sizeof(char *) * field_cnt);
#else
    bzero(summ_tab[howmany].fields, sizeof(char *) * field_cnt);
#endif

    /* Go through the current summary, and break out each individual 
       field. As they are found, copy value into pre-determined slot in
       the fields block */
    
    while (!new_record(summ_buf + sdex)) {
      while (!start_of_field(summ_buf + sdex) && summ_buf[sdex])
	sdex++;
      if (!summ_buf[sdex])
	break;
      field_start = sdex;
      strncpy(field, summ_buf + sdex, 2);
      field[2] = 0;

      /* Check for deleted document */
      if (field[0] == '\n')
	break;
      if (!(summ_form = (format_rec *) hash_Lookup(summ_form_tab, 
                                          hash_HashFunction(field,2),
				          hash_compare, field))) {
	char msg[128];

	sprintf(msg, "Unrequested field \"%s\" in summary record.", field);
	error(msg);
      }
      
      /* Stick a 0 at the end of this field */
      while (!end_of_field(summ_buf + sdex))
	sdex++;
      summ_buf[sdex] = 0;

      /* Copy the field value into the appropriate slot */
      summ_tab[howmany].fields[summ_form->idex] = 
	strdup(summ_buf + field_start);
      
      /* Put the \n back in */
      summ_buf[sdex] = '\n';

    }
    if (summ_buf[sdex]) {
      while ((!new_record(summ_buf + sdex)) && summ_buf[sdex])
	sdex++;
      while ((!start_of_field(summ_buf + sdex)) && summ_buf[sdex])
	sdex++;
    }
  }

  /* Send back the first one */
  send_title(summ_total + 1, build_title(summ_total + 1));
  summ_total = howmany - 1;
  if (logging) {
    packet_rec packet;
    packet.type = TITLE;
    log_trans(packet.type, "");
  }
}

/* If "\n  " signals a continuation of a field, or we could be at the
   end of the record */
end_of_field(char *str)
{
  if (str[0] != '\n')
    return(FALSE);
  if (isupper(str[1]))
    if (isupper(str[2]) || isdigit(str[2]))
      return(TRUE);
  if (strncmp(str + 1, PROMPT_STR, PROMPT_LEN) == 0)
    return(TRUE);
  return(new_record(str));
}

/* Almost like end_of_field, but we want to be past the newline */
start_of_field(char *str)
{
  /* We have to check back one. Kind of iffy programming, but should be ok */
  if (*(str - 1) != '\n')
    return(FALSE);

  /* Consider missing record warning to be field */
  if (strncmp(str, MISSING_STR, MISSING_LEN) == 0)
    return(TRUE);

  if (isupper(str[0]))
    if (isupper(str[1]) || isdigit(str[1]))
      return(TRUE);
  return(FALSE);
}

new_record(char *str)
{
  if (str[0] != '\n')
    return(FALSE);
  if (str[1] != '\n')
    return(!strncmp(str + 1, SUMM_START_STR, SUMM_START_STR_LEN));
  return(!strncmp(str + 2, SUMM_START_STR, SUMM_START_STR_LEN));
}

/* Put together a title string for the current record */
char *build_title(int num)
{
  static char title_buf[MAX_TITLE_LEN];
  int  n, i;
  int sdex, tdex = 0;
  format_rec *summ_form;
  char *data;
  char  done;

  /* Go through each of our title fields, and copy in the current value */
  for (n = 0; n < title_field_cnt; n++) {
    summ_form = title_form_vec[n];
    data = summ_tab[num].fields[summ_form->idex];

    /* Copy in up to width chars from data string. Fix up linebreaks */
    if (data) {
      sdex = ABBREV_INDENT;
      while (data[sdex] == ' ')
	sdex++;
      done = FALSE;
    }
    else
      done = TRUE;
    for (i = 0; i < summ_form->width; i++) {
      if (!done) {
	/* Compress multi-blanks to one */
	if (data[sdex] == ' ') {
	  while (data[sdex] == ' ')
	    sdex++;
	  sdex--;
	}
	if (data[sdex] == 0)
	  done = TRUE;
	else if (data[sdex] == '\n') {
	  if (data[sdex++] == 0)
	    done = TRUE;
	  else {
	    title_buf[tdex++] = ' ';
	    if (++i == summ_form->width)
	      break;
	    while (data[sdex] == ' ')
	      sdex++;
	  }
	}
	if (!done)
	  title_buf[tdex++] = data[sdex++];
      }
      if (done)
	title_buf[tdex++] = ' ';
    }
    if (n < title_field_cnt - 1) {
      title_buf[tdex++] = ' ';
      title_buf[tdex++] = ' ';
    }
  }
  title_buf[tdex] = 0;
  return(title_buf);
}

/* Put together a summary record */
char *build_summ(int num, char abbreviated)
{
  static char *summ_buf = NULL;
  int    tdex, len;
  static int   buf_size = 0;
  format_rec  *summ_form;
  char  *data,*exp_data, *circ;
  char   something = FALSE;
  char  msg_buf[MAX_LINE];
  int n;

  if ((num < 1) || (num > summ_total)) {
    sprintf(msg_buf, "Summary requested is out of range: %d", num);
    misc_warning_hcb(msg_buf);
    return("ERROR");
  }

  if (!summ_buf)
    summ_buf = malloc(buf_size = PARSE_BUF_SIZE);

  sprintf(summ_buf, SUMM_START_FMT, num);
  tdex = strlen(summ_buf);
  for (n = 0; n < summ_field_cnt; n++) {
    summ_form = summ_form_vec[n];
    if (!(data = summ_tab[num].fields[summ_form->idex]))
      continue;
    else
      something = TRUE;
    if (abbreviated) {
      len = strlen(data);
      if ((tdex + len + 2) > buf_size)
	summ_buf = realloc(summ_buf, buf_size += PARSE_BUF_SIZE);
      strcpy(summ_buf + tdex, data);
      tdex += len;
      summ_buf[tdex++] = '\n';
    }
    else {
      exp_data = expand_summ(data);
      len = strlen(exp_data);
      if ((tdex + len + 2) > buf_size)
	summ_buf = realloc(summ_buf, buf_size += PARSE_BUF_SIZE);
      strcpy(summ_buf + tdex, exp_data);
      free(exp_data);
      tdex += len;
    }
  }
  if (!something){
    strcpy(summ_buf,"\n   DOCUMENT DELETED\n");
    return(summ_buf);
  }

  /* Handle circulation status if necessary */
  if (do_circ) {
    if (!summ_tab[num].circ)
      summ_tab[num].circ = get_circ(num);
    if (abbreviated)
      circ = summ_tab[num].circ;
    else
      circ = expand_summ(summ_tab[num].circ);
    if (circ) {
      len = strlen(circ);
      if ((tdex + len + 2) > buf_size)
	summ_buf = realloc(summ_buf, buf_size += PARSE_BUF_SIZE);
      strcpy(summ_buf + tdex, circ);
      tdex += len;
    }
  }
  summ_buf[tdex] = 0;
  return(summ_buf);
}

/* Put together the extensions record */
char *build_ext(int num)
{
  static char *ext_buf = NULL;
  int    tdex, len;
  static int   buf_size = 0;
  format_rec  *summ_form;
  char  *data;
  int n;

  if (!ext_buf)
    ext_buf = malloc(buf_size = PARSE_BUF_SIZE);

  tdex = 0;
  for (n = 0; n < ext_field_cnt; n++) {
    summ_form = ext_form_vec[n];
    if (!(data = summ_tab[num].fields[summ_form->idex]))
      continue;

    len = strlen(data);
    if ((tdex + len + 2) > buf_size)
      ext_buf = realloc(ext_buf, buf_size += PARSE_BUF_SIZE);
    strcpy(ext_buf + tdex, data);
    tdex += len;
    ext_buf[tdex++] = '\n';
  }
  ext_buf[tdex - 1] = 0;
  return(ext_buf);
}

/* Send back a record, either summary/full, abbreviated/short fields,
   if full, queue a command to retrieve it.
 */
record_req_wcb(packet_rec *packet)
{
  int i;
  char fields_buf[MAX_LINE];

  cur_record = packet->var.record_req.ref_num;
  if (packet->var.record_req.summary) {
    send_summ_record(cur_record,
		     build_summ(cur_record,
				packet->var.record_req.abbreviated));
  }
  else {
    /* Check if the user specified fields */
    if (packet->data_len) {
      fields_buf[0] = 0;
      for (i = 0; i < packet->data_len;) {
	/* Leave out "IT" field in LCAT, otherwise we get double "Status" */
	if (!(do_circ && (strcmp(packet->data + i, "IT") == 0))) {
	  strcat(fields_buf, packet->data + i);
	  strcat(fields_buf, ",");
	}
	while(packet->data[i])
	  i++;
	i++;
      }
      sprintf(get_full_cmd, GET_FIELDS_FULL_CMD, fields_buf, cur_record);
      if (packet->var.record_req.abbreviated)
	expanded_to_willow = FALSE;
      else
	expanded_to_willow = TRUE;
      queue_cmd(get_full_cmd, DISP_LONG);
    }
    else
      /* No fields specified, send back an empty record */
      send_full_record(cur_record, "NO RECORD TO DISPLAY -- NO FIELDS SELECTED\n");
  }
}

/* Go through the parse buffer, filter out '##' and pull out extraneous
   blank lines. Used on full-record display, and circulation
   status. Destination buffer gets passed in. Also filter out hilite
   escape chars, but remember where they occured.
 */
char *really_parse_display(void)
{
  int   sdex, tdex;		/* string indices for source and target */
  int   field_dex;
  int   parse_len, field_len, hl_start_len, hl_end_len, ugly_len;
  char *a_buf, *b_buf, *field_buf, *exp_field_buf;
  char  kill_next_hl_end;
  int   buf_size;

  parse_len = strlen(parse_buf);
  a_buf = malloc(buf_size = (2 * parse_len));
  b_buf = malloc(buf_size);
  field_buf = malloc(buf_size);

  tdex = 0;
  field_len = strlen(ABBREV_START);
  hl_start_len = strlen(HL_START_STR);
  hl_end_len   = strlen(HL_END_STR);
  ugly_len = strlen(UGLY_STR);

  /* Find the start of the record */
  for (sdex = 0; sdex < parse_len; sdex++) {
    if (strncmp(parse_buf+sdex, ABBREV_START, field_len) == 0) {
      break;
    }
  }

  /* One pass through to remove unwanted hilights on field-abbreviations
     (if any). Copy into a_buf */
  kill_next_hl_end = FALSE;
  for (; sdex < parse_len; ) {
    if (strncmp(parse_buf+sdex, HL_START_STR, hl_start_len) == 0) {
      if (isupper(parse_buf[sdex+hl_start_len]) && 
	  (parse_buf[sdex+hl_start_len + 1] != ' ') &&
	  (parse_buf[sdex+hl_start_len+2] == ' ') && 
	  (parse_buf[sdex+hl_start_len+3] == ' ') &&
	  (parse_buf[sdex+hl_start_len+4] == ' ')) {
	sdex += hl_start_len;
	kill_next_hl_end = TRUE;
      }
    }
    else if (kill_next_hl_end && 
	     (strncmp(parse_buf+sdex, HL_END_STR, hl_end_len) == 0)) {
      kill_next_hl_end = FALSE;
      sdex += hl_end_len;
    }
    else if (strncmp(parse_buf+sdex, PROMPT_STR, PROMPT_LEN) == 0) {
      break;
    }
    if (strncmp(parse_buf+sdex, UGLY_STR, ugly_len) == 0)
      sdex += ugly_len;
    else
      a_buf[tdex++] = parse_buf[sdex++];
  }
  a_buf[tdex] = 0;

  /* A pass through to expand any abbreviated fields, when needed. Copy
     into b_buf. Will grow.
   */
  if (expanded_to_willow) {
    /* copy first header line in */
    parse_len = strlen(a_buf);
    for (sdex = tdex = 0; a_buf[sdex] != '\n';) 
      b_buf[tdex++] = a_buf[sdex++];
    b_buf[tdex++] = a_buf[sdex++];
    
    for (;sdex < parse_len;) {
      /* We should be at a start of field */
      field_dex = 0;
      for (;;) {
	field_buf[field_dex++] = a_buf[sdex++];
	if (start_of_field(a_buf + sdex))
	  break;
	if (!a_buf[sdex])
	  break;
      }
      field_buf[field_dex] = 0;
      
      /* We have one field in field_buf, now expand it */
      exp_field_buf = expand_summ(field_buf);
      for (field_dex = 0; exp_field_buf[field_dex]; ) {
	b_buf[tdex++] = exp_field_buf[field_dex++];
	if (tdex > buf_size) {
	  a_buf = realloc(a_buf, buf_size += parse_len);
	  b_buf = realloc(b_buf, buf_size);
	}
      }
      free(exp_field_buf);
    }
    b_buf[tdex] = 0;
  }
  else {
    /* swap buffers */
    exp_field_buf = b_buf;
    b_buf = a_buf;
    a_buf = exp_field_buf;
  }

  /* A final pass to record the hilites. They will all be valid now */
  parse_len = strlen(b_buf);
  for (sdex = tdex = 0; sdex < parse_len; ) {
    if (b_buf[sdex] == '\033') {
      if (strncmp(b_buf+sdex, HL_START_STR, hl_start_len) == 0) {
	sdex += hl_start_len;
	hl_start(tdex);
      }
      else if (strncmp(b_buf+sdex, HL_END_STR, hl_end_len) == 0) {
	sdex += hl_end_len;
	hl_end(tdex);
      }
    }
    else
      a_buf[tdex++] = b_buf[sdex++];
  }
  a_buf[tdex] = 0;
  free(b_buf);
  free(field_buf);
  return(a_buf);
}

/* Go through the parse buf to prepare a full-record display. Call 
   really_parse_display to do most of the work. Free the buffer 
   allocated there.
 */
parse_display_hcb(void)
{ 
  char *circ;
  int   buf_size;
  int   circ_len, full_len;
  char *full_rec_buf;

  full_rec_buf = really_parse_display();
  buf_size = strlen(full_rec_buf);

  if (do_circ) {
    if (!summ_tab[cur_record].circ)
      summ_tab[cur_record].circ = get_circ(cur_record);
    if (!expanded_to_willow)
      circ = summ_tab[cur_record].circ;
    else
      circ = expand_summ(summ_tab[cur_record].circ);
    if (circ) {
      circ_len = strlen(circ);
      full_len = strlen(full_rec_buf);

      if ((full_len + circ_len + 2) > buf_size)
	full_rec_buf = realloc(full_rec_buf, buf_size += (circ_len + 2));
      strcat(full_rec_buf, circ);
    }
  }

  send_full_record(cur_record, full_rec_buf);
  free(full_rec_buf);
}

/* Queue command for execution when we are back at READY state.
   If we already at READY, force a state transition.
 */
queue_cmd(char *cmd, int next_state)
{
  states[READY][0].out_string = cmd;
  states[READY][0].next_state = next_state;
  if (state == READY)
    force_transition(0);
}

/* Comparison function to be passed into hash routines */
int hash_compare(char *key, format_rec *element)
{
  return(!strcmp(key, element->field));
}

/* Called when the user sends back a response to a driver request.
   For the uwbrs driver, this is either a request for a user-id or
   a password.
 */
user_resp_wcb(packet_rec *packet)
{
  char *cmd;
  
  if (!packet->var.user_resp.answer)
    error("Login cancelled");

  /* Copy user text out of packet and into table. If we are already
     at proper point in table, force a state transition */
  if (!strcmp(last_sent_req, UID_TAG)) {
    cmd = malloc(strlen(UID_CMD) + strlen(packet->data) + 1);
    sprintf(cmd, UID_CMD, packet->data);
    states[AWAIT_LOGIN][0].out_string = cmd;
    states[AWAIT_LOGIN][0].next_state = AWAIT_PW;
    
    if (state == AWAIT_LOGIN)
      force_transition(0);
  }
  else if (!strcmp(last_sent_req, PW_TAG)) {
    states[AWAIT_PW][0].out_string = append_cr(packet->data);
    states[AWAIT_PW][0].next_state = AWAIT_DB;
    if (state == AWAIT_PW)
      force_transition(0);
  } 
}

/* This is an ugly, and temporary, hack, needed because somebody forgot
   to load the new Medline data into BRS with a period at the end of
   the JT field. This breaks the pathetic reference manager databases
   our users have on their PCs. When we reload Medline in '95, this
   can be removed.
*/
char *lameify(char *text, hl_rec *hl_list)
{
  int len, i;
  int mark = 0;
  char *new_text;

  /* If the current database does not have a JT field, we don't need
     to do anything */
  if (!(hash_Lookup(summ_form_tab, hash_HashFunction("JT", 2),
		    hash_compare, "JT")))
    return(text);

  /* Look for a line starting with "JT   ", and stick a period at the end */
  len = strlen(text);
  for (i = 0; i <= len; i++) {
    if (text[i] == '\n') {
      i++;
      if (strncmp(text + i, "JT   ", 5) == 0) {
	while (text[i] && 
	       !((text[i] == '\n') &&
		 (text[i+1] != ' ')))
	  i++;
	if (text[i - 1] != '.')
	  mark = i;
	break;
      }
    }
  }
  if (mark) {
    /* Hilites will all be off by one if we add a character */
    while (hl_list) {
      if (hl_list->start > mark) {
	hl_list->start++;
	hl_list->end++;
      }
      hl_list = hl_list->next;
    }

    new_text = malloc(len + 2);
    for (i = 0; i < mark; i++)
      new_text[i] = text[i];
    new_text[mark] = '.';
    for (i = mark; i <= len; i++)
      new_text[i+1] = text[i];
    return(new_text);
  }
  else
    return(text);
}
