/***********************************************************************
 *                                                                     *
 *              WILLOW(tm) Information Retrieval Software              *
 *     (Washington Information Looker-upper Layered Over Windows)      *
 *             Copyright 1992-1994 University of Washington            *
 *                                                                     *
 *   Willow is a trademark of the University of Washington.  No        *
 *   commercial use of this trademark or software may be made without  *
 *   prior written permission of the University of Washington.         *
 *                                                                     *
 * Permission to use, copy, modify, and distribute Willow(tm) software *
 * and its documentation for any noncommercial purpose is hereby       *
 * granted, provided that the above copyright notice appears in all    *
 * copies and that both the above copyright notice and this permission *
 * notice appear in supporting documentation, and that the name of the *
 * University of Washington not be used in advertising or publicity    *
 * pertaining to distribution of the software without specific,        *
 * written prior permission.  This software is made available "as is", *
 * and THE UNIVERSITY OF WASHINGTON DISCLAIMS ALL WARRANTIES, EXPRESS  *
 * OR IMPLIED, WITH REGARD TO THIS SOFTWARE, INCLUDING WITHOUT         *
 * LIMITATION ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS    *
 * FOR A PARTICULAR PURPOSE, AND IN NO EVENT SHALL THE UNIVERSITY OF   *
 * WASHINGTON BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL     *
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA  *
 * OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, TORT (INCLUDING       *
 * NEGLIGENCE) OR STRICT LIABILITY, ARISING OUT OF OR IN CONNECTION    *
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.                       *
 *                                                                     *
 * For further information please contact:                             *
 * Willow Project                                                      *
 * Information Systems                                                 *
 * University of Washington                                            *
 * 4545 15th NE, 1st floor                                             *
 * Seattle, Washington, 98105, USA                                     *
 * Internet: willow@cac.washington.edu                                 *
 ***********************************************************************/


/*****************************************************
 * host.h
 *
 * These are definitions used by the driver to log
 * into the BRS system at the University of Wasington, 
 * as implemented under AIX.
 *****************************************************/

#define UID_REQ_TEXT    "Please enter your user-ID for this database"
#define PW_REQ_TEXT     "Please enter your password for this database"

#define QUIT_CMD        "O\n"
#define NEXT_SUMM_CMD   "WILLOW_SUM;%s;%d-%d\n"
#define GET_FIELDS_FULL_CMD    "WILLOW_DISPLAY;%s;%d\n"

#define CPR_CMD          "\n"
#define CPR_LOG_CMD      "l\n"
#ifdef ndef
#define CPR_CMD          "%ss\n"
#define CPR_LOG_CMD      "%sls\n"
#endif

#define DB_CMD           "%s\n"
#define LIMIT_CMD       "WILLOW_LT;%s\n"
#define RESULTS_STR     "\nDOCS "
#define OVERFLOW_STR    "TRUNC_LIMIT"
#define PROMPT_STR      "##-->"
#define PROMPT_LEN      5
#define ERROR_STR       "`"

#define MISSING_STR     "\n     Document Number"
#define MISSING_LEN     21
#define INTERRUPT_CMD   "\003"
#define UID_CMD       "%s\n"
#define SUMM_START_STR  "        Do"
#define SUMM_START_STR_LEN  10
#define SUMM_START_FMT  "Document %d\n"
#define ABBREV_START    "Document"
#define EXPANDED_START  "##"

#define HL_START_STR    "\033[1m"
#define HL_END_STR      "\033[m"
#define UGLY_STR        "\033#5\033M\033[m\n"

#define INFO_START      "##"
#define INFO_END        "##"
#define SUMM_FIELD_SEP  ','

/* The second of these must be multiple of the first */
#define SUMM_BLOCK_SIZE         30
#define SUMM_ALLOC_SIZE         300

#define ABBREV_INDENT           5
#define DFLT_LONG_LEN           16
#define MAX_SUMM_FIELD_NAME_LEN 32
#define MAX_QUERY_LEN           2048
#define MAX_SEARCH              250
#define MAX_DISP_LINE           78
#define HASH_TAB_SIZE           137
#define QUIT_TIMEOUT            150
#define DFLT_CIRC_TIMEOUT       90
#define DFLT_MAX_CIRC           500
#define LIST_TIMEOUT            10

/* Special flag to indicate we need to do LCAT title processing */
#define LCAT_EXACT_TITLE      "LCAT-EXACT-TITLE"
#define LCAT_TITLE_CMD        ".TI."
#define LCAT_SINGLE_TITLE_CMD ".TS."
#define LCAT_CALL_NUM         "LCAT-CALL-NUM"
#define LCAT_CNUM_CMD         ".XC,LN."
#define CIRC_ABBREV           "SS"
#define CIRC_LONG_NAME        "Status:"

/* Define a bunch of host computer states */
#define NO_CON          0
#define AWAIT_LOGIN     1
#define AWAIT_PW	2
#define AWAIT_DB        3
#define AWAIT_SEARCH    4 
#define READY           6 
#define AWAIT_STATS     7 
#define DISP_SHRT       8 
#define DISP_LONG       9 
#define AWAIT_INFO      10
/* Don't let this get above 50 without changing MAX_STATES! */

/* format description for the summary fields */
typedef struct _format_rec { 
  int  idex;       /* position in the summary record field vector */
  char field[8];
  char lname[MAX_SUMM_FIELD_NAME_LEN + 1];
  char title_field;
  char summ_field;
  char ext_field;
  int  width;      /* only applicable to title fields */
  char no_wrap;    /* only applicable to summ_fields  */
} format_rec;


/* Table of extra characters which should be considered part of a word. */
static char valid_extra_chars[] =
                      {' ',
		       '\0',
		       '-',  /* word-joiner */
		       '$',  /* wild-card chars */
		       '?',
		       '\t',
		       '(',
		       ')',
		       '"',
		       '.',
		       '/',
		       };

/* Table of extra characters which are normally stripped, but if they
   appear in a call number must get mapped to hyphen. */
static char callnum_extra_chars[] =
                      {',',
		       '&',
		       };

