/*++isc.c+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/****************************************************************************
 * 									    *
 *                 WILCO(tm) Information Retrieval Software		    *
 *        (Washington Information Looker-upper - Character Oriented)	    *
 * 									    *
 *               Copyright 1993-1994 University of Washington		    *
 *           Wilco is a trademark of the University of Washington	    *
 * 									    *
 * Permission to use, copy, modify, and distribute Wilco(tm) software	    *
 * and its documentation for any purpose and without fee is hereby	    *
 * granted, provided that the above copyright notice appears in all	    *
 * copies and that both the above copyright notice and this permission	    *
 * notice appear in supporting documentation, and that the name of the	    *
 * University of Washington not be used in advertising or publicity	    *
 * pertaining to distribution of the software without specific, written	    *
 * prior permission.  This software is made available "as is", and	    *
 * 									    *
 * THE UNIVERSITY OF WASHINGTON DISCLAIMS ALL WARRANTIES, EXPRESS OR	    *
 * IMPLIED, WITH REGARD TO THIS SOFTWARE, INCLUDING WITHOUT LIMITATION	    *
 * ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR   *
 * PURPOSE, AND IN NO EVENT SHALL THE UNIVERSITY OF WASHINGTON BE LIABLE    *
 * FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES	    *
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN    *
 * ACTION OF CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY,	    *
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS	    *
 * SOFTWARE.								    *
 * 									    *
 * For further information please contact:				    *
 * 									    *
 * Scott Heyano								    *
 * Information Systems							    *
 * Computing and Communications						    *
 * University of Washington						    *
 * 4545 15th, 3rd Floor JE-41						    *
 * Seattle, Washington, 98195, USA					    *
 * Internet: slh@cac.washington.edu					    *
 * 									    *
 * This file is not specifically part of ``Wilco'',			    *
 * but is used by it and is covered by the above copyright notice and	    *
 * usage restrictions.							    *
 ****************************************************************************/
/*
 * File	     : isc.c
 * Related   : isc.h
 * Author    : slh (Scott Heyano, slh@cac.washington.edu, slh@nwnet.net)
 * Copyright : (see above)
 *
 * III Server Client code C implementation file.
 */
/* Log:
 *  0.40  08/01/94 - 08/02/94  slh
 *         LCAT: reversed sense of suppression test (and ch codes)
 *         unusable locations (UNUSABLELOC)
 *         rm !WRAP; appendspaces()
 *  0.35  05/06/94  slh
 *         sortnumericbystringtag(): bug fix, extra checks
 *         put in some checks before copy to arrays
 *  0.34  04/25/94 - 04/26/94  slh
 *         allow leading b or not for bibcode requests
 *  0.33  04/19/94  slh
 *         check that interrupt was caused by our time-out
 *         only retry connect() failure when caused by other time-outs
 *  0.32  04/15/94  slh
 *         accept is[bs]n codes for holding requests
 *  0.31  03/30/94  slh
 *         check that summary holdings field is not suppressed
 *         ch display string constants
 *  0.30  02/16/94 ... 03/21/94  slh
 *         GCAT mode
 *         IscGetHoldingStatus(): pszMode param added
 *         LCAT: do numeric sort for parallel like parts
 *         LCAT: added copy info (945$q)
 *         LCAT: added holds prefix
 *  0.22  12/08/93  slh
 *         incorp portability fixes from mitt (incl vax path)
 *         rm !NEWPARAMS; clean out cond'd out stuff
 *  0.21  11/01/93  slh
 *         extractformatstatus(): bugfix
 *  0.20  10/20/93 - 10/21/93, 10/25/93 - 10/27/93  slh
 *         Isc[GS]etParams(): added
 *         add timeouts (TIMEOUT)
 *         wrapping (WRAP)
 *         IscGetHoldingStatus(): accepts bib or barcode ids as requests
 *         ST -> SS for status short label
 *         remove x subfield from status display
 *  0.14  09/23/93  slh
 *         changed circ lookup failed msg
 *  0.13  09/07/93  slh
 *         IscGetHoldingStatus()/extractformatstatus(): interface change
 *  0.12  08/31/93  slh
 *         added case for when no iii record
 *  0.11  08/20/93, 08/23/93 - 08/25/93  slh
 *         added reserve info to holding status output
 *         holding status output spec re-revised
 *         IscDumpMarcRecord(): added
 *  0.10  08/16/92 - 08/17/93  slh
 *         extractformatstatus(): changed output
 *  0.01   7/ 2/93  slh
 *         IscGetHoldingStatus(): interface change, fTitle, fNonAscii
 *         extractformatstatus():bugfix: removed left over debug
 *  0.00   6/ 8/93 -  6/10/93,  6/21/93 -  6/23/93  slh
 *         create
 */
/* Doc:
 * Protocol:
 *   For all message types:
 *	Length (5 bytes)		: ASCII coded length of entire record
 *	<message code> (3 bytes)	: ASCII coded message code
 *	stuff in "s			: literals
 *
 *   100 Request (received by server)
 *	+--------+-------+---------+-------------------------------------+
 *	| Length | "100" | Op Code | ASCII Search Request                |
 *	+--------+-------+---------+-------------------------------------+
 *	Op Code (1 byte):
 *		C	: Cancel previous request
 *		N	: Next set of records
 *		o	: search by RLIN/OCLC number
 *		.	: search by INNOPAC record number
 *		<tag>	: search by tag where <tag> is any of the lower case
 *			  alphabetic tags used by dartm (atcmlfdsib)
 *	Search Request (variable): ASCII/NUL-terminated search request string
 *   200 Error (sent by server)
 *	+--------+-------+--------------+--------------------------------+
 *	| Length | "200" | Error Number | ASCII Error String             |
 *	+--------+-------+--------------+--------------------------------+
 *	Error Number (5 bytes): ASCII coded error code
 *	Error String (variable): ASCII/NUL-terminated error string
 *   300 Marc Data (sent by server)
 *	+--------+-------+--------------+-----------------+--------------+
 *	| Length | "300" | Record Count | Remaining Count | MARC Records |
 *	+--------+-------+--------------+-----------------+--------------+
 *	Record Count (5 bytes): count of records in message
 *	Remaining Count (5 bytes): count of records after this message
 *	MARC Records (variable): a bunch of MARC records
 *
 * MARC:
 *   All numbers are ASCII-coded
 *   Format:
 *	<header><directory><terminator>{<field><terminator>}<db terminator>
 *	<header>	fixed-length header (below)
 *	<directory>	variable-length directory of fields (below)
 *	<field>		field (below)
 *	<terminator>	ASCII RS
 *	<db terminator>	ASCII GS
 *   Header:
 *	0-4	Record Length (total)
 *	5	Status
 *		N	New
 *		C	Changed
 *		D	Deleted (contains only field 001)
 *	6	Record Type
 *	7-9	Not Used
 *	10	Indicator Count
 *	11	Delimiter Count
 *	12-16	Base Address of Data
 *	17-19	Not Used
 *	20-23	Entry Map
 *		20	length of length of field part
 *		21	start of field position (base-1)
 *		22-23	not used
 *   Directory Entry:
 *	0-2	Tag
 *	3-6	Field Length (total)
 *	7-11	Field Address (offset to base)
 *   Data Fields:
 *	<indicator>{<delimiter><subfield>}+<terminator>
 *	<indicator>	version (Indicator Count chars in length)
 *	<delimiter>	ASCII US
 *	<subfield>	data
 *	<terminator>	ASCII RS (same terminator as above, not another)
 *   Note:
 *	fields 010 and below do not have indicators or subfield codes.
 *
 * Holdings (MARC 945):
 *   a		call number
 *   b		barcode
 *   d-i	volume/part hierarchy
 *   l		location code
 *   m		message
 *   n		public note
 *   o		out date
 *   p		price
 *   r		course record number
 *   s		status
 *   t		I type
 *   u		due date
 *   v		load rule
 *   w		numbeer of current holds
 *   x		Imessage
 *   y		accession number
 *   z		recall
 *   1		Icode 1
 *   9		Icode2 suppression
 *		   -    no suppression, normal record
 *		   n    NO DISPLAY
 *		   d    DELETE CODE
 *		   c    RESRV PHOTOCOP
 *		   p    RESRV PERSONAL
 *		   r    REORDERED
 *		   i    MIS:INV
 *		   f    CL FORGED
 *		   s    MISSING:SRM
 *		   b    AT BINDERY
 *		   a    AT MENDERY
 *		   t    TRAINING ITEM
 *		codes that should actually suppress: n, t
 */
/* Paths:
 * ALWAYSLABEL		ALWAYS construct LABELs
 */


#define  UNUSABLELOC	/* UNUSABLE LOCations functionality */
#define  ALWAYSLABEL	/* ALWAYS write out the title and status LABELs */
/* setting TIMEOUT enables the code and sets the default time-out value */
#define  TIMEOUT	10	/* for reads (secs) (update .h) */
#define  WORKAROUND
#define  GCATGROUPLOC	/* do GCAT record GROUPing by LOCation */
#if 0
#define  GCATSORT	/* do some kind of GCAT record SORTing */
#define  GCATSORTLOC	/* do GCAT record LOCation SORTing */
#define  GCATSORTPARTS	/* do GCAT record volume PARTS SORTing */
#endif

#ifdef  GCATSORTLOC
#ifdef  GCATGROUPLOC
#undef  GCATGROUPLOC
#endif/*GCATGROUPLOC*/
#endif/*GCATSORTLOC*/
#ifdef  GCATSORTLOC
#ifndef  GCATSORT
#define  GCATSORT
#endif/*!GCATSORT*/
#endif/*GCATSORTLOC*/
#ifdef  GCATSORTPARTS
#ifndef  GCATSORT
#define  GCATSORT
#endif/*!GCATSORT*/
#endif/*GCATSORTPARTS*/


#ifndef  vax
#include  <stdlib.h>
#include  <malloc.h>
#endif/*!vax*/
#include  <memory.h>
#include  <string.h>
#include  <stdio.h>
#include  <netdb.h>
#ifdef  TIMEOUT
#include  <errno.h>
#endif/*TIMEOUT*/

#include  <sys/types.h>
#include  <sys/socket.h>
#include  <netinet/in.h>
#include  <arpa/inet.h>
#ifndef  INADDR_NONE		/* e.g. missing in Solaris */
#define  INADDR_NONE	-1
#endif/*!INADDR_NONE*/

#include  "isc.h"

/* data */
/* this needs to come before data files */
typedef  struct _StringToString {
  char*		pszKey;
  char*		pszLock;
}  StringToString, *pStringToString;
#include  "status.h"
#include  "suppression.h"
#include  "location.h"
#include  "reserve.h"
#ifdef   UNUSABLELOC
#include  "unusableloc.h"
#endif /*UNUSABLELOC*/
#include  "status-gcat.h"
#include  "location-gcat.h"
#include  "reserve-gcat.h"


/*----------------------------------------------------------------------------
 *							Macros
 *--------------------------------------------------------------------------*/
/*  III Protocol Stuff  */
#if 000
#define  SZSERVICE_DEFAULT	???
#endif
#define  USPORT_DEFAULT		5060

/* Message Codes */
#define  MCODE_REQUEST		100
#define  MCODE_ERROR		200
#define  MCODE_RESULT		300

/* Message Part Lengths */
#define  CCHMESSAGELENGTH	5	/* msg len length		*/
#define  CCHMESSAGECODE		3	/* msg code length		*/
#define  CCHREQUESTOPCODE	1	/* req msg opcode length	*/
#define  CCHERRORCODE		5	/* error msg code length	*/
#define  CCHRESULTRECORD	5	/* result msg rec count length	*/
#define  CCHRESULTREMAIN	5	/* result msg remain rec ct len	*/

/*  MARC Record Stuff  */
/* for MARC record decoding */
#define  CHMARCDBTERM		ASCII_GS	/* Group Separator	*/
#define  CHMARCTERM		ASCII_RS	/* Record Separator	*/
#define  CHMARCDELIM		ASCII_US	/* Unit Separator	*/

/* MARC offsets and lengths */
#define  CCHMARCTAG		3	/* length of tag		*/
#define  OFFMARCHDR		0	/* offset of header		*/
#define  CCHMARCHDR		24	/* length of header		*/
#define  OFFMARCHDRLENGTH	0	/* offset of header:length	*/
#define  CCHMARCHDRLENGTH	5	/* length of header:length	*/
#define  OFFMARCHDRINDICATOR	10	/* offset of header:indicator len*/
#define  CCHMARCHDRINDICATOR	1	/* length of header:indicator len*/
#define  OFFMARCHDRBASE		12	/* offset of header:base addr	*/
#define  CCHMARCHDRBASE		5	/* length of header:base addr	*/

#define  OFFMARCDIR		24	/* offset of directory		*/
#define  CCHMARCDIRENTRY	12	/* length of directory entry	*/
#define  OFFMARCDIRTAG		0	/* offset of dir entry:tag	*/
#define  OFFMARCDIRLENGTH	3	/* offset of dir entry:length	*/
#define  CCHMARCDIRLENGTH	4	/* length of dir entry:length	*/
#define  OFFMARCDIROFFSET	7	/* offset of dir entry:addr	*/
#define  CCHMARCDIROFFSET	5	/* length of dir entry:addr	*/

/*  Status Output Stuff  */
/* for holdings */
#define  CHBIBREQUESTOPCODE	'.'
#define  CHBARREQUESTOPCODE	'b'
#define  CHISBNREQUESTOPCODE	'i'
#define  CHISSNREQUESTOPCODE	'i'
#define  CCHBIBCODE0		8	/* b prefix not included	*/
#define  CCHBIBCODE		9
#define  CCHBARCODE		14
#define  CCHISBNCODE		10
#define  CCHISSNCODE		9
#if 0	/*???not needed and out of date*/
#define  CCHMAXHOLDINGSTATUSREQUEST	MAX(CCHBIBCODE,CCHBARCODE)
#endif

#define  SZTITLETAG		"245"	/* title field			*/
#define  SZSTATUSTAG		"945"	/* holding status field		*/
#define  SZSUMHOLDTAG		"947"	/* summary holdings field	*/
#define  SZSUMHOLDTAG_GCAT	"984"	/* summary holdings field - GCAT*/

/* status field subfields */
#define  CHCALLNMBRTAG		'a'	/* status call number subfield	*/
#define  CHCOPYNMBRTAG		'q'	/* status copy number subfield	*/
#define  CHLOCATIONTAG		'l'	/* status location subfield	*/
#define  CHSTATUSTAG		's'	/* status status subfield	*/
#define  CHDUEDATETAG		'u'	/* status due date subfield	*/
#define  CHHOLDSTAG		'w'	/* status holds subfield	*/
#define  CHSUPPRESSTAG		'9'	/* status suppress subfield	*/
#define  SZSTATUSSORTALIGNTAGS	"adefghi"
			/* status subfields to be sorted (hi -> lo pri)	*/
#if 0000
#define  SZVOLPARTTAGS		"defghi"
					/* status volume/part subfields	*/
#endif
/* actually should rarely even see 3 d's */
#define  SZSTATUSSORTALIGNTAGS_GCAT	"dq"
			/* status subfields to be sorted (hi -> lo pri)	*/
#define  SZTITLETAGS		"amphbcdefgks"	/* title display subfields*/
#define  SZSTATUSDISPLAYTAGS	"adefghiu9wsnt"
				/* status subfields to be displayed	*/
#define  SZSTATUSDISPLAYTAGS_GCAT	"dquwsnt"
				/* status subfields to be displayed	*/
#if 0000
#define  SZSTATUSALIGNTAGS	"adefghi"
				/* fields after which need to be aligned*/
#endif
#define  SZTOBEDISPLAYEDPTAGS	"iba"
	/* status subfields that deterine if field should be displayed	*/

#define  SZNOTSUPPRESSVALUES	"-prifsba"
#if 0
#define  SZSUPPRESSVALUES	"nt"
#endif
#define  SZNOTSUPPRESSVALUES_GCAT	"-r"

/* summary holding field subfields */
#define  CHSUMHOLDSUPPRESTAG	'9'
#define  SZSUMHOLDNOTSUPPRESS	"-or"

#define  SZCHECKEDOUT	\
"Volumes listed below are checked out.  Check the shelves for other volumes."
#define  SZCHECKSHELF		"CHECK THE SHELVES"
#define  SZCHECKSHELF_GCAT	\
			"No volumes checked out, please check the shelves."
#define  SZLOOKUPFAILED		"CIRCULATION INFORMATION NOT AVAILABLE"
#define  SZLOOKUPFAILED_GCAT	\
			"No volumes checked out, please check the shelves."

#define  SZTITLELABELFULL	"Title: "	/* title label		*/
#define  SZSTATUSLABELFULL	"Status: "	/* status label		*/
#define  SZTITLELABELABBREV	"TI"		/* title label		*/
#define  SZSTATUSLABELABBREV	"SS"		/* status label		*/
#define  CCHSTATUSINDENT1	4		/* indent for each status*/
#define  CCHSTATUSINDENT2	2		/* indent for each status*/
#define  SZDUEDATEPREFIX	"  DUE "	/* duedate prefix (add)	*/
#define  SZCALLNMBRPREFIX	""		/* call # prefix (add)	*/
#define  SZHOLDSPREFIX		"  Holds: "	/* # of holds prefix (add)*/
#define  SZCOPYNMBRPREFIX	"  Copy "	/* copy # prefix (add)	*/
#define  SZVOLPARTPREFIX	" "		/* volume/part prefix (add)*/
#define  SZOTHERPREFIX		"  "		/* other item prefix (add)*/

#define  SZWHITE		" \t\n"		/* white space		*/
#define  SZBREAK		SZWHITE

#define  CCHMAXPARTVAL		32		/* max char for part val*/
#define  CCHMAXLOCCODE		32		/* max char for loc code*/
#define  CCHMAXSTATUSCODE	8		/* max char for status code*/

/**  Intrinsics  **/
#define  SZDIGITS		"0123456789"
#define  NUL			'\0'		/* NUL			*/
#define  ASCII_GS		'\035'		/* GROUP SEPARATOR	*/
#define  ASCII_RS		'\036'		/* RECORD SEPARATOR	*/
#define  ASCII_US		'\037'		/* UNIT SEPARATOR	*/

#ifndef  FALSE
#define  FALSE	0
#define  TRUE	1
#endif/*!FALSE*/

/**  Functions  **/
#define  NUMBER(a)	(sizeof(a) / sizeof(a[0]))
#ifndef  MIN
#define  MIN(a,b)	((a) <= (b) ? (a) : (b))
#define  MAX(a,b)	((a) > (b) ? (a) : (b))
#endif/*!MIN*/


/*----------------------------------------------------------------------------
 *							Globals
 *--------------------------------------------------------------------------*/
#ifdef  TIMEOUT
static  int	gTimeout	= TIMEOUT;
static  int	gfTimedout	= FALSE;
#endif/*TIMEOUT*/


/*----------------------------------------------------------------------------
 *							Private Routines
 *--------------------------------------------------------------------------*/
/*-------------------------------------- General Purpose Routines	*/

#ifdef  TIMEOUT
static
void  tocb  (void)
{
  gfTimedout = TRUE;
}
#endif/*TIMEOUT*/


/*
 * Writes data to file, in parts if necessary.
 * Return: TRUE if successful, FALSE otherwise.
 */
static
int  writedata  (int	fd	,
		 char	*vbData	,
		 int	cbData	)
{
  register  int		cbWrote, cbLeft;
  register  char	*pb;

  for (cbLeft = cbData, pb = vbData ;
       cbLeft > 0 && -1 != (cbWrote = write( fd , pb , cbLeft )) ;
       cbLeft -= cbWrote, pb += cbWrote)
    ;

  return cbLeft == 0;
}


/*
 * Reads data from file, in parts if necessary.
 * Return: TRUE if successful, FALSE otherwise.
 */
static
int  readdata  (int	fd	,
		char	*vbData	,
		int	cbData	)
{
#ifdef  TIMEOUT
  int		id;
#endif/*TIMEOUT*/
  register  int		cbRead, cbLeft;
  register  char	*pb;

#ifdef  TIMEOUT
  if (gTimeout > 0)
    id = malarm_add( gTimeout , tocb );
#endif/*TIMEOUT*/
  for (cbLeft = cbData, pb = vbData ;
#ifdef  TIMEOUT
       cbLeft > 0 &&
       (cbRead = read( fd , pb , cbLeft )) &&
       (cbRead != -1 || errno == EINTR && !gfTimedout) ;
#else /*TIMEOUT*/
       cbLeft > 0 && 0 < (cbRead = read( fd , pb , cbLeft )) ;
#endif/*TIMEOUT*/
       cbLeft -= cbRead, pb += cbRead)
    ;

#ifdef  TIMEOUT
  if (cbLeft == 0)
    if (gTimeout > 0)
      malarm_cancel( id );
  gfTimedout = FALSE;		/* in case was set */
#endif/*TIMEOUT*/

  return cbLeft == 0;
}


static
void*  CwtMalloc  (size_t  sizet)
{
  if (!sizet)
    return (void*)NULL;
  else
    return (void*)malloc( sizet );
}


static
void*  CwtRealloc  (void*	cwp	,
		    size_t	sizet	)
{
  if (!sizet) {
    free( cwp );
    return (void*)NULL;
  } else if (!cwp) {
    return (void*)malloc( sizet );
  } else {
    return (void*)realloc( cwp , sizet );
  }
}


static
void*  CwtCalloc  (size_t	Number	,
		   size_t	sizet	)
{
  if (!(Number * sizet))
    return (void*)NULL;
  else
    return (void*)calloc( Number , sizet );
}


static
void  CwtFree  (void*  cwp)
{
  if (!cwp)
    return;
  else
    free( cwp );
}


/*
 * strips leading and trailing white-space.
 * Return: length of stripped string.
 */
static
int  StrStrip  (char	*str		, /* string to be stripped */
		char	*strWhite	) /* white space characters */
{
  int		cch, cchLead;
  register  int		i;

  cch = strlen( str );
  cchLead = strspn( str , strWhite );

 /* strip leading whilte space */
  while (strspn( str+cch-1 , strWhite ))
    cch--;
  str[cch] = NUL;

 /* strip trailing whilte space */
  cch -= cchLead;
  /* <= copies NUL */
  for (i=0 ; i<=cch ; i++)
    str[i] = str[i+cchLead];

  return cch;
}


/*
 * Return: length of token
 */
static
int  StrNTokLen  (char*	strSrc	, /* token source */
		  char*	strDel	, /* token delimiters */
		  int	cchSrc	) /* length of source (< 0 => ignore) */
{
  char		*pchSrc, *pchDel;

  pchSrc = strSrc;
  while (*pchSrc && (cchSrc < 0 || pchSrc - strSrc < cchSrc)) {
    pchDel = strDel;
    while (*pchDel) {
      if (*pchSrc == *pchDel)
	return pchSrc - strSrc;
      pchDel++;
    }
    pchSrc++;
  }

  return (pchSrc - strSrc);
}


static
char*  keytolock  (StringToString	vsts[]	,
		   int			csts	,
		   char*		pszKey	)
{
  register  int		ists;

  for (ists=0 ; ists<csts ; ists++)
    if (!strcmp( vsts[ists].pszKey , pszKey ))
      return vsts[ists].pszLock;				/*RETURN*/

  return (char*)NULL;						/*RETURN*/
}


/*
 * Params:
 *	host	name or dotted-decimal addr of other system
 *	service	name of service bening requsted (can be NULL iff port > 0)
 *	port	port number of service (host byte order)
 *		== 0 : use port number of service obtained through service
 *		<  0 : bind a local reserved port
 *		>  0 : use as port of service
 * Retuns: socket fd if successful, -1 otherwise.
 * What?: from iii's tcp_open_cl(), cleaned-up and added some stuff.
 */
static
int  connecttoservice  (char*	host	,
			char*	service	,
			u_short	port	)
{
  int		s, resvport;
  unsigned long		ulAddr;
  struct sockaddr_in	ssai;
  struct servent	*psse;
#ifdef  TIMEOUT
  int		id;
  int		Status;
#endif/*TIMEOUT*/
/* Variable */
  struct hostent	*pshe		= (struct hostent*)NULL;

  memset( (void*)&ssai , 0 , sizeof(struct sockaddr_in) );

  if (service && *service) {
    if (!(psse = getservbyname( service , "tcp" )))
      return -1;						/*RETURN*/
    if (port > 0)
      ssai.sin_port = htons( port );
    else
      ssai.sin_port = psse->s_port;
  } else {
    if (port < 0)
      return -1;						/*RETURN*/
    ssai.sin_port = htons( port );
  }

  if (INADDR_NONE != (ulAddr = inet_addr( host ))) {
    ssai.sin_family = AF_INET;
    memcpy( (void*)&ssai.sin_addr.s_addr , (void*)&ulAddr ,
	    sizeof(unsigned long) );
  } else {
    if (!(pshe = gethostbyname( host )))
      return -1;						/*RETURN*/
    ssai.sin_family = pshe->h_addrtype;
/*???should loop throug all addrs below if !pshe*/
    memcpy( (void*)&ssai.sin_addr.s_addr , (void*)pshe->h_addr ,
	    (size_t)pshe->h_length );
  }

  if (port >= 0) {
    if (0 > (s = socket( ssai.sin_family , SOCK_STREAM , 0 )))
      return -1;						/*RETURN*/
  } else {
    resvport = IPPORT_RESERVED - 1;
    if (0 > (s = rresvport( &resvport )))
      return -1;						/*RETURN*/
  }

#ifdef  TIMEOUT
  if (gTimeout > 0)
    id = malarm_add( gTimeout , tocb );
  while (0 > (Status = connect(
		s , (struct sockaddr*)&ssai , sizeof(struct sockaddr_in) )) &&
	 (errno == EINTR && !gfTimedout))
    ;
  if (0 > Status) {
#else /*TIMEOUT*/
  if (0 > connect(
		s , (struct sockaddr*)&ssai , sizeof(struct sockaddr_in) )) {
#endif/*TIMEOUT*/
    (void)shutdown( s , 2 );
    (void)close( s );
    sleep( 5 );
    return -1;
  } else {
#ifdef  TIMEOUT
    if (gTimeout > 0)
      malarm_cancel( id );
    gfTimedout = FALSE;		/* in case was set */
#endif/*TIMEOUT*/
    return s;							/*RETURN*/
  }
}


/*-------------------------------------- MARC Record Routines		*/

static
char*  getdirentry  (char	vchData[]	,
		     int	iOccur		)
{
  char		szBase[CCHMARCHDRBASE+1];
  int		Base;

  (void)memcpy( szBase , vchData+OFFMARCHDRBASE , CCHMARCHDRBASE );
  szBase[CCHMARCHDRBASE] = NUL;
  Base = atoi( szBase );

  if (CCHMARCHDR + iOccur * CCHMARCDIRENTRY > Base - 1)
    return (char*)NULL;
  else
    return vchData+(CCHMARCHDR + iOccur * CCHMARCDIRENTRY);
}


static
char*  findfield  (char		vchData[]	,
		   char*	pszTag		, /* may be NULL */
		   int		iOccur		,
		   int*		pcchField	)
{
  char		szBase[CCHMARCHDRBASE+1];
  char		szLength[CCHMARCDIRLENGTH+1], szOffset[CCHMARCDIROFFSET+1];
  int		Base;
  int		Offset, Length;
  char		*pch;

  (void)memcpy( szBase , vchData+OFFMARCHDRBASE , CCHMARCHDRBASE );
  szBase[CCHMARCHDRBASE] = NUL;
  Base = atoi( szBase );

  for (pch=vchData+CCHMARCHDR ;
       pch<vchData+Base-1 && iOccur>=0 ;
       pch+=CCHMARCDIRENTRY)
    if (!pszTag || !strncmp( pch , pszTag , CCHMARCTAG ))
      iOccur--;

  if (iOccur >= 0) {
    if (pcchField)
      *pcchField = -1;
    return (char*)NULL;
  }

  pch -= CCHMARCDIRENTRY;	/* step back one */

  (void)memcpy( szLength , pch+OFFMARCDIRLENGTH , CCHMARCDIRLENGTH );
  szLength[CCHMARCDIRLENGTH] = NUL;
  Length = atoi( szLength );
  (void)memcpy( szOffset , pch+OFFMARCDIROFFSET , CCHMARCDIROFFSET );
  szOffset[CCHMARCDIROFFSET] = NUL;
  Offset = atoi( szOffset );

  if (pcchField)
    *pcchField = Length;
  return vchData + Base + Offset;
}


static
char*  findsubfield  (char	vchField[]	,
		      int	cchField	,
		      char	chTag		, /* may be NUL */
		      int	iOccur		,
		      int*	pcchSub		) /*return*/
{
  char		*pch, *pchLast;
/* Data */
  static  char	sdszDelim[]		= {			/*STATIC*/
    CHMARCDELIM, CHMARCTERM, NUL,
  };

  for (pch=vchField ;
       pch<vchField+cchField && iOccur>=-1 ;
       pch+=2)
    if (pch = strpbrk( pch , sdszDelim ))
      if (*pch != CHMARCTERM && (!chTag || *(pch+1) == chTag)) {
	if (iOccur == 0)
	  pchLast = pch;
	iOccur--;
      } else if (iOccur <= -1) {
	iOccur--;
      }

  if (iOccur >= -1 || !pch) {
    if (pcchSub)
      *pcchSub = -1;
    return (char*)NULL;
  }

  /* adjust */
  pch -= 2;
  pchLast += 2;

  if (pcchSub)
    *pcchSub = pch - pchLast;
  return pchLast;
}


static
int  dumpsubfield  (FILE*	pfileOut	,
		    char	vchSub[]	,
		    int		cchSub		)
{
  (void)fprintf( pfileOut , "%*.*s" , cchSub , cchSub , vchSub );

  return TRUE;
}


static
int  dumpfield  (FILE*	pfileOut	,
		 char	vchField[]	,
		 int	cchField	)
{
  char		*vchSub;
  int		 cchSub;
  register  int		 iSub;
/* Variable */
  int		fRetVal		= TRUE;

  for (iSub=0 ;
       fRetVal &&
       (vchSub=findsubfield(
			vchField , cchField , '\0' , iSub , &cchSub )) ;
       iSub++) {
    /* since the context is a marc field and
       know the char before is the subfield id */
    if (fRetVal)
      (void)fprintf( pfileOut , "\n\t%c " , vchSub[-1] );
    fRetVal = fRetVal &&
      dumpsubfield( pfileOut , vchSub , cchSub );
  }

  if (iSub == 0)
    (void)fprintf( pfileOut , "%*.*s" ,
		   cchField - 1 , cchField - 1 , vchField );

  return fRetVal;
}


/*-------------------------------------- III Message Protocol Routines	*/

static
int  getmessage  (int		s		,
		  char*		pvchData[]	,
		  int*		pcchData	)
{
  char		szLength[CCHMESSAGELENGTH+1];
/* Variable */
  char		*vch		= (char*)NULL;
  int		 cch		= 0;
  int		fContinue	= TRUE;

  fContinue = fContinue &&
    readdata( s , szLength , CCHMESSAGELENGTH );
  szLength[CCHMESSAGELENGTH] = NUL;

  /* must be greater than zero since includes msg length */
  fContinue = fContinue &&
    (cch = atoi( szLength ));
  if (pvchData) {
    fContinue = fContinue &&
      (vch = (char*)CwtMalloc( (cch + 1)*sizeof(char) ));
    if (fContinue)
      (void)memcpy( (void*)vch , (void*)szLength , (size_t)CCHMESSAGELENGTH );
    fContinue = fContinue &&
      readdata( s , vch+CCHMESSAGELENGTH , cch - CCHMESSAGELENGTH );
    if (fContinue)
      vch[cch] = NUL;
  }

  if (!fContinue) {
    CwtFree( vch );
    vch = (char*)NULL;
    cch = 0;
  }

  if (pvchData)
    *pvchData = vch;
  if (pcchData)
    *pcchData = cch;
  return fContinue;
}


static
int  getmessageinparts  (int		s		,
			 int*		pfError		,
			 int*		pcRecord	,
			 int*		pcRemain	,
			 int*		pError		,
			 char**		pvchData	,
			 int*		pcchData	)
{
  char		szLength[CCHMESSAGELENGTH+1], szCode[CCHMESSAGECODE+1];
  char		szRecord[CCHRESULTRECORD+1], szRemain[CCHRESULTREMAIN+1];
  char		szError[CCHERRORCODE+1];
  int		Length, Code;
/* Variable */
  int		cRecord		= 0;
  int		cRemain		= 0;
  int		Error		= 0;
  char		*vch		= (char*)NULL;
  int		 cch		= 0;
  int		fContinue	= TRUE;

  fContinue = fContinue &&
    readdata( s , szLength , CCHMESSAGELENGTH );
  szLength[CCHMESSAGELENGTH] = NUL;
  if (fContinue)
    Length = atoi( szLength );
  fContinue = fContinue &&
    readdata( s , szCode , CCHMESSAGECODE );
  szCode[CCHMESSAGECODE] = NUL;
  if (fContinue)
    Code = atoi( szCode );

  cch = Length;
  cch -= CCHMESSAGELENGTH + CCHMESSAGECODE;
  switch (Code) {
  case MCODE_RESULT:
    fContinue = fContinue &&
      readdata( s , szRecord , CCHRESULTRECORD );
    szRecord[CCHRESULTRECORD] = NUL;
    if (fContinue)
      cRecord = atoi( szRecord );
    fContinue = fContinue &&
      readdata( s , szRemain , CCHRESULTREMAIN );
    szRemain[CCHRESULTREMAIN] = NUL;
    if (fContinue)
      cRemain = atoi( szRemain );

    cch -= CCHRESULTRECORD + CCHRESULTREMAIN;
    break;
  case MCODE_ERROR:
    fContinue = fContinue &&
      readdata( s , szError , CCHERRORCODE );
    szError[CCHERRORCODE] = NUL;
    if (fContinue)
      Error = atoi( szError );

    cch -= CCHERRORCODE;
    break;
  case MCODE_REQUEST:
  default:
    break;
  }

  if (pvchData) {
    fContinue = fContinue &&
      (!cch ||
       (vch = (char*)CwtMalloc( (cch + 1)*sizeof(char) )));
    if (fContinue)
      vch[0] = NUL;		/* in case read fails */
    fContinue = fContinue &&
      readdata( s , vch , cch );
    if (fContinue)
      vch[cch] = NUL;
  }

  if (!fContinue) {
    CwtFree( vch );
    vch		= (char*)NULL;
    cch		= 0;
    cRecord	= 0;
    cRemain	= 0;
    Error	= 0;
  }

  if (pfError)
    *pfError	= Code == MCODE_ERROR;
  if (pcRecord)
    *pcRecord	= cRecord;
  if (pcRemain)
    *pcRemain	= cRemain;
  if (pError)
    *pError	= Error;
  if (pvchData)
    *pvchData	= vch;
  if (pcchData)
    *pcchData	= cch;
  return fContinue;
}


/*-------------------------------------- Holding Status Routines	*/

static
char*  findsumhold  (char	vchData[]		,
		     char*	pszTags			,
		     char*	pszNotSuppressCodes	,
		     int*	pcchField		)
{
  char		*vchField;
  int		 cchField;
  char		*vchSub;
  int		 cchSub;
  register  int		iField;

  for (iField=0 ;
       (vchField = findfield( vchData , pszTags , iField , &cchField )) &&
       (vchSub = findsubfield(
		vchField , cchField , CHSUMHOLDSUPPRESTAG , 0 , &cchSub )) ;
       iField++)
    if (strchr( pszNotSuppressCodes , *vchSub )) {
      *pcchField = cchField;
      return vchField;
    }

  *pcchField = -1;
  return (char*)NULL;
}


static
void  sortnumericbystringtag  (char*	vvchField[]	,
			       int	vcchField[]	,
			       int	cvchField	,
			       char	chTag		,
			       int	iOccur		)
{
  char		*vchSub;
  int		 cchSub;
  char		sz1[CCHMAXPARTVAL+1], sz2[CCHMAXPARTVAL+1];
  char		*psz1, *psz2;
  int		val1, val2;
  char		*vchT;
  int		cchT;
  int		fswap;
  register  int		ich1, ich2;
  register  int		i, j;

  for (i=0 ; i<cvchField ; i++)
    for (j=cvchField-1 ; j>i ; j--) {
      if (vchSub = findsubfield( vvchField[j-1] , vcchField[j-1] ,
				 chTag , iOccur , &cchSub )) {
	if (cchSub > sizeof(sz1) - 1)
	  cchSub = sizeof(sz1) - 1;
	(void)memcpy( (void*)sz1 , (void*)vchSub , (size_t)cchSub );
	sz1[cchSub] = NUL;
	psz1 = sz1;
      } else {
	psz1 = "";
      }
      if (vchSub = findsubfield( vvchField[j] , vcchField[j] ,
				 chTag , iOccur , &cchSub )) {
	if (cchSub > sizeof(sz2) - 1)
	  cchSub = sizeof(sz2) - 1;
	(void)memcpy( (void*)sz2 , (void*)vchSub , (size_t)cchSub );
	sz2[cchSub] = NUL;
	psz2 = sz2;
      } else {
	psz2 = "";
      }

      /* find beginning of integer */
      ich1 = strcspn( psz1 , SZDIGITS );
      ich2 = strcspn( psz2 , SZDIGITS );
      /* if prefixes not same length... */
      if (ich1 != ich2) {
       /* ...try to backup before any white-space */
	while (ich1 && strchr( SZWHITE , psz1[ich1-1] ))
	  ich1--;
	while (ich2 && strchr( SZWHITE , psz2[ich2-1] ))
	  ich2--;
      }

      /* should do a numeric comparison? */
      if (*psz1 && ich1 == ich2 && !strncmp( psz1 , psz2 , ich1 )) {
	val1 = atoi( psz1+ich1 );
	val2 = atoi( psz2+ich2 );
	fswap = val2 < val1;
      } else {
	fswap = 0 < strcmp( psz1 , psz2 );
      }
      if (fswap) {
	vchT		= vvchField[j-1];
	vvchField[j-1]	= vvchField[j];
	vvchField[j]	= vchT;
	cchT		= vcchField[j-1];
	vcchField[j-1]	= vcchField[j];
	vcchField[j]	= cchT;
      }
    }
}


#ifdef  GCATSORTPARTS
static
void  sortbystringtag  (char*	vvchField[]	,
			int	vcchField[]	,
			int	cvchField	,
			char	chTag		,
			int	iOccur		)
{
  char		*vchSub;
  int		 cchSub;
  char		*vch1, *vch2, *vchT;
  int		cch;
  int		cchT;
  register  int		i, j;

  for (i=0 ; i<cvchField ; i++)
    for (j=cvchField-1 ; j>i ; j--) {
      if (vchSub = findsubfield( vvchField[j-1] , vcchField[j-1] ,
				 chTag , iOccur , &cchSub )) {
	vch1 = vchSub;
	cch = cchSub;
      } else {
	vch1 = "";
	cch = 0;
      }
      if (vchSub = findsubfield( vvchField[j] , vcchField[j] ,
				 chTag , iOccur , &cchSub )) {
	vch2 = vchSub;
	cch = MIN(cch,cchSub);
      } else {
	vch2 = "";
	cch = 0;
      }
/*???nul not even there*/
      cch += 1;		/* so NUL will get in there for diff len strings */

      if (0 < memcmp( vch1 , vch2 , cch )) {
	vchT		= vvchField[j-1];
	vvchField[j-1]	= vvchField[j];
	vvchField[j]	= vchT;
	cchT		= vcchField[j-1];
	vcchField[j-1]	= vcchField[j];
	vcchField[j]	= cchT;
      }
    }
}
#endif/*GCATSORTPARTS*/
#if 0000
static
void  sortbystringtag  (char*		vvchField[]	,
			int		vcchField[]	,
			int		cvchField	,
			char		chTag		)
{
  char		*vchSub;
  int		 cchSub;
  char		*vch1, *vch2, *vchT;
  int		cch;
  int		cchT;
  register  int		i, j;

  for (i=0 ; i<cvchField ; i++)
    for (j=cvchField-1 ; j>i ; j--) {
      if (vchSub = findsubfield( vvchField[j-1] , vcchField[j-1] ,
				 chTag , 0 , &cchSub )) {
	vch1 = vchSub;
	cch = cchSub;
      } else {
	vch1 = "";
	cch = 0;
      }
      if (vchSub = findsubfield( vvchField[j] , vcchField[j] ,
				 chTag , 0 , &cchSub )) {
	vch2 = vchSub;
	cch = MIN(cch,cchSub);
      } else {
	vch2 = "";
	cch = 0;
      }
      cch += 1;		/* so NUL will get in there for diff len strings */

      if (0 < memcmp( vch1 , vch2 , cch )) {
	vchT		= vvchField[j-1];
	vvchField[j-1]	= vvchField[j];
	vvchField[j]	= vchT;
	cchT		= vcchField[j-1];
	vcchField[j-1]	= vcchField[j];
	vcchField[j]	= cchT;
      }
    }
}
#endif


static
void  sortbyloc  (char*			vvchField[]	,
		  int			vcchField[]	,
		  int			cvchField	,
		  StringToString	vstsLocation[]	,
		  int			cstsLocation	)
{
  char		*vchSub;
  int		 cchSub;
  char		szLoc1[CCHMAXLOCCODE+1], szLoc2[CCHMAXLOCCODE+1];
  char		*psz1, *psz2, *vchT;
  int		cchT;
  register  int		i, j;

  for (i=0 ; i<cvchField ; i++)
    for (j=cvchField-1 ; j>i ; j--) {
      if (vchSub = findsubfield( vvchField[j-1] , vcchField[j-1] ,
				 CHLOCATIONTAG , 0 , &cchSub )) {
	if (cchSub > sizeof(szLoc1) - 1)
	  cchSub = sizeof(szLoc1) - 1;
	(void)memcpy( (void*)szLoc1 , (void*)vchSub , (size_t)cchSub );
	szLoc1[cchSub] = NUL;
	(void)StrStrip( szLoc1 , SZWHITE );
	if (!(psz1 = keytolock( vstsLocation , cstsLocation , szLoc1 )))
#if 0000
	if (!(psz1 = keytolock(
			gdastsLocation , NUMBER(gdastsLocation) , szLoc1 )))
#endif
	  psz1 = szLoc1;
      } else {
	psz1 = "";
      }
      if (vchSub = findsubfield( vvchField[j] , vcchField[j] ,
				 CHLOCATIONTAG , 0 , &cchSub )) {
	if (cchSub > sizeof(szLoc2) - 1)
	  cchSub = sizeof(szLoc2) - 1;
	(void)memcpy( (void*)szLoc2 , (void*)vchSub , (size_t)cchSub );
	szLoc2[cchSub] = NUL;
	(void)StrStrip( szLoc2 , SZWHITE );
	if (!(psz2 = keytolock( vstsLocation , cstsLocation , szLoc2 )))
#if 0000
	if (!(psz2 = keytolock(
			gdastsLocation , NUMBER(gdastsLocation) , szLoc2 )))
#endif
	  psz2 = szLoc2;
      } else {
	psz2 = "";
      }

      if (0 < strcmp( psz1 , psz2 )) {
	vchT		= vvchField[j-1];
	vvchField[j-1]	= vvchField[j];
	vvchField[j]	= vchT;
	cchT		= vcchField[j-1];
	vcchField[j-1]	= vcchField[j];
	vcchField[j]	= cchT;
      }
    }
}


#ifdef   GCATGROUPLOC
static
void  groupbyloc  (char*		vvchField[]	,
		   int			vcchField[]	,
		   int			cvchField	)
{
  char		*vchSub;
  int		 cchSub;
  char		szLoc1[CCHMAXLOCCODE+1], szLoc2[CCHMAXLOCCODE+1];
  char		*psz1, *psz2, *vchT;
  int		cchT;
  int		fswap;
  register  int		i, j, k;

  for (i=0 ; i<cvchField-1 ; i++) {
    if (vchSub = findsubfield( vvchField[i] , vcchField[i] ,
			       CHLOCATIONTAG , 0 , &cchSub )) {
      if (cchSub > sizeof(szLoc1) - 1)
	cchSub = sizeof(szLoc1) - 1;
      (void)memcpy( (void*)szLoc1 , (void*)vchSub , (size_t)cchSub );
      szLoc1[cchSub] = NUL;
      (void)StrStrip( szLoc1 , SZWHITE );
      psz1 = szLoc1;
    } else {
      psz1 = "";
    }
    if (vchSub = findsubfield( vvchField[i+1] , vcchField[i+1] ,
			       CHLOCATIONTAG , 0 , &cchSub )) {
      if (cchSub > sizeof(szLoc2) - 1)
	cchSub = sizeof(szLoc2) - 1;
      (void)memcpy( (void*)szLoc2 , (void*)vchSub , (size_t)cchSub );
      szLoc2[cchSub] = NUL;
      (void)StrStrip( szLoc2 , SZWHITE );
      psz2 = szLoc2;
    } else {
      psz2 = "";
    }

    if (strcmp( psz1 , psz2 )) {
      fswap = FALSE;
/*???could be optimized some by moving contiguous runs at one time*/
      for (j=i+2 ; j<cvchField && !fswap ; j++) {
	if (vchSub = findsubfield( vvchField[j] , vcchField[j] ,
				   CHLOCATIONTAG , 0 , &cchSub )) {
	  if (cchSub > sizeof(szLoc2) - 1)
	    cchSub = sizeof(szLoc2) - 1;
	  (void)memcpy( (void*)szLoc2 , (void*)vchSub , (size_t)cchSub );
	  szLoc2[cchSub] = NUL;
	  (void)StrStrip( szLoc2 , SZWHITE );
	  psz2 = szLoc2;
	} else {
	  psz2 = "";
	}

	if (!strcmp( psz1 , psz2 )) {
	  fswap = TRUE;
	  vchT = vvchField[j];
	  cchT = vcchField[j];
	  for (k=j ; k>=i+2 ; k--) {
	    vvchField[k] = vvchField[k-1];
	    vcchField[k] = vcchField[k-1];
	  }
	  vvchField[i+1] = vchT;
	  vcchField[i+1] = cchT;
	}
      }
    }
  }
}
#endif /*GCATGROUPLOC*/


static
void  sortstatus  (char*		vvchField[]	,
		   int			vcchField[]	,
		   int			cvchField	,
		   StringToString	vstsLocation[]	,
		   int			cstsLocation	)
{
  register  int		ich;

  /* sort from low to high priority */
  for (ich=NUMBER(SZSTATUSSORTALIGNTAGS)-2 ; ich>=0 ; ich--)
    sortnumericbystringtag(
	vvchField , vcchField , cvchField , SZSTATUSSORTALIGNTAGS[ich] , 0 );
  sortbyloc(
	vvchField , vcchField , cvchField , vstsLocation , cstsLocation );
}
#if 0000
static
void  sortstatus  (char*		vvchField[]	,
		   int			vcchField[]	,
		   int			cvchField	,
		   StringToString	vstsLocation[]	,
		   int			cstsLocation	)
{
  register  int		ich;

  /* sort from low to high priority */
  for (ich=NUMBER(SZVOLPARTTAGS)-2 ; ich>=0 ; ich--)
    sortbystringtag( vvchField , vcchField , cvchField , SZVOLPARTTAGS[ich] );
  sortbystringtag( vvchField , vcchField , cvchField , CHCALLNMBRTAG );
  sortbyloc(
	vvchField , vcchField , cvchField , vstsLocation , cstsLocation );
}
#endif


#ifdef  GCATSORT
static
void  sortstatusx  (char*		vvchField[]	,
		    int			vcchField[]	,
		    int			cvchField	,
		    StringToString	vstsLocation[]	,
		    int			cstsLocation	)
{
  register  int		ich;
  register  int		ich2;
  register  int		iOccur;

#ifdef  GCATSORTPARTS
  /* sort from low to high priority */
  for (ich=NUMBER(SZSTATUSSORTALIGNTAGS_GCAT)-2 ; ich>=0 ; ich--) {
    for (ich2=ich-1 ;
	 ich2>=0 &&
	 SZSTATUSSORTALIGNTAGS_GCAT[ich2]==SZSTATUSSORTALIGNTAGS_GCAT[ich] ;
	 ich2--)
      ;
    iOccur = ich - ich2 - 1;
    sortbystringtag( vvchField , vcchField , cvchField ,
		      SZSTATUSSORTALIGNTAGS_GCAT[ich] , iOccur );
  }
#endif/*GCATSORTPARTS*/

#ifdef  GCATSORTLOC
  sortbyloc(
	vvchField , vcchField , cvchField , vstsLocation , cstsLocation );
#endif/*GCATSORTLOC*/
}
#endif/*GCATSORT*/


static
int  getmax  (char*		vvchField[]	,
	      int		vcchField[]	,
	      int		cvchField	,
	      char		chTag		,
	      int		iOccur		)
{
  int		 cchSub;
  register  int		i;
/* Variable */
  int		max		= 0;

  for (i=0 ; i<cvchField ; i++) {
    if (!findsubfield(
		vvchField[i] , vcchField[i] , chTag , iOccur , &cchSub ))
      cchSub = 0;
    max = MAX(max,cchSub);
  }

  return max;
}
#if 0
static
int  getmax  (char*		vvchField[]	,
	      int		vcchField[]	,
	      int		cvchField	,
	      char		chTag		)
{
  int		 cchSub;
  register  int		i;
/* Variable */
  int		max		= 0;

  for (i=0 ; i<cvchField ; i++) {
    if (!findsubfield( vvchField[i] , vcchField[i] , chTag , 0 , &cchSub ))
      cchSub = 0;
    max = MAX(max,cchSub);
  }

  return max;
}
#endif


/*
 * NOTE: assumes that if there is repetition it is contiguous.
 */
static
void  getmaxes  (char*	vvchField[]	,
		 int	vcchField[]	,
		 int	cvchField	,
		 char*	pszTags		,
		 int	aMax[]		)
{
  register  int		ich;
  register  int		iOccur;
/* Variable */
  char		chLast		= '\0';		/* can't occur in data */
/* Constant */
  int		cch		= strlen( pszTags );

  for (ich=0 ; ich<cch ; ich++) {
    if (chLast == pszTags[ich])
      iOccur++;
    else
      iOccur = 0;
    aMax[ich] = getmax(
		vvchField , vcchField , cvchField , pszTags[ich] , iOccur );
    chLast = pszTags[ich];
  }
}
#if 0000
static
void  getmaxes  (char*	vvchField[]	,
		 int	vcchField[]	,
		 int	cvchField	,
		 char*	pszTags		,
		 int	aMax[]		)
{
  register  int		ich;

  for (ich=strlen( pszTags )-1 ; ich>=0 ; ich--)
    aMax[ich] = getmax( vvchField , vcchField , cvchField , pszTags[ich] );
}
#endif


static
int  isdateset  (char*	vchSub	,
		 int	cchSub	)
{
  /* should be safe to do since it stops when it sees weirdness and
     a valid date wouldn't have a zero value in it
   */
  return !(!atoi( vchSub ));
}


/*???these two could be combined now and a param added*/
static
int  issuppressed  (char*	vchSub	,
		    int		cchSub	)
{
  return !(int)memchr( (void*)SZNOTSUPPRESSVALUES ,
		       (int)*vchSub , NUMBER(SZNOTSUPPRESSVALUES) - 1 );
#if 0
  return !(!(int)memchr( (void*)SZSUPPRESSVALUES ,
			 (int)*vchSub , NUMBER(SZSUPPRESSVALUES) - 1 ));
#endif
}


static
int  issuppressedGCAT  (char*	vchSub	,
			int	cchSub	)
{
  return !(int)memchr( (void*)SZNOTSUPPRESSVALUES_GCAT ,
		       (int)*vchSub , NUMBER(SZNOTSUPPRESSVALUES_GCAT) - 1 );
}


static
int  anystatustobedisplayed  (char*	vvchField[]	,
			      int	vcchField[]	,
			      int	cvchField	)
{
  char		*vchSub;
  int		 cchSub;
  register  int		ivchField;

  for (ivchField=0 ; ivchField<cvchField ; ivchField++) {
    if ((vchSub = findsubfield( vvchField[ivchField] , vcchField[ivchField] ,
			        CHDUEDATETAG , 0 , &cchSub )) &&
	isdateset( vchSub , cchSub ))
      return TRUE;						/*RETURN*/
/*???maybe should be true if this field not present*/
    if ((vchSub = findsubfield( vvchField[ivchField] , vcchField[ivchField] ,
			        CHSTATUSTAG , 0 , &cchSub )) &&
	*vchSub != '-')
      return TRUE;						/*RETURN*/
  }

  return FALSE;							/*RETURN*/
}


#ifdef   UNUSABLELOC
static
int  anyunusablelocations  (char	vchData[]	,
			    char*	pszTags		)
{
  char		*vchField;
  int		 cchField;
  char		*vchSub;
  int		 cchSub;
  char		szLoc[CCHMAXLOCCODE+1];
  register  int		iField;

  for (iField=0 ;
       (vchField = findfield( vchData , pszTags , iField , &cchField )) &&
       (vchSub = findsubfield( vchField , cchField ,
			       CHLOCATIONTAG , 0 , &cchSub )) ;
       iField++) {
    if (cchSub > sizeof(szLoc) - 1)
      cchSub = sizeof(szLoc) - 1;
    (void)memcpy( (void*)szLoc , (void*)vchSub , (size_t)cchSub );
    szLoc[cchSub] = NUL;
    (void)StrStrip( szLoc , SZWHITE );
    if (keytolock( gdastsUnusableLoc , NUMBER(gdastsUnusableLoc) , szLoc ))
      return TRUE;						/*RETURN*/
  }

  return FALSE;							/*RETURN*/
}
#endif /*UNUSABLELOC*/


#if 0
static
int  anystatustobedisplayed  (char*	vvchField[]	,
			      int	vcchField[]	,
			      int	cvchField	)
{
  char		*vchSub;
  int		 cchSub;
#ifdef   UNUSABLELOC
  char		szLoc[CCHMAXLOCCODE+1];
#endif /*UNUSABLELOC*/
  register  int		ivchField;
#ifdef   UNUSABLELOC
/* Variable */
  int		fDisplay	= FALSE;
#endif /*UNUSABLELOC*/

  for (ivchField=0 ; ivchField<cvchField ; ivchField++) {
    if ((vchSub = findsubfield( vvchField[ivchField] , vcchField[ivchField] ,
			        CHDUEDATETAG , 0 , &cchSub )) &&
	isdateset( vchSub , cchSub ))
      return TRUE;						/*RETURN*/
/*???maybe should be true if this field not present*/
    if ((vchSub = findsubfield( vvchField[ivchField] , vcchField[ivchField] ,
			        CHSTATUSTAG , 0 , &cchSub )) &&
	*vchSub != '-')
      return TRUE;						/*RETURN*/

#ifdef   UNUSABLELOC
    /* if not already determined and location subfield found... */
    if (!fDisplay &&
	(vchSub = findsubfield( vvchField[ivchField] , vcchField[ivchField] ,
			        CHLOCATIONTAG , 0 , &cchSub ))) {
      if (cchSub > sizeof(szLoc) - 1)
	cchSub = sizeof(szLoc) - 1;
      (void)memcpy( (void*)szLoc , (void*)vchSub , (size_t)cchSub );
      szLoc[cchSub] = NUL;
      (void)StrStrip( szLoc , SZWHITE );
      fDisplay = !!keytolock(
		gdastsUnusableLoc , NUMBER(gdastsUnusableLoc) , szLoc );
    }
#endif /*UNUSABLELOC*/
  }

#ifdef   UNUSABLELOC
  return fDisplay;						/*RETURN*/
#else  /*UNUSABLELOC*/
  return FALSE;							/*RETURN*/
#endif /*UNUSABLELOC*/
}
#endif


static
int  statustobedisplayed  (char*	vchField	,
			   int		cchField	,
			   char*	pszToBeTags	)
{
  char		*vchSub;
  int		 cchSub;

/*???maybe should be true if this field not present*/
  if ((vchSub = findsubfield( vchField , cchField ,
			      CHSTATUSTAG , 0 , &cchSub )) &&
      *vchSub != '-')
    return TRUE;						/*RETURN*/
  if ((vchSub = findsubfield( vchField , cchField ,
			      CHDUEDATETAG , 0 , &cchSub )) &&
      isdateset( vchSub , cchSub ))
    return TRUE;						/*RETURN*/
  if (pszToBeTags && *pszToBeTags)
    if ((vchSub = findsubfield( vchField , cchField ,
			        CHSUPPRESSTAG , 0 , &cchSub )) &&
	strchr( pszToBeTags , (int)*vchSub ))
#if 0000
	(int)memchr( (void*)SZTOBEDISPLAYEDPTAGS ,
		     (int)*vchSub , NUMBER(SZTOBEDISPLAYEDPTAGS) - 1 ))
#endif
      return TRUE;						/*RETURN*/
  if (vchSub = findsubfield( vchField , cchField ,
			     CHHOLDSTAG , 0 , &cchSub ))
    return TRUE;						/*RETURN*/

  return FALSE;							/*RETURN*/
}


#if 0
static
int  statustobedisplayedGCAT  (char*	vchField	,
			       int	cchField	)
{
  char		*vchSub;
  int		 cchSub;

/*???maybe should be true if this field not present*/
  if ((vchSub = findsubfield( vchField , cchField ,
			      CHSTATUSTAG , 0 , &cchSub )) &&
      *vchSub != '-')
    return TRUE;						/*RETURN*/
  if ((vchSub = findsubfield( vchField , cchField ,
			      CHDUEDATETAG , 0 , &cchSub )) &&
      isdateset( vchSub , cchSub ))
    return TRUE;						/*RETURN*/
  if (vchSub = findsubfield( vchField , cchField ,
			     CHHOLDSTAG , 0 , &cchSub ))
    return TRUE;						/*RETURN*/

  return FALSE;							/*RETURN*/
}
#endif


static
int  appendwrapped  (char**	pvchOutput	,
		     int*	pcchOutput	,
		     int*	pcchLine	,
		     char*	vchInput	,
		     int	cchInput	,
		     int	fNonAscii	,
		     int	nchLine		,
		     int	nchIndent	,
		     int	fLine		)
{
  int		cchSeg;
  int		cch;

  for ( ; cchInput ; ) {
    cchSeg = StrNTokLen( vchInput , SZBREAK , cchInput );
    /* wrapping on and line has been started (to avoid inf loops) and
       line would be tooooo long... */
    if (nchLine > 0 && *pcchLine != nchIndent && *pcchLine &&
	*pcchLine + cchSeg > nchLine) {
      cch = nchIndent + 1;
      if (!(*pvchOutput = (char*)CwtRealloc(
			*pvchOutput , (*pcchOutput + cch + 1)*sizeof(char) )))
	return FALSE;
      (*pvchOutput)[*pcchOutput] = '\n';
      *pcchOutput += 1;
      (void)memset( (void*)(*pvchOutput+*pcchOutput) ,
		    (int)' ' , (size_t)nchIndent );
      *pcchOutput += nchIndent;
      *pcchLine = nchIndent;
    }
    cch = cchSeg + 1;		/* may be one more than needed */
    if (!(*pvchOutput = (char*)CwtRealloc(
			*pvchOutput , (*pcchOutput + cch + 1)*sizeof(char) )))
      return FALSE;
    for ( ; cchSeg ; cchSeg--,cchInput--,vchInput++) {
      if (fNonAscii || '\000' <= *vchInput && *vchInput <= '\177') {
	(*pvchOutput)[(*pcchOutput)++] = *vchInput;
	(*pcchLine)++;
      }
    }
    if (cchInput) {
      (*pvchOutput)[(*pcchOutput)++] = *vchInput;
      (*pcchLine)++;
      vchInput += 1;
      cchInput -= 1;
    }
  }

  if (fLine) {
    cch = 1;
    *pvchOutput = (char*)CwtRealloc(
			*pvchOutput , (*pcchOutput + cch + 1)*sizeof(char) );
    (*pvchOutput)[*pcchOutput] = '\n';
    *pcchOutput += 1;
    *pcchLine = 0;
  }

  return TRUE;
}


static
int  appendspaces  (char**	pvchOutput	,
		    int*	pcchOutput	,
		    int*	pcchLine	,
		    int		cSpace		,
		    int		fLine		)
{
  if (fLine) {
    if (!(*pvchOutput = (char*)CwtRealloc(
			*pvchOutput , (*pcchOutput + 1 + 1)*sizeof(char) )))
      return FALSE;						/*RETURN*/
    (*pvchOutput)[*pcchOutput] = '\n';
    *pcchOutput += 1;
    *pcchLine = 0;
  }

  if (!(*pvchOutput = (char*)CwtRealloc(
		*pvchOutput , (*pcchOutput + cSpace + 1)*sizeof(char) )))
    return FALSE;						/*RETURN*/
  (void)memset(
	(void*)(*pvchOutput+*pcchOutput) , (int)' ' , (size_t)cSpace );
  *pcchOutput += cSpace;
  *pcchLine += cSpace;

  return TRUE;
}


/*
 * Note: for proper looking output, location subfields must be present.
 */
/*????need to strip suppress field?*/
/*???fContinue?*/
/*???check reallocs (append*())*/
/*???fStatus not used right now (as if always true)*/
static
char*  extractformatstatus  (char	vchData[]	,
			     int	cchData 	,
			     int	fTitle		,
			     int	fStatus		,
			     int	fNonAscii	,
			     int	nchLine		,
			     int	fShortLabels	,
			     int	nchIndent	,
			     int*	pcchStatus	)
{
  char		*vchField;
  int		 cchField;
  char		*vchSub;
  int		 cchSub;
  int		cch;
  char		*pchTmp;
  int		max, maxtotal, sofar;
  int		cchStatusLast;
  char		*pszLocC, *pszLocP, *pszLocT;
#ifdef   UNUSABLELOC
  char		*pszLocCode, *pszLocStatus;
#endif /*UNUSABLELOC*/
  char		szLoc1[CCHMAXLOCCODE+1], szLoc2[CCHMAXLOCCODE+1];
  char		szCode[CCHMAXSTATUSCODE+1];
  int		fDueDate;		/* due date present? */
  char		*strPrefix;
  int		cchLine;
  register  int		ivchField;
  register  char	*pch;
/* Variable */
  char		*vchFieldSH		= (char*)NULL;
  int		 cchFieldSH		= 0;
  char		**vvchField		= (char**)NULL;
  int		 *vcchField		= (int*)NULL;
  int		  cvchField		= 0;
  char		*pszStatus		= (char*)NULL;
  int		 cchStatus		= 0;
/* Data */
  static  char  sdachTitleSub[]		= SZTITLETAGS;
  static  char	sdachStatusAlign[]	= SZSTATUSSORTALIGNTAGS;
#if 0000
  static  char	sdachStatusAlign[]	= SZSTATUSALIGNTAGS;
#endif
/* Variable */
  char		afchTitleSub[NUMBER(sdachTitleSub)];
  int		aMax[NUMBER(sdachStatusAlign)];
/* Constant */
  char		*pszTitleLabel		=
		fShortLabels ? SZTITLELABELABBREV : SZTITLELABELFULL;
  char		*pszStatusLabel		=
		fShortLabels ? SZSTATUSLABELABBREV : SZSTATUSLABELFULL;

  if (nchIndent == -1)
    nchIndent = MAX(strlen( pszTitleLabel ),strlen( pszStatusLabel ));

  /* so there is always room for the NUL added at the end */
  cch = 0;
  pszStatus = (char*)CwtRealloc(
			pszStatus , (cchStatus + cch + 1)*sizeof(char) );
  cchStatus += cch;
  cchLine = 0;

/* Title section */
  /* Title field: "amphbcdefgks" (indent each and term with \n) */
  if (fTitle) {
#ifdef  ALWAYSLABEL
    cch = nchIndent;
    pszStatus = (char*)CwtRealloc(
			pszStatus , (cchStatus + cch + 1)*sizeof(char) );
    (void)sprintf( pszStatus+cchStatus , "%-*.*s" ,
		   cch , cch , pszTitleLabel );
    cchStatus += cch;
    cchLine += cch;
#endif/*ALWAYSLABEL*/

    if ((vchField = findfield( vchData , SZTITLETAG , 0 , &cchField ))) {
      (void)memset( (void*)afchTitleSub ,
		    (int)'0' , sizeof(afchTitleSub) - 1 );
      afchTitleSub[sizeof(afchTitleSub) - 1] = NUL;
      for (pch=sdachTitleSub ; *pch ; pch++)
	if (vchSub = findsubfield(
				vchField , cchField , *pch , 0 , &cchSub )) {
	  /* if not first subfield found... */
	  if (atoi( afchTitleSub )) {
	    (void)appendspaces(
		&pszStatus , &cchStatus , &cchLine , nchIndent , FALSE );
#if 0
	    cch = nchIndent;
	    pszStatus = (char*)CwtRealloc(
			pszStatus , (cchStatus + cch + 1)*sizeof(char) );
	    (void)memset( (void*)(pszStatus+cchStatus) ,
			  (int)' ' , (size_t)cch );
	    cchStatus += cch;
	    cchLine += cch;
#endif
#ifndef  ALWAYSLABEL
	  } else {
	    cch = nchIndent;
	    pszStatus = (char*)CwtRealloc(
			pszStatus , (cchStatus + cch + 1)*sizeof(char) );
	    (void)sprintf( pszStatus+cchStatus , "%-*.*s" ,
			   cch , cch , pszTitleLabel );
	    cchStatus += cch;
	    cchLine += cch;
#endif/*ALWAYSLABEL*/
	  }
	  (void)appendwrapped(
			&pszStatus , &cchStatus , &cchLine , vchSub , cchSub ,
			fNonAscii , nchLine , nchIndent , TRUE );
	  afchTitleSub[pch-sdachTitleSub] = '1';
	}
    }
#ifdef  ALWAYSLABEL
    /* if none of the subfields were found... */
    if (!atoi( afchTitleSub )) {
      (void)appendspaces( &pszStatus , &cchStatus , &cchLine , 0 , TRUE );
#if 0
      cch = 1;
      pszStatus = (char*)CwtRealloc(
			pszStatus , (cchStatus + cch + 1)*sizeof(char) );
      pszStatus[cchStatus] = '\n';
      cchStatus += 1;
      cchLine = 0;
#endif
    }
#endif/*ALWAYSLABEL*/
  }

/* Get summary holdings field */
  vchFieldSH = findsumhold(
		vchData , SZSUMHOLDTAG , SZSUMHOLDNOTSUPPRESS , &cchFieldSH );
#if 0
  vchFieldSH = findfield( vchData , SZSUMHOLDTAG , 0 , &cchFieldSH );
#endif
  
/* Find and sort status fields */
  for ( ;
       vchField = findfield( vchData , SZSTATUSTAG , cvchField , &cchField ) ;
       cvchField++ ) {
/*???checks*/
    vvchField = (char**)CwtRealloc(
				vvchField , (cvchField + 1)*sizeof(char*) );
    vcchField = (int*)CwtRealloc(
				vcchField , (cvchField + 1)*sizeof(int) );
    vvchField[cvchField] = vchField;
    vcchField[cvchField] = cchField;
  }
  sortstatus( vvchField , vcchField , cvchField ,
	      gdastsLocation , NUMBER(gdastsLocation) );
#if 0000
  sortstatus( vvchField , vcchField , cvchField );
#endif
  (void)memset( (void*)aMax , 0 , sizeof(aMax) );
  getmaxes( vvchField , vcchField , cvchField , sdachStatusAlign , aMax );

  maxtotal = 0;
  for (ivchField=0 ; ivchField<NUMBER(aMax) ; ivchField++) {
    maxtotal += aMax[ivchField];
    if (aMax[ivchField])
      switch (sdachStatusAlign[ivchField]) {
      case CHCALLNMBRTAG:
	maxtotal += NUMBER(SZCALLNMBRPREFIX) - 1;
	break;
      case CHCOPYNMBRTAG:
	maxtotal += NUMBER(SZCOPYNMBRPREFIX) - 1;
	break;
      default:
	maxtotal += NUMBER(SZVOLPARTPREFIX) - 1;
	break;
      }
  }
#if 0000
  maxtotal = 0;
  for (ivchField=0 ; ivchField<NUMBER(sdachStatusAlign)-1 ; ivchField++) {
    maxtotal += aMax[ivchField];
    if (aMax[ivchField])
      maxtotal += NUMBER(SZVOLPARTPREFIX) - 1;
  }
  /* take away for an extra and add for the other */
  maxtotal -= NUMBER(SZVOLPARTPREFIX) - 1;
  maxtotal += NUMBER(SZCALLNMBRPREFIX) - 1;
#endif

/* Status section */
  cch = nchIndent;
  pszStatus = (char*)CwtRealloc(
			pszStatus , (cchStatus + cch + 1)*sizeof(char) );
  (void)sprintf( pszStatus+cchStatus , "%-*.*s" ,
		 cch , cch , pszStatusLabel );
  cchStatus += cch;
  cchLine += cch;

  szLoc1[0] = NUL;
  szLoc2[0] = NUL;
  pszLocC = szLoc1;
  pszLocP = szLoc2;
  cchStatusLast = cchStatus;
#if 0	/* outline */
  if (!vchFieldSH && cvchField ||
      vchFieldSH &&
      anystatustobedisplayed( vvchField , vcchField , cvchField )) {
    if (vchFieldSH) {
	SZCHECKEDOUT
    }
    for (ivchField=0 ; ivchField<cvchField ; ivchField++) {
      if ((!(vchSub = findsubfield(
		vchField , cchField , CHSUPPRESSTAG , 0 , &cchSub )) ||
	   !issuppressed( vchSub , cchSub )) &&
	  (!vchFieldSH ||
	   statustobedisplayed(
			vchField , cchField , SZTOBEDISPLAYEDPTAGS ))) {
	if (vchSub = findsubfield(
			vchField , cchField , CHLOCATIONTAG , 0 , &cchSub )) {
	  /* if no summary holding field case or
	     current and previous locations are different... */
	  if (strcmp( pszLocC , pszLocP )) {
	    if (vchFieldSH || cchStatusLast != cchStatus)
	      /* indent */
	    else
	      /* stay at loc */
	  }
	}
      }
    }
  } else if (vchFieldSH) {
    if (anyunusablelocations( vchData , SZSUMHOLDTAG )) {
      SZCHECKSHELF
    } else {
      /* blank */
    }
  } else {		/* no record or ? */
    SZLOOKUPFAILED
  }
#endif
  if (!vchFieldSH && cvchField ||
      vchFieldSH &&
      anystatustobedisplayed( vvchField , vcchField , cvchField )) {
    if (vchFieldSH) {
      (void)appendwrapped( &pszStatus , &cchStatus , &cchLine ,
			   SZCHECKEDOUT , strlen( SZCHECKEDOUT ) ,
			   fNonAscii , nchLine , nchIndent , TRUE );
      cchStatusLast = cchStatus;
    }

    for (ivchField=0 ; ivchField<cvchField ; ivchField++) {
      vchField = vvchField[ivchField];
      cchField = vcchField[ivchField];
/*???maybe only do if field is present?*/
      if ((!(vchSub = findsubfield(
		vchField , cchField , CHSUPPRESSTAG , 0 , &cchSub )) ||
	   !issuppressed( vchSub , cchSub )) &&
	  (!vchFieldSH ||
	   statustobedisplayed(
			vchField , cchField , SZTOBEDISPLAYEDPTAGS ))) {
#if 0000
	  (!vchFieldSH || statustobedisplayed( vchField , cchField ))) {
#endif
       /* Location info */
	if (vchSub = findsubfield(
			vchField , cchField , CHLOCATIONTAG , 0 , &cchSub )) {
	  if (cchSub > CCHMAXLOCCODE)
	    cchSub = CCHMAXLOCCODE;	/* ok since setting again below */
	  (void)memcpy( (void*)pszLocC , (void*)vchSub , (size_t)cchSub );
	  pszLocC[cchSub] = NUL;
	  (void)StrStrip( pszLocC , SZWHITE );
	  if (!(vchSub = keytolock(
			gdastsLocation , NUMBER(gdastsLocation) , pszLocC )))
	    vchSub = pszLocC;
	  cchSub = strlen( vchSub );

#ifdef   UNUSABLELOC
	  pszLocCode = pszLocC;
#endif /*UNUSABLELOC*/

	  /* if no summary holding field case or
	     current and previous locations are different... */
	  if (strcmp( pszLocC , pszLocP )) {
	    if (vchFieldSH || cchStatusLast != cchStatus)
	      cch = nchIndent;
	    else
	      cch = 0;
	    (void)appendspaces(
			&pszStatus , &cchStatus , &cchLine , cch , FALSE );
#if 0
/*???hmmm, this wasn't increassing cchLine*/
	    pszStatus = (char*)CwtRealloc(
			pszStatus , (cchStatus + cch + 1)*sizeof(char) );
	    (void)memset( (void*)(pszStatus+cchStatus) ,
			  (int)' ' , (size_t)cch );
	    cchStatus += cch;
#endif
	    (void)appendwrapped(
			&pszStatus , &cchStatus , &cchLine , vchSub , cchSub ,
			fNonAscii , nchLine , nchIndent , TRUE );

	    /* must be swapped so that they point to different memory areas */
	    pszLocT = pszLocC;
	    pszLocC = pszLocP;
	    pszLocP = pszLocT;
	  }
	}

       /* Status info */
	(void)appendspaces( &pszStatus , &cchStatus , &cchLine ,
			    nchIndent + CCHSTATUSINDENT2 , FALSE );
#if 0
	cch = nchIndent + CCHSTATUSINDENT2;
	pszStatus = (char*)CwtRealloc(
			pszStatus , (cchStatus + cch + 1)*sizeof(char) );
	(void)memset( (void*)(pszStatus+cchStatus) ,
		        (int)' ' , (size_t)cch );
        cchStatus += cch;
	cchLine += cch;
#endif

	sofar = 0;
	fDueDate = FALSE;
	for (pch=SZSTATUSDISPLAYTAGS ; *pch ; pch++)
	  if (vchSub = findsubfield(
				vchField , cchField , *pch , 0 , &cchSub )) {
	    /* if have gone past the alignment subfields and
	       not indented far enough... */
	    if (!strchr( sdachStatusAlign , *pch ) && sofar < maxtotal) {
	      (void)appendspaces( &pszStatus , &cchStatus , &cchLine ,
				  maxtotal - sofar , FALSE );
	      sofar = maxtotal;
#if 0
	      cch = maxtotal - sofar;
	      pszStatus = (char*)CwtRealloc(
			pszStatus , (cchStatus + cch + 1)*sizeof(char) );
	      (void)memset( (void*)(pszStatus+cchStatus) ,
			    (int)' ' , (size_t)cch );
	      cchStatus += cch;
	      cchLine += cch;
	      sofar += cch;
#endif
	    }
	    switch (*pch) {
	    case 'd':
	    case 'e':
	    case 'f':
	    case 'g':
	    case 'h':
	    case 'i':
	      strPrefix = SZVOLPARTPREFIX;
	      break;
	    case 'a':	/* call number */
	      strPrefix = SZCALLNMBRPREFIX;
	      break;
	    case 'u':	/* due date */
	      /* if actually contains a date... */
	      if (isdateset( vchSub , cchSub )) {
		strPrefix = SZDUEDATEPREFIX;
		fDueDate = TRUE;
	      } else {
		vchSub = (char*)NULL;
		cchSub = 0;
	      }
	      break;
	    case '9':	/* suppression */
	      (void)memcpy( (void*)szCode , (void*)vchSub , (size_t)cchSub );
	      szCode[cchSub] = NUL;
	      (void)StrStrip( szCode , SZWHITE );
	      if (!(vchSub = keytolock(
		gdastsSuppression , NUMBER(gdastsSuppression) , szCode )))
		vchSub = szCode;
	      cchSub = strlen( vchSub );
	      if (vchSub && *vchSub) {
		strPrefix = SZOTHERPREFIX;
	      }
	      break;
	    case 's':	/* status */
	      (void)memcpy( (void*)szCode , (void*)vchSub , (size_t)cchSub );
	      szCode[cchSub] = NUL;
	      (void)StrStrip( szCode , SZWHITE );
	      /* special case for - */
#ifdef   UNUSABLELOC
	      pszLocStatus = keytolock(
		gdastsUnusableLoc , NUMBER(gdastsUnusableLoc) , pszLocCode );
#endif /*UNUSABLELOC*/
	      /* special case for - */
#ifdef   UNUSABLELOC
	      if (*szCode != '-' || !fDueDate && !pszLocStatus) {
#else  /*UNUSABLELOC*/
	      if (*szCode != '-' || !fDueDate) {
#endif /*UNUSABLELOC*/
		if (!(vchSub = keytolock(
			gdastsStatus , NUMBER(gdastsStatus) , szCode )))
		  vchSub = szCode;
		cchSub = strlen( vchSub );
		if (vchSub && *vchSub) {
		  strPrefix = SZOTHERPREFIX;
		}
	      } else {
		vchSub = (char*)NULL;
		cchSub = 0;
	      }
	      break;
	    case 't':	/* type of reserve */
	      (void)memcpy( (void*)szCode , (void*)vchSub , (size_t)cchSub );
	      szCode[cchSub] = NUL;
	      (void)StrStrip( szCode , SZWHITE );
	      if (!(vchSub = keytolock(
			gdastsReserve , NUMBER(gdastsReserve) , szCode )))
		vchSub = szCode;
	      cchSub = strlen( vchSub );
	      if (vchSub && *vchSub) {
		strPrefix = SZOTHERPREFIX;
	      }
	      break;
	    case 'q':	/* copy number */
	      strPrefix = SZCOPYNMBRPREFIX;
	      break;
	    case 'w':	/* number of holds */
	      strPrefix = SZHOLDSPREFIX;
	      break;
	    case 'n':	/* reserve related */
	    default:
	      strPrefix = SZOTHERPREFIX;
	      break;
	    }
	    cch = strlen( strPrefix );
	    if (vchSub && *vchSub) {
	      /* if wrapping and have actually written part of line and
		 it would get too long... */
	      if (nchLine > 0 && sofar && cchLine + cch + cchSub > nchLine) {
		(void)appendspaces( &pszStatus , &cchStatus , &cchLine ,
				    nchIndent + CCHSTATUSINDENT2 , TRUE );
#if 0
		cch = nchIndent + CCHSTATUSINDENT2 + 1;
		pszStatus = (char*)CwtRealloc(
			pszStatus , (cchStatus + cch + 1)*sizeof(char) );
		pszStatus[cchStatus] = '\n';
		cchStatus += 1;
		(void)memset( (void*)(pszStatus+cchStatus) ,
			      (int)' ' , (size_t)(cch - 1) );
		cchStatus += cch - 1;
		cchLine = cch - 1;
#endif
		sofar = 0;
	      }

	      /* never wrap here */
	      (void)appendwrapped(
			&pszStatus , &cchStatus , &cchLine , strPrefix , cch ,
			fNonAscii , -1 , nchIndent , FALSE );
	      sofar += cch;
#if 0
	      pszStatus = (char*)CwtRealloc(
			pszStatus , (cchStatus + cch + 1)*sizeof(char) );
	      (void)memcpy( (void*)(pszStatus+cchStatus) ,
			    (void*)strPrefix , (size_t)cch );
	      cchStatus += cch;
	      cchLine += cch;
	      sofar += cch;
#endif

	      (void)appendwrapped(
			&pszStatus , &cchStatus , &cchLine , vchSub , cchSub ,
			fNonAscii , -1 , nchIndent , FALSE );
	      sofar += cchSub;
#if 0
	      cch = cchSub;
	      pszStatus = (char*)CwtRealloc(
			pszStatus , (cchStatus + cch + 1)*sizeof(char) );
	      (void)memcpy( (void*)(pszStatus+cchStatus) ,
			    (void*)vchSub , (size_t)cchSub );
	      cchStatus += cch;
	      cchLine += cch;
	      sofar += cch;
#endif

	      if (pchTmp = strchr( sdachStatusAlign , *pch )) {
		max = aMax[pchTmp - sdachStatusAlign];
		if (max - cchSub > 0) {
		  (void)appendspaces(
				&pszStatus , &cchStatus , &cchLine ,
				max - cchSub + CCHSTATUSINDENT2 , FALSE );
		  sofar += max - cchSub;
#if 0
		  cch = max - cchSub;
		  pszStatus = (char*)CwtRealloc(
			pszStatus , (cchStatus + cch + 1)*sizeof(char) );
		  (void)memset( (void*)(pszStatus+cchStatus) ,
			        (int)' ' , (size_t)cch );
		  cchStatus += cch;
		  cchLine += cch;
		  sofar += cch;
#endif
		}
	      }
	    }
	  }
	(void)appendspaces( &pszStatus , &cchStatus , &cchLine ,
			    0 , TRUE );
#if 0
	cch = 1;
	pszStatus = (char*)CwtRealloc(
			pszStatus , (cchStatus + cch + 1)*sizeof(char) );
	pszStatus[cchStatus] = '\n';
	cchStatus += cch;
	cchLine = 0;
#endif
      }
    }
  } else if (vchFieldSH) {
#ifdef   UNUSABLELOC
    if (anyunusablelocations( vchData , SZSUMHOLDTAG )) {
      (void)appendspaces( &pszStatus , &cchStatus , &cchLine , 0 , TRUE );
#if 0
/*???hmm, wasn't resetting cchLine, but probably didn't matter here*/
      cch = 1;
      pszStatus = (char*)CwtRealloc(
			pszStatus , (cchStatus + cch + 1)*sizeof(char) );
      pszStatus[cchStatus] = '\n';
      cchStatus += cch;
#endif
    } else {
#endif /*UNUSABLELOC*/
      (void)appendwrapped( &pszStatus , &cchStatus , &cchLine ,
			   SZCHECKSHELF , strlen( SZCHECKSHELF ) ,
			   fNonAscii , nchLine , nchIndent , TRUE );
#ifdef   UNUSABLELOC
    }
#endif /*UNUSABLELOC*/
  } else {		/* no record or ? */
    (void)appendwrapped( &pszStatus , &cchStatus , &cchLine ,
			 SZLOOKUPFAILED , strlen( SZLOOKUPFAILED ) ,
			 fNonAscii , nchLine , nchIndent , TRUE );
  }

#ifdef  ALWAYSLABEL
  /* if none of the subfields were found... */
  if (cchStatusLast == cchStatus) {
    (void)appendspaces( &pszStatus , &cchStatus , &cchLine , 0 , TRUE );
#if 0
/*???hmm, wasn't resetting cchLine, but probably didn't matter here*/
    cch = 1;
    pszStatus = (char*)CwtRealloc(
			pszStatus , (cchStatus + cch + 1)*sizeof(char) );
    pszStatus[cchStatus] = '\n';
    cchStatus += cch;
#endif
  }
#endif/*ALWAYSLABEL*/
  pszStatus[cchStatus] = NUL;

  CwtFree( vvchField );
  CwtFree( vcchField );
  *pcchStatus = cchStatus;

  return pszStatus;
}


/*
 * Note: for proper looking output, location subfields must be present.
 */
/*????need to strip suppress field?*/
/*???fContinue?*/
/*???check reallocs*/
/*???fStatus not used right now (as if always true)*/
static
char*  extractformatstatusGCAT  (char	vchData[]	,
				 int	cchData 	,
				 int	fTitle		,
				 int	fStatus		,
				 int	fNonAscii	,
				 int	nchLine		,
				 int	fShortLabels	,
				 int	nchIndent	,
				 int*	pcchStatus	)
{
  char		*vchField;
  int		 cchField;
  char		*vchSub;
  int		 cchSub;
  int		cch;
  char		*pchTmp;
  int		max, maxtotal, sofar;
  int		cchStatusLast;
  char		*pszLocC, *pszLocP, *pszLocT;
  char		szLoc1[CCHMAXLOCCODE+1], szLoc2[CCHMAXLOCCODE+1];
  char		szCode[CCHMAXSTATUSCODE+1];
  int		fDueDate;		/* due date present? */
  char		*strPrefix;
  int		cchLine;
  char		chLast;
  register  int		ivchField;
  register  char	*pch;
  register  int		iSub;
/* Variable */
  char		*vchFieldSH		= (char*)NULL;
  int		 cchFieldSH		= 0;
  char		**vvchField		= (char**)NULL;
  int		 *vcchField		= (int*)NULL;
  int		  cvchField		= 0;
  char		*pszStatus		= (char*)NULL;
  int		 cchStatus		= 0;
/* Data */
  static  char  sdachTitleSub[]		= SZTITLETAGS;
  static  char	sdachStatusAlign[]	= SZSTATUSSORTALIGNTAGS_GCAT;
/* Variable */
  char		afchTitleSub[NUMBER(sdachTitleSub)];	/* room for NUL */
  int		aMax[NUMBER(sdachStatusAlign) - 1];
/* Constant */
  char		*pszTitleLabel		=
		fShortLabels ? SZTITLELABELABBREV : SZTITLELABELFULL;
  char		*pszStatusLabel		=
		fShortLabels ? SZSTATUSLABELABBREV : SZSTATUSLABELFULL;

  if (nchIndent == -1)
    nchIndent = MAX(strlen( pszTitleLabel ),strlen( pszStatusLabel ));

  /* so there is always room for the NUL added at the end */
  cch = 0;
  pszStatus = (char*)CwtRealloc(
			pszStatus , (cchStatus + cch + 1)*sizeof(char) );
  cchStatus += cch;
  cchLine = 0;

/* Title section */
  /* Title field: "amphbcdefgks" (indent each and term with \n) */
  if (fTitle) {
#ifdef  ALWAYSLABEL
    cch = nchIndent;
    pszStatus = (char*)CwtRealloc(
			pszStatus , (cchStatus + cch + 1)*sizeof(char) );
    (void)sprintf( pszStatus+cchStatus , "%-*.*s" ,
		   cch , cch , pszTitleLabel );
    cchStatus += cch;
    cchLine += cch;
#endif/*ALWAYSLABEL*/

    if ((vchField = findfield( vchData , SZTITLETAG , 0 , &cchField ))) {
      (void)memset( (void*)afchTitleSub ,
		    (int)'0' , sizeof(afchTitleSub) - 1 );
      afchTitleSub[sizeof(afchTitleSub) - 1] = NUL;
      for (pch=sdachTitleSub ; *pch ; pch++)
	if (vchSub = findsubfield(
				vchField , cchField , *pch , 0 , &cchSub )) {
	  /* if not first subfield found... */
	  if (atoi( afchTitleSub )) {
	    cch = nchIndent;
	    pszStatus = (char*)CwtRealloc(
			pszStatus , (cchStatus + cch + 1)*sizeof(char) );
	    (void)memset( (void*)(pszStatus+cchStatus) ,
			  (int)' ' , (size_t)cch );
	    cchStatus += cch;
	    cchLine += cch;
#ifndef  ALWAYSLABEL
	  } else {
	    cch = nchIndent;
	    pszStatus = (char*)CwtRealloc(
			pszStatus , (cchStatus + cch + 1)*sizeof(char) );
	    (void)sprintf( pszStatus+cchStatus , "%-*.*s" ,
			   cch , cch , pszTitleLabel );
	    cchStatus += cch;
	    cchLine += cch;
#endif/*ALWAYSLABEL*/
	  }
	  if (fNonAscii && nchLine <= 0) {
	    cch = cchSub + 1;		/* space for string + NL */
	    pszStatus = (char*)CwtRealloc(
			pszStatus , (cchStatus + cch + 1)*sizeof(char) );
	    (void)memcpy( (void*)(pszStatus+cchStatus) ,
			  (void*)vchSub , (size_t)cchSub );
	    cchStatus += cchSub;
	    pszStatus[cchStatus] = '\n';
	    cchStatus += 1;
	    afchTitleSub[pch-sdachTitleSub] = '1';
	  } else {
	    (void)appendwrapped(
			&pszStatus , &cchStatus , &cchLine , vchSub , cchSub ,
			fNonAscii , nchLine , nchIndent , TRUE );
	    afchTitleSub[pch-sdachTitleSub] = '1';
	  }
	}
    }
#ifdef  ALWAYSLABEL
    /* if none of the subfields were found... */
    if (!atoi( afchTitleSub )) {
      cch = 1;
      pszStatus = (char*)CwtRealloc(
			pszStatus , (cchStatus + cch + 1)*sizeof(char) );
      pszStatus[cchStatus] = '\n';
      cchStatus += 1;
      cchLine = 0;
    }
#endif/*ALWAYSLABEL*/
  }

/* Get summary holdings field */
  vchFieldSH = findfield( vchData , SZSUMHOLDTAG_GCAT , 0 , &cchFieldSH );
  
/* Find and sort status fields */
  for ( ;
       vchField = findfield( vchData , SZSTATUSTAG , cvchField , &cchField ) ;
       cvchField++ ) {
/*???checks*/
    vvchField = (char**)CwtRealloc(
				vvchField , (cvchField + 1)*sizeof(char*) );
    vcchField = (int*)CwtRealloc(
				vcchField , (cvchField + 1)*sizeof(int) );
    vvchField[cvchField] = vchField;
    vcchField[cvchField] = cchField;
  }
#ifdef   GCATGROUPLOC
  groupbyloc( vvchField , vcchField , cvchField );
#endif /*GCATGROUPLOC*/
#ifdef  GCATSORT
  sortstatusx( vvchField , vcchField , cvchField ,
	       gdastsLocationGCAT , NUMBER(gdastsLocationGCAT) );
#endif/*GCATSORT*/
  (void)memset( (void*)aMax , 0 , sizeof(aMax) );
  getmaxes( vvchField , vcchField , cvchField , sdachStatusAlign , aMax );

  maxtotal = 0;
  for (ivchField=0 ; ivchField<NUMBER(aMax) ; ivchField++) {
    maxtotal += aMax[ivchField];
    if (aMax[ivchField])
      switch (sdachStatusAlign[ivchField]) {
#if 00000
      case CHCALLNMBRTAG:
	maxtotal += NUMBER(SZCALLNMBRPREFIX) - 1;
	break;
#endif
      case CHCOPYNMBRTAG:
	maxtotal += NUMBER(SZCOPYNMBRPREFIX) - 1;
	break;
      default:
	maxtotal += NUMBER(SZVOLPARTPREFIX) - 1;
	break;
      }
  }

/* Status section */
  cch = nchIndent;
  pszStatus = (char*)CwtRealloc(
			pszStatus , (cchStatus + cch + 1)*sizeof(char) );
  (void)sprintf( pszStatus+cchStatus , "%-*.*s" ,
		 cch , cch , pszStatusLabel );
  cchStatus += cch;
  cchLine += cch;

  szLoc1[0] = NUL;
  szLoc2[0] = NUL;
  pszLocC = szLoc1;
  pszLocP = szLoc2;
  cchStatusLast = cchStatus;
  if (!vchFieldSH && cvchField ||
      vchFieldSH &&
      anystatustobedisplayed( vvchField , vcchField , cvchField )) {
    if (vchFieldSH) {
      (void)appendwrapped( &pszStatus , &cchStatus , &cchLine ,
			   SZCHECKEDOUT , strlen( SZCHECKEDOUT ) ,
			   fNonAscii , nchLine , nchIndent , TRUE );
      cchStatusLast = cchStatus;
    }

    for (ivchField=0 ; ivchField<cvchField ; ivchField++) {
      vchField = vvchField[ivchField];
      cchField = vcchField[ivchField];
/*???maybe only do if field is present?*/
      if ((!(vchSub = findsubfield(
		vchField , cchField , CHSUPPRESSTAG , 0 , &cchSub )) ||
	   !issuppressedGCAT( vchSub , cchSub )) &&
	  (!vchFieldSH ||
		statustobedisplayed( vchField , cchField , (char*)NULL ))) {
       /* Location info */
	if (vchSub = findsubfield(
			vchField , cchField , CHLOCATIONTAG , 0 , &cchSub )) {
	  if (cchSub > CCHMAXLOCCODE)
	    cchSub = CCHMAXLOCCODE;	/* ok since setting again below */
	  (void)memcpy( (void*)pszLocC , (void*)vchSub , (size_t)cchSub );
	  pszLocC[cchSub] = NUL;
	  (void)StrStrip( pszLocC , SZWHITE );
	  if (!(vchSub = keytolock(
		gdastsLocationGCAT , NUMBER(gdastsLocationGCAT) , pszLocC )))
	    vchSub = pszLocC;
	  cchSub = strlen( vchSub );

	  /* if no summary holding field case or
	     current and previous locations are different... */
	  if (strcmp( pszLocC , pszLocP )) {
	    if (vchFieldSH || cchStatusLast != cchStatus)
	      cch = nchIndent;
	    else
	      cch = 0;
	    pszStatus = (char*)CwtRealloc(
			pszStatus , (cchStatus + cch + 1)*sizeof(char) );
	    (void)memset( (void*)(pszStatus+cchStatus) ,
			  (int)' ' , (size_t)cch );
	    cchStatus += cch;
	    (void)appendwrapped(
		&pszStatus , &cchStatus , &cchLine , vchSub , cchSub ,
		fNonAscii , nchLine , nchIndent , TRUE );

	    /* must be swapped so that they point to different memory areas */
	    pszLocT = pszLocC;
	    pszLocC = pszLocP;
	    pszLocP = pszLocT;
	  }
	}

       /* Status info */
	cch = nchIndent + CCHSTATUSINDENT2;
	pszStatus = (char*)CwtRealloc(
			pszStatus , (cchStatus + cch + 1)*sizeof(char) );
	(void)memset( (void*)(pszStatus+cchStatus) ,
		        (int)' ' , (size_t)cch );
        cchStatus += cch;
	cchLine += cch;

	sofar = 0;
	fDueDate = FALSE;
	chLast = '\0';		/* can't occur in data */
	for (pch=SZSTATUSDISPLAYTAGS_GCAT ; *pch ; pch++) {
#if 0
	  for (iSub=0 ;
	       vchSub = findsubfield(
			vchField , cchField , *pch , iSub , &cchSub ) ;
	       iSub++) {
#endif
	  if (chLast == *pch)
	    iSub++;
	  else
	    iSub = 0;
	  chLast = *pch;
	  if (vchSub = findsubfield(
			vchField , cchField , *pch , iSub , &cchSub )) {
	    /* if have gone past the alignment subfields and
	       not indented far enough... */
	    if (!strchr( sdachStatusAlign , *pch ) && sofar < maxtotal) {
	      cch = maxtotal - sofar;
	      pszStatus = (char*)CwtRealloc(
			pszStatus , (cchStatus + cch + 1)*sizeof(char) );
	      (void)memset( (void*)(pszStatus+cchStatus) ,
			    (int)' ' , (size_t)cch );
	      cchStatus += cch;
	      cchLine += cch;
	      sofar += cch;
	    }
	    switch (*pch) {
	    case 'd':
#if 00000
	    case 'e':
	    case 'f':
	    case 'g':
	    case 'h':
	    case 'i':
#endif
	      strPrefix = SZVOLPARTPREFIX;
	      break;
#if 00000
	    case 'a':	/* call number */
	      strPrefix = SZCALLNMBRPREFIX;
	      break;
#endif
	    case 'u':	/* due date */
	      /* if actually contains a date... */
	      if (isdateset( vchSub , cchSub )) {
		strPrefix = SZDUEDATEPREFIX;
		fDueDate = TRUE;
	      } else {
		vchSub = (char*)NULL;
		cchSub = 0;
	      }
	      break;
#if 00000
	    case '9':	/* suppression */
	      (void)memcpy( (void*)szCode , (void*)vchSub , (size_t)cchSub );
	      szCode[cchSub] = NUL;
	      (void)StrStrip( szCode , SZWHITE );
	      if (!(vchSub = keytolock( gdastsSuppressionGCAT ,
				        NUMBER(gdastsSuppressionGCAT) ,
				        szCode )))
		vchSub = szCode;
	      cchSub = strlen( vchSub );
	      if (vchSub && *vchSub) {
		strPrefix = SZOTHERPREFIX;
	      }
	      break;
#endif
	    case 's':	/* status */
	      (void)memcpy( (void*)szCode , (void*)vchSub , (size_t)cchSub );
	      szCode[cchSub] = NUL;
	      (void)StrStrip( szCode , SZWHITE );
	      /* special case for - */
	      if (*szCode != '-' || !fDueDate) {
		if (!(vchSub = keytolock(
			gdastsStatusGCAT , NUMBER(gdastsStatusGCAT) ,
			szCode )))
		  vchSub = szCode;
		cchSub = strlen( vchSub );
		if (vchSub && *vchSub) {
		  strPrefix = SZOTHERPREFIX;
		}
	      } else {
		vchSub = (char*)NULL;
		cchSub = 0;
	      }
	      break;
	    case 't':	/* type of reserve */
	      (void)memcpy( (void*)szCode , (void*)vchSub , (size_t)cchSub );
	      szCode[cchSub] = NUL;
	      (void)StrStrip( szCode , SZWHITE );
	      if (!(vchSub = keytolock(
		gdastsReserveGCAT , NUMBER(gdastsReserveGCAT) , szCode )))
		vchSub = szCode;
	      cchSub = strlen( vchSub );
	      if (vchSub && *vchSub) {
		strPrefix = SZOTHERPREFIX;
	      }
	      break;
	    case 'q':	/* copy number */
	      strPrefix = SZCOPYNMBRPREFIX;
	      break;
	    case 'w':	/* number of holds */
	      strPrefix = SZHOLDSPREFIX;
	      break;
	    case 'n':	/* public note */
	    default:
	      strPrefix = SZOTHERPREFIX;
	      break;
	    }
	    if (vchSub && *vchSub) {
	      /* if wrapping and have actually written part of line and
		 it would get too long... */
	      if (nchLine > 0 && sofar &&
		  cchLine + strlen( strPrefix ) + cchSub > nchLine) {
		cch = nchIndent + CCHSTATUSINDENT2 + 1;
		pszStatus = (char*)CwtRealloc(
			pszStatus , (cchStatus + cch + 1)*sizeof(char) );
		pszStatus[cchStatus] = '\n';
		cchStatus += 1;
		(void)memset( (void*)(pszStatus+cchStatus) ,
			      (int)' ' , (size_t)(cch - 1) );
		cchStatus += cch - 1;
		cchLine = cch - 1;
		sofar = 0;
	      }

	      cch = strlen( strPrefix );
	      pszStatus = (char*)CwtRealloc(
			pszStatus , (cchStatus + cch + 1)*sizeof(char) );
	      (void)memcpy( (void*)(pszStatus+cchStatus) ,
			    (void*)strPrefix , (size_t)cch );
	      cchStatus += cch;
	      cchLine += cch;
	      sofar += cch;

	      cch = cchSub;
	      pszStatus = (char*)CwtRealloc(
			pszStatus , (cchStatus + cch + 1)*sizeof(char) );
	      (void)memcpy( (void*)(pszStatus+cchStatus) ,
			    (void*)vchSub , (size_t)cchSub );
	      cchStatus += cch;
	      cchLine += cch;
	      sofar += cch;

	      if (pchTmp = strchr( sdachStatusAlign , *pch )) {
		max = aMax[pchTmp - sdachStatusAlign + iSub];
		if (max - cchSub > 0) {
		  cch = max - cchSub;
		  pszStatus = (char*)CwtRealloc(
			pszStatus , (cchStatus + cch + 1)*sizeof(char) );
		  (void)memset( (void*)(pszStatus+cchStatus) ,
			        (int)' ' , (size_t)cch );
		  cchStatus += cch;
		  cchLine += cch;
		  sofar += cch;
		}
	      }
	    }
	  }
	}
	cch = 1;
	pszStatus = (char*)CwtRealloc(
			pszStatus , (cchStatus + cch + 1)*sizeof(char) );
	pszStatus[cchStatus] = '\n';
	cchStatus += cch;
	cchLine = 0;
      }
    }
/*???next two yield same output*/
  } else if (vchFieldSH) {
    cch = strlen( SZCHECKSHELF_GCAT );
    (void)appendwrapped(
		&pszStatus , &cchStatus , &cchLine , SZCHECKSHELF_GCAT , cch ,
		fNonAscii , nchLine , nchIndent , TRUE );
  } else {		/* (!vchFieldSH && !cvchField) */
    cch = strlen( SZLOOKUPFAILED_GCAT );
    (void)appendwrapped(
		&pszStatus , &cchStatus , &cchLine ,
		SZLOOKUPFAILED_GCAT , cch ,
		fNonAscii , nchLine , nchIndent , TRUE );
  }

#ifdef  ALWAYSLABEL
  /* if none of the subfields were found... */
  if (cchStatusLast == cchStatus) {
    cch = 1;
    pszStatus = (char*)CwtRealloc(
			pszStatus , (cchStatus + cch + 1)*sizeof(char) );
    pszStatus[cchStatus] = '\n';
    cchStatus += cch;
  }
#endif/*ALWAYSLABEL*/
  pszStatus[cchStatus] = NUL;

  CwtFree( vvchField );
  CwtFree( vcchField );
  *pcchStatus = cchStatus;

  return pszStatus;
}


/*----------------------------------------------------------------------------
 *							Public Routines
 *--------------------------------------------------------------------------*/

int  IscInit  (char*  pszHost)
{
  int		s;

  s = connecttoservice( pszHost , (char*)NULL , USPORT_DEFAULT );
#if 000	/*???use when service name is known */
  /* if cannot connect by service name... */
  if (0 > (s = connecttoservice( pszHost , SZSERVICE_DEFAULT , 0 )))
    s = connecttoservice( pszHost , (char*)NULL , USPORT_DEFAULT );
#endif

  return s;
}


void  IscTerm  (int  s)
{
  (void)shutdown( s , 2 );
  (void)close( s );
}


#ifdef  TIMEOUT
void  IscSetParams  (int  Timeout)
{
  gTimeout = Timeout;
}


void  IscGetParams  (int*  pTimeout)
{
  *pTimeout = gTimeout;
}
#endif/*TIMEOUT*/


int  IscQueryRaw  (int		s		,
		   char		chCode		,
		   char*	pszRequest	,
		   char*	pvchData[]	,
		   int*		pcchData	)
{
  char		*vchData;
  int		 cchData;
/* Variable */
  char		*psz		= (char*)NULL;
  int		 cch		= 0;
  int		fContinue	= TRUE;

  cch = CCHMESSAGELENGTH + CCHMESSAGECODE + CCHREQUESTOPCODE +
	strlen( pszRequest );
  fContinue = fContinue &&
    (psz = (char*)CwtMalloc( (cch + 1)*sizeof(char) ));

  (void)sprintf( psz , "%0*d%*d%c%s" ,
		 CCHMESSAGELENGTH , cch ,
		 CCHMESSAGECODE , MCODE_REQUEST ,
		 chCode , pszRequest );

  fContinue = fContinue &&
    writedata( s , psz , cch );

  fContinue = fContinue &&
    getmessage( s , &vchData , &cchData );

  CwtFree( psz );

  if (!fContinue) {
    CwtFree( vchData );
    vchData	= (char*)NULL;
    cchData	= 0;
  }

  *pvchData	= vchData;
  *pcchData	= cchData;
  return fContinue;
}


int  IscQuery  (int	s		,
		char	chCode		,
		char*	pszRequest	,
		int*	pfError		,
		int*	pcRecord	,
		int*	pcRemain	,
		int*	pError		,
		char*	pvchData[]	,
		int*	pcchData	)
{
/* Variable */
  char		*psz		= (char*)NULL;
  int		 cch		= 0;
  int		fContinue	= TRUE;

  cch = CCHMESSAGELENGTH + CCHMESSAGECODE + CCHREQUESTOPCODE +
	strlen( pszRequest );
  fContinue = fContinue &&
    (psz = (char*)CwtMalloc( (cch + 1)*sizeof(char) ));

  (void)sprintf( psz , "%0*d%*d%c%s" ,
		 CCHMESSAGELENGTH , cch ,
		 CCHMESSAGECODE , MCODE_REQUEST ,
		 chCode , pszRequest );

  fContinue = fContinue &&
    writedata( s , psz , cch );
  CwtFree( psz );

  fContinue = fContinue &&
    getmessageinparts(
	s , pfError , pcRecord , pcRemain , pError , pvchData , pcchData );

  return fContinue;
}


int  IscGetHoldingStatus  (int		s		,
			   char*	pszMode		,
			   char*	pszRequest	,
			   int		fTitle		,
			   int		fNonAscii	,
			   int		nchLine		,
			   int		fShortLabels	,
			   int		nchIndent	,
			   char**	ppszStatus	,
			   int*		pcchStatus	)
{
  int		fError;
  int		cRecord, cRemain, Error;
  int		cchRequest;
  char		chReqCode;
  char		*vchData;
  int		 cchData;
  char		szBibCode[CCHBIBCODE0+1+1];

  if (!pszRequest)
    return FALSE;

  cchRequest = strlen( pszRequest );

  if (!cchRequest)
    return FALSE;

  if        (cchRequest == CCHBIBCODE && pszRequest[0] == 'b') {
    chReqCode = CHBIBREQUESTOPCODE;
  } else if (cchRequest == CCHBIBCODE0) {
    szBibCode[0] = 'b';
    (void)strcpy( szBibCode+1 , pszRequest );
    pszRequest = szBibCode;
    cchRequest += 1;
    chReqCode = CHBIBREQUESTOPCODE;
#ifdef  WORKAROUND	/* request includes the request code */
  } else if (cchRequest == CCHBARCODE+1) {
#else /*WORKAROUND*/
  } else if (cchRequest == CCHBARCODE) {
#endif/*WORKAROUND*/
    chReqCode = CHBARREQUESTOPCODE;
  } else if (cchRequest == CCHISBNCODE) {
    chReqCode = CHISBNREQUESTOPCODE;
  } else if (cchRequest == CCHISSNCODE) {
    chReqCode = CHISSNREQUESTOPCODE;
  } else {
    return FALSE;
  }

#ifdef  WORKAROUND
  if (!IscQuery( s , chReqCode ,
		 chReqCode == CHBARREQUESTOPCODE ? pszRequest+1 : pszRequest ,
#else /*WORKAROUND*/
  if (!IscQuery( s , chReqCode , pszRequest ,
#endif/*WORKAROUND*/
		 &fError , &cRecord , &cRemain , &Error ,
		 &vchData , &cchData ))
    return FALSE;

  if (pszMode && !strcmp( pszMode , "GCAT" ))
    *ppszStatus = extractformatstatusGCAT(
	vchData , cchData ,
	fTitle , TRUE , fNonAscii , nchLine , fShortLabels , nchIndent ,
	pcchStatus );
  else
    *ppszStatus = extractformatstatus(
	vchData , cchData ,
	fTitle , TRUE , fNonAscii , nchLine , fShortLabels , nchIndent ,
	pcchStatus );

  CwtFree( vchData );
  return TRUE;
}


int  IscDumpMarcRecord  (FILE*	pfileOut	,
			 char	vchData[]	,
			 int	cchData		)
{
  char		*pchTag;
  char		*vchField;
  int		 cchField;
  register  int		ichField;
/* Variable */
  int		fRetVal		= TRUE;

  for (ichField=0 ;
       fRetVal &&
       (vchField=findfield( vchData , (char*)NULL , ichField , &cchField )) ;
       ichField++) {
    fRetVal = fRetVal &&
      (pchTag = getdirentry( vchData , ichField ));
    if (fRetVal)
      (void)fprintf( pfileOut , "%*.*s " , CCHMARCTAG , CCHMARCTAG , pchTag );
    fRetVal = fRetVal &&
      dumpfield( pfileOut , vchField , cchField );
    (void)fprintf( pfileOut , "\n" );
  }

  return fRetVal;
}


/*--isc.c-------------------------------------------------------------------*/
