/*++isc.c+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/****************************************************************************
 * 									    *
 *                 WILCO(tm) Information Retrieval Software		    *
 *        (Washington Information Looker-upper - Character Oriented)	    *
 * 									    *
 *               Copyright 1993-1994 University of Washington		    *
 *           Wilco is a trademark of the University of Washington	    *
 * 									    *
 * Permission to use, copy, modify, and distribute Wilco(tm) software	    *
 * and its documentation for any purpose and without fee is hereby	    *
 * granted, provided that the above copyright notice appears in all	    *
 * copies and that both the above copyright notice and this permission	    *
 * notice appear in supporting documentation, and that the name of the	    *
 * University of Washington not be used in advertising or publicity	    *
 * pertaining to distribution of the software without specific, written	    *
 * prior permission.  This software is made available "as is", and	    *
 * 									    *
 * THE UNIVERSITY OF WASHINGTON DISCLAIMS ALL WARRANTIES, EXPRESS OR	    *
 * IMPLIED, WITH REGARD TO THIS SOFTWARE, INCLUDING WITHOUT LIMITATION	    *
 * ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR   *
 * PURPOSE, AND IN NO EVENT SHALL THE UNIVERSITY OF WASHINGTON BE LIABLE    *
 * FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES	    *
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN    *
 * ACTION OF CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY,	    *
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS	    *
 * SOFTWARE.								    *
 * 									    *
 * For further information please contact:				    *
 * 									    *
 * Scott Heyano								    *
 * Information Systems							    *
 * Computing and Communications						    *
 * University of Washington						    *
 * 4545 15th, 3rd Floor JE-41						    *
 * Seattle, Washington, 98195, USA					    *
 * Internet: slh@cac.washington.edu					    *
 * 									    *
 * This file is not specifically part of ``Wilco'',			    *
 * but is used by it and is covered by the above copyright notice and	    *
 * usage restrictions.							    *
 ****************************************************************************/
/*
 * File	     : isc.h
 * Related   : isc.c
 * Author    : slh (Scott Heyano, slh@cac.washington.edu, slh@nwnet.net)
 * Copyright : (see above)
 *
 * III Server Client code public Header file.
 */
/* Log:
 *  0.40  02/16/94  slh
 *         IscGetHoldingStatus(): pszMode param added
 *  0.30  10/27/93  slh
 *         Isc[GS]etParams(): added
 *  0.20  09/07/93  slh
 *         IscGetHoldingStatus(): interface change
 *  0.11  08/23/93  slh
 *         IscDumpMarcRecord(): added
 *  0.10   7/ 2/93  slh
 *         IscGetHoldingStatus(): interface change
 *  0.00   6/22/93 -  6/23/93  slh
 *         create
 */
/* Synopsis:
 * #include  "isc.h"
 *
 * cc ... isc.o malarm.o ...
 */


/*+++
 * IscInit - initiates a session with a server
 * PARAMETERS:
 * pszHost	name of host to which to connect
 * RETURN VALUES:
 * File descriptor of socket if successful, -1 otherwise.
 */
int  IscInit  (char*  pszHost);
/*+++
 * IscTerm - terminates a session with a server
 * PARAMETERS:
 * s		file descriptor of socket (from IscInit())
 */
void  IscTerm  (int  s);

/*+++
 * IscSetParams - set the values of module parameters
 * IscGetParams - get the current values of module parameters
 * PARAMETERS:
 * Timeout	time (in secs) for connect/read time-outs
 * pTimeout	time (in secs) for connect/read time-outs
 */
void  IscSetParams  (int  Timeout);
void  IscGetParams  (int*  pTimeout);

/*+++
 * IscQueryRaw - submit a query and retrieve results
 * PARAMETERS:
 * s		socket connected to server (as returned from IscInit())
 * chCode	request code
 * pszRequest	request string
 * pvchData	result data (return)
 * pcchData	count of bytes in pvchData (return)
 * RETURN VALUES:
 * TRUE if successful, FALSE otherwise.a
 * NOTES:
 * Result data has a NUL byte appended to the end,
 * which is not included in the byte count.
 * Caller is responsible for freeing data returned in pvchData.
 */
int  IscQueryRaw  (int		s		,
		   char		chCode		,
		   char*	pszRequest	,
		   char*	pvchData[]	,
		   int*		pcchData	);
/*+++
 * IscQuery - submit a query and retrieve results
 * Like IscQuery(), except result data is broke into parts.
 * PARAMETERS:
 * s		socket connected to server (as returned from IscInit())
 * chCode	request code
 * pszRequest	request string
 * pcRecord	count of records in data, if no error (return)
 * pcRemain	count of records remaining to be sent, if no error (return)
 * pError	error code, if any (return)
 * pvchData	result data (return)
 * pcchData	count of bytes in pvchData (return)
 * RETURN VALUES:
 * TRUE if successful, FALSE otherwise.a
 * NOTES:
 * Result data has a NUL byte appended to the end,
 * which is not included in the byte count.
 * Caller is responsible for freeing data returned in pvchData.
 */
int  IscQuery  (int	s		,
		char	chCode		,
		char*	pszRequest	,
		int*	pfError		,
		int*	pcRecord	,
		int*	pcRemain	,
		int*	pError		,
		char*	pvchData[]	,
		int*	pcchData	);

/*+++
 * IscGetHoldingStatus - get formatted holding status display string
 * PARAMETERS:
 * s		fd of (connected) socket to use
 * pszMode	mode of operation (currently NULL, "LCAT" or "GCAT")
 * pszRequest	requested item id
 * fTitle	include title section in output?
 * fNonAscii	include non-ASCII chars in output?
 * nchLine	width of output lines in chars (0 => infinite width)
 *		[not currently implemented]
 * fShortLabels	use short labels?
 * nchIndent	width of primary indent in chars (-1 => use label width)
 * ppszStatus	formatted status string (return)
 * pcchStatus	length of status string (return)
 * RETURN VALUES:
 * Returns TRUE if successful, FALSE otherwise.
 * NOTES:
 * Caller is responsible for free'ing data returned in string *ppszStatus.
 */
int  IscGetHoldingStatus  (int		s		,
			   char*	pszMode		,
			   char*	pszRequest	,
			   int		fTitle		,
			   int		fNonAscii	,
			   int		nchLine		,
			   int		fShortLabels	,
			   int		nchIndent	,
			   char**	ppszStatus	,
			   int*		pcchStatus	);

int  IscDumpMarcRecord  (FILE*	pfileOut	,
			 char	vchData[]	,
			 int	cchData		);


/*--isc.c-------------------------------------------------------------------*/
