/***************************************************************************
 *									   *
 *              WILLOW(tm) Information Retrieval Software		   *
 *     (Washington Information Looker-upper Layered Over Windows)	   *
 * 									   *
 *             Copyright 1992 University of Washington			   *
 *       Willow is a trademark of the University of Washington		   *
 * 									   *
 * Permission to use, copy, modify, and distribute Willow(tm) software	   *
 * and its documentation for any purpose and without fee is hereby	   *
 * granted, provided that the above copyright notice appears in all	   *
 * copies and that both the above copyright notice and this permission	   *
 * notice appear in supporting documentation, and that the name of the	   *
 * University of Washington not be used in advertising or publicity	   *
 * pertaining to distribution of the software without specific, written	   *
 * prior permission.  This software is made available "as is", and	   *
 * 									   *
 * THE UNIVERSITY OF WASHINGTON DISCLAIMS ALL WARRANTIES, EXPRESS OR	   *
 * IMPLIED, WITH REGARD TO THIS SOFTWARE, INCLUDING WITHOUT LIMITATION	   *
 * ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR  *
 * PURPOSE, AND IN NO EVENT SHALL THE UNIVERSITY OF WASHINGTON BE LIABLE   *
 * FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES	   *
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN   *
 * ACTION OF CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY,	   *
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS	   *
 * SOFTWARE.								   *
 * 									   *
 * For further information please contact:				   *
 * 									   *
 * Matthew Freedman							   *
 * Information Systems							   *
 * Computing and Communications						   *
 * University of Washington						   *
 * 4545 15th, 4th Floor JE-41						   *
 * Seattle, Washington, 98195, USA					   *
 * Internet: mattf@cac.washington.edu					   *
 ***************************************************************************/

/***************************************************************************
 * malarm.c
 * This file contains the routines that are used to implement the multple
 * alarm handler.
 ***************************************************************************/

/*
 * 04/04/94  slh  malarm_handler(): wasn't free'ing removed entries;
 *                  setupsignal() added
 * 02/16/94  slh  malarm_cancel(): always alarm(0) when one found in queue
 * 12/08/93  slh  incorp portability fixes from mit (incl vax path); misc
 * 11/10/93  slh  rm signal handler when queue is empty
 * 10/26/93  slh  fix signal/system call interupt; malarm_cancel(): bugfix
 *    - 10/27/93
 */


#include <stdio.h>
#include <signal.h>
#include <time.h>

#include  <sys/types.h>

static
void  malarm_handler  (int  sig);

typedef struct _al_rec {
  int    id;
  time_t due;
  void   (*handler)();
  struct _al_rec *next;
} al_rec;

/* The alarm queue */
static al_rec *queue = NULL;


/* External routines */
extern void *malloc (size_t);
extern unsigned int alarm (unsigned int);
#ifndef  vax
extern void *memset (void *, int, size_t);
#endif/*!vax*/
extern void free (void *);


static
void  setupsignal  (void  (*sighandler)())
{
#ifdef  POSIX
  struct  sigaction	ssa;
#else /*POSIX*/
#ifdef  SV_INTERRUPT
  struct  sigvec	ssv;
#endif/*SV_INTERRUPT*/
#endif/*POSIX*/

#ifdef  POSIX
  ssa.sa_handler	= (void(*)())sighandler;
  memset( &ssa.sa_mask , 0 , sizeof(sigset_t) );
  ssa.sa_flags	= 0;
  sigaction( SIGALRM , &ssa , (struct sigaction*)NULL );
#else /*POSIX*/
#ifdef  SV_INTERRUPT
#ifdef  vax
  ssv.sv_handler	= (int(*)())sighandler;
  memset( &ssv.sv_mask , 0 , sizeof(int) );
#else /*vax*/
  ssv.sv_handler	= (void(*)())sighandler;
  memset( &ssv.sv_mask , 0 , sizeof(sigset_t) );
#endif/*vax*/
/*???don't think the cond is necessary, but just in case*/
  ssv.sv_flags		= sighandler == SIG_DFL ? 0 : SV_INTERRUPT;
  sigvec( SIGALRM , &ssv , (struct sigvec*)NULL );
#else /*SV_INTERRUPT*/
  signal( SIGALRM , (void(*)())sighandler );
#endif/*SV_INTERRUPT*/
#endif/*POSIX*/
}


int malarm_add(int interval, void (*handler)())
{
  static int cur_id = 0; /* Next alarm id# to give out */
  al_rec *new;           /* New alarm record to be added to queue */
  al_rec *cur;           /* for traversing the list */
  time_t  now;           /* What time is it */
#if 00
#ifdef  POSIX
  struct  sigaction	ssa;
#else /*POSIX*/
#ifdef  SV_INTERRUPT
  struct  sigvec	ssv;
#endif/*SV_INTERRUPT*/
#endif/*POSIX*/
#endif

  if (interval < 1)
    return(0);

  new = (al_rec *) malloc(sizeof(al_rec));
  new->id = ++cur_id;
  new->handler = handler;
  now = time(NULL);
  new->due = now + interval;
  new->next = NULL;

  if (!queue)
    queue = new;
  else {    /* There is at least one alarm pending already. */
    /* Cancel pending alarm */
    alarm(0);
    
    /* First check and see if new one should be first on the queue */
    if (new->due < queue->due) {
      new->next = queue;
      queue = new;
    }
    else {
      /* Find proper place in queue for new request */
      for (cur = queue; cur; cur = cur->next) {
	if (!cur->next) {
	  /* New one goes on the end */
	  cur->next = new;
	  break;
	}
	if (new->due < cur->next->due) {
	  new->next = cur->next;
	  cur->next = new;
	  break;
	}
      }
    }
  }
  /* Now actually set the alarm */
  if (queue->due > now) {
    setupsignal( malarm_handler );
#if 00
#ifdef  POSIX
    ssa.sa_handler	= (void(*)())malarm_handler;
    memset( &ssa.sa_mask , 0 , sizeof(sigset_t) );
    ssa.sa_flags	= 0;
    sigaction( SIGALRM , &ssa , (struct sigaction*)NULL );
#else /*POSIX*/
#ifdef  SV_INTERRUPT
#ifdef  vax
    ssv.sv_handler	= (int(*)())malarm_handler;
    memset( &ssv.sv_mask , 0 , sizeof(int) );
#else /*vax*/
    ssv.sv_handler	= (void(*)())malarm_handler;
    memset( &ssv.sv_mask , 0 , sizeof(sigset_t) );
#endif/*vax*/
    ssv.sv_flags	= SV_INTERRUPT;
    sigvec( SIGALRM , &ssv , (struct sigvec*)NULL );
#else /*SV_INTERRUPT*/
    signal(SIGALRM, (void(*)())malarm_handler);
#endif/*SV_INTERRUPT*/
#endif/*POSIX*/
#endif
    alarm(queue->due - now);
  }
  else
    malarm_handler( 0 );
  return(new->id);
}


/* This routine is called when an alarm goes off. We have to 
   handle every item on the queue that has a value of 0. To avoid
   complications of the alarm handlers we are calling adding new alarms, 
   don't call any until we have isolated the ones that are ready. */
static
void  malarm_handler  (int  sig)
{
  al_rec *ready_queue, *prev;
  time_t now;
#if 0
#ifdef  POSIX
  struct  sigaction	ssa;
#else /*POSIX*/
#ifdef  SV_INTERRUPT
  struct  sigvec	ssv;
#endif/*SV_INTERRUPT*/
#endif/*POSIX*/
#endif

  now = time(NULL);
  ready_queue = queue;
  queue->due = now;
  while (queue) {
    if (queue->due <= now) {
      prev = queue;
      queue = queue->next;
    }
    else {
      prev->next = 0;
      break;
    }
  }
  /* Must set the first alarm before calling handlers */
  if (queue) {
    alarm(queue->due - now);
  } else {
    setupsignal( SIG_DFL );
#if 00
#ifdef  POSIX
    ssa.sa_handler	= (void(*)())SIG_DFL;
    memset( &ssa.sa_mask , 0 , sizeof(sigset_t) );
    ssa.sa_flags	= 0;
    sigaction( SIGALRM , &ssa , (struct sigaction*)NULL );
#else /*POSIX*/
#ifdef  SV_INTERRUPT
#ifdef  vax
    ssv.sv_handler	= (int(*)())SIG_DFL;
    memset( &ssv.sv_mask , 0 , sizeof(int) );
#else /*vax*/
    ssv.sv_handler	= (void(*)())SIG_DFL;
    memset( &ssv.sv_mask , 0 , sizeof(sigset_t) );
#endif/*vax*/
    ssv.sv_flags	= 0;
    sigvec( SIGALRM , &ssv , (struct sigvec*)NULL );
#else /*SV_INTERRUPT*/
    signal( SIGALRM , (void(*)())SIG_DFL );
#endif/*SV_INTERRUPT*/
#endif/*POSIX*/
#endif
  }
  while (ready_queue) {
    if (ready_queue->handler)
      ready_queue->handler();
    prev = ready_queue;
    ready_queue = ready_queue->next;
    free( prev );
  }
}


/* Cancel an alarm */
void malarm_cancel(int id)
{
  al_rec *cur, *prev;
  time_t  now;
#if 00
#ifdef  POSIX
  struct  sigaction	ssa;
#else /*POSIX*/
#ifdef  SV_INTERRUPT
  struct  sigvec	ssv;
#endif/*SV_INTERRUPT*/
#endif/*POSIX*/
#endif

  now = time(NULL);
  prev = cur = queue;
  while (cur) {
    if (cur->id == id) {
      if (cur == queue) {
	queue = cur->next;
	(void)alarm( 0 );	/* for when no farther alarms will be set */
	if (queue)
	  if (queue->due > now)
	    alarm(queue->due - now);
	  else
	    malarm_handler( 0 );
      }
      else
	prev->next = cur->next;
      free(cur);
      if (!queue) {
	setupsignal( SIG_DFL );
#if 00
#ifdef  POSIX
	ssa.sa_handler	= (void(*)())SIG_DFL;
	memset( &ssa.sa_mask , 0 , sizeof(sigset_t) );
	ssa.sa_flags	= 0;
	sigaction( SIGALRM , &ssa , (struct sigaction*)NULL );
#else /*POSIX*/
#ifdef  SV_INTERRUPT
#ifdef  vax
	ssv.sv_handler	= (int(*)())SIG_DFL;
	memset( &ssv.sv_mask , 0 , sizeof(int) );
#else /*vax*/
	ssv.sv_handler	= (void(*)())SIG_DFL;
	memset( &ssv.sv_mask , 0 , sizeof(sigset_t) );
#endif/*vax*/
	ssv.sv_flags	= 0;
	sigvec( SIGALRM , &ssv , (struct sigvec*)NULL );
#else /*SV_INTERRUPT*/
	signal( SIGALRM , (void(*)())SIG_DFL );
#endif/*SV_INTERRUPT*/
#endif/*POSIX*/
#endif
      }
      return;
    }
    prev = cur;
    cur = cur->next;
  }
}


void malarm_print(void)
{
  al_rec *cur;
  time_t  now;

  now = time(NULL);
  cur = queue;
  while (cur) {
    fprintf(stdout, "%d ", cur->due - now);
    cur = cur->next;
  }
  fprintf(stdout, "<NULL>\n");
}
