#include <stdio.h>
#ifdef SOLARIS
#include <string.h>
#else /* SOLARIS */
#include <strings.h>
#endif /* SOLARIS */
#include <sys/types.h>
#ifdef   _SEQUENT_
#include  <sys/procstats.h>
#endif /*_SEQUENT_*/
#include <ctype.h>
#include <errno.h>
#ifdef   _SEQUENT_
#include <time.h>
#else  /*_SEQUENT_*/
#include <sys/time.h>
#endif /*_SEQUENT_*/
#include "driver.h"
#include "protocol.h"
#include "host.h"

char logging = FALSE;

FILE *log_file = NULL;

#define USECS_IN_ONE_SEC 1000000

/* A couple extra types */
#define TELNETTED 99
#define DBED      100

/* External routines */
#ifdef vax
extern int getpid();
#else
extern pid_t getpid (void);
#endif
extern int error (char *msg);
extern char *getenv ( char *);


/* Return the difference between two timevals as float in seconds */
float
time_diff(struct timeval end, struct timeval start)
{
    struct timeval diff;

    if (end.tv_usec < start.tv_usec)
    { 
	end.tv_usec += USECS_IN_ONE_SEC; 
	end.tv_sec -= 1; 
    } 
    diff.tv_usec = end.tv_usec - start.tv_usec; 
    if (end.tv_sec < start.tv_sec)
    {
	return(0.0);
    }
    diff.tv_sec = end.tv_sec - start.tv_sec;
    return ((float) diff.tv_sec + (float) diff.tv_usec / USECS_IN_ONE_SEC);
}

/*
 * The following are used to convert times back and forth from 
 * floats, to integer rep of microseconds.
 */

float mstof(int ms)
{
    return((float)(ms / USECS_IN_ONE_SEC) + 
	   ((float)(ms % USECS_IN_ONE_SEC) / (float) USECS_IN_ONE_SEC)); 
}

ftoms(float f)
{
    return((int) f * USECS_IN_ONE_SEC +
	   (int) ((f - (int) f) * USECS_IN_ONE_SEC));
}

log_trans(int type, char *comment)
{
  char file_name[128];
  char time_str[128];
  static struct timeval clock1, clock2;
#ifndef   _SEQUENT_
  struct timezone tzp;
#endif /*!_SEQUENT_*/
  static struct tm *time1, *time2;
  char *trans_name;
  int   mills;
  char  mills_s[16];
  char *display;
  char  del_ch;

  if (!log_file) {
#ifdef   _SEQUENT_
    (void)get_process_stats( &clock1 , PS_SELF , NULL , NULL );
#else  /*_SEQUENT_*/
    gettimeofday(&clock1, &tzp);
#endif /*_SEQUENT_*/
    time1 = localtime(&clock1.tv_sec);
    strftime(time_str, 128, "%b_%d_%H:%M:%S", time1);
    sprintf(file_name, "/usr/tmp/wlog_%s_%d", time_str, getpid());
    log_file = fopen(file_name, "w");
    if (!log_file) {
      error("Could not open log file");
      return;
    }
    display = getenv("DISPLAY");
    fprintf(log_file, "%s\n", display ? display : 
	    "unknown DISPLAY"); 
    clock1.tv_sec = 0;
  }
#ifdef   _SEQUENT_
  (void)get_process_stats( &clock2 , PS_SELF , NULL , NULL );
#else  /*_SEQUENT_*/
  gettimeofday(&clock2, &tzp);
#endif /*_SEQUENT_*/
  time2 = localtime(&clock2.tv_sec);
  switch (type) {
  case CONNECT:
    trans_name = "CNCT";
    del_ch = 'S';
    break;
  case CONNECTED:
    trans_name = "CNCTED";
    del_ch = 'S';
    break;
  case TELNETTED:
    trans_name = "TELNET";
    del_ch = 'S';
    break;
  case DBED:
    trans_name = "DB";
    del_ch = 'S';
    break;
  case ERROR:
    trans_name = "ERR";
    del_ch = 'S';
    break;
  case SEARCH:
    trans_name = "SEARCH";
    del_ch = 'U';
    break;
  case STATS:
    trans_name = "STATS";
    del_ch = 'S';
    break;
  case TITLE_REQ: 
    trans_name = "TITLERQ";
    del_ch = 'U';
    break;
  case TITLE:
    trans_name = "TITLE";
    del_ch = 'S';
    break;
  case RECORD_REQ:
    trans_name = "FULLRQ";
    del_ch = 'U';
    break;
  case RECORD:
    trans_name = "FULL";
    del_ch = 'S';
    break;
  case DISCONNECT:
    trans_name = "DISC";
    del_ch = 'U';
    break;
  case CANCEL:
    trans_name = "CNCL";
    del_ch = 'U';
    break;
  case TOGGLE_ECHO:
  case CHEAT:
    return;
  default:
    trans_name = "UNKNOWN";
  }
  strftime(time_str, 128, "%H:%M:%S", time2);

  /* format the milliseconds part of the string */
  mills = clock2.tv_usec / 1000;
  mills_s[0] = mills / 100 + '0';
  mills %= 100;
  mills_s[1] = mills / 10 + '0';
  mills %= 10;
  mills_s[2] = mills + '0';
  mills_s[3] = 0;

  fprintf(log_file, "%s.%s\t%s\t%c\t%7.3f\t%s\n", time_str, mills_s, 
	  trans_name, del_ch, clock1.tv_sec ? time_diff(clock2,clock1): 0.0, 
	  comment);
  clock1 = clock2;
}

#ifdef vax
/*
 * strftime.c
 * Original Author:	G. Haley
 *
 * Taken from GNU lib because VAXen aren't POSIX.
 *
 * Places characters into the array pointed to by s as controlled by the string
 * pointed to by format. If the total number of resulting characters including
 * the terminating null character is not more than maxsize, returns the number
 * of characters placed into the array pointed to by s (not including the
 * terminating null character); otherwise zero is returned and the contents of
 * the array indeterminate.
 */
static int dname_len[7] =
{6, 6, 7, 9, 8, 6, 8};
static char *dname[7] =
{"Sunday", "Monday", "Tuesday", "Wednesday",
 "Thursday", "Friday", "Saturday"};
static int mname_len[12] =
{7, 8, 5, 5, 3, 4, 4, 6, 9, 7, 8, 8};
static char *mname[12] =
{"January", "February", "March", "April",
 "May", "June", "July", "August", "September", "October", "November",
 "December"};
int strftime(s, maxsize, format, tim_p)
	char *s;
	int maxsize;
	char *format;
	struct tm *tim_p;
{
  int count = 0;
  int i;
  for (;;)
    {
      while (*format && *format != '%')
	{
	  if (count < maxsize - 1)
	    s[count++] = *format++;
	  else
	    return 0;
	}
      if (*format == '\0')
	break;
      format++;
      switch (*format)
	{
	case 'a':
	  for (i = 0; i < 3; i++)
	    {
	      if (count < maxsize - 1)
		s[count++] =
		  dname[tim_p->tm_wday][i];
	      else
		return 0;
	    }
	  break;
	case 'A':
	  for (i = 0; i < dname_len[tim_p->tm_wday]; i++)
	    {
	      if (count < maxsize - 1)
		s[count++] =
		  dname[tim_p->tm_wday][i];
	      else
		return 0;
	    }
	  break;
	case 'b':
	  for (i = 0; i < 3; i++)
	    {
	      if (count < maxsize - 1)
		s[count++] =
		  mname[tim_p->tm_mon][i];
	      else
		return 0;
	    }
	  break;
	case 'B':
	  for (i = 0; i < mname_len[tim_p->tm_mon]; i++)
	    {
	      if (count < maxsize - 1)
		s[count++] =
		  mname[tim_p->tm_mon][i];
	      else
		return 0;
	    }
	  break;
	case 'c':
	  if (count < maxsize - 24)
	    {
	      for (i = 0; i < 3; i++)
		s[count++] =
		  dname[tim_p->tm_wday][i];
	      s[count++] = ' ';
	      for (i = 0; i < 3; i++)
		s[count++] =
		  mname[tim_p->tm_mon][i];
	      sprintf (&s[count],
		       " %.2d %2.2d:%2.2d:%2.2d %.4d",
		       tim_p->tm_mday, tim_p->tm_hour,
		       tim_p->tm_min,
		       tim_p->tm_sec, 1900 +
		       tim_p->tm_year);
	      count += 17;
	    }
	  else
	    return 0;
	  break;
	case 'd':
	  if (count < maxsize - 2)
	    {
	      sprintf (&s[count], "%.2d",
		       tim_p->tm_mday);
	      count += 2;
	    }
	  else
	    return 0;
	  break;
	case 'H':
	  if (count < maxsize - 2)
	    {
	      sprintf (&s[count], "%2.2d",
		       tim_p->tm_hour);
	      count += 2;
	    }
	  else
	    return 0;
	  break;
	case 'I':
	  if (count < maxsize - 2)
	    {
	      if (tim_p->tm_hour == 0 ||
		  tim_p->tm_hour == 12)
		{
		  s[count++] = '1';
		  s[count++] = '2';
		}
	      else
		{
		  sprintf (&s[count], "%.2d",
			   tim_p->tm_hour % 12);
		  count += 2;
		}
	    }
	  else
	    return 0;
	  break;
	case 'j':
	  if (count < maxsize - 3)
	    {
	      sprintf (&s[count], "%.3d",
		       tim_p->tm_yday + 1);
	      count += 3;
	    }
	  else
	    return 0;
	  break;
	case 'm':
	  if (count < maxsize - 2)
	    {
	      sprintf (&s[count], "%.2d",
		       tim_p->tm_mon + 1);
	      count += 2;
	    }
	  else
	    return 0;
	  break;
	case 'M':
	  if (count < maxsize - 2)
	    {
	      sprintf (&s[count], "%2.2d",
		       tim_p->tm_min);
	      count += 2;
	    }
	  else
	    return 0;
	  break;
	case 'p':
	  if (count < maxsize - 2)
	    {
	      if (tim_p->tm_hour < 12)
		s[count++] = 'A';
	      else
		s[count++] = 'P';
	      s[count++] = 'M';
	    }
	  else
	    return 0;
	  break;
	case 'S':
	  if (count < maxsize - 2)
	    {
	      sprintf (&s[count], "%2.2d",
		       tim_p->tm_sec);
	      count += 2;
	    }
	  else
	    return 0;
	  break;
	case 'U':
	  if (count < maxsize - 2)
	    {
	      sprintf (&s[count], "%2.2d",
		       (tim_p->tm_yday + 7 -
			tim_p->tm_wday) / 7);
	      count += 2;
	    }
	  else
	    return 0;
	  break;
	case 'w':
	  if (count < maxsize - 1)
	    {
	      sprintf (&s[count], "%1.1d",
		       tim_p->tm_wday);
	      count++;
	    }
	  else
	    return 0;
	  break;
	case 'W':
	  if (count < maxsize - 2)
	    {
	      sprintf (&s[count], "%2.2d",
		       (tim_p->tm_yday + ((8 -
					   tim_p->tm_wday) % 7)) / 7);
	      count += 2;
	    }
	  else
	    return 0;
	  break;
	case 'x':
	  if (count < maxsize - 15)
	    {
	      for (i = 0; i < 3; i++)
		s[count++] =
		  dname[tim_p->tm_wday][i];
	      s[count++] = ' ';
	      for (i = 0; i < 3; i++)
		s[count++] =
		  mname[tim_p->tm_mon][i];
	      sprintf (&s[count],
		       " %.2d %.4d", tim_p->tm_mday,
		       1900 + tim_p->tm_year);
	      count += 8;
	    }
	  else
	    return 0;
	  break;
	case 'X':
	  if (count < maxsize - 8)
	    {
	      sprintf (&s[count],
		       "%2.2d:%2.2d:%2.2d",
		       tim_p->tm_hour, tim_p->tm_min,
		       tim_p->tm_sec);
	      count += 8;
	    }
	  else
	    return 0;
	  break;
	case 'y':
	  if (count < maxsize - 2)
	    {
	      sprintf (&s[count], "%2.2d",
		       tim_p->tm_year);
	      count += 2;
	    }
	  else
	    return 0;
	  break;
	case 'Y':
	  if (count < maxsize - 4)
	    {
	      sprintf (&s[count], "%.4d",
		       1900 + tim_p->tm_year);
	      count += 4;
	    }
	  else
	    return 0;
	  break;
	case 'Z':
	  break;
	case '%':
	  if (count < maxsize - 1)
	    s[count++] = '%';
	  else
	    return 0;
	  break;
	}
      if (*format)
	format++;
      else
	break;
    }
  s[count] = '\0';
  return count;
}
#endif
