/***********************************************************************
 *                                                                     *
 *              WILLOW(tm) Information Retrieval Software              *
 *     (Washington Information Looker-upper Layered Over Windows)      *
 *             Copyright 1992-1994 University of Washington            *
 *                                                                     *
 *   Willow is a trademark of the University of Washington.  No        *
 *   commercial use of this trademark or software may be made without  *
 *   prior written permission of the University of Washington.         *
 *                                                                     *
 * Permission to use, copy, modify, and distribute Willow(tm) software *
 * and its documentation for any noncommercial purpose is hereby       *
 * granted, provided that the above copyright notice appears in all    *
 * copies and that both the above copyright notice and this permission *
 * notice appear in supporting documentation, and that the name of the *
 * University of Washington not be used in advertising or publicity    *
 * pertaining to distribution of the software without specific,        *
 * written prior permission.  This software is made available "as is", *
 * and THE UNIVERSITY OF WASHINGTON DISCLAIMS ALL WARRANTIES, EXPRESS  *
 * OR IMPLIED, WITH REGARD TO THIS SOFTWARE, INCLUDING WITHOUT         *
 * LIMITATION ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS    *
 * FOR A PARTICULAR PURPOSE, AND IN NO EVENT SHALL THE UNIVERSITY OF   *
 * WASHINGTON BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL     *
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA  *
 * OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, TORT (INCLUDING       *
 * NEGLIGENCE) OR STRICT LIABILITY, ARISING OUT OF OR IN CONNECTION    *
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.                       *
 *                                                                     *
 * For further information please contact:                             *
 * Willow Project                                                      *
 * Information Systems                                                 *
 * University of Washington                                            *
 * 4545 15th NE, 1st floor                                             *
 * Seattle, Washington, 98105, USA                                     *
 * Internet: willow@cac.washington.edu                                 *
 ***********************************************************************/

/**********************************************************
 * table.c
 * In this module we initialize the state table for the 
 * willow database driver. We also define a couple other
 * necessary tables.
 **********************************************************/

#define NULL 0

#include "driver.h"
#include "host.h"

/* External references to host-callback routines */

/* These are general routines */
extern bad_db_hcb(void);
extern bad_uid_hcb(void);
extern bad_login_hcb(void);
extern bad_search_hcb(void);
extern connected_hcb(void);
extern set_summ_num_hcb();
extern login_msg_hcb(void);
extern ask_pw_hcb(void);
extern starting_msg_hcb(void);
extern db_select_hcb(void);
extern timed_out_hcb(void);

/* These are highly host-specific */
extern back_to_item_hcb();
extern back_to_short_mode_hcb();
extern parse_display_hcb(void);
extern parse_stats_hcb(void);
extern parse_summary_hcb(void);
extern parse_info_hcb(void);
extern really_disp_long_hcb();
extern db_gone_hcb(void);

/* The state table which actually drives everything */
state_table  states;

/* Initialize the state table which describes the behavior of
   the BRS searchmate interface.
   */
init_table(void)
{
  int   i,j;
  
  for (i = 0; i < MAX_TRANSITIONS; i++)
    for (j = 0; j < MAX_STATES; j++) {
      states[i][j].in_string  = NULL;
      states[i][j].out_string = NULL;
      states[i][j].next_state = 0;
      states[i][j].action     = NULL;
    }
  
  states[NO_CON][0].in_string  = "";
  states[NO_CON][0].out_string = "";
  states[NO_CON][0].next_state = NO_CON;

  states[AWAIT_LOGIN][0].in_string  = "\nlogin:";
  states[AWAIT_LOGIN][0].out_string = NULL;  /* set dynamically */
  states[AWAIT_LOGIN][0].next_state = AWAIT_PW;  /* changeable */
  states[AWAIT_LOGIN][0].action     = login_msg_hcb;

  states[AWAIT_PW][0].in_string  = "Password:";
  states[AWAIT_PW][0].out_string = NULL; /* set dynamically */
  states[AWAIT_PW][0].next_state = AWAIT_DB;  /* changeable */
  states[AWAIT_PW][0].action     = ask_pw_hcb;
  states[AWAIT_PW][1].in_string  = "\nlogin:";
  states[AWAIT_PW][1].out_string = NULL; 
  states[AWAIT_PW][1].next_state = NO_CON;
  states[AWAIT_PW][1].action     = bad_uid_hcb;
  
  states[AWAIT_DB][0].in_string  = "PRESS ENTER TO CONTINUE -->";
  states[AWAIT_DB][0].out_string = NULL; /* Set dynamically */
  states[AWAIT_DB][0].next_state = AWAIT_SEARCH;
  states[AWAIT_DB][0].action     = db_select_hcb;
  states[AWAIT_DB][1].in_string  = "\nlogin:";
  states[AWAIT_DB][1].out_string = "";
  states[AWAIT_DB][1].next_state = NO_CON;
  states[AWAIT_DB][1].action     = bad_login_hcb;
  states[AWAIT_DB][2].in_string  = "invalid login name or password";
  states[AWAIT_DB][2].out_string = "";
  states[AWAIT_DB][2].next_state = NO_CON;
  states[AWAIT_DB][2].action     = bad_login_hcb;
  states[AWAIT_DB][3].in_string  = "what database for CR note:";
  states[AWAIT_DB][3].out_string = "";
#ifdef ndef
  states[AWAIT_DB][3].next_state = AWAIT_INFO;
#endif
  states[AWAIT_DB][3].next_state = AWAIT_DB;

  states[AWAIT_INFO][0].in_string  = "##done";
  states[AWAIT_INFO][0].out_string = "";
  states[AWAIT_INFO][0].next_state = AWAIT_DB;
  states[AWAIT_INFO][0].action     = parse_info_hcb;

  states[AWAIT_SEARCH][0].in_string  = "cat: cannot open";
  states[AWAIT_SEARCH][0].out_string = "";
  states[AWAIT_SEARCH][0].next_state = NO_CON;
  states[AWAIT_SEARCH][0].action     = bad_db_hcb;
  states[AWAIT_SEARCH][1].in_string  = "##-->";
  states[AWAIT_SEARCH][1].out_string = "";
  states[AWAIT_SEARCH][1].next_state = READY;
  states[AWAIT_SEARCH][1].action     = connected_hcb;
  states[AWAIT_SEARCH][2].in_string  = "IS UNAVAILABLE OR INVALID";
  states[AWAIT_SEARCH][2].out_string = "";
  states[AWAIT_SEARCH][2].next_state = NO_CON;
  states[AWAIT_SEARCH][2].action     = bad_db_hcb;
  states[AWAIT_SEARCH][3].in_string  = "DATABASES COULD NOT BE OPENED";
  states[AWAIT_SEARCH][3].out_string = "";
  states[AWAIT_SEARCH][3].next_state = NO_CON;
  states[AWAIT_SEARCH][3].action     = db_gone_hcb;
  states[AWAIT_SEARCH][4].in_string  = "##GO AWAY";
  states[AWAIT_SEARCH][4].out_string = "";
  states[AWAIT_SEARCH][4].next_state = NO_CON;
  states[AWAIT_SEARCH][4].action     = db_gone_hcb;

  states[READY][0].in_string  = "##-->";
  states[READY][0].out_string = "";    /* variable */
  states[READY][0].next_state = READY; /* variable */
  states[READY][1].in_string  = "System Timeout Limit";
  states[READY][1].out_string = "";
  states[READY][1].next_state = NO_CON;
  states[READY][1].action     = timed_out_hcb;

  states[AWAIT_STATS][0].in_string  = "ERROR: ";
  states[AWAIT_STATS][0].out_string = "";
  states[AWAIT_STATS][0].next_state = READY;
  states[AWAIT_STATS][0].action     = bad_search_hcb;
  states[AWAIT_STATS][1].in_string  = "##-->";
  states[AWAIT_STATS][1].out_string = "";           /* This is changeable */
  states[AWAIT_STATS][1].next_state = READY;        /* This too */
  states[AWAIT_STATS][1].action     = parse_stats_hcb;

  states[DISP_SHRT][0].in_string  = "##-->";
  states[DISP_SHRT][0].out_string = "";   
  states[DISP_SHRT][0].next_state = READY;
  states[DISP_SHRT][0].action     = parse_summary_hcb;

  states[DISP_LONG][0].in_string  = "##-->";
  states[DISP_LONG][0].out_string = "";
  states[DISP_LONG][0].next_state = READY;
  states[DISP_LONG][0].action     = parse_display_hcb;
};

