/***********************************************************************
 *                                                                     *
 *              WILLOW(tm) Information Retrieval Software              *
 *     (Washington Information Looker-upper Layered Over Windows)      *
 *             Copyright 1992-1994 University of Washington            *
 *                                                                     *
 *   Willow is a trademark of the University of Washington.  No        *
 *   commercial use of this trademark or software may be made without  *
 *   prior written permission of the University of Washington.         *
 *                                                                     *
 * Permission to use, copy, modify, and distribute Willow(tm) software *
 * and its documentation for any noncommercial purpose is hereby       *
 * granted, provided that the above copyright notice appears in all    *
 * copies and that both the above copyright notice and this permission *
 * notice appear in supporting documentation, and that the name of the *
 * University of Washington not be used in advertising or publicity    *
 * pertaining to distribution of the software without specific,        *
 * written prior permission.  This software is made available "as is", *
 * and THE UNIVERSITY OF WASHINGTON DISCLAIMS ALL WARRANTIES, EXPRESS  *
 * OR IMPLIED, WITH REGARD TO THIS SOFTWARE, INCLUDING WITHOUT         *
 * LIMITATION ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS    *
 * FOR A PARTICULAR PURPOSE, AND IN NO EVENT SHALL THE UNIVERSITY OF   *
 * WASHINGTON BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL     *
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA  *
 * OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, TORT (INCLUDING       *
 * NEGLIGENCE) OR STRICT LIABILITY, ARISING OUT OF OR IN CONNECTION    *
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.                       *
 *                                                                     *
 * For further information please contact:                             *
 * Willow Project                                                      *
 * Information Systems                                                 *
 * University of Washington                                            *
 * 4545 15th NE, 1st floor                                             *
 * Seattle, Washington, 98105, USA                                     *
 * Internet: willow@cac.washington.edu                                 *
 ***********************************************************************/

/***************************************************
 * db.h
 * Declarations for database data and file structures.
 ***************************************************/

/*
 * 05/17/94  slh  allow more than one possible comment char
 */

#define MAX_ENV_SIZE 1024
#define MAX_LISTS    12
#ifndef MAX_LINE
#define MAX_LINE 240
#endif

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif

#define NUM_TEXT_FIELDS 6
#define NUM_BOOL_FIELDS 4

/* Define a bunch of constants for use in parsing config files (db.conf,
   help.conf, and .willowrc). 
 */
#define DB_TAG           0
#define DB_TAG_S           "db_tag"
#define DB_NAME          1 
#define DB_NAME_S          "db_name"
#define DB_LONG          2 
#define DB_LONG_S          "db_long"
#define DRIVER           3 
#define DRIVER_S           "driver"
#define DB_INT_NAME      4 
#define DB_INT_NAME_S      "db_int_name"
#define TITLE_FIELDS     5 
#define TITLE_FIELDS_S     "title_fields"
#define SUMM_FIELDS      6 
#define SUMM_FIELDS_S      "summary_fields"
#define T_FIELD_NAME     7
#define T_FIELD_NAME_S     "t_field_name"
#define T_FIELD_CMD      8 
#define T_FIELD_CMD_S      "t_field_cmd"
#define LIST_NAME        9 
#define LIST_NAME_S        "list_name"
#define LIST_HIER        10
#define LIST_HIER_S        "list_hier"
#define LIST_MNEMONIC    11
#define LIST_MNEMONIC_S    "list_mnemonic"
#define DFLT_AT          12
#define DFLT_AT_S          "dflt_at"
#define L_FIELD_NAME     13
#define L_FIELD_NAME_S     "l_field_name"
#define L_FIELD_CMD      14
#define L_FIELD_CMD_S      "l_field_cmd"
#define OPAC_MODE        15
#define OPAC_MODE_S        "basic_mode"
#define OPAC_ONLY        16
#define OPAC_ONLY_S        "basic_only"
#define OPAC_FIELD       17
#define OPAC_FIELD_S       "basic_field"
#define RANGE_TYPE       18
#define RANGE_TYPE_S       "t_range_type"
#define YEAR_RANGE       19
#define YEAR_RANGE_S       "t_year_range"
#define USE_PN           20
#define USE_PN_S           "use_prev_next"
#define NO_INFO          21
#define NO_INFO_S          "no_info"
#define EXT_LABEL        22
#define EXT_LABEL_S        "extension_label"
#define EXT_PROG         23
#define EXT_PROG_S         "extension_program"
#define EXT_FIELDS       24
#define EXT_FIELDS_S       "extension_fields"
#define ENV_NAME         25
#define ENV_NAME_S         "env_name"
#define ENV_VAL          26
#define ENV_VAL_S          "env_val"
#define BEG_MENU         27
#define BEG_MENU_S         "begin_menu" 
#define END_MENU         28
#define END_MENU_S         "end_menu" 
#define MENU_ITEM        29
#define MENU_ITEM_S        "item"
#define MENU_SEP         30
#define MENU_SEP_S         "separator"
#define IS_LIMIT         31
#define IS_LIMIT_S         "is_limit"

/* These are used in .willowrc */
#define ECHO_HOST         32
#define ECHO_HOST_S         "echo"
#define USE_LOCAL         33
#define USE_LOCAL_S         "use_local"
#define USE_FTP           34
#define USE_FTP_S           "use_ftp"
#define USE_EMAIL         35
#define USE_EMAIL_S         "use_email"
#define FTP_HOST          36
#define FTP_HOST_S          "ftp_host"
#define FTP_UID           37
#define FTP_UID_S           "ftp_user_id"
#define FTP_APPEND        38
#define FTP_APPEND_S        "ftp_append"
#define PRINT_CMD         39
#define PRINT_CMD_S         "print_cmd"
#define MAIL_ADDR         40
#define MAIL_ADDR_S         "mail_addr"
#define MAIL_SUBJ         41
#define MAIL_SUBJ_S         "mail_subj"
#define ABBR_FIELDS       42
#define ABBR_FIELDS_S       "abbrev_fields"
#define USER_INFO         43
#define USER_INFO_S         "user_info"
#define CUSTOM_FIELDS     44
#define CUSTOM_FIELDS_S     "custom_fields"
#define INCLUDE_QUERY     45
#define INCLUDE_QUERY_S     "include_query"
#define NUM_CONFIG_FIELDS 46
#define HASH_TAB_SIZE     101

#define CONFIG_SEP_CHAR     '='
#define SUB_SEP_CHAR        ':'
#define COMMENT_CHARS       "#!"

/* A structure used for hash-lookup of above fields */
typedef struct _conf_hash_elmt {
  char *str;
  int   idex;
} conf_hash_elmt;

/* Values for range types */
#define DAY_S     "DAY"
#define MONTH_S   "MONTH"
#define YEAR_S    "YEAR"
#define NUMERIC_S "NUMERIC"
#define NUMERIC 1
#define YEAR    2
#define MONTH   3
#define DAY     4


/* The biggest allowable difference in year_range.  If this is changed, 
   must edit range.uil to add more buttons. */
#define MAX_YEAR_RANGE 30

/* Put together a structure to hold all the constants necessary to describe
   a database.
 */
typedef struct _db_field_desc_rec {  /* describe a single field */
  char *name;
  char *command;
  char  opac_field;
  char  opac_only;

  /* These are used for range types */
  int    range_type;
  int    first_year;
  int    last_year;

  /* browse lists */
  int   list_cnt;
  char *list_names[MAX_LISTS];   
  char  is_limit;
#ifdef WILLOW
  Widget option_btn;
#endif
  /* These two are actually only used by WILCO */
  char   list_hier;
  char  *list_mnemonic;
  struct _db_field_desc_rec *next;
} db_field_desc_rec;

typedef struct _db_ext_rec { /* describe a multimedia extension */
  char *label;
  char *program;
  struct _db_ext_rec *next;
} db_ext_rec;

/* used for holding a user's response to a driver question */
typedef struct _user_info_rec { 
  char *tag;
  int   type;
  char *response;
  struct _user_info_rec *next;
} user_info_rec;

typedef struct _db_desc_rec { /* describe an entire database */
  char  *tag;
  char  *short_name;
  char  *long_name;
  char  *driver;
  char  *host_name;
  char  *internal_name;
  char  *summary_fields;
  char  *title_fields;
  char  *ext_fields;
  char  *environ;
  db_ext_rec *extensions;
  char **display_fields;
  char   all_default_fields;
  int    env_len;
  char   opac_mode;
  char   use_prev_next;                /* Use prev/next match word btns    */
  char   no_info;                      /* Do not show copyright box        */
  int    num_text_fields;              /* not related to NUM_TEXT_FIELDS   */
  db_field_desc_rec *text_fields;      /* list of text field descriptions  */
  int    num_limit_fields;             /* not related to NUM_BOOL_FIELDS   */
  db_field_desc_rec *limit_fields;     /* list of limit field descriptions */
  db_field_desc_rec *defaults[NUM_TEXT_FIELDS];   /*text fields to display */
  user_info_rec *user_info;            /* store responses to driver questions*/
#ifdef WILLOW
  Widget dflt_btns[NUM_TEXT_FIELDS];   /* pointers to dflt button on options
                                          menu. */
#endif
} db_desc_rec;

/* Define a structure for holding the status of a range button */
typedef struct _range_stat_rec {
  char limited;
  char low;       /* Is this a From (low) or a To? */
  int  month_val;
  int  day_val;
#ifdef WILLOW
  Widget  day;
  Widget  month;
#endif
  int  numeric;   /* this is year or general numeric */
  db_field_desc_rec *field;
} range_stat_rec;


/* This structure is used to hold a temporary linked list of our Database
   menu contruction directives. The directives appear at the head of the
   file, but are not useful until we have read all the db entries,
   so we hang onto them here.
 */
typedef struct _db_menu_rec {
  int   type;
  char *value;
  struct _db_menu_rec *next;
} db_menu_rec;
