/***********************************************************************
 *                                                                     *
 *              WILLOW(tm) Information Retrieval Software              *
 *     (Washington Information Looker-upper Layered Over Windows)      *
 *             Copyright 1992-1994 University of Washington            *
 *                                                                     *
 *   Willow is a trademark of the University of Washington.  No        *
 *   commercial use of this trademark or software may be made without  *
 *   prior written permission of the University of Washington.         *
 *                                                                     *
 * Permission to use, copy, modify, and distribute Willow(tm) software *
 * and its documentation for any noncommercial purpose is hereby       *
 * granted, provided that the above copyright notice appears in all    *
 * copies and that both the above copyright notice and this permission *
 * notice appear in supporting documentation, and that the name of the *
 * University of Washington not be used in advertising or publicity    *
 * pertaining to distribution of the software without specific,        *
 * written prior permission.  This software is made available "as is", *
 * and THE UNIVERSITY OF WASHINGTON DISCLAIMS ALL WARRANTIES, EXPRESS  *
 * OR IMPLIED, WITH REGARD TO THIS SOFTWARE, INCLUDING WITHOUT         *
 * LIMITATION ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS    *
 * FOR A PARTICULAR PURPOSE, AND IN NO EVENT SHALL THE UNIVERSITY OF   *
 * WASHINGTON BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL     *
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA  *
 * OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, TORT (INCLUDING       *
 * NEGLIGENCE) OR STRICT LIABILITY, ARISING OUT OF OR IN CONNECTION    *
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.                       *
 *                                                                     *
 * For further information please contact:                             *
 * Willow Project                                                      *
 * Information Systems                                                 *
 * University of Washington                                            *
 * 4545 15th NE, 1st floor                                             *
 * Seattle, Washington, 98105, USA                                     *
 * Internet: willow@cac.washington.edu                                 *
 ***********************************************************************/

/*******************************************************
 * dialog.c
 * This file contains the general purpose dialog creation
 * routines.
 ******************************************************/

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <Mrm/MrmAppl.h>
#include <Xm/MessageB.h>
#include <Xm/SelectioB.h>
#include <Xm/ToggleB.h>
#include <Xm/Text.h>
#include <Xm/FileSB.h>

#include "willow.h"
#include "db.h"
#include "save_query.h"
#include "protocol.h"

#include "globes/globe1.xbm"
#include "globes/globe2.xbm"
#include "globes/globe3.xbm"
#include "globes/globe4.xbm" 
#include "globes/globe5.xbm" 
#include "globes/globe6.xbm" 
#include "globes/globe7.xbm" 
#include "globes/globe8.xbm" 
#include "globes/globe9.xbm" 
#include "globes/globe10.xbm"
#include "globes/globe11.xbm"
#include "globes/globe12.xbm"
#include "globes/globe13.xbm"
#include "globes/globe14.xbm"
#include "globes/globe15.xbm"
#include "globes/globe16.xbm"
#include "globes/globe17.xbm"
#include "globes/globe18.xbm"
#include "globes/globe19.xbm"
#include "globes/globe20.xbm"
#include "globes/globe21.xbm"
#include "globes/globe22.xbm"
#include "globes/globe23.xbm"
#include "globes/globe24.xbm"
#include "globes/globe25.xbm"
#include "globes/globe26.xbm"
#include "globes/globe27.xbm"
#include "globes/globe28.xbm"
#include "globes/globe29.xbm"
#include "globes/globe30.xbm"
#define GLOBE_WIDTH  128
#define GLOBE_HEIGHT 128

extern void field_pick_ok(void);
extern void field_pick_cancel(void);
extern void starrify_cb(Widget w, char **pw, XmTextVerifyPtr reason);
void send_user_resp(int which, char *text);
extern void really_open_search_cb(Widget w, int *tag, XmFileSelectionBoxCallbackStruct *reason);
extern void really_save_search_cb(Widget w, int *tag, XmFileSelectionBoxCallbackStruct *reason);
extern int display_help (Widget w, char *path, char *file, char **root, char *name, Widget Wpos, char *label, char **defs, int ndefs);
extern int error (char *);
extern Widget display_help_shell (char *H);
extern void pop_warning (char *, char *, Widget);
extern int list_get_selected_cnt (Widget w);
extern void clear_summ_bodies (void);


/* Local function prototypes */
static void finish_about(void);
static int  read_bitmap_file(char *name, 
			     int *width, 
			     int *height, 
			     Pixmap *bitmap);
static void spin_about(void);
static void spin_globe(int tag);
void clear_help (void);
void map_dialog (Widget dialog);
void ForceDialog (Widget w);
void toggle_save_type_cb (Widget w, int *target, XmToggleButtonCallbackStruct *reason);
void q_toggle_save_type_cb (Widget w, int *target, XmToggleButtonCallbackStruct *reason);
void center_dialog(Widget top, Widget bottom);
void save_geometry(Widget shell, geom_rec *geom);
void set_geometry(Widget shell, geom_rec *geom);


extern Widget     widgets[],
                  search_shell,
                  summary_shell,
                  dict_shell,
                  results_shell,
                  toplevel;

extern XtAppContext app_context;
Search_list_ptr  search_head;
extern int opac_mode;

extern MrmHierarchy s_MrmHierarchy;  /* MRM database hierarchy ID */
extern MrmType dummy_class;         /* and class variable. */
extern int spin_delay;
extern int  save_type;
extern int  q_save_type;
extern Boolean ftp_append;
extern Boolean show_search;
extern char *ftp_host;
extern char *ftp_uid;
extern char *ftp_pw;
extern char *ftp_file;
extern char *ftp_q_file;
extern char *mail_addr;
extern char *mail_subj;
extern char *masked_response;
extern char *willow_dir;
extern char *willow_size;
extern db_desc_rec *cur_db;
extern Boolean query_opening;
extern Boolean abbrev_fields;
extern char *db_info;

Widget top_conn_opt_text;

/* Pointers to each shell's help record */
char   *search_hs  = NULL, 
       *basic_hs   = NULL,
       *summary_hs = NULL, 
       *full_hs    = NULL, 
       *dict_hs    = NULL;

/* Save type as of time of popping-up save-box */
int prev_save_type;
int q_prev_save_type;

/* type of currently active user query */
static int cur_query_type;
static char *cur_query_tag;

/* save previous ftp_password in case of save-box cancel */
char prev_ftp_pw[MAX_LINE];

/* Graphics context, initialized in load_globes */
GC     gc = NULL;

/* pixmap for about box */
static Pixmap about_pix = NULL;

/* Depth of current screen */
static int depth;

/* Currently used sub-form of save box (index into widgets[]) */
int cur_save_box = 0;
int q_cur_save_box = 0;

/* Did we open a save-destination box for full or summary? */
Boolean popped_full = FALSE;

/* timeout id */
XtIntervalId globe_timer_id = NULL;
XtIntervalId about_timer_id = NULL;



/* Pop up the help box */
void help_cb(Widget  w, Widget  target, char *reason)
{
  Widget shell;
  char  *dir_name, *file_name;
  int    ndefs;
  char **help_set;
  char  *title;
  char  *defs_list[10];

  /* Find shell-parent of pressed button */
  for (shell = w;
       !XtIsShell(shell);
       shell = XtParent(shell))
    ;
  
  dir_name = XtMalloc(strlen(willow_dir) + strlen(HELP_DIR) + 2);
  sprintf(dir_name, "%s%s", willow_dir, HELP_DIR);
  if (shell == search_shell) {
    if (!opac_mode) {
      file_name = SEARCH_HELP;
      help_set = &search_hs;
      title = "Willow Search Window Help";
    }
    else {
      file_name = BASIC_HELP;
      help_set = &basic_hs;
      title = "Willow Basic Search Help";
    }
  }
  else if (shell == summary_shell) {
    file_name = SUMMARY_HELP;
    help_set = &summary_hs;
    title = "Willow Summaries Window Help";
  }
  else if (shell == results_shell) {
    file_name = FULL_REC_HELP;
    help_set = &full_hs;
    title = "Willow Full-Record Window Help";
  }
  else if (shell == dict_shell) {
    file_name = BROWSE_HELP;
    help_set = &dict_hs;
    title = "Willow Subjects Window Help";
  }
  else
    error("Could not resolve help file");

  ndefs = 0;
  defs_list[ndefs] = NULL;
  if (cur_db)
    defs_list[ndefs++] = cur_db->tag;

  defs_list[ndefs++] = willow_size;

  /* Clear any visible helps */
  clear_help();
  display_help(shell, dir_name, file_name, help_set, "HelpShell", 
	       shell, title, defs_list, ndefs);
}

/* Unmap the current help widgets */
void
clear_help(void)
{
  Widget hw;

  if (hw=display_help_shell(search_hs)) 
    XtDestroyWidget(hw);
  if (hw=display_help_shell(basic_hs)) 
    XtDestroyWidget(hw);
  if (hw=display_help_shell(summary_hs)) 
    XtDestroyWidget(hw);
  if (hw=display_help_shell(full_hs)) 
    XtDestroyWidget(hw);
  if (hw=display_help_shell(dict_hs)) 
    XtDestroyWidget(hw);
}

/* Pop up a message box, can be any of the motif-defined types,
   error, information, question, warning, or working.
   Pass in the type of dialog, the message string, and a mask
   with bits set for each button we want to appear.
 */
void 
pop_message_dialog(unsigned char type, /* valid value for XmNdialogType */
		   char          *message,
		   int            button_mask,
		   Widget         parent)
{
  static  Widget prev_parent = NULL;
  static  unsigned char prev_type = -1;
  Widget  button;
  Arg     args[10];
  int     n;

  if (!parent)
    return;

  /* Fetch the widget from the UID file */
  if ((widgets[k_message] == NULL) || (parent != prev_parent) ||
      (type != prev_type)) {
    if (widgets[k_message]) {
      XtDestroyWidget(widgets[k_message]);
      widgets[k_message] = NULL;
    }
    if (MrmFetchWidget(s_MrmHierarchy, 
		       "message_box", 
		       parent,
		       &widgets[k_message], 
		       &dummy_class) != MrmSUCCESS)
      error("can't fetch message box widget");
  }
  prev_parent = parent;
  prev_type = type;

  /* Set the string and type values. Use CreateLtoR instead of 
     CreateSimple to get recognition of \n */
  n = 0;
  XtSetArg(args[n], XmNmessageString, 
	   XmStringCreateLtoR(message, XmSTRING_DEFAULT_CHARSET));n++;
  XtSetArg(args[n], XmNdialogType, type);n++;
  XtSetValues(widgets[k_message], args, n);
      
  /* Manage buttons we want, and un-manage buttons we don't */

  button = XmMessageBoxGetChild(widgets[k_message], XmDIALOG_OK_BUTTON);
  if (button_mask & OK_BUTTON_MASK) {
    XtManageChild(button);
#ifdef ndef 
    /* This code causes a run-time warning, and I am not sure whether
       or not it is needed */
    XtVaSetValues(widgets[k_message], XmNcancelButton, button, NULL);
#endif
  }
  else
    XtUnmanageChild(button);

  button = XmMessageBoxGetChild(widgets[k_message], XmDIALOG_CANCEL_BUTTON);
  if (button_mask & CANCEL_BUTTON_MASK){
    XtManageChild(button);
#ifdef ndef
    XtVaSetValues(widgets[k_message], XmNcancelButton, button, NULL);
#endif
  }
  else
    XtUnmanageChild(button);

  button = XmMessageBoxGetChild(widgets[k_message], XmDIALOG_HELP_BUTTON);
  if (button_mask & HELP_BUTTON_MASK)
    XtManageChild(button);
  else
    XtUnmanageChild(button);

  /* If no buttons asked for, drop separator */
  button = XmMessageBoxGetChild(widgets[k_message], XmDIALOG_SEPARATOR);
  if (!button_mask) 
    XtUnmanageChild(button);
  else
    XtManageChild(button);
		
  map_dialog(widgets[k_message]);
  ForceDialog(widgets[k_message]);
}

/* clear a dialog via passed in widget */
void 
clear_dialog_cb(Widget  w,
		Widget  target,
		u_long *reason)
{
  if (target)
    XtUnmanageChild(target);
}

/* clear a dialog via passed in tag */
void 
clear_dialog_tag_cb(Widget  w, int  *target, u_long *reason)
{
  if (widgets[*target]) {
    XtUnmanageChild(widgets[*target]);
    if ((*target == k_about_box) && about_pix) {
      XFreePixmap(XtDisplay(search_shell), about_pix);
      about_pix = NULL;
    }
  }
}

/* Pop-up the copyright information dialog */
void pop_info_cb(Widget widg, int *tag, u_long *reason)
{
  int n;
  Arg args[10];
  Widget button;
  Widget dlg;
  Dimension w, h;
  XtTranslations trans;
  char *info_str;

  /* Fetch the widget from the UID file */
  if (widgets[k_info_dlg] == NULL) {
    if (MrmFetchWidget(s_MrmHierarchy, 
		       "message_box", 
		       search_shell,
		       &widgets[k_info_dlg], 
		       &dummy_class) != MrmSUCCESS)
      error("can't fetch copyright  message box widget");

    /* un-manage buttons we don't want */    
    button = XmMessageBoxGetChild(widgets[k_info_dlg], XmDIALOG_CANCEL_BUTTON);
    XtUnmanageChild(button);
    button = XmMessageBoxGetChild(widgets[k_info_dlg], XmDIALOG_HELP_BUTTON);
    XtUnmanageChild(button);

    /* Make close button sensitive to any key */
    trans = XtParseTranslationTable("<Key>: ManagerGadgetSelect()\n");
    XtOverrideTranslations(widgets[k_info_dlg], trans);

    XtVaSetValues(XtParent(widgets[k_info_dlg]), XmNtitle, 
			   "About Current Database", NULL);
  }

  /* Set the string and type values. Use CreateLtoR instead of 
     CreateSimple to get recognition of \n */
  if (!cur_db)
    info_str = NO_DB_INFO_STR;
  else if (db_info) {
    info_str = XtMalloc(strlen(db_info) + MAX_LINE);
    sprintf(info_str, "%s\n%s", db_info, cur_db->tag);
  }
  else {
    info_str = XtMalloc(strlen(NO_DB_INFO_STR) + MAX_LINE);
    sprintf(info_str, "%s\n%s", cur_db->tag, NO_DB_INFO_STR);
  }

  n = 0;
  XtSetArg(args[n], XmNmessageString, 
	   XmStringCreateLtoR(info_str, XmSTRING_DEFAULT_CHARSET));n++;
  if (info_str != NO_DB_INFO_STR)
    XtFree(info_str);
  XtSetArg(args[n], XmNdialogType, XmDIALOG_INFORMATION);n++;
  XtSetValues(widgets[k_info_dlg], args, n);
  dlg = widgets[k_info_dlg];
  while (!XtIsShell(dlg))
    dlg = XtParent(dlg);

  map_dialog(widgets[k_info_dlg]);
  ForceDialog(widgets[k_info_dlg]);

  /* Move the dialog to the righthand corner */
  XtVaGetValues(widgets[k_info_dlg], XmNwidth, &w, XmNheight, &h, NULL);
  XtMoveWidget(dlg, WidthOfScreen(XtScreen(toplevel)) - w - 20,
	       HeightOfScreen(XtScreen(toplevel)) - h - 20);
}


/* Add a named search to the current searches menu */
void 
add_search_cb(Widget  w,
	      int    *tag,
	      u_long *reason)
{
  Widget  button;
  
  if (widgets[k_add_search] == NULL)  {
    if (MrmFetchWidget(s_MrmHierarchy, 
		       "add_search_box", 
		       search_shell,
		       &widgets[k_add_search], 
		       &dummy_class) != MrmSUCCESS)
      error("can't fetch add search window widget");
      
    /* We don't want help button, so get it and remove it */
    button = XmSelectionBoxGetChild(widgets[k_add_search], 
				    XmDIALOG_HELP_BUTTON);
    XtUnmanageChild(button);
    XmTextSetString(XmSelectionBoxGetChild(widgets[k_add_search], 
					   XmDIALOG_TEXT), "");
  }
  XmTextSetString(XmSelectionBoxGetChild(widgets[k_add_search], 
					 XmDIALOG_TEXT), "");
  map_dialog(widgets[k_add_search]);
  ForceDialog(widgets[k_add_search]);
  XmProcessTraversal(XmFileSelectionBoxGetChild(widgets[k_add_search],
		     XmDIALOG_TEXT),XmTRAVERSE_CURRENT);
  XmProcessTraversal(XmFileSelectionBoxGetChild(widgets[k_add_search],
		     XmDIALOG_TEXT),XmTRAVERSE_CURRENT);
}

/* Pop the cheat dialog */
void 
cheat_cb(Widget  w,
	 int    *tag,
	 u_long *reason)
{
  Widget  button;
  
  if (widgets[k_cheat_box] == NULL) {
    if (MrmFetchWidget(s_MrmHierarchy, 
		       "cheat_box", 
		       search_shell,
		       &widgets[k_cheat_box], 
		       &dummy_class) != MrmSUCCESS)
      error("can't fetch cheat-box widget");
    
    /* We don't want help button, so get it and remove it */
    button = XmSelectionBoxGetChild(widgets[k_cheat_box], 
				      XmDIALOG_HELP_BUTTON);
    XtUnmanageChild(button);

    XmTextSetString(XmSelectionBoxGetChild(widgets[k_cheat_box], 
					   XmDIALOG_TEXT),  "");
  }
  map_dialog(widgets[k_cheat_box]);
  ForceDialog(widgets[k_cheat_box]);
  XmProcessTraversal(XmFileSelectionBoxGetChild(widgets[k_cheat_box],
	             XmDIALOG_TEXT),XmTRAVERSE_CURRENT);
  XmProcessTraversal(XmFileSelectionBoxGetChild(widgets[k_cheat_box],
	             XmDIALOG_TEXT),XmTRAVERSE_CURRENT);
}

/* Put up a dialog to ask the user a question from the driver */
void pop_user_query(int type, char *query_text, char *query_tag,
		    char *yes_label, char *no_label)
{
  Widget  button, text;
  user_info_rec *ptr;
  char   title[MAX_LINE];

  if (widgets[k_user_query_box] == NULL) {
    if (MrmFetchWidget(s_MrmHierarchy, 
		       "user_query_box", 
		       search_shell,
		       &widgets[k_user_query_box], 
		       &dummy_class) != MrmSUCCESS)
      error("can't fetch user query widget");
    
    /* We don't want help button, so get and remove it */
    button = XmSelectionBoxGetChild(widgets[k_user_query_box], 
				    XmDIALOG_HELP_BUTTON);
    XtUnmanageChild(button);
    
    /* buttons should not accept keyboard focus */
    button = XmSelectionBoxGetChild(widgets[k_user_query_box], 
				    XmDIALOG_OK_BUTTON);
    XtVaSetValues(button, XmNtraversalOn, FALSE, NULL);
    button = XmSelectionBoxGetChild(widgets[k_user_query_box], 
				    XmDIALOG_CANCEL_BUTTON);
    XtVaSetValues(button, XmNtraversalOn, FALSE, NULL);
  }
  sprintf(title, "%s:%s", cur_db->tag, query_tag);
  XtVaSetValues(XtParent(widgets[k_user_query_box]), XmNtitle, 
		title, NULL);
  button = XmSelectionBoxGetChild(widgets[k_user_query_box], 
				  XmDIALOG_OK_BUTTON);
  XtManageChild(button);
  if (yes_label[1])
    XtVaSetValues(button,XmNlabelString, XmStringCreateSimple(yes_label),NULL);
  else
    XtUnmanageChild(button);

  button = XmSelectionBoxGetChild(widgets[k_user_query_box], 
				  XmDIALOG_CANCEL_BUTTON);
  XtManageChild(button);
  if (no_label[1])
    XtVaSetValues(button, XmNlabelString, XmStringCreateSimple(no_label),NULL);
  else
    XtUnmanageChild(button);

  /* Set the string value. Use CreateLtoR instead of 
     CreateSimple to get recognition of \n */
  text = XmSelectionBoxGetChild(widgets[k_user_query_box], 
				XmDIALOG_SELECTION_LABEL);
  XtVaSetValues(text, XmNlabelString, 
		XmStringCreateLtoR(query_text, XmSTRING_DEFAULT_CHARSET),
		NULL);

  text = XmSelectionBoxGetChild(widgets[k_user_query_box], XmDIALOG_TEXT);

  XtRemoveAllCallbacks(text, XmNmodifyVerifyCallback);
  XmTextSetString(text, "");

  cur_query_type = type;
  switch (type) {
  case BOOL_REQ:
    XtUnmanageChild(text);
    break;
  case TEXT_REQ:
    XtManageChild(text);
    break;
  case MASKED_REQ:
    XtManageChild(text);
    /* Register callback for echoing *'s for mask */
    masked_response[0] = 0;
    XtAddCallback(text, XmNmodifyVerifyCallback, starrify_cb, &masked_response);
    break;
  default:
    pop_warning("Unknown type in user-request packet", NULL, search_shell);
    break;
  }

  /* Look for this query in our buffered list */
  if (type != BOOL_REQ)
    for (ptr = cur_db->user_info; ptr; ptr = ptr->next)
      if (strcasecmp(ptr->tag, query_tag) == 0) {
	XmTextSetString(text, ptr->response);
	break;
      }
  
  XtRealizeWidget(widgets[k_user_query_box]);
  map_dialog(widgets[k_user_query_box]);
  XmProcessTraversal(text, XmTRAVERSE_CURRENT);

  XtFree(query_text);
  XtFree(yes_label);
  XtFree(no_label);
  cur_query_tag = query_tag;
}

/* Callback when the user responds to the user query */
void user_query_cb(Widget w, int *which, char *reason)
{
  char *text;
  user_info_rec *ptr;

  if (cur_query_type == BOOL_REQ)
    text = "";
  else if (cur_query_type == TEXT_REQ)
    text = XmTextGetString(XmSelectionBoxGetChild(widgets[k_user_query_box], 
						  XmDIALOG_TEXT));
  else
    text = strdup(masked_response);
  send_user_resp(*which, text);

  /* If user answered yes, then hang onto the answer to use as
     default if same question gets asked again. */
  if (cur_db && *which && (cur_query_type != BOOL_REQ)) {
    for (ptr = cur_db->user_info; ptr; ptr = ptr->next) {
      if (strcasecmp(ptr->tag, cur_query_tag) == 0)
	break;
    }
    if (ptr) {
      XtFree(cur_query_tag);
      XtFree(ptr->response);
      ptr->response = text;
    }
    else {
      ptr = (user_info_rec *) XtMalloc(sizeof(user_info_rec));
      ptr->tag = cur_query_tag;
      ptr->type = cur_query_type;
      ptr->response = text;
      ptr->next = cur_db->user_info;
      cur_db->user_info = ptr;
    }
  }
}

/* Put up the save options box 
 */
void 
pop_save_box(int which)
{
  Widget button, parent;
  XmToggleButtonCallbackStruct reason;
  
  if (popped_full)
    parent = results_shell;
  else
    parent = summary_shell;

  if (which == NULL){
    if (MrmFetchWidget(s_MrmHierarchy, 
		       "save_file_box", 
		       toplevel,
		       &widgets[k_save_box], 
		       &dummy_class) != MrmSUCCESS)
      error("can't fetch save_file_box widget");    
    
    XtVaSetValues(XtParent(widgets[k_save_box]), XmNtitle, 
			   "Select Save Method", NULL);

    /* Set up the local full-file portion */
    /* We don't want help button, so get it and remove it */
    button = XmSelectionBoxGetChild(widgets[k_local_file_box], 
				    XmDIALOG_HELP_BUTTON);
    XtUnmanageChild(button);        
    XtAddCallback(widgets[k_local_file_box], XmNcancelCallback, 
		  clear_dialog_cb, widgets[k_save_box]);
    XtUnmanageChild(widgets[k_local_file_box]);
    
    /* Set up the ftp portion of the box */
    if (!ftp_pw) {
      ftp_pw = XtMalloc(1);
      ftp_pw[0] = 0;
    }
    XtAddCallback(widgets[k_ftp_pw_text], XmNmodifyVerifyCallback, 
		  starrify_cb, &ftp_pw);
    XtUnmanageChild(widgets[k_ftp_options_box]);

    /* Set-up e-mail part */
    XtUnmanageChild(widgets[k_email_box]);
  }
  else {
    /* Make sure it is cleared */
    XtUnmanageChild(widgets[k_save_box]);

    if (XtIsRealized(widgets[k_save_box])) {
      center_dialog(widgets[k_save_box], parent);
      map_dialog(widgets[k_save_box]);
    }
    else{
      map_dialog(widgets[k_save_box]);
      center_dialog(widgets[k_save_box], parent);
    }
    ForceDialog(widgets[k_save_box]);
    /* Sometimes the first one does not work */
    center_dialog(widgets[k_save_box], parent);
    reason.set = TRUE;
    toggle_save_type_cb(NULL, &which, &reason);
  }
}

/* Put up the save query box
 */
void 
q_pop_save_box(int which)
{
  Widget button;

  if (which == NULL) {
    if (MrmFetchWidget(s_MrmHierarchy, 
		       "q_save_file_box", 
		       search_shell,
		       &widgets[k_q_save_box], 
		       &dummy_class) != MrmSUCCESS)
      error("can't fetch q_save_file_box widget");    

    XtVaSetValues(XtParent(widgets[k_q_save_box]), XmNtitle, 
			   "Query Set Save/Retrieve", NULL);
    
    /* Set up the local full-file portion */
    /* We don't want help button, so get it and remove it */
    button = XmSelectionBoxGetChild(widgets[k_q_fs_box],
				    XmDIALOG_HELP_BUTTON);
    XtUnmanageChild(button);        
    XtAddCallback(widgets[k_q_fs_box], XmNcancelCallback, 
		  clear_dialog_cb, widgets[k_q_save_box]);
    XtUnmanageChild(widgets[k_q_fs_box]);
    
    /* Set up the ftp portion of the box */
    if (!ftp_pw) {
      ftp_pw = XtMalloc(1);
      ftp_pw[0] = 0;
    }
    XtAddCallback(widgets[k_q_ftp_pw_text], XmNmodifyVerifyCallback, 
		  starrify_cb, &ftp_pw);
    XtUnmanageChild(widgets[k_q_ftp_options_box]);
  }
  else {
    XmToggleButtonCallbackStruct reason;
    
    if (!query_opening && !search_head)
      pop_warning("There are no Queries to save!", NULL, search_shell);

    else {
      map_dialog(widgets[k_q_save_box]);
      ForceDialog(widgets[k_q_save_box]);
      reason.set = TRUE;
      q_toggle_save_type_cb(NULL, &which, &reason);
    }
  }
}

/* Toggle which sub-form of the save-file-picker is managed */
void 
toggle_save_type_cb(Widget  w,
		    int  *target,
		    XmToggleButtonCallbackStruct *reason)
{
  int which = *target;

  if (!reason->set)
    return;

  XtManageChild(widgets[which]);
  if (cur_save_box && (cur_save_box != which))
    XtUnmanageChild(widgets[cur_save_box]);
  cur_save_box = which;

  if (which == k_local_file_box) {
    XmProcessTraversal(XmFileSelectionBoxGetChild(
        widgets[k_local_file_box], XmDIALOG_TEXT),XmTRAVERSE_CURRENT);
    XmProcessTraversal(XmFileSelectionBoxGetChild(
        widgets[k_local_file_box], XmDIALOG_TEXT),XmTRAVERSE_CURRENT);
    save_type = SAVE_LOCAL;
    XtVaSetValues(widgets[k_save_box], XmNcancelButton, 
		  XmFileSelectionBoxGetChild(widgets[k_local_file_box], 
					     XmDIALOG_CANCEL_BUTTON), NULL);
  }
  else if (which == k_ftp_options_box) {
    XmProcessTraversal(widgets[k_ftp_host_text], XmTRAVERSE_CURRENT);
    XmProcessTraversal(widgets[k_ftp_host_text], XmTRAVERSE_CURRENT);
    save_type = SAVE_FTP;
    XtVaSetValues(widgets[k_save_box], XmNcancelButton,
		  widgets[k_ftp_opt_cancel], NULL);
  }
  else if (which == k_email_box) {
    XmProcessTraversal(widgets[k_email_text], XmTRAVERSE_CURRENT);
    XmProcessTraversal(widgets[k_email_text], XmTRAVERSE_CURRENT);
    save_type = SAVE_EMAIL;
    XtVaSetValues(widgets[k_save_box], XmNcancelButton,
		  widgets[k_mail_opt_cancel], NULL);
  }
}

/* Toggle which sub-form of the save-file-picker is managed */
void 
q_toggle_save_type_cb(Widget  w,
		      int  *target,
		      XmToggleButtonCallbackStruct *reason)
{
  int which = *target;

  if (!reason->set)
    return;

  XtManageChild(widgets[which]);
  if (q_cur_save_box && (q_cur_save_box != which))
    XtUnmanageChild(widgets[q_cur_save_box]);
  q_cur_save_box = which;

  if (which == k_q_fs_box) {
    XmProcessTraversal(XmFileSelectionBoxGetChild(
        widgets[k_q_fs_box], XmDIALOG_TEXT),XmTRAVERSE_CURRENT);
    XmProcessTraversal(XmFileSelectionBoxGetChild(
        widgets[k_q_fs_box], XmDIALOG_TEXT),XmTRAVERSE_CURRENT);
    q_save_type = SAVE_LOCAL;
    XtVaSetValues(widgets[k_q_save_box], XmNcancelButton, 
		  XmFileSelectionBoxGetChild(widgets[k_q_fs_box], 
					     XmDIALOG_CANCEL_BUTTON), NULL);
  }
  else if (which == k_q_ftp_options_box) {
    XmProcessTraversal(widgets[k_q_ftp_host_text], XmTRAVERSE_CURRENT);
    XmProcessTraversal(widgets[k_q_ftp_host_text], XmTRAVERSE_CURRENT);
    q_save_type = SAVE_FTP;
    XtVaSetValues(widgets[k_q_save_box], XmNcancelButton,
		  widgets[k_q_ftp_opt_cancel], NULL);
  }
}


/* Initial callbacks for popping up the save_box */
void 
open_save_cb(Widget  w,
		     int    *tag,
		     u_long *reason)
{
  popped_full = (w == widgets[k_full_save_btn]);
  if (!popped_full && !list_get_selected_cnt(widgets[k_summary_titles])) {
    pop_warning("You must select some summaries before saving them.", NULL,
		summary_shell);
    return;
  }

  prev_save_type = save_type;
  strcpy(prev_ftp_pw, ftp_pw);
  if (save_type == SAVE_LOCAL)
    pop_save_box(k_local_file_box);
  else if (save_type == SAVE_FTP)
    pop_save_box(k_ftp_options_box);
  else
    pop_save_box(k_email_box);
}

/* Initial callbacks for popping up the save_query_box */
void 
save_search_cb(Widget  w,
	       int    *tag,
	       u_long *reason)
{
  XtRemoveAllCallbacks(widgets[k_q_fs_box], XmNokCallback);
  if (w == widgets[k_open_search_btn]) {
    query_opening = TRUE;
    XtAddCallback(widgets[k_q_fs_box], XmNokCallback,
		  really_open_search_cb, NULL);
    XtVaSetValues(widgets[k_q_local_save_btn], XmNlabelString,
		  XmStringCreateSimple(Q_LOCAL_OPEN_STR), NULL);
    XtVaSetValues(widgets[k_q_ftp_save_btn], XmNlabelString,
		  XmStringCreateSimple(Q_FTP_OPEN_STR), NULL);
  }
  else {
    query_opening = FALSE;
    XtAddCallback(widgets[k_q_fs_box], XmNokCallback,
		  really_save_search_cb, NULL);
    XtVaSetValues(widgets[k_q_local_save_btn], XmNlabelString,
		  XmStringCreateSimple(Q_LOCAL_SAVE_STR), NULL);
    XtVaSetValues(widgets[k_q_ftp_save_btn], XmNlabelString,
		  XmStringCreateSimple(Q_FTP_SAVE_STR), NULL);
  }

  q_prev_save_type = q_save_type;
  strcpy(prev_ftp_pw, ftp_pw);
  if (q_save_type == SAVE_LOCAL)
    q_pop_save_box(k_q_fs_box);
  else if (q_save_type == SAVE_FTP)
    q_pop_save_box(k_q_ftp_options_box);
}


/* Cancel the save box */
void
q_save_box_cancel_cb(Widget  w,
		     int    *tag,
		     u_long *reason)
{
  int which;   /* widget indexer */

  /* Restore save_type button */
  if (q_prev_save_type == SAVE_LOCAL)
    XmToggleButtonSetState(widgets[k_q_local_save_btn], TRUE, TRUE);
  else
    XmToggleButtonSetState(widgets[k_q_ftp_save_btn], TRUE, TRUE);

  /* restore all ftp fields */
  if (ftp_host)
    XmTextSetString(widgets[k_q_ftp_host_text], ftp_host);
  if (ftp_uid)
    XmTextSetString(widgets[k_q_ftp_uid_text], ftp_uid);
  if (ftp_pw[0]) 
    XmTextSetString(widgets[k_q_ftp_pw_text], prev_ftp_pw);

  if (ftp_q_file) 
    XmTextSetString(widgets[k_q_ftp_file_text], ftp_q_file);

  /* clear the box */
  which = k_q_save_box;
  clear_dialog_tag_cb(w, &which, reason);
}

/* Cancel the save box */
void
save_box_cancel_cb(Widget  w,
		   int    *tag,
		   u_long *reason)
{
  int which;   /* widget indexer */

  /* Restore save_type button */
  XmToggleButtonSetState(widgets[k_show_search_btn], show_search, TRUE);
  if (prev_save_type == SAVE_LOCAL)
    XmToggleButtonSetState(widgets[k_local_save_btn], TRUE, TRUE);
  else if (prev_save_type == SAVE_FTP)
    XmToggleButtonSetState(widgets[k_ftp_save_btn], TRUE, TRUE);
  else
    XmToggleButtonSetState(widgets[k_email_save_btn], TRUE, TRUE);

  /* restore all ftp and email fields */
  XmToggleButtonSetState(widgets[k_ftp_append_btn], ftp_append, FALSE);
  XmToggleButtonSetState(widgets[k_ftp_replace_btn], !ftp_append, FALSE);
  if (ftp_host)
    XmTextSetString(widgets[k_ftp_host_text], ftp_host);
  if (ftp_uid)
    XmTextSetString(widgets[k_ftp_uid_text], ftp_uid);
  if (ftp_pw[0]) 
    XmTextSetString(widgets[k_ftp_pw_text], prev_ftp_pw);

  if (ftp_file) 
    XmTextSetString(widgets[k_ftp_file_text], ftp_file);
  if (mail_addr)
    XmTextSetString(widgets[k_email_text], mail_addr);
  if (mail_subj)
    XmTextSetString(widgets[k_email_subj], mail_subj);

  /* clear the box */
  which = k_save_box;
  clear_dialog_tag_cb(w, &which, reason);
}

/* Put up the print_opts dialog. 
 */
void
print_opt_cb(Widget  w,
	     int    *tag,
	     u_long *reason)
{
  Widget button;

  if (widgets[k_print_opt_box] == NULL) {
    if (MrmFetchWidget(s_MrmHierarchy, 
		       "print_opt_box", 
		       search_shell,
		       &widgets[k_print_opt_box], 
		       &dummy_class) != MrmSUCCESS)
      error("can't fetch print_opt_box widget");    

      /* We don't want help button, so get it and remove it */
      button = XmSelectionBoxGetChild(widgets[k_print_opt_box], 
				      XmDIALOG_HELP_BUTTON);
      XtUnmanageChild(button);
  }
  /* Called initially with reason == NULL so widgets can get retrieved */
  if (reason) {
    map_dialog(widgets[k_print_opt_box]);
    ForceDialog(widgets[k_print_opt_box]);
    XmProcessTraversal(XmFileSelectionBoxGetChild(widgets[k_print_opt_box],
		       XmDIALOG_TEXT),XmTRAVERSE_CURRENT);
    XmProcessTraversal(XmFileSelectionBoxGetChild(widgets[k_print_opt_box],
		       XmDIALOG_TEXT),XmTRAVERSE_CURRENT);
  }
}

/* Put up the connection_opts dialog. 
 */
void
connection_opt_cb(Widget  w,
		  int    *tag,
		  u_long *reason)
{
  /* first time we get called fetch the connection dialog */
  if (!widgets[k_conn_options_box]) {
    if (MrmFetchWidget(s_MrmHierarchy, 
		       "conn_options_box", 
		       search_shell,
		       &widgets[k_conn_options_box], 
		       &dummy_class) != MrmSUCCESS)
      error("can't fetch connection_options_box widget");
    
  }
  else {
    map_dialog(widgets[k_conn_options_box]);
    ForceDialog(widgets[k_conn_options_box]);
    XmProcessTraversal(top_conn_opt_text, XmTRAVERSE_CURRENT);
    XmProcessTraversal(top_conn_opt_text, XmTRAVERSE_CURRENT);
  }
}

/* Spin the globe by one image. Yes, this is an unbearably ugly way to
   do this, but it seems to be the most efficient. There is no time spent 
   at start-up parsing files, and we only need a single 1-plane pixmap
   on the server.
 */
static void
spin_globe(int tag)
{
  static int idex = 1;
  Pixmap buf_pix;  /* Intermediate pixmap */

  if (idex == 1)
    buf_pix = XCreateBitmapFromData(
	        XtDisplay(search_shell), XtWindow(search_shell), 
	        globe1_bits, GLOBE_WIDTH, GLOBE_HEIGHT);
  else if (idex == 2)
    buf_pix = XCreateBitmapFromData(
	        XtDisplay(search_shell), XtWindow(search_shell), 
	        globe2_bits, GLOBE_WIDTH, GLOBE_HEIGHT);
  else if (idex == 3)
    buf_pix = XCreateBitmapFromData(
	        XtDisplay(search_shell), XtWindow(search_shell), 
	        globe3_bits, GLOBE_WIDTH, GLOBE_HEIGHT);
  else if (idex == 4)
    buf_pix = XCreateBitmapFromData(
	        XtDisplay(search_shell), XtWindow(search_shell), 
	        globe4_bits, GLOBE_WIDTH, GLOBE_HEIGHT);
  else if (idex == 5)
    buf_pix = XCreateBitmapFromData(
	        XtDisplay(search_shell), XtWindow(search_shell), 
	        globe5_bits, GLOBE_WIDTH, GLOBE_HEIGHT);
  else if (idex == 6)
    buf_pix = XCreateBitmapFromData(
	        XtDisplay(search_shell), XtWindow(search_shell), 
	        globe6_bits, GLOBE_WIDTH, GLOBE_HEIGHT);
  else if (idex == 7)
    buf_pix = XCreateBitmapFromData(
	        XtDisplay(search_shell), XtWindow(search_shell), 
	        globe7_bits, GLOBE_WIDTH, GLOBE_HEIGHT);
  else if (idex == 8)
    buf_pix = XCreateBitmapFromData(
	        XtDisplay(search_shell), XtWindow(search_shell), 
	        globe8_bits, GLOBE_WIDTH, GLOBE_HEIGHT);
  else if (idex == 9)
    buf_pix = XCreateBitmapFromData(
	        XtDisplay(search_shell), XtWindow(search_shell), 
	        globe9_bits, GLOBE_WIDTH, GLOBE_HEIGHT);
  else if (idex == 10)
    buf_pix = XCreateBitmapFromData(
	        XtDisplay(search_shell), XtWindow(search_shell), 
	        globe10_bits, GLOBE_WIDTH, GLOBE_HEIGHT);
  else if (idex == 11)
    buf_pix = XCreateBitmapFromData(
	        XtDisplay(search_shell), XtWindow(search_shell), 
	        globe11_bits, GLOBE_WIDTH, GLOBE_HEIGHT);
  else if (idex == 12)
    buf_pix = XCreateBitmapFromData(
	        XtDisplay(search_shell), XtWindow(search_shell), 
	        globe12_bits, GLOBE_WIDTH, GLOBE_HEIGHT);
  else if (idex == 13)
    buf_pix = XCreateBitmapFromData(
	        XtDisplay(search_shell), XtWindow(search_shell), 
	        globe13_bits, GLOBE_WIDTH, GLOBE_HEIGHT);
  else if (idex == 14)
    buf_pix = XCreateBitmapFromData(
	        XtDisplay(search_shell), XtWindow(search_shell), 
	        globe14_bits, GLOBE_WIDTH, GLOBE_HEIGHT);
  else if (idex == 15)
    buf_pix = XCreateBitmapFromData(
	        XtDisplay(search_shell), XtWindow(search_shell), 
	        globe15_bits, GLOBE_WIDTH, GLOBE_HEIGHT);
  else if (idex == 16)
    buf_pix = XCreateBitmapFromData(
	        XtDisplay(search_shell), XtWindow(search_shell), 
	        globe16_bits, GLOBE_WIDTH, GLOBE_HEIGHT);
  else if (idex == 17)
    buf_pix = XCreateBitmapFromData(
	        XtDisplay(search_shell), XtWindow(search_shell), 
	        globe17_bits, GLOBE_WIDTH, GLOBE_HEIGHT);
  else if (idex == 18)
    buf_pix = XCreateBitmapFromData(
	        XtDisplay(search_shell), XtWindow(search_shell), 
	        globe18_bits, GLOBE_WIDTH, GLOBE_HEIGHT);
  else if (idex == 19)
    buf_pix = XCreateBitmapFromData(
	        XtDisplay(search_shell), XtWindow(search_shell), 
	        globe19_bits, GLOBE_WIDTH, GLOBE_HEIGHT);
  else if (idex == 20)
    buf_pix = XCreateBitmapFromData(
	        XtDisplay(search_shell), XtWindow(search_shell), 
	        globe20_bits, GLOBE_WIDTH, GLOBE_HEIGHT);
  else if (idex == 21)
    buf_pix = XCreateBitmapFromData(
	        XtDisplay(search_shell), XtWindow(search_shell), 
	        globe21_bits, GLOBE_WIDTH, GLOBE_HEIGHT);
  else if (idex == 22)
    buf_pix = XCreateBitmapFromData(
	        XtDisplay(search_shell), XtWindow(search_shell), 
	        globe22_bits, GLOBE_WIDTH, GLOBE_HEIGHT);
  else if (idex == 23)
    buf_pix = XCreateBitmapFromData(
	        XtDisplay(search_shell), XtWindow(search_shell), 
	        globe23_bits, GLOBE_WIDTH, GLOBE_HEIGHT);
  else if (idex == 24)
    buf_pix = XCreateBitmapFromData(
	        XtDisplay(search_shell), XtWindow(search_shell), 
	        globe24_bits, GLOBE_WIDTH, GLOBE_HEIGHT);
  else if (idex == 25)
    buf_pix = XCreateBitmapFromData(
	        XtDisplay(search_shell), XtWindow(search_shell), 
	        globe25_bits, GLOBE_WIDTH, GLOBE_HEIGHT);
  else if (idex == 26)
    buf_pix = XCreateBitmapFromData(
	        XtDisplay(search_shell), XtWindow(search_shell), 
	        globe26_bits, GLOBE_WIDTH, GLOBE_HEIGHT);
  else if (idex == 27)
    buf_pix = XCreateBitmapFromData(
	        XtDisplay(search_shell), XtWindow(search_shell), 
	        globe27_bits, GLOBE_WIDTH, GLOBE_HEIGHT);
  else if (idex == 28)
    buf_pix = XCreateBitmapFromData(
	        XtDisplay(search_shell), XtWindow(search_shell), 
	        globe28_bits, GLOBE_WIDTH, GLOBE_HEIGHT);
  else if (idex == 29)
    buf_pix = XCreateBitmapFromData(
	        XtDisplay(search_shell), XtWindow(search_shell), 
	        globe29_bits, GLOBE_WIDTH, GLOBE_HEIGHT);
  else if (idex == 30)
    buf_pix = XCreateBitmapFromData(
	        XtDisplay(search_shell), XtWindow(search_shell), 
	        globe30_bits, GLOBE_WIDTH, GLOBE_HEIGHT);


  XCopyPlane(XtDisplay(search_shell), buf_pix, 
	     XtWindow(widgets[tag]), gc, 0, 0, 
	     GLOBE_WIDTH, GLOBE_HEIGHT, 0, 0, 1);
  XFreePixmap(XtDisplay(search_shell), buf_pix);

  idex++;
  if (idex > NUM_GLOBES )
    idex = 1;
  if (spin_delay)
    globe_timer_id = XtAppAddTimeOut(app_context, spin_delay, 
				     spin_globe, (XtPointer) tag);
}

/* Initialize globe dialog box
 */

void
load_globes(int tag)
{     
  Arg    args[10];
  int    n;
  XGCValues values;
  XWindowAttributes wa;

  XGetWindowAttributes(XtDisplay(search_shell), 
		       XtWindow(search_shell), &wa);
  depth = wa.depth;
  
  /* Fetch the globe box */
  if (MrmFetchWidget(s_MrmHierarchy, 
		     "globe_box", 
		     tag == k_search_globe_box ? search_shell : summary_shell,
		     &widgets[tag], 
		     &dummy_class) != MrmSUCCESS)
    error("can't fetch globe widget");

  /* Get fore and background */
  n = 0;
  XtSetArg(args[n], XmNbackground, &values.background);n++;
  XtSetArg(args[n], XmNforeground, &values.foreground);n++;
  XtGetValues(widgets[tag], args, n);

  /* Get the graphics context, used globally */
  if (!gc)
    gc = XtGetGC(search_shell, GCForeground | GCBackground, &values); 
}

/* Pop up the globe dialog, used during searching */
void
pop_globe(int tag)
{
  map_dialog(widgets[tag]);
  ForceDialog(widgets[tag]);
  XtSetSensitive(widgets[tag == k_search_globe_box ? k_search_cancel 
			                           : k_full_cancel], TRUE);
  spin_globe(tag == k_search_globe_box ? k_search_globe_image
	     : k_full_globe_image);
}

/* Clear the globe box */
void
clear_globe(int tag)
{
  XtUnmanageChild(widgets[tag]);
  if (globe_timer_id){
    XtRemoveTimeOut(globe_timer_id);
    globe_timer_id = NULL;
  }
}

/* Halt globe */
void
halt_globe(void)
{
  if (globe_timer_id) {
    XtRemoveTimeOut(globe_timer_id);
    globe_timer_id = NULL;
  }
}

static void
spin_about(void)
{
  int status;
  static int idex = 0;
  int width, height;
  char about_name[MAX_LINE];
  Pixmap bitmap;

  if (idex == 26) {
    idex = 0;
    finish_about();
    return;
  }
  if (idex == 0) {
    /* Put up the right half first */
    sprintf(about_name, "%s%s%s", willow_dir, BITMAP_DIR, ABOUT_NAME);
    status = read_bitmap_file(about_name, &width, &height, &bitmap);

    if (status != BitmapSuccess) {
      pop_warning("Could not open about-bitmap: file %s", about_name, search_shell);
      return;
    }    
    XCopyPlane(XtDisplay(search_shell), bitmap,
	       XtWindow(widgets[k_about_image]), gc, 387, 0, 
	       223, height, 387, 0, 1);
    XFreePixmap(XtDisplay(search_shell), bitmap);
  }
  else {
    sprintf(about_name, "%s%sabout%d", willow_dir, BITMAP_DIR, idex+1);
    status = read_bitmap_file(about_name, &width, &height, &bitmap);

    if (status != BitmapSuccess) {
      pop_warning("Could not open 'about': %s", about_name, search_shell);
      return;
    }    
    XCopyPlane(XtDisplay(search_shell), bitmap,
	       XtWindow(widgets[k_about_image]), gc, 0, 0, 
	       width, height, 0, 0, 1);
  
    XFreePixmap(XtDisplay(search_shell), bitmap);
  }
  idex++;
  /* it takes a long time to read each frame, so lower delay */
  about_timer_id = XtAppAddTimeOut(app_context, 
	               spin_delay, spin_about, NULL);
}

static void
finish_about(void)
{
  int   width, height, status;
  Pixmap  bitmap;
  char bm_file[MAX_LINE];

  if (!about_pix) {
    sprintf(bm_file, "%s%s%s", willow_dir, BITMAP_DIR, ABOUT_NAME);
    status = read_bitmap_file(bm_file, &width, &height, &bitmap);
    if (status != BitmapSuccess) {
      pop_warning("Could not open about-bitmap: file %s", bm_file, search_shell);
      return;
    }    
    
    about_pix = XCreatePixmap(XtDisplay(search_shell), XtWindow(search_shell),
			 width, height, depth);
    XCopyPlane(XtDisplay(search_shell), bitmap, about_pix,
	       gc, 0, 0, width, height, 0, 0, (unsigned long) 1);
    XFreePixmap(XtDisplay(search_shell), bitmap);
  }
  XtVaSetValues(widgets[k_about_image], XmNlabelType, XmPIXMAP,
		XmNlabelPixmap, about_pix, NULL);
}

/* Read a bitmap file in our compressed form off the disk */
static int
read_bitmap_file(char *name,
		 int  *width,
		 int  *height,
		 Pixmap *bitmap)
{
  FILE *bmfile;
  char bmident[4];
  int  n;
  char *data;
  char gz_name[1024];
  char cmd[1024];
  struct stat stat_dummy;

   /* Check if the file has .gz extension, then read file as 
      from forked gunzip, otherwise just open the file */
  strcat(strcpy(gz_name, name), ".gz");

  if (stat(gz_name, &stat_dummy) == 0){
    sprintf(cmd, "gunzip -c %s", gz_name);
    if (!(bmfile=popen(cmd, "r")))
      return(BitmapOpenFailed);
  }
  else if (!(bmfile=fopen(name,"r")))
    return(BitmapOpenFailed);

  fread(bmident,4,1,bmfile);
  if (strncmp(bmident,"*BM*",4))
    return(BitmapFileInvalid);

  
  fread(width,4,1,bmfile);
  fread(height,4,1,bmfile);
  fread(&n,4,1,bmfile);

  *width = ntohl(*width);
  *height = ntohl(*height);
  n = ntohl(n);

  if (!(data=(char*)XtMalloc(n))) 
    return(BadAlloc);

  fread(data,1,n,bmfile);
  fclose(bmfile);

  *bitmap = XCreateBitmapFromData(XtDisplay(search_shell),
         XtWindow(search_shell), data, *width, *height);

  if (!*bitmap) {
    XtFree(data);
    return(BitmapFileInvalid);
  }
  XtFree(data);
  return(BitmapSuccess);
}

/* Pop the dialog for configuring a record retrieval */
void 
pop_retrieve_dlg_cb(Widget w, int *tag, u_long *reason)
{
  if (widgets[k_retrieve_dlg] == NULL) {
    if (MrmFetchWidget(s_MrmHierarchy, 
		       "retrieve_box", 
		       search_shell,
		       &widgets[k_retrieve_dlg], 
		       &dummy_class) != MrmSUCCESS)
      error("can't fetch Retrieval dialog widget");    

    XtVaSetValues(XtParent(widgets[k_retrieve_dlg]), XmNtitle, 
			   "Record Format Options", NULL);
  }
  /* Don't pop it the first time we are called, during initialization */
  else {
    XmToggleButtonSetState(widgets[k_long_label], !abbrev_fields, FALSE);
    XmToggleButtonSetState(widgets[k_short_label], abbrev_fields, FALSE);
    map_dialog(widgets[k_retrieve_dlg]);
    ForceDialog(widgets[k_retrieve_dlg]);
  }
}

/* Pop the dialog for selecting from multiple browse */
void 
pop_mbrowse(Boolean map_it)
{
  if (!map_it){
    if (MrmFetchWidget(s_MrmHierarchy, 
		       "mbrowse_box", 
		       search_shell,
		       &widgets[k_mbrowse_box], 
		       &dummy_class) != MrmSUCCESS)
      error("can't fetch mbrowse dialog widget");    
    XtVaSetValues(XtParent(widgets[k_mbrowse_box]), XmNtitle, 
			   "Select Browse Method", NULL);
  }    
  else {
    map_dialog(widgets[k_mbrowse_box]);
    ForceDialog(widgets[k_mbrowse_box]);
  }
}

void retrieve_opts_ok_cb(Widget w, int *tag, u_long *reason)
{
  clear_dialog_cb(NULL, widgets[k_retrieve_dlg], NULL);
  if (abbrev_fields != XmToggleButtonGetState(widgets[k_short_label])) {
    abbrev_fields = !abbrev_fields;
    clear_summ_bodies();
  }
  field_pick_ok();
}

void retrieve_opts_cancel_cb(Widget w, int *tag, u_long *reason)
{
  if (widgets[k_retrieve_dlg]) {
    XmToggleButtonSetState(widgets[k_long_label], !abbrev_fields, FALSE);
    XmToggleButtonSetState(widgets[k_short_label], abbrev_fields, FALSE);
    clear_dialog_cb(NULL, widgets[k_retrieve_dlg], NULL);
  }
  field_pick_cancel();
}

/* Set-up for animating the about box */
void
about_cb(Widget  w,
	 int    *tag,
	 u_long *reason)
{
  if (widgets[k_about_box] == NULL) {
    if (MrmFetchWidget(s_MrmHierarchy, 
		       "about_box", 
		       search_shell,
		       &widgets[k_about_box], 
		       &dummy_class) != MrmSUCCESS)
      error("can't fetch about-box widget");
    
  }
  else 
    XtVaSetValues(widgets[k_about_image], XmNlabelType, XmSTRING, 
		  XmNlabelString, XmStringCreateSimple(""), NULL);

  map_dialog(widgets[k_about_box]);
  XtMapWidget(widgets[k_about_ok]);

  if (spin_delay)
    about_timer_id = XtAppAddTimeOut(app_context, spin_delay, 
				     spin_about, NULL);
  else
    finish_about();
}

/* Center first dialog over second shell */
void center_dialog(Widget dialog, Widget shell)
{
  geom_rec dia_geom, shell_geom;

  while (!XtIsShell(dialog))
    dialog = XtParent(dialog);
  while (!XtIsShell(shell))
    dialog = XtParent(shell);

  save_geometry(dialog, &dia_geom);
  save_geometry(shell, &shell_geom);
  dia_geom.x = shell_geom.x + (shell_geom.w - dia_geom.w)/2;
  if (dia_geom.x < 0)
    dia_geom.x = 20;
  dia_geom.y = shell_geom.y + (shell_geom.h - dia_geom.h)/2;
  if (dia_geom.y < 0)
    dia_geom.y = 20;
  set_geometry(dialog, &dia_geom);
}

/* Record whatever geometry a shell came up in */
void
save_geometry(Widget    shell,
	      geom_rec *geom)
{
  XtVaGetValues(shell, XmNwidth, &geom->w, XmNheight, &geom->h, 
		XmNx, &geom->x, XmNy, &geom->y, NULL);
}

/* Restore shell geometry */
void
set_geometry(Widget    shell,
	     geom_rec *geom)
{
  XtVaSetValues(shell, XmNwidth, geom->w, XmNheight, geom->h, NULL);
  XtMoveWidget(shell, geom->x, geom->y);
}

/* Make sure dialog is managed and mapped */
void
map_dialog(Widget dialog)
{
  XtManageChild(dialog);
  while (!XtIsShell(dialog))
    dialog = XtParent(dialog);
  XtPopup(dialog, XtGrabNone);

}

/*
 * This procedure will ensure that, if a dialog window is being mapped,
 * its contents become visible before returning.  It is intended to be
 * used just before a bout of computing that doesn't service the display.
 * You should still call XmUpdateDisplay() at intervals during this
 * computing if possible.
 *
 * The monitoring of window states is necessary because attempts to map
 * the dialog are redirected to the window manager (if there is one) and
 * this introduces a significant delay before the window is actually mapped
 * and exposed.  This code works under mwm, twm, uwm, and no-wm.  It
 * doesn't work with olwm if the mainwindow is iconified.
 *
 * The argument to ForceDialog is any widget in the dialog (usually it
 * will be the BulletinBoard child of a DialogShell).
 * 
 *  dbrooks@osf.org
 */

void
ForceDialog(Widget w)
{
  Widget diashell, topshell;
  Window diawindow, topwindow;
  Display *dpy;
  XWindowAttributes xwa;
  XtAppContext cxt;

/* Locate the shell we are interested in */
  for (diashell = w;
       !XtIsShell(diashell);
       diashell = XtParent(diashell))
    ;

/* Locate its primary window's shell (which may be the same) */
  for (topshell = diashell;
       !XtIsTopLevelShell(topshell);
       topshell = XtParent(topshell))
    ;

  if (XtIsRealized(diashell) && XtIsRealized(topshell)) {
    dpy = XtDisplay(topshell);
    diawindow = XtWindow(diashell);
    topwindow = XtWindow(topshell);
    cxt = XtWidgetToApplicationContext(diashell);

/* Wait for the dialog to be mapped.  It's guaranteed to become so unless... */

    while (XGetWindowAttributes(dpy, diawindow, &xwa),
	   xwa.map_state != IsViewable) {

/* ...if the primary is (or becomes) unviewable or unmapped, it's
   probably iconified, and nothing will happen. */

      if (XGetWindowAttributes(dpy, topwindow, &xwa),
	  xwa.map_state != IsViewable)
	break;

/* At this stage, we are guaranteed there will be an event of some kind.
   Beware; we are presumably in a callback, so this can recurse. */

/* For unknown reasons processing Alternate Inputs here causes trouble,
   so only process bonafide events */
      if (XtIMXEvent & XtAppPending(cxt))
	XtAppProcessEvent(cxt, XtIMXEvent);
    }
  }

/* The next XSync() will get an expose event if the dialog was unmapped.
   This can recurse also. */

  XmUpdateDisplay(topshell);
}


