/***********************************************************************
 *                                                                     *
 *              WILLOW(tm) Information Retrieval Software              *
 *     (Washington Information Looker-upper Layered Over Windows)      *
 *             Copyright 1992-1994 University of Washington            *
 *                                                                     *
 *   Willow is a trademark of the University of Washington.  No        *
 *   commercial use of this trademark or software may be made without  *
 *   prior written permission of the University of Washington.         *
 *                                                                     *
 * Permission to use, copy, modify, and distribute Willow(tm) software *
 * and its documentation for any noncommercial purpose is hereby       *
 * granted, provided that the above copyright notice appears in all    *
 * copies and that both the above copyright notice and this permission *
 * notice appear in supporting documentation, and that the name of the *
 * University of Washington not be used in advertising or publicity    *
 * pertaining to distribution of the software without specific,        *
 * written prior permission.  This software is made available "as is", *
 * and THE UNIVERSITY OF WASHINGTON DISCLAIMS ALL WARRANTIES, EXPRESS  *
 * OR IMPLIED, WITH REGARD TO THIS SOFTWARE, INCLUDING WITHOUT         *
 * LIMITATION ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS    *
 * FOR A PARTICULAR PURPOSE, AND IN NO EVENT SHALL THE UNIVERSITY OF   *
 * WASHINGTON BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL     *
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA  *
 * OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, TORT (INCLUDING       *
 * NEGLIGENCE) OR STRICT LIABILITY, ARISING OUT OF OR IN CONNECTION    *
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.                       *
 *                                                                     *
 * For further information please contact:                             *
 * Willow Project                                                      *
 * Information Systems                                                 *
 * University of Washington                                            *
 * 4545 15th NE, 1st floor                                             *
 * Seattle, Washington, 98105, USA                                     *
 * Internet: willow@cac.washington.edu                                 *
 ***********************************************************************/

/***********************************************************
 * dict.c
 * This file contains the code which implements the on-line
 * word-list dictionary. 
 ***********************************************************/

#include <stdio.h>
#ifdef SOLARIS
#include <string.h>
#else
#include <strings.h>
#endif

#include <Mrm/MrmAppl.h>
#include <Xm/Protocols.h>
#include <Xm/Text.h>
#include <Xm/PushBG.h>
#include <Xm/ScrollBar.h>
#include <Xm/RowColumn.h>
#include <Xm/ToggleB.h>
#include <Xm/TextP.h>   /* we need this to fix scrolling ugliness */

#include "willow.h"
#include "db.h"
#include "dict.h"

#include "protocol.h"

#include "text-list.h"

extern Widget dict_shell, toplevel, search_shell, widgets[];   
extern char  *willow_dir;
extern XtAppContext app_context;   /* Application Context for XtApp* calls */
extern MrmType dummy_class;         /* and class variable. */
extern MrmHierarchy s_MrmHierarchy;  /* MRM database hierarchy ID */
extern Atom wm_delete_window;
extern int interface_mode;
extern geom_rec dict_geom;
extern int  status;
extern int opac_mode;

/* This is the db that any browse requests should take place on */
static db_desc_rec *browse_db;

/* This is the current list we have loaded */
static char *cur_list = NULL;
static db_field_desc_rec *cur_field = NULL;

/* This is the corresponding text box in the Search window */
static Widget text_box;

/* This is the very large block of chars which holds the list as returned
   by the server */
static char *raw_text = NULL;

/* This is our manipulated version of the text, as displayed */
static char *cooked_text = NULL;

/* Our vertical scrollbar */
static Widget sb;

/* Total number of items in current list */
static int total_records = 0;

/* Total number fields in current list */
static int total_fields = 0;

/* Total number fields in current list which are visible */
static int total_vis_fields = 0;

/* How many lines do we display per record (not counting separator */
static int lines_per_rec;

/* Number of lines visible in our text widget */
static short visible_lines = 0;

/* Number of records visible in our text widget */
static short visible_records = 0;

/* How far down the screen do we want our hits to be? */
static int   record_offset;

/* Last value we set scrollbar to */
static int cur_scroll_value = 0;

/* Hold the descriptions of our fields */
static field_def fields[MAX_FIELDS];

/* Hold the descriptions for currently displayed records */
static record_def records[MAX_RECORDS];

/* For handling selection list */
sel_rec_def *sel_list = NULL;
static int last_sel = 0;

/* Queue of sent but still unanswered browse requests */
browse_q_rec *browse_q = NULL, *browse_q_tail = NULL;

/* If type is SCROLL, where are we scrolling to */
int scroll_to;

/* How many milliseconds to wait before launching an incremental search */
int browse_delay;

/* Timeout Id */
XtIntervalId search_to = NULL;

/* Flag for whether we need to grow the simple screen */
Boolean need_to_grow = FALSE;

/*
 * Forward declarations
 */
void really_search_dict(char *search_str, XtIntervalId *id);
extern int list_lookup (char *db_name, char *list_name, char search_type, char *prefix_st, int lines_wanted, int line_offset);
extern char *strndup (char *s, int n);
extern int atoi (char *);
extern int clear_titles (void);
extern int error (char *);
extern int grow_opac (void);
extern int not_ready_buttons (void);
extern void pop_warning (char *, char *, Widget);
extern int pop_window (int tag);
extern int set_icon (char *icon_name, Widget shell);
extern int start_display (void);
extern void clear_dialog_cb (Widget w, Widget target, u_long *reason);
extern void ctrl_cmd_action(Widget w, XEvent *event, 
			    String *params, Cardinal *num_params);
extern void hide_window (int tag);
extern void list_delete_all (Widget w);
extern void list_deselect_all_items (Widget w);
extern void list_init (Widget w);
extern void list_register_callback (Widget w, int type, void (*proc) (/* ??? */));
extern void list_replace_all_items (Widget w, char *items, int count, Boolean has_header);
extern void list_select_pos (Widget w, int pos, Boolean toggle, Boolean call_cb);
extern void make_query_display(void);
extern void mwm_close_cb(Widget w, caddr_t client_data, caddr_t call_data);
extern void pop_globe (int tag);
extern void pop_mbrowse (Boolean map_it);
extern void save_geometry (Widget shell, geom_rec *geom);
extern void send_search_cmd (int num_text_fields, int num_limit_fields, char *data, int data_len);
int init_browse_widgets (void);
static char *pull_str(char *text, int *sdex_p, char *buf, int max);
static int adjust_scrollbar (int line);
static int highlight (int pos);
static int parse_header(char *header);
static int pull_int(char *header, int *sdex_p);
static void kb_down(Widget w, XEvent *event, String *params, Cardinal *num_params);
static void kb_p_down(Widget w, XEvent *event, String *params, Cardinal *num_params);
static void kb_p_up(Widget w, XEvent *event, String *params, Cardinal *num_params);
static void kb_up(Widget w, XEvent *event, String *params, Cardinal *num_params);
static void new_text(void);
static void set_visible_records(void);
void dict_cancel_cb(Widget w, int *tag, u_long *reason);
void dict_deselect_all_cb(Widget w, list_callback_struct *reason);
void dict_deselect_cb(Widget w, list_callback_struct *reason);
void dict_double_click_cb(Widget w, list_callback_struct *reason);
void dict_hide(void);
void dict_paste_cb(Widget w, int *tag, u_long *reason);
void dict_scroll_down_cb(Widget w, list_callback_struct *reason);
void dict_scroll_up_cb(Widget w, list_callback_struct *reason);
void dict_search_cb(Widget w, int *tag, XmAnyCallbackStruct *reason);
void dict_select_cb(Widget w, list_callback_struct *reason);
void list_kb_down(Widget w, XEvent *event, String *params, Cardinal *num_params);
void list_kb_up(Widget w, XEvent *event, String *params, Cardinal *num_params);
void opac_search_cb(Widget w, int *tag, u_long *reason);
void really_start_browse_cb(Widget w, char *list, u_long *reason);
void resize_handler(Widget w, XEvent *event, String *params, Cardinal *num_params);
void scroll_cb(Widget w, int retain_selection, XmScrollBarCallbackStruct *reason);
void start_browse_cb(Widget w, db_field_desc_rec *list, u_long *reason);


/* Put a information about an out-going browse request into the browse queue */
static void browse_queue(int ref_num, int type, int scroll_to)
{
  browse_q_rec *ptr;

  if (ref_num < 0)
    return;
  ptr = (browse_q_rec *) XtMalloc(sizeof(browse_q_rec));
  ptr->next = ptr->prev = NULL;
  ptr->browse_type = type;
  ptr->scroll_to = scroll_to;
  ptr->ref_num = ref_num;
  
  if (!browse_q)
    browse_q = browse_q_tail = ptr;
  else {
    browse_q_tail->next = ptr;
    ptr->prev = browse_q_tail;
    browse_q_tail = ptr;
  }
}

/* An find a match for an incoming browse response packet on our queue.
   Take out anything older than this request. Return the browse_type.
   Set the global scroll_to.
 */
int browse_de_q(int ref_num)
{
  browse_q_rec *ptr = NULL;
  int type;

  while (browse_q) {
    ptr = browse_q;
    browse_q = browse_q->next;
    if (ptr->ref_num == ref_num)
      break;
    else {
      XtFree(ptr);
      ptr = NULL;
    }
  }
  if (!browse_q)
    browse_q_tail = NULL;
  else 
    browse_q->prev = NULL;

  if (ptr) {
    scroll_to = ptr->scroll_to;
    type = ptr->browse_type;
    XtFree(ptr);
    return(type);
  }
  else
    return(-1);
}


/* Called when a browse result comes in from the driver */
void browse_dcb(packet_rec *packet)
{
  Boolean err = FALSE;
  int     rec;
  int     browse_type;

  /* Is this a header record we are waiting for? */
  rec = packet->var.browse.line_pos;
  if ((browse_type = browse_de_q(packet->var.browse.ref_num)) < 0)
    return;
  if (browse_type == HEADER) {
    if (rec < 0) {
      pop_warning("Could not open list \"%s\"", cur_list, search_shell);
      err = TRUE;
    }
    else if (parse_header(packet->data) < 0) {
      pop_warning("ERROR: \"%s\" List is corrupted", cur_list, search_shell);
      err = TRUE;
    }
    if (err) {
      if (!opac_mode) 
	dict_hide();
      else {
	opac_mode = 0;
	grow_opac();
      }
    }
    /* Send request for first page */
    else
      browse_queue(list_lookup(browse_db->internal_name, cur_list, 'N', "0", 
			       visible_records, record_offset),
		   FIRST, -1);
    return;
  }
  /* Otherwise this is just a normal data page */
  if (rec < 0) {
    pop_warning("Could not read list  \"%s\" from server", cur_list,
		search_shell);
    return;
  }
  if (browse_type == INCR)
    list_deselect_all_items(widgets[k_dict_text]);

  raw_text = strndup(packet->data, packet->data_len);
  new_text();

  /* See if we are supposed to highlight this match */
  if ((browse_type == INCR) || (browse_type == FIRST)) {
    if (rec > record_offset)
      adjust_scrollbar(rec - record_offset);
    else 
      adjust_scrollbar(0);
  }
  if (browse_type == INCR) {
    /* If match is within first few of list, need special handling */
    if (rec < record_offset)
      highlight(rec);
    else
      highlight(record_offset);
  }
  if (browse_type == SCROLL)
    adjust_scrollbar(scroll_to);
  if (need_to_grow) {
    need_to_grow = FALSE;
    grow_opac();
  }
}

void really_start_browse_cb(Widget w, char *list, u_long *reason)
{
  WidgetList  buttons;
  int         num_children, i;

  /* Did list get passed in, or did we get a real callback which requires
     checking all the toggle buttons? */
  if (w) {
    XtVaGetValues(widgets[k_mbrowse_radio], XmNnumChildren, &num_children,
		  XmNchildren, &buttons, NULL);
    for (i = 0; i < num_children; i++)
      if (XmToggleButtonGetState(buttons[i]))
	break;
    XtVaGetValues(buttons[i], XmNuserData, &list, NULL);
    XtVaSetValues(widgets[k_lookup_label], XmNlabelString, 
		  XmStringCreateSimple(list), NULL);
  }
  else
    XtVaSetValues(widgets[k_lookup_label], XmNlabelString, 
		  XmStringCreateSimple(cur_field->name), NULL);

  if (opac_mode) {
    opac_mode = 1;
    grow_opac();
    XtRemoveAllCallbacks(widgets[k_dict_search], XmNvalueChangedCallback);
    XmTextSetString(widgets[k_dict_search], "");
    if (list)
      XtAddCallback(widgets[k_dict_search], XmNvalueChangedCallback,
		    dict_search_cb, NULL);
  }
  else
    pop_window(k_dict_form);


  list_deselect_all_items(widgets[k_dict_text]);

  XtUnmanageChild(widgets[k_dict_search]);
  XtManageChild(widgets[k_dict_search]);
  XmProcessTraversal(widgets[k_dict_search], XmTRAVERSE_CURRENT);
  XmProcessTraversal(widgets[k_dict_search], XmTRAVERSE_CURRENT);

  /* Might be a Keywords search in opac mode */
  if (!list) {
    cur_list = NULL;
    return;
  }
  else
    cur_list = list;

  /* Send request for header record for this list */
  browse_queue(list_lookup(browse_db->internal_name,cur_list, 'H', "0", 1, 0),
	       HEADER, -1);
  XmTextSetString(widgets[k_dict_text], "");
}

/* Callback to do an incremental search, triggered by any change in the
   search-text window. 
 */
void dict_search_cb(Widget w, int *tag, XmAnyCallbackStruct *reason)
{
  char *search_str;

  if (!(browse_db && cur_list))
    return;
  if (search_to)
    XtRemoveTimeOut(search_to);

  search_str = XmTextGetString(widgets[k_dict_search]);
  if (browse_delay)
    search_to = XtAppAddTimeOut(app_context, browse_delay, 
				really_search_dict, strdup(search_str));
  else {
    search_to = NULL;
    really_search_dict(strdup(search_str), NULL);
  }
  XtFree(search_str);
}

void really_search_dict(char *search_str, XtIntervalId *id)
{
  int len;

  search_to = NULL;
  len = strlen(search_str);
  if (len)
    browse_queue(list_lookup(browse_db->internal_name, cur_list, 'C', 
			     search_str, visible_records, record_offset),
		 INCR, -1);
  else
    browse_queue(list_lookup(browse_db->internal_name, cur_list, 'N', "0", 
			     visible_records, record_offset),
		 INCR, -1);
  if (opac_mode == 2)
    need_to_grow = TRUE;
  XtFree(search_str);
}

void scroll_cb(Widget w, int dummy, XmScrollBarCallbackStruct *reason)
{
  char line_number[16];
  browse_q_rec *ptr;
  int diff;

  switch (reason->reason) {
    case XmCR_DECREMENT:
      diff = -1; break;
    case XmCR_INCREMENT: 
      diff = 1;  break;
    case XmCR_PAGE_DECREMENT:
      diff = 0 - (visible_records - 1); break;
    case XmCR_PAGE_INCREMENT: 
      diff = visible_records - 1; break;
    default:
      diff = 0;
  }

  /* Check if there are any pending scroll-requests. If so, we want to
     use it as the basis of next request. */
  if (diff != 0)
    for (ptr = browse_q_tail; ptr; ptr = ptr->prev) {
      if (ptr->browse_type == SCROLL)
	break;
      else if (ptr->browse_type == INCR)
	/* Scroll following an unanswered word search, don't know where
	   base scroll offset from, so drop it. */
	return;
    }

  if (ptr && diff ) {
    /* There was a pending scroll */
    reason->value = ptr->scroll_to + diff;
    if (reason->value < 0)
      reason->value = 0;
    if (reason->value > total_records)
      reason->value = total_records;
  }

  sprintf(line_number, "%d", reason->value);
  browse_queue(list_lookup(browse_db->internal_name, cur_list, 'N', 
			   line_number, visible_records, 0),
	       SCROLL, reason->value);
}

/* Handle a resize of the text widget */
void resize_handler(Widget w, XEvent *event, 
		  String *params, Cardinal *num_params)
{
  char line_number[32];

  if (!(browse_db && cur_list))
    return;
  if (!cur_list)
    return;
  if (opac_mode && (opac_mode < 3))
    return;
  sprintf(line_number, "%d", last_sel + cur_scroll_value);

  XtVaGetValues(widgets[k_dict_text], XmNrows, &visible_lines, NULL);
  set_visible_records();
  browse_queue(list_lookup(browse_db->internal_name, cur_list, 'N', 
			   line_number, visible_records, record_offset),
	       INCR, -1);
}


pop_dict(void)
{
  Boolean do_init;

  do_init = !dict_shell;
  if (do_init) {
    dict_shell = XtVaCreatePopupShell("List Browser", topLevelShellWidgetClass,
				      toplevel, NULL);
    XmAddWMProtocolCallback(dict_shell, wm_delete_window,
			    mwm_close_cb, NULL);
  }
  /* Deal with UIL to get the widgets set up */
  if (MrmFetchWidget(s_MrmHierarchy, 
		     "dict_window", 
		     dict_shell,
		     &widgets[k_dict_form], 
		     &dummy_class) != MrmSUCCESS)
    error("can't fetch dict window widget");



  /* Manage the main part and realize everything. */
  XtManageChild(widgets[k_dict_form]);

  if (do_init) {
    XtRealizeWidget(dict_shell);
    set_icon("subjects.icon", dict_shell);
    save_geometry(dict_shell, &dict_geom);
  }
  init_browse_widgets();
  cooked_text = XtMalloc(TEXT_SIZE);
}


/* Initialize all the callbacks on our vertical scrollbar so we can 
   handle scrolling.
 */
init_browse_widgets(void)
{
  XtTranslations trans;
  XtActionsRec actions[20];
  int  n;

  XtUnmanageChild(widgets[k_opac_search]);
  XtManageChild(widgets[k_dict_cancel]);
  XmProcessTraversal(widgets[k_dict_search], XmTRAVERSE_CURRENT);
  XmProcessTraversal(widgets[k_dict_search], XmTRAVERSE_CURRENT);
  
  XtVaGetValues(XtParent(widgets[k_dict_text]), XmNverticalScrollBar,
			 &sb, NULL);
  XtRemoveAllCallbacks(sb, XmNvalueChangedCallback);
  XtRemoveAllCallbacks(sb, XmNdragCallback);
  XtRemoveAllCallbacks(sb, XmNincrementCallback);
  XtRemoveAllCallbacks(sb, XmNdecrementCallback);
  XtRemoveAllCallbacks(sb, XmNpageIncrementCallback);
  XtRemoveAllCallbacks(sb, XmNpageDecrementCallback);
  XtRemoveAllCallbacks(sb, XmNtoTopCallback);
  XtRemoveAllCallbacks(sb, XmNtoBottomCallback);

  XtAddCallback(sb, XmNvalueChangedCallback, scroll_cb, NULL);
  XtAddCallback(sb, XmNincrementCallback,    scroll_cb, NULL);
  XtAddCallback(sb, XmNdecrementCallback,    scroll_cb, NULL);
  XtAddCallback(sb, XmNpageIncrementCallback,scroll_cb, NULL);
  XtAddCallback(sb, XmNpageDecrementCallback,scroll_cb, NULL);
  XtAddCallback(sb, XmNtoTopCallback,        scroll_cb, NULL);
  XtAddCallback(sb, XmNtoBottomCallback,     scroll_cb, NULL);
  XtVaGetValues(widgets[k_dict_text], XmNrows, &visible_lines, NULL);

  /* Set up actions and translations for return and keyboard movements */
  n = 0;
  actions[n].string = "kb_p_up";
  actions[n].proc = kb_p_up; n++;
  actions[n].string = "kb_p_down";
  actions[n].proc = kb_p_down; n++;
  actions[n].string = "dict_resize";
  actions[n].proc = resize_handler; n++;
  actions[n].string = "kb_up";
  actions[n].proc = kb_up; n++;
  actions[n].string = "kb_down";
  actions[n].proc = kb_down; n++;
  actions[n].string = "ctrl_cmd_action";
  actions[n].proc = ctrl_cmd_action; n++;
  XtAppAddActions(app_context, actions, n);
  
  trans = XtParseTranslationTable(
     "<Key> osfPageUp: kb_p_up()\n\
      <Key> osfPageDown: kb_p_down()\n\
      <ConfigureNotify>: dict_resize()\n\
     ");
  XtOverrideTranslations(widgets[k_dict_text], trans);
  list_init(widgets[k_dict_text]);
  list_register_callback(widgets[k_dict_text], SELECT_CB, 
			 dict_select_cb);
  list_register_callback(widgets[k_dict_text], CLICK2_CB, 
			 dict_double_click_cb);
  list_register_callback(widgets[k_dict_text], SCROLL_UP_CB,
			 dict_scroll_up_cb);
  list_register_callback(widgets[k_dict_text], SCROLL_DN_CB,
			 dict_scroll_down_cb);
  list_register_callback(widgets[k_dict_text], DESELECT_CB,
			 dict_deselect_cb);
  list_register_callback(widgets[k_dict_text], DESELECT_ALL_CB,
			 dict_deselect_all_cb);

  trans = XtParseTranslationTable(
     "<Key> osfUp: kb_up()\n\
      Ctrl<Key>p: kb_up()\n\
      <Key> osfDown: kb_down()\n\
      Ctrl<Key>n: kb_down()\n\
      <Key> osfPageUp: kb_p_up()\n\
      <Key> osfPageDown: kb_p_down()\n\
      <Key> Return: activate()");
  XtOverrideTranslations(widgets[k_dict_search], trans);

  /* Now do basic/advanced specific translations for ctrl-commands */
  if (opac_mode)
    trans = XtParseTranslationTable(
       "Ctrl<Key>v: ctrl_cmd_action(basic,help)\n\
        <Key>F1: ctrl_cmd_action(basic,help)\n\
       ");
  else
    trans = XtParseTranslationTable(
       "Ctrl<Key>v: ctrl_cmd_action(browse,help)\n\
        <Key>F1: ctrl_cmd_action(browse,help)\n\
        Ctrl<Key>c: ctrl_cmd_action(browse,close)\n\
       ");
  XtOverrideTranslations(widgets[k_dict_search], trans);

  /* This stops the text widget from mucking with the scrollbar. 
     Hopefully there are no other side-effects. And hopefully this
     will always work. */
  ((XmTextWidget) widgets[k_dict_text])->text.vsbar_scrolling = TRUE;
}


/* Set which database we are browsing (may be different from cur_db) */
init_browse(db_desc_rec *db)
{
  browse_db = db;
}

void dict_paste_cb(Widget w, int *tag, u_long *reason)
{
  XmTextPosition  pos;
  char            buf[MAX_LINE * 10];
  char           *ptr;
  int             len = 0;
  Boolean         need_quotes = FALSE;
  sel_rec_def    *selected;

  if (opac_mode) {
    opac_search_cb(w, tag, reason);
    return;
  }

  if (sel_list) {
    for (selected = sel_list; selected != NULL; selected = selected->next) {
      /* if this is not the first, put in operator */
      if (len) {
	strncpy(buf + len, DFLT_OP, strlen(DFLT_OP));
	len += strlen(DFLT_OP);
      }
      /* Ugly, but we have to do one pass to see if quotes are needed */
      for (ptr = selected->ret_value; (*ptr); ptr++) 
	if (*ptr == ' ') {
	  need_quotes = TRUE;
	  buf[len++] = '"';
	  break;
	}
      /* Now copy characters in */
      if ((len + strlen(selected->ret_value)) > MAX_LINE)
	goto too_long;
      for (ptr = selected->ret_value; (*ptr); ptr++) 
	buf[len++] = *ptr;
      /* add right paren if needed */
      if (need_quotes) {
	buf[len++] = '"';
	need_quotes = FALSE;
      }
    }
    buf[len] = 0;
    
    pos = XmTextGetLastPosition(text_box);
    if ((pos + len + strlen(DFLT_OP)) >= MAX_LINE) {
too_long:    pop_warning("Total length of your selections is too long", 
	     NULL, dict_shell);
      return;
    }
    if (pos) {
      XmTextReplace(text_box, pos, pos, DFLT_OP);
      pos += strlen(DFLT_OP);
    }
    XmTextReplace(text_box, pos, pos, buf);
  }
  dict_hide();
}

void dict_hide(void)
{
  XtRemoveAllCallbacks(widgets[k_dict_search], XmNvalueChangedCallback);
  XmTextSetString(widgets[k_dict_search], "");
  XtAddCallback(widgets[k_dict_search], XmNvalueChangedCallback,
		dict_search_cb, NULL);
  if (widgets[k_dict_form])
    hide_window(k_dict_form);
  cur_list = NULL;
  list_deselect_all_items(widgets[k_dict_text]);
}

void dict_cancel_cb(Widget w, int *tag, u_long *reason)
{
  list_deselect_all_items(widgets[k_dict_text]);
  dict_hide();
}


extern void start_browse_key(Widget w, XEvent *event, 
			     String *params, Cardinal *num_params);

/* Triggered by List buttons on Search window. If more than one list 
   defined for this field, pop up the picker, otherwise just start the
   browse 
 */
void start_browse_cb(Widget w, db_field_desc_rec *field, u_long *reason)
{
  WidgetList  buttons;
  Widget      button;
  int         num_children;
  int         i;
  char        label_buf[MAX_LINE];
  Arg         args[10];
  int         n;
  static db_field_desc_rec *prev_field = NULL;

  if (widgets[k_mbrowse_box])
    clear_dialog_cb(NULL, widgets[k_mbrowse_box], NULL);

  cur_field = field;
  XtVaGetValues(w, XmNuserData, &text_box, NULL);
  if (field->list_cnt > 1) {
    if (!widgets[k_mbrowse_box])
      pop_mbrowse(NULL);

    if (field != prev_field) {
      prev_field = field;
      sprintf(label_buf, MBROWSE_FMT, field->name);
      XtVaSetValues(widgets[k_mbrowse_label], XmNlabelString,
		    XmStringCreateSimple(label_buf), NULL);
      XtVaGetValues(widgets[k_mbrowse_radio], XmNnumChildren, &num_children,
		    XmNchildren, &buttons, NULL);
      for (i = 0; i < num_children; i++)
	XtDestroyWidget(buttons[i]);
      
      for (i = 0; i < field->list_cnt; i++) {
	n = 0;
	if (i == 0) {
	  XtSetArg(args[n], XmNset, TRUE);
	  n++;
	}
	
	XtSetArg(args[n], XmNlabelString, 
		 XmStringCreateSimple(field->list_names[i]));n++;
	XtSetArg(args[n], XmNuserData, field->list_names[i]);n++;
	XtSetArg(args[n], XmNtraversalOn, FALSE);n++;
	button = XmCreateToggleButton(widgets[k_mbrowse_radio], "toggle",
				      args, n);
	XtManageChild(button);
      }
    }
    pop_mbrowse(TRUE);
  }
  else
    really_start_browse_cb(NULL, field->list_names[0], NULL);
}


/* Used to go through the list of fields in display order, which can
   be different from the order in which the fields come from the server.
   Passed in either -1 to start, or the last value returned.
 */
static int next_index(int match_index)
{
  int match_val;

  if (match_index == -1)
    match_val = 1;
  else
    match_val = fields[match_index].pos + 1;
  for (match_index = 0; match_index < total_fields; match_index++)
    if (fields[match_index].pos == match_val)
      break;
  return(match_index);
}

/* Display new text from server. Take a parameter indicating which record
   our hit is in the list */
static void new_text(void)
{
  XmTextPosition   sdex = 0, tdex = 0;   /* source and target index */
  int   rec_dex = 0;                     /* Which record we are doing */
  XmTextPosition   i, j;
  int              len;
  char  buf[MAX_FIELD_LEN];

  list_delete_all(widgets[k_dict_text]);

  /* Possibly put the field names across the top */
  if (lines_per_rec == 1) {
    for (i = next_index(-1); i < total_fields; i = next_index(i)) {
      len = strlen(fields[i].name);
      for (j= 0; j < len; j++){
	if (fields[i].vis_label) {
	  cooked_text[tdex++] = fields[i].name[j];
	}
	else
	  cooked_text[tdex++] = ' ';
      }
      cooked_text[tdex++] = ' ';
      cooked_text[tdex++] = ' ';
    }
    /* Now a separator line */
    cooked_text[tdex++] = '\n';
    for (i = next_index(-1); i < total_fields; i = next_index(i)) {
      if (fields[i].vis_label) {
	len = strlen(fields[i].name);
	for (j = 0; j < len; j++)
	  cooked_text[tdex++] = '=';
      }
      cooked_text[tdex++] = ' ';
      cooked_text[tdex++] = ' ';
    }
    cooked_text[tdex++] = 0;
  }
  else
    tdex = 0;

  while ((rec_dex < visible_records) && (raw_text[sdex])) {
    /* First locate the start of each field */
    for (i = 0; i < total_fields; i++) {
      if (fields[i].searchable)
	records[rec_dex].ret_ptr = &raw_text[sdex];
      else if (fields[i].browsable)
	records[rec_dex].disp_ptr = &raw_text[sdex];
      fields[i].val_dex = sdex;
      while ((raw_text[sdex] != SEP) && raw_text[sdex] != '\n')
	sdex++;
      sdex++;
    }
    /* Go through the fields in display order */
    for (i = next_index(-1); i < total_fields; i = next_index(i)) {
      /* Copy in the label if it is not visible already */
      if (fields[i].vis_label && (lines_per_rec > 1)) {
	for (j = 0; fields[i].name[j] != 0; j++)
	  cooked_text[tdex++] = fields[i].name[j];
	cooked_text[tdex++] = ' ';
	cooked_text[tdex++] = ' ';
      }
      
      /* Copy in the value. First extract it */
      pull_str(raw_text, &fields[i].val_dex, buf, MAX_FIELD_LEN);
      len = strlen(buf);
      if (fields[i].max_len && (len > fields[i].max_len)){
	buf[fields[i].max_len] = 0;
	len = fields[i].max_len;
      }
      /* Do blank padding, and copy in value */
      if (fields[i].max_len) {
	if (fields[i].rjust) {
	  for (j = 0; j < (fields[i].max_len - len); j++)
	    cooked_text[tdex++] = ' ';
	  for (j = 0; j < len; j++)
	    cooked_text[tdex++] = buf[j];
	}
	else {
	  for (j = 0; j < len; j++)
	    cooked_text[tdex++] = buf[j];
	  for (j = 0; j < (fields[i].max_len - len); j++)
	    cooked_text[tdex++] = ' ';
	}
      }
      else
	for (j = 0; j < len; j++)
	  cooked_text[tdex++] = buf[j];
      cooked_text[tdex++] = ' ';
      cooked_text[tdex++] = ' ';

      if (lines_per_rec > 1)
	cooked_text[tdex++] = '\n';
    }

    /* Put in the separator */
    cooked_text[tdex++] = '\n';
    rec_dex++;
    cooked_text[tdex - 1] = 0;
  }
  /* Terminate string, and kill last \n */
  cooked_text[--tdex] = 0;
  if (lines_per_rec > 1)
    cooked_text[--tdex] = 0;
  rec_dex--;
  list_replace_all_items(widgets[k_dict_text], cooked_text,
			   rec_dex + 1, lines_per_rec == 1);
}

/* Extract the next string from the server record. Adjust sdex_p to point 
   to start of next field. Put string into buf.
 */
static char *pull_str(char *text, int *sdex_p, char *buf, int max)
{
  int sdex, tdex = 0;

  sdex = *sdex_p;
  for (; (text[sdex] != SEP) && (text[sdex] != '\n') && (tdex < (max-1)); 
       sdex++){
    buf[tdex++] = text[sdex];
  }
  sdex++;
  buf[tdex] = 0;
  *sdex_p = sdex;
  return(buf);
}

/* Extract the next integer from the server record. Adjust sdex_p to point 
   to start of next field 
 */
static int pull_int(char *header, int *sdex_p)
{
  int sdex, tdex = 0;
  char buf[MAX_NAME_LEN];

  sdex = *sdex_p;
  for (; header[sdex] && (header[sdex] != SEP); sdex++) {
    buf[tdex++] = header[sdex];
  }
  if (!header[sdex])
    return(-1);
  sdex++;
  buf[tdex] = 0;
  *sdex_p = sdex;
  return(atoi(buf));
}

/* Pull information from list header record */
static int parse_header(char *header)
{
  int  sdex = 0, tdex, i;
  int  longest = 0, len;
  char buf[MAX_NAME_LEN];

  /* Extract the data about the list as a whole */
  if ((total_records = pull_int(header, &sdex)) < 0)
    return(-1);
  if ((total_vis_fields = total_fields = pull_int(header, &sdex)) < 0)
    return(-1);
  if (total_fields > MAX_FIELDS)
    return(-1);
  if ((lines_per_rec = pull_int(header, &sdex)) < 0)
    return(-1);

  /* Parse out relevant info about each field */
  for (i = 0; i < total_fields; i++) {
    pull_str(header, &sdex, fields[i].name, MAX_NAME_LEN);
    if ((fields[i].max_len = pull_int(header, &sdex)) < 0)
      return(-1);
    if ((fields[i].pos = pull_int(header, &sdex)) < 0)
      return(-1);
    if (fields[i].pos == 0)
      fields[i].pos = i + 1;

    /* Read the flags */
    fields[i].browsable  = (header[sdex++] == '1');
    fields[i].searchable = (header[sdex++] == '1');
    if ((fields[i].vis_value  = (header[sdex++] == '1')) == 0) {
      total_vis_fields--;
      fields[i].pos = 0;
    }
    fields[i].vis_label  = (header[sdex++] == '1');
    fields[i].rjust = (header[sdex++] == '1');
    sdex++;

    len = strlen(fields[i].name);
    if ((i > 0) && (len > longest) && (fields[i].vis_label))
      longest = len;
  }

  /* Right-justify the field names if we have multi-line records */
  if (lines_per_rec > 1) {
    for (i = 0; i < total_fields; i++)
      if (fields[i].vis_label) {
	len = longest - strlen(fields[i].name);
	if (len > 0) {
	  strcpy(buf, fields[i].name);
	  for (tdex = 0; tdex < len; tdex++)
	    fields[i].name[tdex] = ' ';
	  strcpy(&fields[i].name[tdex], buf);
	}
      }
  }
  set_visible_records();
  return 0;
}


static void kb_up(Widget w, XEvent *event, 
		  String *params, Cardinal *num_params)
{
  list_kb_up(widgets[k_dict_text], event, params, num_params);
}

static void kb_down(Widget w, XEvent *event, 
		    String *params, Cardinal *num_params)
{
  list_kb_down(widgets[k_dict_text], event, params, num_params);
}

void dict_scroll_up_cb(Widget w, list_callback_struct *reason)
{
  XmScrollBarCallbackStruct sb_reason;
  int  decrement;

  if (last_sel > 0)
    highlight(last_sel - 1);
  else {
    decrement = visible_records - record_offset;

    if (cur_scroll_value > decrement)
      sb_reason.value = cur_scroll_value - decrement;
    else
      sb_reason.value = 0;
    scroll_cb(NULL, TRUE, &sb_reason);
  }
}

void dict_scroll_down_cb(Widget w, list_callback_struct *reason)
{
  XmScrollBarCallbackStruct sb_reason;
  int increment;

  if (last_sel < (visible_records - 1))
    highlight(last_sel + 1);
  else {
    increment = visible_records - record_offset;

    if ((cur_scroll_value + increment) < (total_records - visible_records))
      sb_reason.value = cur_scroll_value + increment;
    else
      sb_reason.value = total_records - visible_records;
    scroll_cb(NULL, TRUE, &sb_reason);
  }
}

static void kb_p_up(Widget w, XEvent *event, 
		    String *params, Cardinal *num_params)
{
  XmScrollBarCallbackStruct reason;

  reason.value = cur_scroll_value - visible_records + 1;
  if (reason.value < 0)
    reason.value = 0;
  scroll_cb(NULL, NULL, &reason);
}

static void kb_p_down(Widget w, XEvent *event, 
		      String *params, Cardinal *num_params)
{
  XmScrollBarCallbackStruct reason;

  reason.value = cur_scroll_value + visible_records - 1;
  if (reason.value > (total_records - visible_records))
    reason.value = total_records - visible_records;
  scroll_cb(NULL, NULL, &reason);
}

/* count how many characters in this field */
int find_len(char *str)
{
  int len = 0;

  if (!str)
    return(0);
  while (str[len] && (str[len] != SEP) && (str[len] != '\n'))
    len++;
  return(len);
}

sel_rec_def *new_selected_rec(int pos)
{
  sel_rec_def *selected;

  selected = (sel_rec_def *) XtMalloc(sizeof(sel_rec_def));
  selected->ret_value = strndup(records[pos].ret_ptr, 
				find_len(records[pos].ret_ptr));
  selected->disp_value = strndup(records[pos].disp_ptr, 
				find_len(records[pos].disp_ptr));
  selected->real_pos = cur_scroll_value + pos;
  selected->next = NULL;
  return(selected);
}

void dict_select_cb(Widget w, list_callback_struct *reason)
{
  sel_rec_def *selected, *cur, *prev;

  last_sel = reason->pos - 1;

  /* Insert into list, in order of real_pos */
  if (!sel_list)
    sel_list = new_selected_rec(last_sel);
  else {
    cur = sel_list;
    prev = NULL;
    while (TRUE) {
      if (cur->real_pos == cur_scroll_value + last_sel)
	break;
      else if (cur->real_pos > cur_scroll_value + last_sel) {
	selected = new_selected_rec(last_sel);
	selected->next = cur;
	if (prev)
	  prev->next = selected;
	else
	  sel_list = selected;
	break;
      }
      if (!cur->next) {
	cur->next = new_selected_rec(last_sel);
	break;
      }
      prev = cur;
      cur = cur->next;
    }
  }
}

/* We got a de-selection from the text-list */
void dict_deselect_cb(Widget w, list_callback_struct *reason)
{
  sel_rec_def *prev, *cur;

  last_sel = reason->pos - 1;

  /* Delete selected from list.*/
  cur = sel_list;
  prev = NULL;
  while (cur) {
    if (cur->real_pos == cur_scroll_value + last_sel) {
      if (prev)
	prev->next = cur->next;
      else
	sel_list = cur->next;
      XtFree(cur->ret_value);
      XtFree(cur->disp_value);
      XtFree(cur);
      break;
    }
    prev = cur;
    cur = cur->next;
  }
}

void dict_deselect_all_cb(Widget w, list_callback_struct *reason)
{
  sel_rec_def *next;

  next = sel_list;
  while (next) {
    next = next->next;
    XtFree(sel_list->ret_value);
    XtFree(sel_list->disp_value);
    XtFree(sel_list);
    sel_list = next;
  }
}

void dict_double_click_cb(Widget w, list_callback_struct *reason)
{
  dict_paste_cb(NULL, NULL, NULL);
}

static highlight(int rec)
{
  list_select_pos(widgets[k_dict_text], rec + 1, FALSE, TRUE);
}


/* Set up the scrollbar to match our new position in remote file */
static adjust_scrollbar(int pos)
{
  int maximum;
  sel_rec_def *selected;

  if (pos < 0)
    pos = 0;
  maximum = total_records;
  cur_scroll_value = pos;
  if ((pos > (maximum - visible_records)) && 
      (maximum > visible_records))
    pos = maximum - visible_records;

  if (maximum < visible_records) 
    maximum += (visible_records - (maximum - pos));

  XtVaSetValues(sb,  XmNmaximum, maximum, XmNsliderSize, visible_records,
		XmNpageIncrement, visible_records - 1, XmNvalue, pos, NULL);

  /* See if text in window needs to be selected */
  for (selected = sel_list; selected != NULL; selected = selected->next) {
    if (cur_scroll_value + visible_records < selected->real_pos)
      break;
    else if (cur_scroll_value <= selected->real_pos) 
      highlight(selected->real_pos - cur_scroll_value);
  }
}

/* Figure out how many records we can fit on the screen, and how many we 
   want our results offset by. */
static void set_visible_records(void)
{
  if (lines_per_rec == 1)
    visible_records = visible_lines;
  else {
    visible_records = visible_lines / (total_vis_fields + 1);
    if ((visible_lines % (lines_per_rec + 1)) == lines_per_rec)
      visible_records++;
  }
  /* Make room for labels across the top */
  if (lines_per_rec == 1)
    visible_records -= 2;
 
  if (visible_records >= MAX_RECORDS)
    error("visible_records too big!");

  /* during opac mode, text can be completely hidden. If so, just fake it */
  else if ((opac_mode == 2) || (opac_mode == 1))
    visible_records = 5;
    
  record_offset = visible_records / 3;
}

void opac_search_cb(Widget w, int *tag, u_long *reason)
{
  char           *ptr;
  char            buf[MAX_LINE * 10];
  int             len = 0;
  sel_rec_def    *selected;
  Boolean         need_quotes = FALSE;

  if (!XtIsSensitive(widgets[k_opac_search]))
    return;

  /* If there is browse list */
  if (cur_list) {
    if (!sel_list) {
      if (opac_mode == 3)
	pop_warning("You must select at least one item from the list before searching", NULL, search_shell);
      else
	pop_warning("You must select at least one item from the list before searching.\n\
In order to display the list, you must type the first few\n\
letters of the item you are interested in.", NULL, search_shell);
      return;
    }
    for (selected = sel_list; selected != NULL; selected = selected->next) {
      /* if this is not the first, put in operator */
      if (len > 1) {
	strncpy(buf + len, DFLT_OP, strlen(DFLT_OP));
	len += strlen(DFLT_OP);
      }
      /* Ugly, but we have to do one pass to see if quotes are needed */
      for (ptr = selected->ret_value; (*ptr); ptr++) 
	if (*ptr == ' ') {
	  need_quotes = TRUE;
	  buf[len++] = '"';
	  break;
	}
      /* Now copy characters in */
      if ((len + strlen(selected->ret_value)) > MAX_LINE)
	goto too_long;
      for (ptr = selected->ret_value; (*ptr); ptr++) 
	buf[len++] = *ptr;
      /* add right paren if needed */
      if (need_quotes) {
	buf[len++] = '"';
	need_quotes = FALSE;
      }
    }
    buf[len] = 0;
    if (len >= MAX_LINE) {
too_long:    pop_warning("Total length of your selections is too long", 
	     NULL, search_shell);
      return;
    }
  }

  /* If no browse list, search what user has typed */
  else {
    ptr = XmTextGetString(widgets[k_dict_search]);
    if (strlen(ptr) == 0) {
      pop_warning("You must type search terms(s) before you can start a search.", NULL, search_shell);
      return;
    }
    strcpy(buf, ptr);
    XtFree(ptr);
    len = strlen(buf);
  }
  len++;
  strcpy(buf + len, cur_field->command);
  len += strlen(cur_field->command) + 1;
  
  /* Send search to driver, prepare interface, just like search_cb() */
  send_search_cmd(1, 0, buf, len);
  
  /* Record a representation of this search for use with saves/prints */
  make_query_display();

  /* Zero the field, but don't trigger a search/scroll */
  ptr = XmTextGetString(widgets[k_dict_search]);
  if (*ptr) {
    if (cur_list) {
      XtRemoveAllCallbacks(widgets[k_dict_search], XmNvalueChangedCallback);
      XmTextSetString(widgets[k_dict_search], "");
      XtAddCallback(widgets[k_dict_search], XmNvalueChangedCallback,
		    dict_search_cb, NULL);
    }
  }
  
  start_display();
  clear_titles();
  pop_globe(k_search_globe_box);
  not_ready_buttons();
  status = SEARCHING;
}
