/***********************************************************************
 *                                                                     *
 *              WILLOW(tm) Information Retrieval Software              *
 *     (Washington Information Looker-upper Layered Over Windows)      *
 *             Copyright 1992-1994 University of Washington            *
 *                                                                     *
 *   Willow is a trademark of the University of Washington.  No        *
 *   commercial use of this trademark or software may be made without  *
 *   prior written permission of the University of Washington.         *
 *                                                                     *
 * Permission to use, copy, modify, and distribute Willow(tm) software *
 * and its documentation for any noncommercial purpose is hereby       *
 * granted, provided that the above copyright notice appears in all    *
 * copies and that both the above copyright notice and this permission *
 * notice appear in supporting documentation, and that the name of the *
 * University of Washington not be used in advertising or publicity    *
 * pertaining to distribution of the software without specific,        *
 * written prior permission.  This software is made available "as is", *
 * and THE UNIVERSITY OF WASHINGTON DISCLAIMS ALL WARRANTIES, EXPRESS  *
 * OR IMPLIED, WITH REGARD TO THIS SOFTWARE, INCLUDING WITHOUT         *
 * LIMITATION ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS    *
 * FOR A PARTICULAR PURPOSE, AND IN NO EVENT SHALL THE UNIVERSITY OF   *
 * WASHINGTON BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL     *
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA  *
 * OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, TORT (INCLUDING       *
 * NEGLIGENCE) OR STRICT LIABILITY, ARISING OUT OF OR IN CONNECTION    *
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.                       *
 *                                                                     *
 * For further information please contact:                             *
 * Willow Project                                                      *
 * Information Systems                                                 *
 * University of Washington                                            *
 * 4545 15th NE, 1st floor                                             *
 * Seattle, Washington, 98105, USA                                     *
 * Internet: willow@cac.washington.edu                                 *
 ***********************************************************************/

/*******************************************************
 * ftp.c
 * This file contains routines for using ftp to save
 * results across the network.
 ******************************************************/

#include <stdio.h>
#include <sys/types.h>

#include <Mrm/MrmAppl.h>
#include <Xm/ToggleBG.h>
#include <Xm/Text.h>

#include "willow.h"

#define IN  0
#define OUT 1

/* forward references */
void ftp_read(caddr_t data, int *fid, XtInputId *iid);
void ftp_get_read(caddr_t data, int *fid, XtInputId *iid);
extern char *get_query_display_str(void);
extern void pop_warning (char *template, char *arg, Widget parent);
extern int pop_working (char *message, Widget parent);
extern int error (char *msg);
extern void clear_dialog_cb (Widget w, Widget target, u_long *reason);
int do_ftp (int which);
int do_ftp_get (void);
extern int desensitize_selection_buttons (void);
#ifdef vax
extern int getpid ();
extern int vfork ();
#else
extern pid_t getpid (void);
extern pid_t vfork (void);
#endif
extern int really_save_summaries (FILE *outfile);
extern void really_save_search_cb (Widget, int *, XmFileSelectionBoxCallbackStruct *);
extern int pipe (int *);
extern int close (int);
extern int dup (int);
extern int execlp (char *, char *, ...);
extern int write (int, void *, unsigned int);
extern int read (int, void *, unsigned int);
extern int sensitize_selection_buttons (void);
extern int unlink (char *);
extern void really_open_search_cb (Widget, int *, XmFileSelectionBoxCallbackStruct *);
extern int not_ready_buttons (void);
extern int clear_working (void);
extern int ready_buttons (void);


extern Widget widgets[], toplevel, search_shell, summary_shell, results_shell;
extern XtAppContext app_context;
extern Boolean saving;
extern Boolean show_search;

/* How much text is currently in the full-rec window */
extern int text_len;

/* Strings we get out of ftp set-up dialog */
char *ftp_host = NULL,
     *ftp_uid  = NULL,
     *ftp_pw   = NULL,      /* initialized in init_application() */
     *ftp_file = NULL,
     *ftp_q_file = NULL;
Boolean ftp_append;

/* Pipes into ftp, and error back */
int to_ftp[2],
    from_ftp[2];

/* process id for ftp */
int ftp_pid = 0;

/* Name of FTP scratch file */
char  tmp_name[MAX_LINE];

/* For query-set ftp's, are we opening or saving? */
Boolean query_opening;

/* Are we currently doing a save of full results? */
extern Boolean popped_full;

/* Pull new string out of text widget, reset old if needed */
char *get_string(char *s, Widget w)
{
  if (s)
    XtFree(s);
  return(XmTextGetString(w));
}

/* Unmanage the ftp options box and collect new parameters */
void ftp_opts_ok_cb(Widget w, int *tag, u_long *reason)
{
  clear_dialog_cb(w, widgets[k_save_box], reason);

  show_search = XmToggleButtonGetState(widgets[k_show_search_btn]);
  
  ftp_host = get_string(ftp_host, widgets[k_ftp_host_text]);
  ftp_uid  = get_string(ftp_uid,  widgets[k_ftp_uid_text]);
  ftp_file = get_string(ftp_file, widgets[k_ftp_file_text]);
  ftp_append = XmToggleButtonGetState(widgets[k_ftp_append_btn]);

  /* Put the query ftp dialog in synch */
  XmTextSetString(widgets[k_q_ftp_host_text], ftp_host);
  XmTextSetString(widgets[k_q_ftp_uid_text], ftp_uid);
  XmTextSetString(widgets[k_q_ftp_pw_text], ftp_pw);

  /* If a capture file has been completely specified, display name.
     Also make sure either both uid and pw are there, or neither are there.*/
  if ((ftp_host[0] && ftp_file[0]) && 
      ((ftp_uid[0] && ftp_pw[0]) || (!ftp_uid[0] && !ftp_pw[0]))) {
    if (popped_full)
      do_ftp(FULL_FLAG);
    else
      do_ftp(SUMM_FLAG);
  }
  else
    pop_warning("Incomplete FTP specification", NULL, 
		popped_full ? results_shell : summary_shell );
}

/* Unmanage the query ftp options box and collect new parameters */
void q_ftp_opts_ok_cb(Widget w, int *tag, u_long *reason)
{
  clear_dialog_cb(w, widgets[k_q_save_box], reason);

  ftp_host = get_string(ftp_host, widgets[k_q_ftp_host_text]);
  ftp_uid  = get_string(ftp_uid,  widgets[k_q_ftp_uid_text]);
  ftp_q_file = get_string(ftp_q_file, widgets[k_q_ftp_file_text]);

  /* Put the results-save ftp dialog in synch */
  XmTextSetString(widgets[k_ftp_host_text], ftp_host);
  XmTextSetString(widgets[k_ftp_uid_text], ftp_uid);
  XmTextSetString(widgets[k_ftp_pw_text], ftp_pw);

  /* If a capture file has been completely specified, display name.
     Also make sure either both uid and pw are there, or neither are there.*/
  if ((ftp_host[0] && ftp_q_file[0]) && 
      ((ftp_uid[0] && ftp_pw[0]) || (!ftp_uid[0] && !ftp_pw[0]))) {
    if (query_opening)
      do_ftp_get();
    else
      do_ftp(QUERY_FLAG);
  }
  /* If nothing was specified, pop a warning */
  else
    pop_warning("Incomplete FTP specification", NULL, search_shell);
}

/* Do the ftp transfer for a full or summary result, or query */
do_ftp(int which)
{
  char *result;
  char ftp_cmd_buf[CHUNK_SIZE];
  FILE *tmp_file;
  XmFileSelectionBoxCallbackStruct reason;  
  char *file_prefix;
  Widget shell;
  char *ftp_type;

  XtSetSensitive(widgets[k_full_save_btn], FALSE);
  desensitize_selection_buttons();
  not_ready_buttons();

  /* First write out the results to a temporary file */
  if (which == FULL_FLAG) {
    file_prefix = FULL_TMP_FILE;
    shell = results_shell;
  }
  else if (which == SUMM_FLAG) {
    file_prefix = SUMM_TMP_FILE;
    shell = summary_shell;
  }
  else {
    file_prefix = QUERY_TMP_FILE;
    shell = search_shell;
  }
  sprintf(tmp_name,"%s%d", file_prefix, getpid());
  if (!(tmp_file = fopen(tmp_name, "w"))) {
    pop_warning("Unable to open %s", tmp_name, shell);
    return;
  }
  saving = TRUE;
  pop_working("FTP in progress...", shell);

  if (which == FULL_FLAG) {
    result = XmTextGetString(widgets[k_text]);
    fprintf(tmp_file, "%s", get_query_display_str());
    fprintf(tmp_file, "%s\n", result);
    XtFree(result);
  }
  else if (which == SUMM_FLAG)
    really_save_summaries(tmp_file);
  else {
    reason.value = XmStringCreateSimple(tmp_name);
    really_save_search_cb(NULL, NULL, &reason);
  }

  fclose(tmp_file);

  /* Do the FTP transfer, via fork and exec. */

  /* set up pipes for communicating with ftp */
  pipe(to_ftp);
  pipe(from_ftp);
   
  /* fork off a child, which will overlay ftp on itself */

  ftp_pid = vfork();

  if (ftp_pid == 0) {
    /* in the child, redirect stdin and stdout. */
    close(fileno(stdin)); 
    dup(to_ftp[IN]);
    close(fileno(stdout));
    dup(from_ftp[OUT]);
    close(to_ftp[IN]);
    close(to_ftp[OUT]); 
    close(from_ftp[IN]); 
    close(from_ftp[OUT]);

    /* Exec ftp */
    if (execlp("ftp", "ftp", FTP_ARGS, ftp_host, NULL) < 0) {
      /* at this point we are in a child process, so we can't do much */
      perror(NULL);
      /* Must _exit after vfork */
      _exit(-1);
    }
  }
  else if (ftp_pid > 0) {
    /* in the parent, close extra pipes */
    close(to_ftp[IN]);
    close(from_ftp[OUT]);
    
    if (which == QUERY_FLAG)
      ftp_type = "put";
    else if (XmToggleButtonGetState(widgets[k_ftp_append_btn]))
      ftp_type = "append";
    else
      ftp_type = "put";

    /* Send the commands down the pipe to ftp */
    if (ftp_uid[0])
      sprintf(ftp_cmd_buf,"user %s %s\n%s %s %s\nquit\n",
	      ftp_uid, ftp_pw,  ftp_type, 
	      tmp_name, (which == QUERY_FLAG ? ftp_q_file : ftp_file));
    else
      sprintf(ftp_cmd_buf,"%s %s %s\nquit\n", ftp_type, tmp_name, 
	      ftp_file);

    write(to_ftp[OUT], ftp_cmd_buf, strlen(ftp_cmd_buf));

    /* Now add a callback to read from pipe whenever something
       appears */
    XtAppAddInput(app_context, from_ftp[IN], 
		  XtInputReadMask, ftp_read, NULL);
  }
  else {
    perror(NULL);
    error("Could not fork ftp");
  }
}

/* Do the ftp get for retrieve query set */
do_ftp_get(void)
{
  char ftp_cmd_buf[CHUNK_SIZE];

  sprintf(tmp_name,"%s%d", QUERY_TMP_FILE, getpid());
  pop_working("FTP in progress...", search_shell);

  /* Do the FTP transfer, via fork and exec. */

  /* set up pipes for communicating with ftp */
  pipe(to_ftp);
  pipe(from_ftp);
   
  /* fork off a child, which will overlay ftp on itself */
  ftp_pid = vfork();
  if (ftp_pid == 0) {
    /* in the child, redirect stdin and stdout. */
    close(fileno(stdin)); 
    dup(to_ftp[IN]);
    close(fileno(stdout));
    dup(from_ftp[OUT]);
    close(to_ftp[IN]);
    close(to_ftp[OUT]); 
    close(from_ftp[IN]); 
    close(from_ftp[OUT]);

    /* Exec ftp */
    if (execlp("ftp", "ftp", FTP_ARGS, ftp_host, NULL) < 0) {
      /* at this point we are in a child process, so we can't do much */
      perror(NULL);
      /* Must _exit after vfork */
      _exit(-1);
    }
  }
  else if (ftp_pid > 0) {
    /* in the parent, close extra pipes */
    close(to_ftp[IN]);
    close(from_ftp[OUT]);

    /* Send the commands down the pipe to ftp */
    if (ftp_uid[0])
      sprintf(ftp_cmd_buf,"user %s %s\nget %s %s\nquit\n",
	      ftp_uid, ftp_pw, ftp_q_file, tmp_name);
    else
      sprintf(ftp_cmd_buf,"get %s %s\nquit\n", ftp_q_file, tmp_name);

    write(to_ftp[OUT], ftp_cmd_buf, strlen(ftp_cmd_buf));

    /* Now add a callback to read from pipe whenever something
       appears */
    XtAppAddInput(app_context, from_ftp[IN], 
		  XtInputReadMask, ftp_get_read, NULL);
  }
  else {
    perror(NULL);
    error("Could not fork ftp");
  }
}

/* Read from standard error from ftp process. Assume we always get enough
   message in the first read.
 */
void ftp_read(caddr_t data, int *fid, XtInputId *iid)
{
  static char ftp_msg_buf[CHUNK_SIZE];
  int  n = 0;
  
  clear_working();
  saving = FALSE;

  /* The -1 is so there is always room to append a \0 */
  n = read(from_ftp[IN], ftp_msg_buf, sizeof(ftp_msg_buf) - 1);  

  sensitize_selection_buttons();
  ready_buttons();
  XtSetSensitive(widgets[k_full_save_btn], TRUE);

  unlink(tmp_name);
  XtRemoveInput(*iid);
  close(to_ftp[OUT]);
  close(from_ftp[IN]);

  if (n > 0) {
    /* Got an error message, ftp failed */
    ftp_msg_buf[n] = 0;
    pop_warning("FTP Error:\n%s", ftp_msg_buf, search_shell);
  }
}

/* Read from standard error from ftp-get process. Assume we always get enough
   message in the first read.
 */
void ftp_get_read(caddr_t data, int *fid, XtInputId *iid)
{
  static char ftp_msg_buf[CHUNK_SIZE];
  int  n = 0;
  XmFileSelectionBoxCallbackStruct reason;  

  clear_working();

  /* The -1 is so there is always room to append a \0 */
  n = read(from_ftp[IN], ftp_msg_buf, sizeof(ftp_msg_buf) - 1);  
  if (n > 0) {
    /* Got an error message, ftp failed */
    ftp_msg_buf[n] = 0;
    pop_warning("FTP Error:\n%s", ftp_msg_buf, search_shell);
  }
  else {
    reason.value = XmStringCreateSimple(tmp_name);
    really_open_search_cb(NULL, NULL, &reason);
    XmStringFree(reason.value);
  }
  unlink(tmp_name);
  XtRemoveInput(*iid);
  close(to_ftp[OUT]);
  close(from_ftp[IN]);
}

