/*

 \input cnoweb
 \title{Help text popup for X}

 \section{Introduction}
  This procedure displays help text in a popup window.
  See the \'Reading the help files' section for details on
  the help file structure.
  */

#define Private static
#define PRINT if(0) printf   /* `if(1)' for debug */

#include <sys/types.h>
#include <sys/stat.h>
#include <netinet/in.h>
 
#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#ifdef SOLARIS
#include <string.h>
#else
#include <strings.h>
#endif

 
extern int errno;

#include <X11/Intrinsic.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <Xm/Xm.h>
#include <Xm/MwmUtil.h>
#include <Xm/PanedW.h>
#include <Xm/ScrolledW.h>
#include <Xm/PushB.h>
#include <Xm/Text.h>
#include <Xm/Label.h>
#include <Xm/Form.h>
#include <Xm/List.h>
#include <Xm/RowColumn.h>
#include <Xm/Frame.h>
#include <Xm/DrawingA.h>

/* We need this for popping up willow error dialogs */
extern Widget search_shell;
  
/* Help widget set. This widget set is used to display help text.
   The structure and all associated data
   are allocated as needed. */

#define MAX_TOPICS 50

typedef struct HelpTop__ {
  caddr_t helpset;             /* parent helpset */
  int    ntopics;              /* number of topics */
  char *topics[MAX_TOPICS];    /* topic title and text  */
  int len[MAX_TOPICS];         /* length of each topic's text */
  struct HelpTop__ *subtop[MAX_TOPICS];    /* subtopic menu */
  char *bmfile[MAX_TOPICS];    /* bitmap file name */
  int  nlines[MAX_TOPICS];     /* number of lines */
} HelpTop_, *HelpTop;


typedef struct HelpSet__ {
  int *root;                   /* a pointer to the helpset */
  Widget w_shell;                /* help popup shell */
  Widget f_list;               /* topics list frame */
  Widget w_list;               /* topics list */
  Widget f_display;            /* help text frame */
  Widget w_display;            /* help text */
  Widget w_quit;               /* quit button */
  HelpTop topics;              /* first level topics */
  char *path;                  /* filename path - for subtopic names */
  char *label;                 /* label - for subtopic labels */
  HelpTop topstk[10];          /* stack for nested topic menus */
  int toplvl;                  /* index to above */
  int maxitems;                /* max number of topics in a menu */
  int width;                   /* width of screen in chars */
  int height;                  /* height of screen in lines */
} HelpSet_, *HelpSet;
  
/* Forward References -- To keep Anal C compilers quiet. */

Private AddHelpTopics();
Private void list_cb();
Private select_topic();
Private void quit_cb();
Private void destroy_cb();
Private drop_subtops();
Private HelpSet create_helpset();
Private HelpTop create_helptopics();
Private position_popup();
  
/* args are used like local vars throughout */

#define ARGN(l,v) XtSetArg(args[n], l, v); n++;
Private Arg args[20];

Private Position width, height;
Private Position xpos, ypos;

Private Arg wh_args[] = {                  /* to get widget size */
  {XtNwidth,	(XtArgVal) &width},
  {XtNheight,	(XtArgVal) &height},
};



/* \newpage */
/* \section{Procedures to display help text} The help text file is scanned,
   all topics are collected, the selected topic is displayed. 
   Bitmap image files are processed only when necessary.

   Two entries are provided.  One positions the help popup by
   coordinates; the other positions it next to a designated widget.
   */

/* \subsection{\"display_help_positioned(...)"} 
    Display help text.
    \item{\"W"} help root widget
    \item{\"path"} path for all other files
    \item{\"file"} help text file name
    \item{\"root"} root of helpset
    \item{\"name"} help window label
    \item{\"x,y"} Position for popup
    \item{\"label"} Label for popup widget
    \item{\"defs"} Defined strings
    \item{\"ndefs"} Number of defined strings
   */

#define STD_MARGIN 5
void quit_cb();
HelpSet create_helpset();
void destroy_cb();
void list_cb();
void display_pop_topic();
Widget AddTopicsForm();
Pixmap get_the_pixmap();
  
int display_help_positioned(W,path,file,root,name,x,y,label,defs,ndefs)
Widget W;
char *path;
char *file;
HelpSet *root;
char *name;
Position x;
Position y;
char *label;
char **defs;
int ndefs;
{
   HelpSet H;
   int n;
   char *starttext;
   char *sp;
   Widget button;
   Widget box;
   Widget text;
   Widget frame;
   Widget form;
   Widget picture;
   XmString s;

   /* If the help popup exists, just be sure it is at the top
      and is on the screen. */

     if (*root) {   /* already popped */
        XMapRaised(XtDisplay((*root)->w_shell),
            XtWindow((*root)->w_shell));
        position_popup((*root)->w_shell, STD_MARGIN, STD_MARGIN);
        return;
     }
      
   /* Read the help text and get the topics */
     
   if (*file=='\0') {
      pop_warning("Sorry, no help available", NULL, search_shell);
      return (0);
   }

   if (!(H=create_helpset(path,file,defs,ndefs))) return;
   H->root = (int*)root;
   *root = H;
   H->label = (char*) XtMalloc(strlen(label)+1);
   strcpy(H->label,label);


   /* Create a popup shell for the help text. */

   n = 0;
#ifdef ndef
   ARGN(XtNx, x);
   ARGN(XtNy, y);
#endif
   ARGN(XtNtitle, label);
   ARGN(XmNmwmDecorations, MWM_DECOR_TITLE|MWM_DECOR_BORDER);
   ARGN(XmNallowShellResize, True);

   H->w_shell = XtCreatePopupShell(name, 
       topLevelShellWidgetClass, W,
       args, n);
   XtAddCallback(H->w_shell,XtNdestroyCallback,
      destroy_cb,H);

  n = 0;
  ARGN(XmNrubberPositioning, True);
  ARGN(XmNresizePolicy, XmRESIZE_ANY);
  ARGN(XmNresizable, True);
  form = XtCreateManagedWidget ("form",
         xmFormWidgetClass,
         H->w_shell, args, n);

  /* Add text display. */

  n = 0;
  ARGN(XmNshadowType, XmSHADOW_IN);
  ARGN(XmNleftAttachment, XmATTACH_FORM);
  ARGN(XmNleftPosition, 1);
  ARGN(XmNtopAttachment, XmATTACH_FORM);
  ARGN(XmNtopPosition, 1);

#ifdef ndef
  ARGN(XmNscrollBarDisplayPolicy, XmAS_NEEDED);
  ARGN(XmNscrollBarPlacement, XmBOTTOM_RIGHT);
  ARGN(XmNscrollingPolicy, XmAUTOMATIC);
  ARGN(XmNscrolledWindowMarginWidth, 2);
  ARGN(XmNscrolledWindowMarginHeight, 2);
#endif
  H->f_display = XtCreateManagedWidget ("textframe",
      xmFrameWidgetClass,
      form, args,n);

  PRINT("Default screen is %dx%d\n",H->width, H->height);
  sp = starttext = (char *)XtMalloc(H->width + (H->height * 2) +2);
  for (n=0;n<H->width;n++) *sp++=' ';
  *sp++ = '\n';
  for (n=1;n<H->height;n++) {
       *sp++ = ' ';
       *sp++ = '\n';
  }
  *sp = '\0';

#ifdef TEXTDISPLAY
  n = 0;
  ARGN(XmNvalue, starttext);
  ARGN(XmNeditable, False);
  ARGN(XmNrecomputeSize, False);
  ARGN(XmNcursorPositionVisible, False);
  H->w_display = XtCreateManagedWidget ("text", xmTextWidgetClass,
             H->f_display, args, n);
#else
  n = 0;
  s = XmStringCreateLtoR(starttext,XmSTRING_DEFAULT_CHARSET);
  ARGN(XmNlabelString, s);
  ARGN(XmNalignment, XmALIGNMENT_BEGINNING);
  ARGN(XmNrecomputeSize, False);
  H->w_display = XtCreateManagedWidget ("text", xmLabelWidgetClass,
             H->f_display, args, n);
  XmStringFree (s);
#endif
  XtFree(starttext);

   n=0;
   width = height = 0;
   ARGN(XmNwidth, &width);
   ARGN(XmNheight, &height);
   XtGetValues(H->w_display, args, n);
   PRINT(".. display w/h = %d/%d\n", width,height);
   width += 20;  /* sloppy allowance for borders */
   height += 20;
   n=0;
   ARGN(XmNwidth, width);
   ARGN(XmNheight, height);
   XtSetValues(H->f_display, args, n);


  /* Add the topics list. */

  n = 0;
  ARGN(XmNshadowType, XmSHADOW_IN);
  ARGN(XmNleftAttachment, XmATTACH_WIDGET);
  ARGN(XmNleftWidget, H->f_display);
  ARGN(XmNtopAttachment, XmATTACH_FORM);
  ARGN(XmNtopPosition, 1);
  ARGN(XmNrightAttachment, XmATTACH_FORM);
  ARGN(XmNrightPosition, 1);
  H->f_list = XtCreateManagedWidget ("frame", xmFrameWidgetClass,
      form, args,n);

  n = 0;
  ARGN(XmNlistSizePolicy, XmVARIABLE);
  ARGN(XmNlistMarginWidth, 3);
  ARGN(XmNlistMarginHeight, 3);
  ARGN(XmNselectionPolicy, XmSINGLE_SELECT);
  ARGN(XmNlistSpacing, 2);
  ARGN(XmNhighlightThickness, 0);
  ARGN(XmNvisibleItemCount, 15);
  H->w_list = XtCreateManagedWidget ("helplist", xmListWidgetClass,
        H->f_list, args, n);
  XtAddCallback (H->w_list, XmNsingleSelectionCallback,
        list_cb, H);

  H->topstk[0] = H->topics;

  /* Add the `go away' button. */

  n = 0;
  ARGN(XmNshadowType, XmSHADOW_IN);
  ARGN(XmNleftAttachment, XmATTACH_FORM);
  ARGN(XmNleftPosition, 1);
  ARGN(XmNtopAttachment, XmATTACH_WIDGET);
  ARGN(XmNtopWidget, H->f_display);
  ARGN(XmNrightAttachment, XmATTACH_FORM);
  ARGN(XmNrightPosition, 1);
  frame = XtCreateManagedWidget ("frame", xmFrameWidgetClass,
      form, args,n);

  n = 0;
  s = XmStringCreateLtoR("Click here to close this window.",
       XmSTRING_DEFAULT_CHARSET);
  ARGN(XmNlabelString, s);
  button = XtCreateManagedWidget ("quit", xmPushButtonWidgetClass,
             frame, args, n);
  XtAddCallback (button, XmNactivateCallback, quit_cb, H);
  XmStringFree (s);

  /* Popup the window and load the first set of topics. */

  XtRealizeWidget(H->w_shell);
  AddHelpTopics(H);

  XtMapWidget(H->w_shell);

  return;

}

/* \subsection{\"display_help(...)"} 
  Display help---position by widget.

    \item{\"W"} help root widget
    \item{\"path"} path for all other files
    \item{\"file"} help text file name
    \item{\"root"} root of helpset
    \item{\"name"} help window label
    \item{\"Wpos"} Position near this widget
    \item{\"label"} Label for popup widget
    \item{\"defs"} Defined strings
    \item{\"ndefs"} Number of defined strings
  */

int display_help(W,path,file,root,name,Wpos,label,defs,ndefs)
Widget W;
char *path;
char *file;
HelpSet *root;
char *name;
Widget Wpos;
char *label;
char **defs;
int ndefs;
{
   Position x,y;

   width = height = 0;
   XtGetValues(Wpos,wh_args,XtNumber(wh_args));
   XtTranslateCoords(Wpos,
      (Position)(width/3), (Position)(0), &x, &y);
   display_help_positioned(W,path,file,
        root,name,x,y,label,defs,ndefs);
}

/* \subsection{\"display_help_shell(H)"} Return the shell widget
   for a helpshell. */

Widget display_help_shell(H)
HelpSet H;
{
   if (H) return (H->w_shell);
   else return (NULL);
}

/* \section{Text selection tools and callbacks}  These respond to
   actions in the topics list.  */

/* \subsection{\"AddHelpTopics(H)"} 
    Add the topics in the current topic set to the help popup \"H". */


Private
AddHelpTopics(H)
HelpSet H;
{
  HelpTop T = H->topstk[H->toplvl];
  char tlabel[128];
  int n;
  XmString *sb, *sp;
  XmString s;
  int itmc;

  sb = sp = (XmString*) XtMalloc(sizeof(XmString*) * T->ntopics);
  for (n=0;n<T->ntopics;n++) {
     *sp++ = XmStringCreate(T->topics[n],XmSTRING_DEFAULT_CHARSET);
  }
  XmListAddItems(H->w_list, sb, T->ntopics, 0);

  sp = sb;
  for (n=0;n<T->ntopics;n++) XmStringFree (*sp++);
  XtFree(sb);

  if (H->toplvl) {
     s = XmStringCreate("<--- previous menu",XmSTRING_DEFAULT_CHARSET);
     XmListAddItem(H->w_list, s, 0);
     XmStringFree (s);
  }

  /* Reposition the window, select the first item,
     and set all items visible. */

  XmListSelectPos(H->w_list, 1, True); /* Select first topic */

  n = 0;
  ARGN(XmNitemCount, &itmc);
  XtGetValues(H->w_list, args, n);
  n = 0;
  ARGN(XmNvisibleItemCount, itmc);
  XtSetValues(H->w_list, args, n);

  position_popup(H->w_shell,STD_MARGIN,STD_MARGIN);
  return;
}

/* \subsection{\"list_cb(w,H,call)"} Callback to respond to 
   a topic selection in the list */

Private
void list_cb(w,H,call)
Widget w;
HelpSet H;
caddr_t call;
{
   HelpTop T = H->topstk[H->toplvl];
   XmListCallbackStruct *cb = (XmListCallbackStruct *) call;

   if (cb->item_position > T->ntopics) {
      XmListDeleteItemsPos(H->w_list, (T->ntopics)+1, 1);
      H->toplvl--;
      AddHelpTopics(H);
   } else {
      select_topic(H,cb->item_position-1);
   }
}



/* \subsection{\"select_topic(w,H,call)"} 
   Set the help display to show a selected topic.
   If it is a subtopic menu then switch to the new set
   of topics and show the first one.  If the selection
   includes a bitmap put it on the display's background.
   */

#define USE_XTBG

Private
select_topic(H,ind)
HelpSet H;
int ind;
{
   HelpTop T = H->topstk[H->toplvl];
   char *saddr;
   int n;
   char *sp;
   Pixmap bmpix;
   char *text;
   XmString s;
   
   /* Is this a subtopic selection */

   if (T->subtop[ind]) {
     n = T->ntopics;
     if (H->toplvl) n++;  /* include the `back' */
     XmListDeleteItemsPos(H->w_list, n, 1);
     H->topstk[++H->toplvl] = T->subtop[ind];
     AddHelpTopics(H);
   
   /* else is text display */

   } else {

     /* Clear any old BG */

#ifdef USE_XTBG
     n = 0;
     ARGN(XmNbackgroundPixmap, XmUNSPECIFIED_PIXMAP);
     XtSetValues(H->w_display, args, n);
#else
     XSetWindowBackgroundPixmap( XtDisplay(H->w_display),
        XtWindow(H->w_display), None);
#endif

     /* Do we need to draw a picture */

     if ((T->bmfile[ind]) &&
         (bmpix=get_the_pixmap(H->w_display,
           H->path,T->bmfile[ind]))) {

#ifdef USE_XTBG
        n=0;
        ARGN(XmNbackgroundPixmap, bmpix);
        XtSetValues(H->w_display, args, n);
#else
        XSetWindowBackgroundPixmap( XtDisplay(H->w_display),
           XtWindow(H->w_display), bmpix);
#endif
        XFreePixmap(XtDisplay(H->w_display),bmpix);
     }

     saddr = T->topics[ind];
     saddr += strlen(saddr) + 1;
     n = H->height - T->nlines[ind];
     if (n<0) n = 0;
     sp = text = (char *)XtMalloc(strlen(saddr) + (n*2) + 2);
     while (*saddr) *sp++ = *saddr++;
     PRINT("Printing %d, lines=%d, fill=%d\n",
          ind, T->nlines[ind], n);
     while (n--) {
       *sp++ = ' ';
       *sp++ = '\n';
     }
     *sp = '\0';
     
#ifdef TEXTDISPLAY
     n = 0;
     ARGN(XmNvalue, text);
     XtSetValues(H->w_display, args, n);
#else
     n = 0;
     s = XmStringCreateLtoR(text,
          XmSTRING_DEFAULT_CHARSET);
     ARGN(XmNlabelString, s);
     XtSetValues(H->w_display, args, n);
     XmStringFree(s);
#endif
     XtFree(text);
        
   }
}

/* \section{Closing down} */

/* \subsection{\"quit_cb(w,H,call)"} Callback to respond to `quit'
   button on help display.  Destroy the popup. Recovery of resources
   will take place during the destroy callback. */

Private
void quit_cb(w,H,call)
Widget w;
HelpSet H;
caddr_t call;
{
    XtDestroyWidget(H->w_shell);
}

/* \subsection{\"close_help_display(root)"} External user call
   to close a helpset. */

close_help_display(root)
HelpSet *root;
{
   if (*root) quit_cb(NULL,*root,NULL);
   *root = NULL;
}


/* \subsection{\"destroy_cb(w,H,call)"} Callback to respond to 
   destruction of a help popup. Reset the helpset's root pointer,
   free any subtopics,
   and free the helpset's memory. */

Private
void destroy_cb(w,H,call)
Widget w;
HelpSet H;
caddr_t call;
{
   drop_subtops(H->topics);
   XtFree(H->path);
   XtFree(H->label);
   *H->root = NULL;
   XtFree (H);
}

/* \subsection{\"drop_subtops(T)"} Drop subtopics from a topic list */

Private
drop_subtops(T)
HelpTop T;
{
   int t;

   if (!T) return;
   for (t=0;t<T->ntopics;t++) {
       if (T->subtop[t]) drop_subtops(T->subtop[t]);
       if (T->bmfile[t]) XtFree(T->bmfile[t]);
   }
   XtFree(T);
}

/* \section{Reading the help files} The help files are ascii
   text.  The first character of a line indicates its function.

   \item{\"\#"} comment line(s)
   \item{\"!"}  topic label
   \item{\"\$"}  subtopic filename.
   \item{\"\%"}  bitmap file.
   \item{else}  help text
       
   Notes:

   \item{1} The comments are optional.
   \item{2} A subtopic overrides helptext and a bitmap file. 
   \item{4} All filenames are based on the \"path".
   
   */
   
/* \subsection{\"create_helpset(path,file)"} Create a helpset
    form the specified path and filename.

    Return the HelpSet. */

HelpTop create_helptopics();

Private
HelpSet create_helpset(path,file,defs,ndefs)
char *path;
char *file;
char **defs;
int ndefs;
{
    HelpSet H = (HelpSet) XtMalloc(sizeof(HelpSet_));

    H->path = (char*) XtMalloc(strlen(path)+1);
    strcpy(H->path,path);
    H->root = NULL;
    H->w_shell = NULL;
    H->w_display = NULL;
    H->w_quit = NULL;
    H->maxitems = 0;
    H->width = 0;
    H->height = 0;
    H->topics = create_helptopics(file,H,NULL,defs,ndefs);
    H->toplvl = 0;

    if (!H->topics) {
       XtFree(H->path);
       XtFree (H);
       return (NULL);
    }

    return (H);
}

/* \subsection{\"create_helptopics(file,H,root)"} Allocate a
   read the help text from file, setup pointers and etc,
   setup subtopic info.

   Returns the HelpTop. */

#define MAXARGS 50

Private int check_defs();
#define DEFS_TRUE (check_defs(&args[prmi], defs, ndefs))
#define DEFS_FALSE (!check_defs(&args[prmi], defs, ndefs))

Private
HelpTop create_helptopics(file,H,root,defs,ndefs)
char *file;
HelpSet H;
HelpTop root;
char **defs;
int ndefs;
{
  HelpTop T;
  FILE *helpfil = NULL;
  int help_len;
  int nb;
  int nt = 0;
  int ct;
  int tl;
  char *helpmem;
  char *hp, *he;
  char *s,*l;
  char fullname[1024];
  int nlines;
  char *cmd, *prm;
  int prmi;
  int i;
  char *args[MAXARGS];
  int argn;
 
  int ifc[10], ifg[10], ifp;
 
   /* Get memory, read the help file,
      and fill the list and topic structures */

   strcat(strcpy(fullname,H->path),file);
   helpfil = fopen(fullname,"r");

   if (!helpfil) {
      pop_warning("Could not open help file: %s", fullname, search_shell);
      return (NULL);
   }

   fseek(helpfil,0,2);
   help_len = ftell(helpfil);
   fseek(helpfil,0,0);
   
   /* We might possibly need more than \'help_len' space due
      to the conversion of null lines to blank lines (see below).
      The 1000 is that extra space---sloppy, but it seems the 
      most efficient. */

   nb = help_len + 1000 + sizeof(HelpTop_);
   helpmem = (char *)XtMalloc(nb);
   T = (HelpTop)helpmem;
   T->helpset = (caddr_t) H;

   ifc[0] = ifg[0] = 1;
   ifp = 0;
     
   /* Now parse the help text into the buffer we allocated. */

   hp = helpmem+sizeof(HelpTop_);
   he = hp + help_len + 1000;

   while ((hp<he) && fgets(hp,he-hp,helpfil)) {

      switch (*hp) {

       /* Comment */

       case '#': {
           break;
       }

       /* New topic */

       case '!' : {

          if (!ifg[ifp]) break;

          if (nt<MAX_TOPICS) {
            if (nt!=0) {
               T->nlines[ct] = nlines;
               T->len[ct] = (hp-(T->topics[ct]));
               *hp++ = '\0';
            } else {
               hp++;  /* skip the `!' */
            }
            T->topics[nt] = hp;
            T->subtop[nt] = NULL;
            T->bmfile[nt] = NULL;
            nlines = 0;
            ct = nt++;
            tl = strlen(hp);
            hp += tl-1;
            while (isspace(*hp)) hp--;
            *++hp = '\0';
            hp++;
          }
          break;
       }

       /* command */

       case '$': {
          cmd = hp+1;
          if (l=strchr(cmd,'\n')) *l = '\0';
          for (i=0;i<MAXARGS;i++) {
             while (isspace(*cmd)) cmd++;
             args[i] = cmd;
             while (*cmd && (*cmd!=' ')) cmd++;
             if (*cmd) *cmd++ = '\0';
          }
          cmd = args[0];
          argn = 1;
         
          /* allow some ligatures */

          if (strcmp(args[argn],"if")==0) {
             strcat(args[argn-1],args[argn]);
             argn++;
          }
          if (strcmp(args[argn],"not")==0) {
             strcat(args[argn-1],args[argn]);
             argn++;
          }
          prmi = argn;
          prm = args[argn];

          /* subtopic menu */

          if (!strcmp(cmd,"help")) {
             if (!ifg[ifp]) break;
             if (!prm) pop_warning("No filename", NULL, search_shell);
             else T->subtop[ct] = create_helptopics(prm,H,T,defs,ndefs);
          }

          /* Bitmap file */

          else if (!strcmp(cmd,"picture")) {
             if (!ifg[ifp]) break;
             if (!prm) pop_warning("No filename", NULL, search_shell);
             else {
                T->bmfile[ct] = (char*) XtMalloc(strlen(prm)+1);
                strcpy(T->bmfile[ct],prm);
             }
          }

          /* control directive */

          else if (!strcmp(cmd,"if")) {
             if (DEFS_TRUE) ifc[++ifp] = 1;
             else ifc[++ifp] = 0;
             ifg[ifp] = ifg[ifp-1];
             if (ifg[ifp]) ifg[ifp] = ifc[ifp];
          }

          else if (!strcmp(cmd,"ifnot")) {
             if (DEFS_FALSE) ifc[++ifp] = 1;
             else ifc[++ifp] = 0;
             ifg[ifp] = ifg[ifp-1];
             if (ifg[ifp]) ifg[ifp] = ifc[ifp];
          }

          else if (!strcmp(cmd,"elseif")) {
             if (ifc[ifp]) ifg[ifp] = 0;
             else {
               if (DEFS_TRUE) ifc[ifp] = 1;
               ifg[ifp] = ifg[ifp-1];
               if (ifg[ifp]) ifg[ifp] = ifc[ifp];
             }
          }

          else if (!strcmp(cmd,"elseifnot")) {
             if (ifc[ifp]) ifg[ifp] = 0;
             else {
               if (DEFS_FALSE) ifc[ifp] = 1;
               ifg[ifp] = ifg[ifp-1];
               if (ifg[ifp]) ifg[ifp] = ifc[ifp];
             }
          }

          else if (!strcmp(cmd,"else")) {
             ifc[ifp] = 1-ifc[ifp];
             if (ifg[ifp-1]) ifg[ifp] = ifc[ifp];
          }

          else if (!strcmp(cmd,"endif")) {
             if (--ifp < 0) pop_warning("unexpected endif",NULL,search_shell);
          }

          else pop_warning("Unrecognized directive: %s", cmd, search_shell);

          break;
       }

       /* else just help text */

       case '\n': {     /* Label ignores empty lines */
          *hp++ = ' ';
          *hp++ = '\n';
          break;
       }
       default: {
          if (!ifg[ifp]) break;
          if (nt) {
             hp += strlen(hp);
             nlines++;
          } else { 
             if (strncmp(hp,"width ",6)==0) {
                hp+=5;
                while (isspace(*++hp));
                if (!H->width) {
                   sscanf(hp,"%d",&H->width);
                   PRINT("width is %s %d\n",hp, H->width);
                }
             } else if (strncmp(hp,"height ",7)==0) {
                if (!H->height) {
                   sscanf(hp,"height %d",&H->height);
                   PRINT("height is %s %d\n",hp, H->height);
                }
             }
          }
       }
     } /* end switch */
   }
   fclose(helpfil);
         
   /* finish setting up */

   if (nt==0) {
      pop_warning("Invalid help file", NULL, search_shell);
      return (NULL);
   }
   if (ifp!=0) {
      pop_warning("unbalanced if directives", NULL, search_shell);
      return (NULL);
   }
   *hp++ = '\0';;
   T->nlines[ct] = nlines;
   T->len[ct] = (hp-(T->topics[ct]));
   T->topics[nt] = NULL;
   T->len[nt] = 0;
   T->ntopics = nt;
   if (nt>H->maxitems) H->maxitems = nt;

   return (T);
}

/* \subsection{\"check_defs()"} Check the defs argument. */

#define OP_OR  1
#define OP_AND 2

Private
int check_defs(args,defs,ndefs)
char **args;
char **defs;
int ndefs;
{
   int p,i;
   int op = 0;
   int v, vv;
   char c;

   for (p=0;*args[p];p++) {
      for (i=0;i<ndefs;i++) if (!strcmp(args[p],defs[i])) break;
      vv = (i<ndefs);
      if (op==OP_OR) v = v || vv;
      else if (op==OP_AND) v = v && vv;
      else v = vv;
      
      if (c = *args[++p]) {
         if (c=='|') op = OP_OR;
         else if (c=='&') op = OP_AND;
         else op = 0;   /* error actually */
      } else break;
   }
   return (v);
}
    
     
   
   
/* \section{X tools}  These are some general purpose tools
   that possibly should go into a library sometime. */

/* \subsection{\"position_popup(w,xmargin,ymargin)"} Assure that the
   popup widget \"w" will be on the screen.  With right margin \|xmargin|
   and bottom margin \|ymargin|.
 */

Private
position_popup(w,xmargin,ymargin)
Widget w; 
int xmargin;
int ymargin;
{
   int movd = 0;
   int o;
   Position x,y;
   Position vwx,vwy;
   int screen_width;           /* width of user's screen */
   int screen_height;          /* height of user's screen */
   Display *display = XtDisplay(w);

   screen_width = DisplayWidth(display,DefaultScreen(display));
   screen_height = DisplayHeight(display,DefaultScreen(display));

   XtRealizeWidget(w);
   width = height = 0;
   XtGetValues(w,wh_args,XtNumber(wh_args));
   XtTranslateCoords(w,
      (Position)(0), (Position)(0), &x, &y);
   PRINT("P: w is %d/%d at %d,%d\n",width,height,x,y);
   
   /* Check horizontal positioning. */

   if ((o=(x+width-screen_width+xmargin))>0) {
      PRINT("adjusting x by %d\n",o);
      x -= o;
      movd = 1;
   }
   
   /* Check vertical positioning. */

   if ((o=(y+height-screen_height+ymargin))>0) {
      PRINT("adjusting y by %d\n",o);
      y -= o;
      movd = 2;
   }

   /* Move the widget */

   if (movd) {
      if (y<0) y = 0;
      if (x<0) x = 0;
      XtMoveWidget(w,x,y);
   }


}

/* \subsection{\"get_the_pixmap()"} Read a bitmap file and
   generate an appropriate pixmap for the background of
   the window. */

/* #define SYNC XSync(XtDisplay(w),False) */
#define SYNC

Pixmap get_the_pixmap(w,path,file)
Widget w;
char *path;
char *file;
{
   int n;
   unsigned int bmw, bmh;
   int  bmx, bmy;
   XGCValues vals;
   Pixel fg,bg;
   int dx,dy,dw,dh;
   Pixmap bmmap, bmpix;
   GC gc;
   char fullname[1024];
   char cmd[1024];
   FILE *bmfile;
   char bmident[4];
   char *data;
   struct stat stat_dummy;

   /* Read the bitmap from the file.  If the first word of the
      file is the code `*BM*' then we have a binary file
      generated by \"cvtxbm". */

   strcat(strcpy(fullname,path),file);
   bmmap = NULL;

   /* Check if the file has .gz extension, then read file as 
      from forked gunzip, otherwise just open the file */
   strcat(fullname, ".gz");
   if (stat(fullname, &stat_dummy) == 0){
     sprintf(cmd, "gunzip -c %s", fullname);
     if (!(bmfile=popen(cmd, "r"))) {
       pop_warning("Couldn't open %s", cmd, search_shell);
       return (NULL);
     }
   }
   else {
     strcat(strcpy(fullname,path),file);
     if (!(bmfile=fopen(fullname,"r"))) {
       pop_warning("Couldn't open %s", fullname, search_shell);
       return (NULL);
     }
   }
   PRINT("Incl picture from [%s]\n",fullname);

   fread(bmident,4,1,bmfile);
   if (strncmp(bmident,"*BM*",4)) {
     /* character bitmap file */
     fclose(bmfile);
     PRINT(" <character foramt>\n");
     n = XReadBitmapFile(XtDisplay(w),
         XtWindow(w),
         fullname,
         &bmw, &bmh, &bmmap, &bmx, &bmy);
     SYNC;
     PRINT("  stat=%d, w=%d, h=%d, map=%x\n",
         n, bmw, bmh, bmmap);
     if (n) return (NULL);
   } else {
     /* binary bitmap file */
     PRINT(" <binary foramt>\n");
     fread(&bmw,4,1,bmfile);
     fread(&bmh,4,1,bmfile);
     fread(&n,4,1,bmfile);

     bmw = ntohl(bmw);
     bmh = ntohl(bmh);
     n = ntohl(n);

     PRINT("  size=%d, w=%d, h=%d\n",n, bmw, bmh);
     if (!(data=(char*)XtMalloc(n))) {
        pop_warning("No bitmap malloc", NULL, search_shell);
        return (NULL);
     }
     fread(data,1,n,bmfile);
     fclose(bmfile);
     bmmap = XCreateBitmapFromData(XtDisplay(w),
         XtWindow(w), data, bmw, bmh);
     SYNC;
     if (!bmmap) {
         pop_warning("CreateBitmap failed", NULL, search_shell);
         XtFree(data);
         return (NULL);
     }
     XtFree(data);

   }

   /* Get some screen values */

   n=0;
   width = height = 0;
   ARGN(XmNwidth, &width);
   ARGN(XmNheight, &height);
   ARGN(XmNdepth, &bmy);
   ARGN(XmNforeground, &fg);
   ARGN(XmNbackground, &bg);
   XtGetValues(w, args, n);
   PRINT("depth=%d, w/h = %d/%d\n",bmy, width,height);
   n = 0;
/* #ifdef NOTDEF */
   if (width<bmw) {
      ARGN(XmNwidth, bmw);
      width = bmw;
   }
   if (height<bmh) {
      ARGN(XmNheight, bmh);
      height = bmh;
   }
   if (n) XtSetValues(w, args, n);
/* #endif */

   /* Get a clear pixmap with background color */

   vals.foreground = bg;
   gc = XtGetGC(w, GCForeground, &vals);

   bmpix = XCreatePixmap(XtDisplay(w),
       XtWindow(w), width, height, bmy);
   SYNC;
   if (!bmpix) {
       XtReleaseGC(w,gc);
       XFreePixmap(XtDisplay(w),bmmap);
       return (NULL);
   }

   XFillRectangle(XtDisplay(w),
       bmpix, gc, 0, 0, width, height);
   XtReleaseGC(w,gc);

   /* Copy the bitmap into the pixmap */

   if ((width-bmw)>=0) {
      dx = (width-bmw);
      dw = bmw;
   } else {
      dx = 0;
      dw = width;
   }
   if ((height-bmh)>=0) {
      dy = (height-bmh);
      dh = bmh;
   } else {
      dy = 0;
      dh = height;
   }
  
   dx = dy = 0;

   vals.foreground = fg;
   vals.background = bg;
   gc = XtGetGC(w, GCForeground|GCBackground, &vals);
   SYNC;

   XCopyPlane(XtDisplay(w),
       bmmap, bmpix, gc, 0, 0, dw, dh, dx, dy, 1);

   XtReleaseGC(w,gc);
   XFreePixmap(XtDisplay(w),bmmap);

   return (bmpix);
}


/* \endc */

