/***********************************************************************
 *                                                                     *
 *              WILLOW(tm) Information Retrieval Software              *
 *     (Washington Information Looker-upper Layered Over Windows)      *
 *             Copyright 1992-1994 University of Washington            *
 *                                                                     *
 *   Willow is a trademark of the University of Washington.  No        *
 *   commercial use of this trademark or software may be made without  *
 *   prior written permission of the University of Washington.         *
 *                                                                     *
 * Permission to use, copy, modify, and distribute Willow(tm) software *
 * and its documentation for any noncommercial purpose is hereby       *
 * granted, provided that the above copyright notice appears in all    *
 * copies and that both the above copyright notice and this permission *
 * notice appear in supporting documentation, and that the name of the *
 * University of Washington not be used in advertising or publicity    *
 * pertaining to distribution of the software without specific,        *
 * written prior permission.  This software is made available "as is", *
 * and THE UNIVERSITY OF WASHINGTON DISCLAIMS ALL WARRANTIES, EXPRESS  *
 * OR IMPLIED, WITH REGARD TO THIS SOFTWARE, INCLUDING WITHOUT         *
 * LIMITATION ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS    *
 * FOR A PARTICULAR PURPOSE, AND IN NO EVENT SHALL THE UNIVERSITY OF   *
 * WASHINGTON BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL     *
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA  *
 * OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, TORT (INCLUDING       *
 * NEGLIGENCE) OR STRICT LIABILITY, ARISING OUT OF OR IN CONNECTION    *
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.                       *
 *                                                                     *
 * For further information please contact:                             *
 * Willow Project                                                      *
 * Information Systems                                                 *
 * University of Washington                                            *
 * 4545 15th NE, 1st floor                                             *
 * Seattle, Washington, 98105, USA                                     *
 * Internet: willow@cac.washington.edu                                 *
 ***********************************************************************/

/*******************************************************
 * mail.c
 * This file contains routines for mailing full record
 * results and summaries
 ******************************************************/

#include <stdio.h>
#include <sys/types.h>

#include <Mrm/MrmAppl.h>
#include <Xm/Text.h>

#include "willow.h"

#define IN  0
#define OUT 1

extern Widget widgets[], search_shell, summary_shell, results_shell, toplevel;
extern XtAppContext app_context;
extern char *summaries[];
extern Boolean saving;
extern Boolean show_search;

void mail_read(caddr_t data, int *fid, XtInputId *iid);
extern void clear_dialog_cb (Widget w, Widget target, u_long *reason);
extern int write (int, void *, unsigned int);
extern int close (int);
extern Boolean list_get_selected (Widget w, int **pos_list, int *num);
extern void force_summaries (void);
extern int desensitize_selection_buttons (void);
extern int pipe (int *);
#ifdef vax
extern int vfork ();
#else
extern pid_t vfork (void);
#endif
extern int dup2 (int, int);
extern int execlp (char *, char *, ...);
extern int read (int, void *, unsigned int);
extern int sensitize_selection_buttons (void);
extern int not_ready_buttons (void);
extern int clear_working (void);
extern int ready_buttons (void);

/* Pipes into mail, and error back */
int to_mail[2],
    from_mail[2];

/* E-Mail Address we send to */
char *mail_addr = NULL;

/* Subject line we use */
char *mail_subj = NULL;

/* How much text is currently in the full-rec window */
extern int text_len;

extern char *get_query_display_str(void);
extern void pop_warning (char *template, char *arg, Widget parent);
int mail_full (void);
int mail_summ (void);
int open_mail_pipe (void);
extern int pop_working (char *message, Widget parent);
extern int error (char *msg);

/* Are we currently doing a save of full results? */
extern Boolean popped_full;


/* Unmanage the save box and collect new parameters */
void mail_ok_cb(Widget w, int *tag, u_long *reason)
{
  clear_dialog_cb(w, widgets[k_save_box], reason);
  show_search = XmToggleButtonGetState(widgets[k_show_search_btn]);

  /* If this is a *real* callback, we need to free the old mail_addr */
  if (reason) {
    XtFree(mail_addr);
    XtFree(mail_subj);
  }
  mail_addr = XmTextGetString(widgets[k_email_text]);
  mail_subj = XmTextGetString(widgets[k_email_subj]);

  /* If address has been completely specified, display name */
  if (mail_addr[0]) {
    if (popped_full)
      mail_full();
    else
      mail_summ();
  }
  else if (reason) /* Only pop warning if this is a *real* callback */
    pop_warning("Incomplete e-mail address specification", NULL,
		popped_full ? results_shell : summary_shell );
}

/* Send a full-record result through a one way pipe to the mail command.
 */
mail_full(void)
{
  char *text;

  open_mail_pipe();
  text = XmTextGetString(widgets[k_text]);

  write(to_mail[OUT], get_query_display_str(), 
	strlen(get_query_display_str()));
  write(to_mail[OUT], text, strlen(text));
  close(to_mail[OUT]);
  XtFree(text);
}

/* Send the summaries through one way pipe to mail command. */
mail_summ(void)
{
  char *result;
  int  *position_list;
  int   position_count, i;


  if (list_get_selected(widgets[k_summary_titles], &position_list, 
			   &position_count)) {
    open_mail_pipe();
    write(to_mail[OUT], get_query_display_str(), 
	  strlen(get_query_display_str()));
    force_summaries();
    for (i = 0; i < position_count; i++) {
      result = summaries[position_list[i]];
      write(to_mail[OUT], result, strlen(result));
      write(to_mail[OUT], "\n", 1);
    }
    close(to_mail[OUT]);
    XtFree(position_list);
  }
}

/* Open a pipe to the mail command */
open_mail_pipe(void)
{
  int mail_pid = 0;
  char cmd_buf[2*MAX_LINE];

  saving = TRUE;
  XtSetSensitive(widgets[k_full_save_btn], FALSE);
  desensitize_selection_buttons();
  not_ready_buttons();

  pop_working("Save via E-Mail in progress...", 
	      popped_full ? results_shell : summary_shell );

  /* set up pipes for communicating with mail */
  pipe(to_mail);
  pipe(from_mail);
   
  /* fork off a child, which will overlay mail on itself */
  mail_pid = vfork();

  if (mail_pid == 0) {
    /* in the child, redirect stdin, stdout, and stderr. */
    dup2(to_mail[IN], fileno(stdin));
    dup2(from_mail[OUT],fileno(stdout));
    dup2(from_mail[OUT],fileno(stderr));
    close(to_mail[IN]);
    close(to_mail[OUT]); 
    close(from_mail[IN]); 
    close(from_mail[OUT]);

    /* Exec sh, pass in mail command */
    sprintf(cmd_buf, DEFAULT_MAIL_CMD, mail_subj, mail_addr);
    if (execlp("/bin/sh", "sh", "-c", cmd_buf, NULL) < 0) {
      /* at this point we are in a child process, so we can't do much */
      perror(NULL);
      /* Must _exit after vfork */
      _exit(-1);
    }
  }
  else if (mail_pid > 0) {
    /* in the parent, close extra pipes */
    close(to_mail[IN]);
    close(from_mail[OUT]);
    
    /* Now add a callback to read from pipe whenever something
       appears */
    XtAppAddInput(app_context, from_mail[IN], 
		  XtInputReadMask, mail_read, NULL);
  }
  else {
    perror(NULL);
    error("Could not fork sh/mail");
  }
}

/* Read from standard error from mail process. 
 */
void mail_read(caddr_t data, int *fid, XtInputId *iid)
{
  static char mail_msg_buf[CHUNK_SIZE];
  int  n = 0;
  static int err_len = 0;

  clear_working();
  saving = FALSE;

  /* The -1 is so there is always room to append a \0 */
  n = read(from_mail[IN], mail_msg_buf + err_len, 
	   sizeof(mail_msg_buf) - err_len - 1);  

  if (n)
    err_len += n;
    
  else {
    sensitize_selection_buttons();
    ready_buttons();
    XtSetSensitive(widgets[k_full_save_btn], TRUE);

    XtRemoveInput(*iid);
    close(from_mail[IN]);

    if (err_len) {
      /* Got an error message, mail failed */
      mail_msg_buf[err_len] = 0;
      pop_warning("Mailing Error:\n%s", mail_msg_buf, 
		popped_full ? results_shell : summary_shell );
      err_len = 0;
    }
  }
}
