/*
   Finds a window and messes with it.

   1) Set TARGET to the name of the target window
   1) Set MSGTYP to the ClientMessage type

   2) Run this program
      at the prompt
 
      >>q       quit
      >>i       iconify the window
      >>u       Unmap the window
      >>m       Map the window
      >>sxxxxx  send ClientMessage event with xxxxx as data

      */

#define TARGET "Summary List"
#define MSGTYP "Willow"

#define RESET_CM      0
#define MAP_CM    1
#define UNMAP_CM  2
#define SELECT_DB_CM  3
#define QUIT_CM   4

#include <stdio.h>
#include <ctype.h>
#include <X11/Xos.h>
#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <X11/Xutil.h>

char *ProgramName;

static void usage ()
{
    fprintf (stderr,
	     "usage:  %s  [-display dpy] \n", ProgramName);
    exit (1);
}

Window xev_win = NULL;

char resp[128];

XEvent msg;

main (argc, argv)
    int argc;
    char **argv;
{
    int i;
    char *displayname = NULL;
    Bool all_screens = True;
    Bool verbose = False;
    Display *dpy;
    int maxcmdlen = 10000;
    Atom xa_HELLO;

    int x,y;
    unsigned int w, h, bw, d;
    Window rw;

    ProgramName = argv[0];

    dpy = XOpenDisplay (displayname);
    if (!dpy) {
	fprintf (stderr, "%s:  unable to open display \"%s\"\r\n",
		 ProgramName, XDisplayName (displayname));
	exit (1);
    }

    xa_HELLO = XInternAtom(dpy,MSGTYP, False);

    printf("Looking for %s\n", TARGET);
    lookat (dpy, RootWindow(dpy,0), verbose, maxcmdlen);


    msg.xclient.type = ClientMessage;
    msg.xclient.message_type = xa_HELLO;
    msg.xclient.format = 8;
    msg.xclient.window = xev_win;

    if (xev_win) {
      XTranslateCoordinates(dpy, xev_win, RootWindow(dpy,0),
         0, 0, &x, &y, &rw);
      printf(".. x/y=%d/%d   w/h=%d/%d   bw=%d\n",
         x,y,w,h,bw);
    }
    while (xev_win) {
      
      printf(">>"); fflush(stdout);
      if (!gets(resp)) break;
      if (*resp=='i') XIconifyWindow(dpy,xev_win,0);
/*      if (*resp=='m') XMapWindow(dpy,xev_win); */
      if (*resp=='u') {
	msg.xclient.data.b[0] = UNMAP_CM;
	XSendEvent(dpy,xev_win, False, -1, &msg);
      }
      if (*resp=='m') {
	msg.xclient.data.b[0] = MAP_CM;
	XSendEvent(dpy,xev_win, False, -1, &msg);
      }
      if (*resp=='r') {
	msg.xclient.data.b[0] = RESET_CM;
	XSendEvent(dpy,xev_win, False, -1, &msg);
      }
      if (*resp=='s') {
	msg.xclient.data.b[0] = SELECT_DB_CM;
	strcpy(msg.xclient.data.b+1,resp+1);
	XSendEvent(dpy,xev_win, False, -1, &msg);
      }
      if (*resp=='q') {
	msg.xclient.data.b[0] = QUIT_CM;
	XSendEvent(dpy,xev_win, False, -1, &msg);
      }
      XFlush(dpy);
    }

    XCloseDisplay (dpy);
    exit (0);
}


lookat (dpy, root, verbose, maxcmdlen)
    Display *dpy;
    Window root;
    Bool verbose;
    int maxcmdlen;
{
    Window dummy, *children = NULL, client;
    unsigned int i, nchildren = 0;


    /*
     * then, get the list of windows
     */
    if (!XQueryTree (dpy, root, &dummy, &dummy, &children, &nchildren)) {
	return;
    }

    for (i = 0; i < nchildren; i++) {
	client = XmuClientWindow (dpy, children[i]);
	if (client != None)
	  check_client_properties (dpy, client, verbose, maxcmdlen);
    }
}

static char *Nil = "(nil)";

check_client_properties (dpy, w, verbose, maxcmdlen)
    Display *dpy;
    Window w;
    Bool verbose;
    int maxcmdlen;
{
    char **cliargv = NULL;
    int i, cliargc;
    XTextProperty nametp, machtp, tp;
    int charsleft = maxcmdlen;

	if (XGetWMName (dpy, w, &tp)) {
	    print_text_field (dpy, "  Icon Name:  ", &tp);
            if (strcmp(tp.value,TARGET)==0) {
               xev_win = w;
               printf("... that's the one %x\n",w);
            }
	    if (tp.value) XFree ((char *) tp.value);
	}


}


print_text_field (dpy, s, tp)
    Display *dpy;
    char *s;
    XTextProperty *tp;
{
    if (tp->encoding == None || tp->format == 0) {
	printf ("''");
	return;
    }

    if (s) printf ("%s", s);
    if (tp->encoding == XA_STRING && tp->format == 8) {
	printf ("%s", tp->value ? (char *) tp->value : Nil);
    } else {
	unknown (dpy, tp->encoding, tp->format);
    }
    if (s) putchar ('\n');
}

unknown (dpy, actual_type, actual_format)
    Display *dpy;
    Atom actual_type;
    int actual_format;
{
    char *s;

    printf ("<unknown type ");
    if (actual_type == None) printf ("None");
    else if ((s = XGetAtomName (dpy, actual_type)) != NULL) {
	fputs (s, stdout);
	XFree (s);
    } else {
	fputs (Nil, stdout);
    }
    printf (" (%ld) or format %d>", actual_type, actual_format);
}


