/***********************************************************************
 *                                                                     *
 *              WILLOW(tm) Information Retrieval Software              *
 *     (Washington Information Looker-upper Layered Over Windows)      *
 *             Copyright 1992-1994 University of Washington            *
 *                                                                     *
 *   Willow is a trademark of the University of Washington.  No        *
 *   commercial use of this trademark or software may be made without  *
 *   prior written permission of the University of Washington.         *
 *                                                                     *
 * Permission to use, copy, modify, and distribute Willow(tm) software *
 * and its documentation for any noncommercial purpose is hereby       *
 * granted, provided that the above copyright notice appears in all    *
 * copies and that both the above copyright notice and this permission *
 * notice appear in supporting documentation, and that the name of the *
 * University of Washington not be used in advertising or publicity    *
 * pertaining to distribution of the software without specific,        *
 * written prior permission.  This software is made available "as is", *
 * and THE UNIVERSITY OF WASHINGTON DISCLAIMS ALL WARRANTIES, EXPRESS  *
 * OR IMPLIED, WITH REGARD TO THIS SOFTWARE, INCLUDING WITHOUT         *
 * LIMITATION ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS    *
 * FOR A PARTICULAR PURPOSE, AND IN NO EVENT SHALL THE UNIVERSITY OF   *
 * WASHINGTON BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL     *
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA  *
 * OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, TORT (INCLUDING       *
 * NEGLIGENCE) OR STRICT LIABILITY, ARISING OUT OF OR IN CONNECTION    *
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.                       *
 *                                                                     *
 * For further information please contact:                             *
 * Willow Project                                                      *
 * Information Systems                                                 *
 * University of Washington                                            *
 * 4545 15th NE, 1st floor                                             *
 * Seattle, Washington, 98105, USA                                     *
 * Internet: willow@cac.washington.edu                                 *
 ***********************************************************************/

/******************************************************
 protocol.c
 This file contains routines to put the willow/driver
 protocol packets in and out of network byte order.
 ******************************************************/

#include <sys/types.h>
#include <netinet/in.h>

#include "protocol.h"

packet_rec *netify(packet_rec *packet);
packet_rec *denetify(packet_rec *packet);


/* Put all netint (4 byte ints) into network byte order. We return
   a copy of the packet, kept in static memory here.
 */
packet_rec *netify(packet_rec *packet)
{
  static packet_rec new_pack;

#if defined(POSIX) || defined (_POSIX_SOURCE)
    memmove(&new_pack, packet, PACKET_NET_SIZE);
#else
    bcopy(packet, &new_pack, PACKET_NET_SIZE);
#endif
  
  new_pack.type = htonl(new_pack.type);
  new_pack.data_len = htonl(new_pack.data_len);

  /* Since all the record variants consist of a bunch of 4-byte ints,
     we can ignore what type the record actually is for conversion 
     purposes.
   */
  new_pack.var.conversion.int1 = htonl(new_pack.var.conversion.int1);
  new_pack.var.conversion.int2 = htonl(new_pack.var.conversion.int2);
  new_pack.var.conversion.int3 = htonl(new_pack.var.conversion.int3);
  new_pack.var.conversion.int4 = htonl(new_pack.var.conversion.int4);

  return(&new_pack);
}

/* Put the netints into host order. Return the same packet that was
   passed in (unlike netify()).
 */
packet_rec *denetify(packet_rec *packet)
{
  packet->type = ntohl(packet->type);
  packet->data_len = ntohl(packet->data_len);

  /* Since all the record variants consist of a bunch of 4-byte ints,
     we can ignore what type the record actually is for conversion 
     purposes.
   */
  packet->var.conversion.int1 = ntohl(packet->var.conversion.int1);
  packet->var.conversion.int2 = ntohl(packet->var.conversion.int2);
  packet->var.conversion.int3 = ntohl(packet->var.conversion.int3);
  packet->var.conversion.int4 = ntohl(packet->var.conversion.int4);
  return(packet);
}
