/***********************************************************************
 *                                                                     *
 *              WILLOW(tm) Information Retrieval Software              *
 *     (Washington Information Looker-upper Layered Over Windows)      *
 *             Copyright 1992-1994 University of Washington            *
 *                                                                     *
 *   Willow is a trademark of the University of Washington.  No        *
 *   commercial use of this trademark or software may be made without  *
 *   prior written permission of the University of Washington.         *
 *                                                                     *
 * Permission to use, copy, modify, and distribute Willow(tm) software *
 * and its documentation for any noncommercial purpose is hereby       *
 * granted, provided that the above copyright notice appears in all    *
 * copies and that both the above copyright notice and this permission *
 * notice appear in supporting documentation, and that the name of the *
 * University of Washington not be used in advertising or publicity    *
 * pertaining to distribution of the software without specific,        *
 * written prior permission.  This software is made available "as is", *
 * and THE UNIVERSITY OF WASHINGTON DISCLAIMS ALL WARRANTIES, EXPRESS  *
 * OR IMPLIED, WITH REGARD TO THIS SOFTWARE, INCLUDING WITHOUT         *
 * LIMITATION ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS    *
 * FOR A PARTICULAR PURPOSE, AND IN NO EVENT SHALL THE UNIVERSITY OF   *
 * WASHINGTON BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL     *
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA  *
 * OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, TORT (INCLUDING       *
 * NEGLIGENCE) OR STRICT LIABILITY, ARISING OUT OF OR IN CONNECTION    *
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.                       *
 *                                                                     *
 * For further information please contact:                             *
 * Willow Project                                                      *
 * Information Systems                                                 *
 * University of Washington                                            *
 * 4545 15th NE, 1st floor                                             *
 * Seattle, Washington, 98105, USA                                     *
 * Internet: willow@cac.washington.edu                                 *
 ***********************************************************************/

/*******************************************************
 * range.c
 * This file contains the routines for handling numeric range data,
 * including dates.
 ******************************************************/

#include <stdio.h>
#include <time.h>
#include <Mrm/MrmAppl.h>
#include <Xm/ToggleB.h>
#include <Xm/Text.h>

#include "willow.h"
#include "db.h"

extern int error (char *msg);
extern int atoi (char *);
int cur_year (void);

extern Widget     widgets[],
                  search_shell,
                  summary_shell,
                  dict_shell,
                  results_shell,
                  toplevel;
extern MrmHierarchy s_MrmHierarchy;  /* MRM database hierarchy ID */
extern MrmType dummy_class;         /* and class variable. */
extern db_desc_rec *cur_db;

static Widget cur_range_btn = NULL;

/* Reset all range selectors */
static void
range_reset(void)
{
  Widget        submenu;
  WidgetList    buttons;

  /* Reset option menus to first button */
  XtVaGetValues(widgets[k_month_range], XmNsubMenuId, &submenu, NULL);
  XtVaGetValues(submenu, XmNchildren, &buttons, NULL);
  XtVaSetValues(widgets[k_month_range], XmNmenuHistory, buttons[0], NULL);

  XtVaGetValues(widgets[k_day_range], XmNsubMenuId, &submenu, NULL);
  XtVaGetValues(submenu, XmNchildren, &buttons, NULL);
  XtVaSetValues(widgets[k_day_range], XmNmenuHistory, buttons[0], NULL);

  XtVaGetValues(widgets[k_year_range], XmNsubMenuId, &submenu, NULL);
  XtVaGetValues(submenu, XmNchildren, &buttons, NULL);
  XtVaSetValues(widgets[k_year_range], XmNmenuHistory, buttons[0], NULL);

  /* Clear the text */
  XmTextSetString(widgets[k_num_range], "");
}

/* Put up the range box 
 */
void 
pop_range_dlg_cb(Widget w, int *tag, u_long *reason)
{
  int cur_year_wid;
  char       label_str[MAX_LINE];
  char      *adjective;
  range_stat_rec *stat_ptr;

  if (widgets[k_range_dlg] == NULL) {
    if (MrmFetchWidget(s_MrmHierarchy, 
		       "range_dialog", 
		       search_shell,
		       &widgets[k_range_dlg], 
		       &dummy_class) != MrmSUCCESS)
      error("can't fetch range dialog widget");    

    /* Just a call to force retrieval of widgets */
    if (!reason) 
      return;
  }    
  else
    range_reset();

  cur_range_btn = w;

  /* Setup depending on type and values of corresponding button */
  XtVaGetValues(w, XmNuserData, &stat_ptr, NULL);

  /* First the two toggle button */
  if (stat_ptr->low)
    if (stat_ptr->field->range_type == NUMERIC)
      adjective = LOW_NUMERIC_LABEL;
    else
      adjective = LOW_DATE_LABEL;
  else
    if (stat_ptr->field->range_type == NUMERIC)
      adjective = HIGH_NUMERIC_LABEL;
    else
      adjective = HIGH_DATE_LABEL;
  
  /* Get rid of unused widgets */

  if (stat_ptr->field->first_year) {
    /* Put appropriate values in the menu pane for year pop-up */
    Widget  pane;
    int   year, which;
    char  year_str[16];
    WidgetList years;

    /* Need to refer to currently managed year widget */
    cur_year_wid = k_year_range;
    XtUnmanageChild(widgets[k_num_range]);
    XtManageChild(widgets[k_year_range]);

    XtVaGetValues(widgets[k_year_range], XmNsubMenuId, &pane, NULL);
    XtVaGetValues(pane, XmNchildren, &years, NULL);
    for (which = 0, year = stat_ptr->field->last_year; 
	 which < MAX_YEAR_RANGE; which++, year--) {
      if (year >= stat_ptr->field->first_year) {
	XtManageChild(years[which]);
	sprintf(year_str, "%d", year);
	XtVaSetValues(years[which], XmNuserData, year, 
		      XmNlabelString, XmStringCreateSimple(year_str), NULL);
	if (stat_ptr->numeric == year)
	  XtVaSetValues(widgets[k_year_range], XmNmenuHistory, years[which],
			NULL);
      }
      else 
	XtUnmanageChild(years[which]);
    }
  }
  else {
    cur_year_wid = k_num_range;
    XtUnmanageChild(widgets[k_year_range]);
    XtManageChild(widgets[k_num_range]);
  }

  XtVaSetValues(widgets[k_range_sep], 
		XmNtopWidget, widgets[cur_year_wid], NULL);
  switch (stat_ptr->field->range_type) {
    case NUMERIC:
      XtVaSetValues(widgets[k_year_label], 
		    XmNrightAttachment, XmATTACH_NONE,
		    XmNtopAttachment,   XmATTACH_FORM,
		    XmNlabelString, 
		    XmStringCreateSimple(stat_ptr->field->name),
		    NULL);
      XtVaSetValues(widgets[k_num_range], 
		    XmNtopAttachment, XmATTACH_FORM,
		    XmNleftWidget,    widgets[k_year_label],
		    XmNcolumns,       MAX_NUMERIC_DIGITS,
		    XmNmaxLength,     MAX_LINE, 
		    NULL);
      XtVaSetValues(widgets[k_year_range], 
		    XmNtopAttachment, XmATTACH_FORM,
		    XmNleftWidget,    widgets[k_year_label],
		    NULL);
      XtUnmanageChild(widgets[k_month_range]);
      XtUnmanageChild(widgets[k_month_label]);
      XtUnmanageChild(widgets[k_day_range]);
      XtUnmanageChild(widgets[k_day_label]);
      break;
    case YEAR:
      XtVaSetValues(widgets[k_year_label], 
		    XmNrightAttachment, XmATTACH_NONE,
		    XmNlabelString,     XmStringCreateSimple(YEAR_LABEL), 
		    XmNtopAttachment,   XmATTACH_FORM, 
		    NULL);
      XtVaSetValues(widgets[k_num_range], 
		    XmNcolumns, 4,    XmNmaxLength, 4, 
		    XmNtopAttachment, XmATTACH_FORM,
		    XmNleftWidget,    widgets[k_year_label], 
		    NULL);
      XtVaSetValues(widgets[k_year_range], 
		    XmNtopAttachment, XmATTACH_FORM,
		    XmNleftWidget,    widgets[k_year_label], 
		    NULL);
      XtUnmanageChild(widgets[k_month_range]);
      XtUnmanageChild(widgets[k_month_label]);
      XtUnmanageChild(widgets[k_day_range]);
      XtUnmanageChild(widgets[k_day_label]);
      break;
    case MONTH:
      XtVaSetValues(widgets[k_month_range], 
		    XmNtopAttachment, XmATTACH_FORM, NULL);
      XtVaSetValues(widgets[k_month_label], 
		    XmNtopAttachment, XmATTACH_FORM, NULL);
      XtVaSetValues(widgets[k_num_range], 
		    XmNcolumns, 4,    XmNmaxLength, 4, 
		    XmNtopAttachment, XmATTACH_WIDGET, 
		    XmNleftWidget,    widgets[k_month_label],
		    XmNtopWidget,     widgets[k_month_range], 
		    NULL);
      XtVaSetValues(widgets[k_year_range], 
		    XmNtopAttachment, XmATTACH_WIDGET, 
		    XmNleftWidget,    widgets[k_month_label],
		    XmNtopWidget,     widgets[k_month_range], 
		    NULL);
      XtVaSetValues(widgets[k_year_label], 
		    XmNlabelString,     XmStringCreateSimple(YEAR_LABEL),
		    XmNtopAttachment,   XmATTACH_WIDGET, 
		    XmNtopWidget,       widgets[k_month_range],
		    XmNrightAttachment, XmATTACH_WIDGET,
		    XmNrightWidget,     widgets[cur_year_wid], 
		    NULL);
      XtUnmanageChild(widgets[k_day_range]);
      XtUnmanageChild(widgets[k_day_label]);
      XtManageChild(widgets[k_month_range]);
      XtManageChild(widgets[k_month_label]);
      break;
    case DAY:
      XtVaSetValues(widgets[k_month_range], 
		    XmNtopAttachment, XmATTACH_FORM, NULL);
      XtVaSetValues(widgets[k_month_label], 
		    XmNtopAttachment, XmATTACH_FORM, NULL);
      XtVaSetValues(widgets[k_day_range], 
		    XmNtopAttachment, XmATTACH_WIDGET, 
		    XmNtopWidget, widgets[k_month_range], NULL);
      XtVaSetValues(widgets[k_day_label], 
		    XmNtopAttachment, XmATTACH_WIDGET, 
		    XmNtopWidget, widgets[k_month_range], NULL);
      XtVaSetValues(widgets[k_num_range], 
		    XmNcolumns, 4,XmNmaxLength, 4, 
		    XmNtopAttachment, XmATTACH_WIDGET, 
		    XmNtopWidget, widgets[k_day_range],
		    XmNleftWidget, widgets[k_month_label], NULL);
      XtVaSetValues(widgets[k_year_range], 
		    XmNtopAttachment, XmATTACH_WIDGET, 
		    XmNtopWidget, widgets[k_day_range],
		    XmNleftWidget, widgets[k_month_label], NULL);
      XtVaSetValues(widgets[k_year_label], 
		    XmNlabelString, XmStringCreateSimple(YEAR_LABEL),
		    XmNtopAttachment, XmATTACH_WIDGET, 
		    XmNtopWidget, widgets[k_day_range],
		    XmNrightAttachment, XmATTACH_WIDGET, 
		    XmNrightWidget, widgets[cur_year_wid], NULL);
      XtManageChild(widgets[k_month_range]);
      XtManageChild(widgets[k_month_label]);
      XtManageChild(widgets[k_day_range]);
      XtManageChild(widgets[k_day_label]);
      break;
  }

  sprintf(label_str, "%s %s", adjective, stat_ptr->field->name);
  XtVaSetValues(widgets[k_unlimit_range], XmNlabelString, 
		XmStringCreateSimple(label_str), 
		XmNset, !stat_ptr->limited, NULL);
  
  sprintf(label_str, "Specify %s", stat_ptr->field->name);
  XtVaSetValues(widgets[k_limit_range], XmNlabelString, 
		  XmStringCreateSimple(label_str), 
		  XmNset, stat_ptr->limited, NULL);

  if (!stat_ptr->limited)
    XtUnmanageChild(widgets[k_range_ctrl]);
  else
    XtManageChild(widgets[k_range_ctrl]);

  if (stat_ptr->month)
    XtVaSetValues(widgets[k_month_range], XmNmenuHistory, stat_ptr->month, NULL);
  if (stat_ptr->day)
    XtVaSetValues(widgets[k_day_range], XmNmenuHistory, stat_ptr->day, NULL);
  if (stat_ptr->numeric) {
    /* If we have pop-up year, initial value has already been set */
    if (!stat_ptr->field->first_year) {
      sprintf(label_str, "%d", stat_ptr->numeric);
      XmTextSetString(widgets[k_num_range], label_str);
    }
  }

  map_dialog(widgets[k_range_dlg]);
  ForceDialog(widgets[k_range_dlg]);
  XmProcessTraversal(widgets[k_num_range], XmTRAVERSE_CURRENT);
}

/* Called when the user toggles the "non-limited" button */
void toggle_range_limit_cb(Widget w, int *target, XmToggleButtonCallbackStruct *reason)
{
  if (reason->set)
    XtUnmanageChild(widgets[k_range_ctrl]);
  else
    XtManageChild(widgets[k_range_ctrl]);
  XmProcessTraversal(widgets[k_num_range], XmTRAVERSE_CURRENT);
}

/* Dialog Ok button pressed, or we could be unpacking a saved search.
   Reason tells us which. Set up button label to reflect state of
   dialog box widgets.
 */
void range_ok_cb(Widget w, int *tag, u_long *reason)
{
  range_stat_rec *stat_ptr;
  char           *numeric_str;
  char           *month_abbrev;
  char            label_str[MAX_LINE];
  Widget          range_btn, year;

  if (reason) {
    XtManageChild(widgets[k_range_ctrl]);
    XtUnmanageChild(widgets[k_range_dlg]);
    range_btn = cur_range_btn;
  }
  else
    range_btn = w;

  XtVaGetValues(range_btn, XmNuserData, &stat_ptr, NULL);

  /* If this a real callback, fill in stat_ptr from dialog box */
  if (reason) {
    stat_ptr->day_val = stat_ptr->month_val = stat_ptr->numeric = 0;
    stat_ptr->limited = XmToggleButtonGetState(widgets[k_limit_range]);
    if (stat_ptr->limited) {
      if (stat_ptr->field->range_type == DAY) {
	XtVaGetValues(widgets[k_day_range], XmNmenuHistory, 
		      &stat_ptr->day, NULL); 
	XtVaGetValues(stat_ptr->day, XmNuserData, &stat_ptr->day_val, NULL);
      }
      if (stat_ptr->field->range_type >= MONTH) {
	XtVaGetValues(widgets[k_month_range], XmNmenuHistory, 
		      &stat_ptr->month, NULL); 
	XtVaGetValues(stat_ptr->month, XmNuserData, &stat_ptr->month_val, NULL);
      }
      if (stat_ptr->field->first_year) {
	XtVaGetValues(widgets[k_year_range], XmNmenuHistory, &year, NULL); 
	XtVaGetValues(year, XmNuserData, &stat_ptr->numeric, NULL);
      }
      else {
	numeric_str = XmTextGetString(widgets[k_num_range]);
	stat_ptr->numeric = atoi(numeric_str);
	XtFree(numeric_str);
      }
      if (stat_ptr->field->range_type >= YEAR) {
	if (stat_ptr->numeric == 0)
	  stat_ptr->numeric = cur_year();
	else if (stat_ptr->numeric < 100)
	  stat_ptr->numeric += DEFAULT_RANGE_CENTURY;
      }
    }
  }
  /* Adjust button to reflect stat_ptr */
  if (stat_ptr->limited) {
    if (stat_ptr->field->range_type >= MONTH)
      switch (stat_ptr->month_val) {
        case 1: month_abbrev = "Jan.";break;
        case 2: month_abbrev = "Feb.";break;
        case 3: month_abbrev = "Mar.";break;
        case 4: month_abbrev = "Apr.";break;
        case 5: month_abbrev = "May";break;
        case 6: month_abbrev = "June";break;
        case 7: month_abbrev = "July";break;
        case 8: month_abbrev = "Aug.";break;
        case 9: month_abbrev = "Sept.";break;
        case 10: month_abbrev = "Oct.";break;
        case 11: month_abbrev = "Nov.";break;
        case 12: month_abbrev = "Dec.";break;
      }
    switch (stat_ptr->field->range_type) {
      case DAY:
        sprintf(label_str, "%s %d, %d", month_abbrev, stat_ptr->day_val, 
		stat_ptr->numeric); break;
      case MONTH:
        sprintf(label_str, "%s %d", month_abbrev, 
		stat_ptr->numeric); break;
      default:
        sprintf(label_str, "%d", stat_ptr->numeric); break;
    }
  }
  else {  /* Not limited */
    if (stat_ptr->low) {
      if (stat_ptr->field->range_type == NUMERIC)
	strcpy(label_str, LOW_NUMERIC_LABEL);
      else
	strcpy(label_str, LOW_DATE_LABEL);
    }
    else {
      if (stat_ptr->field->range_type == NUMERIC)
	strcpy(label_str, HIGH_NUMERIC_LABEL);
      else
	strcpy(label_str, HIGH_DATE_LABEL);
    }
  }
  XtUnmanageChild(range_btn);
  XtVaSetValues(range_btn, XmNlabelString, 
		XmStringCreateSimple(label_str), NULL);
  XtManageChild(range_btn);
}

void range_cancel_cb(Widget w, int *tag, u_long *reason)
{
  if (widgets[k_range_ctrl]) {
    XtManageChild(widgets[k_range_ctrl]);
    XtUnmanageChild(widgets[k_range_dlg]);
  }
}

/* figure out what year it is */
cur_year(void)
{
  time_t secs;
  struct tm *real_time;

  secs = time(NULL);
  real_time = localtime(&secs);
  return(real_time->tm_year + DEFAULT_RANGE_CENTURY);
}

/* When field button is pressed, check to see if it is a range, and
   if so, set stuff up. If reason == NULL, we are being called from
   clear_cb() or unpack_search(). */
void setup_range_cb(Widget w, int which, u_long *reason)
{
  XmString label;
  db_field_desc_rec *field;
  range_stat_rec *stat_ptr;

  /* Get the field pointer for this button */
  XtVaGetValues(w, XmNuserData, &field, NULL);

  if (!field)
    return;

  /* Make range picker visible/invisible as appropriate */
  if (!field->range_type) {
    XtUnmanageChild(widgets[k_range1 + which]);
    XtManageChild(widgets[k_field1 + which]);
    return;
  }
  XtManageChild(widgets[k_range1 + which]);
  XtUnmanageChild(widgets[k_field1 + which]);
  
  /* Turn off browse button */
  XtVaSetValues(widgets[k_list_btn], XmNlabelString,
		XmStringCreateLtoR(DFLT_LIST_BTN, XmSTRING_DEFAULT_CHARSET), NULL);
  XtSetSensitive(widgets[k_list_btn], FALSE);

  XtVaGetValues(widgets[k_from1 + which], XmNuserData, &stat_ptr, NULL);

  /* If there was no change, leave as is, unless we are being called 
     from clear_cb() or unpack_search() */
  if ((stat_ptr->field == field) && reason)
    return;

  /* Initialize the status records */
#if defined(POSIX) || defined (_POSIX_SOURCE)
  memset(stat_ptr, 0, sizeof(range_stat_rec));
#else /* defined(POSIX) || defined (_POSIX_SOURCE) */
  bzero(stat_ptr, sizeof(range_stat_rec));
#endif /* defined(POSIX) || defined (_POSIX_SOURCE) */
  stat_ptr->field = field;
  stat_ptr->low   = TRUE;

  XtVaGetValues(widgets[k_to1 + which], XmNuserData, &stat_ptr, NULL);
#if defined(POSIX) || defined (_POSIX_SOURCE)
  memset(stat_ptr, 0, sizeof(range_stat_rec));
#else /* defined(POSIX) || defined (_POSIX_SOURCE) */
  bzero(stat_ptr, sizeof(range_stat_rec));
#endif /* defined(POSIX) || defined (_POSIX_SOURCE) */
  stat_ptr->field = field;
  stat_ptr->low   = FALSE;

  /* Set button labels */
  if (field->range_type == NUMERIC)
    label = XmStringCreateSimple(LOW_NUMERIC_LABEL);
  else
    label = XmStringCreateSimple(LOW_DATE_LABEL);
  XtVaSetValues(widgets[k_from1 + which], XmNlabelString, label, NULL);
  XmStringFree(label);

  if (field->range_type == NUMERIC)
    label = XmStringCreateSimple(HIGH_NUMERIC_LABEL);
  else
    label = XmStringCreateSimple(HIGH_DATE_LABEL);
  XtVaSetValues(widgets[k_to1 + which], XmNlabelString, label, NULL);
  XmStringFree(label);
}

/* When re-creating a saved query off disk, figure out widget-based
   portion of range_status record */
Widget range_get_widget(int type, int val)
{
  Widget     pane;
  static WidgetList months = NULL, days = NULL;

  if (!widgets[k_range_dlg])
    pop_range_dlg_cb(NULL, NULL, NULL);

  if (type == MONTH) {
    if (!months) {
      XtVaGetValues(widgets[k_month_range], XmNsubMenuId, &pane, NULL);
      XtVaGetValues(pane, XmNchildren, &months, NULL);
    }
    if ((val >= 0) && (val <= 12))
      return(months[val ? val - 1 : 0]);
    else
      error("Bad Value in range_get_widget(); corrupt Query File?");
  }
  else if (type == DAY) {
    if (!days) {
      XtVaGetValues(widgets[k_day_range], XmNsubMenuId, &pane, NULL);
      XtVaGetValues(pane, XmNchildren, &days, NULL);
    }
    if ((val >= 0) && (val <= 31))
      return(days[val ? val - 1 : 0]);
    else
      error("Bad Value in range_get_widget(); corrupt Query File?");
  }
}

/* Convert current settings of both range buttons to text for 
   transmission to driver. Caller should free return value */
char *range_to_text(int which)
{
  range_stat_rec *from_range, *to_range;
  char  *text = NULL;
  
  XtVaGetValues(widgets[k_from1+which], XmNuserData, &from_range, NULL);
  XtVaGetValues(widgets[k_to1+which], XmNuserData, &to_range, NULL);
  
  if (!from_range->limited && !to_range->limited)
    return("");

  text = XtMalloc(MAX_LINE);
  sprintf(text, "%d/%d/%d %d/%d/%d", 
	  from_range->month_val, from_range->day_val, from_range->numeric,
	  to_range->month_val, to_range->day_val, to_range->numeric);
  return(text);
}
