/***********************************************************************
 *                                                                     *
 *              WILLOW(tm) Information Retrieval Software              *
 *     (Washington Information Looker-upper Layered Over Windows)      *
 *             Copyright 1992-1994 University of Washington            *
 *                                                                     *
 *   Willow is a trademark of the University of Washington.  No        *
 *   commercial use of this trademark or software may be made without  *
 *   prior written permission of the University of Washington.         *
 *                                                                     *
 * Permission to use, copy, modify, and distribute Willow(tm) software *
 * and its documentation for any noncommercial purpose is hereby       *
 * granted, provided that the above copyright notice appears in all    *
 * copies and that both the above copyright notice and this permission *
 * notice appear in supporting documentation, and that the name of the *
 * University of Washington not be used in advertising or publicity    *
 * pertaining to distribution of the software without specific,        *
 * written prior permission.  This software is made available "as is", *
 * and THE UNIVERSITY OF WASHINGTON DISCLAIMS ALL WARRANTIES, EXPRESS  *
 * OR IMPLIED, WITH REGARD TO THIS SOFTWARE, INCLUDING WITHOUT         *
 * LIMITATION ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS    *
 * FOR A PARTICULAR PURPOSE, AND IN NO EVENT SHALL THE UNIVERSITY OF   *
 * WASHINGTON BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL     *
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA  *
 * OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, TORT (INCLUDING       *
 * NEGLIGENCE) OR STRICT LIABILITY, ARISING OUT OF OR IN CONNECTION    *
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.                       *
 *                                                                     *
 * For further information please contact:                             *
 * Willow Project                                                      *
 * Information Systems                                                 *
 * University of Washington                                            *
 * 4545 15th NE, 1st floor                                             *
 * Seattle, Washington, 98105, USA                                     *
 * Internet: willow@cac.washington.edu                                 *
 ***********************************************************************/

/****************************************************************
 * resources.c
 * This file contains code for setting up Willow's initial
 * resources.
 ****************************************************************/

#include <Mrm/MrmAppl.h>
#ifdef SOLARIS
#include <string.h>
#else
#include <strings.h>
#endif
#include <stdio.h>

#include "resources.h"
#include "willow.h"

extern Widget toplevel;

/* These are global variables initialized through resources */
extern char *willow_dir;
extern char *db_config_file_name;
extern char *print_cmd;
extern char *mail_subj;
extern char *mail_addr;
extern char *ftp_host;
extern char *ftp_uid;
extern Boolean ftp_append;
extern Boolean show_search;
extern Boolean show_print_dialog;
extern char *user_info_list;
extern char *custom_fields_list;
extern int   spin_delay;
extern int   debug_level;
extern int   save_type;
extern Boolean enable_cheat;
extern int     interface_mode;
extern Boolean initial_echo;
extern Boolean abbrev_fields;
extern Boolean no_basic_mode;
extern int   browse_delay;
char *willow_size;

/* The annoying resource structure required by XtGetApplicationResources */
static OptionsRec r;
#define Offset(field) XtOffsetOf(OptionsRec, field)

/* Define an XtResource array for resources we will be looking at */
XtResource resources[] = {
  {
    "size", XtCString, XtRString, sizeof (char *),
    Offset(willow_size), XtRString, "default",
  },
  {
    "defaultFont", XtCString, XtRString, sizeof (char *),
    Offset(default_font), XtRString, "default",
  },
  {
    "buttonFont", XtCString, XtRString, sizeof (char *),
    Offset(button_font), XtRString, "default",
  },
  {
    "textFont", XtCString, XtRString, sizeof (char *),
    Offset(text_font), XtRString, "default",
  },
  {
    "defaultBackground", XtCString, XtRString, sizeof (char *),
    Offset(default_bg), XtRString, "SkyBlue1",
  },
  {
    "defaultForeground", XtCString, XtRString, sizeof (char *),
    Offset(default_fg), XtRString, "black",
  },
  {
    "menuBarBackground", XtCString, XtRString, sizeof (char *),
    Offset(menubar_bg), XtRString, "SkyBlue1",
  },
  {
    "menuBarForeground", XtCString, XtRString, sizeof (char *),
    Offset(menubar_fg), XtRString, "black",
  },
  {
    "textBackground", XtCString, XtRString, sizeof (char *),
    Offset(text_bg), XtRString, "SlateGray1",
  },
  {
    "textForeground", XtCString, XtRString, sizeof (char *),
    Offset(text_fg), XtRString, "black",
  },
  {
    "imageBackground", XtCString, XtRString, sizeof (char *),
    Offset(image_bg), XtRString, "white",
  },
  {
    "imageForeground", XtCString, XtRString, sizeof (char *),
    Offset(image_fg), XtRString, "black",
  },
  {
    "buttonBackground", XtCString, XtRString, sizeof (char *),
    Offset(button_bg), XtRString, "SkyBlue1",
  },
  {
    "buttonForeground", XtCString, XtRString, sizeof (char *),
    Offset(button_fg), XtRString, "black",
  },
  {
    "fieldButtonBackground", XtCString, XtRString, sizeof (char *),
    Offset(field_btn_bg), XtRString, "SkyBlue1",
  },
  {
    "fieldButtonForeground", XtCString, XtRString, sizeof (char *),
    Offset(field_btn_fg), XtRString, "black",
  },
  {
    "checkMarkBackground", XtCString, XtRString, sizeof (char *),
    Offset(check_mark_bg), XtRString, "red",
  },
  {
    "willowDir", XtCString, XtRString, sizeof (char *),
    Offset(willow_dir), XtRString, "/usr/lib/X11/willow/",
  },
  {
    "databaseConfigFile", XtCString, XtRString, sizeof (char *),
    Offset(db_config_file_name), XtRString, "db.conf",
  },
  {
    "printCommand", XtCString, XtRString, sizeof (char *),
    Offset(print_cmd), XtRString, "",
  },
  {
    "interfaceMode", XtCString, XtRString, sizeof (char *),
    Offset(interface_mode_name), XtRString, "normal",
  },
  {
    "saveType", XtCString, XtRString, sizeof (char *),
    Offset(save_type_name), XtRString, "local",
  },
  {
    "mailSubject", XtCString, XtRString, sizeof (char *),
    Offset(mail_subj), XtRString, DEFAULT_MAIL_SUBJ,
  },
  {
    "mailAddress", XtCString, XtRString, sizeof (char *),
    Offset(mail_addr), XtRString, "",
  },
  {
    "ftpHost", XtCString, XtRString, sizeof (char *),
    Offset(ftp_host), XtRString, "",
  },
  {
    "ftpUserID", XtCString, XtRString, sizeof (char *),
    Offset(ftp_uid), XtRString, "",
  },
  {
    "userInfo", XtCString, XtRString, sizeof (char *),
    Offset(user_info), XtRString, "",
  },
  {
    "customFields", XtCString, XtRString, sizeof (char *),
    Offset(custom_fields), XtRString, "",
  },
  {
    "animationDelay", "Spin", XtRInt, sizeof (int),
    Offset(spin_delay), XtRImmediate, (void *) DEFAULT_SPIN,
  },
  {
    "browseDelay", "BrowseDelay", XtRInt, sizeof (int),
    Offset(browse_delay), XtRImmediate, (void *) DEFAULT_BROWSE_DELAY,
  },
  {
    "debug", "Debug", XtRInt, sizeof (int),
    Offset(debug_level), XtRImmediate, 0,
  },
  {
    "enableCheat", XtCBoolean, XtRBoolean, sizeof (Boolean),
    Offset(enable_cheat), XtRImmediate, (void *) FALSE,
  },
  {
    "abbreviatedFields", XtCBoolean, XtRBoolean, sizeof (Boolean),
    Offset(abbrev_fields), XtRImmediate, (void *) FALSE,
  },
  {
    "echo", XtCBoolean, XtRBoolean, sizeof (Boolean),
    Offset(initial_echo), XtRImmediate, (void *) FALSE,
  },
  {
    "noBasicMode", XtCBoolean, XtRBoolean, sizeof (Boolean),
    Offset(no_basic_mode), XtRImmediate, (void *) FALSE,
  },
  {
    "reverseVideo", XtCBoolean, XtRBoolean, sizeof (Boolean),
    Offset(inverse), XtRImmediate, (void *) FALSE,
  },
  {
    "ftpAppend", XtCBoolean, XtRBoolean, sizeof (Boolean),
    Offset(ftp_append), XtRImmediate, (void *) TRUE,
  },
  {
    "includeQuery", XtCBoolean, XtRBoolean, sizeof (Boolean),
    Offset(show_search), XtRImmediate, (void *) FALSE,
  },
  {
    "printDialog", XtCBoolean, XtRBoolean, sizeof (Boolean),
    Offset(show_print_dialog), XtRImmediate, (void *) TRUE,
  },
};


/* Initialize our resources. Size has 6 possible values.
   large, medium, small, xlarge: set geometries of windows appropriately.
   default: Look at size of screen, and pick appropriate size. 
   custom:  Individual shell geometries must be set.

   There are three font resources: defaultFont, buttonFont, and textFont.
   Normally the value will be "default" for all of them, and this routine
   will automatically plug in values appropriate to the size, however
   a user can over-ride any or all of them. Note that the word "default" is
   over-loaded here.

   For colors, first check if the screen supports color. If so, set the
   color resources. If not, set up a monochrome version.
 */

extern int error (char *);

init_resources(void)
{
  XrmDatabase db;
  int len;
  char *new_dir;
  char *rclass;
  XrmValue rval;

  /* Load our custom resources */
  XtGetApplicationResources(toplevel, &r, resources,
			    XtNumber(resources), NULL, 0);

  /* Copy globals out of OptionsRec */
  willow_dir = r.willow_dir;
  db_config_file_name = r.db_config_file_name;
  print_cmd = r.print_cmd;
  mail_subj = r.mail_subj;
  mail_addr = r.mail_addr;
  ftp_host = r.ftp_host;
  ftp_uid = r.ftp_uid;
  ftp_append = r.ftp_append;
  show_search = r.show_search;
  user_info_list = r.user_info;
  custom_fields_list = r.custom_fields;
  spin_delay = r.spin_delay;
  debug_level = r.debug_level;
  save_type = r.save_type;
  enable_cheat = r.enable_cheat;
  interface_mode = r.interface_mode;
  initial_echo = r.initial_echo;
  abbrev_fields = r.abbrev_fields;
  no_basic_mode = r.no_basic_mode;
  willow_size = r.willow_size;
  browse_delay = r.browse_delay;
  show_print_dialog = r.show_print_dialog;

  /* Make sure willowDir has a trailing / */
  len = strlen(willow_dir);
  if (willow_dir[len - 1] != '/') {
    new_dir = XtMalloc(len + 2);
    sprintf(new_dir, "%s/", willow_dir);
    willow_dir = new_dir;
  }

  db = XtDatabase(XtDisplay(toplevel));


  /* If the user has not already put in some textTranslations, install
     the emacs bindings */
  if (!XrmGetResource(db, "Willow.textTranslations", XmCTranslations, &rclass,
		      &rval))
    XrmPutStringResource(&db, "Willow*textTranslations", EMACS_BINDINGS);

  /* This resource turns off "close" from mwm menu */
  XrmPutLineResource(&db, "Willow*deleteResponse : DO_NOTHING");

  /* Deal with size */
  if (strncasecmp(willow_size, "default", strlen("default")) == 0) {
    if ((WidthOfScreen(XtScreen(toplevel)) > MIN_SCREEN_WIDTH) &&
	(HeightOfScreen(XtScreen(toplevel)) > MIN_SCREEN_HEIGHT))
      willow_size = "large";
    else
      willow_size = "small";
  }
  if (strncasecmp(willow_size, "small", strlen("small")) == 0) {
    XrmPutStringResource(&db, "Willow*Search.geometry",     SMALL_SEARCH_GM);
    XrmPutStringResource(&db, "Willow*Summaries.geometry",  SMALL_SUMM_GM);
    XrmPutStringResource(&db, "Willow*List Browser.geometry",SMALL_LIST_GM);
    XrmPutStringResource(&db, "Willow*Full-Record.geometry",SMALL_FULL_GM);  
    XrmPutStringResource(&db, "Willow*HelpShell.geometry",  SMALL_HELP_GM);  
    XrmPutStringResource(&db, "Willow*XmForm*verticalSpacing", 
			 SMALL_VERT_SPACING);
    XrmPutStringResource(&db, "Willow*XmForm*horizontalSpacing", 
			 SMALL_HORIZ_SPACING);
    XrmPutStringResource(&db, "Willow*XmDialogShell*XmForm*verticalSpacing", 
			 SMALL_DIALOG_VERT_SPACING);
    XrmPutStringResource(&db, "Willow*XmDialogShell*XmForm*horizontalSpacing", 
			 SMALL_DIALOG_HORIZ_SPACING);
    XrmPutStringResource(&db, "Willow*Search*search_box*XmRowColumn.topOffset",
			 SMALL_FIELD_OFFSET);

  }
  else if (strncasecmp(willow_size, "medium", strlen("medium")) == 0) {
    XrmPutStringResource(&db, "Willow*Search.geometry",     MEDIUM_SEARCH_GM);
    XrmPutStringResource(&db, "Willow*Summaries.geometry",  MEDIUM_SUMM_GM);
    XrmPutStringResource(&db, "Willow*List Browser.geometry",MEDIUM_LIST_GM);
    XrmPutStringResource(&db, "Willow*Full-Record.geometry",MEDIUM_FULL_GM);  
    XrmPutStringResource(&db, "Willow*HelpShell.geometry",  MEDIUM_HELP_GM);  
    XrmPutStringResource(&db, "Willow*XmForm*verticalSpacing", 
			 MEDIUM_VERT_SPACING);
    XrmPutStringResource(&db, "Willow*XmForm*horizontalSpacing", 
			 MEDIUM_HORIZ_SPACING);
    XrmPutStringResource(&db, "Willow*XmDialogShell*XmForm*verticalSpacing", 
			 MEDIUM_DIALOG_VERT_SPACING);
    XrmPutStringResource(&db, "Willow*XmDialogShell*XmForm*horizontalSpacing", 
			 MEDIUM_DIALOG_HORIZ_SPACING);
    XrmPutStringResource(&db, "Willow*Search*search_box*XmRowColumn.topOffset",
			 MEDIUM_FIELD_OFFSET);

  }
  else if (strncasecmp(willow_size, "large", strlen("large")) == 0) {
    XrmPutStringResource(&db, "Willow*Search.geometry",     LARGE_SEARCH_GM);
    XrmPutStringResource(&db, "Willow*Summaries.geometry",  LARGE_SUMM_GM);
    XrmPutStringResource(&db, "Willow*List Browser.geometry",LARGE_LIST_GM);
    XrmPutStringResource(&db, "Willow*Full-Record.geometry",LARGE_FULL_GM);  
    XrmPutStringResource(&db, "Willow*HelpShell.geometry",  LARGE_HELP_GM);  
    XrmPutStringResource(&db, "Willow*XmForm*verticalSpacing", 
			 LARGE_VERT_SPACING);
    XrmPutStringResource(&db, "Willow*XmForm*horizontalSpacing", 
			 LARGE_HORIZ_SPACING);
    XrmPutStringResource(&db, "Willow*XmDialogShell*XmForm*verticalSpacing", 
			 LARGE_DIALOG_VERT_SPACING);
    XrmPutStringResource(&db, "Willow*XmDialogShell*XmForm*horizontalSpacing", 
			 LARGE_DIALOG_HORIZ_SPACING);
    XrmPutStringResource(&db, "Willow*Search*search_box*XmRowColumn.topOffset",
			 LARGE_FIELD_OFFSET);
  }
  else if (strncasecmp(willow_size, "extra-large", strlen("extra-large")) == 0) {
    XrmPutStringResource(&db, "Willow*Search.geometry",     XLARGE_SEARCH_GM);
    XrmPutStringResource(&db, "Willow*Summaries.geometry",  XLARGE_SUMM_GM);
    XrmPutStringResource(&db, "Willow*List Browser.geometry",XLARGE_LIST_GM);
    XrmPutStringResource(&db, "Willow*Full-Record.geometry",XLARGE_FULL_GM);  
    XrmPutStringResource(&db, "Willow*HelpShell.geometry",  XLARGE_HELP_GM);  
    XrmPutStringResource(&db, "Willow*XmForm*verticalSpacing", 
			 XLARGE_VERT_SPACING);
    XrmPutStringResource(&db, "Willow*XmForm*horizontalSpacing", 
			 XLARGE_HORIZ_SPACING);
    XrmPutStringResource(&db, "Willow*XmDialogShell*XmForm*verticalSpacing", 
			 XLARGE_DIALOG_VERT_SPACING);
    XrmPutStringResource(&db, "Willow*XmDialogShell*XmForm*horizontalSpacing", 
			 XLARGE_DIALOG_HORIZ_SPACING);
    XrmPutStringResource(&db, "Willow*Search*search_box*XmRowColumn.topOffset",
			 XLARGE_FIELD_OFFSET);

  }
  else if (strncasecmp(willow_size, "custom", strlen("custom")) != 0) 
    error("Invalid specification for Willow*size");
  /* Do Fonts */
  if (strncasecmp(r.default_font, "default", strlen("default")) == 0) {
    if (strncasecmp(willow_size, "small", strlen("small")) == 0)
      r.default_font = SMALL_DEFAULT_FT;
    else if (strncasecmp(willow_size, "medium", strlen("medium")) == 0)
      r.default_font = MEDIUM_DEFAULT_FT;
    else if (strncasecmp(willow_size, "large", strlen("large")) == 0)
      r.default_font = LARGE_DEFAULT_FT;
    else if (strncasecmp(willow_size, "extra-large", strlen("extra-large")) == 0)
      r.default_font = XLARGE_DEFAULT_FT;
  }
  if (strncasecmp(r.button_font, "default", strlen("default")) == 0) {
    if (strncasecmp(willow_size, "small", strlen("small")) == 0)
      r.button_font = SMALL_BUTTON_FT;
    else if (strncasecmp(willow_size, "medium", strlen("medium")) == 0)
      r.button_font = MEDIUM_BUTTON_FT;
    else if (strncasecmp(willow_size, "large", strlen("large")) == 0)
      r.button_font = LARGE_BUTTON_FT;
    else if (strncasecmp(willow_size, "extra-large", strlen("extra-large")) == 0)
      r.button_font = XLARGE_BUTTON_FT;
  }
  if (strncasecmp(r.text_font, "default", strlen("default")) == 0) {
    if (strncasecmp(willow_size, "small", strlen("small")) == 0)
      r.text_font = SMALL_TEXT_FT;
    else if (strncasecmp(willow_size, "medium", strlen("medium")) == 0)
      r.text_font = MEDIUM_TEXT_FT;
    else if (strncasecmp(willow_size, "large", strlen("large")) == 0)
      r.text_font = LARGE_TEXT_FT;
    else if (strncasecmp(willow_size, "extra-large", strlen("extra-large")) == 0)
      r.text_font = XLARGE_TEXT_FT;
  }

  XrmPutStringResource(&db,"Willow*fontList", r.default_font);
  XrmPutStringResource(&db,"Willow*opacify_btn*fontList", r.default_font);
  XrmPutStringResource(&db,"Willow*XmPushButton*fontList", r.button_font);
  XrmPutStringResource(&db,"Willow*XmDialogShell*XmPushButton*fontList", 
		       r.default_font);
  XrmPutStringResource(&db,"Willow*XmText*fontList", r.text_font);
  XrmPutStringResource(&db,"Willow*XmTextField*fontList", r.text_font);
  XrmPutStringResource(&db,"Willow*XmList*fontList", r.text_font);
  XrmPutStringResource(&db,"Willow*XmRowColumn*fontList", r.default_font);
  XrmPutStringResource(&db,"Willow*from_btn*fontList", r.default_font);
  XrmPutStringResource(&db,"Willow*to_btn*fontList", r.default_font);
  XrmPutStringResource(&db,"Willow*HelpShell*XmLabel.fontList", r.text_font);
  XrmPutStringResource(&db,"Willow*HelpShell*XmList.fontList", r.default_font);

  /* Increase shadow around buttons */
  XrmPutStringResource(&db,"Willow*XmPushButton*shadowThickness", SHADOW_THICKNESS_S);
  XrmPutStringResource(&db,"Willow*summary_window*XmCascadeButton*shadowThickness", "0");
  XrmPutStringResource(&db,"Willow*summary_window*XmRowColumn.shadowThickness", SHADOW_THICKNESS_S);
  XrmPutStringResource(&db,"Willow*menu_bar*XmRowColumn*shadowThickness", "2");
  XrmPutStringResource(&db,"Willow*search_box*XmRowColumn*shadowThickness", "2");
  XrmPutStringResource(&db,"Willow*search_box*XmRowColumn*XmCascadeButtonGadget*shadowThickness", SHADOW_THICKNESS_S);
  XrmPutStringResource(&db,"Willow*range_ctrl_box*XmRowColumn*shadowThickness", "2");
  XrmPutStringResource(&db,"Willow*range_ctrl_box*XmRowColumn*XmCascadeButtonGadget*shadowThickness", SHADOW_THICKNESS_S);
  /* This one fixes the pull-down extensions menu */
  XrmPutStringResource(&db,"Willow*ext_menu*shadowThickness", "2");

  /* Silly stuff to make Field buttons show up one color */
  XrmPutStringResource(&db,"Willow*search_box.XmRowColumn*highlightThickness", "0");
  XrmPutStringResource(&db,"Willow*search_box.XmRowColumn.marginHeight", "0");
  XrmPutStringResource(&db,"Willow*search_box.XmRowColumn.marginWidth", "0");
  XrmPutStringResource(&db,"Willow*search_box.XmRowColumn.spacing", "0");
  XrmPutStringResource(&db,"Willow*search_box*XmRowColumn*XmLabelGadget.marginWidth", "0");
  XrmPutStringResource(&db,"Willow*search_box*XmRowColumn*XmLabelGadget.shadowThickness", "0");

  XrmPutStringResource(&db,"Willow*range_ctrl_box*XmRowColumn*highlightThickness", "0");
  XrmPutStringResource(&db,"Willow*range_ctrl_box*XmRowColumn.marginHeight", "0");
  XrmPutStringResource(&db,"Willow*range_ctrl_box*XmRowColumn.marginWidth", "0");
  XrmPutStringResource(&db,"Willow*range_ctrl_box*XmRowColumn.spacing", "0");
  XrmPutStringResource(&db,"Willow*range_ctrl_box*XmRowColumn*XmLabelGadget.marginWidth", "0");
  XrmPutStringResource(&db,"Willow*range_ctrl_box*XmRowColumn*XmLabelGadget.shadowThickness", "0");
  /* Do colors */
  if (DefaultDepthOfScreen(XtScreen(toplevel)) > 1) {
    XrmPutStringResource(&db,"Willow*background", r.default_bg);
    XrmPutStringResource(&db,"Willow*foreground", r.default_fg);
    XrmPutStringResource(&db,"Willow*XmText*background", r.text_bg);
    XrmPutStringResource(&db,"Willow*XmTextField*background", r.text_bg);
    XrmPutStringResource(&db,"Willow*XmText*foreground", r.text_fg);
    XrmPutStringResource(&db,"Willow*XmTextField*foreground", r.text_fg);
    XrmPutStringResource(&db,"Willow*XmList*background", r.text_bg);
    XrmPutStringResource(&db,"Willow*XmList*foreground", r.text_fg);
    XrmPutStringResource(&db,"Willow*XmPushButton*background", r.button_bg);
    XrmPutStringResource(&db,"Willow*XmPushButton*foreground", r.button_fg);
    XrmPutStringResource(&db,"Willow*search_box*XmRowColumn*background", r.field_btn_bg);
    XrmPutStringResource(&db,"Willow*search_box*XmRowColumn*foreground", r.field_btn_fg);
    XrmPutStringResource(&db,"Willow*XmDialogShell*range_ctrl_box*XmRowColumn*background", r.field_btn_bg);
    XrmPutStringResource(&db,"Willow*XmDialogShell*range_ctrl_box*XmRowColumn*foreground", r.field_btn_fg);
    XrmPutStringResource(&db,"Willow*opac_field*background", r.field_btn_bg);
    XrmPutStringResource(&db,"Willow*opac_field*foreground", r.field_btn_fg);
    XrmPutStringResource(&db,"Willow*from_btn.background",   r.field_btn_bg);
    XrmPutStringResource(&db,"Willow*from_btn.foreground",   r.field_btn_fg);
    XrmPutStringResource(&db,"Willow*to_btn.background",   r.field_btn_bg);
    XrmPutStringResource(&db,"Willow*to_btn.foreground",   r.field_btn_fg);
    
    XrmPutStringResource(&db,"Willow*menu_bar*background", r.menubar_bg);
    XrmPutStringResource(&db,"Willow*menu_bar*foreground", r.menubar_fg);

    XrmPutStringResource(&db,"Willow*Search*status_label.background", r.text_bg);
    XrmPutStringResource(&db,"Willow*Search*status_label.foreground", r.text_fg);


    XrmPutStringResource(&db,"Willow*XmToggleButton*selectColor",r.check_mark_bg);
    XrmPutStringResource(&db,"Willow*XmDialogShell*background", r.default_bg);
    XrmPutStringResource(&db,"Willow*XmDialogShell*foreground", r.default_fg);
    XrmPutStringResource(&db,"Willow*XmDialogShell*globe_image*background", r.image_bg);
    XrmPutStringResource(&db,"Willow*XmDialogShell*globe_image*foreground", r.image_fg);
    XrmPutStringResource(&db,"Willow*XmDialogShell*XmPushButton*foreground", r.default_fg);
    XrmPutStringResource(&db,"Willow*XmDialogShell*XmText*background", r.text_bg);
    XrmPutStringResource(&db,"Willow*XmDialogShell*XmTextField*background", r.text_bg);
    XrmPutStringResource(&db,"Willow*XmDialogShell*XmText*foreground", r.text_fg);
    XrmPutStringResource(&db,"Willow*XmDialogShell*XmTextField*foreground", r.text_fg);
    XrmPutStringResource(&db,"Willow*XmDialogShell*XmList*background", r.text_bg);
    XrmPutStringResource(&db,"Willow*XmDialogShell*XmList*foreground", r.text_fg);
    XrmPutStringResource(&db,"Willow*XmDialogShell*globe_box*background", r.image_bg);
    XrmPutStringResource(&db,"Willow*XmDialogShell*globe_box*foreground", r.image_fg);
    XrmPutStringResource(&db,"Willow*XmDialogShell*globe_box*cancel_btn.foreground", r.button_fg);
    XrmPutStringResource(&db,"Willow*XmDialogShell*globe_box*cancel_btn.background", r.button_bg);
    XrmPutStringResource(&db,"Willow*about_box_popup*background", r.image_bg);
    XrmPutStringResource(&db,"Willow*about_box_popup*foreground", r.image_fg);
    XrmPutStringResource(&db,"Willow*HelpShell*XmLabel*background", r.text_bg);
    XrmPutStringResource(&db,"Willow*HelpShell*XmLabel*foreground", r.text_fg);
    XrmPutStringResource(&db,"Willow*HelpShell*XmList*background", r.default_bg);
    XrmPutStringResource(&db,"Willow*HelpShell*XmList*foreground", r.default_fg);
  }
  else {
    /* Set up a monochrome version */
    if (r.inverse) {
      XrmPutLineResource(&db, "Willow*background : black");
      XrmPutLineResource(&db, "Willow*foreground : white");
    }
    else {
      XrmPutLineResource(&db, "Willow*background : white");
      XrmPutLineResource(&db, "Willow*foreground : black");
    }
  }

  /* Map from save_type_name string to save_type integer */
  if (strncasecmp(r.save_type_name, "email", strlen("email")) == 0)
    save_type = SAVE_EMAIL;
  else if (strncasecmp(r.save_type_name, "ftp", strlen("ftp")) == 0)
    save_type = SAVE_FTP;
  else
    save_type = SAVE_LOCAL;

  /* Map from interface_mode_name string to interface_mode integer */
  if (strncasecmp(r.interface_mode_name, "anonymous", strlen("anonymous")) == 0)
    interface_mode = ANONYMOUS_MODE;
  else if (strncasecmp(r.interface_mode_name, "managed", strlen("managed")) == 0)
    interface_mode = MANAGED_MODE;
  else
    interface_mode = NORMAL_MODE;
};

