/***********************************************************************
 *                                                                     *
 *              WILLOW(tm) Information Retrieval Software              *
 *     (Washington Information Looker-upper Layered Over Windows)      *
 *             Copyright 1992-1994 University of Washington            *
 *                                                                     *
 *   Willow is a trademark of the University of Washington.  No        *
 *   commercial use of this trademark or software may be made without  *
 *   prior written permission of the University of Washington.         *
 *                                                                     *
 * Permission to use, copy, modify, and distribute Willow(tm) software *
 * and its documentation for any noncommercial purpose is hereby       *
 * granted, provided that the above copyright notice appears in all    *
 * copies and that both the above copyright notice and this permission *
 * notice appear in supporting documentation, and that the name of the *
 * University of Washington not be used in advertising or publicity    *
 * pertaining to distribution of the software without specific,        *
 * written prior permission.  This software is made available "as is", *
 * and THE UNIVERSITY OF WASHINGTON DISCLAIMS ALL WARRANTIES, EXPRESS  *
 * OR IMPLIED, WITH REGARD TO THIS SOFTWARE, INCLUDING WITHOUT         *
 * LIMITATION ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS    *
 * FOR A PARTICULAR PURPOSE, AND IN NO EVENT SHALL THE UNIVERSITY OF   *
 * WASHINGTON BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL     *
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA  *
 * OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, TORT (INCLUDING       *
 * NEGLIGENCE) OR STRICT LIABILITY, ARISING OUT OF OR IN CONNECTION    *
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.                       *
 *                                                                     *
 * For further information please contact:                             *
 * Willow Project                                                      *
 * Information Systems                                                 *
 * University of Washington                                            *
 * 4545 15th NE, 1st floor                                             *
 * Seattle, Washington, 98105, USA                                     *
 * Internet: willow@cac.washington.edu                                 *
 ***********************************************************************/

/****************************************************************
 * summary.c
 * This file contains the code necessary for manipulating the 
 * summary window.
 ****************************************************************/
#include <stdio.h>
#ifdef SOLARIS
#include <string.h>
#else
#include <strings.h>
#endif


#include <Mrm/MrmAppl.h>
#include <Xm/Text.h>
#include <Xm/PushBG.h>
#include <Xm/ToggleB.h>
#include "willow.h"
#include "db.h"
#include "text-list.h"

extern char* print_cmd;
extern Widget widgets[], summary_shell;
extern int status;
extern FILE *capture_file;
extern XtAppContext app_context;
extern int save_type;
extern Boolean abbrev_fields;

/* This an array which holds pointers to the text of each summary, needs
   to be +1 because we index from 1 */
char *summaries[MAX_SUMMARIES + 1];

/* This an array which holds pointers to the text of each extension, needs
   to be +1 because we index from 1 */
char *extensions[MAX_SUMMARIES + 1];

/* Hold pointers to titles (with marks) of each summary */
char *summ_titles[MAX_SUMMARIES + 1];

/* This is the BRS-assigned number of the current search */
int search_num;

/* Total number of results found in current search */
int num_results = 0;

/* Number of summaries currently entered in list */
int cur_summary = 0;

/* Which, if any, summary is currently being shown in the box */
int displayed_summary = 0;

/* Flag indicates we are waiting for a full-text */
Boolean displaying = FALSE;

/* Flag indicates we are currently doing a summary save */
Boolean printing = FALSE;

/* Flag indicates we are currently doing a summary print */
Boolean saving = FALSE;

/* Are we paused in the middle of a summary retrieval? */
Boolean paused = FALSE;

/* List of currently selected titles */
int *display_titles = NULL;
int  display_titles_count = 0;

/* current database pointer */
extern db_desc_rec *cur_db;       

/* Translation table for click/ctrl-click mucking */
XtTranslations trans = NULL;

/* This is the manual summary that we just asked for, if any */
int needed_summary = 0;

/* how many summary bodies are still needed */
int summ_needed_cnt = 0;

/* how many extension records are still needed */
int ext_needed_cnt = 0;

/* Forward routines */
int desensitize_selection_buttons (void);
int really_save_summaries (FILE *outfile);
int sensitize_selection_buttons (void);
void force_summaries(void);
void title_double_click_cb(Widget w, list_callback_struct *reason);

/* Extern routines */
extern Boolean list_get_selected (Widget w, int **pos_list, int *num);
extern char *get_query_display_str(void);
extern int clear_working (void);
extern int not_ready_buttons (void);
extern int pop_working (char *, Widget);
extern int ready_buttons (void);
extern int start_display (void);
extern void clear_globe (int tag);
extern void hide_window (int tag);
extern void list_add_item (Widget w, char *item);
extern void list_delete_all (Widget w);
extern void list_select_pos (Widget w, int pos, Boolean toggle, Boolean call_cb);
extern void pop_globe (int tag);
extern void send_record_req_cmd (int item, char summary, char abbreviated);
extern void send_title_req_cmd (int title_num);
extern void send_ext_req_cmd(int ext_num);

/* Clear the title list */
clear_titles(void)
{
  int   n, i;
  Arg   args[10];
  char  label[MAX_LINE];

  displayed_summary = needed_summary = summ_needed_cnt = ext_needed_cnt = 0;
  desensitize_selection_buttons();
  XtSetSensitive(widgets[k_summ_next_block],FALSE);
  XmTextSetString(widgets[k_summary_summaries], "");

  n = 0;
  sprintf(label, SCALE_FMT, 0, 0);
  XtSetArg(args[n], XmNlabelString, XmStringCreateSimple(label));n++;
  XtSetValues(widgets[k_summary_scale], args, n);

  list_delete_all(widgets[k_summary_titles]);
  for (i = 1; i <= cur_summary; i++) {
    XtFree(summ_titles[i]);
    if (summaries[i]) {
      XtFree(summaries[i]);
      summaries[i] = NULL;
    }
    if (extensions[i]) {
      XtFree(extensions[i]);
      extensions[i] = NULL;
    }
  }

  cur_summary = 0;
  
  /* Close the window too */
  hide_window(k_summary_titles);
}

/* Get rid of the Summary bodies we are currently buffering, because 
   of a toggle of field abbrev.
*/
void clear_summ_bodies(void)
{
  int i;

  needed_summary = summ_needed_cnt = 0;
  for (i = 1; i <= cur_summary; i++)
    if (summaries[i]) {
      XtFree(summaries[i]);
      summaries[i] = NULL;
    }
  if (displayed_summary) {
    needed_summary = displayed_summary;
    send_record_req_cmd(displayed_summary, TRUE, abbrev_fields);
    XmTextSetString(widgets[k_summary_summaries], "\n\n    RETRIEVING SUMMARY TEXT...");
    displayed_summary = 0;
  }
}

/* Get ready to display a new set of summaries. Free previous saved
   summaries, if any, set up scale, clear the list and show passed in
   search stats in the text window.
   Request first summary.

   NOTE! Do not pass in a status string of NULL.
   Pass in the empty string "" instead.
 */ 
init_summary(int num, char *status)
{ 
  int n;
  Arg args[10];
  char label[MAX_LINE];
  char msg[1024];

  displayed_summary = needed_summary = summ_needed_cnt = ext_needed_cnt = 0;
  if (num > SUMM_CHUNK) {
    if (status[0])
      sprintf(msg,
"%s\n\
Your query matched %d items.\n\
The first %d summaries are being retrieved.\n\
To add more titles to the list, click the %d MORE button.\n\
To view the summary for a title, click it.\n\
   To print it:               click the Print button.\n\
   To e-mail or ftp it:       click the Save/Mail button.\n\
   To view the full record:   click the Retrieve Full button\n\
To select multiple items, press the Ctrl or Control key and\n\
click each title.  Selected titles will be highlighted.\n\
For more information, click the Help button.\n",
	    status, num, SUMM_CHUNK, SUMM_CHUNK);
    else
      sprintf(msg,
"Your query matched %d items.\n\
The first %d summaries are being retrieved.\n\
To add more titles to the list, click the %d MORE button.\n\
\n\
To view the summary for a title, click it.\n\
   To print it:               click the Print button.\n\
   To e-mail or ftp it:       click the Save/Mail button.\n\
   To view the full record:   click the Retrieve Full button\n\
\n\
To select multiple items, press the Ctrl or Control key and\n\
click each title.  Selected titles will be highlighted.\n\
\n\
For more information, click the Help button.\n",
	    num, SUMM_CHUNK, SUMM_CHUNK);
  }
  else if (num > 0) {
    if (status[0])
      sprintf(msg,
"%s\n\
Your query matched %d items. The summaries are being retrieved.\n\
To view the summary for a title, click it.\n\
   To print it:               click the Print button.\n\
   To e-mail or ftp it:       click the Save/Mail button.\n\
   To view the full record:   click the Retrieve Full button\n\
To select multiple items, press the Ctrl or Control key and\n\
click each title.  Selected titles will be highlighted.\n\
\n\
For more information, click the Help button.\n",
	      status, num);
    else
      sprintf(msg,
"Your query matched %d items. The summaries are being retrieved.\n\
\n\
To view the summary for a title, click it.\n\
   To print it:               click the Print button.\n\
   To e-mail or ftp it:       click the Save/Mail button.\n\
   To view the full record:   click the Retrieve Full button\n\
\n\
To select multiple items, press the Ctrl or Control key and\n\
click each title.  Selected titles will be highlighted.\n\
\n\
For more information, click the Help button.\n",
	      num);
  }
  else
    msg[0] = NULL;

  XmTextSetString(widgets[k_summary_summaries], msg);
  
  if (num > MAX_SUMMARIES)
    num = MAX_SUMMARIES;

  if (num)
    send_title_req_cmd(1);

  paused = FALSE;
  n = 0;
  sprintf(label, SCALE_FMT, 0, num);
  XtSetArg(args[n], XmNlabelString, XmStringCreateSimple(label));n++;
  XtSetValues(widgets[k_summary_scale], args, n);
  num_results = num;
}

/* This semantic action routine is called when we have read the last 
   summary. 
 */
summaries_done(void)
{
  status = IDLE;
}

/* Take title and display it in scrolled list,
 */
capture_title(char *title_str)
{
  char *full_title_str;
  XmString str;
  char     label[MAX_LINE];
  int      n;
  Arg      args[10];
  char     number[32];

  if (cur_summary == 0) {
    clear_globe(k_search_globe_box);
    ready_buttons();
  }
  if ((cur_summary < MAX_SUMMARIES) && 
      (status == SUMMARIZING) &&
      !displaying)
    cur_summary++;
  else
    return;

  /* Truncate title to a reasonable length */
  if (strlen(title_str) > MAX_TITLE_CHARS)
    title_str[MAX_TITLE_CHARS] = 0;

  full_title_str = XtMalloc(strlen(title_str) + strlen(TITLE_FORM) + 10);
 
  /* Get the numbering in */
  if (cur_summary < 10)
    sprintf(number, "%d    ", cur_summary);
  else if (cur_summary < 100)
    sprintf(number, "%d   ", cur_summary);
  else if (cur_summary < 1000)
    sprintf(number, "%d  ", cur_summary);
  else
    sprintf(number, "%d ", cur_summary);

  sprintf(full_title_str, TITLE_FORM, number, title_str);

  str = XmStringCreateSimple(full_title_str);
  list_add_item(widgets[k_summary_titles], full_title_str);
  XmStringFree(str);
  summ_titles[cur_summary] = full_title_str;

  n = 0;
  sprintf(label, SCALE_FMT, cur_summary, num_results);
  XtSetArg(args[n], XmNlabelString, XmStringCreateSimple(label));n++;
  XtSetValues(widgets[k_summary_scale], args, n);
  
  if (cur_summary == num_results) {
    summaries_done();
    if (num_results == 1) {
      list_select_pos(widgets[k_summary_titles], 1, TRUE, TRUE);
    }
  }
  else if (cur_summary % SUMM_CHUNK)
    send_title_req_cmd(cur_summary + 1);
  else {
    status = IDLE;
    paused = TRUE;
    XtSetSensitive(widgets[k_summ_next_block],TRUE);
  }
}

/* An extension record just came in */
capture_ext(char *ext_str, int num)
{
  if (extensions[num])
    XtFree(extensions[num]);
  extensions[num] = XtMalloc(strlen(ext_str)+1);
  strcpy(extensions[num], ext_str);
  if (ext_needed_cnt)
    ext_needed_cnt--;
}

/* Capture and/or display a summary body that just came in */
void display_summary(int num, char *summary_body)
{
  if (summaries[num])
    XtFree(summaries[num]);
  summaries[num] = XtMalloc(strlen(summary_body)+1);
  strcpy(summaries[num], summary_body);
  if (summ_needed_cnt)
    summ_needed_cnt--;

  if (needed_summary == num) {
    XmTextSetString(widgets[k_summary_summaries], summaries[num]);
    displayed_summary = num;
    needed_summary = NULL;
  }
}

/* Continue paused summaries */
void summ_cont_cb(Widget w, int *tag, XtPointer reason)
{
  paused = FALSE;
  status = SUMMARIZING;
  XtSetSensitive(widgets[k_summ_next_block],FALSE);
  send_title_req_cmd(cur_summary + 1);
}

/* Callback triggered when a title is clicked on. Causes us to 
   display the associated text in the summary text window.
 */
void title_select_cb(Widget w, list_callback_struct *reason)
{
  int which;

  sensitize_selection_buttons();
  if (reason)
    which = reason->pos;
  else
    which = 1;

  if (!summaries[which]){
    needed_summary = which;
    send_record_req_cmd(which, TRUE, abbrev_fields);
    XmTextSetString(widgets[k_summary_summaries], "\n\n    RETRIEVING SUMMARY TEXT...");
  }
  else {
    XmTextSetString(widgets[k_summary_summaries], summaries[which]);
    displayed_summary = which;
  }
}

/* This callback is called when the summary display button
   is pressed. 
 */
void title_display_cb(Widget w, int *tag, char *reason)
{
  char msg[MAX_LINE];

  if (displaying)
    return;
  if (display_titles)
    XtFree(display_titles);
  if (list_get_selected(widgets[k_summary_titles], &display_titles, 
			   &display_titles_count)) {
    not_ready_buttons();
    XtSetSensitive(widgets[k_summ_disp], FALSE);
    XtSetSensitive(widgets[k_ext_btn], FALSE);
    XtSetSensitive(widgets[k_summ_next_block], FALSE);

    sprintf(msg, FULL_GLOBE_MSG_FMT, 0, display_titles_count);
    XtVaSetValues(widgets[k_full_globe_label], XmNlabelString, 
		  XmStringCreateLtoR(msg, XmSTRING_DEFAULT_CHARSET), NULL);
    pop_globe(k_full_globe_box);
    send_record_req_cmd(display_titles[0], FALSE, abbrev_fields);
    start_display();
    displaying = TRUE;
  }
}

/* This callback is called when a summary title is double clicked. 
   Call title_select_cb to select this title, and set flag so
   we can start retrieval as soon as this summary is done.
 */
void title_double_click_cb(Widget w, list_callback_struct *reason)
{
  title_display_cb(NULL, NULL, NULL);
}

/* Sensitize the buttons which are only applicable when there is valid
   title currently selected.
 */
sensitize_selection_buttons(void)
{
  if (!displaying && !printing && !saving) {
    XtSetSensitive(widgets[k_summ_disp], TRUE);
    XtSetSensitive(widgets[k_ext_btn], TRUE);
    XtSetSensitive(widgets[k_summ_save_btn], TRUE);
    if (print_cmd[0])
      XtSetSensitive(widgets[k_summ_print],TRUE);
  }
}

/* Sensitize the buttons which are only applicable when there is valid
   title currently selected.
 */
desensitize_selection_buttons(void)
{
  XtSetSensitive(widgets[k_summ_disp], FALSE);
  XtSetSensitive(widgets[k_ext_btn], FALSE);
  XtSetSensitive(widgets[k_summ_print],FALSE);
  XtSetSensitive(widgets[k_summ_save_btn],FALSE);
}

/* save the current summaries */

void results_summ_save(void)
{
  if (!capture_file)
    return;
  saving = TRUE;

  desensitize_selection_buttons();
  not_ready_buttons();

  pop_working("Saving summaries to disk...", summary_shell);
  really_save_summaries(capture_file);

  clear_working();
  saving = FALSE;
  sensitize_selection_buttons();
  ready_buttons();
}

/* Make sure all the selected summaries are really here. We need 
   our own event loop to do it. */
void force_summaries(void)
{
  int *position_list;
  int  position_count, i;

  if (list_get_selected(widgets[k_summary_titles], &position_list, 
			   &position_count)) {
    for (i = 0; i < position_count; i++) {
      if (!summaries[position_list[i]]) {
	summ_needed_cnt++;
	send_record_req_cmd(position_list[i], TRUE, abbrev_fields);
      }
    }
    XtFree(position_list);
    while (summ_needed_cnt) {
      XtAppProcessEvent(app_context, XtIMAlternateInput | XtIMTimer | 
			XtIMXEvent);
    }
  }
}

/* Make sure the extension records for all selected titles are 
   really here. We need our own event loop to do it. */
void force_extensions(void)
{
  int *position_list;
  int  position_count, i;

  if (list_get_selected(widgets[k_summary_titles], &position_list, 
			   &position_count)) {
    for (i = 0; i < position_count; i++) {
      if (!extensions[position_list[i]]) {
	ext_needed_cnt++;
	send_ext_req_cmd(position_list[i]);
      }
    }
    XtFree(position_list);
    while (ext_needed_cnt) {
      XtAppProcessEvent(app_context, XtIMAlternateInput | XtIMTimer | 
			XtIMXEvent);
    }
  }
}

/* Really save the current summaries */
really_save_summaries(FILE *outfile)
{
  char *result;
  int *position_list;
  int  position_count, i;

  fprintf(outfile, "%s", get_query_display_str());
  force_summaries();
  if (list_get_selected(widgets[k_summary_titles], &position_list, 
			   &position_count)) {
    for (i = 0; i < position_count; i++) {
      result = summaries[position_list[i]];
      fprintf(outfile, "%s\n", result);
    }
    fflush(outfile);    
    XtFree(position_list);
  }
}

/* Center the control button in the summary window */
center_summ(Widget w, XtPointer data, XEvent *event, Boolean *cont)
{
  Dimension form_width, btn_width, btn_offset;

  /* The "30 More" form */
  XtVaGetValues(w, XmNwidth, &form_width, NULL);
  XtVaGetValues(widgets[k_summ_next_block], XmNwidth, &btn_width, NULL);
  XtVaSetValues(widgets[k_summ_next_block], XmNleftOffset, 
		(form_width - btn_width)/2, NULL);

  /* The other buttons */
  XtVaGetValues(widgets[k_summ_disp], XmNwidth, &btn_width, NULL);

  XtVaGetValues(widgets[k_summ_btn_form], XmNwidth, &form_width, NULL);
  btn_offset = (form_width - btn_width) / 2;
  XtVaSetValues(widgets[k_summ_disp], XmNleftOffset, btn_offset, NULL);
  XtVaSetValues(widgets[k_ext_menu], XmNleftOffset, btn_offset, NULL);
  XtVaSetValues(widgets[k_ext_btn], XmNwidth, btn_width - 2 * SHADOW_THICKNESS, NULL);
  XtVaSetValues(widgets[k_summ_print], XmNleftOffset, btn_offset, 
		XmNwidth, btn_width, NULL);
  XtVaSetValues(widgets[k_summ_save_btn], XmNleftOffset, btn_offset, 
		XmNwidth, btn_width, NULL);
  XtVaSetValues(widgets[k_summ_help_btn], XmNleftOffset, btn_offset, NULL);
  XtVaSetValues(widgets[k_summ_ok_btn], XmNrightOffset, btn_offset, NULL);
}

