/***********************************************************************
 *                                                                     *
 *              WILLOW(tm) Information Retrieval Software              *
 *     (Washington Information Looker-upper Layered Over Windows)      *
 *             Copyright 1992-1994 University of Washington            *
 *                                                                     *
 *   Willow is a trademark of the University of Washington.  No        *
 *   commercial use of this trademark or software may be made without  *
 *   prior written permission of the University of Washington.         *
 *                                                                     *
 * Permission to use, copy, modify, and distribute Willow(tm) software *
 * and its documentation for any noncommercial purpose is hereby       *
 * granted, provided that the above copyright notice appears in all    *
 * copies and that both the above copyright notice and this permission *
 * notice appear in supporting documentation, and that the name of the *
 * University of Washington not be used in advertising or publicity    *
 * pertaining to distribution of the software without specific,        *
 * written prior permission.  This software is made available "as is", *
 * and THE UNIVERSITY OF WASHINGTON DISCLAIMS ALL WARRANTIES, EXPRESS  *
 * OR IMPLIED, WITH REGARD TO THIS SOFTWARE, INCLUDING WITHOUT         *
 * LIMITATION ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS    *
 * FOR A PARTICULAR PURPOSE, AND IN NO EVENT SHALL THE UNIVERSITY OF   *
 * WASHINGTON BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL     *
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA  *
 * OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, TORT (INCLUDING       *
 * NEGLIGENCE) OR STRICT LIABILITY, ARISING OUT OF OR IN CONNECTION    *
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.                       *
 *                                                                     *
 * For further information please contact:                             *
 * Willow Project                                                      *
 * Information Systems                                                 *
 * University of Washington                                            *
 * 4545 15th NE, 1st floor                                             *
 * Seattle, Washington, 98105, USA                                     *
 * Internet: willow@cac.washington.edu                                 *
 ***********************************************************************/

/****************************************************************
 * text-list.c
 * This file contains the code necessary for creating the functionality
 * of an XmList using an XmText widget. This should probably be 
 * a genuine subclassed widget, but who's got time for that kind
 * of thing?
 ****************************************************************/
#include <stdio.h>
#ifdef SOLARIS
#include <string.h>
#else
#include <strings.h>
#endif


#include <Mrm/MrmAppl.h>
#include <Xm/Text.h>
#include <Xm/TextP.h>

#include "willow.h"
#include "text-list.h"

/* Externals */
extern XtAppContext app_context;   /* Application Context for XtApp* calls */
extern Widget widgets[];
void         usleep(unsigned int nusecs);
Boolean      xmtext_at_bottom(Widget w);

/* Forward declarations */
static void tl_click(Widget w, XEvent *event, String *params, 
		  Cardinal *num_params);
static void tl_drag(Widget w, XEvent *event, String *params, 
		  Cardinal *num_params);
static void tl_mouse_up(Widget w, XEvent *event, String *params, 
		  Cardinal *num_params);
static void tl_2click(Widget w, XEvent *event, String *params, 
		      Cardinal *num_params);
static void tl_out(Widget w, XEvent *event, String *params, 
		   Cardinal *num_params);
static void tl_in(Widget w, XEvent *event, String *params, 
		   Cardinal *num_params);
void list_kb_up(Widget w, XEvent *event, String *params, Cardinal *num_params);
void list_kb_down(Widget w, XEvent *event, String *params, Cardinal *num_params);
void list_select_pos(Widget w, int pos, Boolean toggle, Boolean call_cb);
void list_deselect_pos(Widget w, int pos, Boolean toggle, Boolean call_cb);

extern Widget search_shell;

/* What is duration for a multi-click? */
static int multi_click_time = 0;

/* We really only need one, because clicks close enough in time should
   be in the same window */
static int last_click_time = 0;

list_rec *dict_desc, *summ_desc;

/* Are we currently doing an auto-scroll due to LeaveNotify? */
static Boolean auto_scrolling = FALSE;

#define SET_DESC desc = (w == widgets[k_summary_titles] ? \
			 summ_desc : dict_desc);\
                 if (!desc) return;
  
/* register a "callback" for the list */
void list_register_callback(Widget w, int type, void (*proc)())
{
  list_rec *desc;

  SET_DESC;

  switch (type) {
    case SELECT_CB:
      desc->select_cb = proc;
      break;
    case CLICK2_CB:
      desc->click2_cb = proc;
      break;
    case SCROLL_UP_CB:
      desc->scroll_up_cb = proc;
      break;
    case SCROLL_DN_CB:
      desc->scroll_down_cb = proc;
      break;
    case DESELECT_CB:
      desc->deselect_cb = proc;
      break;
    case DESELECT_ALL_CB:
      desc->deselect_all_cb = proc;
      break;
  }
}

/* Set-up our translations tables for a list */
void list_init(Widget w)
{
  list_rec *desc;
  XtTranslations trans;
  XtActionsRec actions[20];
  int  n;

  if (!multi_click_time)
    multi_click_time = XtGetMultiClickTime(XtDisplay(search_shell));

  /* create a new list */
  desc = (list_rec *) XtMalloc(sizeof(list_rec));
  desc->num_items = 0;
  desc->num_slots = LIST_CHUNK_ITEMS;
  desc->num_selected = 0;
  desc->last_selected = 0;
  desc->last_clicked = 0;
  desc->items = (list_item *) XtMalloc(sizeof(list_item) * 
				       LIST_CHUNK_ITEMS);
  desc->len = 0;
  desc->select_cb = NULL;
  desc->click2_cb = NULL;
  desc->scroll_up_cb = NULL;
  desc->scroll_down_cb = NULL;
  desc->deselect_cb = NULL;
  desc->deselect_all_cb = NULL;
  if (w == widgets[k_summary_titles])
    summ_desc = desc;
  else
    dict_desc = desc;

  /* Set up actions and translations for return and keyboard movements */
  n = 0;
  actions[n].string = "tl_click";
  actions[n].proc = tl_click; n++;
  actions[n].string = "tl_2click";
  actions[n].proc = tl_2click; n++;
  actions[n].string = "tl_mouse_up";
  actions[n].proc = tl_mouse_up; n++;
  actions[n].string = "tl_drag";
  actions[n].proc = tl_drag; n++;
  actions[n].string = "tl_out";
  actions[n].proc = tl_out; n++;
  actions[n].string = "tl_in";
  actions[n].proc = tl_in; n++;
  actions[n].string = "list_kb_up";
  actions[n].proc = list_kb_up; n++;
  actions[n].string = "list_kb_down";
  actions[n].proc = list_kb_down; n++;
  XtAppAddActions(app_context, actions, n);
  
  /* We don't register double-click here because of motion-compression
     problems. Check for it in tl_click() */
  trans = XtParseTranslationTable(
     "<Btn1Down>: tl_click()\n\
      <Btn1Up>: tl_mouse_up()\n\
      Button1<Motion>:tl_drag()\n\
      Button1<LeaveNotify>: tl_out()\n\
      Button1<EnterNotify>: tl_in()\n\
      <Key>osfUp: list_kb_up()\n\
      Ctrl<Key>p: list_kb_up()\n\
      <Key>osfDown: list_kb_down()\n\
      Ctrl<Key>n: list_kb_down()\n\
      <Key>Return: activate()\n\
      ");
  XtOverrideTranslations(w, trans);
}

void list_destroy(Widget w)
{
  list_rec *desc;

  SET_DESC;
  XtFree(desc->items);
  XtFree(desc);
  if (w == widgets[k_summary_titles])
    summ_desc = NULL;
  else
    dict_desc = NULL;
}

/* Add an item to the end of a list. We start with pos == 1 for
   compatibility with list widget */
void list_add_item(Widget w, char *item)
{
  list_rec *desc;

  SET_DESC;
  if (desc->num_items == desc->num_slots - 1) {
    desc->items = (list_item *) XtRealloc(desc->items, sizeof(list_item) *
		         (desc->num_slots + LIST_CHUNK_ITEMS));
    desc->num_slots += LIST_CHUNK_ITEMS;
  }
  if (desc->len)
    XmTextReplace(w, desc->len, desc->len, item);
  else
    XmTextSetString(w, item);
  desc->num_items++;
  desc->items[desc->num_items].offset = desc->len;
  desc->items[desc->num_items].selected = FALSE;
  desc->items[desc->num_items].len = strlen(item);
  desc->len += desc->items[desc->num_items].len;
}

/* Replace the entire contents of the list with a block of text.
   Block comes in as one long chunk, with 0's at end of each
   item. We replace 0's with \n, except for the last one.
 */
void list_replace_all_items(Widget w, char *items, int count, 
			    Boolean has_header)
{
  list_rec *desc;
  int       n, sdex, mark_dex = 0;

  SET_DESC;
  for (n = 0, sdex = 0; n < count; sdex++)
    if (!items[sdex]) {
      n++;
      if (has_header && (n == 1)) {
	has_header = FALSE;
	n--;
      }
      else {
	desc->items[n].offset = mark_dex;
	desc->items[n].selected = FALSE;
	desc->items[n].len = sdex - mark_dex;
      }
      if (n < count)
	items[sdex] = '\n';
      mark_dex = sdex + 1;
    }
  /* The last one seems to get a len 1 to long */
  desc->items[n].len--;
  desc->num_items = count;
  desc->len = sdex - 1;
  desc->num_selected = 0;
  desc->last_selected = 0;
  desc->last_clicked = 0;
  XmTextSetString(w, items);
}

/* Replace item in a list */
void list_replace_item(Widget w, char *item, int pos)
{
  list_rec *desc;
  int  len_diff;
  int  n;

  SET_DESC;
  if (desc->items[pos].selected) {
    desc->items[pos].selected = FALSE;
    desc->num_selected--;
    XmTextSetHighlight(w, desc->items[pos].offset, 
		       desc->items[pos].offset + desc->items[pos].len,
		       XmHIGHLIGHT_NORMAL);
  }
  XmTextReplace(w, desc->items[pos].offset,
		desc->items[pos].offset + desc->items[pos].len,
		item);

  /* Adjust offsets */
  len_diff = strlen(item) - desc->items[pos].len;
  desc->items[pos].len += len_diff;
  for (n = pos + 1; n <= desc->num_items; n++)
    desc->items[n].offset += len_diff;
}

/* make pos selected. If already selected, make unselected, unless
   toggle is FALSE.
 */
void list_select_pos(Widget w, int pos, Boolean toggle, Boolean call_cb)
{
  list_rec *desc;
  list_callback_struct reason;

  SET_DESC;
  if (desc->items[pos].selected) {
    if (!toggle)
      return;
    desc->num_selected--;
    desc->items[pos].selected = FALSE;
    XmTextSetHighlight(w, desc->items[pos].offset, 
		       desc->items[pos].offset + desc->items[pos].len,
		       XmHIGHLIGHT_NORMAL);
    if (call_cb && desc->deselect_cb) {
      reason.pos = pos;
      desc->deselect_cb(w, &reason);
    }
  }
  else {
    desc->num_selected++;
    desc->items[pos].selected = TRUE;
    XmTextSetHighlight(w, desc->items[pos].offset, 
		       desc->items[pos].offset + desc->items[pos].len,
		       XmHIGHLIGHT_SELECTED);
    if (call_cb && desc->select_cb) {
      reason.pos = pos;
      desc->select_cb(w, &reason);
    }
  }
  desc->last_selected = pos;
}

/* make pos deselected. If already deselected, make selected, unless
   toggle is FALSE.
 */
void list_deselect_pos(Widget w, int pos, Boolean toggle, Boolean call_cb)
{
  list_rec *desc;
  list_callback_struct reason;

  SET_DESC;
  if (desc->items[pos].selected) {
    desc->num_selected--;
    desc->items[pos].selected = FALSE;
    XmTextSetHighlight(w, desc->items[pos].offset, 
		       desc->items[pos].offset + desc->items[pos].len,
		       XmHIGHLIGHT_NORMAL);
    if (call_cb && desc->deselect_cb) {
      reason.pos = pos;
      desc->deselect_cb(w, &reason);
    }
  }
  else {
    if (!toggle)
      return;
    desc->num_selected++;
    desc->items[pos].selected = TRUE;
    XmTextSetHighlight(w, desc->items[pos].offset, 
		       desc->items[pos].offset + desc->items[pos].len,
		       XmHIGHLIGHT_SELECTED);
    if (call_cb && desc->select_cb) {
      reason.pos = pos;
      desc->select_cb(w, &reason);
    }
  }
  desc->last_selected = pos;
}

/* Delete an item, pass in pos */
void list_delete_all(Widget w)
{
  list_rec *desc;

  SET_DESC;
  desc->num_items = 0;
  desc->len = 0;
  desc->num_selected = 0;
  desc->last_selected = 0;
  desc->last_clicked = 0;
  XmTextSetString(w, "");
}

/* De-select all items */
void list_deselect_all_items(Widget w)
{
  list_rec *desc;
  int       n;
  list_callback_struct reason;
  
  SET_DESC;
  if (desc->deselect_cb)
    desc->deselect_all_cb(w, &reason);

  desc->last_selected = 0;
  desc->last_clicked = 1;
  desc->num_selected = 0;
  for (n = 1; n <= desc->num_items; n++) 
    desc->items[n].selected = FALSE;

  XmTextSetHighlight(w, 0, desc->len, XmHIGHLIGHT_NORMAL);
}

/* Return a list of all pos's which are selected. Caller should free.
   Return TRUE if anything selected */
Boolean list_get_selected(Widget w, int **pos_list, int *num)
{
  list_rec *desc;
  int  n, i;

  SET_DESC;
  if (!desc->num_selected)
    return(FALSE);

  *pos_list = (int *) XtMalloc(desc->num_selected * sizeof(int));
  *num = desc->num_selected;
  for (n = 1, i = 0; n <= desc->num_items; n++) 
    if (desc->items[n].selected)
      (*pos_list)[i++] = n;
  return(TRUE);
}

/* Return a count of how many items are selected */
int list_get_selected_cnt(Widget w)
{
  list_rec *desc;

  SET_DESC;
  return(desc->num_selected);
}

/* Translation-table triggered routines */
static void tl_click(Widget w, XEvent *event, 
		  String *params, Cardinal *num_params)
{
  list_rec *desc;
  XmTextPosition text_offset;
  int pos;
  int n, last_sel;
  Boolean do_sel = TRUE;

  if (event->xbutton.time && 
      ((event->xbutton.time - last_click_time) < multi_click_time))
    tl_2click(w, event, params, num_params);
  else {
    SET_DESC;

    /* Convert xy to text position */
    text_offset = XmTextXYToPos(w, event->xbutton.x, event->xbutton.y);
    
    /* Which record has that text_offset? */
    for (pos = 1; pos <= desc->num_items; pos++)
      if ((text_offset < (desc->items[pos].offset + desc->items[pos].len)) ||
	  (pos == desc->num_items)) {
	/* If this is not a ctrl-click, deselect the rest */
	if (event->xbutton.state & ShiftMask) {
	  last_sel = desc->last_selected;
	  do_sel = desc->items[last_sel].selected;
	  for (n = last_sel; n != pos; pos < last_sel ? n-- : n++) {
	    if (do_sel)
	      list_select_pos(w, n, FALSE, desc == dict_desc);
	    else
	      list_deselect_pos(w, n, FALSE, desc == dict_desc);
	  }
	}
	else if (!(event->xbutton.state & ControlMask))
	  list_deselect_all_items(w);
	desc->last_clicked = pos;
	if (do_sel)
	  list_select_pos(w, pos, TRUE, TRUE);
	desc->max_dragged = desc->min_dragged = desc->last_dragged = pos;
	break;
      }
  }
  if (event->xbutton.time)
    last_click_time = event->xbutton.time;
}

static void tl_2click(Widget w, XEvent *event, 
		  String *params, Cardinal *num_params)
{
  list_rec *desc;
  XmTextPosition text_offset;
  int n;
  list_callback_struct reason;

  SET_DESC;
  if (!desc->click2_cb)
    return;

  /* Conver xy to text position */
  text_offset = XmTextXYToPos(w, event->xbutton.x, event->xbutton.y);
  
  /* Which record has that text_offset? */
  for (n = 1; n <= desc->num_items; n++)
    if (text_offset < (desc->items[n].offset + desc->items[n].len)) {
      /* if this is not the same place we got the last click,
	 this is not really a double click, pretend it is a click */
      if (n != desc->last_clicked) {
	event->xbutton.time = 0;
	tl_click(w, event, params, num_params);
	return;
      }

      /* If this is not selected, we must have just unselected it */
      if (!desc->items[n].selected)
	list_select_pos(w, n, TRUE, TRUE);
      
      reason.pos = n;
      desc->click2_cb(w, &reason);
      break;
    }
}

/* Translation-table triggered routines */
static void tl_drag(Widget w, XEvent *event, 
		  String *params, Cardinal *num_params)
{
  list_rec *desc;
  XmTextPosition text_offset;
  int pos, i;

  SET_DESC;
  /* Convert xy to text position */
  text_offset = XmTextXYToPos(w, event->xbutton.x, event->xbutton.y);

  /* Which record has that text_offset? */
  for (pos = 1; pos <= desc->num_items; pos++)
    if (text_offset < (desc->items[pos].offset + desc->items[pos].len))
      break;
  if (pos > desc->num_items) pos = desc->num_items;
  if (!pos)
    return;
  if (pos < desc->last_clicked) {
    /* Check for "lost" items */
    if (desc->max_dragged > desc->last_clicked) {
      for (i = desc->last_clicked + 1; i <= desc->max_dragged; i++)
	list_deselect_pos(w, i, FALSE, desc == dict_desc);
      desc->max_dragged = desc->last_clicked;
    }
    if (pos < desc->min_dragged)
      for (i = desc->min_dragged; i >= pos; i--)
	list_select_pos(w, i, FALSE, desc == dict_desc);
    else if (pos > desc->min_dragged)
      for (i = desc->min_dragged; i < pos; i++)
	list_deselect_pos(w, i, FALSE, desc == dict_desc);
    desc->min_dragged = pos;
  }
  else if (pos > desc->last_clicked) {
    /* Check for "lost" items */
    if (desc->min_dragged < desc->last_clicked) {
      for (i = desc->last_clicked - 1; i >= desc->min_dragged; i--)
	list_deselect_pos(w, i, FALSE, desc == dict_desc);
      desc->min_dragged = desc->last_clicked;
    }
    if (pos > desc->max_dragged)
      for (i = desc->max_dragged; i <= pos; i++)
	list_select_pos(w, i, FALSE, desc == dict_desc);
    else if (pos < desc->max_dragged)
      for (i = desc->max_dragged; i > pos; i--)
	list_deselect_pos(w, i, FALSE, desc == dict_desc);
    desc->max_dragged = pos;
  }
  else 
    if (desc->last_dragged != desc->last_clicked)
      list_deselect_pos(w, desc->last_dragged, FALSE, desc == dict_desc);
  desc->last_dragged = pos;
}

/* Mouse leaving the window, trigger an auto-scroll if button down */
static void tl_out(Widget w, XEvent *event, 
		  String *params, Cardinal *num_params)
{
  list_rec *desc;
  Dimension h;
  int  scroll;
  Window dum_win;
  int    dum_pos;
  unsigned int mask;

  SET_DESC;
  XtVaGetValues(w, XmNheight, &h, NULL);
  if (event->xcrossing.y <= 0)
    scroll = -1;
  else if (event->xcrossing.y >= h)
    scroll = 1;
  else 
    return;

  auto_scrolling = TRUE;
  while (TRUE) {
    XQueryPointer(XtDisplay(search_shell), XtWindow(search_shell), &dum_win,
		  &dum_win, &dum_pos, &dum_pos, &dum_pos, &dum_pos,  &mask);
    if (mask & Button1Mask) {
      if (scroll > 0) {
	if (!xmtext_at_bottom(w))
	  XmTextScroll(w, scroll);
      }
      else
	XmTextScroll(w, scroll);
      
      while (XtAppPending(app_context))
	XtAppProcessEvent(app_context, XtIMAlternateInput | XtIMTimer | 
			  XtIMXEvent);
      if (!auto_scrolling)
	/* We got canceled by an EnterNotify during event processing */
	break;

     tl_drag(w, event, params, num_params);

      /* Wait a bit */
      usleep(50000);
    }
    else
      break;
  }
  auto_scrolling = FALSE;
}

/* We only handle EnterNotify to cancel an auto-scroll */
static void tl_in(Widget w, XEvent *event, 
		  String *params, Cardinal *num_params)
{
  auto_scrolling = FALSE;
}

/* Need this for finishing off a drag */
static void tl_mouse_up(Widget w, XEvent *event, 
		  String *params, Cardinal *num_params)
{
  list_rec *desc;
  XmTextPosition text_offset;
  int pos;
  list_callback_struct reason;

  SET_DESC;
  
  /* Convert xy to text position */
  text_offset = XmTextXYToPos(w, event->xbutton.x, event->xbutton.y);

  /* Which record has that text_offset? */
  for (pos = 1; pos <= desc->num_items; pos++)
    if (text_offset < (desc->items[pos].offset + desc->items[pos].len))
      break;
  if (pos > desc->num_items) pos = desc->num_items;
  if ((pos != desc->last_clicked) && desc->select_cb && (desc == summ_desc)) {
    reason.pos = pos;
    desc->select_cb(w, &reason);
  }
}

/* Called by our modified translation table when arrow-keys are pressed, 
   we need to spoof an appropriate scrollbar event */
void list_kb_up(Widget w, XEvent *event, 
		  String *params, Cardinal *num_params)
{
  list_rec *desc;
  int selection;
  list_callback_struct reason;
  short  rows;
  XmTextPosition top;

  SET_DESC;
  selection = desc->last_selected - 1;
  if (selection > 0) {
    list_deselect_all_items(w);
    list_select_pos(w, selection, FALSE, TRUE);

    /* If off screen, page up */
    XtVaGetValues(w, XmNtopCharacter, &top, XmNrows, &rows, NULL);
    if (desc->items[selection].offset < top)
      XmTextScroll(w, 0 - 2 * (rows / 3));
  }
  else if (desc->scroll_up_cb)
    desc->scroll_up_cb(w, reason);
}

void list_kb_down(Widget w, XEvent *event, 
		    String *params, Cardinal *num_params)
{
  list_rec *desc;
  int selection;
  list_callback_struct reason;
  short  rows;
  XmTextPosition top;

  SET_DESC;
  selection = desc->last_selected + 1;
  if (selection <= desc->num_items) {
    list_deselect_all_items(w);
    list_select_pos(w, selection, FALSE, TRUE);

    /* If off screen, page up */
    XtVaGetValues(w, XmNtopCharacter, &top, XmNrows, &rows, NULL);

    /* Note that Motif does give us a way to easily look at what the
       last visible position is, so we are cheating here and looking
       at inside the widget. Hopefully this will not break on us
       in the future. */
    if ((desc->items[selection].offset + desc->items[selection].len) >
	((XmTextWidget) w)->text.bottom_position)
      XmTextScroll(w, 2 * rows / 3);
  }
  else if (desc->scroll_down_cb)
    desc->scroll_down_cb(w, reason);
}
