/***********************************************************************
 *                                                                     *
 *              WILLOW(tm) Information Retrieval Software              *
 *     (Washington Information Looker-upper Layered Over Windows)      *
 *             Copyright 1992-1994 University of Washington            *
 *                                                                     *
 *   Willow is a trademark of the University of Washington.  No        *
 *   commercial use of this trademark or software may be made without  *
 *   prior written permission of the University of Washington.         *
 *                                                                     *
 * Permission to use, copy, modify, and distribute Willow(tm) software *
 * and its documentation for any noncommercial purpose is hereby       *
 * granted, provided that the above copyright notice appears in all    *
 * copies and that both the above copyright notice and this permission *
 * notice appear in supporting documentation, and that the name of the *
 * University of Washington not be used in advertising or publicity    *
 * pertaining to distribution of the software without specific,        *
 * written prior permission.  This software is made available "as is", *
 * and THE UNIVERSITY OF WASHINGTON DISCLAIMS ALL WARRANTIES, EXPRESS  *
 * OR IMPLIED, WITH REGARD TO THIS SOFTWARE, INCLUDING WITHOUT         *
 * LIMITATION ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS    *
 * FOR A PARTICULAR PURPOSE, AND IN NO EVENT SHALL THE UNIVERSITY OF   *
 * WASHINGTON BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL     *
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA  *
 * OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, TORT (INCLUDING       *
 * NEGLIGENCE) OR STRICT LIABILITY, ARISING OUT OF OR IN CONNECTION    *
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.                       *
 *                                                                     *
 * For further information please contact:                             *
 * Willow Project                                                      *
 * Information Systems                                                 *
 * University of Washington                                            *
 * 4545 15th NE, 1st floor                                             *
 * Seattle, Washington, 98105, USA                                     *
 * Internet: willow@cac.washington.edu                                 *
 ***********************************************************************/

/****************************************************************
 * willow.c
 * This is the source code Willow, the Washington Information 
 * Looker-upper Layered Over Windows.
 ****************************************************************/
#include <stdio.h>
#ifdef SOLARIS
#include <string.h>
#else
#include <strings.h>
#endif

#include <ctype.h>
#include <pwd.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/types.h>
#include <netinet/in.h>

#include <Mrm/MrmAppl.h>
#include <Xm/Protocols.h>
#include <Xm/Text.h>
#include <Xm/RowColumn.h>
#include <Xm/ToggleB.h>
#include <Xm/PushB.h>
#include <Xm/Separator.h>
#include <Xm/MessageB.h>
#include <Xm/CascadeB.h>

#include "willow.h"
#include "protocol.h"
#include "db.h"
#include "text-list.h"
#include "save_query.h"
#include "hash.h"
#include "field_pick.h"

/*
 * Global Data
 */

/* Directory where various willow files live (settable as resource) */
char *willow_dir;

Widget toplevel,            /* non-realized  root widget     */
  main_window,   	    /* root of MRM widget tree       */
  search_shell,             /* shell to hold search window   */
  results_shell,            /* shell to hold results window  */
  summary_shell,            /* shell to hold summary window  */
  dict_shell = NULL,        /* shell to hold subjects list   */
  widgets[NUM_WIDGETS];     /* place to keep rest of widgets */

XtAppContext app_context;   /* Application Context for XtApp* calls */

MrmHierarchy s_MrmHierarchy = NULL;  /* MRM database hierarchy ID */
MrmType  dummy_class;         /* and class variable. */
char *db_filename_vec[] =     /* Mrm.hierarchy file list. */
{"uid/willow.uid",            /* basic uid file */
 "uid/dict.uid",              /* subject dict. uid file */
 "uid/range.uid",             /* range-picker uid file */
 };            

/* Default geometries for our shells */
geom_rec search_geom, results_geom, summary_geom, dict_geom;

int db_filename_num =           
  (sizeof db_filename_vec / sizeof db_filename_vec [0]);

/* The # of chars which have been appended to the view widget */
int     text_len = 0;

/* Head/tail of a linked list of hilite markers */
hl_rec *hl_head = NULL, *hl_tail = NULL;

/* Current status of connection with remote host */
int status = UNCONNECTED;

/* Delayed selection command */
char stacked_select[MAX_LINE];

/* This is the currently selected database */
struct _db_desc_rec *cur_db = NULL;

/* Top text widget in Connections Option box. Need this for focus-setting */
Widget top_conn_opt_text;

/* The current results capture files and names */
FILE *capture_file = NULL;
char *capture_name = NULL;

/* Initial state of echo */
Boolean initial_echo;

/* Are we currently running in OPAC mode? 0 for no, 1-3 for steps */
int opac_mode = 0;

/* Default print command */
char *print_cmd;

/* milliseconds between globe spins, if 0, don't use globe at all */
int spin_delay;

/* Used for trapping mwm close event */
Atom wm_delete_window;

/* Used for session management protocol */
Atom xa_willow;

/* How do we save? Initial value set in resources.c */
int save_type;

/* Retrieve abbreviated or full fields? Set in resources.c */
Boolean abbrev_fields;

/* How do we save queries? */
int q_save_type;

/* This is a hash table of available databases */
hash_tbl *db_hash_tab = NULL;

/* List of currently selected titles */
extern int *display_titles;
extern int  display_titles_count;

/* Most recently received summary */
extern int cur_summary;

/* Are we currently waiting a full? */
extern Boolean displaying;

/* Are we currently doing a save of full results? */
extern Boolean popped_full;

/* Global vars for passwords */
extern char   *ftp_pw;

/* ftp parameters */
extern char *ftp_host;
extern char *ftp_uid;
extern char *ftp_file;
extern char *ftp_q_file;
extern Boolean ftp_append;

/* Query-representation pre-pended to search results? */
extern Boolean show_search;

/* Mail parameters */
extern char *mail_addr;
extern char *mail_subj;

/* Are we paused in summaries? */
extern Boolean paused;

/* config file currently being parsed */
extern char *config_file_name;

/* Debug level for output, settable as resource */
int debug_level = 0;

/* Name of database configuration file, settable as resource */
char *db_config_file_name;

/* What line in current configuration file */
extern int config_line_num;

/* Should we give user a cheat box? */
Boolean enable_cheat;

/* Should we disallow basic mode? */
Boolean no_basic_mode;

/* Are we normal, anonymous, or managed (by session manager)? Initialized in
   resources.c */
int interface_mode;

/* List of user_info settings from X-resource */
char *user_info_list;

/* Current info block from driver about database */
char *db_info = NULL;

/* Which of search window text fields was last to receive focus */
static int last_focused_field = 1;

/* Masked text (password) the user types in as a response the driver */
char *masked_response; /* gets initialized in init_application() */

/*
 * Forward declarations
 */
XtEventHandler client_message_handler(Widget w, XtPointer cl, XEvent *ev, caddr_t call);
char *bdup(char *s, int size);
char *capture_output();
char *get_line(char *s, FILE *f);
/* char *strdup(char *s); */
char *xmstring_to_string(XmString string);
int  quit_handler(void);
int append_to_view (char *buffer);
void ctrl_cmd_action(Widget w, XEvent *event, 
		     String *params, Cardinal *num_params);
int config_interface (void);
int disconnect (void);
int error (char *msg);
int free_hl_list (void);
int grow_opac (void);
int hide_all_windows (void);
int init_application (void);
int not_ready_buttons (void);
int parse_user_info(char *str);
void parse_user_info_list(void);
int pop_results (void);
int pop_summary (void);
void pop_user_query(int type, char *query_text, char *query_tag,
		    char *yes_label, char *no_label);
int pop_window (int tag);
int pop_working (char *message, Widget parent);
int read_options (void);
int ready_buttons (void);
int really_select_db (Widget w, db_desc_rec *db_desc, u_long *reason);
void reset_field_labels(void);
int select_by_name (char *name);
int set_icon (char *icon_name, Widget shell);
int set_status_label (int status, char *msg);
int start_display (void);
void browse_dcb(packet_rec *packet);
void build_db_menu(db_menu_rec *db_menu_list);
void center_status(Widget w, XtPointer data, XEvent *event, Boolean *cont);
void clear_cb(Widget w, int *tag, XmAnyCallbackStruct *reason);
void clear_search_file_box();
void conn_opts_ok_cb(Widget w, int *tag, u_long *reason);
void connected_dcb(packet_rec *packet);
void create_cb(Widget w, int *tag, u_long *reason);
void de_opacify(void);
void echo_opt_cb(Widget w, int *tag, u_long *reason);
void error_dcb(packet_rec *packet);
void exts_dcb(packet_rec *packet);
void field_focus_cb(Widget w, int *tag, u_long *reason);
void hide_window(int tag);
void hide_window_cb(Widget w, int *tag, u_long *reason);
void info_dcb(packet_rec *packet);
void match_next_cb(Widget w, int *tag, u_long *reason);
void match_prev_cb(Widget w, int *tag, u_long *reason);
void mwm_close_cb(Widget w, caddr_t client_data, caddr_t call_data);
void no_cr_cb(Widget w, int *tag, XmTextVerifyPtr reason);
void no_db_cb(Widget w, int *tag, u_long *reason);
void opacify(Boolean force);
void opacify_cb(Widget w, int *tag, XtPointer reason);
void pop_info_cb(Widget w, int *tag, u_long *reason);
void pop_warning(char *template, char *arg, Widget parent);
void quit_cb(Widget w, int *tag, u_long *reason);
void read_input();
void really_cheat_cb(Widget w, int *tag, XmSelectionBoxCallbackStruct *reason);
void really_open_save_file_cb(Widget w, int *tag, XmFileSelectionBoxCallbackStruct *reason);
void record_dcb(packet_rec *packet);
void results_full_save(void);
void save_options_cb(Widget w, int *tag, XmSelectionBoxCallbackStruct *reason);
void search_cancel_cb(Widget w, int *tag, u_long *reason);
void search_cb(Widget w, int *tag, u_long *reason);
void select_db_cb(Widget w, db_desc_rec *db, u_long *reason);
void set_focus_cb(Widget w, int which, u_long *reason);
void starrify_cb(Widget w, char **pw, XmTextVerifyPtr reason);
void stats_dcb(packet_rec *packet);
void title_dcb(packet_rec *packet);
void user_req_dcb(packet_rec *packet);
void window_to_top_cb(Widget w, int *tag, u_long *reason);

/*
 * External declarations
 */
extern Boolean list_get_selected (Widget w, int **pos_list, int *num);
extern Search_ptr package_search (void);
extern Widget XmSelectionBoxGetChild (Widget, unsigned char);
extern center_summ(Widget w, XtPointer data, XEvent *event, Boolean *cont);
extern char *getlogin(void);
extern char *range_to_text (int which);
extern char *str_clean (char *s);
extern char *get_query_display_str(void);
extern int config_dbs(char *db_config_file_name, char *path, hash_tbl **db_hash_tab, db_menu_rec **db_menu_list);
extern int atoi ( char *);
extern int capture_exts (char *ext_str, int num);
extern int capture_title (char *title_str);
extern int clear_titles (void);
extern int del_all_searches (void);
extern int desensitize_selection_buttons (void);
extern int do_ftp (int which);
extern int init_browse (db_desc_rec *db);
extern int init_browse_widgets (void);
extern int init_resources (void);
extern int init_summary (int num, char *status);
extern void make_query_display(void);
extern int parse_field_pick_str(char *str);
extern void parse_field_pick_list();
extern int pop_dict (void);
extern int scan_config (FILE *infile, char *value);
extern int summaries_done (void);
extern uid_t getuid (void);
extern void XmListSelectPos (Widget, int, Boolean);
extern void about_cb(Widget w, int *tag, u_long *reason);
extern void add_search_cb(Widget w, int *tag, u_long *reason);
extern void cancel_print_opt_cb(Widget w, int *tag, u_long *reason);
extern void cheat_cb(Widget w, int *tag, u_long *reason);
extern void clear_dialog_cb(Widget w, Widget target, u_long *reason);
extern void clear_dialog_tag_cb(Widget w, int *target, u_long *reason);
extern void clear_globe (int tag);
extern void clear_help (void);
extern void connection_opt_cb(Widget w, int *tag, u_long *reason);
extern void del_search_cb(Widget w, int *tag, u_long *reason);
extern void dict_cancel_cb(Widget w, int *tag, u_long *reason);
extern void dict_hide(void);
extern void dict_paste_cb(Widget w, int *tag, u_long *reason);
extern void dict_search_cb(Widget w, int *tag, XmTextVerifyCallbackStruct *reason);
extern void display_summary (int num, char *summary_body);
extern void field_pick_all_cb(Widget w, int *tag, u_long *reason);
extern void field_pick_clear_cb(Widget w, int *tag, u_long *reason);
extern void field_pick_defaults_cb(Widget w, int *tag, u_long *reason);
extern void free_search (Search_ptr search);
extern void ftp_opts_ok_cb(Widget w, int *tag, u_long *reason);
extern void got_field_list(int count, char *fields);
extern void halt_globe (void);
extern void help_cb(Widget w, Widget target, char *reason);
extern void init_extensions(db_desc_rec *db);
extern void list_destroy (Widget w);
extern void list_init (Widget w);
extern void list_register_callback (Widget w, int type, void (*proc) (/* ??? */));
extern void mail_ok_cb(Widget w, int *tag, u_long *reason);
extern void ok_print_opt_cb(Widget w, int *tag, u_long *reason);
extern void opac_search_cb(Widget w, int *tag, u_long *reason);
extern void open_save_cb(Widget w, int *tag, u_long *reason);
extern void pop_globe (int tag);
extern void pop_message_dialog(unsigned char type, char *message, int button_mask, Widget parent);
extern void pop_range_dlg_cb(Widget w, int *tag, u_long *reason);
extern void pop_retrieve_dlg_cb(Widget w, int *tag, u_long *reason);
extern void pop_save_box(int which);
extern void print_full_cb(Widget w, XtPointer data, u_long *reason);
extern void print_opt_cb(Widget w, int *tag, u_long *reason);
extern void print_summ_cb(Widget w, XtPointer data, u_long *reason);
extern void q_ftp_opts_ok_cb(Widget w, int *tag, u_long *reason);
extern void q_pop_save_box(int which);
extern void q_save_box_cancel_cb(Widget w, int *tag, u_long *reason);
extern void q_toggle_save_type_cb(Widget w, int *target, XmToggleButtonCallbackStruct *reason);
extern void range_cancel_cb(Widget w, int *tag, u_long *reason);
extern void range_ok_cb(Widget w, int *tag, u_long *reason);
extern void really_add_search_cb(Widget w, int *tag, XmSelectionBoxCallbackStruct *reason);
extern void really_del_search_cb(Widget w, int *tag, XmListCallbackStruct *reason);
extern void really_open_search_cb(Widget w, int *tag, XmFileSelectionBoxCallbackStruct *reason);
extern void really_save_search_cb(Widget w, int *tag, XmFileSelectionBoxCallbackStruct *reason);
extern void really_start_browse_cb(Widget w, char *list, u_long *reason);
extern void results_summ_save(void);
extern void retrieve_opts_cancel_cb(Widget w, int *tag, u_long *reason);
extern void retrieve_opts_ok_cb(Widget w, int *tag, u_long *reason);
extern void save_box_cancel_cb(Widget w, int *tag, u_long *reason);
extern void save_geometry (Widget shell, geom_rec *geom);
extern void save_search_cb(Widget w, int *tag, u_long *reason);
extern void send_cheat_cmd(char *command);
extern char *send_connect_cmd(db_desc_rec *db, char *config_path);
extern void send_disconnect_cmd(void);
extern void send_echo_cmd (void);
extern void send_record_req_cmd (int, char, char);
extern void send_search_cmd (int, int, char *, int);
extern void send_search_cmd(int num_text_fields, int num_limit_fields, char *data, int data_len);
extern void send_search_cancel_cmd(void);
extern void send_title_req_cmd (int);
extern void set_geometry (Widget shell, geom_rec *geom);
extern void setup_range_cb(Widget w, int which, u_long *reason);
extern void splice (char **dest, int start, int end, char *source, int source_len);
extern void start_browse_cb(Widget w, db_field_desc_rec *list, u_long *reason);
extern void summ_cont_cb(Widget w, int *tag, XtPointer reason);
extern void title_display_cb(Widget w, int *tag, char *reason);
extern void title_double_click_cb(Widget w, list_callback_struct *reason);
extern void title_select_cb(Widget w, list_callback_struct *reason);
extern void toggle_range_limit_cb(Widget w, int *target, XmToggleButtonCallbackStruct *reason);
extern void toggle_save_type_cb(Widget w, int *target, XmToggleButtonCallbackStruct *reason);
extern void unpack_search_cb (Widget w, Search_ptr search, caddr_t reason);
extern void user_query_cb(Widget w, int *which, char *reason);
extern void wait_child(int sig, int code, caddr_t  data);
int db_compare(char *key, db_desc_rec *element);

/* The names and addresses of things that Mrm.has to bind.  The names do
 * not have to be in alphabetical order.  */

MrmRegisterArg reglist[] = {
  {"about_cb",  		  (caddr_t) about_cb},
  {"add_search_cb", 		  (caddr_t) add_search_cb},
  {"cancel_print_opt_cb", 	  (caddr_t) cancel_print_opt_cb},
  {"cheat_cb", 			  (caddr_t) cheat_cb},
  {"clear_cb",			  (caddr_t) clear_cb}, 
  {"clear_dialog_tag_cb",	  (caddr_t) clear_dialog_tag_cb}, 
  {"connection_opt_cb", 	  (caddr_t) connection_opt_cb}, 
  {"conn_opts_ok_cb", 		  (caddr_t) conn_opts_ok_cb}, 
  {"create_cb", 		  (caddr_t) create_cb}, 
  {"del_search_cb", 		  (caddr_t) del_search_cb},
  {"dict_cancel_cb",  		  (caddr_t) dict_cancel_cb},
  {"dict_paste_cb",  		  (caddr_t) dict_paste_cb},
  {"dict_search_cb",  		  (caddr_t) dict_search_cb},
  {"echo_opt_cb", 		  (caddr_t) echo_opt_cb},
  {"field_focus_cb", 		  (caddr_t) field_focus_cb},
  {"field_pick_all_cb",	          (caddr_t) field_pick_all_cb},
  {"field_pick_clear_cb",	  (caddr_t) field_pick_clear_cb},
  {"field_pick_defaults_cb",	  (caddr_t) field_pick_defaults_cb},
  {"ftp_opts_ok_cb", 		  (caddr_t) ftp_opts_ok_cb}, 
  {"q_ftp_opts_ok_cb", 		  (caddr_t) q_ftp_opts_ok_cb}, 
  {"help_cb", 			  (caddr_t) help_cb},
  {"hide_window_cb", 		  (caddr_t) hide_window_cb}, 
  {"mail_ok_cb", 		  (caddr_t) mail_ok_cb}, 
  {"match_next_cb",  		  (caddr_t) match_next_cb},
  {"match_prev_cb",  		  (caddr_t) match_prev_cb},
  {"no_cr_cb",  		  (caddr_t) no_cr_cb},
  {"no_db_cb",  		  (caddr_t) no_db_cb},
  {"ok_print_opt_cb", 		  (caddr_t) ok_print_opt_cb},
  {"opac_search_cb", 		  (caddr_t) opac_search_cb},
  {"opacify_cb", 		  (caddr_t) opacify_cb},
  {"open_save_cb", 	          (caddr_t) open_save_cb},
  {"pop_info_cb", 		  (caddr_t) pop_info_cb},
  {"pop_range_dlg_cb", 		  (caddr_t) pop_range_dlg_cb},
  {"pop_retrieve_dlg_cb",	  (caddr_t) pop_retrieve_dlg_cb},
  {"print_full_cb",               (caddr_t) print_full_cb}, 
  {"print_opt_cb",                (caddr_t) print_opt_cb},
  {"print_summ_cb",               (caddr_t) print_summ_cb}, 
  {"quit_cb",                     (caddr_t) quit_cb},
  {"range_cancel_cb",             (caddr_t) range_cancel_cb},
  {"range_ok_cb",                 (caddr_t) range_ok_cb},
  {"really_add_search_cb",        (caddr_t) really_add_search_cb},
  {"really_cheat_cb",             (caddr_t) really_cheat_cb}, 
  {"really_del_search_cb",        (caddr_t) really_del_search_cb},
  {"really_open_save_file_cb",    (caddr_t) really_open_save_file_cb},
  {"really_save_search_cb", 	  (caddr_t) really_save_search_cb},
  {"really_start_browse_cb", 	  (caddr_t) really_start_browse_cb},
  {"retrieve_opts_ok_cb", 	  (caddr_t) retrieve_opts_ok_cb}, 
  {"retrieve_opts_cancel_cb", 	  (caddr_t) retrieve_opts_cancel_cb}, 
  {"save_options_cb", 		  (caddr_t) save_options_cb},
  {"summ_cont_cb", 		  (caddr_t) summ_cont_cb},
  {"save_box_cancel_cb", 	  (caddr_t) save_box_cancel_cb},
  {"q_save_box_cancel_cb", 	  (caddr_t) q_save_box_cancel_cb},
  {"save_search_cb", 		  (caddr_t) save_search_cb},
  {"search_cb", 		  (caddr_t) search_cb}, 
  {"search_cancel_cb", 		  (caddr_t) search_cancel_cb}, 
  {"title_display_cb", 		  (caddr_t) title_display_cb},
  {"toggle_range_limit_cb", 	  (caddr_t) toggle_range_limit_cb},
  {"toggle_save_type_cb", 	  (caddr_t) toggle_save_type_cb},
  {"user_query_cb",               (caddr_t) user_query_cb},
  {"q_toggle_save_type_cb", 	  (caddr_t) q_toggle_save_type_cb},
  {"window_to_top_cb", 		  (caddr_t) window_to_top_cb},
};
int reglist_num = (sizeof reglist / sizeof reglist [0]);

#define USAGE "USAGE: willow [-cf config_file | - ] [standard X args]"

void main(unsigned int argc, char **argv)
                                         /* Command line argument count. */
                                         /* Pointers to command line args. */
{
  char *uid_name;
  Boolean override_config = FALSE;
  int   i;
  int   db_count;
  db_menu_rec *db_menu_list;

/* Following is a bunch of stuff for MIT's benefit. */
#ifdef WILLOW_SEARCH_PATH
  char *oldpath, *newpath;
  extern char* getenv();

  oldpath = getenv("XFILESEARCHPATH");

#if defined(mips) || defined(_AIX) || defined(sparc)
  if (!oldpath) {
      newpath = (char *) malloc (200);
      sprintf (newpath, "XFILESEARCHPATH=%s", WILLOW_SEARCH_PATH);
  }
  else {
      newpath = (char *) malloc (200 + strlen (oldpath));
      sprintf (newpath, "XFILESEARCHPATH=%s:%s", WILLOW_SEARCH_PATH, oldpath);
  }
  putenv (newpath);
#else
  if (!oldpath) {
      newpath = (char *) malloc (100);
      strcpy (newpath, WILLOW_SEARCH_PATH);
  }
  else {
      newpath = (char *) malloc (100 + strlen (oldpath));
      sprintf (newpath, "%s:%s", WILLOW_SEARCH_PATH, oldpath);
  }
  setenv ("XFILESEARCHPATH", newpath, 1);

#endif
#endif

  init_application();
  MrmInitialize();     
  
  /* Initialize the X Toolkit. We get back a top level shell widget. */
  
  toplevel = XtAppInitialize(&app_context,
			     "Willow",
			     NULL, /* No options list */
			     0,
			     &argc,
			     argv,
			     NULL, /* No fallbacks */
			     NULL, /* No args */
			     0);

  /* realizing and unrealizing the toplevel fixes a Motif 1.2.1 bug, 
     where otherwise all dialogs would be modal */
#ifdef MOTIF_1_2_1
  XtRealizeWidget(toplevel);
  XtUnrealizeWidget(toplevel);
#endif

  init_resources();

  /* Process the non-X command line arguments (override the resources */
  /* The only flag is -cf <db config file> */
  argc--; argv++;
  while (argc) {
    if (strcmp("-cf", argv[0]) == 0) {
      if (argv[1]) {
	db_config_file_name = argv[1];
	override_config = TRUE;
	argc -= 2;
	argv += 2;
      }
      else
	error(USAGE);
    }
    else
      error(USAGE);
  }

  /* prepend willow home directory name to uid file names */
  for (i = 0; i < db_filename_num; i++) {
    uid_name = XtMalloc(strlen(willow_dir) + strlen(db_filename_vec[i]) + 1);
    strcpy(uid_name, willow_dir);
    strcat(uid_name, db_filename_vec[i]);
    db_filename_vec[i] = uid_name;
  }

  /* Open the UID files (the output of the UIL compiler) in the hierarchy*/
  if (MrmOpenHierarchy(db_filename_num, /* Number of files. */
		       db_filename_vec, /* Array of file names.  */
		       NULL,            /* Default OS extenstion. */
		       &s_MrmHierarchy) /* Pointer to returned MRM ID */
      !=MrmSUCCESS)
    error("can't open hierarchy");
  
  /* Register the items MRM needs to bind for us. */
  MrmRegisterNames(reglist, reglist_num);


  search_shell = XtVaCreatePopupShell("Search", topLevelShellWidgetClass,
				      toplevel,
				      XmNtitle,
				      FULL_SEARCH_TITLE, NULL);
  /* Add handler for session-manager messages */
#define NonMaskableMask ((EventMask)0x80000000L)
  xa_willow = XmInternAtom(XtDisplay(toplevel), WILLOW_ATOM, False);

  /* Trap the mwm close event in the search shell */
  wm_delete_window = XmInternAtom(XtDisplay(toplevel),
				  "WM_DELETE_WINDOW", False);
  XmAddWMProtocolCallback(search_shell, wm_delete_window,
			  mwm_close_cb, NULL);

  /* Add emergency X-error handler */
  XSetIOErrorHandler((XIOErrorHandler)quit_handler);

  /* Go get the main part of the application. */
  if (MrmFetchWidget(s_MrmHierarchy, "main_window", search_shell,
		     &main_window, &dummy_class) != MrmSUCCESS)
    error("can't fetch main window");

  /* Set up event handler for resize so we can center the status-frame */
  XtInsertEventHandler(widgets[k_status_frame], StructureNotifyMask,
		       FALSE, center_status, NULL, XtListTail);
  XtInsertEventHandler(widgets[k_opacify], StructureNotifyMask,
		       FALSE, center_status, NULL, XtListTail);

  /* Manage the main part and realize everything. */
  XtManageChild(main_window);

  /* This needs to be done before pop-ups */
  set_status_label(UNCONNECTED, "");

  /* put up the windows */
  pop_results();  
  pop_summary();

  /* Handle session manager request via summary window */
  XtAddRawEventHandler(summary_shell, NonMaskableMask, True, 
		       client_message_handler, NULL);
  pop_dict();

  XtRealizeWidget(search_shell);
  if (interface_mode != MANAGED_MODE)
    XMapRaised(XtDisplay(toplevel), XtWindow(search_shell));
  set_icon("search.icon", search_shell);
  save_geometry(search_shell, &search_geom);

  /* Load the globes */
  load_globes(k_search_globe_box);
  load_globes(k_full_globe_box);

  /* Configure for whatever databases are available */
  db_count = config_dbs(db_config_file_name, override_config?NULL:willow_dir,
			&db_hash_tab, &db_menu_list);
  build_db_menu(db_menu_list);

  /* parse the per-database settings from X-resource, now that database
     list has been constructed */
  parse_user_info_list();
  parse_field_pick_list();

  /* Load the users personal preferences, if any */
  read_options();

  /* Configure user-interface */
  config_interface();

  /* Set our focus on first text-box */
  XmProcessTraversal(widgets[k_field1], XmTRAVERSE_CURRENT);
  XmProcessTraversal(widgets[k_field1], XmTRAVERSE_CURRENT);
  /* Did we get an early call to select_by_name? */
  if (stacked_select[0])
    select_by_name(stacked_select);

  /* If there is only one db in the list, may as well log into it */
  if (db_count == 1)
    select_by_name(((db_desc_rec *)hash_FirstEntry(db_hash_tab))->tag);

  /* Sit around forever waiting to process X-events.  We never leave
   * XtMainLoop. From here on, we only execute our callback routines. */
  XtAppMainLoop(app_context);
}

/* Configure the user interface. Mostly take out a bunch of stuff
   if we are operating in anonymous mode.
 */
config_interface(void)
{
  int n;
  int i;
  XmToggleButtonCallbackStruct reason;
  XtTranslations trans;
  XtActionsRec actions[20];

  /* Make status box pop-up database information */
  n = 0;
  actions[n].string = "pop_info_cb";
  actions[n].proc = pop_info_cb; n++;
  XtAppAddActions(app_context, actions, n);
  trans = XtParseTranslationTable(
     "<Btn1Down>: pop_info_cb()\n\
      ");
  XtOverrideTranslations(widgets[k_status_label], trans);

  /* Add some commands to each of the text fields on search window.
     Basic Mode and browse window are handled in init_browse_widgets(),
     since they are continually destroyed/created. ctrl_cmd_action()
     was registered there.*/
  trans = XtParseTranslationTable(
     "Ctrl<Key>t: ctrl_cmd_action(search,browse)\n\
      Shift Ctrl<Key>c: ctrl_cmd_action(search,clear_all)\n\
      Meta Ctrl<Key>c: ctrl_cmd_action(search,clear_all)\n\
      Ctrl<Key>c: ctrl_cmd_action(search,clear)\n\
      Ctrl<Key>v: ctrl_cmd_action(search,help)\n\
      <Key>F1: ctrl_cmd_action(search,help)\n\
      Ctrl<Key>x: ctrl_cmd_action(search,fields)\n\
      ");
  for (n = 0; n < NUM_TEXT_FIELDS; n++)
    XtOverrideTranslations(widgets[k_field1 + n], trans);
  
  /* Add commands to the summary window */
  trans = XtParseTranslationTable(
     "<Key>-: previous-page()\n\
      \" \": next-page()\n\
      Ctrl<Key>m: ctrl_cmd_action(summ,more)\n\
      Ctrl<Key>3: ctrl_cmd_action(summ,more)\n\
      Ctrl<Key>v: ctrl_cmd_action(summ,help)\n\
      <Key>F1: ctrl_cmd_action(summ,help)\n\
      Ctrl<Key>c: ctrl_cmd_action(summ,close)\n\
      Ctrl<Key>y: ctrl_cmd_action(summ,print)\n\
      Ctrl<Key>s: ctrl_cmd_action(summ,save)\n\
      ");
  XtOverrideTranslations(widgets[k_summary_titles], trans);

  /* Add commands to the record retrieval window */
  trans = XtParseTranslationTable(
     "<Key>-: previous-page()\n\
      \" \": next-page()\n\
      Ctrl<Key>v: ctrl_cmd_action(results,help)\n\
      <Key>F1: ctrl_cmd_action(results,help)\n\
      Ctrl<Key>c: ctrl_cmd_action(results,close)\n\
      Ctrl<Key>y: ctrl_cmd_action(results,print)\n\
      Ctrl<Key>s: ctrl_cmd_action(results,save)\n\
      Ctrl<Key>n: ctrl_cmd_action(results,next)\n\
      Ctrl<Key>p: ctrl_cmd_action(results,prev)\n\
      ");
  XtOverrideTranslations(widgets[k_text], trans);

  /* De-sensitize some buttons */
  XtSetSensitive(widgets[k_search], FALSE);
  XtSetSensitive(widgets[k_list_btn], FALSE);

  /* Turn off field buttons */
  for (i = 0; i < NUM_TEXT_FIELDS; i++)
    XtSetSensitive(widgets[k_label1 + i], FALSE);

  /* Use cheat box? */
  if (!enable_cheat)
    XtUnmanageChild(widgets[k_cheat_btn]);

  XtUnmanageChild(widgets[k_opacify]);
  XtUnmanageChild(widgets[k_ext_menu]);

  /* Set widgets to values set through resources or .willowrc */
  connection_opt_cb(NULL, NULL, NULL);
  XmToggleButtonSetState(widgets[k_echo], initial_echo, FALSE);
  XmToggleButtonSetState(widgets[k_ftp_append_btn], ftp_append, FALSE);
  XmToggleButtonSetState(widgets[k_ftp_replace_btn], !ftp_append, FALSE);
  XmToggleButtonSetState(widgets[k_show_search_btn], show_search, TRUE);

  XmTextSetString(widgets[k_ftp_host_text], ftp_host);
  XmTextSetString(widgets[k_q_ftp_host_text], ftp_host);
  XmTextSetString(widgets[k_ftp_uid_text], ftp_uid);
  XmTextSetString(widgets[k_q_ftp_uid_text], ftp_uid);
  XmTextSetString(widgets[k_email_text], mail_addr);
  XmTextSetString(widgets[k_email_subj], mail_subj);

  field_pick_hide();

  if (interface_mode != NORMAL_MODE) {
    save_type = (save_type == SAVE_LOCAL) ? SAVE_EMAIL : save_type;
    q_save_type = SAVE_FTP;
    XtUnmanageChild(widgets[k_print_opt_btn]);
    XtUnmanageChild(widgets[k_conn_opt_btn]);
    XtUnmanageChild(widgets[k_save_opt_btn]);
    XtUnmanageChild(widgets[k_local_save_btn]);
    XtUnmanageChild(widgets[k_q_local_save_btn]);
  }
  else {
    q_save_type = (save_type == SAVE_FTP) ? SAVE_FTP : SAVE_LOCAL;
    /* Set default print cmd if none already set */
    if (strncasecmp(print_cmd, NO_PRINT_CMD, strlen(NO_PRINT_CMD)) == 0)
      print_cmd = "";
    else if (!print_cmd[0])
      print_cmd = DEFAULT_PRINT_CMD;
  }

  /* Set print command into widget (must be after above section) */
  XmTextSetString(XmSelectionBoxGetChild(widgets[k_print_opt_box], 
					 XmDIALOG_TEXT), print_cmd);

  /* Set initial toggle in save_box */
  reason.set = TRUE;
  if (save_type == SAVE_LOCAL)
    XmToggleButtonSetState(widgets[k_local_save_btn], TRUE, TRUE);
  else if (save_type == SAVE_FTP)
    XmToggleButtonSetState(widgets[k_ftp_save_btn], TRUE, TRUE);
  else 
    XmToggleButtonSetState(widgets[k_email_save_btn], TRUE, TRUE);

  if (q_save_type == SAVE_LOCAL)
    XmToggleButtonSetState(widgets[k_q_local_save_btn], TRUE, TRUE);
  else
    XmToggleButtonSetState(widgets[k_q_ftp_save_btn], TRUE, TRUE);

  /* If we are running under session manager, no quit option */
  if (interface_mode == MANAGED_MODE)
    XtUnmanageChild(widgets[k_quit_btn]);

  /* Initialize the text-list pseudo-widget */
  list_init(widgets[k_summary_titles]);
  list_register_callback(widgets[k_summary_titles], SELECT_CB, 
			 title_select_cb);
  list_register_callback(widgets[k_summary_titles], CLICK2_CB, 
			 title_double_click_cb);
}


/* Build the Database menu. We are passed in the list of menu
   directives, and the hash table with the database records in it.
   add_db_submenu is recursive.
 */
static db_menu_rec *add_db_submenu(Widget menu, db_menu_rec *db_menu_list)
{
  db_desc_rec *db;
  char  err_buf[256];
  Arg    args[10];
  int    n;
  Widget button, new_menu;
  db_menu_rec *prev;

  while (db_menu_list) {
    n = 0;
    switch (db_menu_list->type) {
    case MENU_ITEM:
      db = (db_desc_rec *) 
	hash_Lookup(db_hash_tab,
		    hash_HashFunction(db_menu_list->value,
				      strlen(db_menu_list->value)),
		    db_compare, db_menu_list->value);
      if (!db) {
	sprintf(err_buf, "Error in Database menu specification.\nCould not match '%s'",
		db_menu_list->value);
	error(err_buf);
      }
      XtSetArg(args[n], XmNlabelString, 
	       XmStringCreateSimple(db->long_name));n++;
      XtSetArg(args[n], XmNuserData, db);n++;
      button = XmCreatePushButton(menu, "db_button", args, n);
      XtAddCallback(button, XmNactivateCallback, select_db_cb, db);
      XtManageChild(button);
      break;
    case MENU_SEP:
      XtManageChild(XmCreateSeparator(menu, "separator", NULL, NULL));
      break;
    case BEG_MENU:
      new_menu = XmCreatePulldownMenu(menu, "pulldown", NULL, 0);
      XtSetArg(args[n], XmNlabelString, 
	       XmStringCreateSimple(db_menu_list->value));n++;
      XtSetArg(args[n], XmNsubMenuId, new_menu);n++;
      button = XmCreateCascadeButton(menu, "db_button", args, n);
      XtManageChild(button);
      db_menu_list = add_db_submenu(new_menu, db_menu_list->next);
      break;
    case END_MENU:
      return(db_menu_list);
    }
    prev = db_menu_list;
    db_menu_list = db_menu_list->next;
    XtFree(prev->value);
    XtFree(prev);
  }
  return(db_menu_list);
}
void build_db_menu(db_menu_rec *db_menu_list)
{
  if (db_menu_list->type == BEG_MENU) {
    /* Skip the first one which is just a begin_menu */
    add_db_submenu(widgets[k_db_pdm], db_menu_list->next);
    XtFree(db_menu_list->value);
    XtFree(db_menu_list);
  }
  else
    add_db_submenu(widgets[k_db_pdm], db_menu_list);
}

/* Open the results window */
pop_results(void)
{
  results_shell = XtVaCreatePopupShell("Full-Record", topLevelShellWidgetClass,
				      toplevel, XmNtitle,
				      FULL_REC_TITLE, NULL);
  XmAddWMProtocolCallback(results_shell, wm_delete_window,
			  mwm_close_cb, NULL);
  if (MrmFetchWidget(s_MrmHierarchy, 
		     "results_window", 
		     results_shell,
		     &widgets[k_results], 
		     &dummy_class) != MrmSUCCESS)
    error("can't fetch results window widget");
  
  XtSetSensitive(widgets[k_full_save_btn], FALSE);
  XtSetSensitive(widgets[k_full_print], FALSE);
  XtSetSensitive(widgets[k_match_next], FALSE);
  XtSetSensitive(widgets[k_match_prev], FALSE);

  XtManageChild(widgets[k_results]);
  XtRealizeWidget(results_shell);
  set_icon("full_record.icon", results_shell);
  save_geometry(results_shell, &results_geom);
}

/* Open the summary window */
pop_summary(void)
{
  summary_shell = XtVaCreatePopupShell("Summaries", topLevelShellWidgetClass,
				      toplevel, XmNtitle,
				      SUMMARY_TITLE, NULL);
  XmAddWMProtocolCallback(summary_shell, wm_delete_window,
			  mwm_close_cb, NULL);  
  if (MrmFetchWidget(s_MrmHierarchy, 
		     "summary_window", 
		     summary_shell,
		     &widgets[k_summary], 
		     &dummy_class) != MrmSUCCESS)
    error("can't fetch summary window widget");
  
  desensitize_selection_buttons();
  XtSetSensitive(widgets[k_summ_next_block],FALSE);

  /* Set up event handler for resize so we can center the 50 More button */
  XtInsertEventHandler(widgets[k_sum_dum], StructureNotifyMask,
		       FALSE, center_summ, NULL, XtListTail);

  XtManageChild(widgets[k_summary]);
  XtRealizeWidget(summary_shell);
  set_icon("summaries.icon", summary_shell);
  save_geometry(summary_shell, &summary_geom);
}

/* Set up an icon */
set_icon(char *icon_name, Widget shell)
{
  int   x_hot, y_hot, status;
  unsigned int width, height;
  Pixmap  pix;
  int  n;
  Arg  args[10];
  char icon_file[MAX_LINE];

  sprintf(icon_file, "%s%s%s", willow_dir, BITMAP_DIR, icon_name);
  status = XReadBitmapFile(XtDisplay(shell), XtWindow(shell),
			   icon_file, &width, &height, &pix, &x_hot, &y_hot);
  if (status != BitmapSuccess)
    return;

  n = 0;
  XtSetArg(args[n], XmNiconPixmap, pix);n++;
  XtSetValues(shell, args, n);
}


/*
 * X Callback functions
 */

/* Called via XclientMessage, session manager is directing us to
   pick new database by name.
 */
select_by_name(char *name)
{
  db_desc_rec *db;

  if (!db_hash_tab) {
    /* We have not initialized the databases yet */
    strcpy(stacked_select, name);
    return;
  }

  if (status != UNCONNECTED)
    disconnect();

  db = (db_desc_rec *) 
    hash_Lookup(db_hash_tab,
		hash_HashFunction(name, strlen(name)),
		db_compare, name);
  if (db)
    select_db_cb(NULL, db, NULL);
}

/*
   Called when a database is selected. We need to tell the database
   driver to connect us to a new database. How we proceed depends on 
   whether we are currently connected. If so, we must disconnect first, 
   then we get called again recursively. We need to tell send_connect_cmd 
   what database to log us into.
 */
void select_db_cb(Widget w, db_desc_rec *db, u_long *reason)
{
  static db_desc_rec *next_db;
  int    i;
  char   *msg;

  /* Turn field buttons back on */
  for (i = 0; i < NUM_TEXT_FIELDS; i++)
    XtSetSensitive(widgets[k_label1 + i], TRUE);

  if (cur_db == db)
    return;
  if (db)
    next_db = db;
  
  if (status == SUMMARIZING) 
    status = IDLE;
  if (opac_mode)
    de_opacify();
  not_ready_buttons();
  start_display();
  clear_titles();

  dict_hide();
  init_browse(db);
  clear_help();

  if (status == IDLE) {
    pop_working("Changing database...", search_shell);

    /* in this case, we end up back in this routine */
    disconnect();
    status = CONNECTING;
    select_db_cb(w, db, reason);
  }
  else if ((status == UNCONNECTED) || (status == CONNECTING)) {
    set_status_label(CONNECTING, next_db->short_name);
    pop_working("Starting database driver program...", search_shell);
    status = CONNECTING;
    if (msg = send_connect_cmd(next_db, willow_dir)) {
      pop_warning(msg, NULL, search_shell);
      no_db_cb(NULL, NULL, NULL);
    }
    else
      really_select_db(NULL, next_db, NULL);
  }
}

/* Called by the "No Database" button, disconnect from current db. */
void no_db_cb(Widget w, int *tag, u_long *reason)
{
  int i;

  if (status == UNCONNECTED)
    return;
  if (status != IDLE)
    summaries_done();
  disconnect();
  if (opac_mode)
    de_opacify();
  XtUnmanageChild(widgets[k_opacify]);

  /* Turn off field buttons */
  for (i = 0; i < NUM_TEXT_FIELDS; i++)
    XtSetSensitive(widgets[k_label1 + i], FALSE);
}

really_select_db(Widget w, db_desc_rec *db_desc, u_long *reason)
{
  int  i, j, which;
  Arg  args[10];
  int  n;
  Widget menu_pane, dflt_btn;
  WidgetList  old_buttons,     /* buttons to be changed */
              new_buttons;     /* buttons to be created   */
  int   old_cnt,new_cnt;       /* count of buttons modified/created */
  int  num_children;
  db_field_desc_rec *db_field, *dflt_field;
  static Boolean first_time = TRUE;
  Search_ptr query;

  /* Hold on to current search */
  query = package_search();
  clear_cb(NULL, NULL, NULL);

  cur_db = db_desc;
  new_buttons= (Widget*) XtMalloc(db_desc->num_text_fields * sizeof(Widget*));
  /* Go through the Text fields, and rebuild the options menus */
  for (i = 0; i < NUM_TEXT_FIELDS; i++) {
    which = i + k_label1;
    n = 0;
    XtSetArg(args[n], XmNsubMenuId, &menu_pane);n++;
    XtGetValues(widgets[which], args, n);

    /* Now, go through the menu-pane, and change the labels. Create 
       new buttons as needed. */
    n = 0;
    XtSetArg(args[n], XmNnumChildren, &num_children);n++;
    XtSetArg(args[n], XmNchildren, &old_buttons);n++;
    XtGetValues(menu_pane, args, n);
    /* Need to get old_buttons into our own storage, else Xt will re-use it */
    old_buttons 
      = (WidgetList) bdup((char *) old_buttons, num_children * sizeof(old_buttons[0]));

    /* Go through the list of Text fields in the db structure,
       changing or creating buttons as needed */
    dflt_field = db_desc->defaults[i];
    db_field = db_desc->text_fields;
    new_cnt = old_cnt = 0;
    while (db_field != NULL) {
      if (!db_field->opac_only) {
	n = 0;
	XtSetArg(args[n], XmNlabelString,
		 XmStringCreateSimple(db_field->name));n++;
	XtSetArg(args[n], XmNuserData, db_field);n++;
	if (old_cnt < num_children) {
	  XtSetValues(old_buttons[old_cnt], args, n);
	  if (dflt_field == db_field)
	    dflt_btn = old_buttons[old_cnt];
	  if (!XtIsManaged(old_buttons[old_cnt]))
	    XtManageChild(old_buttons[old_cnt]);
	  
	  /* We have to add callback to our single pre-built button */
	  if (first_time && (old_cnt == 0)) {
	    XtAddCallback(old_buttons[old_cnt], XmNactivateCallback,
			  setup_range_cb, i);
	    XtAddCallback(old_buttons[old_cnt], XmNactivateCallback,
			  set_focus_cb, i);
	  }
	  old_cnt++;
	}
	else {
	  new_buttons[new_cnt] = 
	    XmCreatePushButton(menu_pane, "button", args, n);
	  XtAddCallback(new_buttons[new_cnt], XmNactivateCallback,
			setup_range_cb, i);
	  XtAddCallback(new_buttons[new_cnt], XmNactivateCallback,
			set_focus_cb, i);
	  if (dflt_field == db_field)
	    dflt_btn = new_buttons[new_cnt];
	  new_cnt++;
	}
      }
      db_field = db_field->next;
    }
    if (old_cnt < num_children) 
      XtUnmanageChildren(&old_buttons[old_cnt], num_children - old_cnt);

    XtManageChildren(new_buttons, new_cnt);
    /* Motif won't recognize changed label if XmNmenuHistory
       button has same ID, so spoof it. Won't work if there is
       only one button, but that is unlikely.
     */
    n = 0;
    if (dflt_btn != old_buttons[0]){
      XtSetArg(args[n], XmNmenuHistory, old_buttons[0]);n++;
    }
    else if (old_cnt > 1){
      XtSetArg(args[n], XmNmenuHistory, old_buttons[1]);n++;
    }
    else {
      XtSetArg(args[n], XmNmenuHistory, new_buttons[0]);n++;
    }
    XtSetValues(widgets[which], args, n);

    /* Now set MenuHistory to real value */
    n = 0;
    XtSetArg(args[n], XmNmenuHistory, dflt_btn);n++;
    XtSetValues(widgets[which], args, n);

    /* Check if it is a range limit */
    setup_range_cb(dflt_btn, i, NULL);

    /* Save default button for later use with Clear */
    db_desc->dflt_btns[i] = dflt_btn;
  }
    
  /* Go through the list of Limit fields, and adjust the interface */
  db_field = db_desc->limit_fields;
  j = 0;
  for (which = k_bool1; which < k_bool1 + NUM_BOOL_FIELDS; which++) {
    XtUnmanageChild(widgets[which]);
    if (db_field) {
      n = 0;
      XtSetArg(args[n], XmNlabelString, 
	       XmStringCreateSimple(db_field->name));n++;
      XtSetArg(args[n], XmNset, FALSE);n++;
      XtSetArg(args[n], XmNuserData, db_field);n++;
      XtSetValues(widgets[which], args, n);
      new_buttons[j++] = widgets[which];
      db_field = db_field->next;
    }
  }
  if (first_time)
    first_time = FALSE;
  XtManageChildren(new_buttons, db_desc->num_limit_fields);
  XtFree(new_buttons);
  XtFree(old_buttons);
  if (db_desc->opac_mode)
    opacify(FALSE);
  else
    XtUnmanageChild(widgets[k_opacify]); 

  init_extensions(db_desc);

  /* Get rid of/restore prev/next on full-rec */
  if (db_desc->use_prev_next) {
    XtManageChild(widgets[k_match_prev]);
    XtManageChild(widgets[k_match_next]);
    XtManageChild(widgets[k_match_label]);
  }
  else {
    XtUnmanageChild(widgets[k_match_prev]);
    XtUnmanageChild(widgets[k_match_next]);
    XtUnmanageChild(widgets[k_match_label]);
  }
  unpack_search_cb(NULL, query, NULL);
  free_search(query);
}

/* Turn the interface into simplified OPAC (Online Public Access Catalog) 
   mode. */
void opacify_cb(Widget w, int *tag, XtPointer reason)
{
  if (opac_mode)
    de_opacify();
  else
    opacify(TRUE);
}

void opacify(Boolean force) 
{
  db_field_desc_rec *field, *first_field = NULL;
  char label[MAX_LINE];
  char *blank;
  Widget button;
  Dimension form_w, rc_w;

  if (!cur_db)
    return;

  if (no_basic_mode) {
    XtVaSetValues(widgets[k_opacify], XmNlabelString, 
		  XmStringCreateLtoR(OPACIFY_LABEL, XmSTRING_DEFAULT_CHARSET),
		  NULL);
    XtManageChild(widgets[k_opacify]);
    if (!force)
      return;
  }
  
  XtVaSetValues(search_shell, XmNtitle, OPAC_SEARCH_TITLE, NULL);
  XtUnmanageChild(widgets[k_search_form]);
  list_destroy(widgets[k_dict_text]);
  XtDestroyWidget(dict_shell);
  dict_shell = NULL;

  widgets[k_dict_form] = NULL;
  hide_window(k_summary_titles);
  hide_window(k_results);
  if (widgets[k_mbrowse_box])
    clear_dialog_cb(NULL, widgets[k_mbrowse_box], NULL);

  if (MrmFetchWidget(s_MrmHierarchy, 
		     "opac_form", 
		     widgets[k_dummy_form],
		     &widgets[k_opac_form], 
		     &dummy_class) != MrmSUCCESS)
    error("can't fetch opac form widget");

  XtVaSetValues(widgets[k_opac_form], XmNtopWidget, 
		widgets[k_status_frame], NULL);

  /* Create the choices buttons based on current lists */
  field = cur_db->text_fields;
  while (field) {
    if (field->opac_field) {
      if (!first_field)
	first_field = field;
      strcpy(label, field->name);
      /* Change ' ' to \n in name */
      if (blank = strchr(label, ' '))
	*blank = '\n';
      button = XtVaCreateWidget("opac_field", xmPushButtonWidgetClass, 
	widgets[k_opac_choices], XmNtraversalOn, FALSE, XmNlabelString, 
	XmStringCreateLtoR(label, XmSTRING_DEFAULT_CHARSET), NULL);
      XtManageChild(button);
      XtAddCallback(button, XmNactivateCallback, start_browse_cb, field);
    }
    field = field->next;
  }
  XtVaSetValues(widgets[k_opacify], XmNlabelString, 
	  XmStringCreateLtoR(DE_OPACIFY_LABEL, XmSTRING_DEFAULT_CHARSET),
	     NULL);
  XtManageChild(widgets[k_opacify]);
  
  XtUnmanageChild(widgets[k_searches_btn]);
  XtManageChild(widgets[k_opac_form]);
  XtManageChild(widgets[k_opac_choices]);

  /* Now center the button box */
  XtVaGetValues(widgets[k_opac_choices], XmNwidth, &rc_w, NULL);
  XtVaGetValues(search_shell, XmNwidth, &form_w, NULL);
  XtVaSetValues(widgets[k_opac_choices], XmNleftOffset,
		(form_w > rc_w) ? (form_w - rc_w) / 2 : 0, NULL);

  XtUnmanageChild(widgets[k_dict_paste]);
  XtSetSensitive(widgets[k_opac_search], XtIsSensitive(widgets[k_search]));
  grow_opac();

  /* Trigger browse on first list */
  init_browse_widgets();
  XtManageChild(widgets[k_opac_search]);
}

/* Take opacization one more step */
grow_opac(void) {
  Dimension  rc_h, rc_top, mb_h, status_h, text_top;

  if (opac_mode == 0) {
    /* Shrink the form to height of button box, plus top and bottom margin, 
       plus height of top menu bar and the status frame area*/
    XtVaGetValues(widgets[k_opac_choices], XmNheight, &rc_h,
		  XmNy, &rc_top, NULL);
    XtVaGetValues(widgets[k_menu_bar], XmNheight, &mb_h, NULL);
    XtVaGetValues(widgets[k_opac_form], XmNy, &status_h, NULL);
    XtVaSetValues(search_shell, XmNheight, mb_h + rc_h + status_h +
		  (rc_top * 2), NULL);
  }

  else if (opac_mode == 1){
    /* Grow to old height, plus type-in area */
    XtVaGetValues(widgets[k_opac_choices], XmNheight, &rc_h,
		  XmNy, &rc_top, NULL);
    XtVaGetValues(widgets[k_menu_bar], XmNheight, &mb_h, NULL);
    XtVaGetValues(widgets[k_dict_stext], XmNy, &text_top, NULL);
    XtVaGetValues(widgets[k_opac_form], XmNy, &status_h, NULL);
    XtVaSetValues(search_shell, XmNheight, mb_h + rc_h + (rc_top * 4) + 
		  status_h + text_top + 1, NULL);
    XtUnmanageChild(widgets[k_dict_cancel]);
  }
  
  else if (opac_mode == 2) 
    XtVaSetValues(search_shell, XmNheight, (Dimension) (search_geom.h * 1.4),
		  NULL);

  opac_mode++;
}

/* Turn off opac mode */
void de_opacify(void)
{
  XtVaSetValues(search_shell, XmNtitle, FULL_SEARCH_TITLE, NULL);
  opac_mode = FALSE;
  XtUnmanageChild(widgets[k_opac_form]);
  list_destroy(widgets[k_dict_text]);
  XtDestroyWidget(widgets[k_opac_form]);

  widgets[k_opac_form] = NULL;
  if (widgets[k_mbrowse_box])
    clear_dialog_cb(NULL, widgets[k_mbrowse_box], NULL);

  XtVaSetValues(widgets[k_opacify], XmNlabelString, 
	  XmStringCreateLtoR(OPACIFY_LABEL, XmSTRING_DEFAULT_CHARSET),
		NULL);

  /* Restore shell size */
  XtVaSetValues(search_shell, XmNheight, search_geom.h,	NULL);

  XtManageChild(widgets[k_search_form]);
  XtManageChild(widgets[k_searches_btn]);
  pop_dict();
}

/* 
   When focus comes into a new text field, we have to adjust the list
   browse button.
*/
void field_focus_cb(Widget w, int *tag, u_long *reason)
{
  XmString label;
  char    *label_str;
  char     btn_label_str[MAX_LINE];
  db_field_desc_rec *field;
  Widget   button;
  static Widget prev_w = NULL;

  last_focused_field = *tag - k_field1;
  if (!cur_db)
    return;
  
  /* cancel the browse list if it is up */
  if (prev_w != w)
    dict_cancel_cb(NULL, NULL, NULL);
  prev_w = w;

  /* First get the id of the menu-button of the corresponding option menu */
  XtVaGetValues(widgets[k_label1 + last_focused_field], XmNmenuHistory, 
		&button, NULL);
  
  /* Then get the string value of the current setting for that menu */
  XtVaGetValues(button, XmNlabelString, &label, NULL);
  label_str = xmstring_to_string(label);
  XmStringFree(label);

  /* Find out if there is a corresponding browse list */
  for (field = cur_db->text_fields; field != NULL; field = field->next)
    if (strcmp(field->name, label_str) == 0)
      break;

  XtUnmanageChild(widgets[k_list_btn]);
  if (field->list_cnt) {
    XtSetSensitive(widgets[k_list_btn], TRUE);
    sprintf(btn_label_str, LIST_BTN_FMT, field->name);
    XtVaSetValues(widgets[k_list_btn], XmNlabelString,
         XmStringCreateLtoR(btn_label_str, XmSTRING_DEFAULT_CHARSET), NULL);
    XtRemoveAllCallbacks(widgets[k_list_btn], XmNactivateCallback);
    XtAddCallback(widgets[k_list_btn], 
		  XmNactivateCallback, start_browse_cb, field);
    /* Add the corresponding text box to the user data field of widget */
    XtVaSetValues(widgets[k_list_btn], XmNuserData, w, NULL);
  }
  else {
    XtVaSetValues(widgets[k_list_btn], XmNlabelString,
         XmStringCreateLtoR(DFLT_LIST_BTN, XmSTRING_DEFAULT_CHARSET), NULL);
    XtSetSensitive(widgets[k_list_btn], FALSE);
  }
  XtManageChild(widgets[k_list_btn]);
  XtFree(label_str);
}

/* Set the focus to a new text field when option menu is used */
void set_focus_cb(Widget w, int which, u_long *reason)
{
  XmProcessTraversal(widgets[k_field1 + which], XmTRAVERSE_CURRENT);
}

/* Set the status label to reflect the appropriate database(s) */
set_status_label(int status, char *msg)
{
  char  str[MAX_LINE];
  char *status_str;

  switch (status) {
  case UNCONNECTED:
    status_str = UNCONNECTED_S;
    break;
  case CONNECTING:
    status_str = CONNECTING_S;
    break;
  case IDLE:
    status_str = IDLE_S;
    break;
  }
    
  sprintf(str,"%s%s", status_str, msg);

  /* Need this to fix a centering glitch */
  XtVaSetValues(widgets[k_status_frame], XmNleftOffset, 5, NULL);

  XtVaSetValues(widgets[k_status_label], XmNlabelString,
	      XmStringCreateLtoR(str, XmSTRING_DEFAULT_CHARSET), NULL);
}

/* Unmanage the connection options box */
void conn_opts_ok_cb(Widget w, int *tag, u_long *reason)
{
  clear_dialog_cb(w, widgets[k_conn_options_box], reason);
}

/* Log each widget id in our array as widgets are created */
void create_cb(Widget w, int *tag, u_long *reason)
{
  int i;
  range_stat_rec *stat_ptr;

  if ((*tag == k_search_cancel) && widgets[*tag])
    widgets[k_full_cancel] = w;
  else if ((*tag == k_search_globe_image) && widgets[*tag])
    widgets[k_full_globe_image] = w;
  else if ((*tag == k_search_globe_label) && widgets[*tag])
    widgets[k_full_globe_label] = w;
  else if ((*tag == k_from1) || (*tag == k_to1)) {
    for (i = *tag; widgets[i] != NULL; i++)
      ;
    widgets[i] = w;

    /* Attach a range status record to the button */
    stat_ptr = (range_stat_rec *) XtMalloc(sizeof(range_stat_rec));
#if defined(POSIX) || defined (_POSIX_SOURCE)
    memset(stat_ptr, 0, sizeof(range_stat_rec));
#else /* defined(POSIX) || defined (_POSIX_SOURCE) */
    bzero(stat_ptr, sizeof(range_stat_rec));
#endif /* defined(POSIX) || defined (_POSIX_SOURCE) */
    XtVaSetValues(w, XmNuserData, stat_ptr, NULL);
  }
  else
    widgets[*tag] = w;
}

/* Called when the echo button is pressed in the options dialog.*/
void echo_opt_cb(Widget w, int *tag, u_long *reason)
{
  if (status != UNCONNECTED)
    send_echo_cmd();
}

/* Jump to next hilite in Full-Record display. Scan list of hilites, until
   we find first one that is beyond the end of visible portion of screen,
   call ShowPosition to put it at bottom line of widget, then scroll
   to center it.
  */
void match_next_cb(Widget w, int *tag, u_long *reason)
{
  int     top;     /* Position of first visible character */
  hl_rec *cur;
  Position dummy;
  short   rows;
  int     n;
  Arg     args[10];
  Boolean didit = FALSE;

  if (!hl_head)
    return;
  
  top = XmTextGetTopCharacter(widgets[k_text]);
  n = 0;
  XtSetArg(args[n], XmNrows, &rows);n++;
  XtGetValues(widgets[k_text], args, n);

  cur = hl_head;
  while (cur) {
    if ((cur->start > top) && !didit &&  
	(!XmTextPosToXY(widgets[k_text], cur->start, &dummy, &dummy))) {
      XmTextShowPosition(widgets[k_text], cur->start);
      XmTextScroll(widgets[k_text], rows/2);
      didit = TRUE;
    }
    /* As long as we are traversing the list, refresh the highlights, since
       they can get zapped by selection. */
    XmTextSetHighlight(widgets[k_text], cur->start, cur->end, 
		       XmHIGHLIGHT_SELECTED);
    cur = cur->next;
  }
}

void match_prev_cb(Widget w, int *tag, u_long *reason)
{
  int     top;     /* Position of first visible character */
  hl_rec *cur;
  short   rows;
  int     n;
  Arg     args[10];
  Boolean didit = FALSE;

  if (!hl_head)
    return;
  
  top = XmTextGetTopCharacter(widgets[k_text]);
  n = 0;
  XtSetArg(args[n], XmNrows, &rows);n++;
  XtGetValues(widgets[k_text], args, n);

  cur = hl_tail;
  while (cur) {
    if ((cur->start < top) && !didit)  {
      XmTextShowPosition(widgets[k_text], cur->start);
      XmTextScroll(widgets[k_text], -(rows/2));
      didit = TRUE;
    }
    XmTextSetHighlight(widgets[k_text], cur->start, cur->end, 
		       XmHIGHLIGHT_SELECTED);
    cur = cur->prev;
  }
}

/* Pop-up the working dialog. */
pop_working(char *message, Widget parent)
{
  int button_mask;

  button_mask = 0;
  pop_message_dialog(XmDIALOG_WORKING, message, button_mask, parent);
}

/* Pop-up a warning dialog. If second parameter is not null, then
   sprintf it into first */
void pop_warning(char *template, char *arg, Widget parent)
{
  char *message;
  int button_mask;

  button_mask = OK_BUTTON_MASK;
  if (arg == NULL)
    pop_message_dialog(XmDIALOG_WARNING, template, button_mask, parent);
  else {
    message = XtMalloc(strlen(template) + strlen(arg) + 1);
    sprintf(message, template, arg);
    pop_message_dialog(XmDIALOG_WARNING, message, button_mask, parent);
    XtFree(message);
  }
}

/* Clear the working dialog */
clear_working(void)
{
  Arg args[10];
  int  n;
  unsigned char type;

  if (!widgets[k_message])
    return;

  /* Check to see if the current dialog is in fact a working dialog */
  n = 0;
  XtSetArg(args[n], XmNdialogType, &type);n++;
  XtGetValues(widgets[k_message], args, n);

  if (type == XmDIALOG_WORKING)
    XtUnmanageChild(widgets[k_message]);
}

/* Called if the user selects 'close' from mwm menu */
void mwm_close_cb(Widget w, caddr_t client_data, caddr_t call_data)
{
  pop_warning("If you really want to quit, select\nQuit from the File menu", NULL, search_shell);
}

/* This function totally resets willow's state */
reset(void)
{
  db_desc_rec *db;
  user_info_rec *info_ptr;
  int i;

  /* Go back to default geometry */
  set_geometry(search_shell, &search_geom);
  set_geometry(summary_shell, &summary_geom);
  set_geometry(results_shell, &results_geom);
  if (dict_shell)
    set_geometry(dict_shell, &dict_geom);

  /* Zap all the text and boolean fields */
  clear_cb(NULL, NULL, NULL);
  reset_field_labels();

  /* Disconnect us */
  if ((status == SEARCHING) || (status == SUMMARIZING)) {
    search_cancel_cb(NULL, NULL, NULL);
    clear_globe(k_search_globe_box);
    clear_globe(k_full_globe_box);
  }
  if (status != UNCONNECTED) 
    no_db_cb(NULL, NULL, NULL);

  /* Delete all the saved queries */
  del_all_searches();

  /* Restore save options to default state */
  XmToggleButtonSetState(widgets[k_email_save_btn], TRUE, TRUE);
  XmToggleButtonSetState(widgets[k_show_search_btn], FALSE, FALSE);
  XmTextSetString(widgets[k_email_text],    "");
  XmTextSetString(widgets[k_email_subj], DEFAULT_MAIL_SUBJ);
  XmTextSetString(widgets[k_ftp_host_text], "");
  XmTextSetString(widgets[k_ftp_uid_text],  "");
  XmTextSetString(widgets[k_ftp_file_text], "");
  XmTextSetString(widgets[k_ftp_pw_text], "");
  XmTextSetString(widgets[k_q_ftp_host_text], "");
  XmTextSetString(widgets[k_q_ftp_uid_text],  "");
  XmTextSetString(widgets[k_q_ftp_file_text], "");
  XmTextSetString(widgets[k_q_ftp_pw_text], "");

  /* We need to reset these too, otherwise old values get restored of
     save-box is canceled before next ok */
  mail_addr = ftp_host = ftp_uid = ftp_pw = ftp_file = ftp_q_file = "";
  mail_subj = DEFAULT_MAIL_SUBJ;

  XmTextSetString(widgets[k_dict_search], "");
  
  /* Any globals that need to be reset? (ugly!) */
  displaying = FALSE;

  /* Password needs special handling */
  XtFree(ftp_pw);
  ftp_pw = XtMalloc(1);
  ftp_pw[0] = 0;

  /* Clear any dialogs that are up */
  clear_working();
  clear_dialog_cb(NULL, widgets[k_save_box], NULL);
  clear_dialog_cb(NULL, widgets[k_q_save_box], NULL);
  clear_dialog_cb(NULL, widgets[k_cheat_box], NULL);
  clear_dialog_cb(NULL, widgets[k_about_box], NULL);
  clear_dialog_cb(NULL, widgets[k_del_search], NULL);
  clear_dialog_cb(NULL, widgets[k_add_search], NULL);
  clear_dialog_cb(NULL, widgets[k_info_dlg], NULL);
  clear_dialog_cb(NULL, widgets[k_user_query_box], NULL);
  abbrev_fields = FALSE;
  retrieve_opts_cancel_cb(NULL, NULL, NULL);
  dict_cancel_cb(NULL, NULL, NULL);

  /* Reset per-database saved preferences (user-info and full-rec-fields) */
  for (db = (db_desc_rec *) hash_FirstEntry(db_hash_tab);
       db; db = (db_desc_rec *) hash_NextEntry(db_hash_tab)) {
    for (info_ptr = db->user_info; info_ptr; info_ptr = db->user_info) {
      if (info_ptr->tag)
	XtFree(info_ptr->tag);
      if (info_ptr->response)
	XtFree(info_ptr->response);
      db->user_info = info_ptr->next;
      XtFree(info_ptr);
    }
    if (db->display_fields) {
      for (i = 0; db->display_fields[i]; i++)
	XtFree(db->display_fields[i]);
      XtFree(db->display_fields);
      db->display_fields = NULL;
    }
  }
}

/* Send the newly arrived driver-packet to appropriate routine */
dispatch_packet(packet_rec *packet)
{
  char msg[256];

  switch (packet->type) {
  case CONNECTED:
    connected_dcb(packet);
    break;
  case ERROR:
    error_dcb(packet);
    break;
  case STATS:
    stats_dcb(packet);
    break;
  case TITLE:
    title_dcb(packet);
    break;
  case EXTS:
    exts_dcb(packet);
    break;
  case RECORD:
    record_dcb(packet);
    break;
  case INFO:
    info_dcb(packet);
    break;
  case USER_REQ:
    user_req_dcb(packet);
    break;
  case BROWSE:
    browse_dcb(packet);
    break;
  default:
    sprintf(msg, "unknown packet type: %d", packet->type);
    error(msg);
  }
}


XtEventHandler client_message_handler(Widget w, XtPointer cl, XEvent *ev, caddr_t call)
{
  static Boolean first_unmap = TRUE;

  if ((ev->xclient.type==ClientMessage) && 
      (ev->xclient.message_type == xa_willow))  {
    switch (ev->xclient.data.b[0]){
    case RESET_CM :
      if (debug_level){
	fprintf(stderr, "===>SESSION MANAGER RESET REQUEST<===\n");
	fflush(stderr);
      }
      reset();
      break;
    case MAP_CM :
      if (debug_level){
	fprintf(stderr, "===>SESSION MANAGER MAP REQUEST<===\n");
	fflush(stderr);
      }
      pop_window(k_search);
      break;
    case UNMAP_CM :
      if (debug_level){
	fprintf(stderr, "===>SESSION MANAGER UNMAP REQUEST<===\n");
	fflush(stderr);
      }
      if (first_unmap) {
	/* Lame hack to temporarily fix "ghost windows" */
	first_unmap = FALSE;
	
	pop_window(k_results);
	ForceDialog(results_shell);
	pop_window(k_summary_titles);
	ForceDialog(summary_shell);

	if (dict_shell) {
	  /* If host is *really* slow, this might come in before reset()
	     is finished, so check that we have dict_shell */
	  pop_window(k_dict_form);
	  ForceDialog(dict_shell);
	  hide_window(k_dict_form);
	}
	hide_window(k_results);
	hide_window(k_summary_titles);
      }
      hide_all_windows();
      break;
    case SELECT_DB_CM :
      if (debug_level){
	fprintf(stderr, "===>SESSION MANAGER SELECT %s REQUEST<===\n",
		&ev->xclient.data.b[1]);
	fflush(stderr);
      }
      select_by_name(&ev->xclient.data.b[1]);
      break;
    case QUIT_CM :
      if (debug_level){
	fprintf(stderr, "===>SESSION MANAGER QUIT REQUEST<===\n");
	fflush(stderr);
      }
      quit_cb(NULL, NULL, NULL);
      break;
    default:
      if (debug_level){
	fprintf(stderr, "===>UNKNOWN SESSION MANAGER REQUEST???<===\n");
	fflush(stderr);
      }
      printf("Unrecognized protocol thang! %d", ev->xclient.message_type);fflush(stdout);
    } 
  }
}

/* Send a command directly to the host */ 
void really_cheat_cb(Widget w, int *tag, XmSelectionBoxCallbackStruct *reason)
{
  char *command;

  command = xmstring_to_string(reason->value);
  XmTextSetString(XmSelectionBoxGetChild(widgets[k_cheat_box], XmDIALOG_TEXT),
		  "");

  if (command) {
    command = XtRealloc(command, strlen(command) + 2);
    strcat(command, "\n");
  }
  else {
    command = XtMalloc(2);
    command = "\n";
  }
  send_cheat_cmd(command);
  XtFree(command);
}


/*
 * Database Callback Routines
 */
void connected_dcb(packet_rec *packet)
{
  status = IDLE;
  clear_working();
  ready_buttons();
  set_status_label(IDLE, cur_db->short_name);
}

void info_dcb(packet_rec *packet)
{
  if (packet->var.info.type == COPYRIGHT_INFO) {
    db_info = strdup(packet->data);
    if (!cur_db->no_info)
      pop_info_cb(NULL, NULL, NULL);
  }
  else /* it is FIELDS_INFO */
    got_field_list(packet->var.info.fields_cnt, packet->data);
}


disconnect(void)
{
  if (cur_db) {
    dict_cancel_cb(NULL, NULL, NULL);
    send_disconnect_cmd();
  }
  clear_working();
  if (db_info) {
    XtFree(db_info);
    db_info = NULL;
  }
  set_status_label(UNCONNECTED, "");
  field_pick_hide();
  XtSetSensitive(widgets[k_db_btn], TRUE);
  XtSetSensitive(widgets[k_search], FALSE);
  start_display();
  clear_titles();
  cur_db = NULL;
  status = UNCONNECTED;
  displaying = FALSE;
}

void error_dcb(packet_rec *packet)
{
  switch (packet->var.error.type) {
  case BAD_SEARCH:
    clear_globe(k_search_globe_box);
    if (packet->data_len == 0)
      pop_warning("Search statement rejected by database", NULL, search_shell);
    else
      pop_warning("ERROR from database driver:\n  %s", packet->data, search_shell);
    status = IDLE;
    ready_buttons();
    break;
  case MISC:
    if ((status == SEARCHING) || (status == SUMMARIZING))
      clear_globe(k_search_globe_box);
    pop_warning("ERROR from database driver:\n%s", packet->data, search_shell);
    no_db_cb(NULL, NULL, NULL);
    break;
  case WARNING:
    pop_warning("WARNING from database driver:\n  %s", packet->data, search_shell);
    break;
  case UPDATE:
    pop_working(packet->data, search_shell);
    break;
  case DEAD:
    if (status != UNCONNECTED) {
      if ((status == SEARCHING) || (status == SUMMARIZING))
	clear_globe(k_search_globe_box);
      pop_warning("The database driver has unexpectedly died", NULL, search_shell);
      no_db_cb(NULL, NULL, NULL);
    }
    break;
  default:
    error("Unknown error from driver???");
    break;
  }
}

/* Received a record packet, need to display the text, mark the hilites,
   and adjust the interface, or put it in summary */
void record_dcb(packet_rec *packet)
{
  int n;
  hl_rec *hl = NULL, *hl_prev = NULL;
  int     hl_offset;
  static  int title_dex;
  char    msg[MAX_LINE];
  netint  hl_buf;

  if (packet->var.record.summary) {
    display_summary(packet->var.record.ref_num, packet->data + 
		    (packet->var.record.num_hilites * 2 * sizeof(netint)));
    return;
  }
    
  /* First full rec for this request? */
  if (!text_len){
    pop_window(k_results);
    ForceDialog(results_shell);
    title_dex = 0;
    free_hl_list();
    hl_head = hl_tail = NULL;
    XtSetSensitive(widgets[k_match_next], TRUE);
    XtSetSensitive(widgets[k_match_prev], TRUE);
  }
  else
    append_to_view("\n");

  title_dex++;
  sprintf(msg, FULL_GLOBE_MSG_FMT, title_dex, display_titles_count);
  XtVaSetValues(widgets[k_full_globe_label], XmNlabelString, 
		XmStringCreateLtoR(msg, XmSTRING_DEFAULT_CHARSET), NULL);

  if ((title_dex == display_titles_count) ||
      !XtIsSensitive(widgets[k_full_cancel])) {
    int *dummy, dummy_cnt;

    XtSetSensitive(widgets[k_full_cancel], TRUE);
    display_titles_count = 0;
    clear_globe(k_full_globe_box);

    /* Check to see if there is nothing else highlighted now */
    if (list_get_selected(widgets[k_summary_titles], &dummy, &dummy_cnt)){
      XtSetSensitive(widgets[k_summ_disp], TRUE);
      XtSetSensitive(widgets[k_ext_btn], TRUE);
      XtFree(dummy);
    }
    else
      desensitize_selection_buttons();

    if (print_cmd[0])
      XtSetSensitive(widgets[k_full_print], TRUE);
    XtSetSensitive(widgets[k_full_save_btn], TRUE);
    displaying = FALSE;
    ready_buttons();
    if (status == SUMMARIZING) {
      send_title_req_cmd(cur_summary + 1);
    } 
    if ((status == IDLE) && paused)
      XtSetSensitive(widgets[k_summ_next_block], TRUE);
  }
  else
    send_record_req_cmd(display_titles[title_dex], FALSE, abbrev_fields);

  hl_offset = text_len;
  append_to_view(packet->data + 
		 (packet->var.record.num_hilites * 2 * sizeof(netint)));

  /* Set up hiliting as needed */

  for (n = 0; n < packet->var.record.num_hilites; n++) {
    hl = (hl_rec *) XtMalloc(sizeof(hl_rec));
    hl->next = hl->prev = NULL;
#if defined(POSIX) || defined (_POSIX_SOURCE)
    memmove(&hl_buf, packet->data + (n * 2 * sizeof(netint)), sizeof(netint));
#else /* defined(POSIX) || defined (_POSIX_SOURCE) */
    bcopy(packet->data + (n * 2 * sizeof(netint)), &hl_buf, sizeof(netint));
#endif /* defined(POSIX) || defined (_POSIX_SOURCE) */
    hl->start = ntohl(hl_buf) + hl_offset;

#if defined(POSIX) || defined (_POSIX_SOURCE)
    memmove(&hl_buf, packet->data + (n * 2 * sizeof(netint))+sizeof(netint), 
	  sizeof(netint));
#else /* defined(POSIX) || defined (_POSIX_SOURCE) */
    bcopy(packet->data + (n * 2 * sizeof(netint))+sizeof(netint), &hl_buf, 
	  sizeof(netint));
#endif /* defined(POSIX) || defined (_POSIX_SOURCE) */
    hl->end = ntohl(hl_buf) + hl_offset;
    XmTextSetHighlight(widgets[k_text], hl->start, hl->end, 
		       XmHIGHLIGHT_SELECTED);
    
    if (!hl_head) {
      hl_head = hl_prev = hl;
    }
    else {
      if (!hl_prev)
	hl_prev = hl_tail;
      hl_prev->next = hl;
      hl->prev = hl_prev;
      hl_prev = hl;
    }
  }
  if (hl)
    hl_tail = hl;
}

/* Free up memory from old hilite list, if any */
free_hl_list(void)
{
  hl_rec *cur, *prev;

  if (!hl_head)
    return;
  cur = hl_head;
  while (cur) {
    prev = cur;
    cur = cur->next;
    XtFree(prev);
  }
  hl_head = hl_tail = NULL;
}

void stats_dcb(packet_rec *packet)
{
  status = SUMMARIZING;
  if (packet->data_len > 0)
      init_summary(packet->var.stats.num_results, packet->data);
  else
      init_summary(packet->var.stats.num_results, "");

  if (packet->var.stats.num_results > MAX_SUMMARIES) {
    char msg[1024];

    sprintf(msg, "The first %d records are available for retrieval.", 
	    MAX_SUMMARIES);
    pop_warning(msg, NULL, search_shell);
  } 

  if (packet->var.stats.num_results == 0) {
    clear_globe(k_search_globe_box);
    pop_warning("No documents found.  Check spelling, or open \"Help\" and\nchoose \"Searching\" for help with your search strategy.", NULL, search_shell);
    status = IDLE;
    ready_buttons();
  }
  else
    pop_window(k_summary_titles);
}

void title_dcb(packet_rec *packet)
{
    capture_title(packet->data);
}

void user_req_dcb(packet_rec *packet)
{
  char *ptr;
  char *query_text, *query_tag, *yes_label, *no_label;

  /* Separate out the individual string components */
  ptr = packet->data;

  query_text = strdup(ptr);
  ptr += strlen(ptr) + 1;

  query_tag = strdup(ptr);
  ptr += strlen(ptr) + 1;

  yes_label = strdup(ptr);
  ptr += strlen(ptr) + 1;

  no_label = strdup(ptr);

  pop_user_query((int) packet->var.user_req.type, query_text, query_tag, 
		 yes_label, no_label);
}

void exts_dcb(packet_rec *packet)
{
    capture_ext(packet->data, packet->var.exts.ext_num);
}

/* Set applicable buttons to sensitive because we are in READY state */
ready_buttons(void)
{
  XtSetSensitive(widgets[k_search], TRUE);
  if (opac_mode)
    XtSetSensitive(widgets[k_opac_search], TRUE);
  XtSetSensitive(widgets[k_db_btn], TRUE);
}

/* Set applicable buttons to insensitive because we left READY state */
not_ready_buttons(void)
{
  XtSetSensitive(widgets[k_search], FALSE);
  if (opac_mode)
    XtSetSensitive(widgets[k_opac_search], FALSE);
  XtSetSensitive(widgets[k_db_btn], FALSE);
}

/* quit the program via a call back, triggered by menu-item */
void quit_cb(Widget w, int *tag, u_long *reason)
{
  if (status != UNCONNECTED)
    disconnect();
  exit(0);
}

/* Emergency X error handler */
int quit_handler(void)
{
  quit_cb(NULL, NULL, NULL);
}

/* Set the options menu back to default position */
void reset_field_labels(void)
{
  int    i;
  int n;
  Arg args[10];

  if (!cur_db)
    return;
  for (i = 0; i < NUM_TEXT_FIELDS; i++){
    n = 0;
    XtSetArg(args[n], XmNmenuHistory, 
	     cur_db->dflt_btns[i]);n++;
    XtSetValues(widgets[i + k_label1], args, n);
    setup_range_cb(cur_db->dflt_btns[i], i, NULL);
  }

  /* Fix up the browse button */
  n = k_field1 + last_focused_field;
  field_focus_cb(widgets[n], &n, NULL);
}


/* Clear the search screen. Depends on adjacency of #defines in 
   willow.h
 */
void clear_cb(Widget w, int *tag, XmAnyCallbackStruct *reason)
{
  int    i;
  Widget button;

  range_cancel_cb(NULL, NULL, NULL);
  for (i = 0; i < NUM_TEXT_FIELDS; i++) {
    XmTextSetString(widgets[i + k_field1], "");
    XtVaGetValues(widgets[k_label1 + i], XmNmenuHistory, &button, NULL);
    setup_range_cb(button, i, NULL);
  }

  for (i = k_bool1; i <= k_bool4; i++) 
    XmToggleButtonSetState(widgets[i], FALSE, FALSE);

  if (reason)
    if ((reason->event->xbutton.state & ControlMask) ||
	(reason->event->xbutton.state & ShiftMask))
      reset_field_labels();
}


/* Called when the search button is pressed. Pull together all needed
   strings, and pass to send_search_cmd().
 */
void search_cb(Widget w, int *tag, u_long *reason)
{
  char *text;		/* text returned from widget */
  int   i;
  char *command;	/* field specifier command */
  Widget button;	/* option-menu button for current text field */
  db_field_desc_rec *field_desc;  /* DB field descriptor */
  char  data[MAX_LINE * NUM_TEXT_FIELDS * 2]; /* data area for search packet */
  char  tmp_data[MAX_LINE * NUM_TEXT_FIELDS * 2];
  int   data_ptr = 0, tmp_ptr = 0;
  int   text_cnt = 0;
  int   limit_cnt = 0;
  Boolean is_text_term = FALSE; /* Make sure we have at least one text-term */

  if (!XtIsSensitive(widgets[k_search]))
    return;
  if (status == SUMMARIZING) 
    status = IDLE;
  if (status != IDLE) {
    return;
  }

  for (i = 0; i < NUM_TEXT_FIELDS; i++){
    /* Find the current field data structure */
    XtVaGetValues(widgets[k_label1 + i], XmNmenuHistory, &button, NULL);
    XtVaGetValues(button, XmNuserData, &field_desc, NULL);

    /* Check if range or type-in field, and get text */
    if (field_desc->range_type)
      text = range_to_text(i);
    else
      text = XmTextGetString(widgets[k_field1 + i]);

    if (*text) {
      /* Store the text from current search line into the data area */
      if (field_desc->is_limit) {
	limit_cnt++;
	strcpy(tmp_data + tmp_ptr, str_clean(text));
	tmp_ptr += strlen(text) + 1;
      }
      else {
	/* Save type into first byte */
	if (!(data[data_ptr++] = (char) field_desc->range_type))
	  is_text_term = TRUE;

	text_cnt++;
	strcpy(data+data_ptr, str_clean(text));
	data_ptr += strlen(text) + 1;

	/* Store the command string after the text (preserving \0) */
	command = field_desc->command;
	if (!command) 
	  command = "";
	strcpy(data+data_ptr, command);
	data_ptr += strlen(command) + 1;
      }
      XtFree(text);
    }
  }
  
  /* If we had any text-field limits, put them into the limit area */
  if (limit_cnt) {
#if defined(POSIX) || defined (_POSIX_SOURCE)
    memmove(data+data_ptr, tmp_data, tmp_ptr);
#else /* defined(POSIX) || defined (_POSIX_SOURCE) */
    bcopy(tmp_data, data+data_ptr, tmp_ptr);
#endif /* defined(POSIX) || defined (_POSIX_SOURCE) */
    data_ptr += tmp_ptr;
  }

  /* Retrieve any set boolean fields, and save into data area  */
  for (i = 0; i <  cur_db->num_limit_fields; i++){
    if (XmToggleButtonGetState(widgets[k_bool1 + i])) {
      limit_cnt++;
      /* find this button in database structure */ 
      XtVaGetValues(widgets[k_bool1 + i], XmNuserData, &field_desc, NULL);

      /* Store the command string in the data area */
      command = field_desc->command;
      strcpy(data+data_ptr, command);
      data_ptr += strlen(command) + 1;
    }  
  }
  
  if (is_text_term) {
    /* Note that data_ptr is now equal to total length */
    send_search_cmd(text_cnt, limit_cnt, data, data_ptr);
    start_display();
    clear_titles();
    pop_globe(k_search_globe_box);
    not_ready_buttons();
    status = SEARCHING;
    make_query_display();
  }
  else
    pop_warning("You must fill in at least one text field before searching",
		NULL, search_shell);
}

/* Called when user hits the search cancel button, on search globe or
   on full-rec globe */
void search_cancel_cb(Widget w, int *tag, u_long *reason)
{
  if (w == widgets[k_search_cancel]) {
    if (status == SEARCHING) {
      send_search_cancel_cmd();
      XtSetSensitive(widgets[k_search_cancel], FALSE);
      halt_globe();
    }
    else
      halt_globe();
  }
  else {
    XtSetSensitive(widgets[k_full_cancel], FALSE);
    halt_globe();
  }
}

/* This callback really opens the results capture file. If the file
   exists, append to it, if not, create it. If called from Save button,
   do the save too.
 */
void really_open_save_file_cb(Widget w, int *tag, XmFileSelectionBoxCallbackStruct *reason)
{
  /* Sometimes Xt calls us by mistake (because of XmNokCallback) */
  if (save_type != SAVE_LOCAL)
    return;

  show_search = XmToggleButtonGetState(widgets[k_show_search_btn]);

  clear_dialog_cb(w, widgets[k_save_box], (u_long *) reason);

  if (capture_file) {
    fclose(capture_file);
    XtFree(capture_name);
    capture_file = NULL;
  }

  capture_name = xmstring_to_string(reason->value);
  capture_file = fopen(capture_name,"a");
  if (!capture_file) {
    pop_warning("Unable to open %s", capture_name, 
		popped_full ? results_shell : summary_shell );
    XtFree(capture_name);
    capture_name = NULL;
  }
  else if (popped_full)
    results_full_save();
  else
    results_summ_save();
}


/* save the current result */
void results_full_save(void)
{
  char *result;

  if (!text_len)
    return;

  if (!capture_file)
    return;
    
  pop_working("Saving text to disk...", results_shell);
  result = XmTextGetString(widgets[k_text]);
  fprintf(capture_file, "%s", get_query_display_str());
  fprintf(capture_file, "%s\n", result);
  fflush(capture_file);
  XtFree(result);
  
  /* Clear the dialog after 1 second, so there is time to see it */
  XtAppAddTimeOut(app_context, 1000, clear_working, NULL);
}

/* Convert the typed-in password text to stars */
void starrify_cb(Widget w, char **pw, XmTextVerifyPtr reason)
{
  int i;

  /* First put the real text into the saved password.
     If this is a programmatic change, assume we should be over-writing
     the whole thing. For some reason we get a value of 0 in endPos */

  splice(pw, reason->startPos, 
	 reason->event ? reason->endPos : strlen(*pw),
	 reason->text->ptr, reason->text->length);

  /* Put *'s into text to be added to window */
  for (i = 0; i < reason->text->length; i++)
    reason->text->ptr[i] = PW_ECHO_CHAR;
}

/* This is necessary for motif 1.1 -- we have to use XmText because
   XmTextField does not have focusCallback, but XmText allows paste
   of carriage return. */
void no_cr_cb(Widget w, int *tag, XmTextVerifyPtr reason)
{
  int i;

  for (i = 0; i < reason->text->length; i++)
    if (reason->text->ptr[i] == '\n')
      reason->text->ptr[i] = ' ';
}

/* Record the current state of the all the options to ~/.willowrc */
void save_options_cb(Widget w, int *tag, XmSelectionBoxCallbackStruct *reason)
{
  struct passwd *pw;   /* password file entry */
  char   file_name[MAX_LINE];
  FILE  *outfile;
  char  *login;
  db_desc_rec *db;
  user_info_rec *info_ptr;
  int    i;

  login = getlogin();
  if (login == NULL)
    pw = getpwuid(getuid());
  else if (strlen(login) == 0)
    pw = getpwuid(getuid());
  else
    pw = getpwnam(login);

  strcpy(file_name, pw->pw_dir);
  strcat(file_name, OPTIONS_FILE_NAME);
  
  if (!(outfile = fopen(file_name,"w"))) {
    pop_warning("Unable to open %s", file_name, search_shell);
    return;
  }
  pop_working("Saving options to ~/.willowrc", search_shell);

  fprintf(outfile, "%d\n", OPTION_FILE_COOKIE); 
  fprintf(outfile, "%s\n", DONT_EDIT_WARNING);

  fprintf(outfile, "%s %c %d\n", ECHO_HOST_S, CONFIG_SEP_CHAR,
	  XmToggleButtonGetState(widgets[k_echo]));
  if (save_type == SAVE_FTP)
    fprintf(outfile, "%s ", USE_FTP_S);
  else if (save_type == SAVE_EMAIL)
    fprintf(outfile, "%s ", USE_EMAIL_S);
  else
    fprintf(outfile, "%s ", USE_LOCAL_S);
  fprintf(outfile, "%c %d\n", CONFIG_SEP_CHAR, TRUE);

  if (ftp_host)
    fprintf(outfile, "%s %c %s\n", FTP_HOST_S, CONFIG_SEP_CHAR, ftp_host);
  if (ftp_uid)
    fprintf(outfile, "%s %c %s\n", FTP_UID_S,  CONFIG_SEP_CHAR, ftp_uid);
  fprintf(outfile, "%s %c %d\n", FTP_APPEND_S,  CONFIG_SEP_CHAR, ftp_append);
  if (mail_addr)
    fprintf(outfile, "%s %c %s\n", MAIL_ADDR_S,  CONFIG_SEP_CHAR, mail_addr);
  if (mail_subj)
    fprintf(outfile, "%s %c %s\n", MAIL_SUBJ_S,  CONFIG_SEP_CHAR, mail_subj);
  if (XmToggleButtonGetState(widgets[k_show_search_btn]))
    fprintf(outfile, "%s %c %d\n", INCLUDE_QUERY_S,  CONFIG_SEP_CHAR, TRUE);

  fprintf(outfile, "%s %c %d\n", ABBR_FIELDS_S, CONFIG_SEP_CHAR, abbrev_fields);
  fprintf(outfile, "%s %c %s\n", PRINT_CMD_S, CONFIG_SEP_CHAR, print_cmd);

  /* Save any user-info records that have been created. Loop through
     all of our databases, then loop through the user_info list for each.
     Save the tag and response for each TEXT_REQ (we do not saved 
     the MASKED_REQ, because they are presumably secure).

     While looping through the databases, also save any full-record
     field customizations.
   */
  for (db = (db_desc_rec *) hash_FirstEntry(db_hash_tab);
       db; db = (db_desc_rec *) hash_NextEntry(db_hash_tab)) {
    for (info_ptr = db->user_info; info_ptr; info_ptr = info_ptr->next) {
      if (info_ptr->type == TEXT_REQ) {
	fprintf(outfile, "%s %c%s%c%s%c%s\n", USER_INFO_S, CONFIG_SEP_CHAR,
		db->tag, SUB_SEP_CHAR, info_ptr->tag, SUB_SEP_CHAR,
		info_ptr->response);
      }
    }
    if (db->display_fields && !db->all_default_fields) {
      fprintf (outfile, "%s %c%s%c", CUSTOM_FIELDS_S, CONFIG_SEP_CHAR,
	       db->tag, SUB_SEP_CHAR);
      for (i = 0; db->display_fields[i]; i++)
	fprintf(outfile, "%s%c", db->display_fields[i], FIELD_PICK_SEP);
      fprintf(outfile,"\n");
    }
  }

  /* Clear the dialog after 2 seconds, so there is time to see it */
  XtAppAddTimeOut(app_context, 2000, clear_working, NULL);
  fclose(outfile);
}

/* Read the options from ~/.willowrc, if it exists */
read_options(void)
{
  struct passwd *pw;   /* password file entry */
  char   file_name[MAX_LINE],
         line_buf[MAX_LINE];
  FILE  *infile;
  int    field;               /* Token # for current field */
  char   value[MAX_LINE];     /* Value for current field */
  int    cookie;

  pw = getpwuid(getuid());
  
  /* Need to retrieve all the widgets in the save and print dialogs */
  pop_save_box(NULL);
  q_pop_save_box(NULL);
  print_opt_cb(NULL, NULL, NULL);
  pop_retrieve_dlg_cb(NULL, NULL, NULL);

  strcpy(file_name, pw->pw_dir);
  strcat(file_name, OPTIONS_FILE_NAME);

  if (!(infile = fopen(file_name,"r"))) {
    return;
  }

  config_line_num = 0;
  config_file_name = file_name;

  /* Check that first line of file contains 'magic cookie' */
  get_line(line_buf, infile);
  sscanf(line_buf, "%d", &cookie);
  if (cookie != OPTION_FILE_COOKIE) {
    pop_warning("%s appears to be invalid or wrong version.", file_name, search_shell);
    return;
  }
  
  /* Loop through file */
  while ((field = scan_config(infile, value)) != EOF) {
    switch (field) {
    case ECHO_HOST:
      initial_echo = atoi(value);
      break;
    case USE_FTP:
      if (atoi(value))
	save_type = SAVE_FTP;
      break;
    case USE_EMAIL:
      if (atoi(value))
	save_type = SAVE_EMAIL;
      break;
    case USE_LOCAL:
      if (atoi(value))
	save_type = SAVE_LOCAL;
      break;
    case INCLUDE_QUERY:
      show_search = atoi(value);
      break;
    case FTP_HOST:
      ftp_host = strdup(value);
      break;
    case FTP_UID:
      ftp_uid = strdup(value);
      break;
    case FTP_APPEND:
      ftp_append = atoi(value);
      break;
    case MAIL_ADDR:
      mail_addr = strdup(value);
      break;
    case MAIL_SUBJ:
      mail_subj = strdup(value);
      break;
    case PRINT_CMD:
      print_cmd = strdup(value);
      break;
    case ABBR_FIELDS:
      abbrev_fields = atoi(value);
      break;
    case USER_INFO:
      if (parse_user_info(value) < 0)
	pop_warning("%s is malformed.", file_name, search_shell);
      break;
    case CUSTOM_FIELDS:
      if (parse_field_pick_str(value) < 0)
	pop_warning("%s is malformed.", file_name, search_shell);
      break;
    default:
      pop_warning("%s is malformed.", file_name, search_shell);
      return;
    }
  }
  fclose(infile);
}

/* Put a user-info string into the user_info list for a database. String
   is of form database-tag:query-tag:value. Can be found either in 
   .willowrc or X-resource.
 */
int parse_user_info(char *str)
{
  char *db_tag, *query_tag, *value;
  db_desc_rec *db;
  user_info_rec *info_ptr;

  db_tag = str;
  if (!(query_tag = strchr(db_tag, SUB_SEP_CHAR)))
    return -1;
  *query_tag = 0;
  query_tag++;

  if (!(value = strchr(query_tag, SUB_SEP_CHAR)))
    return -1;

  *value = 0;
  value++;

  db = (db_desc_rec *) hash_Lookup(db_hash_tab,
				   hash_HashFunction(db_tag, strlen(db_tag)),
				   db_compare, db_tag);
  /* If the referenced database does not exist now, no big deal */
  if (!db)
    return (0);

  for (info_ptr = db->user_info; info_ptr; info_ptr = info_ptr->next) {
    if (strcasecmp(info_ptr->tag, query_tag) == 0)
      break;
  }
  if (info_ptr) {
    XtFree(info_ptr->response);
    info_ptr->response = strdup(value);
  }
  else {
    info_ptr = (user_info_rec *) XtMalloc(sizeof(user_info_rec));
    info_ptr->tag = strdup(query_tag);
    info_ptr->type = TEXT_REQ;
    info_ptr->response = strdup(value);
    info_ptr->next = db->user_info;
    db->user_info = info_ptr;
  }
  return 0;
}

/* Parse the userInfo resource. Could not do this at resource reading
   time, because db.conf had not been read yet.
 */
void parse_user_info_list(void)
{
  char *p, *idex;
  int len, total_len;

  p = user_info_list;
  total_len = strlen(p);
  while (p < (user_info_list + total_len)) {
    if (idex = strchr(p, '\n'))
      *idex = 0;
    len = strlen(p);
    parse_user_info(p);
    p += len + 1;
  }
}

pop_window(int tag)
{
  int dummy;

  dummy = tag;
  window_to_top_cb(NULL, &dummy, NULL);
}

/* Put a specified window on top, de-iconizing if need be. Take 
   a pointer to tag, or tag itself */
void window_to_top_cb(Widget w, int *tag, u_long *reason)
{
  Widget shell;
  XWindowAttributes xwa;

  if (*tag ==  k_dict_form) {
    shell = dict_shell;
    XmProcessTraversal(widgets[k_dict_search], XmTRAVERSE_CURRENT);
    XmProcessTraversal(widgets[k_dict_search], XmTRAVERSE_CURRENT);
  }
  else if (*tag == k_summary_titles){
    shell = summary_shell;
  }
  else if (*tag == k_results){
    shell = results_shell;
  }
  else if (*tag == k_search){
    shell = search_shell;
  }
  else
    error("Bad tag passed to window_to_top_cb()");

  if (!shell)
    return;

  if (interface_mode == MANAGED_MODE) {
    XGetWindowAttributes(XtDisplay(shell), XtWindow(shell), &xwa);
    if (xwa.map_state == IsViewable)
      XtUnmapWidget(shell);
  }
  XMapRaised(XtDisplay(toplevel), XtWindow(shell));
}

/* Unmap all of the windows. Note that this won't unmap any icons, but
   this function is only used in a session manager environment that does
   not use icons anyway.
 */
hide_all_windows(void)
{
  hide_window(k_summary_titles);
  hide_window(k_results);
  hide_window(k_search);

  dict_hide();
  clear_help();
}

void hide_window(int tag) 
{
  hide_window_cb(NULL, &tag, NULL);
}

/* Turn off the iconic resource, otherwise screws up hide/restore */
turn_off_iconic(Widget shell)
{
  Arg args[10];
  int n = 0;

  XtSetArg(args[n], XmNiconic, FALSE);n++;
  XtSetArg(args[n], XmNinitialState, NormalState);n++;
  XtSetValues(shell, args, n);
}

/* Hide a window (not even an icon left) */
void hide_window_cb(Widget w, int *tag, u_long *reason)
{
  Widget shell;

  if (*tag ==  k_dict_form) {
    shell = dict_shell;
  }
  else if (*tag == k_summary_titles){
    clear_globe(k_full_globe_box);
    shell = summary_shell;
  }
  else if (*tag == k_results){
    shell = results_shell;
  }
  else if (*tag == k_search){
    shell = search_shell;
  }
  else
    error("Bad tag passed to hide_window_cb()");
  
  if (shell) {
    turn_off_iconic(shell);
    XtUnmapWidget(shell);
  }
}

/*
 * Miscellaneous Junk
 */

/* initialization */
init_application(void)
{
  int  i;

  stacked_select[0] = 0;
  ftp_pw = XtMalloc(1);
  ftp_pw[0] = 0;
  masked_response = XtMalloc(1);
  masked_response[0] = 0;

  for (i=0; i<NUM_WIDGETS; i++)
    widgets[i] = NULL;

  signal(SIGCHLD, wait_child);
  signal(SIGPIPE, SIG_IGN);
}

/* error handling function */
error(char *msg)
{
  int n;
  Arg args[20];
  Widget  dlg = NULL, button;
  char buffer[2048];

  fprintf(stderr, "\n\nFATAL ERROR IN WILLOW: %s\n",msg);

  if (!s_MrmHierarchy)
    exit(0);
  if (MrmFetchWidget(s_MrmHierarchy, 
		     "message_box", 
		     search_shell,
		     &dlg,
		     &dummy_class) != MrmSUCCESS)
    exit(0);

  sprintf(buffer, "FATAL ERROR IN WILLOW:\n%s", msg);
  n = 0;
  XtSetArg(args[n], XmNmessageString, 
	   XmStringCreateLtoR(buffer, XmSTRING_DEFAULT_CHARSET));n++;
  XtSetArg(args[n], XmNdialogType, XmDIALOG_ERROR);n++;
  XtSetArg(args[n], XmNdialogStyle, XmDIALOG_FULL_APPLICATION_MODAL);n++;
  XtSetValues(dlg, args, n);

  /* un-manage buttons we don't want */    
  button = XmMessageBoxGetChild(dlg, XmDIALOG_CANCEL_BUTTON);
  XtUnmanageChild(button);
  button = XmMessageBoxGetChild(dlg, XmDIALOG_HELP_BUTTON);
  XtUnmanageChild(button);

  /* Add quit callback */
  button = XmMessageBoxGetChild(dlg, XmDIALOG_OK_BUTTON);
  XtAddCallback(button, XmNactivateCallback, quit_cb, NULL);

  map_dialog(dlg);
  ForceDialog(dlg);

  /* Go into infinite event loop, because caller expects us not to return, 
     but we need events so user can acknowledge dialog */
  while (TRUE)
    XtAppProcessEvent(app_context, XtIMAlternateInput | XtIMTimer | 
		      XtIMXEvent);
}

/* Prepare to start displaying a new result */
start_display(void)
{
  /* Clear the results window */
  XmTextSetString(widgets[k_text], "");
  text_len = 0;
  XtSetSensitive(widgets[k_full_print], FALSE);
  XtSetSensitive(widgets[k_full_save_btn], FALSE);
  XtSetSensitive(widgets[k_match_next], FALSE);
  XtSetSensitive(widgets[k_match_prev], FALSE);

  /* Close the window too */
  hide_window(k_results);
}

/* Append the passed in buffer to the view window text. The 
   global variable text_len is tracking how long the text is */
append_to_view(char *buffer)
                       /* The characters (null terminated) */
{
  XmTextReplace(widgets[k_text], text_len, text_len, buffer);
  text_len += strlen(buffer);
}

/* Need this out here, so dbio.c can be X-free */
int echo_on(void)
{
 return XmToggleButtonGetState(widgets[k_echo]);
}

/* Center the status frame */
void center_status(Widget w, XtPointer data, XEvent *event, Boolean *cont)
{
  Dimension frame_width, btn_width, form_width;
  int       offset, btn_offset;

  XtVaGetValues(widgets[k_status_frame], XmNwidth, &frame_width, NULL);
  if (XtIsManaged(widgets[k_opacify]))
    XtVaGetValues(widgets[k_opacify], XmNwidth, &btn_width, 
		  XmNleftOffset, &btn_offset, NULL);
  else
    btn_offset = btn_width = 0;
  XtVaGetValues(search_shell, XmNwidth, &form_width, NULL);

  offset = (form_width - btn_width - frame_width - btn_offset)/2;
  XtVaSetValues(widgets[k_status_frame], XmNleftOffset, 
		offset > 0 ? offset : 0, NULL);
}


/* Execute a button command via a control key. We expect
   two parameters, the name of the shell, and the name of the
   command. See config_interface() */
void ctrl_cmd_action(Widget w, XEvent *event, 
		     String *params, Cardinal *num_params)
{
  if (*num_params != 2)
    return;

  /* Pop up the fields menu. This is the tricky one */
#ifdef ndef
  if (strcmp(params[1], "fields") == 0) {
    XEvent dummy;
    for (n = 0; n < NUM_TEXT_FIELDS; n++)
      if (widgets[k_field1 + n] == w)
	break;

    XtCallActionProc(widgets[k_label1 + n], "ManagerGadgetSelect", &dummy, 
		     NULL, NULL);
  }
#endif
  if (strcmp(params[1], "browse") == 0) {
    if (XtIsSensitive(widgets[k_list_btn]) && !opac_mode)
      XtCallCallbacks(widgets[k_list_btn], XmNactivateCallback, NULL);
  }
  else if (strcmp(params[1], "clear") == 0) {
    if (!opac_mode)
      clear_cb(NULL, NULL, NULL);
  }
  else if (strcmp(params[1], "clear_all") == 0) {
    if (!opac_mode) {
      clear_cb(NULL, NULL, NULL);
      reset_field_labels();
    }
  }
  else if (strcmp(params[1], "help") == 0) {
    if (strcmp(params[0], "search") == 0)
      help_cb(search_shell, NULL, NULL);
    else if (strcmp(params[0], "basic") == 0)
      help_cb(search_shell, NULL, NULL);
    else if (strcmp(params[0], "summ") == 0)
      help_cb(summary_shell, NULL, NULL);
    else if (strcmp(params[0], "results") == 0)
      help_cb(results_shell, NULL, NULL);
    else if (strcmp(params[0], "browse") == 0)
      help_cb(dict_shell, NULL, NULL);
  }
  else if (strcmp(params[1], "more") == 0) {
    if (XtIsSensitive(widgets[k_summ_next_block]))
      summ_cont_cb(NULL, NULL, NULL);
  }
  else if (strcmp(params[1], "close") == 0){
    if (strcmp(params[0], "summ") == 0)
      hide_window(k_summary_titles);
    else if (strcmp(params[0], "results") == 0)
      hide_window(k_results);
    else if (strcmp(params[0], "browse") == 0)
      dict_cancel_cb(NULL, NULL, NULL);
  }
  else if (strcmp(params[1], "print") == 0){
    if (strcmp(params[0], "summ") == 0) {
      if (XtIsSensitive(widgets[k_summ_print]))
	print_summ_cb(NULL, NULL, NULL);
    }
    else if (strcmp(params[0], "results") == 0) {
      if (XtIsSensitive(widgets[k_full_print]))
	print_full_cb(NULL, NULL, NULL);
    }
  }
  else if (strcmp(params[1], "save") == 0){
    if (strcmp(params[0], "summ") == 0) {
      if (XtIsSensitive(widgets[k_summ_save_btn]))
	open_save_cb(widgets[k_summ_save_btn], NULL, NULL);
    }
    else if (strcmp(params[0], "results") == 0) {
      if (XtIsSensitive(widgets[k_full_save_btn]))
	open_save_cb(widgets[k_full_save_btn], NULL, NULL);
    }
  }
  else if (strcmp(params[1], "next") == 0)
    match_next_cb(NULL, NULL, NULL);
  else if (strcmp(params[1], "prev") == 0)
    match_prev_cb(NULL, NULL, NULL);
}
