/***********************************************************************
 *                                                                     *
 *              WILLOW(tm) Information Retrieval Software              *
 *     (Washington Information Looker-upper Layered Over Windows)      *
 *             Copyright 1992-1994 University of Washington            *
 *                                                                     *
 *   Willow is a trademark of the University of Washington.  No        *
 *   commercial use of this trademark or software may be made without  *
 *   prior written permission of the University of Washington.         *
 *                                                                     *
 * Permission to use, copy, modify, and distribute Willow(tm) software *
 * and its documentation for any noncommercial purpose is hereby       *
 * granted, provided that the above copyright notice appears in all    *
 * copies and that both the above copyright notice and this permission *
 * notice appear in supporting documentation, and that the name of the *
 * University of Washington not be used in advertising or publicity    *
 * pertaining to distribution of the software without specific,        *
 * written prior permission.  This software is made available "as is", *
 * and THE UNIVERSITY OF WASHINGTON DISCLAIMS ALL WARRANTIES, EXPRESS  *
 * OR IMPLIED, WITH REGARD TO THIS SOFTWARE, INCLUDING WITHOUT         *
 * LIMITATION ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS    *
 * FOR A PARTICULAR PURPOSE, AND IN NO EVENT SHALL THE UNIVERSITY OF   *
 * WASHINGTON BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL     *
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA  *
 * OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, TORT (INCLUDING       *
 * NEGLIGENCE) OR STRICT LIABILITY, ARISING OUT OF OR IN CONNECTION    *
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.                       *
 *                                                                     *
 * For further information please contact:                             *
 * Willow Project                                                      *
 * Information Systems                                                 *
 * University of Washington                                            *
 * 4545 15th NE, 1st floor                                             *
 * Seattle, Washington, 98105, USA                                     *
 * Internet: willow@cac.washington.edu                                 *
 ***********************************************************************/

/***************************************************
 * willow.h
 * This is the main header file for willow
 ***************************************************/

#ifndef MAX
#define MAX(a,b) ((a)>(b)?(a):(b))
#endif
#define MAX_LINE 240  /* This is defined again in db.h */
#define MAX_TITLE_CHARS 100
#define CHUNK_SIZE 2048
#define SUMM_CHUNK      30
#define MAX_SUMMARIES   10000
#define LIST_TIMEOUT    10
#define DISCON_TIMEOUT  20000
#define CONNECT_TIMEOUT 30000
#define NUM_GLOBES      30
#define SEARCH_FILE_COOKIE 345672
#define OPTION_FILE_COOKIE 9271993
#define SHADOW_THICKNESS    3
#define SHADOW_THICKNESS_S "3"
#define OPTIONS_FILE_NAME "/.willowrc"
#define DICT_DIR          "dict/"
#define DRIVER_DIR        "drivers/"
#define BITMAP_DIR        "xbm/"
#define HELP_DIR          "help/"
#define SEARCH_HELP       "search.txt"
#define BASIC_HELP        "basic.txt"
#define SUMMARY_HELP      "summaries.txt"
#define FULL_REC_HELP     "full-record.txt"
#define BROWSE_HELP       "browse.txt"
#define DONT_EDIT_WARNING "### WILLOW PRIVATE FILE, DO NOT EDIT ###"
#define OPEN_DEST_STR    "Save Destination..."
#define ABOUT_NAME       "about"
#define PW_ECHO_CHAR '*'
#define Q_LOCAL_OPEN_STR  "Retrieve Query Set from Local File"
#define Q_FTP_OPEN_STR    "Retrieve via File Transfer Protocol"
#define Q_LOCAL_SAVE_STR  "Save Query Set to Local File"
#define Q_FTP_SAVE_STR    "Save via File Transfer Protocol"
#define NO_DB_INFO_STR    "No database information currently available"

#define TITLE_FORM         "%s%s\n"
#define GLOBE_FILE_FMT     "%s%sglobe%d.xbm"
#define SCALE_FMT          "%d out of %d"
#define MBROWSE_FMT        "Browse %s by"
#define FULL_GLOBE_MSG_FMT "Retrieving Records\n%d out of %d"
#define LIST_BTN_FMT       "Browse %s List"
#define DFLT_LIST_BTN      "Browse <NO LIST>"
#define DE_OPACIFY_LABEL   "Go To Advanced Searching"
#define OPACIFY_LABEL      "Go To Basic Searching"
#define LOW_DATE_LABEL     "Oldest"
#define HIGH_DATE_LABEL    "Newest"
#define LOW_NUMERIC_LABEL  "Lowest"
#define HIGH_NUMERIC_LABEL "Highest"
#define MAX_NUMERIC_DIGITS  12

#define FULL_SEARCH_TITLE  "Advanced Searching"
#define OPAC_SEARCH_TITLE  "Basic Searching"
#define SUMMARY_TITLE      "Summary List"
#define FULL_REC_TITLE     "Record Retrieval"
#define RETRIEVE_LABEL     "Retrieve %d Records"
#define RETRIEVE_LABEL1    "Retrieve 1 Record"

#define YEAR_LABEL        "Year"
#define DEFAULT_PRINT_CMD "lpr"
#define DEFAULT_MAIL_CMD  "mail -s \"%s\" %s"
#define DEFAULT_MAIL_SUBJ "Willow Search Results"
#define NO_PRINT_CMD      "none"

#define DEFAULT_RANGE_CENTURY 1900

/* Types for client message protocol */
#define WILLOW_ATOM   "Willow"
#define RESET_CM      0
#define MAP_CM        1
#define UNMAP_CM      2
#define SELECT_DB_CM  3
#define QUIT_CM       4

/* Bit masks for dialog buttons */
#define OK_BUTTON_MASK  1
#define CANCEL_BUTTON_MASK 2
#define HELP_BUTTON_MASK 4

/* Possible values for save_type */
#define SAVE_LOCAL      1
#define SAVE_FTP        2
#define SAVE_EMAIL      3

/* Possible values for interface_mode */
#define NORMAL_MODE     0
#define ANONYMOUS_MODE  1
#define MANAGED_MODE    2

/* Possible values for connection status */
#define UNCONNECTED     0
#define IDLE            1
#define CONNECTING      2
#define SUMMARIZING     3
#define SEARCHING       4

/* Possible values for "which" parameter in do_ftp() */
#define SUMM_FLAG       0
#define FULL_FLAG       1
#define QUERY_FLAG      2

#define UNCONNECTED_S   "UNCONNECTED: Choose from Database menu to connect"
#define IDLE_S          "CONNECTED to: "
#define CONNECTING_S    "CONNECTING to: "
#define NO_CAPTURE_S    "<UNSPECIFIED>"
#define USING_MAIL_S    "Via E-Mail"
#define USING_FTP_S     "Via File Transfer"
#define USING_LOCAL_S   "To File"
#define FULL_TMP_FILE   "/usr/tmp/WillowFullFTP."
#define SUMM_TMP_FILE   "/usr/tmp/WillowSummFTP."
#define QUERY_TMP_FILE  "/usr/tmp/WillowQueryFTP."
#define FTP_ARGS     "-in"

/* Define some constant widget tags shared by willow.c and willow.uil */
#define k_retrieve_dlg          0
#define k_results               1
#define k_text                  2
#define k_cheat_btn             3
#define k_message               4
#define k_db_btn                5
#define k_num_range             6
#define k_search                7
#define k_field1                8
#define k_field2                9
#define k_field3               10
#define k_field4               11
#define k_field5               12
#define k_field6               13
#define k_label1               14
#define k_label2               15
#define k_label3               16
#define k_label4               17
#define k_label5               18
#define k_label6               19
#define k_summ_disp            20
#define k_summ_next            21
#define k_summ_prev            22
#define k_searches_pdm         23
#define k_add_search           24
#define k_del_search           25
#define k_echo                 26
#define k_local_file_box       27
#define k_print_opt_btn        28
#define k_full_save_btn        29
#define k_bool1                30
#define k_bool2                31
#define k_bool3                32
#define k_bool4                33
#define k_summary_titles       34
#define k_summary_summaries    35
#define k_summary_scale        36
#define k_conn_options_box     37
#define k_conn_db_sep          38
#define k_user_query_box       39
#define k_field_pick_rc        40
#define k_search_globe_box     41
#define k_search_globe_image   42
#define k_db_pdm               43
#define k_status_label         44
#define k_search_form          45
#define k_summary              46
#define k_search_cancel        47
#define k_cheat_box            48
#define k_dict_form            49
#define k_dict_text            50
#define k_dict_search          51
#define k_ftp_options_box      52
#define k_ftp_host_text        53
#define k_ftp_uid_text         54
#define k_ftp_pw_text          55
#define k_ftp_file_text        56
#define k_conn_opt_btn         57
#define k_no_db_btn            58
#define k_summ_save_btn        59
#define k_save_opt_btn         60
#define k_summ_help_btn        61
#define k_summ_print           62
#define k_full_print           63
#define k_print_opt_box        64
#define k_open_search_btn      65
#define k_about_box            66
#define k_local_save_btn       67
#define k_ftp_save_btn         68
#define k_email_save_btn       69
#define k_save_box             70
#define k_email_box            71
#define k_email_text           72
#define k_ftp_ok_btn           73
#define k_email_ok_btn         74
#define k_summ_ok_btn          75
#define k_summ_btn_form        76
#define k_match_next           77
#define k_match_prev           78
#define k_about_image          79
#define k_about_ok             80
#define k_quit_btn             81
#define k_summ_next_block      82
#define k_sum_dum              83
#define k_list_btn             84
#define k_status_frame         85
#define k_range_dlg            86
#define k_limit_range          87
#define k_unlimit_range        88
#define k_month_range          89
#define k_lookup_label         90
#define k_opac_form            91
#define k_opacify              92
#define k_dummy_form           93
#define k_opac_choices         94
#define k_dict_paste           95
#define k_opac_search          96
#define k_dict_cancel          97
#define k_searches_btn         98
#define k_menu_bar             99
#define k_dict_stext          100
#define k_full_cancel         101
#define k_full_globe_image    102
#define k_full_globe_box      103
#define k_search_globe_label  104
#define k_full_globe_label    105
#define k_day_range           106
#define k_range1              107
#define k_range2              108
#define k_range3              109
#define k_range4              110
#define k_range5              111
#define k_range6              112
#define k_from1               113
#define k_from2               114
#define k_from3               115
#define k_from4               116
#define k_from5               117
#define k_from6               118
#define k_to1                 119
#define k_to2                 120
#define k_to3                 121
#define k_to4                 122
#define k_to5                 123
#define k_to6                 124
#define k_range_ctrl          125
#define k_day_label           126
#define k_month_label         127
#define k_year_label          128
#define k_email_subj          129
#define k_match_label         130
#define k_q_fs_box            131
#define k_q_ftp_file_text     132
#define k_q_ftp_host_text     133
#define k_q_ftp_options_box   134
#define k_q_ftp_pw_text       135
#define k_q_ftp_save_btn      136
#define k_q_ftp_uid_text      137
#define k_q_local_save_btn    138
#define k_q_save_box          139
#define k_mbrowse_label       140
#define k_mbrowse_radio       141
#define k_mbrowse_box         142
#define k_short_label         143
#define k_long_label          144
#define k_year_range          145
#define k_range_sep           146
#define k_info_dlg            147
#define k_info_btn            148
#define k_ext_menu            149
#define k_ext_btn             150
#define k_ext_pane            151
#define k_ftp_append_btn      152
#define k_ftp_replace_btn     153
#define k_q_ftp_opt_cancel    154
#define k_ftp_opt_cancel      155
#define k_mail_opt_cancel     156
#define k_show_search_sep     157

#define k_field_pick_form     159
#define k_field_pick_sep      160
#define k_retrieve_sep        161

#define k_field_pick_label    163
#define k_show_search_btn     164
#define NUM_WIDGETS	      175

/* Define a structure for holding geometry of shell */
typedef struct _geom_rec {
  Dimension w, h, x, y;
} geom_rec;
