/*
 *  Copyright (C) 1994 by the Massachusetts Institute of Technology
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice appear in all
 * copies and that both that copyright notice and this permission
 * notice appear in supporting documentation, and that the name of
 * M.I.T. not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior
 * permission.  M.I.T. makes no representations about the suitability
 * of this software for any purpose.  It is provided "as is" without
 * express or implied warranty.
 */

/* Author: Tom Owens MIT Library Systems Office */
/*
 * 06/14/94  slh  R6/date: treat u's like blanks and bars (OCLCFIX)
 * 05/31/94  slh  cleaned out cond'd out stuff
 *    - 06/02/94    fix MESHSubject/R2 rule
 * 05/16/94  slh  added changes for mattf
 * 05/13/94  slh  bugfixes for R3 and R7
 * 05/05/94  slh  fix summary title stuff (FIXSUMTIT)
 *  ... 05/08/94
 * 04/18/94  slh  rule data/params (RULEDATA); R3; various fixes;
 *  ... 04/26/94
 */
#define  OCLCFIX	/* workaround FIX for OCLC non-conformity */
#define  FIXSUMTIT	/* FIX SUMmary TITle stuff */
#define  RULEDATA	/* DATA/params allowed for RULEs */


#include <stdio.h>
#include <ctype.h>
#include "marc.h"   /* marc routines header file */
#include "bib.h"  
#include <pwd.h>
#include <string.h>
#include "msg.h"

#define BLANK_STRING " "
#define BLANK_CHAR ' '
#define TRUE 1
#define FALSE 0
#define SUMM_TAB_LEN 2
#define FULL_TAB_LEN LABEL_LENGTH
#define MAX_LINE 256 
#define FAILED_IND_TEST 1
#define ERR_OUT(m,i)  msg_make(DEFAULT_MSG,m); return i
#define BAD_ALLOC "memory error in GetBrief"
#define BAD_RECORD "could not build record"

#ifdef   FIXSUMTIT
#ifndef  MIN
#define  MIN(a,b)	((a) <= (b) ? (a) : (b))
#define  MAX(a,b)	((a) > (b) ? (a) : (b))
#endif/*!MIN*/
#endif /*FIXSUMTIT*/

#define DEBUG

static BIB *briefHead = NULL,
      *briefTail = NULL,
       *fullHead = NULL,
        *fullTail = NULL,
        *head = NULL,
        *curAtom = NULL;      /* current atom in memory */

static char   inString[80], /*BIB definition line */
       *buildStr = NULL,    /*string currently being built for this atom*/
       *displayStr = NULL, /*string being built for this display */
        *tagPrefixNeeded = NULL, /* flag for moving tag prefix properly */
        *tagData = NULL, /* the position of the current tag */
	summ_tab[SUMM_TAB_LEN+1],
	full_tab[FULL_TAB_LEN+1];

static int     entry, inPosition,tagLength;

static FUNC_POINTER ruleTable[MAX_RULE + 1];

static MARC_RECORD *curMarc; /* the MARC record currently in memory */


/* create the linked list of bib definitions */
int MakeBib (type)
int type;
{
    FILE *fp,*OpenBibFile();
    int docFlag, makeNew, length, returnValue;
    static int madeRuleTable = FALSE;


    head= NULL;
    if (type==FULL)
	fp=OpenBibFile(FULL_FILE);
    else
	fp=OpenBibFile(BRIEF_FILE);

    if (fp==NULL) {
	ERR_OUT("Could not open bibfile",BAD_BIB);
    }
    /* and initialize the tabs */
     memset(summ_tab,BLANK_CHAR,SUMM_TAB_LEN);
     summ_tab[SUMM_TAB_LEN]=NULL;
     memset(full_tab,BLANK_CHAR,FULL_TAB_LEN);
     full_tab[FULL_TAB_LEN]=NULL;
    /*we've opened the file, make the linked list */
     entry =0;
     docFlag=FALSE;
     makeNew=TRUE;
     if (!madeRuleTable) {
	 MakeRuleTable();
	 madeRuleTable++;
     }

     while ((fscanf (fp, "%s", inString)) != EOF) {
	 if (inString[0] == '/')              /*either open or close a comment */
	     docFlag ^= TRUE;
	 else
	     if (!(docFlag)) {
		 /* we need a new atom */
		 if (makeNew) { 
		     entry++;
		     if (NewBib (&head, &curAtom) != GOOD) {
			 ERR_OUT ("no memory!",BAD_BIB);
		     }
		     if ((length = strlen (inString)) > LABEL_LENGTH) {
			 ERR_OUT ("illegal name",BAD_BIB);
		     }
		     InsertBlanks(inString);
		     memset(curAtom->label, BLANK_CHAR,LABEL_LENGTH);
		     curAtom->label[LABEL_LENGTH]=NULL;
		     strncpy(curAtom->label+LABEL_LENGTH-length, inString, length);

#ifdef NEVER
		    InsertBlanks(curAtom->label);
		    /*pad the label with blanks  */
		    while (length<LABEL_LENGTH) {
			strcat (curAtom->label, BLANK_STRING);
			length++;
		    }
#endif

		     makeNew = FALSE;
		 }
		 /*indicators? */
		 else if (curAtom->indOne == NULL)  {
		     returnValue = ParseIndicators(&(curAtom->indOne));
		     if (returnValue != GOOD) {
			 ERR_OUT("could not form indicator one",returnValue);
		     }
		 }
		 else if (curAtom->indTwo == NULL) {
		     returnValue = ParseIndicators(&(curAtom->indTwo));
		     if (returnValue != GOOD) {
			 ERR_OUT("could not form indicator two", returnValue);
		     }
		 }
		 /* have we defined the tag yet? */
		 else if (!curAtom->tag) {
		     returnValue = ParseTag ();
		     if (returnValue != GOOD) {
			 ERR_OUT("could not build tag",returnValue);
		     }
		 }

		 /* if we have label and tag, next comes subfields used */
		 else if (curAtom->sfds == NULL) {
		     returnValue=ParseSfd();
		     if (returnValue != GOOD) {
			 ERR_OUT("could not build subfields",returnValue);
		     }
		 }
		
		 /* and finally, how much of it is used in the summary? */
		 else if (curAtom->charsInSummary == 0) {
		     returnValue = ParseCharsInSummary();
		     if (returnValue != GOOD) {
			 ERR_OUT("could not parse summary info",returnValue);
		     }
		     makeNew = TRUE;
		 }
	     }
     }
#ifdef DUBUG
    /* show the list we created */
     ShowList (head);
#endif DEBUG
     if (type == FULL) {
	 fullHead = head;
	 fullTail = curAtom;
     }
     else {
	 briefHead = head;
	 briefTail = curAtom;
     }
     return GOOD;
}

/* helper function for MakeBibKey:
 Scan src up to end string or newline
 whichever comes first.  If no chars in string,
 return null.  Otherwise copy the resultant string
 into dest.  */
char *fetchtoken(src, dst)
char *src, *dst;
{
    int c;

    while (((c = *src) != '\0') && (c != ',') && (!isspace (c))) {
	*dst++ = c;
	src++;
    }

    *dst = '\0';

    if (*src == ',') src++;
    while (isspace(*src)) src++;

    if (*src == '\0') return NULL;
    else return src;
}

/* create linked list of bib definitions
 from a string passed in that has the same format
 as the bib files */
int MakeBibKey (key, key_head, key_tail)
char *key;
BIB **key_head, **key_tail;	/* value/return */
{
    int docFlag, makeNew, length, returnValue;
    static int madeRuleTable = FALSE;
    char *sptr;

    head= NULL;

    /* and initialize the tabs */
     memset(summ_tab,BLANK_CHAR,SUMM_TAB_LEN);
     summ_tab[SUMM_TAB_LEN]=NULL;
     memset(full_tab,BLANK_CHAR,FULL_TAB_LEN);
     full_tab[FULL_TAB_LEN]=NULL;
    /*we've opened the file, make the linked list */
     entry =0;
     docFlag=FALSE;
     makeNew=TRUE;
     if (!madeRuleTable) {
	 MakeRuleTable();
	 madeRuleTable++;
     }
     sptr = key;

     while ((sptr = fetchtoken(sptr, inString)) != NULL) {
	 if (inString[0] == '/')              /*either open or close a comment */
	     docFlag ^= TRUE;
	 else
	     if (!(docFlag)) {
		 /* we need a new atom */
		 if (makeNew) { 
		     entry++;
		     if (NewBib (&head, &curAtom) != GOOD) {
			 ERR_OUT ("no memory!",BAD_BIB);
		     }
		     if ((length = strlen (inString)) > LABEL_LENGTH) {
			 ERR_OUT ("illegal name",BAD_BIB);
		     }
		     InsertBlanks(inString);
		     memset(curAtom->label, BLANK_CHAR,LABEL_LENGTH);
		     curAtom->label[LABEL_LENGTH]=NULL;
		     strncpy(curAtom->label+LABEL_LENGTH-length, inString, length);

#ifdef NEVER
		    InsertBlanks(curAtom->label);
		    /*pad the label with blanks  */
		    while (length<LABEL_LENGTH) {
			strcat (curAtom->label, BLANK_STRING);
			length++;
		    }
#endif

		     makeNew = FALSE;
		 }
		 /*indicators? */
		 else if (curAtom->indOne == NULL)  {
		     returnValue = ParseIndicators(&(curAtom->indOne));
		     if (returnValue != GOOD) {
			 ERR_OUT("could not form indicator one",returnValue);
		     }
		 }
		 else if (curAtom->indTwo == NULL) {
		     returnValue = ParseIndicators(&(curAtom->indTwo));
		     if (returnValue != GOOD) {
			 ERR_OUT("could not form indicator two", returnValue);
		     }
		 }
		 /* have we defined the tag yet? */
		 else if (!curAtom->tag) {
		     returnValue = ParseTag ();
		     if (returnValue != GOOD) {
			 ERR_OUT("could not build tag",returnValue);
		     }
		 }

		 /* if we have label and tag, next comes subfields used */
		 else if (curAtom->sfds == NULL) {
		     returnValue=ParseSfd();
		     if (returnValue != GOOD) {
			 ERR_OUT("could not build subfields",returnValue);
		     }
		 }
		
		 /* and finally, how much of it is used in the summary? */
		 else if (curAtom->charsInSummary == 0) {
		     returnValue = ParseCharsInSummary();
		     if (returnValue != GOOD) {
			 ERR_OUT("could not parse summary info",returnValue);
		     }
		     makeNew = TRUE;
		 }
	     }
     }
#ifdef DUBUG
    /* show the list we created */
     ShowList (head);
#endif DEBUG
     *key_head = head;
     *key_tail = curAtom;
     return GOOD;
}

/* given a MARC_RECORD structure, and a key
 create a string that contains the marc belonging to
 those keys. */

int GetExtract(marc, outExtract, key)
MARC_RECORD *marc;
char **outExtract;
char *key;
{
    BIB *key_head, *key_tail;
    int returnValue, occurrence;

    if ( (returnValue = MakeBibKey (key, &key_head, &key_tail)) != GOOD) {
	ERR_OUT("Cannot generate Bib Key", BAD_BIB);
    }

   curMarc = marc;

    returnValue = InitStrings();
    if (returnValue != GOOD) {
	ERR_OUT("could not init working strings",returnValue);
    }

    *outExtract = (char *) malloc(sizeof(char));
    if (*outExtract == NULL) {
	ERR_OUT(BAD_ALLOC, MEMORY_ERROR);
    }
    (*outExtract)[0] = NULL;

    curAtom = key_head;
    if (curAtom == NULL) {
	ERR_OUT("bad display list",BAD_FULL);
    }

    while (curAtom != NULL) {
	if  (curAtom->SpecialRule == NULL ) {
	    occurrence=0;
	    while ( (returnValue = GetTag (++occurrence) ) == GOOD) {
		tagPrefixNeeded = curAtom->tagPrefix;
		if (curAtom->sfdOrderRule == recordOrder)
		    returnValue = AppendBySfdOccurrence();
		else
		    returnValue = AppendBySfdName();
		if (returnValue != GOOD) {
		    ERR_OUT(BAD_RECORD,returnValue);
		}
		if (buildStr != NULL) {
		    if (strlen(curAtom->tagSuffix) > 0) {
			returnValue = AppendString(curAtom->tagSuffix);
			if (returnValue != GOOD) {
			    ERR_OUT("could not add tag suffix",returnValue);
			}
		    }
		    returnValue = MakeDisplayable(buildStr,MAX_CHARS_IN_LINE);
		    if (returnValue != GOOD) {
			ERR_OUT(BAD_RECORD,returnValue);
		    }
		    free(buildStr);
		    buildStr = NULL;
		}
	    }
	    if (returnValue == MEMORY_ERROR) {
		ERR_OUT(BAD_RECORD,returnValue);
	    }
	}
	else {
	   /* GetSpecialRule(); */
	    returnValue = curAtom->SpecialRule();
	    if (returnValue != GOOD) {
		ERR_OUT(BAD_RECORD,BAD_FULL);
	    }
#if   defined(RULEDATA) && defined(FIXSUMTIT)
	    if (buildStr) {
		if (GOOD != WrapDisplayable( buildStr )) {
		    ERR_OUT("memory error",MEMORY_ERROR);
		}
/*???not totally sure what this extra char is for (nl?)*/
		if (displayStr) {
		    if (!(displayStr = (char*)realloc( displayStr , (strlen( displayStr ) + 2)*sizeof(char) ))) {
			ERR_OUT("memory error",MEMORY_ERROR);
		    }
		}
		free( buildStr );
		buildStr = (char*)NULL;
	    }
#endif /*RULEDATA && FIXSUMTIT*/
	   
	}
	curAtom = curAtom->next;
    }

    *outExtract = (char *) realloc(*outExtract, (strlen(displayStr)+1) *sizeof(char) );
    if (*outExtract == NULL) {
	ERR_OUT(BAD_ALLOC,MEMORY_ERROR);
    }
    strcpy(*outExtract,displayStr);

    free (displayStr);

    
    return GOOD;
}

/* given a MARC_RECORD structure, create a displayable full string from its data */
int GetFull(marc,outFull)
MARC_RECORD *marc;
char **outFull;
{

    int returnValue, occurrence;

    static int fullMade = FALSE;

    /* if the first time called, create the linked list */
    if (!(fullMade)) {
	if ( (returnValue = MakeBib (FULL) ) != GOOD) {
	    ERR_OUT("bad full attempt",BAD_FULL);
	}
	fullMade++;
    }

    curMarc = marc;

    returnValue = InitStrings();
    if (returnValue != GOOD) {
	ERR_OUT("could not init working strings",returnValue);
    }

    *outFull = (char *) malloc(sizeof(char));
    if (*outFull == NULL) {
	ERR_OUT(BAD_ALLOC, MEMORY_ERROR);
    }
    (*outFull)[0] = NULL;

    curAtom = fullHead;
    if (curAtom == NULL) {
	ERR_OUT("bad display list",BAD_FULL);
    }

    while (curAtom != NULL) {
	if  (curAtom->SpecialRule == NULL ) {
	    occurrence=0;
	    while ( (returnValue = GetTag (++occurrence) ) == GOOD) {
		tagPrefixNeeded = curAtom->tagPrefix;
		if (curAtom->sfdOrderRule == recordOrder)
		    returnValue = AppendBySfdOccurrence();
		else
		    returnValue = AppendBySfdName();
		if (returnValue != GOOD) {
		    ERR_OUT(BAD_RECORD,returnValue);
		}
		if (buildStr != NULL) {
		    if (strlen(curAtom->tagSuffix) > 0) {
			returnValue = AppendString(curAtom->tagSuffix);
			if (returnValue != GOOD) {
			    ERR_OUT("could not add tag suffix",returnValue);
			}
		    }
		    returnValue = MakeDisplayable(buildStr,MAX_CHARS_IN_LINE);
		    if (returnValue != GOOD) {
			ERR_OUT(BAD_RECORD,returnValue);
		    }
		    free(buildStr);
		    buildStr = NULL;
		}
	    }
	    if (returnValue == MEMORY_ERROR) {
		ERR_OUT(BAD_RECORD,returnValue);
	    }
	}
	else {
	   /* GetSpecialRule(); */
	    returnValue = curAtom->SpecialRule();
	    if (returnValue != GOOD) {
		ERR_OUT(BAD_RECORD,BAD_FULL);
	    }
#if   defined(RULEDATA) && defined(FIXSUMTIT)
	    if (buildStr) {
		if (GOOD != WrapDisplayable( buildStr )) {
		    ERR_OUT("memory error",MEMORY_ERROR);
		}
/*???not totally sure what this extra char is for (nl?)*/
		if (displayStr) {
		    if (!(displayStr = (char*)realloc( displayStr , (strlen( displayStr ) + 2)*sizeof(char) ))) {
			ERR_OUT("memory error",MEMORY_ERROR);
		    }
		}
		free( buildStr );
		buildStr = (char*)NULL;
	    }
#endif /*RULEDATA && FIXSUMTIT*/
	   
	}
	curAtom = curAtom->next;
    }

    *outFull = (char *) realloc(*outFull, (strlen(displayStr)+1) *sizeof(char) );
    if (*outFull == NULL) {
	ERR_OUT(BAD_ALLOC,MEMORY_ERROR);
    }
    strcpy(*outFull,displayStr);

    free (displayStr);
    return GOOD;
}

/* given a MARC_RECORD structure, create a displayable brief string from its data */
int     GetBrief (marc, outBrief, outSummary)
MARC_RECORD *marc; /* the marc record structure */
char  **outBrief, **outSummary; /* the brief record and summary record to be returned */
{
    char    lastName[LABEL_LENGTH+1];

#ifdef   FIXSUMTIT
    char	*psz;
    int		cchSegUse;
    int		cchSegPrint;
    int		oldSummaryLength;	/* doesn't include NUL */
    int		returnValue;
#else  /*FIXSUMTIT*/
     int    preSummaryLength,
            summaryLength,
            returnValue,
            newSummaryLength ;
#endif /*FIXSUMTIT*/

    static int  briefMade = FALSE;

    /* if the first time called, create the linked list */
    if (!(briefMade)) {
	if ((returnValue=MakeBib (BRIEF))!=GOOD) {
	    ERR_OUT("bad brief attempt",BAD_BRIEF);
	}
	briefMade++;
    }

    curMarc = marc;

    /* initialize our strings -- if nothing else, we'll send back 0 length strings */
    returnValue = InitStrings();
    if (returnValue != GOOD) {
	ERR_OUT("could not initialize working strings",returnValue);
    }

    lastName[0] = NULL;

    *outSummary= (char *) malloc(sizeof(char));	
    if (*outSummary==NULL) {
	ERR_OUT(BAD_ALLOC, MEMORY_ERROR);
    }
    (*outSummary)[0]=NULL;
#ifdef   FIXSUMTIT
    oldSummaryLength = 0;
#endif /*FIXSUMTIT*/

    *outBrief= (char *) malloc(sizeof(char));	
    if (*outBrief==NULL) {
	ERR_OUT(BAD_ALLOC, MEMORY_ERROR);
    }
    (*outBrief)[0]=NULL;


    /* loop thtough the linked list */
    curAtom = briefHead;
    if (curAtom==NULL) {
	ERR_OUT("bad display list",BAD_BRIEF);
    }
    do {
	/* a summary will be required even if we find nothing */
#ifndef   FIXSUMTIT
	if ( (curAtom->charsInSummary>0) && (strcmp(lastName,curAtom->label)!=0)
	    && (curAtom->charsInSummary != MAX_SUMMARY_SIZE) ){
	    preSummaryLength = strlen(*outSummary);
	    newSummaryLength = preSummaryLength+LABEL_LENGTH+SUMM_TAB_LEN;
	    *outSummary = (char *) realloc(*outSummary,(newSummaryLength+1)*sizeof(char));
	    if (*outSummary == NULL) {
		ERR_OUT("no room for summary",MEMORY_ERROR);
	    }
	    strcat(*outSummary,summ_tab);
	    while (strlen(*outSummary)<newSummaryLength)
		strcat(*outSummary,BLANK_STRING);
	}
	strcpy(lastName,curAtom->label);
#endif /*!FIXSUMTIT*/

	if (curAtom->SpecialRule == NULL) {

	    /* does the tag exist? */
	    if ((returnValue = GetTag (1) ) == GOOD) {
		tagPrefixNeeded = curAtom->tagPrefix;
		if (curAtom->sfdOrderRule == recordOrder)
		    returnValue=AppendBySfdOccurrence();
		else 
		    returnValue=AppendBySfdName();
		if (returnValue != GOOD) { 
		    ERR_OUT(BAD_RECORD,returnValue);
		}
	    }
	    else if (returnValue == MEMORY_ERROR) {
		ERR_OUT(BAD_ALLOC,MEMORY_ERROR);
	    }
	
	    /* we got something!*/
	    if (buildStr!=NULL) {
		if (strlen( curAtom->tagSuffix)>0) {
		    returnValue = AppendString(curAtom->tagSuffix);
		    if (returnValue != GOOD) {
			ERR_OUT("unable to add tag suffix in brief",returnValue);
		    }
		}
	        /* make it look pretty */
		returnValue = MakeDisplayable(buildStr,MAX_CHARS_IN_LINE);
		if (returnValue!=GOOD) {
		    ERR_OUT(BAD_RECORD,returnValue);
		}
	    }
	}
	else {
	    returnValue = curAtom->SpecialRule();
	    if (returnValue != GOOD) {
		ERR_OUT("could not build brief special rule",returnValue);
	    }
	    if (buildStr) {
		if (GOOD != WrapDisplayable( buildStr )) {
		    ERR_OUT("memory error",MEMORY_ERROR);
		}
/*???not totally sure what this extra char is for (nl?)*/
		if (displayStr) {
		    if (!(displayStr = (char*)realloc( displayStr , (strlen( displayStr ) + 2)*sizeof(char) ))) {
			ERR_OUT("memory error",MEMORY_ERROR);
		    }
		}
	    }
	}

#ifdef   FIXSUMTIT
	if ((buildStr ||
	     curAtom->next && strcmp( curAtom->label , curAtom->next->label)) &&
	    (curAtom->charsInSummary == MAX_SUMMARY_SIZE && buildStr ||
	     0 < curAtom->charsInSummary &&
	     curAtom->charsInSummary < MAX_SUMMARY_SIZE)) {
	  if (!buildStr)
	    cchSegUse = 0;
	  else
	    cchSegUse = (psz = strchr( buildStr+LABEL_LENGTH , '\n' ))
			? psz - (buildStr+LABEL_LENGTH)
			: strlen( buildStr+LABEL_LENGTH );
	  cchSegPrint = cchSegUse;
	  if (curAtom->charsInSummary == MAX_SUMMARY_SIZE) {
	    cchSegPrint = cchSegUse;
	  } else {
	    cchSegPrint = curAtom->charsInSummary;
	    cchSegUse = MIN(cchSegUse,cchSegPrint);
	  }
	    *outSummary = (char*)realloc( *outSummary , (oldSummaryLength + cchSegPrint + SUMM_TAB_LEN + 1)*sizeof(char) );
	    if (*outSummary == NULL) {
		ERR_OUT("no memory for the summary",MEMORY_ERROR);
	    }
	    (void)sprintf( (*outSummary)+oldSummaryLength , "%-*.*s%s" ,
			   cchSegPrint , cchSegUse ,
			   buildStr ? buildStr + LABEL_LENGTH : "" ,
			   summ_tab );
	    oldSummaryLength += cchSegPrint + SUMM_TAB_LEN;
	}
	(void)strcpy( lastName , curAtom->label );
	if (buildStr != NULL) {
#else  /*FIXSUMTIT*/
        /* some of it goes in the summary? */
	 if (buildStr != NULL) {
	     if (curAtom->charsInSummary>0) {
		 summaryLength = strlen(buildStr+LABEL_LENGTH);
		 if (summaryLength > curAtom->charsInSummary)
		     summaryLength = curAtom->charsInSummary;

		 if (curAtom->charsInSummary != MAX_SUMMARY_SIZE) {
		     strncpy((*outSummary)+preSummaryLength,buildStr+LABEL_LENGTH,summaryLength);
		 }
		 else {
		     *outSummary = (char *)	realloc(*outSummary,(strlen(*outSummary)+summaryLength+1)*sizeof(char));
		     if (*outSummary == NULL) {
			 ERR_OUT("no memory for the summary",MEMORY_ERROR);
		     }
		     strcat(*outSummary,buildStr+LABEL_LENGTH);
		 }
	     }
#endif /*FIXSUMTIT*/
	    
	     free(buildStr);
	     buildStr=NULL;
	    
	     /* find the next label */
	     while ( (curAtom!=NULL) && (strcmp(lastName,curAtom->label)==0) )
		 curAtom=curAtom->next;
	 }
	
	/* we didn't find anything for this label */
	else 
	    curAtom = curAtom->next;
	

    } while (curAtom!=NULL);


    /* okay,we've built a nice display string, send it back */
    *outBrief = (char *) realloc(*outBrief,(strlen (displayStr) + 1) * sizeof (char));
    if (*outBrief == NULL) {
	ERR_OUT(BAD_ALLOC, MEMORY_ERROR);
    }
    strcpy (*outBrief, displayStr);
    free (displayStr);
    return GOOD;
}

void ErrOut (errStr)
char    errStr[];
{
    printf ("error at entry %d\n%s", entry, errStr);
    exit (1);
}


int     GetHoldings ()
{

#define SUMMARY_HOLDINGS_TAG 930
#define ITEM_HOLDINGS_TAG 966
#define SUMMARY_FIELDS "lmdji"
#define ITEM_FIELDS "lms"
#define NUMBER_HOLDINGS_TAGS 2
#define CRLF "\n"

    int     returnValue,
            occurrence ,
             tag,
            loop = 0;
    char newLine[FULL_TAB_LEN + 3];

    if (curAtom->sfds[0] == USE_DEFAULT_FLAG)
	free (curAtom->sfds);
	

    strcpy(newLine,CRLF);
    strcat(newLine,full_tab);

    while (loop < NUMBER_HOLDINGS_TAGS) {
	switch (loop) {
	    case 0:
		curAtom->sfds = SUMMARY_FIELDS;
		tag = SUMMARY_HOLDINGS_TAG;
		break;
	    default:
		curAtom->sfds = ITEM_FIELDS;
		tag = ITEM_HOLDINGS_TAG;
		break;
	}
	loop++;
	occurrence = 0;
	
	while((returnValue=mrFindTagName(curMarc,tag,++occurrence,&tagData,&tagLength))==GOOD) {
	    if (buildStr != NULL) {
		returnValue = AppendString(newLine) ;
		if (returnValue != GOOD) {
		    ERR_OUT("could not create holdings",returnValue);
		}
	    }
	    returnValue= AppendBySfdName();
	    if (returnValue != GOOD) {
		ERR_OUT("could not find holdings data",returnValue);
	    }
	}
	if (returnValue == MEMORY_ERROR) {
	    ERR_OUT("memory error in holdings",MEMORY_ERROR);
	}
    }
#ifndef   FIXSUMTIT
    MakeSimpleDisplayable();
#endif /*!FIXSUMTIT*/

    return GOOD;
}

int MESHSubject ()
{
#define MESH_SUBJECT_TAG 650
#define MESH_SUBFIELDS "a"
/* MAJOR_STR relies on Label length, and old name to
 fall such that overwrite replaces leading spaces and "MESH"
 with "MESH MAJOR" */
#define MESH_MAJOR_STR "MESH MAJOR"
#define MESH_MINOR_STR "MESH MINOR"
#define LEGAL_SECOND_IND '2'
#define INDICATOR_TWO 1

    int     returnValue,
            occurrence = 0,
            cycle = 0;
#ifdef   FIXSUMTIT
    int		cchLast;
#endif /*FIXSUMTIT*/

    if (curAtom->sfds[0] == USE_DEFAULT_FLAG) {
	free (curAtom->sfds);
	curAtom->sfds = MESH_SUBFIELDS;
    }

#ifdef   FIXSUMTIT
    cchLast = 0;
#endif /*FIXSUMTIT*/
    while((returnValue=mrFindTagName(curMarc, MESH_SUBJECT_TAG, ++occurrence, &tagData, &tagLength))==GOOD) {
	if (tagData[INDICATOR_TWO] == LEGAL_SECOND_IND) {
	    returnValue= AppendMESHBySfdName();
	    if (returnValue != GOOD) {
		ERR_OUT("could not find MESH Subject data",returnValue);
	    }
	    cycle++;
#ifdef   FIXSUMTIT
	    /* write label */
	    (void)strncpy( buildStr+cchLast , curAtom->label , LABEL_LENGTH );
	    /* overwrite label */
#endif /*FIXSUMTIT*/
	    switch (cycle) {
		case 1:
#ifdef   FIXSUMTIT
		    strncpy (buildStr+cchLast, MESH_MAJOR_STR, strlen(MESH_MAJOR_STR));
#else  /*FIXSUMTIT*/
		    strncpy (buildStr, MESH_MAJOR_STR, strlen(MESH_MAJOR_STR));
#endif /*FIXSUMTIT*/
		    break;
		case 2:
#ifdef   FIXSUMTIT
		    strncpy (buildStr+cchLast, MESH_MINOR_STR, strlen(MESH_MINOR_STR));
#else  /*FIXSUMTIT*/
		    strncpy (buildStr, MESH_MINOR_STR, strlen(MESH_MINOR_STR));
#endif /*FIXSUMTIT*/
		    break;
		default:
#ifdef   FIXSUMTIT
		    strncpy (buildStr+cchLast, full_tab, FULL_TAB_LEN);
#else  /*FIXSUMTIT*/
		    strncpy (buildStr, full_tab, FULL_TAB_LEN);
#endif /*FIXSUMTIT*/
		    break;
	    }
#ifdef   FIXSUMTIT
	    if (buildStr && strlen( buildStr ) != cchLast) {
		AppendString( "\n" );
		cchLast = strlen( buildStr );
		AppendString( full_tab );
	    }
#else  /*FIXSUMTIT*/
	    MakeSimpleDisplayable();
#endif /*FIXSUMTIT*/
	}
    }
    if (returnValue == MEMORY_ERROR) {
	ERR_OUT("memory error in MESH Subject",MEMORY_ERROR);
    }
#ifdef   FIXSUMTIT
    if (buildStr) {
	/* effectively remove extra nl and indent */
	buildStr[cchLast] = '\0';
    }
#endif /*FIXSUMTIT*/

    return GOOD;
}


#ifdef   RULEDATA
/*+++
 * rule			: sublabels - [R3] varargs substitute labels
 * rule paramters	:
 *	0	"R3"
 *	1	(decimal) field tag
 *	2	first substitute label (maybe empty)
 *	3	second substitute label (maybe empty)
 *	.	.
 *	:	:
 *	n	n-1th substitute label (maybe empty)
 */
static
int  sublabels  ()
{
  char		*psz;
  int		cch;
  int		returnValue;
  int		FieldTag;
#ifdef   FIXSUMTIT
  int		cchLastStart, cchLastCheck;
  char		szBlankLabel[LABEL_LENGTH+1];
#endif /*FIXSUMTIT*/
/* Variable */
  char		*pszLeft		= curAtom->pszRuleData;
  int		occurrence		= 0;

  (void)memset( szBlankLabel , BLANK_CHAR , LABEL_LENGTH );
  szBlankLabel[LABEL_LENGTH] = '\0';

  FieldTag = (int)strtol( pszLeft , &psz , 10 );
  if (psz == pszLeft || psz[0] != RULE_PARAM_SEP) {
    ERR_OUT("R3/sublabels: malformed rule data (field tag)",BAD_BIB);
  }
  pszLeft = psz + 1;

  while (GOOD == (returnValue = mrFindTagName(
	curMarc , FieldTag , ++occurrence , &tagData , &tagLength ))) {
    if ((curAtom->indOne[0] == USE_DEFAULT_FLAG ||
	 curAtom->indOne[0] == tagData[0]) &&
	(curAtom->indTwo[0] == USE_DEFAULT_FLAG ||
	 curAtom->indTwo[0] == tagData[1])) {
#ifdef   FIXSUMTIT
      if (buildStr) {
	cchLastStart = strlen( buildStr );
	if (GOOD != AppendString( "\n" )) {
	  ERR_OUT("R3/sublabels: memory error",MEMORY_ERROR);
	}
      } else {
	cchLastStart = 0;
	if (!(buildStr = (char*)malloc( (1)*sizeof(char) ))) {
	  ERR_OUT("R3/sublabels: memory error",MEMORY_ERROR);
	}
	buildStr[0] = '\0';
      }
      if (GOOD != AppendString( szBlankLabel )) {
	ERR_OUT("R3/sublabels: memory error",MEMORY_ERROR);
      }
      cchLastCheck = strlen( buildStr );
#endif /*FIXSUMTIT*/
      if (curAtom->sfdOrderRule == recordOrder)
	returnValue = AppendBySfdOccurrence(  );
      else
	returnValue = AppendBySfdName(  );
      if (returnValue != GOOD) {
	ERR_OUT("R3/sublabels: could not find data",returnValue);
      }
      /* if something was written... */
#ifdef   FIXSUMTIT
      if (buildStr && strlen( buildStr ) != cchLastCheck) {
#else  /*FIXSUMTIT*/
      if (buildStr) {
#endif /*FIXSUMTIT*/
	psz = strchr( pszLeft , RULE_PARAM_SEP );
	cch = psz - pszLeft;
	if (cch > LABEL_LENGTH) {
	  ERR_OUT("R3/sublabels: label too long",BAD_BIB);
	}
#ifdef   FIXSUMTIT
	if (psz) {
	  (void)strncpy( buildStr+cchLastCheck-cch , pszLeft , cch );
	  pszLeft = psz + 1;
	}
#else  /*FIXSUMTIT*/
	if (psz) {
	  (void)memset( buildStr , BLANK_CHAR , LABEL_LENGTH );
	  (void)strncpy( buildStr+LABEL_LENGTH-cch , pszLeft , cch );
	  pszLeft = psz + 1;
	} else {
	  (void)strncpy( buildStr , full_tab , FULL_TAB_LEN );
	}
#endif /*FIXSUMTIT*/
#ifndef   FIXSUMTIT
	(void)MakeSimpleDisplayable(  );
#endif /*!FIXSUMTIT*/
      }
    }
  }
  if (returnValue == MEMORY_ERROR) {
    ERR_OUT("R3/sublabels: memory error",MEMORY_ERROR);
  }

#ifdef   FIXSUMTIT
  if (buildStr) {
   /* remove excess */
    if (cchLastCheck == strlen( buildStr )) {
      buildStr[cchLastStart] = '\0';
    }
  }
#endif /*FIXSUMTIT*/

  return GOOD;
}


/*+++
 * rule			: fixedoffset - [R4] extract data from a fixed offset
 * rule paramters	:
 *	0	"R4"
 *	1	(decimal)
 *	2	(decimal/hex/octal) beginning char offset
 *	3	(decimal/hex/octal) ending char offset (must be >= beginning)
 * rule notes		:
 *	offset values should include indicators, if any.
 */
static
int  fixedoffset  ()
{
  int		returnValue;
  char		*psz;
  int		FieldTag;
  int		iBeg, iEnd;
/* Variable */
  int		occurrence	= 0;
  char		*pszLeft	= curAtom->pszRuleData;

  FieldTag = (int)strtol( pszLeft , &psz , 10 );
  if (psz == pszLeft || psz[0] != RULE_PARAM_SEP) {
    ERR_OUT("R4/fixedoffsets: malformed rule data (field tag)",BAD_BIB);
  }
  pszLeft = psz + 1;
  iBeg = (int)strtol( pszLeft , &psz , 0 );
  if (psz == pszLeft || psz[0] != RULE_PARAM_SEP) {
    ERR_OUT("R4/fixedoffsets: malformed rule data (begin index)",BAD_BIB);
  }
  pszLeft = psz + 1;
  iEnd = (int)strtol( pszLeft , &psz , 0 );
  if (psz == pszLeft || psz[0] != RULE_PARAM_SEP) {
    ERR_OUT("R4/fixedoffsets: malformed rule data (end index)",BAD_BIB);
  }
  pszLeft = psz + 1;

  if (iBeg > iEnd) {
    ERR_OUT("R4/fixedoffsets: begin index after end index",BAD_BIB);
  }

  while (GOOD == (returnValue = mrFindTagName(
	curMarc , FieldTag , ++occurrence , &tagData , &tagLength ))) {
    if ((curAtom->indOne[0] == USE_DEFAULT_FLAG ||
	 curAtom->indOne[0] == tagData[0]) &&
	(curAtom->indTwo[0] == USE_DEFAULT_FLAG ||
	 curAtom->indTwo[0] == tagData[1])) {
#ifdef   FIXSUMTIT
      if (buildStr) {
	if (GOOD != AppendString( "\n" )) {
	  ERR_OUT("R4/fixedoffsets: memory error",MEMORY_ERROR);
	}
      } else {
	if (!(buildStr = (char*)malloc( (1)*sizeof(char) ))) {
	  ERR_OUT("R4/fixedoffsets: memory error",MEMORY_ERROR);
	}
	buildStr[0] = '\0';
      }
      if (GOOD != AppendString( curAtom->label )) {
	ERR_OUT("R4/fixedoffsets: memory error",MEMORY_ERROR);
      }
#endif /*FIXSUMTIT*/
      if (GOOD != AppendCharsNoSP( tagData+iBeg , iEnd - iBeg + 1 )) {
	ERR_OUT("R4/fixedoffsets: memory error",MEMORY_ERROR);
      }
    }
#ifndef   FIXSUMTIT
    (void)MakeSimpleDisplayable(  );
#endif /*!FIXSUMTIT*/
  }
  if (returnValue == MEMORY_ERROR) {
    ERR_OUT("R4/fixedoffsets: memory error",MEMORY_ERROR);
  }

  return GOOD;
}


/*+++
 * rule			: subdata - [R5] varargs substitute data
 * rule paramters	:
 *	0	"R5"
 *	1	(decimal) field tag
 *	2	first substitute data (maybe empty)
 *	3	second substitute data (maybe empty)
 *	.	.
 *	:	:
 *	n	n-1th substitute data (maybe empty)
 */
static
int  subdata  ()
{
  int		FieldTag;
  char		*psz;
  int		returnValue;
/* Variable */
  char		*pszLeft		= curAtom->pszRuleData;
  int		occurrence		= 0;

  FieldTag = (int)strtol( pszLeft , &psz , 10 );
  if (psz == pszLeft || psz[0] != RULE_PARAM_SEP) {
    ERR_OUT("R5/subdata: malformed rule data (field tag)",BAD_BIB);
  }
  pszLeft = psz + 1;

  while (GOOD == (returnValue = mrFindTagName(
	curMarc , FieldTag , ++occurrence , &tagData , &tagLength ))) {
    if ((curAtom->indOne[0] == USE_DEFAULT_FLAG ||
	 curAtom->indOne[0] == tagData[0]) &&
	(curAtom->indTwo[0] == USE_DEFAULT_FLAG ||
	 curAtom->indTwo[0] == tagData[1])) {
      psz = strchr( pszLeft , RULE_PARAM_SEP );
#ifdef   FIXSUMTIT
      if (buildStr) {
	if (GOOD != AppendString( "\n" )) {
	  ERR_OUT("R5/subdata: memory error",MEMORY_ERROR);
	}
      } else {
	if (!(buildStr = (char*)malloc( (1)*sizeof(char) ))) {
	  ERR_OUT("R5/subdata: memory error",MEMORY_ERROR);
	}
	buildStr[0] = '\0';
      }
      if (GOOD != AppendString( curAtom->label )) {
	ERR_OUT("R5/subdata: memory error",MEMORY_ERROR);
      }
#endif /*FIXSUMTIT*/
      if (psz) {
	if (GOOD != AppendCharsNoSP( pszLeft , psz - pszLeft )) {
	  ERR_OUT("R5/subdata: memory error",MEMORY_ERROR);
	}
	pszLeft = psz + 1;
#ifndef   FIXSUMTIT
      } else {
	/* makes label be written */
	if (GOOD != AppendCharsNoSP( " " , 1 )) {
	  ERR_OUT("R5/subdata: memory error",MEMORY_ERROR);
	}
#endif /*!FIXSUMTIT*/
      }
#ifndef   FIXSUMTIT
      (void)MakeSimpleDisplayable(  );
#endif /*!FIXSUMTIT*/
    }
  }
  if (returnValue == MEMORY_ERROR) {
    ERR_OUT("R5/subdata: memory error",MEMORY_ERROR);
  }

  return GOOD;
}


#define  CCHOFFCONTROL	6
#define  CCHOFFDATE1	7
#define  CCHOFFDATE2	11
#define  SZFMTDATE1	"%4.4s-%4.4s"
#define  SZFMTDATE2	"%4.4s, %4.4s"
#define  SZFMTDATE3	"%s%*.*s, %4.4s"	/* this should be longest */
#define  SZFMTDATE4	"%4.4s"
#define  SZFMTDATE5	"%4.4s?"
#define  CCHMAXLENDATE	(9 + 1 + 2 + 2 + 4 + 10)
#define  SZBLANKS	"    "
#define  SZBARS		"||||"
#ifdef   OCLCFIX
#define  SZUS		"uuuu"
#endif /*OCLCFIX*/
/*+++
 * rule			: date - [R6] extact and format date
 * rule paramters	:
 *	0	"R6"
 *	1	(decimal) field tag
 */
static
int  date  ()
{
  int		returnValue;
  char		*psz;
  int		FieldTag;
  char		szMonthCode[2+1];
  int		Month;
  int		cchDay;
  char		szDate[CCHMAXLENDATE+1];
/* Variable */
  int		occurrence	= 0;
  char		*pszLeft	= curAtom->pszRuleData;
/* Data */
  static  char	*aszMonth[]	= {				/*STATIC*/
    "[ERROR]" ,
    "January", "February", "March", "April", "May", "June",
    "July", "August", "September", "October", "November", "December",
  };

  FieldTag = (int)strtol( pszLeft , &psz , 10 );
  if (psz == pszLeft || psz[0] != RULE_PARAM_SEP) {
    ERR_OUT("R6/date: malformed rule data (field tag)",BAD_BIB);
  }
  pszLeft = psz + 1;

  while (GOOD == (returnValue = mrFindTagName(
	curMarc , FieldTag , ++occurrence , &tagData , &tagLength ))) {
    if ((curAtom->indOne[0] == USE_DEFAULT_FLAG ||
	 curAtom->indOne[0] == tagData[0]) &&
	(curAtom->indTwo[0] == USE_DEFAULT_FLAG ||
	 curAtom->indTwo[0] == tagData[1])) {
      if (CCHOFFDATE2+4 >= tagLength) {
	ERR_OUT("R6/date: short marc field",BAD_BIB);
      }
      szDate[0] = '\0';

      /* if date type 1... */
      if        ((tagData[CCHOFFCONTROL] == 'i' ||
		  tagData[CCHOFFCONTROL] == 'k' ||
		  tagData[CCHOFFCONTROL] == 'm') ||
		 (tagData[CCHOFFCONTROL] == 'd' &&
		  curMarc->level == 's')) {
	(void)sprintf( szDate , SZFMTDATE1 ,
		       tagData+CCHOFFDATE1 , tagData+CCHOFFDATE2 );
      /* if date type 2... */
      } else if ((tagData[CCHOFFCONTROL] == 'p' ||
		  tagData[CCHOFFCONTROL] == 'r') ||
		 (tagData[CCHOFFCONTROL] == 'c' &&
		  curMarc->level != 's' &&
		  (strncmp( tagData+CCHOFFDATE2 , SZBLANKS , 4 ) &&
#ifndef   OCLCFIX
		   strncmp( tagData+CCHOFFDATE2 , SZBARS , 4 )))) {
#else  /*!OCLCFIX*/
		   strncmp( tagData+CCHOFFDATE2 , SZBARS , 4 ) &&
		   strncmp( tagData+CCHOFFDATE2 , SZUS , 4 )))) {
#endif /*!OCLCFIX*/
	(void)sprintf( szDate , SZFMTDATE2 ,
		       tagData+CCHOFFDATE1 , tagData+CCHOFFDATE2 );
      /* if date type 3 or 4... */
      } else if (tagData[CCHOFFCONTROL] == 'd' &&
		 curMarc->level != 's') {
	szMonthCode[0] = (tagData+CCHOFFDATE2)[0];
	szMonthCode[1] = (tagData+CCHOFFDATE2)[1];
	szMonthCode[2] = '\0';
	cchDay = 0;
	if ((tagData+CCHOFFDATE2)[3] != ' ')
	  cchDay++;
	if ((tagData+CCHOFFDATE2)[2] != ' ')
	  cchDay++;
	Month = atoi( szMonthCode );
	/* if date type 3... */
	if (1 <= Month && Month <= 12)
	  (void)sprintf( szDate , SZFMTDATE3 ,
			 aszMonth[Month] ,
			 cchDay , cchDay ? cchDay + 1 : 0 ,
			 tagData+CCHOFFDATE2+2 ,
			 tagData+CCHOFFDATE1 );
	/* if date type 4... */
	else
	  (void)sprintf( szDate , SZFMTDATE4 , tagData+CCHOFFDATE1 );
      /* if date type 5... */
      } else if (tagData[CCHOFFCONTROL] == 'q') {
	(void)sprintf( szDate , SZFMTDATE5 , tagData+CCHOFFDATE1 );
      /* if date type 4... */
      } else if (strncmp( tagData+CCHOFFDATE1 , SZBLANKS , 4 ) &&
#ifndef   OCLCFIX
		 strncmp( tagData+CCHOFFDATE1 , SZBARS , 4 )) {
#else  /*!OCLCFIX*/
		 strncmp( tagData+CCHOFFDATE1 , SZBARS , 4 ) &&
		 strncmp( tagData+CCHOFFDATE1 , SZUS , 4 )) {
#endif /*!OCLCFIX*/
	(void)sprintf( szDate , SZFMTDATE4 , tagData+CCHOFFDATE1 );
      }

      if (*szDate) {
#ifdef   FIXSUMTIT
	if (buildStr) {
	  if (GOOD != AppendString( "\n" )) {
	    ERR_OUT("R6/date: memory error",MEMORY_ERROR);
	  }
	} else {
	  if (!(buildStr = (char*)malloc( (1)*sizeof(char) ))) {
	    ERR_OUT("R6/date: memory error",MEMORY_ERROR);
	  }
	  buildStr[0] = '\0';
	}
	if (GOOD != AppendString( curAtom->label )) {
	  ERR_OUT("R6/date: memory error",MEMORY_ERROR);
	}
#endif /*FIXSUMTIT*/
	if (GOOD != AppendString( szDate )) {
	  ERR_OUT("R6/date: memory error",MEMORY_ERROR);
	}
      }
    }
#ifndef   FIXSUMTIT
    (void)MakeSimpleDisplayable(  );
#endif /*!FIXSUMTIT*/
  }
  if (returnValue == MEMORY_ERROR) {
    ERR_OUT("R6/date: memory error",MEMORY_ERROR);
  }

  return GOOD;
}


/*+++
 * rule			: subfieldterm - [R7] subfield terminators
 * rule paramters	:
 *	0	"R7"
 *	1	(decimal) field tag
 *	2	subfield terminator for all but last field (maybe empty)
 *	2	subfield terminator for last field (maybe empty)
 */
static
int  subfieldterm  ()
{
  int		FieldTag;
  char		*psz;
  char		*pszMost, *pszLast;
  int		cchMost, cchLast;
  int		returnValue;
#ifdef   FIXSUMTIT
  int		cchLastStart, cchLastCheck;
#endif /*FIXSUMTIT*/
/* Variable */
  char		*pszLeft		= curAtom->pszRuleData;
  int		occurrence		= 0;

 /* Field tag param */
  FieldTag = (int)strtol( pszLeft , &psz , 10 );
  if (psz == pszLeft || psz[0] != RULE_PARAM_SEP) {
    ERR_OUT("R7/subfieldterm: malformed rule data (field tag)",BAD_BIB);
  }
  pszLeft = psz + 1;
 /* All-but-last terminator */
  psz = strchr( pszLeft , RULE_PARAM_SEP );
  if (psz[0] != RULE_PARAM_SEP) {
    ERR_OUT("R7/subfieldterm: malformed rule data (most terminator)",BAD_BIB);
  }
  pszMost = pszLeft;
  cchMost = psz - pszLeft;
  pszLeft = psz + 1;
 /* Last terminator */
  psz = strchr( pszLeft , RULE_PARAM_SEP );
  if (psz[0] != RULE_PARAM_SEP) {
    ERR_OUT("R7/subfieldterm: malformed rule data (last terminator)",BAD_BIB);
  }
  pszLast = pszLeft;
  cchLast = psz - pszLeft;
  pszLeft = psz + 1;

  while (GOOD == (returnValue = mrFindTagName(
	curMarc , FieldTag , ++occurrence , &tagData , &tagLength ))) {
    if ((curAtom->indOne[0] == USE_DEFAULT_FLAG ||
	 curAtom->indOne[0] == tagData[0]) &&
	(curAtom->indTwo[0] == USE_DEFAULT_FLAG ||
	 curAtom->indTwo[0] == tagData[1])) {
#ifdef   FIXSUMTIT
      if (buildStr) {
	cchLastStart = strlen( buildStr );
	if (GOOD != AppendString( "\n" )) {
	  ERR_OUT("R7/subfieldterm: memory error",MEMORY_ERROR);
	}
      } else {
	cchLastStart = 0;
	if (!(buildStr = (char*)malloc( (1)*sizeof(char) ))) {
	  ERR_OUT("R7/subfieldterm: memory error",MEMORY_ERROR);
	}
	buildStr[0] = '\0';
      }
      if (GOOD != AppendString( curAtom->label )) {
	ERR_OUT("R7/subfieldterm: memory error",MEMORY_ERROR);
      }
      cchLastCheck = strlen( buildStr );
#endif /*FIXSUMTIT*/
      if (curAtom->sfdOrderRule == recordOrder)
	returnValue = AppendBySfdOccurrenceWithTerm(
				pszMost , cchMost , pszLast , cchLast );
      else
	returnValue = AppendBySfdNameWithTerm(
				pszMost , cchMost , pszLast , cchLast );
      if (returnValue != GOOD) {
	ERR_OUT("R7/subfieldterm: memory error",MEMORY_ERROR);
      }
#ifndef   FIXSUMTIT
      (void)MakeSimpleDisplayable(  );
#endif /*!FIXSUMTIT*/
    }
  }
  if (returnValue == MEMORY_ERROR) {
    ERR_OUT("R7/subfieldterm: memory error",MEMORY_ERROR);
  }

#ifdef   FIXSUMTIT
  if (buildStr) {
   /* remove excess */
    if (cchLastCheck == strlen( buildStr )) {
      buildStr[cchLastStart] = '\0';
    }
  }
#endif /*FIXSUMTIT*/

  return GOOD;
}
#endif /*RULEDATA*/


int WrapDisplayable (string)
char *string;
{
    int position, returnValue;
    char *nlis;

    /* can we take advantage of embedded newlines? */
    while ((nlis = strchr(string, '\n')) != NULL) {
	position = nlis - string;
	if ((returnValue = WrapDispRecurse(string, position, MAX_CHARS_IN_LINE)) != GOOD)
	    return returnValue;
	string = nlis + 1;
    }
    if (*string == '\0') return GOOD;
    return WrapDispRecurse(string, strlen(string), MAX_CHARS_IN_LINE);
}

int WrapDispRecurse(string, count, linelim)
char *string;
int count, linelim;
{
    int position,returnValue;
    char *workString = string;
    int cutSpaces = 0;

   /* we can fit it all onto this line */
    if (count < linelim)
	return (AppendDisplayString(string, count));

    /* we have to cut it into several lines */
	
    /* let's look for a place where we can cut the line */
    for (position = linelim - 1;
	 position > 0;
	 position--) {
	/*here's a space within our limit, cut the line here */
	if (isspace(workString[position])){
	    /* if this is first time through, cover
	     case of finding first space at the label. */
	    if ((linelim == MAX_CHARS_IN_LINE) &&
/*???should be <= ?*/
		(position < LABEL_LENGTH)) {
		position = 0;
		break;
	    } else {
		if ((returnValue = AppendDisplayString(workString, position)) != GOOD)
		    return returnValue;
		break;
	    }
	}
    }

   /* wdc 23 Feb 94:  If there is no whitespace on the 
     line, position is zero, but we infinitely recurse.
     New policy: if there's no space to wrap at, just
     cut the line at the end. */

    if (position == 0) {
	if ((returnValue=AppendDisplayString(workString, linelim))!=GOOD)
	    return returnValue;
	position = linelim;
	cutSpaces = 1;
    }

    workString += position;
    count -= position;

    /* we found something to add, make a tab at the next line */
    displayStr=(char *) realloc(displayStr, (strlen(displayStr) + LABEL_LENGTH+1 + cutSpaces ) * sizeof(char));
    if (displayStr == NULL) {
	ERR_OUT("not enough memory for display",MEMORY_ERROR);
    }
    strcat(displayStr, full_tab);
    while ((cutSpaces--) > 0) strcat(displayStr, " ");
    /* first time called, not a recursion */
    if (linelim == MAX_CHARS_IN_LINE)
	linelim -= LABEL_LENGTH;
    /* start the recursion */
    return WrapDispRecurse(workString, count, linelim);
}


int MakeSimpleDisplayable()
{
    if (buildStr != NULL) {
	if (WrapDisplayable(buildStr) != GOOD) {
	    ERR_OUT("not enough memory for display",MEMORY_ERROR);
	}
	free(buildStr);
	buildStr = NULL;
    }
    if (displayStr != NULL) {
	displayStr = (char *) realloc(displayStr,(strlen(displayStr)+2)*sizeof(char) );
	if (displayStr == NULL) {
	    ERR_OUT("no memory for special rule",MEMORY_ERROR);
	}
   }
    return GOOD;
}

int NewBib (h,t)
BIB ** h, **t;
{
    BIB *new;

    new = (BIB *) malloc (sizeof (BIB));
    if (new == NULL)
	return MEMORY_ERROR;

    new->sfds = new->tagPrefix  =  new->indOne = new->indTwo = NULL ;
    new->next = NULL;
    new->charsInSummary= new->tag = 0;
    new->SpecialRule = NULL;
#ifdef   RULEDATA
    new->pszRuleData = (char*)NULL;
#endif /*RULEDATA*/

    if (*t != NULL)
	((*t)->next = new);
    if (*h == NULL)
	*h = new;
    *t = new;

    return GOOD;
}

/* show the linked list, given a current (head?) position */
void ShowList (h)
BIB * h;
{
    
    while (h != NULL) {
/*	printf ("%s\t%s%d%s\t%s\t%d\n", h->label, h->tagPrefix, h->tag, h->tagSuffix, h->sfds, h-> charsInSummary); */
	printf ("label = %s \n", h->label); 
       printf("tag = (%s) %d (%s) \n", h->tagPrefix, h->tag, h->tagSuffix);
       if (h->indOne[0] == USE_DEFAULT_FLAG) 
	   printf("any first indicator\n");
       else 
	   printf("first indicator must be one of %s\n",h->indOne);
       if (h->indTwo[0] == USE_DEFAULT_FLAG) 
	   printf("any second indicator\n");
       else
	   printf("second indicator must be one of %s\n",h->indTwo);
       if (h->sfds[0] == USE_DEFAULT_FLAG)
	   printf("use default subfields\n");
       else {
	   if (h->sfdInclusionRule == doNotUse)
	       printf("do not use ");
	   else
	       printf("use ");
	   printf("subfields %s\n",h->sfds);
       }
       if (h->sfdOrderRule == recordOrder)
	   printf("in record order\n");
       else
	   printf("in listed order\n");
       printf("use %d characters in the summary\n",h->charsInSummary);
	h = h->next;
    }
}

/* append a string of a certain length to the string we are building */
int     AppendChars (charArray, charsToAdd)
char   *charArray;
int     charsToAdd;
{
    return DoAppendChars(charArray, charsToAdd, TRUE);
}

int     AppendCharsNoSP (charArray, charsToAdd)
char   *charArray;
int     charsToAdd;
{
    return DoAppendChars(charArray, charsToAdd, FALSE);
}

 /* append a string of a certain length to the string we are building */
int     DoAppendChars (charArray, charsToAdd, flag)
char   *charArray;
int     charsToAdd;
int     flag;

{
    int     buildLength,currentLength;

    /* we haven't found memory for the string yet */
    if (buildStr == NULL) {
	if (MakeBuildStr() != GOOD)
	    return MEMORY_ERROR;
    }

    currentLength=strlen(buildStr);
    buildLength = currentLength + charsToAdd;
    /*current length + how much to add + space + NULL */
    buildStr = (char *) realloc (buildStr, (buildLength + 2) * sizeof (char));

    if (buildStr == NULL)
	return MEMORY_ERROR;

    strncpy (&buildStr[currentLength], charArray, charsToAdd);
    buildStr[buildLength] = NULL;

    if (flag) strcat (buildStr, " ");

    return GOOD;
}

int     AppendString  (string)
char *string;
{
    if (buildStr == NULL) {
	if (MakeBuildStr() != GOOD)
	    return MEMORY_ERROR;
    }
    buildStr = (char *) realloc (buildStr, (strlen(buildStr)+strlen(string)+ 1) * sizeof (char));
    if (buildStr == NULL)
	return MEMORY_ERROR;
    strcat(buildStr,string);

    return GOOD;
}

/* given a pointer to a tag, append the first occurrence of a specific subfield */
int     Appendsfd (tagData, tagLength, sfd)
char   *tagData,
        sfd;
int     tagLength;

{
    int     returnValue,
            sfdLength;
    char   *sfdData;

    if ((returnValue = mrFindSubFieldName (sfd, 1, tagData, tagLength, &sfdData, &sfdLength)) == GOOD) {
	if ((returnValue = AppendChars (sfdData, sfdLength)) != GOOD)
	    return returnValue;
    }
    else
	if (returnValue == MEMORY_ERROR)
	    return returnValue;
    return GOOD;
}

/* break lines and indent continuing labels 
   this is a recursive function*/
int MakeDisplayable (string,currentLineLimit)
char *string;
int currentLineLimit; /* how much room is left on the line we are building? */
{

    int position,returnValue;
    char *workString;
    int cutSpaces = 0;
    
    /* skip leading spaces */
    position=0;
 /*   while (isspace(string[position]))
	    position++; */
    workString = &string[position];

    /* we can fit it all onto this line */
    if ((position=strlen(workString))<currentLineLimit)
	return (AppendDisplayString(workString,position));
    
    /* we have to cut it into several lines */

    /* let's look for a place where we can cut the line */
    for (position=currentLineLimit-1;position>0;position--) {
	/*here's a space within our limit, cut the line here */
	if (isspace(workString[position])){
	    /* if this is first time through, cover
	     case of finding first space at the label. */
	    if ((currentLineLimit==MAX_CHARS_IN_LINE) &&
		(position < LABEL_LENGTH)) {
		position = 0;
		break;
	    } else {
		if ((returnValue = AppendDisplayString (workString, position))!=GOOD)
		    return returnValue;
		break;
	    }
	}
    }
    /* wdc 23 Feb 94:  If there is no whitespace on the
     line, position is zero, but we infinitely recurse.
     New policy: if there's no space to wrap at, just
     cut the line at the end. */

    if (position == 0) {
	if ((returnValue=AppendDisplayString(workString, currentLineLimit))!=GOOD)
	    return returnValue;
	position = currentLineLimit;
	cutSpaces = 1;
    }

    /* we found something to add, make a tab at the next line */
    if (strlen(&workString[position])>0) {
	 displayStr=(char *) realloc(displayStr,(strlen(displayStr) + LABEL_LENGTH+1 + cutSpaces) * sizeof(char));
	 strcat(displayStr,full_tab);
	 while ((cutSpaces--) > 0) strcat(displayStr, " ");
	 /* first time called, not a recursion */
	 if (currentLineLimit==MAX_CHARS_IN_LINE)
	     currentLineLimit-= LABEL_LENGTH;
	 /* start the recursion */
	 return (MakeDisplayable(&workString[position],currentLineLimit));
    }
    else 
	return GOOD;
}

int MakeBuildStr ()
{
    buildStr = (char *) malloc((LABEL_LENGTH+1)*sizeof(char));
    if (buildStr == NULL)
	return MEMORY_ERROR;
    strcpy(buildStr,curAtom->label);
    return GOOD;
}


/* add a string to our pretty little buffer */
int AppendDisplayString (string,stringSize)
char *string;
int stringSize;

{
    int newSize;

    newSize=strlen(displayStr)+stringSize;
    /* leave room for a null and carriage return */
    displayStr=(char *) realloc(displayStr,(newSize+2)*sizeof(char));
    if (displayStr == NULL)
	return MEMORY_ERROR;
    strncat(displayStr,string,stringSize);
    displayStr[newSize]=NULL;
    strcat(displayStr,"\n");
    return GOOD;
}

/* given a pointer to a tag, add each subfield as it occurs in the record */
int AppendBySfdOccurrence ()

{
    int sfdOccurrence,returnValue;
    char foundSfd,*sfdData,*sfdMatch;
    int sfdLength;

    /* loop through the subfields */
    for (sfdOccurrence=1; ;sfdOccurrence++) {
	returnValue = mrGetSubField (sfdOccurrence, tagData, tagLength, &foundSfd, &sfdData, &sfdLength);
	if (returnValue != GOOD)
	    break;
	sfdMatch = strchr(curAtom->sfds,foundSfd);
	if ( (curAtom->sfdInclusionRule == doNotUse) && (sfdMatch !=NULL) )
	    continue;
	if ( (curAtom->sfds[0] != USE_DEFAULT_FLAG) && (curAtom->sfdInclusionRule == use) && (sfdMatch == NULL) )
	    continue;

	if ( (tagPrefixNeeded!=NULL) && (strlen(curAtom->tagPrefix)>0) ) {
	    if (AppendString(curAtom->tagPrefix) !=GOOD)
		return MEMORY_ERROR;
	    tagPrefixNeeded = NULL;
	}

	if (AppendChars (sfdData, sfdLength) != GOOD)
	    return MEMORY_ERROR;
	
    }
    if (returnValue == MEMORY_ERROR)
	return MEMORY_ERROR;
    return GOOD;
}

/* given a pointer to a tag, add all occurrences of specified subfield names */   
int AppendBySfdName ()
{
    int ptr,i,sfdOccurrence,returnValue,sfdLength;
    char *sfdData;
    
    /* loop through the subfields listed */
    for (ptr = 0, i = strlen (curAtom->sfds); ptr < i; ptr++) {
	/*loop through the occurrences of the subfield */
	sfdOccurrence = 1;
	while ((returnValue = mrFindSubFieldName (curAtom->sfds[ptr], sfdOccurrence, tagData,tagLength, &sfdData, &sfdLength)) == GOOD) {
	    if (tagPrefixNeeded != NULL) {
		if (AppendString(curAtom->tagPrefix)!=GOOD)
		    return MEMORY_ERROR;
		tagPrefixNeeded=NULL;
	    }
	    if (AppendChars (sfdData, sfdLength) != GOOD)
		return MEMORY_ERROR;
	    sfdOccurrence++;
	}
	if (returnValue == MEMORY_ERROR)
	    return MEMORY_ERROR;
    }
    return GOOD;
}


#ifdef   RULEDATA
/* given a pointer to a tag, add each subfield as it occurs in the record */
int  AppendBySfdOccurrenceWithTerm  ( pszTermMost , cchTermMost , pszTermLast , cchTermLast )
char	*pszTermMost;
int	 cchTermMost;
char	*pszTermLast;
int	 cchTermLast;
{
    int sfdOccurrence,returnValue;
    char foundSfd,*sfdData,*sfdMatch;
    int sfdLength;
    int		cch;
    int		fOutput		= FALSE;

    /* loop through the subfields */
    for (sfdOccurrence=1; ;sfdOccurrence++) {
	returnValue = mrGetSubField (sfdOccurrence, tagData, tagLength, &foundSfd, &sfdData, &sfdLength);
	if (returnValue != GOOD)
	    break;
	sfdMatch = strchr(curAtom->sfds,foundSfd);
	if ( (curAtom->sfdInclusionRule == doNotUse) && (sfdMatch !=NULL) )
	    continue;
	if ( (curAtom->sfds[0] != USE_DEFAULT_FLAG) && (curAtom->sfdInclusionRule == use) && (sfdMatch == NULL) )
	    continue;

	if ( (tagPrefixNeeded!=NULL) && (strlen(curAtom->tagPrefix)>0) ) {
	    if (AppendString(curAtom->tagPrefix) !=GOOD)
		return MEMORY_ERROR;
	    tagPrefixNeeded = NULL;
	}

	if (AppendCharsNoSP (sfdData, sfdLength) != GOOD)
	    return MEMORY_ERROR;
	if (AppendCharsNoSP( pszTermMost , cchTermMost ) != GOOD)
	    return MEMORY_ERROR;
	fOutput = TRUE;
	
    }
    if (returnValue == MEMORY_ERROR)
	return MEMORY_ERROR;
    if (fOutput) {
	cch = strlen( buildStr );
	buildStr[cch-cchTermMost] = '\0';
	if (cchTermMost >= cchTermLast) {
	    (void)strncpy(
		buildStr+cch-cchTermMost , pszTermLast , cchTermLast );
	    buildStr[cch-cchTermMost+cchTermLast] = '\0';
	} else {
	    if (AppendCharsNoSP( pszTermLast , cchTermLast ) != GOOD)
		return MEMORY_ERROR;
	}
    }
    return GOOD;
}


int  AppendBySfdNameWithTerm  ( pszTermMost , cchTermMost , pszTermLast , cchTermLast )
char	*pszTermMost;
int	 cchTermMost;
char	*pszTermLast;
int	 cchTermLast;
{
    int ptr,i,sfdOccurrence,returnValue,sfdLength;
    char *sfdData;
    int		cch;
    int		fOutput		= FALSE;

    /* loop through the subfields listed */
    for (ptr = 0, i = strlen (curAtom->sfds); ptr < i; ptr++) {
	/*loop through the occurrences of the subfield */
	sfdOccurrence = 1;
	while ((returnValue = mrFindSubFieldName (curAtom->sfds[ptr], sfdOccurrence, tagData,tagLength, &sfdData, &sfdLength)) == GOOD) {
	    if (tagPrefixNeeded != NULL) {
		if (AppendString(curAtom->tagPrefix)!=GOOD)
		    return MEMORY_ERROR;
		tagPrefixNeeded=NULL;
	    }
	    if (AppendCharsNoSP( sfdData , sfdLength ) != GOOD)
		return MEMORY_ERROR;
	    if (AppendCharsNoSP( pszTermMost , cchTermMost ) != GOOD)
		return MEMORY_ERROR;
	    fOutput = TRUE;
	    sfdOccurrence++;
	}
	if (returnValue == MEMORY_ERROR)
	    return MEMORY_ERROR;
    }
    if (fOutput) {
	cch = strlen( buildStr );
	buildStr[cch-cchTermMost] = '\0';
	if (cchTermMost >= cchTermLast) {
	    (void)strncpy(
		buildStr+cch-cchTermMost , pszTermLast , cchTermLast );
	    buildStr[cch-cchTermMost+cchTermLast] = '\0';
	} else {
	    if (AppendCharsNoSP( pszTermLast , cchTermLast ) != GOOD)
		return MEMORY_ERROR;
	}
    }
    return GOOD;
}
#endif /*RULEDATA*/


/* given a pointer to a tag, add all occurrences of specified subfield names */   
int AppendMESHBySfdName ()
{
    int ptr,i,sfdOccurrence,returnValue,sfdLength;
    char *sfdData;
    char newLine[FULL_TAB_LEN + 3];
    char *sfdPart;
    int sfdPartLen;

    strcpy(newLine,CRLF);
    strcat(newLine,full_tab);
    
    /* loop through the subfields listed */
    for (ptr = 0, i = strlen (curAtom->sfds); ptr < i; ptr++) {
	/*loop through the occurrences of the subfield */
	sfdOccurrence = 1;
	while ((returnValue = mrFindSubFieldName (curAtom->sfds[ptr], sfdOccurrence, tagData,tagLength, &sfdData, &sfdLength)) == GOOD) {
	    if (tagPrefixNeeded != NULL) {
		if (AppendString(curAtom->tagPrefix)!=GOOD)
		    return MEMORY_ERROR;
		tagPrefixNeeded=NULL;
	    }
	    if (sfdOccurrence > 1)
		if (AppendString(newLine) != GOOD) return MEMORY_ERROR;

	    /* This block will append the chars and skip
	     occurances of '*' */
#ifdef SUPPRESS_STARS
	    sfdPartLen = sfdLength;
	    sfdPart = sfdData;
	    while (sfdPartLen > 0) {
		int count = 0;
		while (sfdPartLen > 0 && *sfdPart == '*') {
		    sfdPartLen--;
		    sfdPart++;
		}
		while ((sfdPartLen > 0) && (*(sfdPart + count) != '*')) {
		    count++;
		    sfdPartLen--;
		}			
		if (count)
		    if (AppendCharsNoSP (sfdPart, count) != GOOD)
			return MEMORY_ERROR;
		sfdPart += count;
	    }
#else	    
	    if (AppendChars (sfdData, sfdLength) != GOOD)
		return MEMORY_ERROR;
#endif
	    sfdOccurrence++;
	}
	if (returnValue == MEMORY_ERROR)
	    return MEMORY_ERROR;
    }
    return GOOD;
}

/*
 * Open BibFile
 *
 * Accepts as input a filename
 *
 * Searches for it in the following places, in the
 * following order:
 *
 * <willowDir>/drivers/z3950.format/<file>.<bib_suffix>
 * <willowDir>/drivers/z3950.format/<file>
 * <default-bib-directory><file>
 *
 * willowDir is passed in from willow via get_env.
 * bib_suffix is passed in from willow via get_env.
 * default-bib-directory is defined in bib.h and
 *    is assumed to end with a trailing slash.
 * Currently there are TWO names defined for file.
 * They are defined in bib.h.
 * They are assumed NOT to have slashes in them, and
 *    not to begin with a period.
 *
 * (Modifications 1/19/94 wdc@mit.edu)
 * 4/8/94 wdc@mit.edu --
 *  bib files in the home dir are no longer supported
 *  bib_suffix in db.conf added.
 *
 */

#define BIB_FILE_PATH "/drivers/z3950.format/"
extern char *config_dir;

FILE *OpenBibFile(file)
char *file;
{
  
    int getuid();
    char *env_get(), *suffix;
    char bibFile[MAX_LINE];
    struct passwd *pw, /* all used for finding the user's default directory */
           *getpwuid(),
           *getpwnam();

    FILE *fp = NULL;

    suffix = env_get("bib_suffix");

    if (suffix)
      if (suffix[0] != '\0') {
	strcpy(bibFile, config_dir);
	strcat(bibFile, BIB_FILE_PATH);
	strcat(bibFile, file);
	strcat(bibFile, ".");
	strcat(bibFile, suffix);
	fp = fopen (bibFile, "r");
      }

    /* falls through to here if suffix nonexistant */
    if (fp == NULL) { 
	strcpy(bibFile, config_dir);
	strcat(bibFile, BIB_FILE_PATH);
	strcat(bibFile, file);
	fp = fopen (bibFile, "r");
    }

    
    /* not there -- use the default */
    if (fp == NULL) {
	strcpy(bibFile,DEFAULT_BIB_DIRECTORY);
	strcat(bibFile,file);
	fp = fopen (bibFile, "r") ;
    }

    return fp;
}

int ParseLiteral( string,stopChar,stringLen)
char **string;
int stringLen;
char stopChar;
{
    register int position;
    
    position=inPosition;
    while ( (position<stringLen) && (inString[position]!=stopChar) )
	position++;

    if (position>=stringLen)
	return BAD_BIB;
    if (inString[position] == stopChar) {
	*string= (char *) malloc((position-inPosition+1)*sizeof(char));
	if (string ==NULL)
	    return MEMORY_ERROR;
	strncpy(*string,&inString[inPosition],position-inPosition);
	*(*string+position-inPosition)=NULL;
	inPosition+= position;
    }
    return GOOD;
}

int ParseIndicators (string)
char **string;
{
    *string = (char *) malloc((strlen(inString)+1) *sizeof(char));
    if (*string == NULL)
	return MEMORY_ERROR;
    strcpy(*string,inString);
    return GOOD;
}
int ParseTag ()
{
    int ptr;
    inPosition=0;
    curAtom->tagPrefix = (char *) malloc (sizeof(char));
    if (curAtom->tagPrefix == NULL) {
	ERR_OUT("no memory for tag prefix",MEMORY_ERROR);
    }
    curAtom->tagPrefix[0] = NULL;

    curAtom->tagSuffix = (char *) malloc(sizeof(char));
    if (curAtom->tagSuffix ==NULL) {
	ERR_OUT("no memory for tag suffix",MEMORY_ERROR);
    }
    curAtom->tagSuffix[0] = NULL;

    if (inString[inPosition] == SPECIAL_RULE_FLAG) {
	inPosition++;
	if (AssignSpecialRule() != GOOD) {
	    ERR_OUT("illegal rule",BAD_BIB);
	}
	curAtom->tag = USE_SPECIAL_RULE;
	return GOOD;
    }

    if (inString[inPosition]==PREFIX_START)  {
	inPosition++;
	free(curAtom->tagPrefix);
	if ( ParseLiteral(&curAtom->tagPrefix,PREFIX_END,strlen(inString))!=GOOD) {
	    ERR_OUT("bad tag prefix",BAD_BIB);
	}
	InsertBlanks(curAtom->tagPrefix);
    }
    
    if  (strlen(&inString[inPosition]) <TAG_MARK_LEN) {
	ERR_OUT("illegal tag",BAD_BIB);
    }
    for (ptr=0; ptr < TAG_MARK_LEN; ptr++) {
	if (!isdigit (inString[inPosition+ptr])) {
	    ERR_OUT ("illegal tag",BAD_BIB);
	}
	curAtom->tag = (curAtom->tag*10) + (int) inString[inPosition+ptr] - 48;
    }
    if((curAtom->tag<MIN_TAG_ALLOWED)||(curAtom->tag> MAX_TAG_ALLOWED)) {
	ERR_OUT ("illegal tag",BAD_BIB);
    }

    inPosition+=ptr;

    if (inPosition>= strlen(inString))
	return GOOD;
    
    else if (inString[inPosition] != SUFFIX_START) {
	ERR_OUT("illegal tag",BAD_BIB);
    }

    if (inString[inPosition]==SUFFIX_START)  {
	inPosition++;
	free (curAtom->tagSuffix);
	if ( ParseLiteral(&curAtom->tagSuffix,SUFFIX_END,strlen(inString) )!=GOOD) {
	    ERR_OUT("bad tag suffix",BAD_BIB);
	    	}
	InsertBlanks(curAtom->tagSuffix);
    }
    
    return GOOD;
}
int ParseSfd ()
{

    if (inString[0]==USE_DEFAULT_FLAG) {
	curAtom->sfds = (char *) malloc (2*sizeof (char));
	if (curAtom->sfds == NULL) {
	    ERR_OUT("no memory for subfields",MEMORY_ERROR);
	}
	curAtom->sfds[0] = USE_DEFAULT_FLAG;
	curAtom->sfds[1] = NULL;
	curAtom->sfdOrderRule = recordOrder;
	curAtom->sfdInclusionRule = use;

    }
    else if (inString[0] == DO_NOT_USE_FLAG) {
	/*notice that we have room for the null because we move the string from
	    offset one */
	curAtom->sfds = (char *) malloc ( strlen(inString)*sizeof(char));
	if (curAtom->sfds == NULL) {
	    ERR_OUT("no memory for subfields",MEMORY_ERROR);
	}
	strcpy(curAtom->sfds,&inString[1]);
	curAtom->sfdInclusionRule = doNotUse;
	curAtom->sfdOrderRule = recordOrder;
    }
    else {
	inPosition=0;
	if (inString[inPosition] == SPECIFIC_ORDER_FLAG)  {
	    curAtom->sfdOrderRule = listedOrder;
	    inPosition++;
	}
	else 
	    curAtom->sfdOrderRule = recordOrder;
	curAtom->sfdInclusionRule = use;
	curAtom->sfds = (char *) malloc ((strlen (&inString[inPosition])+1) * sizeof (char));
	if (curAtom->sfds == NULL) {
	    ERR_OUT("no memory for subfields",MEMORY_ERROR);
	}
	strcpy (curAtom->sfds, &inString[inPosition]);
    }
    return GOOD;
}

int ParseCharsInSummary ()
{
    int ptr,length;

    if (inString[0]==USE_DEFAULT_FLAG)
	curAtom->charsInSummary = MAX_SUMMARY_SIZE;
    else if (inString[0]==DO_NOT_USE_FLAG)
	curAtom->charsInSummary = 0-1;
    else {
	if ((length=strlen (inString)) > MAX_SUMMARY_DIGITS) {
	    ERR_OUT ("illegal summary statement",BAD_BRIEF);
	}
	for (ptr = 0; ptr < length; ptr++)
	    if (!isdigit (inString[ptr])) {
		ERR_OUT ("illegal summary statement",BAD_BRIEF);
	    }
	curAtom->charsInSummary = atoi (inString);
	if ((curAtom->charsInSummary < 1) || (curAtom->charsInSummary > MAX_SUMMARY_SIZE)) {
	    ERR_OUT ("illegal summary",BAD_BRIEF);
	}
    }
    return GOOD;
}

int GetTag(occurrence)
int occurrence;
{
    int returnValue;

    returnValue = mrFindTagName(curMarc,curAtom->tag,occurrence,&tagData,&tagLength);
    if (returnValue !=GOOD)
	return returnValue;
    if ( (curAtom->indOne[0] != USE_DEFAULT_FLAG) && (strchr(curAtom->indOne,tagData[0])==NULL) )
	return FAILED_IND_TEST;
    if ( (curAtom->indTwo[0] != USE_DEFAULT_FLAG) && (strchr(curAtom->indTwo,tagData[1]) == NULL) )
	return FAILED_IND_TEST;

    return GOOD;
}

int InitStrings ()
{
    buildStr = NULL;
    displayStr = (char *) malloc(sizeof(char));
    if (displayStr == NULL)
	return MEMORY_ERROR;
    displayStr[0] = NULL;
    return GOOD;
}

void InsertBlanks (string)
char *string;
{
    char *blankPosition;

    while ( (blankPosition = strchr(string,BLANK_CODE) ) != NULL)
	blankPosition[0] = BLANK_CHAR;

    return;
}

int AssignSpecialRule ()
{
#ifdef   RULEDATA
    char	*psz;
    int		rule;

    rule = (int)strtol( inString+inPosition , &psz , 0 );
    if (psz == inString+inPosition ||
	psz[0] != '\0' && psz[0] != RULE_PARAM_SEP) {
        ERR_OUT("illegal rule",BAD_BIB);
    }

    /* if rule data... */
    if (psz) {
        psz++;
	curAtom->pszRuleData = (char*)malloc(
					(strlen( psz ) + 1)*sizeof(char) );
	if (curAtom->pszRuleData == NULL) {
	    ERR_OUT("no memory for rule data",MEMORY_ERROR);
	}
	(void)strcpy( curAtom->pszRuleData , psz );
	InsertBlanks( curAtom->pszRuleData );
    }
#else  /*RULEDATA*/
    int ptr,stringLen, rule=0;

    stringLen = strlen (&inString[inPosition]);
    if (stringLen >MAX_RULE_LENGTH) {
	ERR_OUT("illegal rule",BAD_BIB); 
    }

    for (ptr=0; ptr<stringLen; ptr++) {
	if (!isdigit (inString[inPosition+ptr])) {
	    ERR_OUT ("illegal rule",BAD_BIB);
	}
	rule = (rule*10) + (int) inString[inPosition+ptr] - 48;
    }
#endif /*RULEDATA*/

    if ( (rule < 1) || (rule > MAX_RULE) ) {
	ERR_OUT("illegal rule number",BAD_BIB);
    }
    curAtom->SpecialRule = ruleTable[rule];
    return GOOD;
}
/*
void GetSpecialRule ()
{
    switch (curAtom->rule) {
	case 1 :
	    SpecialRule = &GetHoldings;
	    break;
	default:
	    SpecialRule = &BadRule;
    }
    return;
}*/



int BadRule()
{
    ERR_OUT("bad rule called", BAD_BIB);
}

void MakeRuleTable()
{
    ruleTable[1] = GetHoldings;
    ruleTable[2] = MESHSubject;
#ifdef   RULEDATA
    ruleTable[3] = sublabels;
    ruleTable[4] = fixedoffset;
    ruleTable[5] = subdata;
    ruleTable[6] = date;
    ruleTable[7] = subfieldterm;
#endif /*RULEDATA*/
    return;
}

