/***********************************************************************
 *									   *
 *              WILLOW(tm) Information Retrieval Software		   *
 *     (Washington Information Looker-upper Layered Over Windows)	   *
 *             Copyright 1992-1994 University of Washington                *
 * 									   *
 *   Willow is a trademark of the University of Washington.  No        *
 *   commercial use of this trademark or software may be made without  *
 *   prior written permission of the University of Washington.         *
 *                                                                     *
 * Permission to use, copy, modify, and distribute Willow(tm) software	   *
 * and its documentation for any noncommercial purpose is hereby       *
 * granted, provided that the above copyright notice appears in all	   *
 * copies and that both the above copyright notice and this permission	   *
 * notice appear in supporting documentation, and that the name of the	   *
 * University of Washington not be used in advertising or publicity	   *
 * pertaining to distribution of the software without specific,        *
 * written prior permission.  This software is made available "as is", *
 * and THE UNIVERSITY OF WASHINGTON DISCLAIMS ALL WARRANTIES, EXPRESS  *
 * OR IMPLIED, WITH REGARD TO THIS SOFTWARE, INCLUDING WITHOUT         *
 * LIMITATION ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS    *
 * FOR A PARTICULAR PURPOSE, AND IN NO EVENT SHALL THE UNIVERSITY OF   *
 * WASHINGTON BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL     *
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA  *
 * OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, TORT (INCLUDING       *
 * NEGLIGENCE) OR STRICT LIABILITY, ARISING OUT OF OR IN CONNECTION    *
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.                       *
 * 									   *
 * For further information please contact:				   *
 * Willow Project  							   *
 * Information Systems							   *
 * University of Washington						   *
 * 4545 15th NE, 1st floor 						   *
 * Seattle, Washington, 98105, USA					   *
 * Internet: willow@cac.washington.edu					   *
 ***********************************************************************/

  /***************************************************
  * driver.c
  * This is core of the database driver.
  *
  * It is a hacked up version of the uwbrs driver.
  * 
  * Some routines have the suffix _wcb or _hcb on their names. These
  * are "Callback" routines, modeled after those in X. They are
  * triggered when certain events occur. _wcb routines are triggered by
  * incoming packets from Willow, while _hcb routines are triggered by
  * incoming strings from the database host.
  ********************************************************/

/*
 * Now contains code from Stanford Z39.50 client:
 *  Copyright (c) 1992 Stanford Leland Junior University
 *
 *  Commercial use prohibited
 *
 *  Author: Harold Finkbeiner (Harold@Forsythe.Stanford.Edu)
 */

/*
 *  Copyright (C) 1994 by the Massachusetts Institute of Technology
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice appear in all
 * copies and that both that copyright notice and this permission
 * notice appear in supporting documentation, and that the name of
 * M.I.T. not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior
 * permission.  M.I.T. makes no representations about the suitability
 * of this software for any purpose.  It is provided "as is" without
 * express or implied warranty.
 */

/*
 * This unholy amalgamation performed by Bill Cattey
 * MIT Information Systems
 * wdc@mit.edu
 */

/*
 * Open issues:
 *
 * Assembling the search string:
 * Searches sent by Willow are arrays of strings, keyword, 
 * and type with an implied and operation between them.
 * Searches sent to Z39.50 are rpn queries; type following
 * keyword.
 * I am reasonably sure that I have botched the conversion
 * between these when it comes to use of parentheses.
 * Willow ALWAYS surrounded each clause and the search
 * as a whole with parentheses.  I left that code out.
 *
 */


#include <stdio.h>
#ifdef SOLARIS
#include <string.h>
#else
#include <strings.h>
#endif /* SOLARIS */

#include <sys/types.h>
#include <netinet/in.h>
#include <sys/wait.h>
#include <ctype.h>
#include <signal.h>
#include <errno.h>
#if defined(_AIX) || defined (_SEQUENT_)
#include <sys/select.h>
#endif

#include "license.h"
#include "license_status.h"

static void updater(void);

#include "driver.h"
#include "protocol.h"
#define W_RECORD 8
#if (W_RECORD != RECORD)
??? Value of record in protocol.h has changed ???
#endif
#undef RECORD

#include "host.h"

#include <constant.h>
#include <netio.h>
#include <berutil.h>
#include <msg.h>
#include <pdutags.h>
#include <putget.h>
#include <session.h>
#include <strings.h>
#include <z3950.h>
#include <recinfo.h>
#include <api.h>
#include <resource.h>

#define MAXPATHLEN 1024

/* External variables */
extern int text_lines;

/* Extern routines */
extern char *malloc();
extern char *realloc();
extern char *msg_hcb(char *msg);
extern int malarm_add (int interval, void (*handler) (/* ??? */));
extern int hcb_record_msg(RECORD_INFO *info);
extern int do_io_select(int operation);
extern packet_rec *netify(packet_rec *packet);
extern packet_rec *denetify(packet_rec *packet);
extern void validate_class(int sock, char *class, char *class_file);
void browse_wcb(SESSION **p_session, packet_rec *packet);

/* Forward Routines */
char *dump_pac_type (int pac_type);
void packet_operate(packet_rec *p_packet);
void read_from_client(packet_rec *p_packet);
char *env_get(char *s);
void parse_format_string(void);
void get_info_wcb(SESSION **p_session, packet_rec *ignored);
void send_title();
void send_record(int num, char *text, int summ_flag);
void send_ext();
void send_user_req(int type, char *query_text, char *query_tag, char *yes_label, char *no_label);
int user_resp_wcb(packet_rec *packet, char **p_response);
void user_resource_control_confirm(SESSION *session, long report[MAX_RESOURCE_REPORT_BIB1], int *p_complete, STRING *message);


/***** GLOBALS *****/

FILE *log;

char *logname = LOG_DEFAULT;

int g_search_num = 0;          /* reference number for this result set */

/* Kludge for hook into Stanford Z code. */
FILE *out = stdout;

/* Hold text read between matches in this buffer */
char *parse_buf;

/* Level of debugging, passed in on command line */
int debug_level = 0;

/* Kludge for hook into Stanford Z code. */
int debug = 0;

/* Are we being accessed via a network socket rather than a local pipe? */
char socketed = FALSE;

/* file descriptors for communication with willow. Could be a pipe
   or a socket */
int to_willow, from_willow;

/* Current characters in parse_buf */
int    parse_buf_size;
int    parse_buf_used = 0;

char   fresh_input;  /* flag */

int    host_pid = NULL;     /* process id */

extern state_table states;           /* The state table */

/* The current state of the remote computer */
int    state;

/* The current state of our connection with willow */
char   connected_to_willow = TRUE;

/* Are we currently echoing interactions with host? */
char   echo = FALSE;

/* a couple of pipes for talking to telnet */
int  to_host[2],
     from_host[2];

/* File descriptor mask for select call */
fd_set read_fds;
int    num_fds;

/* Head and tail of linked list of hilite points */
hl_rec *hl_head = NULL, *hl_tail = NULL;
int     num_hilites = 0;

/* Environment list */
env_rec env[MAX_ENV];
int env_cnt = 0;

/* Maximum number of circ requests we will do */
int max_circ;

/* Timeout value for cir requests */
int circ_time;

/* Key for fetching MARC Elts in Extension */
char *ext_fields;

/* Salt away the preferred syntax at init time
 so we don't need to fetch it every time for search */
char *w_preferred_syntax = "";

/* structure for license manager, global cause we allocate in one routine,
   but free in another */
static  struct ticket_str *db_ticket=NULL; 

/* ticket-updater alarm id */
int updater_id = NULL;

/* Name of the file containing IP address class info */
char *class_file = NULL;

/* I hope this stuff with the summary fields is temporary.
 It would be nice if the MARC formatter would do this
 neatly */

char *summary_fields = 0;    /* Which fields we use for summaries */

struct format_rec summary_formats[128]; 
int  summary_long_len = 0;   /* How wide are the expanded field names */
int  summary_field_cnt = 0;  /* How many fields in our summary */

/* directory where willow config files live. Not used by uwbrs,
   but needed by z3950 driver */
char *config_dir = "";

struct format_rec title_formats[128];
int  title_field_cnt = 0;

#define USAGE "USAGE: z3950 [-debug <number>] [-socketed] [-outfile <file>]\n\
  [-cfg_path <path>] [-class_file <file>]"

/* Takes one argument, an integer representing the debugging level */
main(argc, argv)
int   argc;
char *argv[];
{
    char *logname = LOG_DEFAULT;
    extern int Client_FD;
    FILE *outfile;
    char *outfile_name = "/dev/null";
    char  redirect = FALSE;

    /* Process the command line args */
    argc--; argv++;
    while (argc) {
      if (strcmp("-debug", argv[0]) == 0) {
	if (argv[1]) {
	  debug_level = atoi(argv[1]);
	  argc -= 2;
	  argv += 2;
	}
	else
	  error(USAGE);
      }
      else if (strcmp("-socketed", argv[0]) == 0){
	  socketed = TRUE;
	  argc -= 1;
	  argv += 1;
      }
      else if (strcmp("-outfile", argv[0]) == 0) {
	  redirect = TRUE;
	  if (argv[1]) {
	      outfile_name = argv[1];
	      argc -= 2;
	      argv += 2;
	  }
	  else
	      error(USAGE);
      }
      else if (strcmp("-cfg_path", argv[0]) == 0) {
	if (argv[1]) {
	  config_dir = argv[1];
	  argc -= 2;
	  argv += 2;
	}
	else
	  error(USAGE);
      }
      else if (strcmp("-class_file", argv[0]) == 0) {
	if (argv[1]) {
	  class_file = argv[1];
	  argc -= 2;
	  argv += 2;
	}
	else
	  error(USAGE);
    }
      else
	  error(USAGE);
    }

    /* initialize willow file descriptors. If we are running socketed,
     then 0 is socket, passed from inetd. */
    if (socketed)
	to_willow = from_willow = 0;
    else {
	to_willow = FROM_DRIVER_FD;
	from_willow = TO_DRIVER_FD;
    }

    /* Redirect stdout and stderr. Must do this if socketed, otherwise 
     optional. */
    if (socketed || redirect) {
	if (!(outfile = fopen(outfile_name, "a"))) {
	    char err_buf[256];

	    sprintf(err_buf,"Could not open outfile '%s'", outfile_name);
	    error(err_buf);
	}
	close(fileno(stdout));
	close(fileno(stderr));
	dup2(fileno(outfile),fileno(stdout));
	dup2(fileno(outfile),fileno(stderr));
    }

    Client_FD = get_client_fd();

    parse_buf = malloc(PARSE_BUF_SIZE);
    parse_buf[0] = 0;
    parse_buf_size = PARSE_BUF_SIZE;
    state = NO_CON;

    num_fds = 0;	/* initialize */

    /* If a log file indicated then open the file */

    if (logname[0] != EOS)
        {
            log = fopen(logname,"a");
            if (log == NULL)
                fprintf(stderr, "Unable to open the log file %s.\n", logname);
        }
    else
        log = (FILE *) NULL;

    main_loop();

    if (log) fclose(log);
}

/* This is the main loop of the program. We block on select until
  we get input from one of the pipes we are listening on, then
  take appropriate action.
  */
main_loop(void)
{
    while (1) {
#ifdef LATER
	/* Asynchronous operation to willow */
	FD_ZERO(&read_fds);
	if (connected_to_willow)
	{
	    num_fds = MAX(num_fds, (from_willow) + 1);
	    FD_SET(from_willow, &read_fds);
	}
	if (state != NO_CON)
	{
	    num_fds = MAX(num_fds, (from_host[R]) + 1);
	    FD_SET(from_host[R], &read_fds);
	}

	/* Block forever on select */
	if (select(num_fds, &read_fds, NULL, NULL, NULL) < 0) {
#ifdef ANONYMOUS
	    if (errno == EINTR){  /* Interrupt due to license-update timer */
		errno = 0;
		FD_ZERO(&read_fds);
		continue;
	    }
	    else 
		error("error on select");
#else /* ANONYMOUS */
	    error("error on select");
#endif /* ANONYMOUS */
	}
	/* Decide which pipe has input */
	if (FD_ISSET(from_willow, &read_fds)) {
	    read_from_willow();
	}
	else if (FD_ISSET(from_host[R], &read_fds)) {
	    read_from_host();
	}
	else
	    error("select on unknown pipe");
#else /* LATER */
	if (connected_to_willow) {
	    if (debug_level) {
		fprintf(stderr, "Calling do_io_select from main loop -- OPERATE.\n");
		fflush (stderr);
	    }
	    do_io_select(OPERATE);
	} else
	    exit(0);
#endif /* LATER */
    }
}

typedef struct packet_node 
  {
      packet_rec packet;
      struct packet_node *next;
  } packet_node, *queue_ptr;


int get_client_fd()
/* Returns the willow file descriptor */
{
    return(from_willow);
}


/* This routine takes a packet in off the pipe from the client, and
  takes an appropriate action, enqueue or operate on it.  Returns
  CANCEL_SEARCH if a cancel packet is received, USER_OK if a YES or OK
  user response is received, USER_CANCEL if a NO or CANCEL user
  response is received, NOTOK if called with neither OPERATE or ENQUEUE,
  or OK otherwise.
  */

int do_io_select(int operation)
{
    /* externals */
    extern int debug_level;

    SESSION *session;
    packet_rec packet;

    static queue_ptr packet_queue=(queue_ptr) NULL;
    queue_ptr tmp_ptr = (queue_ptr) NULL;
 

    /* On Operates, first flush out any enqueued packets */
    if (operation == OPERATE)
	while (packet_queue != (queue_ptr) NULL) {
	    if (debug_level || echo) {
		fprintf(stderr,"DEQUEUED request, type: %s \n", dump_pac_type(packet_queue->packet.type));
		fprintf(stderr, "packet_queue is %d.\n", packet_queue);
		fprintf(stderr, "Next is %d.\n", packet_queue->next);
		fflush(stderr);
	    }	
	    packet_operate(&packet_queue->packet);
	    if (debug_level || echo) {
		fprintf(stderr, "After operate, next is %d.\n", packet_queue->next);
		fprintf(stderr, "packet_queue is %d.\n", packet_queue);
		fflush(stderr);
	    }	
	    tmp_ptr = packet_queue->next;
	    free(packet_queue);
	    packet_queue = tmp_ptr;
	}

    /* Read in the packet */
    read_from_client(&packet);

    /* Call a handler routine for this packet type */
    if (debug_level || echo) {
	fprintf(stderr,"DRIVER RECEIVED %s\n", dump_pac_type(packet.type));
	fflush(stderr);
    }

    if (operation == ENQUEUE) {
	if (debug_level  || echo) {
	    fprintf(stderr,"ENQUEUED request, type: %s \n", dump_pac_type(packet.type));
	    fflush(stderr);
	}

	if (packet.type == USER) {
	    if (packet.var.user_resp.answer == 1)
		return USER_OK;
	    else /* packet.var.user_resp.answer == 0 */
		return USER_CANCEL;
	}

	if (packet.type == CANCEL) {
	    /* Ignore previous packets? */
	    while (packet_queue != (queue_ptr) NULL) {
		free(packet_queue->packet.data);
		tmp_ptr	= packet_queue->next;
		free(packet_queue);
		packet_queue = tmp_ptr;
	    }
	    return CANCEL_SEARCH;
	}
	else { /* enqueue packet, return */
	    if (packet_queue == (queue_ptr) NULL) {
		packet_queue = (queue_ptr) malloc(sizeof(packet_node));
		packet_queue->packet = packet;
		packet_queue->next = (queue_ptr) NULL;
		if (debug_level) {
		    fprintf(stderr, "After enqueue, new packet_queue is %d.\n", packet_queue);
		    fflush(stderr);
		}
		return OK;
	    }

	    tmp_ptr = packet_queue;

	    while(tmp_ptr->next != (queue_ptr) NULL)
		tmp_ptr = tmp_ptr->next;

	    tmp_ptr->next = (queue_ptr) malloc(sizeof(packet_node));
	    tmp_ptr->next->packet = packet;
	    tmp_ptr->next->next = (queue_ptr) NULL;
	    
	    if (debug_level) {
		fprintf(stderr, "After enqueue, next is %d.\n", packet_queue->next);
		fprintf(stderr, "packet_queue is %d.\n", packet_queue);
		fflush(stderr);
	    }	
	    return OK;
	}
    }
    else if (operation == OPERATE) {
	if (debug_level || echo) {
	    fprintf(stderr,"OPERATE request, type: %s\n", dump_pac_type(packet.type));
	    fflush(stderr);
	}

	/* operate on current packet */
	packet_operate(&packet);

	return OK;
    }
    else {
	return msg_make(DEFAULT_MSG,"Bad operation type in io_select");
    }
}

void read_from_client(packet_rec *p_packet)
{
    extern int Client_FD;

    int  got_count = 0;  /* how many bytes of packet have we gotten */
    int  n = 0;
    
    /* First read in just the header */
    while (got_count < PACKET_NET_SIZE) {
	if (debug_level) {
	    fprintf(stderr,"Reading from client...\n");
	    fflush(stderr);
	}
	n = read(Client_FD, p_packet + got_count, 
		 PACKET_NET_SIZE - got_count);
	if (debug_level) {
	    fprintf(stderr,"Finished reading from client.\n");
	    fflush(stderr);
	}
	if (n == 0) {
	    error("Willow has unexpectedly died");
	}
	if (n < 0) {
	    if (errno == EINTR)
	        errno = 0;
	    else {
	        perror(NULL);
		error("error reading input pipe (header)");
	    }
	}
	if (n > 0)
	  got_count += n;
    }
    denetify(p_packet);

    /* Grab enough storage, and read in rest of packet */
    p_packet->data = (char *) malloc(p_packet->data_len);

    /* Now read until we get the entire packet in */
    got_count = 0;
    while (got_count != (p_packet->data_len)) {
	n = read(Client_FD, p_packet->data+got_count, p_packet->data_len - got_count);
	if (n == 0) {
	    error("Willow has unexpectedly died");
	    return;
	}
	if (n < 0) {
  	    if (errno == EINTR) 
	        errno = 0;
	    else {
 	        perror(NULL);
	        error("error reading input pipe (data)");
	    }
	}
	if (n > 0)
	  got_count += n;
    }

}

char *dump_pac_type (int pac_type)
{
    switch(pac_type) 
    {
	case CONNECT:
	    return ("CONNECT");
	case SEARCH:
	    return("SEARCH");
	case CANCEL:
	    return("CANCEL");
	case TITLE_REQ:
	    return("TITLE_REQ");
	case RECORD_REQ:
	    return("RECORD_REQ");
	case EXT_REQ:
	    return("EXT_REQ");
	case DISCONNECT:
	    return("DISCONNECT");
	case TOGGLE_ECHO:
	    return("TOGGLE_ECHO");
	case CHEAT:
	    return("CHEAT");
	case BROWSE_REQ:
	    return("BROWSE_REQ");
	case USER:
	    return("USER");
	default:
	    return("Unknown");

    }
}

/* Dispatches to handler depending on packet type received */
void packet_operate(packet_rec *p_packet)
{
    static SESSION *session;

    switch (p_packet->type) {
	case CONNECT:
	    /* zclient SELECT command */
	    select_command(&session, p_packet);
	    break;
	case SEARCH:
	    /* zclient FIND command */
	    search_wcb(&session, p_packet);
	    break;
	case CANCEL:
	    search_cancel_wcb(&session, p_packet);
	    break;
	case TITLE_REQ:
	    /* zclient DISPLAY command (for now) */
	    title_req_wcb(&session, p_packet);
	    break;
	case RECORD_REQ:
	    record_req_wcb(&session, p_packet);
	    break;
	case EXT_REQ:
	    ext_req_wcb(&session, p_packet);
	    break;
	case DISCONNECT:
	    /* zclient CLOSE command */
	    disconnect_wcb(&session);
	    break;
	case TOGGLE_ECHO:
	    toggle_echo_wcb(p_packet);
	    break;
	case CHEAT:
	    cheat_wcb(p_packet);
	    break;
#ifndef OLD_LISTS
	case BROWSE_REQ:
	    browse_wcb(&session,p_packet);
	    break;
#endif
	case USER:
	    error_wcb(&session, "not expecting user input now");
	    break;
	default:
	    error_wcb(&session, "unknown packet type");
    }
    free(p_packet->data);

}

/* Send some data down the pipe back to willow. If len is NULL, send 
  strlen() bytes, plus 1 for terminator.
      */
willow_send(void *data, int len)
{
    if (len)
	write(to_willow, data, len);
    else
	write(to_willow, data, strlen(data) + 1);
}

/* Dump the string to wherever we want all the interactions dumped.
  For now, just send it to standard err. 
  */
dump(char *str)
{
    fprintf(stdout, "%s", str);
    fflush(stdout);
}

/*
  * These are the non-host-specific willow-callback routines.
  */

/*------------------------------------------------------------*
 |                    SELECT_COMMAND                          |
 | Function:                                                  |
 |    Select a database and server for searching.             |
*------------------------------------------------------------*/

select_command(p_session, packet)
SESSION **p_session;
packet_rec *packet;
{
    char    *uid = NULL;
    char    *pw;
    char    *db_name;
    char    *title_fields;
    char    *summ_fields;
    char    *ptr;
    char    *machine = NULL;
    char    *port;
    char    *dummy;
    char    *response;
    char    *db_pool;
    int      rc;
    char     err_msg[256];
    static char *serve_list[2];

    packet_rec user_packet;
    SESSION *session;
    RECORD_INFO *info;
    int  code;

    /* we get the login id, password, and db name from willow */
    ptr = packet->data;

    db_name = strdup(ptr);
    ptr += strlen(ptr) + 1;

    title_fields = strdup(ptr);
    ptr += strlen(ptr) + 1;

    summ_fields = strdup(ptr);
    ptr += strlen(ptr) + 1;

    ext_fields = strdup(ptr);
    ptr += strlen(ptr) + 1;

  /* Build environ list */
    env_cnt = 0;
    while ((ptr - packet->data) < packet->data_len) {
	if (env[env_cnt].name)
	    free(env[env_cnt].name);
	env[env_cnt].name = strdup(ptr);
	ptr += strlen(ptr) + 1;

	if (env[env_cnt].val)
	    free(env[env_cnt].val);
	env[env_cnt].val = strdup(ptr);
	ptr += strlen(ptr) + 1;
	env_cnt++;
    }

    if (dummy = env_get("MaxCirc"))
	max_circ =  atoi(dummy);
    if (dummy = env_get("CircTimeout"))
	circ_time = atoi(dummy);
    else
	circ_time = DFLT_CIRC_TIMEOUT;

    if (db_pool = env_get("db_pool")) {
      /* We need to get a license from the server. The profile 
	 portion of a Z39.50 ticket follows this format: 
	   0: z3950 target host name
	   1: z3950 port number
	   2: user id
	   3: password
	   4: copyright information string
	   5: magic cookie/authenticator string/struct.
       */
      if (debug_level) {
	update_hcb("Getting license from server...");
	fprintf(stderr,"Requesting ticket from license server\n");
	fflush(stderr);
      }

      /* Before asking for ticket, if this is a socketed driver we may
	 want to make the sure the client is coming from a place that
	 matches the class it wants to be */
      if (socketed && class_file)
	validate_class(from_willow, env_get("class"), class_file);
    
      /* Check if a license-manager host was specified */
      if (serve_list[0] = env_get("LicenseHost")) {
	serve_list[1] = NULL;
	license_server(serve_list);
      }

      if ((rc = get_ticket(&db_ticket, db_pool, env_get("class"), 
			   NULL)) != LICENSE_SUCCESS) {
	if (rc == LICENSE_FAILURE)
	  sprintf(err_msg, "Too many users already logged in. Please try again later.\nCould not get database ticket");
	else
	  sprintf(err_msg, "Could not get database ticket\nError Code = %s",
		  license_message(rc));
	misc_error_hcb(err_msg);
	force_quit_hcb();
      }
      machine = db_ticket->profile[0];
      port    = db_ticket->profile[1];
      uid = malloc(strlen(db_ticket->profile[2]) + 
		   strlen(db_ticket->profile[3]) + 2);
      sprintf(uid, "%s/%s", db_ticket->profile[2], db_ticket->profile[3]);
      got_info(db_ticket->profile[4]);

      /* Set up an alarm so we can periodically update the ticket.*/
      if (db_ticket->update)
	updater_id = malarm_add(db_ticket->update, updater);
    }

    if (env_get("ManualLogin")) {
	/* use willow user info protocol to get pw */

	if (uid) free (uid);
	send_user_req(MASKED_REQ, "Please enter your authentication (userid/password) for this database", "Authentication", "Ok", "Cancel");

	read_from_client(&user_packet);

	if (debug_level) {
	    fprintf(stderr,"While waiting for user's authentication string, packet type = %s\n",dump_pac_type(user_packet.type));
	    fflush(stderr);
	}

	if (user_packet.type != USER) {
	    misc_error_hcb("User input expected from Willow, other type received");
	    force_quit_hcb();

	} else { /* user response */
	    code = user_resp_wcb(&user_packet,&response);
	    if (code == CANCEL) {
		if (debug_level) {
		    fprintf(stderr,"User canceled login\n");
		    fflush(stderr);
		}
		misc_error_hcb("User canceled authentication request, cannot access database");
		force_quit_hcb();
	    } else { /* code == OK */
		uid = response;
		free(user_packet.data);
	    }
	}

    } 
    else if (env_get("auth_id")) { 
      /* use password from db.conf */
      if (debug_level) {
	fprintf(stderr,"attempting willow user info style login\n");
	fflush(stderr);
      }
      if (uid) free (uid);
      uid = strdup(env_get("auth_id"));
    } /* end- login using password from db.conf */

    if (!uid)
      uid = strdup("");
    
    /* Allow db.conf to override Host and port specified in ticket. Do
       we really  want to do this? */
    if (env_get("HostName")) {
      if (machine)
	free(machine);
      machine = strdup(env_get("HostName"));
    }
    if (!machine) {
      misc_error_hcb("No host address specified for this database");
      force_quit_hcb();
    }

    if (env_get("Z39.50Port")) {
      if (port)
	free(port);
      port = strdup(env_get("Z39.50Port"));
    }
    else
      port = strdup("210");


    if (env_get("preferred_syntax"))
      w_preferred_syntax = strdup(env_get("preferred_syntax"));

    update_hcb("Establishing network connection to database server...");

    code = z39_initial(p_session, db_name, machine, port, "", w_preferred_syntax, uid);

    if (echo) {
	fprintf(stderr, "Z39.50_init -- db: %s, machine: %s,  uid: %s.\n", db_name, machine, uid);
	fprintf(stderr, "Returns: %d.\n", code);
	fflush(stderr);
    }

    if (code) {
	misc_error_hcb("Unable to establish connection.");
	force_quit_hcb();
    }
    else {
	connected_hcb();
	tags_into_table(summ_fields, summary_formats, &summary_field_cnt);
	tags_into_table(title_fields, title_formats, &title_field_cnt);
    }

    /* insert resource_control function into session */
    (*p_session)->resource_control = user_resource_control_confirm;

}

/* Called by malarm() when it is time to update the ticket */
static void updater(void)
{
  int      rc;
  char     err_msg[256];

  if ((rc = update_ticket(db_ticket)) != LICENSE_SUCCESS) {
    sprintf(err_msg, "License manager unable to update database ticket\nError Code = %s", license_message(rc));
    misc_error_hcb(err_msg);
    force_quit_hcb();
  }
  updater_id = malarm_add(db_ticket->update, updater);
}

/* Called when we get an echo toggle packet */
toggle_echo_wcb(packet)
packet_rec *packet;
{
    echo = !echo;
}

search_cancel_wcb(p_session, packet)
SESSION **p_session;
packet_rec *packet;
{
    /* What do we do if we're still waiting for the search? */
    search_error_hcb("Search Canceled.");
}

/*
 * For now, this holds the Z39.50 display command.
 *
 * Note:  This routine always assumes you're going to ask
 * for a summary record of higher index than the last one
 * you asked for. */

title_req_wcb(p_session, packet)
SESSION **p_session;
packet_rec *packet;
{
    static RECORD_INFO *info;
    char setname[BUFSIZ], elemname[BUFSIZ];
    int  position, display, code, index, status;
    static int title_total;

    setname[0] = '\0';
    elemname[0] = '\0';
    position = packet->var.title_req.title_num;
    display = DISPLAY_NUMBER;

    if (position == 1) /* First time through */
	title_total = 0;

    if (debug_level) {
	fprintf (stderr, "Title Request #%d.\n", position);
	fflush (stderr);
    }

    if (position >= title_total) {
	if (info != NULL) recinfo_dispose (&info);
	info = recinfo_new();

	code = z39_present(*p_session, setname, elemname, w_preferred_syntax, position, display, info);

	if (echo) {
	    fprintf(stderr, "Z39.50_present -- setname: %s, position: %d, count: %d.\n", setname, position, display);
	    fprintf(stderr, "Returns: %d.\n", code);
	    fflush(stderr);
	}

	if (code == NOTOK ) {
	    misc_warning_hcb("z39_present returned NOTOK status.");
	    title_total = 0;
	    send_title(position, "BAD RECORD -- No title");
	    return;
	}
	status = recinfo_get_status(info);

	if (echo) {
	    fprintf(stderr, "Returns status: %d.\n", status);
	    fflush(stderr);
	}

	if (status == 5) {
	    msg_make(DEFAULT_MSG, "Present returned FAILURE status");
	}
	if (hcb_record_msg (info) == ARE_MSGS) {
	    title_total = 0;
	    misc_warning_hcb("Error diagnostics returned:");
	    send_title(position, "BAD RECORD -- No title");
	    return;
	} else {
	    title_total = recinfo_get_position(info) + recinfo_get_returned(info);
	}
    }
    wcb_record_title(info, position);
}


record_req_wcb(p_session, packet)
SESSION **p_session;
packet_rec *packet;
{
    RECORD_INFO *info;
    char setname[BUFSIZ], elemname[BUFSIZ];
    int  position;
    int  display;
    int  code;
    int status;

    setname[0] = '\0';
    elemname[0] = 'F';
    elemname[1] = '\0';
    position = packet->var.record_req.ref_num;

    display = 1;    /* getting a single full record */

    if (debug_level) {
	fprintf (stderr, "Record %s Request #%d.\n", ((packet->var.record_req.summary) ? "Summary" : ""), position);
	fflush (stderr);
    }

    info = recinfo_new();

    if (packet->var.record_req.summary && (summary_field_cnt == 0) &&
	strcmp(w_preferred_syntax, "SUTR") == 0)
	strcpy(elemname, "B");

    code = z39_present(*p_session, setname, elemname, w_preferred_syntax, position, display, info);

    if (echo) {
	fprintf(stderr, "Z39.50_present -- setname: %s, position: %d, count: %d.\n", setname, position, display);
	fflush(stderr);
    }

    if (code == NOTOK)
	misc_warning_hcb("z39_present returned NOTOK status.");
    else
    {
	status = recinfo_get_status(info);

	if (echo) {
	    fprintf(stderr, "Returns status: %d.\n", status);
	    fflush(stderr);
	}

	make_msg_single_rec_present(status);

	if (hcb_record_msg (info) == ARE_MSGS) {
	    misc_warning_hcb("Error diagnostics returned:");
	    send_record(position, "FAILED PRESENT\n No Record.\n",
			packet->var.record_req.summary);
	} else 
	    wcb_record(info, position, packet->var.record_req.summary, packet->var.record_req.abbreviated);
    }

    recinfo_dispose(&info);
}

/* Strain out unwelcome characters */
char strain(ch)
     char ch;
{
  int i;

  if (isalpha(ch) || isdigit(ch))
    return(ch);
  
  for (i = 0; i < sizeof(valid_extra_chars); i++)
    if (ch == valid_extra_chars[i])
      return(ch);
  
#ifdef LATER
  /* If we are in "Any Words" field, leave periods for now, because they 
     could be field specifiers. */
  if (!cur->command[0])
    if (ch == '.')
      return(ch);
#endif /* LATER */

  return((char) -1);
}


#define ANDSTR " and ("

search_wcb(p_session, packet)
SESSION **p_session;
packet_rec *packet;
{
    search_desc_rec *limit_fields = NULL,
    *last_field = NULL;

    char *data_ptr, *find, *tfield, *cfield;
    int i, sdex, tdex, find_len, token_len, rfield;
    static int set_number = 0;
    RECORD_INFO *info;
    char database[BUFSIZ];
    char setname[BUFSIZ];
    int  ch, code, diag_status;
    int close_and, empty_term;

    database[0] = '\0';
    setname[0] = '\0';

    data_ptr = packet->data;
    find = malloc(1);
    find[0] = '\0';
    find_len = 1;
    tdex = 0;

    /* Simple search string creation */
    /* Append command if any.  Append text if any. */
    /* TODO: deal with parens. */

    for (i = 0; i < packet->var.search.num_text_fields; i++) {
	if (*data_ptr <= DAY)
	    rfield = (int) (*data_ptr++);
	else
	    rfield = NONE;

	tfield = data_ptr;  /* text field */
	data_ptr += (strlen(data_ptr) + 1);
	cfield = data_ptr;  /* Command field */
	data_ptr += (strlen(data_ptr) + 1);

	/* For now IGNORE ranges */
	if (rfield != NONE) continue;

	/* test for term consisting of only whitespace */
	/* We expect token_len NEVER to be 0 */
	empty_term = TRUE;
	token_len = strlen(tfield);
	for (sdex=0; sdex < token_len; sdex++) {
	    if (!isspace(tfield[sdex])) {
		empty_term = FALSE;
		break;
	    }
	}

	if (empty_term) continue;

	close_and = FALSE;
	/* If there was a clause, append an  " and (" before
	 the next clause */
	if (find[0] != '\0') {
	    find_len += strlen(ANDSTR);
	    find = realloc (find, find_len);
	    strcat(find, ANDSTR);
	    tdex += strlen(ANDSTR);
	    close_and = TRUE;
	}

	/* For now, do not realloc find for case tdex<sdex */
	token_len = strlen(cfield);
	if (token_len > 0) {
	    find_len +=	token_len + 1;	/* for the space */
	    find = realloc(find, find_len);
	    for (sdex=0; sdex <= token_len; sdex++) {
		ch = cfield[sdex];
		if (ch == '\0') {
		    find[tdex++] = ' ';
		    find[tdex] = '\0';
		}
		else
		    find[tdex++] = ch;
	    }
	}
	token_len = strlen(tfield);
	find_len += token_len + (close_and ? 2: 0);
	find = realloc(find, find_len);
	for (sdex=0; sdex <= token_len; sdex++) {
	    ch = tfield[sdex];
	    if (ch == '\0')
		find[tdex] = '\0';
	    else
		find[tdex++] = ch;
	}
	if (close_and) {
	    find[tdex++] = ' ';
	    find[tdex++] = ')';
	    find[tdex] = '\0';
	}
    }

    for (i = 0; i < packet->var.search.num_limit_fields; i++) {
	close_and = FALSE;
	/* If there was a clause, append an  " and " before
	 the next clause */
	if (find[0] != '\0') {
	    find_len += strlen(ANDSTR);
	    find = realloc (find, find_len);
	    strcat(find, ANDSTR);
	    tdex += strlen(ANDSTR);
	    close_and = TRUE;
	}
	tfield = data_ptr;  /* Command and text field */
	data_ptr += (strlen(data_ptr) + 1);
	token_len = strlen(tfield);

	find_len += token_len + (close_and ? 2: 0);
	find = realloc(find, find_len);
	for (sdex=0; sdex <= token_len; sdex++) {
	    ch = tfield[sdex];
	    if (ch == '\0')
		find[tdex] = '\0';
	    else
		find[tdex++] = ch;
	}
	if (close_and) {
	    find[tdex++] = ' ';
	    find[tdex++] = ')';
	    find[tdex] = '\0';
	}

    }

    if (debug_level) dump (find);

    info = recinfo_new();

    code = z39_search(*p_session, find, database, setname, w_preferred_syntax, info);

    free(find);

    diag_status=hcb_record_msg(info);


    if (code == NOTOK)
	search_error_hcb("z39_search returned NOTOK status.");
    else
	if (diag_status==ARE_MSGS)
	    search_error_hcb("Error diagnostics returned:");
	else
	{
	    int hits = recinfo_get_result(info);

	    if (hits == -1) { /* Bug in Newton */
		search_error_hcb("Internal error on what should have been a valid search.");
	    } else {
		g_search_num++;
		send_stats_cmd(NULL, hits, g_search_num);
	    }
	}

    recinfo_dispose(&info);
}

/* This will later pass back a copyright notice or
 Some helpful string from the DB connection.
*/
void get_info_wcb(SESSION **p_session, packet_rec *ignored)
{
  packet_rec packet;

  char *info = "When the Z39.50 EXPLAIN query gets implemented,\nthis space will tell you about the database you have just connected to.";
  packet.type = INFO;
  packet.data_len = strlen(info) + 1;
  packet.data = NULL;
  willow_send(netify(&packet), PACKET_NET_SIZE);
  willow_send(info, packet.data_len);
}

cheat_wcb(packet)
packet_rec *packet;
{
    update_hcb("You cannot cheat under Z39.50.");
}

got_info(char *info)
{
  packet_rec packet;

  if (!info)
    info = NO_INFO_STR;
  else if (!info[0])
    info = NO_INFO_STR;
  packet.type = INFO;
  packet.var.info.type = COPYRIGHT_INFO;
  packet.data_len = strlen(info) + 1;
  packet.data = NULL;
  willow_send(netify(&packet), PACKET_NET_SIZE);
  willow_send(info, packet.data_len);
}


/*
  * These are the non-host-specific host-callback routines.
  */

/* Called from the state table when we have finished connecting */
connected_hcb()
{
    packet_rec packet;

    packet.type = CONNECTED;
    packet.data_len = 0;
    packet.data = NULL;
    willow_send(netify(&packet), PACKET_NET_SIZE);
}


misc_error_hcb(msg)
char *msg;
{
    packet_rec packet;
    char *whole_msg;

    whole_msg=msg_hcb(msg);
    packet.type = ERROR;
    packet.data_len = strlen(whole_msg) + 1;
    packet.data = NULL;

    packet.var.error.type = MISC;
    willow_send(netify(&packet), PACKET_NET_SIZE);
    willow_send(whole_msg, packet.data_len);
}

misc_warning_hcb(char *msg)
{
  packet_rec packet;
  char *whole_msg;

  whole_msg=msg_hcb(msg);
  packet.type = ERROR;
  packet.data_len = strlen(whole_msg) + 1;
  packet.data = NULL;
  
  packet.var.error.type = WARNING;
  willow_send(netify(&packet), PACKET_NET_SIZE);
  willow_send(whole_msg, packet.data_len);
}

search_error_hcb(msg)
char *msg;
{
    packet_rec packet;
    char *whole_msg;

    whole_msg=msg_hcb(msg);
    packet.type = ERROR;
    packet.data_len = strlen(whole_msg) + 1;
    packet.data = NULL;

    packet.var.error.type = BAD_SEARCH;
    willow_send(netify(&packet), PACKET_NET_SIZE);
    willow_send(whole_msg, packet.data_len);
}

update_hcb(msg)
char *msg;
{
    packet_rec packet;

    packet.type = ERROR;
    packet.data_len = strlen(msg) + 1;
    packet.data = NULL;

    packet.var.error.type = UPDATE;
    willow_send(netify(&packet), PACKET_NET_SIZE);
    willow_send(msg, packet.data_len);
}

bad_search_hcb()
{
    packet_rec packet;

    packet.type = ERROR;
    packet.data_len = 0;
    packet.data = NULL;

    packet.var.error.type = BAD_SEARCH;
    willow_send(netify(&packet), PACKET_NET_SIZE);
}

force_quit_hcb()
{
    cleanup();
}


/* Close things up, and quit */
cleanup()
{
  /* We are out of here */
  exit(0);
}

/* close off our connection to Willow, we are on our own */
close_willow()
{
    connected_to_willow = FALSE;
    close(from_willow);
    if (!socketed)  /* don't want to close socket twice */
	close(to_willow);

    if (!echo) {
	close(fileno(stdout));
	close(fileno(stderr));
	close(fileno(stdin));
    }
}

disconnect_wcb(p_session)
SESSION **p_session;
{
    int code;

    code = z39_close(p_session, NULL);

    close_willow();

    if (db_ticket) {
      malarm_cancel(updater_id);
      return_ticket(db_ticket);
      db_ticket = NULL;
    }
}


/* Send the stats back to willow */
send_stats_cmd(stats, num_results, search_num)
char *stats;
int   num_results;
int   search_num;
{
    packet_rec packet;

    packet.type = STATS;
    packet.data = NULL;
    packet.var.stats.result_id = search_num; 
    packet.var.stats.num_results = num_results;

    if (stats == NULL) {
	packet.data_len = 0;
	willow_send(netify(&packet), PACKET_NET_SIZE);
    } else {
	packet.data_len = strlen(stats) + 1;
	willow_send(netify(&packet), PACKET_NET_SIZE);
	willow_send(stats, packet.data_len);
    }
}

/* Send a summary back to willow */
void send_title(int num, char *title)
{
    packet_rec packet;

    packet.type = TITLE;
    packet.var.title.title_num = num;
    packet.data = NULL;
    packet.data_len = strlen(title) + 1;
    willow_send(netify(&packet), PACKET_NET_SIZE);
    willow_send(title, NULL);
}

/* Send back an extension record */
void send_ext(int num, char *text)
{
  packet_rec packet;

  packet.type = EXTS;
  packet.data = NULL;
  packet.var.exts.ext_num = num;
  packet.data_len = strlen(text) + 1;
  willow_send(netify(&packet), PACKET_NET_SIZE);
  willow_send(text, NULL);
}

/* Called when willow asks us for an extensions record */

ext_req_wcb(p_session, packet)
SESSION **p_session;
packet_rec *packet;
{
    RECORD_INFO *info;
    char setname[BUFSIZ], elemname[BUFSIZ];
    int  position;
    int  display;
    int  code, status;
    char *extraction;

    setname[0] = '\0';
    elemname[0] = '\0';
    position = packet->var.record_req.ref_num;

    display = 1;    /* getting a single full record */

    if (debug_level) {
	fprintf (stderr, "Ext Record %s Request #%d.\n", ((packet->var.record_req.summary) ? "Summary" : ""), position);
	fflush (stderr);
    }

    info = recinfo_new();

    code = z39_present(*p_session, setname, elemname, w_preferred_syntax, position, display, info);

    if (echo) {
	fprintf(stderr, "Ext: Z39.50_present -- setname: %s, position: %d, count: %d.\n", setname, position, display);
	fprintf(stderr, "Returns: %d.\n", code);
	fflush(stderr);
    }

    if (code == NOTOK) {
	misc_warning_hcb("z39_present returned NOTOK status.");
	send_ext(packet->var.record_req.ref_num, "");
    } else {
	status = recinfo_get_status(info);

	if (echo) {
	    fprintf(stderr, "Returns status: %d.\n", status);
	    fflush(stderr);
	}

	make_msg_single_rec_present(status);

	if (hcb_record_msg(info)==ARE_MSGS) {
	    misc_warning_hcb("Error diagnostics returned:");
	    send_ext(packet->var.record_req.ref_num, "");
	} else
	    if (((code = wcb_record_extract(info, position, ext_fields, &extraction)) == OK)
		&& extraction) {
		send_ext(packet->var.record_req.ref_num, 
			 extraction);
		free(extraction);
	    } else  {
		misc_warning_hcb("Extraction of Article view path failed.");
		send_ext(packet->var.record_req.ref_num, "");
	    }
    }

    recinfo_dispose(&info);
}


#ifdef NEVER
/* Send back a short record */
void send_summ_record(int num, char abbreviated)
{
  packet_rec packet;
  char *body;

  packet.type = W_RECORD;
  packet.var.record.num_hilites = 0;
  packet.data = NULL;
  packet.var.record.ref_num = num;
  
  if (abbreviated)
    body = summ_tab[num].summ;
  else {
    if (!summ_tab[num].expanded_summ)
      summ_tab[num].expanded_summ = expand_summ(summ_tab[num].summ);
    body = summ_tab[num].expanded_summ;
  }
  packet.data_len = strlen(body) + 1;

  if (do_circ) {
    if (abbreviated) {
      if (!summ_tab[num].circ) 
	summ_tab[num].circ = get_circ(summ_tab[num].invis, max_circ, 
				      abbreviated, ABBREV_INDENT);
      packet.data_len += strlen(summ_tab[num].circ);
    }
    else {
      if (!summ_tab[num].expanded_circ) 
	summ_tab[num].expanded_circ = get_circ(summ_tab[num].invis, max_circ,
				        abbreviated, summary_long_len);
      packet.data_len += strlen(summ_tab[num].expanded_circ);
    }
  }
  
  willow_send(netify(&packet), PACKET_NET_SIZE);
  if (do_circ) {
    willow_send(body, strlen(body));
    if (abbreviated)
      willow_send(summ_tab[num].circ, NULL);
    else
      willow_send(summ_tab[num].expanded_circ, NULL);
  }
  else
    willow_send(body, NULL);
}
#endif

/* Send back full text retrieval */
/* Note: we'll have to add back in the circ hacks */
void send_record(int num, char *text, int summ_flag)
{
    packet_rec packet;
    hl_rec    *hl, *hl_prev;
    netint     hl_buf;

    packet.type = W_RECORD;
    packet.var.record.summary = summ_flag;
    packet.var.record.num_hilites = num_hilites;
    packet.data = NULL;
    packet.var.record.ref_num = num;
    packet.data_len = strlen(text) + 1 + 
      packet.var.record.num_hilites * 2 * sizeof(int);

    willow_send(netify(&packet), PACKET_NET_SIZE);

    /* Send out the hi-lite information, and destroy the list */
    hl = hl_head;
    while (hl) {
      hl_buf = htonl(hl->start);
      willow_send(&hl_buf, sizeof(netint));
      hl_buf = htonl(hl->end);
      willow_send(&hl_buf,   sizeof(netint));
      hl_prev = hl; 
      hl = hl->next;
      free(hl_prev);
    }
    num_hilites = 0;
    hl_head = hl_tail = NULL;

    willow_send(text, NULL);
}

/* Miscellaneous */
error_wcb(SESSION **p_session, char *msg)
{
    fprintf(stderr, "ERROR IN DRIVER: %s\n",msg);fflush(stderr);
    disconnect_wcb(p_session);
}

error(msg)
char *msg;
{
    fprintf(stderr, "FATAL ERROR IN DRIVER: %s\n",msg);fflush(stderr);
    exit(-1);
}

/* Set up an alarm handler */
set_alarm(time_out)
int time_out;
{
    signal(SIGALRM, force_quit_hcb);
    alarm(time_out);
}

/* append a carriage return to given malloced string */
char *append_cr(s)
char *s;
{
    char *new;

    new = malloc(strlen(s) + 2);
    strcpy(new, s);
    strcat(new, "\n");
    free(s);
    return(new);
}

/* Create a node in our linked list of hilite points, then record the
  starting offset for it. */
hl_start(point)
int point;
{
    hl_rec *hl;

    num_hilites++;
    hl = (hl_rec *) malloc(sizeof(hl_rec));
    hl->start = point;
    hl->end = -1;
    hl->next = NULL;

    if (!hl_head)
	hl_head = hl_tail = hl;
    else {
	hl_tail->next = hl;
	hl_tail = hl;
    }
}

/* Record the end-offset of a hilite region, check for errors to avoid
    crashing because of bad data, but don't bother reporting them. */
hl_end(point)
int point;
{
    if (hl_tail)
	if (hl_tail->end == -1)
	    hl_tail->end = point;
}

char *env_get(char *s)
{
  int i;
  
  for (i = 0; i < env_cnt; i++)
    if (strcmp(s, env[i].name) == 0)
      return(env[i].val);
  return(NULL);
}


/* Send a user request packet */
void send_user_req(int type, char *query_text, char *query_tag, char *yes_label, char *no_label)
{
  packet_rec packet;

  packet.type = USER_REQ;
  packet.var.user_req.type = type;
  packet.data = NULL;
  packet.data_len = strlen(query_text) + 1 +
                    strlen(query_tag) + 1 +
		    strlen(yes_label) + 1 +
		    strlen(no_label) + 1;
  willow_send(netify(&packet), PACKET_NET_SIZE);
  willow_send(query_text, NULL);
  willow_send(query_tag, NULL);
  willow_send(yes_label, NULL);
  willow_send(no_label, NULL);
}




int user_resp_wcb(packet_rec *packet, char **p_response)
/* respond to a user response 
   returns OK if the user hits OK(or YES), 
   CANCEL if the user hits cancel(or NO) */
{
  
  if (!packet->var.user_resp.answer) {
      return CANCEL;
  }

  /* Copy user text out of packet and into response */
  if (packet->data_len != 0) {
      *p_response = malloc(strlen(packet->data) + 1);
      strcpy(*p_response,packet->data);
  } else
      *p_response = (char *) NULL;

  return OK;

}


void user_resource_control_confirm(SESSION *session, long report[MAX_RESOURCE_REPORT_BIB1], int *p_complete, STRING *message)

{
    char line[255];
    char *usermsg = NULL;
    int count;
    packet_rec user_packet;

    /* put report information if any in usermsg */
    usermsg = malloc(2048);
    *usermsg = NULL;

    sprintf(usermsg,"The server has issued a resource control request, asking whether\nyou would like to continue the operation.\n");

    if (report[0] < 0)
	strcat(usermsg,"There is no resource report information available.\n");
    else {
	strcat(usermsg,"The following resource report information has been returned:\n");
	for (count = 1; count < MAX_RESOURCE_REPORT_BIB1; count++)
	{
	    if (report[count] >= 0) {
		sprintf(line,"%s : %d\n",
			resource_report_bib1_msg[count],
			report[count]);
		strcat(usermsg,line);
	    }
	}
    }

    if (message && message->length) {
	sprintf(line,"%.*s\n", message->length, message->location);
	strcat(usermsg,line);
    }

    /* ask user using this information */
    if (debug_level) {
	fprintf(stderr,"Sending resource_control user request, message = \n%s\n",usermsg);
	fflush(stderr);
    }

    strcat(usermsg,"\nDo you want to continue the current operation?\n");
    send_user_req(BOOL_REQ, usermsg, "Continue operation?", "Yes", "No");
    free(usermsg);

    count = do_io_select(ENQUEUE);

    if (debug_level) {
	fprintf(stderr,"While waiting for user's resource control response, do_io_select(ENQUEUE) returns %d\n",count);
	fflush(stderr);
    }

    if (count == USER_OK) {
	*p_complete = TRUE;
    } else if (count == USER_CANCEL) {
	misc_warning_hcb("Canceling operation.");
	sleep(1);
	*p_complete = FALSE;
    } else { /* other packet type received (and enqueued) */
	msg_make(DEFAULT_MSG,"You ignored the question.  Response will be taken as No.");
	*p_complete = FALSE;
    }
}
