/***********************************************************************
 *                                                                     *
 *              WILLOW(tm) Information Retrieval Software              *
 *     (Washington Information Looker-upper Layered Over Windows)      *
 *             Copyright 1992-1994 University of Washington            *
 *                                                                     *
 *   Willow is a trademark of the University of Washington.  No        *
 *   commercial use of this trademark or software may be made without  *
 *   prior written permission of the University of Washington.         *
 *                                                                     *
 * Permission to use, copy, modify, and distribute Willow(tm) software *
 * and its documentation for any noncommercial purpose is hereby       *
 * granted, provided that the above copyright notice appears in all    *
 * copies and that both the above copyright notice and this permission *
 * notice appear in supporting documentation, and that the name of the *
 * University of Washington not be used in advertising or publicity    *
 * pertaining to distribution of the software without specific,        *
 * written prior permission.  This software is made available "as is", *
 * and THE UNIVERSITY OF WASHINGTON DISCLAIMS ALL WARRANTIES, EXPRESS  *
 * OR IMPLIED, WITH REGARD TO THIS SOFTWARE, INCLUDING WITHOUT         *
 * LIMITATION ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS    *
 * FOR A PARTICULAR PURPOSE, AND IN NO EVENT SHALL THE UNIVERSITY OF   *
 * WASHINGTON BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL     *
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA  *
 * OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, TORT (INCLUDING       *
 * NEGLIGENCE) OR STRICT LIABILITY, ARISING OUT OF OR IN CONNECTION    *
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.                       *
 *                                                                     *
 * For further information please contact:                             *
 * Willow Project                                                      *
 * Information Systems                                                 *
 * University of Washington                                            *
 * 4545 15th NE, 1st floor                                             *
 * Seattle, Washington, 98105, USA                                     *
 * Internet: willow@cac.washington.edu                                 *
 ***********************************************************************/

/************************************************
 * driver.h
 * Contains definitions used by all drivers. Host
 * specific defs should be somewhere else.
 ************************************************/

#define FALSE 0
#define TRUE  1
#define MAX_LINE 256
#define MAX_TITLE_LEN 256
#define PARSE_BUF_SIZE 8192 
#define IN_BUF_SIZE 2048
#define NO_INFO_STR     "NO DATABASE INFORMATION AVAILABLE"

/* Range types */
#define NONE    0
#define NUMERIC 1
#define YEAR    2
#define MONTH   3
#define DAY     4

/* Define a record type for keeping state/string info in */
typedef struct _transition_rec {
    char    *in_string;
    char    *out_string;
    int      next_state;
    int     (*action)();
} transition_rec;

/* Define a two dimensional array to describe a state table. Each state
   can have up to 6 inputs */
#define MAX_TRANSITIONS 7
#define MAX_STATES      50
typedef transition_rec state_table[MAX_STATES][MAX_TRANSITIONS];

/* Define a structure for building linked lists of search elements. */
typedef struct _search_desc_rec {  /* describe a single field */
  int   range_type;
  char *text;     /* Not used for limit searches */
  char *command;
  struct _search_desc_rec *next;
} search_desc_rec;
 
/* Entry in our table of summaries */
typedef struct _summ_rec { 
  char *title;
  char *summ;
  char *circ;
  char *expanded_summ;
  char *expanded_circ;
  char *invis;
} summ_tab_entry;

/* A type for environment list */
#define MAX_ENV 64
typedef struct _env_rec {
  char *name;
  char *val;
} env_rec;

