/*
 *  Copyright (C) 1994 by the Massachusetts Institute of Technology
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice appear in all
 * copies and that both that copyright notice and this permission
 * notice appear in supporting documentation, and that the name of
 * M.I.T. not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior
 * permission.  M.I.T. makes no representations about the suitability
 * of this software for any purpose.  It is provided "as is" without
 * express or implied warranty.
 */

/* Author: Tom Owens MIT Library Systems Office */

/* MARC leader positions, etc */

/* leader data positions*/
#define RECLEN_OFFSET 0
#define STATUS_OFFSET 5
#define TYPE_OFFSET 6
#define LEVEL_OFFSET 7
#define BASEADR_OFFSET 12
#define ENCOD_OFFSET 17
#define DCF_OFFSET 18
#define LINKING_OFFSET 19

#define	DIRECTORY_OFFSET 24	/* position of the directory */
#define	DIRECTORY_LENGTH 12	/* length of a directory entry */
#define	RECORD_LENGTH_LENGTH 5	/* length of the record length field */
#define BASEADR_LENGTH 5
#define DIRECTORY_TAG_LENGTH 3
#define DIRECTORY_LENGTH_LENGTH 4
#define DIRECTORY_OFFSET_LENGTH 5

/* special MARC characters */

#define DELIMITER '\037'
#define END_OF_FIELD '\036'
#define END_OF_RECORD '\035'

/* miscellaneous */


#define	MAX_MARC_SIZE	65000	/* we don't expect records larger than 20K */
#define	MIN_MARC_SIZE	36	/* leader plus one directory entry */
#define MAX_TAG_NUMBER  999     /* highest tag number possible */

/*  errors */
#define GOOD 0
#define TAG_NOT_FOUND -1
#define OCCURRENCE_NOT_FOUND -2
#define SUBFIELD_NOT_FOUND -3
#define MEMORY_ERROR -4
#define BAD_MARC_SIZE -5
#define TOO_MANY_TAGS -6
#define BAD_OCCURRENCE_REQUESTED -7
#define BAD_TAG_DIRECTORY -8
#define BAD_TAG_REQUESTED -9
#define BAD_TAG_DATA -10

#define UNKNOWN_ERROR -9000

typedef struct tagMapEntry {
    int tag;
    int tagLength;
    int	tagOffset;
} TAG_MAP_ENTRY;

/* the structure of our dreams */

typedef struct marcRec {
  int length;	    /* leader data */
  char status;
  char type;
  char level;
  char encod;
  char dcf;
  char linking;
  TAG_MAP_ENTRY *tagMap;
  char *data;	    /* the actual data of the tags */
  int tagOccurrences[MAX_TAG_NUMBER+1]; /* total number of tags -- by tag number (offset 0=total in record] */
  int lastTag;	    /* last tag read by mrFindTagNam */
  int lastTagOccurrence;    /* occurrence of last tag read by mrFindTagNam */
  int lastTagPosition;    /* field number of last tag read by mrfindTagNam */
} MARC_RECORD;

/* prototypes */

#ifdef _STDC_
int mrInit (MARC_RECORD *marc);
#else
int mrInit();
#endif

#ifdef _STDC_
int mrMapMarc (char *rawRec, MARC_RECORD *marc);
#else
int mrMapMarc();
#endif

#ifdef _STDC_
int mrGetField (MARC_RECORD *marc, int fieldRequested, int *tagFound, char **tagDataOffset, int *tagDataLength);
#else
int mrGetField();
#endif

#ifdef _STDC_
int mrGetSubField (int requestedOccurrence, char *tagData, const int tagLength, char *subFieldName, char **subFieldData, int *subFieldLength);
#else 
int mrGetSubField();
#endif

#ifdef _STDC_
int mrFindTagName (MARC_RECORD *marc, const int tagRequested, int requestedOccurrence, char **tagDataOffset, int *tagDataLength );
#else
int mrFindTagName();
#endif

#ifdef _STDC_
int mrFindSubFieldName(const char subFieldRequested, int requestedOccurrence, char *tagData, const int tagDataLength, char **subFieldOffset, int *subFieldLength);
#else
int mrFindSubFieldName();
#endif

#ifdef _STDC_
void mrFree(MARC_RECORD *marc);
#else
void mrFree();
#endif
