/*
 *  Copyright (C) 1994 by the Massachusetts Institute of Technology
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice appear in all
 * copies and that both that copyright notice and this permission
 * notice appear in supporting documentation, and that the name of
 * M.I.T. not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior
 * permission.  M.I.T. makes no representations about the suitability
 * of this software for any purpose.  It is provided "as is" without
 * express or implied warranty.
 */

/* Author: James Chang MIT Information Systems */

/* msg_hcb.c
Routines for retrieving accumulated message strings from msg_make calls.  */


#include <stdio.h>
#include <string.h>
#include <constant.h>
#include <recinfo.h>
#include <record.h>
#include <series.h>
#include <msg.h>

void rev_msg(EMSG msg);  /* Function Prototypes */
char *msg_hcb(char *msg);
int hcb_record_msg(RECORD_INFO *info);

extern EMSG Messages; /* Message stack used by msg_make */


static char *Accum=(char *) NULL; /* Use to accumulate messages */


void rev_msg(EMSG msg)
/* Puts message stack contents with newlines in correct(FIFO) order
   in global Accum, a dynamically allocated string.  Deallocates
   message stack elements(msg and its children). */

{   
    if (msg != (EMSG) NULL) {
	rev_msg(msg->next);
	Accum=(char *) realloc(Accum,strlen(Accum)+strlen(msg->message)+2);
	Accum=strcat(Accum,msg->message);
	sprintf(Accum,"%s\n",Accum);
	free(msg);
    }
}


char *msg_hcb(char *msg)
/* Returns the concatenation of msg and the accumulated messages
   from msg_make.  Clears message stack pointer(Messages). */

{
    if (Accum!=(char *) NULL)
	free(Accum);
    Accum=(char *) malloc(strlen(msg) + 2);
    sprintf(Accum,"%s\n",msg);
    if (Messages!=(EMSG) NULL) {
	rev_msg(Messages);
	Messages=(EMSG) NULL;
    }
    return(Accum);
}

int hcb_record_msg(RECORD_INFO *info)
/*puts any diagnostics in info into the accumulated messages stack.
  Returns ARE_MSGS if there are diagnostic messages, NO_MSGS if none. */
{
    /* Diagnostic messages from Z39.50 protocol */
    static char *diagnostic_msgs[] = { 
        "Unknown diagnostic message", /* 0 not used */
        "Permanent system error",
        "Temporary system error",
        "Unsupported search",
        "Terms only exclusion (stop) words",
        "Too many argument words",
        "Too many boolean operators",
        "Too many truncated words",
        "Too many incomplete subfields",
        "Truncated words too short",
        "Invalid format for record number (search term)",
        "Too many characters in search statement",
        "Too many records retrieved",
        "Present request out-of-range",
        "System error in presenting records",
        "Record not authorized to be sent intersystem",
        "Record exceeds Preferred-message-size",
        "Record exceeds Maximum-record-size",
        "Result set not supported as a search term",
        "Only SINGLE result set as search term supported",
        "Only AND-ing of a SINGLE result set as search term supported",
        "Result set exists and replace indicator off",
        "Result set naming not supported",
        "Combination of specified databases not supported",
        "Element set names not supported",
        "Specified element set name not valid for specified database",
        "Only a single element set name supported",
        "Result set no longer exists - unilaterally deleted by target",
        "Result set is in use",
        "One of the specified databases is locked",
        "Specified result set does not exist",
        "Resources exhausted - no results available",
        "Resources exhausted - unpredictable partial results available",
        "Resources exhausted - valid subset of results available",
        "Unknown diagnostic message", /* 34-99 currently not used */
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unspecified error",
        "Access-control failure",
        "Security challenge required but could not be issued - request terminated",
        "Security challenge required but could not be issued - record not included",
        "Security challenge failed - record not included",
        "Terminated by negative continue response from client",
        "No abstract syntaxes agreed to for this record",
        "Query type not supported",
        "Malformed query",
        "Database unavailable",
        "Operator unsupported",
        "Too many databases specified",
        "Too many result sets created",
        "Unsupported attribute type",
        "Unsupported Use attribute",
        "Unsupported value for Use attribute",
        "Use attribute required but not supported",
        "Unsupported Relation attribute",
        "Unsupported Structure attribute",
        "Unsupported Position attribute",
        "Unsupported Truncation attribute",
        "Unsupported Attribute set",
        "Unsupported Completeness attribute",
        "Unsupported attribute combination",
        "Unsupported coded value for term",
        "Malformed search term",
        "Illegal term value for attribute",
        "Unparsable format for un-normalized value",
        "Illegal result set name",
        "Proximity search of sets not supported",
        "Illegal result set in proximity search",
        "Unsupported proximity relation",
	"Unsupported proximity unit code",
        "Unknown diagnostic message", /* 133-200 currently not used */
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
        "Unknown diagnostic message",
	"Proximity not supported with this attribute combination",
	"Unsupported distance for proximity",
	"Ordered flag not supported for proximity",
	"Exclusion flag not supported for proximity",
	"Only zero step size supported for Scan",
	"Specified step size not supported for Scan",
	"Cannot sort according to sequence",
	"No result set name supplied on Sort",
	"Generic sort not supported (db specific only)",
	"DB specific sort not supported",
	"Too many sort keys",
	"Duplicate sort keys",
	"Unsupported missing data action",
	"Illegal sort relation",
	"Illegal case value",
	"Illegal missing data action",
	"Cannot guarantee records will fit in specified segments",
	"Extended Services: Package name already in use",
	"Extended Services: no such package, on modify/delete",
	"Extended Services: quota exceeded",
	"Extended Services: service type not supported",
	"Extended Services: id not authorized on this service",
	"Extended Services: cannot modify or delete",
	"Extended Services: immediate execution failed",
	"Extended Services: immediate execution not supported for this service",
	"Extended Services: immediate execution not supported for these parameters",
	"No data avaliable in requested record syntax",
	"Scan: Malformed scan",
	"Term type not supported",
	"Sort: too many input results",
	"Sort: incompatible record formats" };

#define MAX_DIAG_IDX    ((sizeof(diagnostic_msgs) / sizeof(diagnostic_msgs[0])) - 1)


    SERIES *series = info->errors;
    RECORD        *error;
    EMSG msg;
    
    for (; series_get_total(series) > 0; series->total--, error++) {
	error = (RECORD *)series_item(series, series_get_total(series));
	msg=(EMSG) malloc(sizeof(ERRMSG));
	if (error->condition > MAX_DIAG_IDX)
	    sprintf(msg->message,"Unknown diagnostic returned: %d", error->condition);
	else
	    sprintf(msg->message,"DIAGNOSTIC: %s(%d)",diagnostic_msgs[error->condition],error->condition);
	msg->next=Messages;
	Messages=msg;
    }
    if (Messages==(EMSG) NULL)
	return(NO_MSGS);
    else
	return(ARE_MSGS);
}

make_msg_single_rec_present(int status)
{
    /* We only ask for one record so status tells us
	 about THIS record */
    switch (status) {
	case 0:
	    break;	/* NO PROB! */
	case 1:
	    msg_make(DEFAULT_MSG, "partial-1: Request terminated by access control.");
	    break;
	case 2:
	    msg_make(DEFAULT_MSG, "partial-2: Record is too big to fit in preferred message size.\nRecompile driver with bigger values for PREFERRED_MESSAGE and MAXIMUM_MESSAGE in z3950.h.");
	    break;
	case 3:
	    msg_make(DEFAULT_MSG, "partial-3: Request terminated by resource control at origin.");
	    break;
	case 4:
	    msg_make(DEFAULT_MSG, "partial-4: Request terminated by resource control at target.");
	    break;
	case 5:
	    msg_make(DEFAULT_MSG, "Failure: The requested record could be returned.");
	    break;
	default:
	    msg_make(DEFAULT_MSG, "Bogus status (greater than 5).");
	    break;
    }

}
