#include <strings.h>
#include <scaninfo.h>
#include <series.h>
#include <strlst.h>
#include <constant.h>


STRLST *strlst_new()
/* returns a new, empty stringlist */
{
    STRLST *tmp;

    tmp = (STRLST *) malloc(sizeof(STRLST));
    if (tmp == (STRLST *) NULL)
	return (STRLST *) NULL;
    tmp->numitems = 0;
    tmp->head = NULL;
    return tmp;
}


int strlst_get_numitems(STRLST *sl)
/* returns the number of items in the stringlist */
{
    return sl->numitems;
}


int strlst_get_occurrences(STRLST *sl, int i)
/* returns the occurrences for the ith string from the list */
{
    int j;
    STRLST_ITEM *tmp;

    tmp = sl->head;
    if (strlst_get_numitems(sl) < i)
	return NOTOK;
    else for (j=1;j<i;j++)
	tmp = tmp->next;
    return tmp->occurrences;
}


char *strlst_get_item(STRLST *sl, int i)
/* returns the ith string from the list */
{
    int j;
    STRLST_ITEM *tmp;

    tmp = sl->head;
    if (strlst_get_numitems(sl) < i)
	return NULL;
    else for (j=1;j<i;j++)
	tmp = tmp->next;
    return tmp->item;
}


int strlst_set_item(STRLST *sl, int i, char *item, int occurrences)
/* sets the ith string in the list to item */
{
    int j;
    STRLST_ITEM *tmp;

    tmp = sl->head;
    if (strlst_get_numitems(sl) < i)
	return NOTOK;
    else for (j=1;j<i;j++)
	tmp = tmp->next;
    tmp->item = item;
    tmp->occurrences = occurrences;
    return OK;
}

int strlst_insert(STRLST *sl, int i, char *item, int occurrences)
/* inserts item in the list before the ith string */
{
    int j;
    STRLST_ITEM *tmp, *newitem;

    newitem = (STRLST_ITEM *) malloc(sizeof(STRLST_ITEM));
    if (newitem == (STRLST_ITEM *) NULL)
	return NOTOK;
    newitem->item = item;
    newitem->occurrences = occurrences;

    if (i == 1) {
	newitem->next = sl->head;
	sl->head = newitem;
	sl->numitems = sl->numitems + 1;
	return OK;
    }
	
    tmp = sl->head;
    if (strlst_get_numitems(sl) < i - 1)
	return NOTOK;
    else for (j=2;j<i;j++)
	tmp = tmp->next;
    newitem->next = tmp->next;
    tmp->next = newitem;
    sl->numitems = sl->numitems + 1;
    return OK;
}
    

int strlst_remove(STRLST *sl, int i)
/* removes the ith string from the list */
{
    int j;
    STRLST_ITEM *tmp, *tmp2;

    if (strlst_get_numitems(sl) < i)
	return NOTOK;

    if (i == 1) {
	tmp = sl->head;
	sl->head = tmp->next;
	free(tmp);
	sl->numitems = sl->numitems - 1;
	return OK;
    } 
	
    tmp = sl->head;
    for (j=2;j<i;j++)
	tmp = tmp->next;
    tmp2 = tmp->next;
    tmp->next = tmp2->next;
    free(tmp2);
    sl->numitems = sl->numitems - 1;
    return OK;
}


int strlst_dispose(STRLST **psl)
/* deallocates all memory used by the list */
{
    STRLST_ITEM *tmp, *tmp2;
    STRLST *sl;

    if (psl == NULL)
	return NOTOK;
    sl = *psl;
    if (sl == (STRLST *) NULL)
	return NOTOK;
    tmp = sl->head;
    while (tmp != NULL) {
	tmp2 = tmp->next;
	free(tmp);
	tmp = tmp2;
    }
    free(sl);
    return OK;
}


int strlst_combine(STRLST *sl1, STRLST **psl2)
/* combines two stringlists, puts the result in sl1, and deallocates sl2 */
{
    STRLST_ITEM *tmp;
    STRLST *sl2;

    sl2 = *psl2;
    if (strlst_get_numitems(sl1) == 0) {
	sl1->head = sl2->head;
	sl1->numitems = sl2->numitems;
	free(sl2);
	return OK;
    }
    if (strlst_get_numitems(sl2) == 0) {
	free(sl2);
	return OK;
    }

    tmp = sl1->head;
    while (tmp->next != NULL)
	tmp = tmp->next;
    tmp->next = sl2->head;
    sl1->numitems += strlst_get_numitems(sl2);
    free(sl2);
    return OK;
}


void strlst_display(STRLST *sl)
/* prints the contents of the list to stderr */
{
    STRLST_ITEM *tmp;

    fprintf(stderr,"strlst_display: outputting contents, %d items\n",strlst_get_numitems(sl));
    tmp = sl->head;
    while (tmp != NULL) {
	fprintf(stderr,"Item: %s (%d occurrences)\n",tmp->item,tmp->occurrences);
	tmp = tmp->next;
    }
    fprintf(stderr,"strlst_display: done displaying.\n");
}


int estrcmp(char *s1, char *s2)
/* PSUEDO-strcmp for EBCDIC collating sequence.  Doesn't really do
   true EBCDIC comparison, just puts digits higher than other chars */
{

    while ((s1 != NULL) && (s2 != NULL)) {

	if ((*s1 == NULL) && (*s2 == NULL))
	    return 0; /* reached end, s1 == s2 */
	if (*s1 == NULL)
	    return -1;
	if (*s2 == NULL)
	    return 1;

	if (*s1 == *s2) { /* move to next char */
	    s1++;
	    s2++;
	    continue;
	}

	if ((isdigit(*s1) && isdigit(*s2)) ||
	    (!isdigit(*s1) && !isdigit(*s1)))
	    /* both are in same "group" of char */
	    if (*s1 > *s2)
		return 1;
	    else /* *s1 < *s2 */
		return -1;

	if (isdigit(*s1) && !isdigit(*s2))
	    return 1;
	else /* !isdigit(*s1) && isdigit(*s2) */
	    return -1;
    }


}

	
int strlst_get_position(STRLST *sl, char *item, int collate)
/* searches a stringlist for an item; returns
   0 if list is empty,
   i if item matches the ith item,
   -i if not matched but i is the first item lexicographically > item,
   -(number of items + 1) if item is > all the items.
   uses collating sequence specified by collate */

{
    int i,cmpval;
    STRLST_ITEM *tmp;

    i = 1;
    tmp = sl->head;
    if (tmp == NULL)
	return 0;
    while (tmp != NULL) {
	if (collate == EBCDIC_COLLATING)
	    cmpval=estrcmp(tmp->item,item);
	else /* collate == ASCII_COLLATING */ 
	    cmpval=strcmp(tmp->item,item);

	if (cmpval == 0)
	    return i;
	if (cmpval < 0) {
	    tmp = tmp->next;
	    i++;
	}
	if (cmpval > 0) {
	    return -i;
	}
    }
    return -(i+1);
}
	    

STRLST *series2strlst(SERIES *ser)
/* converts a SERIES of SCAN items to a stringlist */
{
    STRLST *sl;
    SCAN *scanitem;
    STRING scanterm;
    char *term;
    int i;

    sl = strlst_new();
    
    for (i=1;i<=ser->total;i++) {
	scanitem = series_item(ser,i);
	scanterm = scanitem->term;
	term = (char *) malloc(scanterm.length + 1);
	if (term == (char *) NULL)
	    return (STRLST *) NULL;
	*term = NULL;
	strncat(term,scanterm.location,scanterm.length);
	/* note occurrences is spelled occurences in type SCAN */
	strlst_insert(sl,strlst_get_numitems(sl)+1,term,scanitem->occurences);
    }
    return sl;

}


SERIES *strlst2series(STRLST *sl)
/* converts a stringlist to a SERIES of SCAN items */
{
    SERIES *ser;
    SCAN *scanitem;
    STRING scanterm;
    char *term;
    int i;

    ser = series_new(sizeof(SCAN));
    for (i=1;i<=strlst_get_numitems(sl);i++) {
	term = strlst_get_item(sl,i);
	scanitem = (SCAN *) malloc(sizeof(SCAN));
	if (scanitem == (SCAN *) NULL)
	    return ser;
	scanitem->dbname = NULL;
	scanitem->occurences = strlst_get_occurrences(sl,i);
	scanitem->condition = 0;
	scanitem->term.location = term;
	scanitem->term.length = strlen(term);
	series_append(ser,scanitem);
    }
    return ser;
}


int strlst_prune_list(STRLST *sl, int i)
/* removes items from the list, starting with the ith and ending with the last, inclusive */
{
    STRLST_ITEM *tmp, *tmp2;
    int j;

    if (i > strlst_get_numitems(sl))
	return NOTOK;
    if (strlst_get_numitems(sl) == 0)
	return OK;

    tmp = sl->head;
    if (i == 1) {
	while (tmp != NULL) {
	    tmp2=tmp->next;
	    free(tmp);
	    tmp=tmp2;
	}
	sl->head = NULL;
	sl->numitems = 0;
	return OK;
    }

    for (j=1;j<i-1;j++) 
	tmp = tmp->next;
    tmp2 = tmp->next;
    tmp->next = NULL;
    while (tmp2 != NULL) {
	tmp=tmp2->next;
	free(tmp2);
	tmp2=tmp;
    }
    sl->numitems = i-1;
    return OK;
}

