/*
 *  Copyright (C) 1994 by the Massachusetts Institute of Technology
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice appear in all
 * copies and that both that copyright notice and this permission
 * notice appear in supporting documentation, and that the name of
 * M.I.T. not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior
 * permission.  M.I.T. makes no representations about the suitability
 * of this software for any purpose.  It is provided "as is" without
 * express or implied warranty.
 */

/*       This module Interfaces to record.c and is derived from it.
 *
 *  Copyright (c) 1992 Stanford Leland Junior University
 *
 *  Commercial use prohibited
 *
 *  Author: Harold Finkbeiner (Harold@Forsythe.Stanford.Edu)
 *
 * Hacked to death with a Broad Sword to pass data to willow
 * instead of printing it to the tty
 * by Bill Cattey, MIT Information Systems, wdc@mit.edu
 */

#include <stdio.h>

#include <constant.h>
#include <msg.h>
#include <series.h>
#include <session.h>
#include <recinfo.h>
#include <record.h>
#include <series.h>

void send_record(int num, char *text, int summ_flag);


/*--------------------------------------------------------------*
 |                    USMARC_DISPLAY                            Place holder for something called by recinfo.c
   If it ever gets called, we sould warn of an internal
   error.
|
 *--------------------------------------------------------------*/

int usmarc_display(SESSION *session,
                   RECORD *record,
                   int position)
{
    msg_make(DEFAULT_MSG, "Internal error.\nCall made to record_display instead of wcb_record display routines.");
    return NOTOK;
}


/*--------------------------------------------------------------*
 |                    WCB_OUTPUT_MARC                      |
 | Function:                                                    |
 |    Output a US MARC record.                                  |
 | Parameters:                                                  |
 |    message   : Pointer to the US Marc record.                |
 |    length    : Length of the record.                         |
 |    condition : Condition code of record (NOT USED).          |
 |    position  : Ordinal position of record in the result set. |
 |    summ_flag: True if we send Brief Record.                  |
 |    abbr_flag: True if pass on abbreviations unexpanded       |
  | Returns:                                                     |
 |    Nothing.                                                  |
 | Description:                                                 |
 |    Output a record in US Marc format.                        |
 | Note:                                                        |
 |                                                              |
 *--------------------------------------------------------------*/

void wcb_output_marc(RECORD *record,
		 int  position,
		 int summ_flag,
		 int abbr_flag)
{
    wcb_marc(position, record->location, record->length, summ_flag, abbr_flag);
}

/*--------------------------------------------------------------*
 |                    WCB_OUTPUT_MARC_TITLE                      |
 | Function:                                                    |
 |    Output a US MARC record.                                  |
 | Parameters:                                                  |
 |    message   : Pointer to the US Marc record.                |
 |    length    : Length of the record.                         |
 |    condition : Condition code of record (NOT USED).          |
 |    position  : Ordinal position of record in the result set. |
 |    summ_flag: True if we send Brief Record.                  |
 |    abbr_flag: True if pass on abbreviations unexpanded       |
  | Returns:                                                     |
 |    Nothing.                                                  |
 | Description:                                                 |
 |    Output a record in US Marc format.                        |
 | Note:                                                        |
 |                                                              |
 *--------------------------------------------------------------*/

void wcb_output_marc_title(RECORD *record,
		 int  position)
{
    wcb_marc_title(position, record->location, record->length);
}

/*--------------------------------------------------------------*
 |                    WCB_RECORD                        |
 | Function:                                                    |
 |    The summary records of the series are output.             |
 | Parameters:                                                  |
 |    info     : Pointer to the record information.             |
 |    count    : The ordinal position in the result set.        |
 |    summ_flag: True if we send Brief Record.                  |
 |    abbr_flag: True if pass on abbreviations unexpanded       |
 | Returns:                                                     |
 |    Nothing.                                                  |
 | Description:                                                 |
 |    Depending on the type of record, output the record.       |
 | Note:                                                        |
 |                                                              |
 *--------------------------------------------------------------*/

void wcb_record(RECORD_INFO *info, int count, int summ_flag, int abbr_flag)
{
    SERIES *series = info->records;
    RECORD        *record;
    int pos = (count - info->position) + 1;

    record = (RECORD *)series_item(series, pos);

    switch (record->syntax)
    {
	case USMARC:
	    wcb_output_marc(record, count, summ_flag, abbr_flag);
	    break;
	case SUTR:
	    sutr_willow_record(record, count, summ_flag, abbr_flag);
	    break;
	case DIAG_B1:
	    diag_bib1_display(record, count);
	    send_record(count, "DIAG RECORD HERE\n", summ_flag);
	    break;
	case UNKNOWN:
	    misc_warning_hcb("UNKNOWN record format.");
	    send_record(count, "UNKNOWN record format.\n", summ_flag);
	    break;
	default:
	    misc_warning_hcb("UNKNOWN record format.");
	    send_record(count, "UNKNOWN record format.\n", summ_flag);
	    break;
    }
}

int wcb_record_extract(RECORD_INFO *info, int count, char *key, char **retval)
{
    SERIES *series = info->records;
    RECORD        *record;
    int pos = (count - info->position) + 1;

    record = (RECORD *)series_item(series, pos);

    *retval = NULL; /* prepare for the worst */

    switch (record->syntax)
    {
	case USMARC:
	    return wcb_extract_marc(count, record->location, record->length, key, retval);
	    break;
	case SUTR:
	    sutr_willow_record(record, count, 0, 0); /* output whole record */
	    break;
	case DIAG_B1:
	    diag_bib1_display(record, count);
	    break;
	case UNKNOWN:
	    misc_warning_hcb("wcb_record_extract: unknown record format.");
	    break;
	default:
	    misc_warning_hcb("wcb_record_extract: unknown record type.");
	    break;
    }
    return NOTOK;
}

/*--------------------------------------------------------------*
 |                    WCB_RECORD_TITLE                       |
 | Function:                                                    |
 |    The summary records of the series are output.             |
 | Parameters:                                                  |
 |    info     : Pointer to the record information.             |
 |    count    : The ordinal position in the result set.        |
 |    summ_flag: True if we send Brief Record.                  |
 |    abbr_flag: True if pass on abbreviations unexpanded       |
 | Returns:                                                     |
 |    Nothing.                                                  |
 | Description:                                                 |
 |    Depending on the type of record, output the record.       |
 | Note:                                                        |
 |                                                              |
 *--------------------------------------------------------------*/

void wcb_record_title(RECORD_INFO *info, int count)
{
    SERIES *series = info->records;
    RECORD        *record;
    int pos = (count - info->position) + 1;

    record = (RECORD *)series_item(series, pos);

    switch (record->syntax)
    {
	case USMARC:
	    wcb_output_marc_title(record,
				 count);
	    break;
	case SUTR:
	    sutr_willow_title(record, count);
	    break;
	case DIAG_B1:
	    diag_bib1_display(record, count);
	    send_title(count, "DIAG RECORD HERE\n");
	    break;
	case UNKNOWN:
	    misc_warning_hcb("UNKNOWN title format.");
	    send_title(count, "UNKNOWN title format.");
	    break;
	default:
	    misc_warning_hcb("UNKNOWN title format.");
	    send_record(count, "UNKNOWN title format.", 1);
	    break;
    }
}
