
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: support.C,v 2.3 1995/06/24 15:01:04 davidsi Exp $
// 
// $Log: support.C,v $
// Revision 2.3  1995/06/24  15:01:04  davidsi
// - Added GNU copyright message.
//
// Revision 2.2  1995/04/03  20:37:50  davidsi
// - Added a allocstringlist() to provide a consistent mechanism for allocating
//   string lists.
//
// Revision 2.1  1995/02/27  22:17:21  davidsi
// - Added the freestringlist() function.
//
// Revision 2.0  1995/02/05  18:15:05  davidsi
// - Initial version.
//
//

#include <stdio.h>
#include <ctype.h>
#define _STDARG_H
#include <stdlib.h>
#include <dial.h>
#include <sys/types.h>

static char rcsid[]="$Id: support.C,v 2.3 1995/06/24 15:01:04 davidsi Exp $";

//
// alpha_strcmp string1 string2
//
// Compares the two strings ignoring the case of characters
//
// INPUTS : 	string1		First string to be compared 
// 		string2		Second string to be compared 
// RETURNS:	Position of difference or 0 if strings are equal when
//		case is ignored.
//
int
alpha_strcmp(char *str1,char *str2)
{
	while (toupper(*str1)==toupper(*str2)) {
		if (*str1=='\0')
			return(0);
		str1++;
		str2++;
	}
	return((int) (toupper(*str1) - toupper(*str2) ));
}



//
// strcpy_2_space dst src size
//
// Copies strings using nulls, spaces or buffer full for the termination
// condition.
//
// INPUTS :	dst	Destination for copy
//		src	Source string
//		size	size of destination area
// RETURNS:	Pointer to next string or to null terminator.
//
char *
strcpy_2_space( char *dst, char *src, int size)
{
register int length=0;

	while (*src==' ')			// skip leading spaces
		src++;

	while (*src!=' ' && *src!='\0' && length<size-1 )
		*dst++=*src++;
	*dst='\0';
	return(src);
}


//
// remove_lt_spaces string
//
// Removes leading and trailing spaces from a string.
// condition.
//
// INPUTS :	string	String whose leading spaces are to be removed
// RETURNS:	Pointer to substring within the original string where
//		all leading/trailing spaces have been removed
//
//		*NOTE* The string passed in can be modified by this
//		function.
char *
remove_lt_spaces( register char *string)
{
register char *s;

	while (*string==' ')
		string++;
	s=string+strlen(string)-1;
	while (*s==' ' && s>string)
		--s;
	*(s+1)='\0';
	return(string);
}


/*
 * freestringlist list
 *
 * Removes all entries in a list of XmStrings, then frees the memory used
 * the by the actual list structure.
 *
 * INPUTS : list     A ptr to an area of memory storing ascii strings
 *          size     Number of XmStrings in the memory area
 */
void
freestringlist( char **ilist )
{
	char **olist=ilist;

	if (ilist==NULL)
		return;

	while(*ilist) {
		free(*ilist);
		ilist++;
	}

	free(olist);
}

char **
allocstringlist(int length)
{
	return( (char **)malloc( sizeof(char *) * (length+1)) );
}
