
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: symbol.C,v 2.2 1995/06/24 15:01:04 davidsi Exp $
//
// $Log: symbol.C,v $
// Revision 2.2  1995/06/24  15:01:04  davidsi
// - Added GNU copyright message.
//
// Revision 2.1  1995/06/17  20:36:27  davidsi
// - There was a need for multiple references to the same symbol table for
//   value_structs.  Added functions that support reference counting on
//   a symbol table, with the symbol table only being deleted when the
//   count becomes zero.
//
// Revision 2.0  1995/01/19  21:48:50  davidsi
// - Initial version.
//
//


#include "stdlib.h"
#include "stdio.h"

#include "symbol.h"
#include "localmacros.h"
#include "error.h"

extern errstream cerror;
extern diagstream cinfo;



//
// Constructor for the symbol_value class. Typically the symbol_value
// class will be inherited by other classes that need to use the
// symbol table.
//
symbol_value::symbol_value()
{
}


//
// Destructor for the symbol_value class
//
symbol_value::~symbol_value()
{
}



//
// Convenience function to save the need for casts elsewhere
//
class symbol *
symbol::next() 
{
	return( LISTREMAP(symbol,list_node,next,this) );
}

//
// symbol::name
//
// Returns the name of a symbol
//
// INPUTS :	None
// RETURNS:	The name of the symbol
//
char *
symbol::name()
{
	return( symname );
}


//
// symbol::symbol_value
//
// Returns the value of a symbol
//
// INPUTS :	None
// RETURNS:	The value of the symbol
//
//
class symbol_value *
symbol::symbol_value()
{
	return(value);
}


//
// Constructor for the symbol class
//
symbol::symbol(char *name,class symbol_value *v)
{
	INFO(cinfo << "Creating symbol " << name <<"\n";)

	symname=(char *)malloc(strlen(name)+1);
	strcpy(symname,name);
	value=v;
}


//
// Destructor for the symbol class
//
symbol::~symbol()
{
	free(symname);
}


//
// Convenience function to save the need for casts elsewhere
//
class symbol *
symboltable::head() 
{
	return( LISTREMAP(symbol,list,head,this) );
}

//
// symboltable::findsymbol name
//
// Locates a symbol within a symbol table.
//
// INPUTS :	name		The name of the symbol
// RETURNS:	NULL if not found, else the value of the symbol.
//
class symbol_value *
symboltable::findsymbol(char *symname)
{
	class symbol *sptr=(class symbol *)head();
	
	while (!eol(sptr)) {
		if ( !strcmp(sptr->symname,symname) )
			return(sptr->value);
		sptr=sptr->next();
	}

	return (NULL);
}


//
// symboltable::addsymbol name value
// 
// Adds a symbol to a symbol table with the specified value.
//
// INPUTS :	name		The name of the symbol
//		value		The value of the symbol
// RETURNS:	None
void
symboltable::addsymbol(char *name,class symbol_value *v)
{
	class symbol *sptr;

	cinfo << "Adding symbol " << name << "\n";

	sptr=new class symbol(name,v);

	insert_head(sptr);
}


//
// symboltable::SymboltableReferenced
//
// References a symbol table. There are occasions (the value_struct class
// is one) where a symbol table may be referenced by multiple instances
// of a class (or classes) and the table must remain around until all
// those instances have gone.  Provided the class(es) cooperate and
// use SymboltableReferenced/SymboltableUnreferenced then this will
// be acheived. 
//
// INPUTS :	None
// RETURNS:	None
//
void 
symboltable::SymboltableReferenced()
{
	refcnt++;
}


//
// symboltable::SymboltableUnreferenced
//
// Unreferences a symbol table. If there are no references, then
// delete the table.
//
// INPUTS :	None
// RETURNS:	None
//
void 
symboltable::SymboltableUnreferenced()
{
	refcnt--;
	if (refcnt==0)
		delete this;
}


//
// Constructor
//
symboltable::symboltable()
{
	refcnt=0;
}

//
// Destructor
//
symboltable::~symboltable()
{
}
