
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: symbol.h,v 2.2 1995/06/24 15:06:13 davidsi Exp $
//
// $Log: symbol.h,v $
// Revision 2.2  1995/06/24  15:06:13  davidsi
// - Added GNU copyright message.
//
// Revision 2.1  1995/06/17  20:36:02  davidsi
// - There was a need for multiple references to the same symbol table for
//   value_structs.  Added functions that support reference counting on
//   a symbol table, with the symbol table only being deleted when the
//   count becomes zero.
//
// Revision 2.0  1995/01/19  21:48:50  davidsi
// - Initial version.
//
//

#ifndef __SYMBOL_H
#define __SYMBOL_H

#include "list.h"

class symbol_value {
	
public:
	symbol_value();
	~symbol_value();
};

class symbol : public list_node {
	char *symname;
	class symbol_value *value;

	friend class symboltable;
	
public:
	class symbol *next();

	//
	// The following functions are useful for debug purposes
	//
	char *name();
	class symbol_value *symbol_value(); 

	symbol(char *name,class symbol_value *val);
	~symbol();
};


class symboltable :  public list {
	int refcnt;		// number of references to this table
	
public:
	class symbol_value *findsymbol(char *symname);
	void addsymbol(char *name,class symbol_value *val);
	void SymboltableReferenced();
	void SymboltableUnreferenced();

	//
	// Prevent the need for casts
	//
	class symbol *head();

	symboltable();
	~symboltable();
};
#endif
