
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: text.h,v 2.2 1995/06/24 15:06:13 davidsi Exp $
//
// $Log: text.h,v $
// Revision 2.2  1995/06/24  15:06:13  davidsi
// - Added GNU copyright message.
//
// Revision 2.1  1995/02/28  22:18:52  davidsi
// - Added locking for text instances, this allows a Text instance to
//   be modified once its being displayed: TextrealClear() sets the
//   lock, TextrealComplete() clears the lock, and TextrealIslocked()
//   returns the state of the lock.
//
// Revision 2.0  1995/02/02  21:17:49  davidsi
// - Many changes.  Factored out much of the functionality into 2 new classes.
//   The attr (attribute) class looks after text attributes. The textline
//   class looks after lines of text and their attributes.  This has resulted
//   in a much cleaner text class, without all the weird array manipulations.
//
// Revision 1.1  1994/12/11  21:38:35  davidsi
// Initial revision
//
//
//
#ifndef __TEXT_H
#define __TEXT_H

#include "draw.h"

#define MAXFIELDS	16

typedef unsigned long attr_t;
//
// attr_t values
//
#define FIELDATTR_LEFT		0x1
#define FIELDATTR_RIGHT		0x2
#define FIELDATTR_CENTRE	0x4

//
// Derived attribute values
//
#define FIELDATTR_TEXTPOS (FIELDATTR_LEFT | FIELDATTR_RIGHT | FIELDATTR_CENTRE)


class attr {
	attr_t flags;
	colour_e colour;
public:
	attr_t extract_attr();
	colour_e extract_colour();
	
	attr(class attr *);
	attr(attr_t,colour_e);
	attr(attr_t);
	~attr();
};

class textline {
	char *linetext;
	int *fieldlens;		// ptr to array of lengths for each string
	class attr **attrs;

public:
	char *readtext(int fieldno);
	class attr *readattribute(int fieldno);
	void setattr(int fieldno,class attr *aptr);
	textline(int fields,char *txt,int *fmax,class attr **,char fieldend);
	~textline();
};

//
// The Textreal class provides real text management facilities
// 
class Textreal {
private:
	int ref_cnt;		// number of text instances referencing this one
	int max_n_lines;	// total number of lines space has been 
				// allocated for - this will increase
	class textline **lines;
	class attr  **df_attr;	// default attributes 
	int *fieldmax;		// max field widths
	char fieldend;		// character delimeter for fields
	int fields;		// number of fields (maximum)
	int n_lines;		// number of lines in 'lines' array
	unsigned long textflags;// flags see TEXT_FLAGS_XXX

public:
	void TextrealAdd(char *line);
	void TextrealClear();
	char *TextrealRead(int lineno);
	char *TextrealRead(int lineno,int field);
	class attr *TextrealReadattribute(int lineno,int field);
	virtual void TextrealOutput(int lineno);
	int TextrealNlines();
	int TextrealNfields();
	int TextrealCline();
	void TextrealSetdefaultattr(int fi,class attr *);
	void TextrealSetattr(int line,int fi,class attr *);
	void TextrealSetattr(int line,class attr *);
	void TextrealUnref();
	void TextrealRef();
	int TextrealIslocked();
	void TextrealComplete();

	Textreal(int number_of_fields);
	virtual ~Textreal();
};


//
// The Text class.
//
class Text {
private:
	class Textreal *text;
public:
	void TextAdd(char *line);
	void TextClear();
	char *TextRead(int lineno);
	char *TextRead(int lineno,int field);
	class attr *TextReadattribute(int lineno,int field);
	virtual void TextOutput(int lineno);
	int TextNlines();
	int TextNfields();
	int TextCline();
	void TextSetdefaultattr(int fi,class attr *);
	void TextSetattr(int line,int fi,class attr *);
	void TextSetattr(int line,class attr *);
	void TextAssign(class Text *);
	int TextIslocked();
	void TextComplete();

	Text();
	virtual ~Text();
};

//
// textflags values
//
#define TEXT_FLAGS_LOCKED	0x1	// don't access (for display updates)
					// as internal structures are not
					// yet consistent

#endif
