
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: textdialog.C,v 2.4 1995/06/24 15:01:04 davidsi Exp $
//
// $Log: textdialog.C,v $
// Revision 2.4  1995/06/24  15:01:04  davidsi
// - Added GNU copyright message.
//
// Revision 2.3  1995/04/16  20:26:18  davidsi
// - Added TextdialogSetprintertitle() that allows a title to be set that
//   will be used when printing out.
//
// Revision 2.2  1995/04/03  20:37:21  davidsi
// - BUG FIX. Textdialogprint::~Textdialogprint() must check that the
//   d (Draw) member has been set, since it is possible to delete an
//   instance without having set this member. The Textdialogprint
//   constructor now sets this to NULL.
//
// Revision 2.1  1995/03/01  22:15:39  davidsi
// - When mapping colour information use inheritated members rather
//   than the global information.
// - Add support for centred text (FIELD_ATTR_CENTRE).
// - Add TextdialogRedraw() to allow a forced update - useful if
//   the fundamental data to be displayed changes.
// - BUG FIX. Textdialog::TextdialogPrint() now handles the situation
//   when Printer::GimmeDraw() fails to return a Drawtext instance.
// - BUG FIX. Textdialogprint() requires a destructor so that it can
//   delete the Drawtext instance.
//
// Revision 2.0  1995/02/02  21:18:35  davidsi
// - Initial version.
//
//
//
//




//
// Include Files
//
#include <stdio.h>
#include <varargs.h>
#define _STDARG_H
#include <stdlib.h>
#include <string.h>
#include <dial.h>
#include <math.h>

#define FULL_INCLUDE

#include "localmacros.h"

#include "textdialog.h"
#include "draw.h"

extern dialog d_root;
//
// The Textdialogprint class
//
// The sole function of this class is to support the textdialog class.
// It provides an issolated class that can inherit from the Printer 
// class without any interactions (ie. from Menu or Button classes).
//
class Textdialogprint : public Printer {
	class Drawtext *dt;
	class Draw *d;
private:
	//
	// Overrides the virtual functions in the Printer class
	//
	virtual void PrinterPage(int pno);
public:
	void TextdialogSetdraw(class Drawtext *,class Draw *);
	Textdialogprint();
	virtual ~Textdialogprint();
};

//
// This assignment is pernament - this will take the responsibility
// for delting the Drawtext and the draw class
//
void
Textdialogprint::TextdialogSetdraw(class Drawtext *idt,class Draw *id)
{
	dt=idt;
	d=id;
}

void
Textdialogprint::PrinterPage(int pageno)
{
	dt->DrawtextOutput(pageno);
}

//
// Textdialogprint constructor
//
// This sets up the textdialog pointer so that PrinterPage() member
// can render the output.
//
Textdialogprint::Textdialogprint()
{
	dt=NULL;
	d=NULL;
}

Textdialogprint::~Textdialogprint()
{
	//
	// If Drawtext exists, delete it.  We need to do this!  In
	// the case of printer based output, this is what triggers
	// the closing of the relevant files and the files being
	// passed to the print spooler.
	//
	if (dt)
		delete dt;

	//
	// This will result in the draw dialog going away, unless
	// someone else is still using it...
	//
	if (d)
		d->DrawUnref();
}

//
// accessfield thisptr lineno fieldno string len col attr
//
// This function is statically declared so that it can act as
// a callback to the dialog functions that actually handle the
// text windows.  It takes a ptr to the Textdialog
//
void
Textdialog::accessfield(char *cbv,
		int l,
		int f,
		char **string,
		int *len,
		char **col,
		int *attr)
{
	class Textdialog *rthis=(class Textdialog *)cbv;
	colour_e colval;

	//
	// This is special case, indicating that the background colour
	// should be returned!
	//
	if (l==-1 && f==-1) {
		*col=rthis->ColourString( rthis->ColourBackground() );
		if (rthis->TextIslocked())  {
			*attr= -1;
		} else
			*attr= 0;
		return;
	}

	*string=rthis->TextRead(l,f);
	*len=strlen(*string);
	if (col) {
		colval=(rthis->TextReadattribute(l,f))->extract_colour();
		*col=rthis->ColourString(colval);
	}
	if (attr)
		switch ((rthis->TextReadattribute(l,f))->extract_attr()) {
		case FIELDATTR_LEFT:
			*attr=FIELD_ATTR_LEFT;
			break;
		case FIELDATTR_RIGHT:
			*attr=FIELD_ATTR_RIGHT;
			break;
		case FIELDATTR_CENTRE:
			*attr=FIELD_ATTR_CENTRE;
			break;
	}
}


// 
// TextdialogRedraw
//
// Redisplays the current data no additional processing is performed.
//
// INPUTS :	None
// RETURNS:	None
//
void
Textdialog::TextdialogRedraw()
{
	dialog_text_redraw(ti_ptr);
}


// 
// TextdialogOutput
//
// Outputs the Text class information via a text dialog.
//
// INPUTS :	None
// RETURNS:	None
//
void
Textdialog::TextdialogOutput()
{
	static new_d_text nt= {
		8,                      // number of fields
		15,                     // field spacing
		MTEXT_HEADER,           // flags
		"Untitled",
		"prop2",                // font
		650,                    // width
		250,                    // height
		NULL 
	};
	char str[2048];

	if (!ti_ptr) {
		nt.accessfield=accessfield;
		nt.accessfieldv=(char *)this;
		if (title)
			nt.name=title;
		ti_ptr=dialog_text(d_root,&nt,NULL);
		MenuDisplay(ti_ptr);
	} 
	dialog_text_setlines(ti_ptr,TextNlines());
	dialog_realise(ti_ptr);
}


//
// TextdialogPrint
//
// Handles the printing of text within a Textdialog instance.
//
// INPUTS :	None
// RETURNS:	None
//
void
Textdialog::TextdialogPrint()
{
	class Draw *d;
	//
	// Free off any old printer information
	//
	if (printer)
		delete printer;

	printer=new class Textdialogprint();

	//
	// If PrinterSetup() then the user has hit CANCEL
	// or some other problem has arisen which means
	// nothing should be printed.
	//
	do {
		if (printer->PrinterSetup() == -1)
			return;
		d=printer->GimmeDraw();
	} while (d==NULL);

	//
	// Drawtext will bump the reference count on Draw
	//
	dt=new class Drawtext(d);
        dt->DrawtextTitle(printer_title);

	//
	// Note - we don't need to call DrawRef(), because we are
	// effectively transferring ownership.
	//
	printer->TextdialogSetdraw(dt,d);

	dt->TextAssign(this);
	dt->DrawStartpage();
        dt->DrawSetfont("Helvetica",12);
	dt->DrawtextPrepare();

	if (printer->PrinterOutput(1,dt->DrawtextNpages()) == 0)
		printer=NULL;
}


//
// TextdialogSettitle title
//
// Sets the title for a Textdialog instance.
//
// INPUTS :	ititle		The title to set
// RETURNS:	None
//
void
Textdialog::TextdialogSettitle(char *ititle)
{
	//
	// If no registration mechanism is specified, the use sensible
	// defaults.
	//
	if (!ititle)
		ititle="Untitled";

	title=(char *)malloc(strlen(ititle)+1);
	strcpy(title,ititle);
}

//
// TextdialogSetprintertitle ititle
//
// Sets the title for the printer output with a textdialog
//
// INPUTS :	ititle 		The title to set
// RETURNS:	None
//
void
Textdialog::TextdialogSetprintertitle(char *iptitle)
{
	if (printer_title) 
		free(printer_title);
	if (iptitle) {
		printer_title=(char *)malloc(strlen(iptitle)+1);
		strcpy(printer_title,iptitle);
	} else
		printer_title=NULL;
}



//
// Constructor function
//
// A draw dialog requires a drawd_reg_t structure to specify how the
// dialog will be managed.
//
Textdialog::Textdialog()
{
	ti_ptr=NULL;
	title=NULL;
	dt=NULL;
	printer=NULL;
	printer_title=NULL;
}



//
// Destructor function
//
Textdialog::~Textdialog()
{

	if (title)
		free(title);

	if (printer_title)
		free(printer_title);

	if (ti_ptr);
		dialog_destroy(ti_ptr);
	//
	// Delete printer information - this will delete the draw
	// text if it still exists.
	//
	if (printer)
		delete printer;
}
