
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: trans.C,v 2.7 1995/06/23 20:50:28 davidsi Exp $
//
// $Log: trans.C,v $
// Revision 2.7  1995/06/23  20:50:28  davidsi
// - Don't allow transactions to be entered on a group type
//   security.
//
// Revision 2.6  1995/06/02  19:55:16  davidsi
// - Tidied up names and titles of the dialogs.
//
// Revision 2.5  1995/05/18  20:32:26  davidsi
// - The make_callback() function is now CallbackMake().
// - DialogentryCreate() now requires an argument specifying total number
//   of visible columns.
// - BUG FIX. The call to add_load_vect() should occur from within the
//   trans_list constructor otherwise the actual list may not exist when
//   the add_load_vect() function is called.
//
// Revision 2.4  1995/04/25  21:07:12  davidsi
// - Added support for a busy cursor.
// - Additional menu options to allow the user to specify which transactions
//   they are interested in.
//
// Revision 2.3  1995/04/16  20:25:58  davidsi
// - Added Security() function member that extracts a security from a
//   transaction. Rarely useful, but now there is a need for it (see
//   dtrans.C).
//
// Revision 2.2  1995/03/08  21:21:34  davidsi
// - Factored the transaction class into two classes, the transaction class
//   provides all the main transaction handling functionality and the
//   Transactionedit class allows transactions to be edited.  This reduces
//   the amount of memory req'd to represent a transaction.
// - The extract() member of security was renamed to SecurityExtract(), so
//   so this module modified to reflect that.
// - Revised to use the new Holding class which contains transaction
//   information rather than the security class.
// - BUG FIX. When a SELL transaction was being created, the number being
//   checked was the number member rather than the number just entered by
//   the user.
//
// Revision 2.1  1995/03/01  22:25:19  davidsi
// - Uses Dialog class. Calls to dialog_xxx functions have been removed.
//
// Revision 2.0  1995/01/19  21:09:09  davidsi
// - Renamed all calls to date2string() which is now getdate().
// - trans_load now must take a first ptr argument, that it can ignore. Please
//   see the related changes in the callback class.
// - In transaction::create_transaction() and trans_load cause all
//   callbacks to occur.
// - Removed the time members of the trans_list class.  Dates are now
//   managed using the date class.
// - Removed trans2date() member.
// - Added extract() member to transactions.
// - Removed the time members of the trans_list class.  Dates are now
//   managed using the date class.
// - Removed trans2date() member.
// - Added extract() member to transactions.
// - Added add_vect to the trans_list class to allow callbacks when
//   transactions are added.
//
// Revision 1.3  1994/11/05  21:08:55  davidsi
// - BUG FIX.  Removed the "Number" labelentry dialog in addw_date()
//   as this dialog is already created by transaction::addw_number().
// - Various modifications to allow this file to compile using g++.
//
// Revision 1.2  1994/10/17  20:58:08  davidsi
// - All calls to extract() in the security class now require a further
//   argument, so that the type of security can be determined using flags
//   rather than strings.
// - Now that per security transaction lists are circular the termination
//   condition for scanning the list must be updated for having completed
//   the entire list.
// - When a security name list has been extracted, it no longer should be
//   freed because it is a shared resource and not private to the caller.
// - Tidied up transaction::addw_sec_name() which no longer needed seperate
//   cases for security name list building.
// - BUG FIX. The d_entrylist was not being correctly initialized despite
//   widgets being created for it.
// - BUG FIX.  The regex pattern used by transaction::trans_load() was
//   not matching all intended patterns.
//
// Revision 1.1  1994/08/06  16:35:49  davidsi
// Initial revision
//
//
//



#include <stdio.h>
#include <varargs.h>
#define _STDARG_H
#include <stdlib.h>
#include <dial.h>

#include "localmacros.h"
#include "menu.h"
#include "trans.h"


//
// Identifiers for the dialogs used in this module
//
enum dialogids_e {
	ID_DATE,
	ID_NUMBER,
	ID_NUMBEREXE,
	ID_PRICEPERSHARE,
	ID_TOTAL,
	BUTTON_OK,
	BUTTON_CANCEL
	};



//
// Variable Declarations 
///
trans_list *main_trans_list;
static struct {
		char *name;		// name of transaction type for files
		int value;		// internal id for this transaction type
	} trans_types[]=
	{
	{"BUY",TRANS_BUY} , 
	{"SELL",TRANS_SELL} ,
	{"SPLIT",TRANS_SPLIT} ,
	{"BUYR",TRANS_BUYRIGHTS},
	{"SELR",TRANS_SELRIGHTS},
	{"EXR",TRANS_EXRIGHTS} ,
	{"USER",TRANS_USERIGHTS},
	{"GETR",TRANS_GETRIGHTS} ,
	{"BONUS",TRANS_BONUS},
	{"VOID",TRANS_VOID},
	{NULL,NULL} 
	};





//
// Transactionedit::ButtonCallback buttonid flag
//
// Handles the call back when a OK/CANCEL button is pressed. For OK
// the transaction must be checked and then stored both in memory and
// on the file.  For CANCEL the dialog is removed from the display.
// Note that the OK/CANCEL refers to the string and not necessarily
// the label on the dialog which might be PURCHASE, etc.
// 
// INPUTS :	buttonid	The button identifier
//		flag		State of button
// RETURNS:	None
//
void 
Transactionedit::ButtonCallback(unsigned long id,buttonstate_e flag)
{
	char *str;
	double number_to_exe;
	class date cdate;
	double cnumber=0.0;
	double cprice=0.0;
	double ctotal=0.0;
	class security *sec;
	class Holding *holding;
	transaction *newtrans;
	sectype_e sectype;

	if (id == BUTTON_CANCEL)
		goto done;

	str=d_entrylist.DialogentryGet(ID_DATE);
	if (str) {
		if (cdate.setdate(str) == -1) {
			DialogMessage("No such date");
			return;
		}
	} else {
		DialogMessage("A date must be entered");
		return;
	}

	str=d_entrylist.DialogentryGet(ID_NUMBER);
	if (str)
		cnumber=atof(str);

	str=d_entrylist.DialogentryGet(ID_PRICEPERSHARE);
	if (str) 	// only if price dialog was put up
		cprice=atof(str);

	str=d_entrylist.DialogentryGet(ID_TOTAL);
	if (str) 	// only if price dialog was put up
		ctotal=atof(str);
	
	str=d_secname->OptionlistinstanceGet();
	holding=portfolio->HoldinglistFind(str);
	if (holding==NULL) {
		DialogMessage("Cannot add a transaction for a non-existent security");
		return;
	}
	sec=holding->HoldingSecurity();

	//
	// Transactions on a "group" security don't make sense - so don't
	// allow them.
	//
	sec->SecurityExtract(NULL,NULL,NULL,NULL,&sectype);
	if (sectype==SEC_TYPE_GROUP) {
		DialogMessage("Transactions are not possible on this security");
		return;
	}

	if (TRANS_SELL==trans_type && 	// can't sell more than owned
		cnumber >(holding->HoldingTransactionhead())->number(&cdate) ) {
		DialogMessage("Do not own that number of securities");
		return;
	}
	//
	// If we are exercising rights then we must update the share
	// structure relating to those rights-as well as the real
	// share (which is being obtained).
	//
	if ( trans_type==TRANS_EXRIGHTS ) {
		char *rstr;
		class security *sec;
		class Holding *holding;

		rstr=d_entrylist.DialogentryGet(ID_NUMBEREXE);
		if (rstr)
			number_to_exe=atof(rstr);
		else
			number_to_exe=0.0;

		rstr=d_rsecname->OptionlistinstanceGet();
		holding=portfolio->HoldinglistFind(str);
		if (holding==NULL) {
			DialogMessage("Cannot add a transaction for a non-existent share");
			return;
		}
		sec=holding->HoldingSecurity();
		if (number_to_exe>(holding->HoldingTransactionhead()) ->number(&cdate) ) {
			DialogMessage("Do not own that number of rights");
			return; 
		}

		newtrans=new transaction;
		if (newtrans->TransactionCreate(rstr,TRANS_USERIGHTS,
					&cdate,number_to_exe,0,0) == -1) {
			DialogMessage("Cannot add transaction");
			delete newtrans;
			return;
		}
	}

	newtrans=new transaction;
	if (newtrans->TransactionCreate(str, 
			trans_type,&cdate,cnumber,cprice, ctotal)== -1) {
		DialogMessage("Cannot add transaction");
		delete newtrans;
		return;
	}
	main_trans_list->save();
done:
	delete this;
#undef _L_BUFSIZE
}


//
// Transactionedit::addw_buttons
// 
// Creates the buttons which allow the user finally to say PURCHASE, SELL,
// CANCEL (etc).
// 
// INPUTS :	None
// RETURNS:	None
//
void 
Transactionedit::addw_buttons()
{
	char *lname;

	d_buthoriz.DialoghorizontalCreate(d_vert);

	switch (trans_type) {
	case TRANS_BUYRIGHTS:
	case TRANS_BUY:
	case TRANS_EXRIGHTS:
		lname=" PURCHASE ";
		break;
	case TRANS_SELL:
	case TRANS_SELRIGHTS:
		lname=" SELL ";
		break;
	case TRANS_BONUS:
	case TRANS_SPLIT:
	case TRANS_GETRIGHTS:
		lname=" ACCEPT ";
		break;
	default:
		DialogPanic("addw_buttons:bad value in select");
	}
	(void)ButtonPush(d_buthoriz,lname,BUTTON_OK);
	(void)ButtonPush(d_buthoriz," CANCEL ",BUTTON_CANCEL);
}



//
// Transactionedit::addw_sec_name parent mode
//
// Creates a list that allows the selection of a stock name.
// 
// INPUTS :	mode		TRUE if to give share list
//				FALSE if to give rights
// RETURNS:	The newly created optionlist
//
Optionlistinstance  *
Transactionedit::addw_sec_name(
	int mode
	)
{
	char rights_flag;		// TRUE when dealing with rights/
					// warrants etc
	//
	// List definition for share names
	//
	static Newoptionlist nolist = {
		20,				// width
		5,				// height
		NULL,				// list of items
		"",				// name of list
		"",				// name of item
		OLIST_FLAGS_SEARCH,		// flags=search for matches
		NULL,				// default font
	};

	rights_flag=(trans_type==TRANS_BUYRIGHTS||trans_type==TRANS_SELRIGHTS ||
		trans_type==TRANS_EXRIGHTS || trans_type==TRANS_GETRIGHTS );

	nolist.listitem=main_sec_list->extract();

	if (mode && trans_type==TRANS_EXRIGHTS)
		nolist.selname="Stock To Which Rights Apply";
	else if (mode && !rights_flag)
		nolist.selname="Stock Name";
	else
		nolist.selname="Warrant Name";

	if ( !rights_flag || (mode && trans_type==TRANS_EXRIGHTS) )
		nolist.listname="Stock Names";
	else
		nolist.listname="Warrant Names";

	return ( OptionlistCreate(d_horiz,nolist,0) );
}





//
// Transactionedit::addw_prices 
//
// Adds the dialogs to a parent dialog that allow the user to specify
// amounts spent/received, numbers bought, etc.
// 
// INPUTS :	None
// RETURNS:	None
//
void 
Transactionedit::addw_prices()
{

	switch(trans_type) {
	case TRANS_BUY:
	case TRANS_EXRIGHTS:
	case TRANS_BUYRIGHTS:
		d_entrylist.DialogentrySet("Price per Stock","",ID_PRICEPERSHARE);
		d_entrylist.DialogentrySet("Total","",ID_TOTAL);
		if (trans_type==TRANS_EXRIGHTS)
			d_entrylist.DialogentrySet("Number to exercise","",ID_NUMBEREXE);
		break;
	case TRANS_SELL:
	case TRANS_SELRIGHTS:
		d_entrylist.DialogentrySet("Price per Stock","",ID_PRICEPERSHARE);
		d_entrylist.DialogentrySet("Total","",ID_TOTAL);
		break;
    }
}



//
// Transactionedit::userentry title trans_type
//
// Builds the user input dialog.
// 
// INPUTS :	title		The title of the new dialog
//		trans_type	The types of transaction (TRANS_XXX)
// RETURNS:	None
//
void 
Transactionedit::userentry(
	char *title,
	int ttype
	)
{
	class date tdate;	// date

	trans_type=ttype;
	d_shell.DialogshellCreate(title);
	d_vert.DialogverticalCreate(d_shell);

	d_horiz.DialoghorizontalCreate(d_vert);

	if (trans_type==TRANS_EXRIGHTS)
	d_rsecname=addw_sec_name(FALSE);
	d_secname=addw_sec_name(TRUE);

	d_entrylist.DialogentryCreate(d_horiz,12);

	tdate.setdate();
	d_entrylist.DialogentrySet("Date",tdate.getdate(),ID_DATE);
	d_entrylist.DialogentrySet("Number","",ID_NUMBER);
	addw_prices();

	addw_buttons();
	d_shell.DialogRealize();
}


//
// Constructor for Transactionedit
//
Transactionedit::Transactionedit()
{
}

//
// Destructor for Transactionedit
//
Transactionedit::~Transactionedit()
{
}


//
// transaction::TransactionCreate secname transtype date number price total
//
// INPUTS :	secname		The security to which the transaction applies
//		transtype	A TRANS_XXX value.
//		date		Date that the transaction was performed at
//		number		Number of securities involved
//		price		Price involved
//		total 		Total for the transaction (i.e. either
//				received or spent).
// RETURNS:	-1 if error otherwise 0
//
int
transaction::TransactionCreate( 
	char *cname,
	unsigned long ctype,
	class date *cdate,
	double cnumber,
	double cprice,
	double ctotal
	)
{
	class Holding *holding;

	holding=portfolio->HoldinglistFind(cname);
	if (!holding)
		return -1;
	sec=holding->HoldingSecurity();

	//
	// Check transaction type is valid
	//
	switch (ctype) {
	case TRANS_BUY:
	case TRANS_SELL:
	case TRANS_SPLIT:
	case TRANS_BUYRIGHTS:
	case TRANS_SELRIGHTS:
	case TRANS_EXRIGHTS:
	case TRANS_USERIGHTS:
	case TRANS_GETRIGHTS:
	case TRANS_BONUS:
	case TRANS_VOID:
		break;
	default:
		DialogPanic("transaction::TransactionCreate - Bad transaction type");
	}

	trans_type=ctype;
	date=*cdate;
	number=cnumber;
	price=cprice;
	total=ctotal;

	//
	// Add it to the list of all transactions
	//
        main_trans_list->insert_tail(this);

	//
	// Add it to the list of transactions for the specific security
	//
	holding->HoldingTransactionadd(this);

	//
	// Issue the callback
	//
	main_trans_list->add_vect.CallbackMake(0,NULL);
	
	return 0;
}


//
// transaction::extract type date number price total
//
// extracts per transaction information
//
// INPUTS :	None
// RETURNS:	type		see TRANS_XXX (ptr for return) 
//		date		The date for the transaction (ptr for return)
//		number		Number of securities involved (ptr for return)
//		price		Price per security (ptr for return)
//		total		Total amount involved (ptr for return)
//
void
transaction::extract(unsigned long *ctrans_type, 
		class date *cdate, 
		double *cnumber,
		double *cprice,	
		double *ctotal)
{
	if (ctrans_type)
		*ctotal=trans_type;
	if (cdate)
		*cdate=date;
	if (cnumber)
		*cnumber=number;
	if (cprice)
		*cprice=price;
	if (ctotal)
		*ctotal=total;
}


//
// transaction::Security
//
// Returns the security that the transaction applies to
//
// INPUTS :	None
// RETURNS:	Returns a pointer to the security that the transaction
//		applies to.
//
class security *
transaction::Security()
{
	return( sec);
}



//
// sectrans_list::extract sdate edate number_ptr spent_ptr recd_ptr avgdate_ptr
// 
// Extracts transaction information for a complete list of transactions
// 
//
// INPUTS :	sdate		Start of period
//		edate		End of period
// RETURNS:	number_ptr	Amount of shares held
//		spent_ptr	Amount spend on shares in period
// 		recd_ptr	Amount recived from sales of shares in period
//		avgdate_ptr	Average date of purchases
//
void 
sectrans_list::extract(
			class date *s_date,
			class date *e_date,
			double *number,
			double *spent,
			double *recd,
			class date *avgdate
			)
{
	class transaction *trans;
	class transaction *head_trans;
	class date datesum;
	class date cavgdate;
	int numtrans;
	double cnumber;
	double cspent;
	double crecd;

	cnumber=0.0;
	cspent=0.0;
	crecd=0.0;
	numtrans=0;

	head_trans=dual_head();
	trans=head_trans;

	while (trans!=NULL) {		// for each transaction do..
		if ( trans->date.inrange(s_date,e_date) ) {
			datesum.add(trans->date.daysince1970());
			numtrans++;
			switch (trans->trans_type) {
			case TRANS_BUYRIGHTS:
			case TRANS_BUY:
			case TRANS_EXRIGHTS:
			case TRANS_GETRIGHTS:
			case TRANS_BONUS:cnumber += trans->number;
				cspent+= trans->total; //bonus set this to 0
				break;
			case TRANS_USERIGHTS:
			case TRANS_SELRIGHTS:
			case TRANS_SELL:
				cnumber -= trans->number;
				crecd += trans->total;
				break;
			case TRANS_SPLIT:
					cnumber += trans->number;
					break;
			default: // should check that this only for rights issues
				;// silly compiler wants this ; - I ask you...
			}
		}
		trans=trans->dual_next();
		if (trans==head_trans)
			break;
	}
	if (numtrans!=0)
		cavgdate.daysince1970(datesum.daysince1970()/numtrans);

	if (number)
		*number=cnumber;
	if (spent)
		*spent=cspent;
	if (recd)
		*recd=crecd;
	if (avgdate)
		*avgdate=cavgdate;

}





//
// transaction::trans_load rec 
//
// Reads transactions from the transactions file, thus generating
// a portfolio.  Security information must be loaded prior to this.
// 
// INPUTS :	rec		A pointer to the input record
// RETURNS:	None
//
void
transaction::trans_load( 
	class record *rec
	)
{
	char tstr[256];
	char *string;
	char oldv;
	int iter;
	class Holding *holding;

	//
	// Look for old versions of the file
	//
	oldv=!(strcmp(rec->version(),"VERSION:ODT11:1") );

	rec->compile("[|:\n]");

	if ( (string=rec->uptomatch()) == NULL)
		goto error;

	holding=portfolio->HoldinglistFind(string);
	if (holding==NULL) {
		sprintf(tstr,"Found transactions for a non-existent security %s",string);
		DialogPanic(tstr);
	}
	sec=holding->HoldingSecurity();

	if ( (string=rec->uptomatch()) == NULL)
		goto error;

	//
	// If VERSION:ODT11:1 then split off the TRANS_ at the start of
	// the type string
	//
	if (oldv) {
		static char *input[] = {
		      "TRANS_BUY","TRANS_SELL","TRANS_SPLIT","TRANS_BUYRIGHTS",
		      "TRANS_SELRIGHTS", "TRANS_EXRIGHTS", "TRANS_USERIGHTS",
		      "TRANS_GETRIGHTS", "TRANS_BONUS", "TRANS_VOID", NULL };
		static char *output[]= {
		      "BUY","SELL","SPLIT","BUYR",
		      "SELR", "EXR", "USER",
		      "GETR", "BONUS", "VOID", NULL };

		for(iter=0;;iter++) {
			if (input[iter]==NULL)
				break;
			if (!strcmp(input[iter],string)) {
				string=output[iter];
				break;
			}
		}
	}

	trans_type=string2type(string);

	if ( (string=rec->uptomatch()) == NULL) {
		fprintf(stderr,"Failed to read date\n");
		goto error;
	}

	if (date.setdate(string)== -1 ) {
		fprintf(stderr,"Failed to convert date(%s)\n",string);
		goto error;
	}

	if ( (string=rec->uptomatch()) == NULL) {
		fprintf(stderr,"Failed to read number\n");
		goto error;
	}
	number=atof(string);

	if ( (string=rec->uptomatch()) == NULL) {
		fprintf(stderr,"Failed to read price\n");
		goto error;
	}
	price=atof(string);

	if ( (string=rec->uptomatch()) == NULL) {
		fprintf(stderr,"Failed to read total\n");
		goto error;
	}
	total=atof(string);

	//
	// Add it to the list of all transactions
	//
        main_trans_list->insert_tail(this);

	//
	// Add it to the list of transactions for the specific security
	//
	holding->HoldingTransactionadd(this);

	//
	// Issue the callback
	//
	main_trans_list->add_vect.CallbackMake(0,NULL);
	
	return;
error:
 	DialogPanic("Failed during transactions file read.");
}


//
// trans_load varargs
//
// Loads transactions.  The varargs argument is the record pointer
// that contains the information for the transaction.  This function
// acts as a wrapper from a C callback to C++.
//
void
trans_load(char *,va_list ap)
{
	class record *rec;
	transaction *newtrans=new transaction;

        rec=va_arg(ap,class record *);
	newtrans->trans_load(rec);
}


//
// transaction::save_record record
// 
// Saves the transactions to the appropriate portfolio file.  Note that
// all the transactions are written to the file, it is not a case
// of simply writing out the changes.  At the moment this a bit dubious
// since we don't take the precaution of backing up the old file -- more
// to be done..
// 
// INPUTS:	record		The output record to be created.
// RETURNS:	None
//
void
transaction::save_record(class record *rec)
{
	char string[100];
	char *secname;

	sec->SecurityExtract(&secname,NULL,NULL,NULL,NULL);
	rec->create(secname);
	rec->append("|");
	rec->append(type2string());
	sprintf(string,"|%s|",date.getdate());
	rec->append(string);
	sprintf(string,"%f|",number);
	rec->append(string);
	sprintf(string,"%f|",price);
	rec->append(string);
	sprintf(string,"%f|",total);
	rec->append(string);
}



//
// transaction::type2string
//
// Returns an ascii string describing the type of transaction
//
// INPUTS :	None
// RETURNS:	String containing name of transaction type
//
char *
transaction::type2string()
{
int     iter;

	for(iter=0;;iter++)
		if ( (trans_types[iter].value==NULL) ||
		     (trans_types[iter].value==trans_type) )
			break;

	if ( trans_types[iter].value==NULL ) 
		DialogPanic("Illegal transaction type");

	return(trans_types[iter].name);
}

//
// string2type name
//
// Takes an ascii string describing a type of transaction and
// returns the equivalent integer value.
//
// INPUTS :	name		The name of the transaction type
// RETURNS:	Integer representation of that name.
//
int
transaction::string2type(char *name)
{
int iter;

	for(iter=0;;iter++) {
		if (trans_types[iter].value==NULL) 
			break;
		if (!strcmp(name,trans_types[iter].name))
			break;
	}

	if ( trans_types[iter].value==NULL ) 
		DialogPanic("Unsupported transaction type");

	return (trans_types[iter].value);
}


//
// trans2string mode mask secname
//
// Returns a complete string representaton of a transaction.
//
// INPUTS :	mode		NAMEFIRST or DATEFIRST to describe which
//				information shall be displayed first
//		mask 		Mask containing transaction types (TRANS_XX)
//				that are to be displayed.
//		secname		Name of security
//
char *
transaction::trans2string(int mode,
	unsigned long transmask,
	char *secname
	)
{
#define _L_MODE_FULL	1	// display it all
#define _L_MODE_NOPR	2	// no prices/ totals to be shown

	char *trans_str;	// build up area
	char str[256];		// build up space, hope this is big enough
	char *datebuf;
	char sharename[16];	// share name, if you change this you must	
				// change the sprintfs above and below!
	int dmode;		// takes values given by _L_MODE_XXX
				// selects type of output to be written

	if (!(transmask & trans_type))
		return(NULL);	// user does not want this trans type

	datebuf=date.getdate();

	switch (trans_type) {	// deal with each transaction type
	case TRANS_BUY:dmode=_L_MODE_FULL;
		break;
	case TRANS_SELL:dmode=_L_MODE_FULL;
		break;
	case TRANS_SPLIT:dmode=_L_MODE_NOPR;
		break;
	case TRANS_BUYRIGHTS:dmode=_L_MODE_FULL;
		break;
	case TRANS_SELRIGHTS:dmode=_L_MODE_FULL;
		break;
	case TRANS_EXRIGHTS: dmode=_L_MODE_FULL;
		break;
	case TRANS_USERIGHTS: dmode=_L_MODE_NOPR;
		break;
	case TRANS_GETRIGHTS: dmode=_L_MODE_NOPR;
		break;
	case TRANS_BONUS: dmode=_L_MODE_NOPR;
		break;
	default:
		DialogPanic("transaction::trans2string - Bad trans_type");
	}

	trans_str=type2string();
	strncpy(sharename,secname,15);
	sharename[15]='\0';	// 16th character must terminate the string
	switch (dmode) {
	case _L_MODE_FULL:	// give all information 
		if (mode==NAMEFIRST)
			sprintf(str,"%s\t%s\t%s\t%.4f\t%.3f\t%.3f",
			sharename,trans_str,datebuf,number,price,
			total);
		else if (mode==DATEFIRST)
			sprintf(str,"%s\t%s\t%s\t%.4f\t%.3f\t%.3f",
			datebuf,sharename,trans_str,number,price,
			total);
		else {
			perror("portfolio/append:Bad mode select\n");
			exit(1);
			}
		break;
	case _L_MODE_NOPR:	// no cost and total values 
		if (mode==NAMEFIRST)
			sprintf(str,"%s\t%s\t%s\t%.4f\t%s\t%s",
			sharename,trans_str,datebuf,number,"n/a","n/a");
		else if (mode==DATEFIRST)
			sprintf(str,"%s\t%s\t%s\t%.4f\t%s\t%s",
			datebuf,sharename,trans_str,number,"n/a","n/a");
		else {
			perror("portfolio/append:Bad mode select\n");
			exit(1);
		}
		break;
	default:
		DialogPanic("transaction::trans2string - Bad dmode");
	}
	return(str);		// something has been added
}



//
// Constructor function for the transaction class
//
transaction::transaction()
{
	number=0.0;
	trans_type=TRANS_VOID;
}

//
//////////////////////////////////////////////////////////////////////////////
//
// ALL THE REMAINING FUNCTIONS ARE GLOBAL TO THE PROGRAM 
//
//////////////////////////////////////////////////////////////////////////////
//



//
// buyorsell title ttype
//
// Builds buy/sell dialogs for all transaction types.
//
// INPUTS :	title		The name of dialog
//	        ttype		The type of the transaction
// RETURNS:	None
//
void
buyorsell(
	char *title,
	int trans_type
	)
{
	Transactionedit *newtrans;
	dcursortype_e cursor;

	cursor=DialogCursor( DCURSOR_WAIT );

	newtrans=new Transactionedit;
	newtrans->userentry(title,trans_type);

	DialogCursor( cursor );
}

//
// purchase_share 
//
// Displays a dialog to allow the user to enter information about 
// purchasing a share.
// 
// INPUTS :	None
// RETURNS:	None
//
static void 
purchase_share(local_va_dcl)
{
	va_alist=va_alist;	// stop compiler warnings
	buyorsell("Buy Stock",TRANS_BUY);
}




//
// sell_share
//
// Displays a dialog to allow the user to enter information about 
// selling a share.
//
// INPUTS :	None
// RETURNS:	The newly created dialog
//
static void
sell_share(local_va_dcl)
{
	va_alist=va_alist;	// stop compiler warnings
	buyorsell("Sell Stock",TRANS_SELL);
}



//
// purchase_rights parent
//
// Displays a dialog to allow the user to enter information about 
// purchasing some rights.
//
// INPUTS :	None
// RETURNS:	The newly created dialog
//
static void
purchase_rights(local_va_dcl)
{
	va_alist=va_alist;	// stop compiler warnings
	buyorsell("Purchase Rights",TRANS_BUYRIGHTS);
}



//
// sell_rights
//
// Displays a dialog to allow the user to enter information about 
// selling some rights.
//
// INPUTS :	None
// RETURNS:	The newly created dialog
//
static void
sell_rights(local_va_dcl)
{
	va_alist=va_alist;	// stop compiler warnings
	buyorsell("Sell Rights",TRANS_SELRIGHTS);
}



//
// receive_rights 
//
// Displays a dialog to allow the user to enter information about 
// receiving some rights.
//
// INPUTS :	None
// RETURNS:	The newly created dialog
//
static void
receive_rights(local_va_dcl)
{
	va_alist=va_alist;	// stop compiler warnings
	buyorsell("Receive Rights",TRANS_GETRIGHTS);
}




//
// exercise_rights parent
//
// Displays a dialog to allow the user to enter information about 
// exercising rights.
//
// INPUTS :	None
// RETURNS:	The newly created dialog
//
static void
exercise_rights(local_va_dcl)
{
	va_alist=va_alist;	// stop compiler warnings
	buyorsell("Exercise Rights",TRANS_EXRIGHTS);
}





//
// share_bonus parent
//
// Displays a dialog to allow the user to enter information about a 
// share bonus.
//
// INPUTS :	None
// RETURNS:	The newly created dialog
//
//
static void
share_bonus(local_va_dcl)
{
	va_alist=va_alist;	// stop compiler warnings
	buyorsell("Receive Bonus Stocks",TRANS_BONUS);
}



//
// share_split parent
//
// Displays a dialog to allow the user to enter information about a 
// share split.
//
// INPUTS :	None
// RETURNS:	The newly created dialog
//
static void
share_split(local_va_dcl)
{
	va_alist=va_alist;	// stop compiler warnings
	buyorsell("Split Stocks",TRANS_SPLIT);
}


//
// number datestamp
//
// Convenience function that returns the number of securities
// held at a given datestamp.  
//
// INPUTS:	datestamp		The date at which the number of
//					securities is to apply
// RETURNS:	Number of securities 
//
double
sectrans_list::number(class date *dateval)
{
	double number;

	extract(0,dateval,&number,NULL,NULL,NULL);
	return(number);
}


//
// Constructor function for the secondary transaction list
//
sectrans_list::sectrans_list()
{
}



//
// Constructor function for the primary transaction list
//
trans_list::trans_list()
{
	set_filename("transactions",WRITE_BY_LINE);
	//
        // Register our decoder for building the list of
        // transaction
        //
        add_load_vect(trans_load);
}

//
// These are to prevent the need for casts
//
class transaction *
sectrans_list::dual_head()
{
	return( LISTREMAP(transaction,dslist_head,dual_head,this) );
}

class transaction *
transaction::dual_next()
{
	return( LISTREMAP(transaction,dslist_node,dual_next,this) );
}


//
// trans_init parent menu_bar
//
// Initialises the portfolio module:
//	- Sets up the menu for the main menu bar
// 
// INPUTS :	d_parent	The parent dialog
//		menu_bar	A handle (dialog) on the main menu bar
// RETURNS:	None
//
void
trans_init(menu_bar *mb)
{
	MENU("Transactions","Purchase",purchase_share,0);
	MENU("Transactions","Sell",sell_share,0);
	MENU("Transactions","Receive Bonus",share_bonus,0);
	MENU("Transactions","Split",share_split,0);
	MENU("Transactions","Receive rights",receive_rights,0);
	MENU("Transactions","Exercise rights",exercise_rights,0);
	MENU("Transactions","Purchase rights",purchase_rights,0);
	MENU("Transactions","Sell rights",sell_rights,0);

}
