
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: trans.h,v 2.4 1995/05/17 21:26:35 davidsi Exp $
//
// $Log: trans.h,v $
// Revision 2.4  1995/05/17  21:26:35  davidsi
// - Added TRANS_ALL transaction type, which is used to specify all types.
// - Renamed create_transaction() to TransactionCreate().
//
// Revision 2.3  1995/04/16  20:25:25  davidsi
// - Added Security() function member that extracts a security from a
//   transaction. Rarely useful, but now there is a need for it (see
//   dtrans.C).
//
// Revision 2.2  1995/03/08  21:19:22  davidsi
// - Factored the transaction class into two classes, the transaction class
//   provides all the main transaction handling functionality and the
//   Transactionedit class allows transactions to be edited.  This reduces
//   the amount of memory req'd to represent a transaction.
// - Made main_trans_list externally visible.
//
// Revision 2.1  1995/03/01  22:24:46  davidsi
// - Removed all dialog references - use the Dialog class instead.
//
// Revision 2.0  1995/01/19  21:08:06  davidsi
// - Modified all time based declarations to use the new date class.
// - Added callback for when a transaction is added.
//
// Revision 1.1  1994/08/06  16:36:00  davidsi
// Initial revision
//
//
#ifndef __TRANS_H
#define __TRANS_H

#include <stdio.h>
#include <string.h>

#include "security.h"
#include "dslist.h"
#include "record.h"
#include "date.h"
#include "dialog.h"
#include "optionl.h"
#include "button.h"


//
// The secondary list transaction class - used to contain transactions for
// a single securities
//
class sectrans_list : public dslist_head {
	double total_number;		// total number of securities held
public:
	double number(class date *);
	void extract(class date *,class date *, double *, 
			double *, double *, class date *);

	//
	// These are to prevent needing casts
	//
        class transaction *dual_head();

	sectrans_list();
};


//
// The list transaction class - used to contain transactions for all
// securities
//
class trans_list : public dslist {
public:
	class callback add_vect;	// callbacks for when a transaction
					// is added.
	trans_list();
};

//
// The Transactionedit class
//
class Transactionedit : private Button, private Optionlist
{
	unsigned long trans_type;

	Dialogshell d_shell; 	        // top most dialog 
	Dialogvertical d_vert;		// main vertical dialog
	Dialoghorizontal d_horiz;	// main horizontal dialog
	Dialogentry d_entrylist;	// dialog for data entry
	Dialoghorizontal d_buthoriz;	// horizontal for buttons
	Optionlistinstance *d_secname;	// selection dialog for name of security
	Optionlistinstance *d_rsecname;	// selection dialog for name of rights

private:
	void addw_figures();
	Optionlistinstance *addw_sec_name(int);
	void addw_buttons();
	void addw_prices();
	//
	// Override callback in button class
	//
	virtual void ButtonCallback(unsigned long id,buttonstate_e flag);
public:
	void userentry( char *title, int trans_type);

	Transactionedit();
	virtual ~Transactionedit();
};



//
// The transaction class
//
class transaction : public dslist_node
{
	unsigned long trans_type;
	class date date;		// date of transaction 
	double number;			// number of securities in transaction 
	double price;			// price received/spent per securities 
	double total;			// total spent/ recevied 
	class security *sec;		// security to which this applies

private:
	char *type2string();
	int string2type(char *);
	//
	// Friends
	//
	friend class sectrans_list;
	//
	// Public functions
	// 
public:
	int TransactionCreate(char *,unsigned long,class date *,
				double,double,double);
	void extract( unsigned long *trans_type, class date *date, 
				double *number, double *price, double *total);
	char *trans2string(int mode, unsigned long transmask, char *secname);
	class security *Security();


	//
	// Inheiritated from the slist class
	// 
        virtual void save_record(class record *);
	void trans_load( class record *rec );

	//
	// These are to prevent needing casts
	//
        class transaction *dual_next();

	transaction();
};


//
// Different possible transaction types
//
#define TRANS_VOID      0	// no transaction
#define TRANS_BUY       1       // securities wre purchased
#define TRANS_SELL      2       // securities were sold
#define TRANS_SPLIT     4	// securities were split
#define TRANS_BONUS     8	// bonus number of securities
#define TRANS_GETRIGHTS 16	// rights were given free
#define TRANS_SELRIGHTS 32	// rights were sold
#define TRANS_BUYRIGHTS 64	// rights were purchased
#define TRANS_EXRIGHTS  128     // securities obtained by exercising rights 
#define TRANS_USERIGHTS 256     // rights were exercised
#define TRANS_ALL (TRANS_BUY|TRANS_SELL|TRANS_SPLIT|			\
		   TRANS_BONUS| TRANS_GETRIGHTS|TRANS_SELRIGHTS|	\
		   TRANS_BUYRIGHTS| TRANS_EXRIGHTS| TRANS_USERIGHTS)



//
// Modes available with trans2string
//
#define NAMEFIRST 1
#define DATEFIRST 2

//
// Externally available variables
//
extern trans_list *main_trans_list;

#endif
