
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: ttext.C,v 2.3 1995/06/24 15:01:04 davidsi Exp $
//
// $Log: ttext.C,v $
// Revision 2.3  1995/06/24  15:01:04  davidsi
// - Added GNU copyright message.
//
// Revision 2.2  1995/04/16  19:56:53  davidsi
// - BUG FIX. The read_pagelist() function must write a header line,
//   otherwise the first price in the file will be read (and lost)
//   as if it was a header.
//
// Revision 2.1  1995/02/27  22:18:55  davidsi
// - Replaced all calls to dialog_message() to use the new DialogMessage()
//   function.
// - Now dialog free.
//
// Revision 2.0  1995/01/19  21:05:47  davidsi
// - Removed time.h include.
//
// Revision 1.2  1994/11/05  21:09:51  davidsi
// - Various modifications to allow this file to compile using g++.
//
// Revision 1.1  1994/10/18  19:58:50  davidsi
// Initial revision
//
//
//
//

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <dial.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <ctype.h>


#include "localmacros.h"

#include "ttext.h"
#include "dialog.h"



//
// read_page file name
//
// Reads a teletext page, extracting the specified rows and columns.
// 
// INPUTS :	file	File handle for output file
//		name	Name of file to read from
// RETURNS:	None
//
void
ttext::read_page( FILE *file,char *name)
{
#define _L_BUFFERSIZE	256		// size of work buffer - price server
					// should not wrt a string out larger 
					// than this 
	FILE *ptr;			// pointer to file 
	char buffer[_L_BUFFERSIZE];	// input area 
	char copybuffer[_L_BUFFERSIZE];	// secondary input area
	int current_line;
	int i;
	int processing;

	if ( (ptr=fopen(name,"r"))==NULL) {
		perror("bad");
		sprintf(buffer,"Cannot open file %s",name);
		DialogMessage(buffer);
		return;
	}
	processing=0;
	current_line=1;
	for(;;) {
		if (current_line==max_line)
			break;
		if (current_line==start_line)
			processing=1;
		if (current_line>stop_line)
			processing=0;

		if (fgets(buffer,_L_BUFFERSIZE,ptr)==NULL) {
			sprintf(buffer,"File %s is too short (%d lines)",name,current_line);
			DialogMessage(buffer);
			fclose(ptr);
			return;
		};

		current_line++;
		
		if (!processing)
			continue;

		for(i=0;i<max_columns;i++) {
			strncpy(copybuffer,&(buffer[start_row[i]]),stop_row[i]-start_row[i]);
			copybuffer[stop_row[i]-start_row[i]]='\0';
			fprintf(file,"%s\n",copybuffer);
		}
	}
	fclose(ptr);
}

//
// read_pagelist binary_name name
// 
// INPUTS :	binary_name	Name of executable that will fetch
//				teletext pages and write the filenames
//				that they have been saved in on stdout
//		name		Name of file to write the processed data to
// RETURNS:	None
//
void
ttext::read_pagelist(char *binary_name,char *oname)
{
	char buffer[_L_BUFFERSIZE];	// input area 
	FILE *iptr;			// pointer to file 
	FILE *optr;			// pointer to file 

	if ((iptr=popen(binary_name,"r"))==NULL) {
		DialogMessage("Failed to start page loader)");
		return;
	}
	if ((optr=fopen(oname,"w"))==NULL) {
		DialogMessage("Failed to open output file");
		return;
	}
	fprintf(optr,"TELETEXT UNIX\n");
	for(;;) {
		if (fgets(buffer,_L_BUFFERSIZE,iptr)==NULL) 
			break;
		buffer[strlen(buffer)-1]='\0';
		read_page(optr,buffer);
	}
	fclose(optr);
	pclose(iptr);
}

//
// set maxline maxcol startl stopl startr[] stopr[] 
//
// Defines the area on a page that is to be loaded
//
// INPUTS :	maxline		The maximum line to load to
//		maxcol		The maximum number of columns
//		startl		The line where processing starts
//		stopl		The line where processing stops
//		startr[]	The starting columns
//		stoprp[]	The stop columns
// RETURNS:	None
//
void
ttext::set(int c_max_line,
		int c_max_columns,
		int c_start_line,
		int c_stop_line,
		int *c_start_row,
		int *c_stop_row 
	)
{
	max_line=c_max_line;
	max_columns=c_max_columns;
	start_line=c_start_line;
	stop_line=c_stop_line;
	memcpy(start_row,c_start_row,sizeof(start_row));
	memcpy(stop_row,c_stop_row,sizeof(stop_row));
}


//
// Constructor function
//
ttext::ttext()
{
}


//
// Destructor function
//
ttext::~ttext()
{
}
