.ad l
.nh
.TH CONVERT 1 "10 October 1992" "ImageMagick"
.SH NAME
convert - converts an input file using one image format to an output
file with a differing image format.
.SH SYNOPSIS
.B "convert"
[ \fIoptions\fP ... ] \fIfile\fP \fIfile\fP
.SH DESCRIPTION
.PP
.I Convert
converts an input file using one image format to an output file with a
differing image format. By default, the image format is determined by
its magic number. To specify a particular image format, precede the 
filename with an image format name and a colon (i.e.  mtv:image) or specify 
the image type as the filename suffix (i.e. image.mtv).  Specify \fIfile\fP as 
\fI-\fP for standard input or output.  If \fIfile\fP has the extension 
\fB.Z\fP, the file is decoded with \fIuncompress\fP.

\fIConvert\fP recognizes the following image formats:

.TP 6
.B Tag
\fBDescription\fP
.PP
-------------------------------------------------------------
.TP 6
.B AVS
Advanced Visualization System image file format.
.TP 6
.B CMYK	
Raw cyan, magenta, yellow, and black bytes.
.TP 6
.B FAX
Group 3.
.TP 6
.B GIF
Graphic Image Format.
.TP 6
.B GRAY
Raw gray bytes.
.TP 6
.B JPEG
.TP 6
.B MIFF
Machine Independant file format.
.TP 6
.B MTV
.TP 6
.B PNM
Portable bitmap.
.TP 6
.B PS
Postscript.
.TP 6
.B RGB
Raw red, green, and blue bytes.
.TP 6
.B RLE
Utah Raster Toolkit; read only.
.TP 6
.B SUN
SUN raster.
.TP 6
.B TEXT
raw text file; read only.
.TP 6
.B TIFF
Tagged Image File Format.
.TP 6
.B VICAR
read only.
.TP 6
.B YUV
Raw Y, U, and V bytes.  U and V, normally -0.5 through 0.5, are normalized
to the range 0 through 255 to fit within a byte.
.TP 6
.B X
select image from X server screen; read only.
.TP 6
.B XC
constant image of X server background color.
.TP 6
.B XBM
X11 bitmap.
.TP 6
.B XWD
X11 window dump.
.PP
.SH EXAMPLES
.PP
To convert a \fIMIFF\fP image of a cockatoo to a SUN raster image, use:
.PP
     convert cockatoo.miff sun:cockatoo.ras
.PP
To convert a \fIMTV\fP image of a molecule to a JPEG image, use:
.PP
     convert molecule.mtv molecule.jpeg
.PP
To convert a raw \fIGRAY\fP image to a portable graymap, use:
.PP
     convert -geometry 768x512 gray:raw image.pnm
.SH OPTIONS
.TP 5
.B "-alpha"
store alpha channel if the image has one.
.TP 5
.B "-colors \fIvalue\fP"
preferred number of colors in the image.

The actual number of colors in the image may be less than your request,
but never more.  Note, this is a color reduction option.  Images with
less unique colors than specified with this option will remain unchanged.
Refer to \fBQuantize(9)\fP for more details.

Note, options \fB-dither\fP, \fB-colorspace\fP, and \fB-treedepth\fP affect
the color reduction algorithm.
.TP 5
.B "-colorspace \fIvalue\fP"
the type of colorspace: \fIGRAY\fP, \fIRGB\fP, \fIXYZ\fP, \fIYIQ\fP, or
\fIYUV\fP.

Color reduction, by default, takes place in the RGB color space.
Empirical evidence suggests that distances in color spaces such as YUV
or YIQ correspond to perceptual color differences more closely
than do distances in RGB space.  These color spaces may give better
results when color reducing an image.  Refer to \fBQuantize(9)\fP for
more details.

The \fB-colors\fP or \fB-monochrome\fP option is required for this option
to take effect.
.TP 5
.B "-compress \fItype\fP"
the type of image compression: \fIQEncoded\fP or \fIRunlengthEncoded\fP.

Specify \fB\+compress\fP to store the binary image in an uncompressed format.
The default is the compression type of the specified image file.
.TP 5
.B "-density \fI<width>x<height>
vertical and horizonal density of the image.

This option specifies an image density whose interpretation changes
with the type of image.  The default is 72 dots per inch in the
horizonal and vertical direction for Postscript.  Text files default to
80 characters in width and 60 lines in height.  Use this option to
alter the default density.
.TP 5
.B "-display \fIhost:display[.screen]\fP"
specifies the X server to contact; see \fBX(1)\fP.
.TP 5
.B "-dither"
apply Floyd/Steinberg error diffusion to the image.

The basic strategy of dithering is to trade intensity resolution for
spatial resolution by averaging the intensities of several neighboring
pixels.  Images which suffer from severe contouring when reducing colors
can be improved with this option.

The \fB-colors\fP option is required for dithering to take effect.
.TP 5
.B "-geometry \fI<width>x<height>\fP"
the width and height of the image.  

Use this option to specified the width and height of raw images whose
dimensions are unknown such as GRAY, RGB, and CMYK.  This option can
also change the default 8.5 by 11 width and height of the Postscript
canvas.
.TP 5
.B "-quality \fIvalue\fP"
JPEG quality setting.  Quality is 0 (worst) to 100 (best). The default is
75.
.TP 5
.B "-scene \fIvalue\fP"
image scene number.
.TP 5
.B "-treedepth \fIvalue\fP"
Normally, this integer value is zero or one.  A zero or one tells
\fIDisplay\fP to choose a optimal tree depth for the color reduction
algorithm.

An optimal depth generally allows the best representation of the source
image with the fastest computational speed and the least amount of
memory.  However, the default depth is inappropriate for some images.
To assure the best representation, try values between 2 and 8 for this
parameter.  Refer to \fBQuantize(9)\fP for more details.

The \fB-colors\fP option is required for this option to take effect.
.TP 5
.B -verbose
print detailed information about the image.

This information is printed: image scene number;  image name;  converted
image name;  image size;  the image class (\fIDirectClass\fP or 
\fIPseudoClass\fP);  the total number of unique colors;  and the number
of seconds to read and convert the image.
.PP
Change '-' to '+' in any option above to reverse its effect.  For
example, specify +alpha to store the image without its alpha channel.
.PP
To specify a particular image format prefix \fIfile\fP with the image
type and a colon (i.e. mtv:image) or specify the image type as the
filename suffix (i.e. image.mtv).  See the beginning of this document
for a list of valid image formats.  If \fIfile\fP has the extension
\fB.Z\fP, the file size is reduced using Lempel-Ziv coding with
\fBcompress\fP.  If \fIfile\fP already exists, you will be prompted as
to whether it should be overwritten.

When you specify \fBX\fP as your image type, the filename has special
meaning.  It specifies an X window by id, name, or \fBroot\fP.  If no
filename is specified, the window is selected by clicking the mouse in
the desired window.
.SH ENVIRONMENT
.PP
.TP 5
.B DISPLAY
To get the default host, display number, and screen.
.SH SEE ALSO
display(1), animate(1), mogrify(1), Quantize(9), X(1), MIFF(5)
.SH COPYRIGHT
Copyright 1992 E. I. du Pont de Nemours & Company
.PP
Permission to use, copy, modify, distribute, and sell this software and
its documentation for any purpose is hereby granted without fee,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of E. I. du Pont de Nemours
& Company not be used in advertising or publicity pertaining to
distribution of the software without specific, written prior
permission.  E. I. du Pont de Nemours & Company makes no representations
about the suitability of this software for any purpose.  It is provided
"as is" without express or implied warranty.
.PP
E. I. du Pont de Nemours & Company disclaims all warranties with regard
to this software, including all implied warranties of merchantability
and fitness, in no event shall E. I. du Pont de Nemours & Company be
liable for any special, indirect or consequential damages or any
damages whatsoever resulting from loss of use, data or profits, whether
in an action of contract, negligence or other tortious action, arising
out of or in connection with the use or performance of this software.
.SH AUTHORS
John Cristy, E.I. du Pont De Nemours & Company Incorporated
